import 'dart:convert';

/// Hanzi ranked by the frequency of their use in chinese

class HanziUtils {
  static List<Map<String, Object>> get data {
    return _hanziRankFrequency;
  }

  static List<HanziRankInfo> findHanziRankBySimplifiedPinyin(
      String simplifiedPinyin) {
    return _findByValue(simplifiedPinyin, 'sp');
  }

  static List<HanziRankInfo> findHanziRankByPinyin(String pinyin) {
    return _findByValue(pinyin, 'p');
  }

  static List<HanziRankInfo> findHanziRankByHanzi(String hanzi) {
    return _findByValue(hanzi, 'h');
  }

  static List<HanziRankInfo> _findByValue(String value, String keyName) {
    final data = _hanziRankFrequency
        .where((h) => h[keyName] == value)
        .map(
          (e) => HanziRankInfo.fromMap(e),
        )
        .toList();
    data.sort((a, b) => a.fequencyRank.compareTo(b.fequencyRank));
    return data;
  }
}

class HanziRankInfo {
  String hanzi;
  String pinyin;
  String simplifiedPinyin;
  String radical;
  int strokeCount;
  int hskLevel;

  /// the smaller the rank, the most frequent the hanzi
  /// in Chinese language based on
  /// Jun Da's Modern Chinese Character Frequency List
  int fequencyRank;
  HanziRankInfo({
    required this.hanzi,
    required this.pinyin,
    required this.simplifiedPinyin,
    required this.radical,
    required this.strokeCount,
    required this.fequencyRank,
    required this.hskLevel,
  });

  factory HanziRankInfo.fromMap(Map data) {
    return HanziRankInfo(
      hanzi: data['h'],
      pinyin: data['p'],
      simplifiedPinyin: data['sp'],
      radical: data['r'],
      strokeCount: data['sc'],
      fequencyRank: data['f'],
      hskLevel: data['l'],
    );
  }

  @override
  String toString() {
    return jsonEncode(toMap());
  }

  Map<String, dynamic> toMap() {
    return {
      'hanzi': hanzi,
      'pinyin': pinyin,
      'simplifiedPinyin': simplifiedPinyin,
      'radical': radical,
      'strokeCount': strokeCount,
      'fequencyRank': fequencyRank,
      'hskLevel': hskLevel,
    };
  }
}

const _hanziRankFrequency = [
  {"h": "的", "p": "de", "sp": "de", "r": "白", "sc": 8, "l": 1, "f": 1},
  {"h": "一", "p": "yī", "sp": "yi", "r": "一", "sc": 1, "l": 1, "f": 2},
  {"h": "是", "p": "shì", "sp": "shi", "r": "日", "sc": 9, "l": 1, "f": 3},
  {"h": "不", "p": "bù", "sp": "bu", "r": "一", "sc": 4, "l": 1, "f": 4},
  {"h": "了", "p": "le", "sp": "le", "r": "亅", "sc": 2, "l": 1, "f": 5},
  {"h": "在", "p": "zài", "sp": "zai", "r": "土", "sc": 6, "l": 1, "f": 6},
  {"h": "人", "p": "rén", "sp": "ren", "r": "人", "sc": 2, "l": 1, "f": 7},
  {"h": "有", "p": "yǒu", "sp": "you", "r": "月", "sc": 6, "l": 1, "f": 8},
  {"h": "我", "p": "wǒ", "sp": "wo", "r": "戈", "sc": 7, "l": 1, "f": 9},
  {"h": "他", "p": "tā", "sp": "ta", "r": "人", "sc": 5, "l": 1, "f": 10},
  {"h": "这", "p": "zhè", "sp": "zhe", "r": "辵", "sc": 7, "l": 1, "f": 11},
  {"h": "个", "p": "gè", "sp": "ge", "r": "丨", "sc": 3, "l": 1, "f": 12},
  {"h": "们", "p": "men", "sp": "men", "r": "人", "sc": 5, "l": 1, "f": 13},
  {"h": "中", "p": "zhōng", "sp": "zhong", "r": "丨", "sc": 4, "l": 1, "f": 14},
  {"h": "来", "p": "lái", "sp": "lai", "r": "木", "sc": 7, "l": 1, "f": 15},
  {"h": "上", "p": "shàng", "sp": "shang", "r": "一", "sc": 3, "l": 1, "f": 16},
  {"h": "大", "p": "dà", "sp": "da", "r": "大", "sc": 3, "l": 1, "f": 17},
  {"h": "为", "p": "wèi", "sp": "wei", "r": "丶", "sc": 4, "l": 2, "f": 18},
  {"h": "和", "p": "hé", "sp": "he", "r": "口", "sc": 8, "l": 1, "f": 19},
  {"h": "国", "p": "guó", "sp": "guo", "r": "囗", "sc": 8, "l": 1, "f": 20},
  {"h": "地", "p": "de", "sp": "de", "r": "土", "sc": 6, "l": 1, "f": 21},
  {"h": "地", "p": "di", "sp": "di", "r": "土", "sc": 6, "l": 1, "f": 21},
  {"h": "到", "p": "dào", "sp": "dao", "r": "刀", "sc": 8, "l": 1, "f": 22},
  {"h": "以", "p": "yǐ", "sp": "yi", "r": "人", "sc": 4, "l": 1, "f": 23},
  {"h": "说", "p": "shuō", "sp": "shuo", "r": "言", "sc": 9, "l": 1, "f": 24},
  {"h": "时", "p": "shí", "sp": "shi", "r": "日", "sc": 7, "l": 1, "f": 25},
  {"h": "要", "p": "yào", "sp": "yao", "r": "襾", "sc": 9, "l": 1, "f": 26},
  {"h": "就", "p": "jiù", "sp": "jiu", "r": "尢", "sc": 12, "l": 1, "f": 27},
  {"h": "出", "p": "chū", "sp": "chu", "r": "凵", "sc": 5, "l": 1, "f": 28},
  {"h": "会", "p": "huì", "sp": "hui", "r": "人", "sc": 6, "l": 1, "f": 29},
  {"h": "可", "p": "kě", "sp": "ke", "r": "口", "sc": 5, "l": 1, "f": 30},
  {"h": "也", "p": "yě", "sp": "ye", "r": "乙", "sc": 3, "l": 1, "f": 31},
  {"h": "你", "p": "nǐ", "sp": "ni", "r": "人", "sc": 7, "l": 1, "f": 32},
  {"h": "对", "p": "duì", "sp": "dui", "r": "寸", "sc": 5, "l": 1, "f": 33},
  {"h": "生", "p": "shēng", "sp": "sheng", "r": "生", "sc": 5, "l": 1, "f": 34},
  {"h": "能", "p": "néng", "sp": "neng", "r": "肉", "sc": 10, "l": 1, "f": 35},
  {"h": "而", "p": "ér", "sp": "er", "r": "而", "sc": 6, "l": 2, "f": 36},
  {"h": "子", "p": "zi", "sp": "zi", "r": "子", "sc": 3, "l": 1, "f": 37},
  {"h": "那", "p": "nà", "sp": "na", "r": "邑", "sc": 6, "l": 1, "f": 38},
  {"h": "得", "p": "dé", "sp": "de", "r": "彳", "sc": 11, "l": 1, "f": 39},
  {"h": "于", "p": "yú", "sp": "yu", "r": "二", "sc": 3, "l": 2, "f": 40},
  {"h": "着", "p": "zhe", "sp": "zhe", "r": "目", "sc": 11, "l": 1, "f": 41},
  {"h": "下", "p": "xià", "sp": "xia", "r": "一", "sc": 3, "l": 1, "f": 42},
  {"h": "自", "p": "zì", "sp": "zi", "r": "自", "sc": 6, "l": 2, "f": 43},
  {"h": "之", "p": "zhī", "sp": "zhi", "r": "丿", "sc": 3, "l": 3, "f": 44},
  {"h": "年", "p": "nián", "sp": "nian", "r": "干", "sc": 6, "l": 1, "f": 45},
  {"h": "过", "p": "guò", "sp": "guo", "r": "辵", "sc": 6, "l": 1, "f": 46},
  {"h": "发", "p": "fā", "sp": "fa", "r": "又", "sc": 5, "l": 2, "f": 47},
  {"h": "后", "p": "hòu", "sp": "hou", "r": "口", "sc": 6, "l": 1, "f": 48},
  {"h": "作", "p": "zuò", "sp": "zuo", "r": "人", "sc": 7, "l": 1, "f": 49},
  {"h": "里", "p": "lǐ", "sp": "li", "r": "里", "sc": 7, "l": 1, "f": 50},
  {"h": "用", "p": "yòng", "sp": "yong", "r": "用", "sc": 5, "l": 1, "f": 51},
  {"h": "道", "p": "dào", "sp": "dao", "r": "辵", "sc": 12, "l": 1, "f": 52},
  {"h": "行", "p": "xíng", "sp": "xing", "r": "行", "sc": 6, "l": 2, "f": 53},
  {"h": "所", "p": "suǒ", "sp": "suo", "r": "戶", "sc": 8, "l": 2, "f": 54},
  {"h": "然", "p": "rán", "sp": "ran", "r": "火", "sc": 12, "l": 2, "f": 55},
  {"h": "家", "p": "jiā", "sp": "jia", "r": "宀", "sc": 10, "l": 1, "f": 56},
  {"h": "种", "p": "zhǒng", "sp": "zhong", "r": "禾", "sc": 9, "l": 3, "f": 57},
  {"h": "事", "p": "shì", "sp": "shi", "r": "亅", "sc": 8, "l": 1, "f": 58},
  {"h": "成", "p": "chéng", "sp": "cheng", "r": "戈", "sc": 6, "l": 2, "f": 59},
  {"h": "方", "p": "fāng", "sp": "fang", "r": "方", "sc": 4, "l": 1, "f": 60},
  {"h": "多", "p": "duō", "sp": "duo", "r": "夕", "sc": 6, "l": 1, "f": 61},
  {"h": "经", "p": "jīng", "sp": "jing", "r": "糸", "sc": 8, "l": 2, "f": 62},
  {"h": "么", "p": "me", "sp": "me", "r": "丿", "sc": 3, "l": 1, "f": 63},
  {"h": "去", "p": "qù", "sp": "qu", "r": "厶", "sc": 5, "l": 1, "f": 64},
  {"h": "法", "p": "fǎ", "sp": "fa", "r": "水", "sc": 8, "l": 2, "f": 65},
  {"h": "学", "p": "xué", "sp": "xue", "r": "子", "sc": 8, "l": 1, "f": 66},
  {"h": "如", "p": "rú", "sp": "ru", "r": "女", "sc": 6, "l": 2, "f": 67},
  {"h": "都", "p": "dōu", "sp": "dou", "r": "邑", "sc": 10, "l": 1, "f": 68},
  {"h": "同", "p": "tóng", "sp": "tong", "r": "口", "sc": 6, "l": 1, "f": 69},
  {"h": "现", "p": "xiàn", "sp": "xian", "r": "玉", "sc": 8, "l": 1, "f": 70},
  {"h": "当", "p": "dāng", "sp": "dang", "r": "小", "sc": 6, "l": 2, "f": 71},
  {"h": "没", "p": "méi", "sp": "mei", "r": "水", "sc": 7, "l": 1, "f": 72},
  {"h": "动", "p": "dòng", "sp": "dong", "r": "力", "sc": 6, "l": 1, "f": 73},
  {"h": "面", "p": "miàn", "sp": "mian", "r": "面", "sc": 9, "l": 1, "f": 74},
  {"h": "起", "p": "qǐ", "sp": "qi", "r": "走", "sc": 10, "l": 1, "f": 75},
  {"h": "看", "p": "kàn", "sp": "kan", "r": "目", "sc": 9, "l": 1, "f": 76},
  {"h": "定", "p": "dìng", "sp": "ding", "r": "宀", "sc": 8, "l": 2, "f": 77},
  {"h": "天", "p": "tiān", "sp": "tian", "r": "大", "sc": 4, "l": 1, "f": 78},
  {"h": "分", "p": "fēn", "sp": "fen", "r": "刀", "sc": 4, "l": 1, "f": 79},
  {"h": "还", "p": "hái", "sp": "hai", "r": "辵", "sc": 7, "l": 1, "f": 80},
  {"h": "进", "p": "jìn", "sp": "jin", "r": "辵", "sc": 7, "l": 1, "f": 81},
  {"h": "好", "p": "hǎo", "sp": "hao", "r": "女", "sc": 6, "l": 1, "f": 82},
  {"h": "小", "p": "xiǎo", "sp": "xiao", "r": "小", "sc": 3, "l": 1, "f": 83},
  {"h": "部", "p": "bù", "sp": "bu", "r": "邑", "sc": 10, "l": 2, "f": 84},
  {"h": "其", "p": "qí", "sp": "qi", "r": "八", "sc": 8, "l": 2, "f": 85},
  {"h": "些", "p": "xiē", "sp": "xie", "r": "二", "sc": 8, "l": 1, "f": 86},
  {"h": "主", "p": "zhǔ", "sp": "zhu", "r": "丶", "sc": 5, "l": 2, "f": 87},
  {"h": "样", "p": "yàng", "sp": "yang", "r": "木", "sc": 10, "l": 1, "f": 88},
  {"h": "理", "p": "lǐ", "sp": "li", "r": "玉", "sc": 11, "l": 2, "f": 89},
  {"h": "心", "p": "xīn", "sp": "xin", "r": "心", "sc": 4, "l": 2, "f": 90},
  {"h": "她", "p": "tā", "sp": "ta", "r": "女", "sc": 6, "l": 1, "f": 91},
  {"h": "本", "p": "běn", "sp": "ben", "r": "木", "sc": 5, "l": 1, "f": 92},
  {"h": "前", "p": "qián", "sp": "qian", "r": "刀", "sc": 9, "l": 1, "f": 93},
  {"h": "开", "p": "kāi", "sp": "kai", "r": "廾", "sc": 4, "l": 1, "f": 94},
  {"h": "但", "p": "dàn", "sp": "dan", "r": "人", "sc": 7, "l": 2, "f": 95},
  {"h": "因", "p": "yīn", "sp": "yin", "r": "囗", "sc": 6, "l": 2, "f": 96},
  {"h": "只", "p": "zhǐ", "sp": "zhi", "r": "口", "sc": 5, "l": 1, "f": 97},
  {"h": "从", "p": "cóng", "sp": "cong", "r": "人", "sc": 4, "l": 1, "f": 98},
  {"h": "想", "p": "xiǎng", "sp": "xiang", "r": "心", "sc": 13, "l": 1, "f": 99},
  {"h": "实", "p": "shí", "sp": "shi", "r": "宀", "sc": 8, "l": 2, "f": 100},
  {"h": "日", "p": "rì", "sp": "ri", "r": "日", "sc": 4, "l": 1, "f": 101},
  {"h": "军", "p": "jūn", "sp": "jun", "r": "車", "sc": 6, "l": 5, "f": 102},
  {"h": "者", "p": "zhě", "sp": "zhe", "r": "老", "sc": 8, "l": 2, "f": 103},
  {"h": "意", "p": "yì", "sp": "yi", "r": "心", "sc": 13, "l": 1, "f": 104},
  {"h": "无", "p": "wú", "sp": "wu", "r": "无", "sc": 4, "l": 4, "f": 105},
  {"h": "力", "p": "lì", "sp": "li", "r": "力", "sc": 2, "l": 2, "f": 106},
  {"h": "它", "p": "tā", "sp": "ta", "r": "宀", "sc": 5, "l": 2, "f": 107},
  {"h": "与", "p": "yǔ", "sp": "yu", "r": "一", "sc": 3, "l": 4, "f": 108},
  {"h": "长", "p": "zhǎng", "sp": "zhang", "r": "長", "sc": 4, "l": 2, "f": 109},
  {"h": "把", "p": "bǎ", "sp": "ba", "r": "手", "sc": 7, "l": 1, "f": 110},
  {"h": "机", "p": "jī", "sp": "ji", "r": "木", "sc": 6, "l": 1, "f": 111},
  {"h": "十", "p": "shí", "sp": "shi", "r": "十", "sc": 2, "l": 1, "f": 112},
  {"h": "民", "p": "mín", "sp": "min", "r": "氏", "sc": 5, "l": 3, "f": 113},
  {"h": "第", "p": "dì", "sp": "di", "r": "竹", "sc": 11, "l": 1, "f": 114},
  {"h": "公", "p": "gōng", "sp": "gong", "r": "八", "sc": 4, "l": 1, "f": 115},
  {"h": "此", "p": "cǐ", "sp": "ci", "r": "止", "sc": 6, "l": 3, "f": 116},
  {"h": "已", "p": "yǐ", "sp": "yi", "r": "己", "sc": 3, "l": 2, "f": 117},
  {"h": "工", "p": "gōng", "sp": "gong", "r": "工", "sc": 3, "l": 1, "f": 118},
  {"h": "使", "p": "shǐ", "sp": "shi", "r": "人", "sc": 8, "l": 2, "f": 119},
  {"h": "情", "p": "qíng", "sp": "qing", "r": "心", "sc": 11, "l": 2, "f": 120},
  {"h": "明", "p": "míng", "sp": "ming", "r": "日", "sc": 8, "l": 1, "f": 121},
  {"h": "性", "p": "xìng", "sp": "xing", "r": "心", "sc": 8, "l": 3, "f": 122},
  {"h": "知", "p": "zhī", "sp": "zhi", "r": "矢", "sc": 8, "l": 1, "f": 123},
  {"h": "全", "p": "quán", "sp": "quan", "r": "入", "sc": 6, "l": 2, "f": 124},
  {"h": "三", "p": "sān", "sp": "san", "r": "一", "sc": 3, "l": 1, "f": 125},
  {"h": "又", "p": "yòu", "sp": "you", "r": "又", "sc": 2, "l": 2, "f": 126},
  {"h": "关", "p": "guān", "sp": "guan", "r": "八", "sc": 6, "l": 1, "f": 127},
  {"h": "点", "p": "diǎn", "sp": "dian", "r": "火", "sc": 9, "l": 1, "f": 128},
  {"h": "正", "p": "zhèng", "sp": "zheng", "r": "止", "sc": 5, "l": 1, "f": 129},
  {"h": "业", "p": "yè", "sp": "ye", "r": "一", "sc": 5, "l": 2, "f": 130},
  {"h": "外", "p": "wài", "sp": "wai", "r": "夕", "sc": 5, "l": 1, "f": 131},
  {"h": "将", "p": "jiāng", "sp": "jiang", "r": "爿", "sc": 9, "l": 3, "f": 132},
  {"h": "两", "p": "liǎng", "sp": "liang", "r": "一", "sc": 7, "l": 1, "f": 133},
  {"h": "高", "p": "gāo", "sp": "gao", "r": "高", "sc": 10, "l": 1, "f": 134},
  {"h": "间", "p": "jiān", "sp": "jian", "r": "門", "sc": 7, "l": 1, "f": 135},
  {"h": "由", "p": "yóu", "sp": "you", "r": "田", "sc": 5, "l": 2, "f": 136},
  {"h": "问", "p": "wèn", "sp": "wen", "r": "門", "sc": 6, "l": 1, "f": 137},
  {"h": "很", "p": "hěn", "sp": "hen", "r": "彳", "sc": 9, "l": 1, "f": 138},
  {"h": "最", "p": "zuì", "sp": "zui", "r": "曰", "sc": 12, "l": 1, "f": 139},
  {"h": "重", "p": "zhòng", "sp": "zhong", "r": "里", "sc": 9, "l": 1, "f": 140},
  {"h": "并", "p": "bìng", "sp": "bing", "r": "干", "sc": 6, "l": 3, "f": 141},
  {"h": "物", "p": "wù", "sp": "wu", "r": "牛", "sc": 8, "l": 2, "f": 142},
  {"h": "手", "p": "shǒu", "sp": "shou", "r": "手", "sc": 4, "l": 1, "f": 143},
  {"h": "应", "p": "yīng", "sp": "ying", "r": "广", "sc": 7, "l": 2, "f": 144},
  {"h": "战", "p": "zhàn", "sp": "zhan", "r": "戈", "sc": 9, "l": 4, "f": 145},
  {"h": "向", "p": "xiàng", "sp": "xiang", "r": "口", "sc": 6, "l": 2, "f": 146},
  {"h": "头", "p": "tóu", "sp": "tou", "r": "大", "sc": 5, "l": 2, "f": 147},
  {"h": "文", "p": "wén", "sp": "wen", "r": "文", "sc": 4, "l": 1, "f": 148},
  {"h": "体", "p": "tǐ", "sp": "ti", "r": "人", "sc": 7, "l": 1, "f": 149},
  {"h": "政", "p": "zhèng", "sp": "zheng", "r": "攴", "sc": 9, "l": 4, "f": 150},
  {"h": "美", "p": "měi", "sp": "mei", "r": "羊", "sc": 9, "l": 2, "f": 151},
  {"h": "相", "p": "xiāng", "sp": "xiang", "r": "目", "sc": 9, "l": 2, "f": 152},
  {"h": "见", "p": "jiàn", "sp": "jian", "r": "見", "sc": 4, "l": 1, "f": 153},
  {"h": "被", "p": "bèi", "sp": "bei", "r": "衣", "sc": 10, "l": 3, "f": 154},
  {"h": "利", "p": "lì", "sp": "li", "r": "刀", "sc": 7, "l": 2, "f": 155},
  {"h": "什", "p": "shén", "sp": "shen", "r": "人", "sc": 4, "l": 1, "f": 156},
  {"h": "二", "p": "èr", "sp": "er", "r": "二", "sc": 2, "l": 1, "f": 157},
  {"h": "等", "p": "děng", "sp": "deng", "r": "竹", "sc": 12, "l": 1, "f": 158},
  {"h": "产", "p": "chǎn", "sp": "chan", "r": "立", "sc": 6, "l": 3, "f": 159},
  {"h": "或", "p": "huò", "sp": "huo", "r": "戈", "sc": 8, "l": 2, "f": 160},
  {"h": "新", "p": "xīn", "sp": "xin", "r": "斤", "sc": 13, "l": 1, "f": 161},
  {"h": "己", "p": "jǐ", "sp": "ji", "r": "己", "sc": 3, "l": 2, "f": 162},
  {"h": "制", "p": "zhì", "sp": "zhi", "r": "刀", "sc": 8, "l": 3, "f": 163},
  {"h": "身", "p": "shēn", "sp": "shen", "r": "身", "sc": 7, "l": 1, "f": 164},
  {"h": "果", "p": "guǒ", "sp": "guo", "r": "木", "sc": 8, "l": 1, "f": 165},
  {"h": "加", "p": "jiā", "sp": "jia", "r": "力", "sc": 5, "l": 2, "f": 166},
  {"h": "西", "p": "xī", "sp": "xi", "r": "襾", "sc": 6, "l": 1, "f": 167},
  {"h": "斯", "p": "sī", "sp": "si", "r": "斤", "sc": 12, "l": 4, "f": 168},
  {"h": "月", "p": "yuè", "sp": "yue", "r": "月", "sc": 4, "l": 1, "f": 169},
  {"h": "话", "p": "huà", "sp": "hua", "r": "言", "sc": 8, "l": 1, "f": 170},
  {"h": "合", "p": "hé", "sp": "he", "r": "口", "sc": 6, "l": 2, "f": 171},
  {"h": "回", "p": "huí", "sp": "hui", "r": "囗", "sc": 6, "l": 1, "f": 172},
  {"h": "特", "p": "tè", "sp": "te", "r": "牛", "sc": 10, "l": 2, "f": 173},
  {"h": "代", "p": "dài", "sp": "dai", "r": "人", "sc": 5, "l": 3, "f": 174},
  {"h": "内", "p": "nèi", "sp": "nei", "r": "入", "sc": 4, "l": 3, "f": 175},
  {"h": "信", "p": "xìn", "sp": "xin", "r": "人", "sc": 9, "l": 2, "f": 176},
  {"h": "表", "p": "biǎo", "sp": "biao", "r": "衣", "sc": 8, "l": 2, "f": 177},
  {"h": "化", "p": "huà", "sp": "hua", "r": "匕", "sc": 4, "l": 3, "f": 178},
  {"h": "老", "p": "lǎo", "sp": "lao", "r": "老", "sc": 6, "l": 1, "f": 179},
  {"h": "给", "p": "gěi", "sp": "gei", "r": "糸", "sc": 9, "l": 1, "f": 180},
  {"h": "世", "p": "shì", "sp": "shi", "r": "一", "sc": 5, "l": 3, "f": 181},
  {"h": "位", "p": "wèi", "sp": "wei", "r": "人", "sc": 7, "l": 2, "f": 182},
  {"h": "次", "p": "cì", "sp": "ci", "r": "欠", "sc": 6, "l": 1, "f": 183},
  {"h": "度", "p": "dù", "sp": "du", "r": "广", "sc": 9, "l": 2, "f": 184},
  {"h": "门", "p": "mén", "sp": "men", "r": "門", "sc": 3, "l": 1, "f": 185},
  {"h": "任", "p": "rèn", "sp": "ren", "r": "人", "sc": 6, "l": 3, "f": 186},
  {"h": "常", "p": "cháng", "sp": "chang", "r": "巾", "sc": 11, "l": 1, "f": 187},
  {"h": "先", "p": "xiān", "sp": "xian", "r": "儿", "sc": 6, "l": 1, "f": 188},
  {"h": "海", "p": "hǎi", "sp": "hai", "r": "水", "sc": 10, "l": 2, "f": 189},
  {"h": "通", "p": "tōng", "sp": "tong", "r": "辵", "sc": 10, "l": 2, "f": 190},
  {"h": "教", "p": "jiào", "sp": "jiao", "r": "攴", "sc": 11, "l": 1, "f": 191},
  {"h": "儿", "p": "ér", "sp": "er", "r": "儿", "sc": 2, "l": 1, "f": 192},
  {"h": "原", "p": "yuán", "sp": "yuan", "r": "厂", "sc": 10, "l": 2, "f": 193},
  {"h": "东", "p": "dōng", "sp": "dong", "r": "一", "sc": 5, "l": 1, "f": 194},
  {"h": "声", "p": "shēng", "sp": "sheng", "r": "士", "sc": 7, "l": 2, "f": 195},
  {"h": "提", "p": "tí", "sp": "ti", "r": "手", "sc": 12, "l": 2, "f": 196},
  {"h": "立", "p": "lì", "sp": "li", "r": "立", "sc": 5, "l": 3, "f": 197},
  {"h": "及", "p": "jí", "sp": "ji", "r": "又", "sc": 3, "l": 3, "f": 198},
  {"h": "比", "p": "bǐ", "sp": "bi", "r": "比", "sc": 4, "l": 1, "f": 199},
  {"h": "员", "p": "yuán", "sp": "yuan", "r": "口", "sc": 7, "l": 2, "f": 200},
  {"h": "解", "p": "jiě", "sp": "jie", "r": "角", "sc": 13, "l": 3, "f": 201},
  {"h": "水", "p": "shuǐ", "sp": "shui", "r": "水", "sc": 4, "l": 1, "f": 202},
  {"h": "名", "p": "míng", "sp": "ming", "r": "口", "sc": 6, "l": 1, "f": 203},
  {"h": "真", "p": "zhēn", "sp": "zhen", "r": "目", "sc": 10, "l": 1, "f": 204},
  {"h": "论", "p": "lùn", "sp": "lun", "r": "言", "sc": 6, "l": 2, "f": 205},
  {"h": "处", "p": "chù", "sp": "chu", "r": "夂", "sc": 5, "l": 2, "f": 206},
  {"h": "走", "p": "zǒu", "sp": "zou", "r": "走", "sc": 7, "l": 1, "f": 207},
  {"h": "义", "p": "yì", "sp": "yi", "r": "丿", "sc": 3, "l": 3, "f": 208},
  {"h": "各", "p": "gè", "sp": "ge", "r": "口", "sc": 6, "l": 3, "f": 209},
  {"h": "入", "p": "rù", "sp": "ru", "r": "入", "sc": 2, "l": 2, "f": 210},
  {"h": "几", "p": "jǐ", "sp": "ji", "r": "几", "sc": 2, "l": 1, "f": 211},
  {"h": "口", "p": "kǒu", "sp": "kou", "r": "口", "sc": 3, "l": 1, "f": 212},
  {"h": "认", "p": "rèn", "sp": "ren", "r": "言", "sc": 4, "l": 1, "f": 213},
  {"h": "条", "p": "tiáo", "sp": "tiao", "r": "木", "sc": 7, "l": 1, "f": 214},
  {"h": "平", "p": "píng", "sp": "ping", "r": "干", "sc": 5, "l": 1, "f": 215},
  {"h": "系", "p": "xì", "sp": "xi", "r": "糸", "sc": 7, "l": 1, "f": 216},
  {"h": "气", "p": "qì", "sp": "qi", "r": "气", "sc": 4, "l": 1, "f": 217},
  {"h": "题", "p": "tí", "sp": "ti", "r": "頁", "sc": 15, "l": 2, "f": 218},
  {"h": "活", "p": "huó", "sp": "huo", "r": "水", "sc": 9, "l": 2, "f": 219},
  {"h": "尔", "p": "ěr", "sp": "er", "r": "小", "sc": 5, "l": 5, "f": 220},
  {"h": "更", "p": "gèng", "sp": "geng", "r": "曰", "sc": 7, "l": 2, "f": 221},
  {"h": "别", "p": "bié", "sp": "bie", "r": "刀", "sc": 7, "l": 1, "f": 222},
  {"h": "打", "p": "dǎ", "sp": "da", "r": "手", "sc": 5, "l": 1, "f": 223},
  {"h": "女", "p": "nǚ", "sp": "nu", "r": "女", "sc": 3, "l": 1, "f": 224},
  {"h": "变", "p": "biàn", "sp": "bian", "r": "又", "sc": 8, "l": 2, "f": 225},
  {"h": "四", "p": "sì", "sp": "si", "r": "囗", "sc": 5, "l": 1, "f": 226},
  {"h": "神", "p": "shén", "sp": "shen", "r": "示", "sc": 9, "l": 3, "f": 227},
  {"h": "总", "p": "zǒng", "sp": "zong", "r": "八", "sc": 9, "l": 3, "f": 228},
  {"h": "何", "p": "hé", "sp": "he", "r": "人", "sc": 7, "l": 3, "f": 229},
  {"h": "电", "p": "diàn", "sp": "dian", "r": "曰", "sc": 5, "l": 1, "f": 230},
  {"h": "数", "p": "shù", "sp": "shu", "r": "攴", "sc": 13, "l": 2, "f": 231},
  {"h": "安", "p": "ān", "sp": "an", "r": "宀", "sc": 6, "l": 2, "f": 232},
  {"h": "少", "p": "shǎo", "sp": "shao", "r": "小", "sc": 4, "l": 1, "f": 233},
  {"h": "报", "p": "bào", "sp": "bao", "r": "手", "sc": 7, "l": 2, "f": 234},
  {"h": "才", "p": "cái", "sp": "cai", "r": "手", "sc": 3, "l": 2, "f": 235},
  {"h": "结", "p": "jié", "sp": "jie", "r": "糸", "sc": 9, "l": 2, "f": 236},
  {"h": "反", "p": "fǎn", "sp": "fan", "r": "又", "sc": 4, "l": 3, "f": 237},
  {"h": "受", "p": "shòu", "sp": "shou", "r": "又", "sc": 8, "l": 2, "f": 238},
  {"h": "目", "p": "mù", "sp": "mu", "r": "目", "sc": 5, "l": 2, "f": 239},
  {"h": "太", "p": "tài", "sp": "tai", "r": "大", "sc": 4, "l": 1, "f": 240},
  {"h": "量", "p": "liàng", "sp": "liang", "r": "里", "sc": 12, "l": 2, "f": 241},
  {"h": "再", "p": "zài", "sp": "zai", "r": "冂", "sc": 6, "l": 1, "f": 242},
  {"h": "感", "p": "gǎn", "sp": "gan", "r": "心", "sc": 13, "l": 2, "f": 243},
  {"h": "建", "p": "jiàn", "sp": "jian", "r": "廴", "sc": 8, "l": 3, "f": 244},
  {"h": "务", "p": "wù", "sp": "wu", "r": "夂", "sc": 5, "l": 2, "f": 245},
  {"h": "做", "p": "zuò", "sp": "zuo", "r": "人", "sc": 11, "l": 1, "f": 246},
  {"h": "接", "p": "jiē", "sp": "jie", "r": "手", "sc": 11, "l": 2, "f": 247},
  {"h": "必", "p": "bì", "sp": "bi", "r": "心", "sc": 5, "l": 2, "f": 248},
  {"h": "场", "p": "chǎng", "sp": "chang", "r": "土", "sc": 6, "l": 1, "f": 249},
  {"h": "件", "p": "jiàn", "sp": "jian", "r": "人", "sc": 6, "l": 2, "f": 250},
  {"h": "计", "p": "jì", "sp": "ji", "r": "言", "sc": 4, "l": 2, "f": 251},
  {"h": "管", "p": "guǎn", "sp": "guan", "r": "竹", "sc": 14, "l": 3, "f": 252},
  {"h": "期", "p": "qī", "sp": "qi", "r": "月", "sc": 12, "l": 1, "f": 253},
  {"h": "市", "p": "shì", "sp": "shi", "r": "巾", "sc": 5, "l": 2, "f": 254},
  {"h": "直", "p": "zhí", "sp": "zhi", "r": "目", "sc": 8, "l": 2, "f": 255},
  {"h": "德", "p": "dé", "sp": "de", "r": "彳", "sc": 15, "l": 3, "f": 256},
  {"h": "资", "p": "zī", "sp": "zi", "r": "貝", "sc": 10, "l": 3, "f": 257},
  {"h": "命", "p": "mìng", "sp": "ming", "r": "口", "sc": 8, "l": 3, "f": 258},
  {"h": "山", "p": "shān", "sp": "shan", "r": "山", "sc": 3, "l": 1, "f": 259},
  {"h": "金", "p": "jīn", "sp": "jin", "r": "金", "sc": 8, "l": 3, "f": 260},
  {"h": "指", "p": "zhǐ", "sp": "zhi", "r": "手", "sc": 9, "l": 3, "f": 261},
  {"h": "克", "p": "kè", "sp": "ke", "r": "儿", "sc": 7, "l": 2, "f": 262},
  {"h": "许", "p": "xǔ", "sp": "xu", "r": "言", "sc": 6, "l": 2, "f": 263},
  {"h": "统", "p": "tǒng", "sp": "tong", "r": "糸", "sc": 9, "l": 4, "f": 264},
  {"h": "区", "p": "qū", "sp": "qu", "r": "匸", "sc": 4, "l": 3, "f": 265},
  {"h": "保", "p": "bǎo", "sp": "bao", "r": "人", "sc": 9, "l": 3, "f": 266},
  {"h": "至", "p": "zhì", "sp": "zhi", "r": "至", "sc": 6, "l": 3, "f": 267},
  {"h": "队", "p": "duì", "sp": "dui", "r": "阜", "sc": 4, "l": 2, "f": 268},
  {"h": "形", "p": "xíng", "sp": "xing", "r": "彡", "sc": 7, "l": 3, "f": 269},
  {"h": "社", "p": "shè", "sp": "she", "r": "示", "sc": 7, "l": 3, "f": 270},
  {"h": "便", "p": "biàn", "sp": "bian", "r": "人", "sc": 9, "l": 2, "f": 271},
  {"h": "空", "p": "kōng", "sp": "kong", "r": "穴", "sc": 8, "l": 2, "f": 272},
  {"h": "决", "p": "jué", "sp": "jue", "r": "冫", "sc": 6, "l": 3, "f": 273},
  {"h": "治", "p": "zhì", "sp": "zhi", "r": "水", "sc": 8, "l": 4, "f": 274},
  {"h": "展", "p": "zhǎn", "sp": "zhan", "r": "尸", "sc": 10, "l": 3, "f": 275},
  {"h": "马", "p": "mǎ", "sp": "ma", "r": "馬", "sc": 3, "l": 1, "f": 276},
  {"h": "科", "p": "kē", "sp": "ke", "r": "禾", "sc": 9, "l": 2, "f": 277},
  {"h": "司", "p": "sī", "sp": "si", "r": "口", "sc": 5, "l": 2, "f": 278},
  {"h": "五", "p": "wǔ", "sp": "wu", "r": "二", "sc": 4, "l": 1, "f": 279},
  {"h": "基", "p": "jī", "sp": "ji", "r": "土", "sc": 11, "l": 3, "f": 280},
  {"h": "眼", "p": "yǎn", "sp": "yan", "r": "目", "sc": 11, "l": 2, "f": 281},
  {"h": "书", "p": "shū", "sp": "shu", "r": "丨", "sc": 4, "l": 1, "f": 282},
  {"h": "非", "p": "fēi", "sp": "fei", "r": "非", "sc": 8, "l": 1, "f": 283},
  {"h": "则", "p": "zé", "sp": "ze", "r": "刀", "sc": 6, "l": 4, "f": 284},
  {"h": "听", "p": "tīng", "sp": "ting", "r": "口", "sc": 7, "l": 1, "f": 285},
  {"h": "白", "p": "bái", "sp": "bai", "r": "白", "sc": 5, "l": 1, "f": 286},
  {"h": "却", "p": "què", "sp": "que", "r": "卩", "sc": 7, "l": 4, "f": 287},
  {"h": "界", "p": "jiè", "sp": "jie", "r": "田", "sc": 9, "l": 3, "f": 288},
  {"h": "达", "p": "dá", "sp": "da", "r": "辵", "sc": 6, "l": 3, "f": 289},
  {"h": "光", "p": "guāng", "sp": "guang", "r": "儿", "sc": 6, "l": 1, "f": 290},
  {"h": "放", "p": "fàng", "sp": "fang", "r": "攴", "sc": 8, "l": 1, "f": 291},
  {"h": "强", "p": "qiáng", "sp": "qiang", "r": "弓", "sc": 12, "l": 3, "f": 292},
  {"h": "即", "p": "jí", "sp": "ji", "r": "卩", "sc": 7, "l": 4, "f": 293},
  {"h": "像", "p": "xiàng", "sp": "xiang", "r": "人", "sc": 13, "l": 2, "f": 294},
  {"h": "难", "p": "nán", "sp": "nan", "r": "又", "sc": 10, "l": 1, "f": 295},
  {"h": "且", "p": "qiě", "sp": "qie", "r": "一", "sc": 5, "l": 2, "f": 296},
  {"h": "权", "p": "quán", "sp": "quan", "r": "木", "sc": 6, "l": 4, "f": 297},
  {"h": "思", "p": "sī", "sp": "si", "r": "心", "sc": 9, "l": 2, "f": 298},
  {"h": "王", "p": "wáng", "sp": "wang", "r": "玉", "sc": 4, "l": 1, "f": 299},
  {"h": "象", "p": "xiàng", "sp": "xiang", "r": "豕", "sc": 11, "l": 3, "f": 300},
  {"h": "完", "p": "wán", "sp": "wan", "r": "宀", "sc": 7, "l": 2, "f": 301},
  {"h": "设", "p": "shè", "sp": "she", "r": "言", "sc": 6, "l": 3, "f": 302},
  {"h": "式", "p": "shì", "sp": "shi", "r": "弋", "sc": 6, "l": 3, "f": 303},
  {"h": "色", "p": "sè", "sp": "se", "r": "色", "sc": 6, "l": 2, "f": 304},
  {"h": "路", "p": "lù", "sp": "lu", "r": "足", "sc": 13, "l": 1, "f": 305},
  {"h": "记", "p": "jì", "sp": "ji", "r": "言", "sc": 5, "l": 1, "f": 306},
  {"h": "南", "p": "nán", "sp": "nan", "r": "十", "sc": 9, "l": 1, "f": 307},
  {"h": "品", "p": "pǐn", "sp": "pin", "r": "口", "sc": 9, "l": 3, "f": 308},
  {"h": "住", "p": "zhù", "sp": "zhu", "r": "人", "sc": 7, "l": 1, "f": 309},
  {"h": "告", "p": "gào", "sp": "gao", "r": "口", "sc": 7, "l": 1, "f": 310},
  {"h": "类", "p": "lèi", "sp": "lei", "r": "大", "sc": 9, "l": 3, "f": 311},
  {"h": "求", "p": "qiú", "sp": "qiu", "r": "水", "sc": 7, "l": 2, "f": 312},
  {"h": "据", "p": "jù", "sp": "ju", "r": "手", "sc": 11, "l": 3, "f": 313},
  {"h": "程", "p": "chéng", "sp": "cheng", "r": "禾", "sc": 12, "l": 3, "f": 314},
  {"h": "北", "p": "běi", "sp": "bei", "r": "匕", "sc": 5, "l": 1, "f": 315},
  {"h": "边", "p": "biān", "sp": "bian", "r": "辵", "sc": 5, "l": 1, "f": 316},
  {"h": "死", "p": "sǐ", "sp": "si", "r": "歹", "sc": 6, "l": 3, "f": 317},
  {"h": "张", "p": "zhāng", "sp": "zhang", "r": "弓", "sc": 7, "l": 2, "f": 318},
  {"h": "该", "p": "gāi", "sp": "gai", "r": "言", "sc": 8, "l": 2, "f": 319},
  {"h": "交", "p": "jiāo", "sp": "jiao", "r": "亠", "sc": 6, "l": 1, "f": 320},
  {"h": "规", "p": "guī", "sp": "gui", "r": "見", "sc": 8, "l": 3, "f": 321},
  {"h": "万", "p": "wàn", "sp": "wan", "r": "一", "sc": 3, "l": 2, "f": 322},
  {"h": "取", "p": "qǔ", "sp": "qu", "r": "又", "sc": 8, "l": 2, "f": 323},
  {"h": "拉", "p": "lā", "sp": "la", "r": "手", "sc": 8, "l": 2, "f": 324},
  {"h": "格", "p": "gé", "sp": "ge", "r": "木", "sc": 10, "l": 3, "f": 325},
  {"h": "望", "p": "wàng", "sp": "wang", "r": "月", "sc": 11, "l": 2, "f": 326},
  {"h": "觉", "p": "jué", "sp": "jue", "r": "見", "sc": 9, "l": 1, "f": 327},
  {"h": "术", "p": "shù", "sp": "shu", "r": "木", "sc": 5, "l": 3, "f": 328},
  {"h": "领", "p": "lǐng", "sp": "ling", "r": "頁", "sc": 11, "l": 3, "f": 329},
  {"h": "共", "p": "gòng", "sp": "gong", "r": "八", "sc": 6, "l": 2, "f": 330},
  {"h": "确", "p": "què", "sp": "que", "r": "石", "sc": 12, "l": 2, "f": 331},
  {"h": "传", "p": "chuán", "sp": "chuan", "r": "人", "sc": 6, "l": 3, "f": 332},
  {"h": "师", "p": "shī", "sp": "shi", "r": "巾", "sc": 6, "l": 1, "f": 333},
  {"h": "观", "p": "guān", "sp": "guan", "r": "見", "sc": 6, "l": 2, "f": 334},
  {"h": "清", "p": "qīng", "sp": "qing", "r": "水", "sc": 11, "l": 2, "f": 335},
  {"h": "今", "p": "jīn", "sp": "jin", "r": "人", "sc": 4, "l": 1, "f": 336},
  {"h": "切", "p": "qiè", "sp": "qie", "r": "刀", "sc": 4, "l": 3, "f": 337},
  {"h": "院", "p": "yuàn", "sp": "yuan", "r": "阜", "sc": 9, "l": 1, "f": 338},
  {"h": "让", "p": "ràng", "sp": "rang", "r": "言", "sc": 5, "l": 2, "f": 339},
  {"h": "识", "p": "shí", "sp": "shi", "r": "言", "sc": 7, "l": 1, "f": 340},
  {"h": "候", "p": "hòu", "sp": "hou", "r": "人", "sc": 10, "l": 1, "f": 341},
  {"h": "带", "p": "dài", "sp": "dai", "r": "巾", "sc": 9, "l": 2, "f": 342},
  {"h": "导", "p": "dǎo", "sp": "dao", "r": "寸", "sc": 6, "l": 3, "f": 343},
  {"h": "争", "p": "zhēng", "sp": "zheng", "r": "亅", "sc": 6, "l": 3, "f": 344},
  {"h": "运", "p": "yùn", "sp": "yun", "r": "辵", "sc": 7, "l": 2, "f": 345},
  {"h": "笑", "p": "xiào", "sp": "xiao", "r": "竹", "sc": 10, "l": 1, "f": 346},
  {"h": "飞", "p": "fēi", "sp": "fei", "r": "乙", "sc": 3, "l": 1, "f": 347},
  {"h": "风", "p": "fēng", "sp": "feng", "r": "風", "sc": 4, "l": 1, "f": 348},
  {"h": "步", "p": "bù", "sp": "bu", "r": "止", "sc": 7, "l": 2, "f": 349},
  {"h": "改", "p": "gǎi", "sp": "gai", "r": "攴", "sc": 7, "l": 2, "f": 350},
  {"h": "收", "p": "shōu", "sp": "shou", "r": "攴", "sc": 6, "l": 2, "f": 351},
  {"h": "根", "p": "gēn", "sp": "gen", "r": "木", "sc": 10, "l": 3, "f": 352},
  {"h": "干", "p": "gàn", "sp": "gan", "r": "干", "sc": 3, "l": 1, "f": 353},
  {"h": "造", "p": "zào", "sp": "zao", "r": "辵", "sc": 10, "l": 3, "f": 354},
  {"h": "言", "p": "yán", "sp": "yan", "r": "言", "sc": 7, "l": 2, "f": 355},
  {"h": "联", "p": "lián", "sp": "lian", "r": "耳", "sc": 12, "l": 3, "f": 356},
  {"h": "持", "p": "chí", "sp": "chi", "r": "手", "sc": 9, "l": 3, "f": 357},
  {"h": "组", "p": "zǔ", "sp": "zu", "r": "糸", "sc": 8, "l": 2, "f": 358},
  {"h": "每", "p": "měi", "sp": "mei", "r": "毋", "sc": 7, "l": 2, "f": 359},
  {"h": "济", "p": "jì", "sp": "ji", "r": "水", "sc": 9, "l": 3, "f": 360},
  {"h": "车", "p": "chē", "sp": "che", "r": "車", "sc": 4, "l": 1, "f": 361},
  {"h": "亲", "p": "qīn", "sp": "qin", "r": "立", "sc": 9, "l": 3, "f": 362},
  {"h": "极", "p": "jí", "sp": "ji", "r": "木", "sc": 7, "l": 3, "f": 363},
  {"h": "林", "p": "lín", "sp": "lin", "r": "木", "sc": 8, "l": 4, "f": 364},
  {"h": "服", "p": "fú", "sp": "fu", "r": "月", "sc": 8, "l": 1, "f": 365},
  {"h": "快", "p": "kuài", "sp": "kuai", "r": "心", "sc": 7, "l": 1, "f": 366},
  {"h": "办", "p": "bàn", "sp": "ban", "r": "力", "sc": 4, "l": 2, "f": 367},
  {"h": "议", "p": "yì", "sp": "yi", "r": "言", "sc": 5, "l": 3, "f": 368},
  {"h": "往", "p": "wǎng", "sp": "wang", "r": "彳", "sc": 8, "l": 2, "f": 369},
  {"h": "元", "p": "yuán", "sp": "yuan", "r": "儿", "sc": 4, "l": 1, "f": 370},
  {"h": "英", "p": "yīng", "sp": "ying", "r": "艸", "sc": 8, "l": 1, "f": 371},
  {"h": "士", "p": "shì", "sp": "shi", "r": "士", "sc": 3, "l": 4, "f": 372},
  {"h": "证", "p": "zhèng", "sp": "zheng", "r": "言", "sc": 7, "l": 3, "f": 373},
  {"h": "近", "p": "jìn", "sp": "jin", "r": "辵", "sc": 7, "l": 2, "f": 374},
  {"h": "失", "p": "shī", "sp": "shi", "r": "大", "sc": 5, "l": 3, "f": 375},
  {"h": "转", "p": "zhuǎn", "sp": "zhuan", "r": "車", "sc": 8, "l": 2, "f": 376},
  {"h": "夫", "p": "fū", "sp": "fu", "r": "大", "sc": 4, "l": 3, "f": 377},
  {"h": "令", "p": "lìng", "sp": "ling", "r": "人", "sc": 5, "l": 5, "f": 378},
  {"h": "准", "p": "zhǔn", "sp": "zhun", "r": "冫", "sc": 10, "l": 1, "f": 379},
  {"h": "布", "p": "bù", "sp": "bu", "r": "巾", "sc": 5, "l": 3, "f": 380},
  {"h": "始", "p": "shǐ", "sp": "shi", "r": "女", "sc": 8, "l": 2, "f": 381},
  {"h": "怎", "p": "zěn", "sp": "zen", "r": "心", "sc": 9, "l": 1, "f": 382},
  {"h": "呢", "p": "ne", "sp": "ne", "r": "口", "sc": 8, "l": 1, "f": 383},
  {"h": "存", "p": "cún", "sp": "cun", "r": "子", "sc": 6, "l": 3, "f": 384},
  {"h": "未", "p": "wèi", "sp": "wei", "r": "木", "sc": 5, "l": 4, "f": 385},
  {"h": "远", "p": "yuǎn", "sp": "yuan", "r": "辵", "sc": 7, "l": 1, "f": 386},
  {"h": "叫", "p": "jiào", "sp": "jiao", "r": "口", "sc": 5, "l": 1, "f": 387},
  {"h": "台", "p": "tái", "sp": "tai", "r": "口", "sc": 5, "l": 1, "f": 388},
  {"h": "单", "p": "dān", "sp": "dan", "r": "八", "sc": 8, "l": 2, "f": 389},
  {"h": "影", "p": "yǐng", "sp": "ying", "r": "彡", "sc": 15, "l": 1, "f": 390},
  {"h": "具", "p": "jù", "sp": "ju", "r": "八", "sc": 8, "l": 3, "f": 391},
  {"h": "罗", "p": "luō", "sp": "luo", "r": "网", "sc": 8, "l": 5, "f": 392},
  {"h": "字", "p": "zì", "sp": "zi", "r": "子", "sc": 6, "l": 1, "f": 393},
  {"h": "爱", "p": "ài", "sp": "ai", "r": "爪", "sc": 10, "l": 1, "f": 394},
  {"h": "击", "p": "jī", "sp": "ji", "r": "凵", "sc": 5, "l": 5, "f": 395},
  {"h": "流", "p": "liú", "sp": "liu", "r": "水", "sc": 10, "l": 2, "f": 396},
  {"h": "备", "p": "bèi", "sp": "bei", "r": "夂", "sc": 8, "l": 1, "f": 397},
  {"h": "兵", "p": "bīng", "sp": "bing", "r": "八", "sc": 7, "l": 4, "f": 398},
  {"h": "连", "p": "lián", "sp": "lian", "r": "辵", "sc": 7, "l": 3, "f": 399},
  {"h": "调", "p": "diào", "sp": "diao", "r": "言", "sc": 10, "l": 3, "f": 400},
  {"h": "深", "p": "shēn", "sp": "shen", "r": "水", "sc": 11, "l": 3, "f": 401},
  {"h": "商", "p": "shāng", "sp": "shang", "r": "口", "sc": 11, "l": 1, "f": 402},
  {"h": "算", "p": "suàn", "sp": "suan", "r": "竹", "sc": 14, "l": 2, "f": 403},
  {"h": "质", "p": "zhì", "sp": "zhi", "r": "厂", "sc": 8, "l": 4, "f": 404},
  {"h": "团", "p": "tuán", "sp": "tuan", "r": "囗", "sc": 6, "l": 3, "f": 405},
  {"h": "集", "p": "jí", "sp": "ji", "r": "隹", "sc": 12, "l": 3, "f": 406},
  {"h": "百", "p": "bǎi", "sp": "bai", "r": "白", "sc": 6, "l": 1, "f": 407},
  {"h": "需", "p": "xū", "sp": "xu", "r": "雨", "sc": 14, "l": 3, "f": 408},
  {"h": "价", "p": "jià", "sp": "jia", "r": "人", "sc": 6, "l": 3, "f": 409},
  {"h": "花", "p": "huā", "sp": "hua", "r": "艸", "sc": 7, "l": 2, "f": 410},
  {"h": "党", "p": "dǎng", "sp": "dang", "r": "儿", "sc": 10, "l": 5, "f": 411},
  {"h": "华", "p": "huá", "sp": "hua", "r": "十", "sc": 6, "l": 3, "f": 412},
  {"h": "城", "p": "chéng", "sp": "cheng", "r": "土", "sc": 9, "l": 3, "f": 413},
  {"h": "石", "p": "shí", "sp": "shi", "r": "石", "sc": 5, "l": 3, "f": 414},
  {"h": "级", "p": "jí", "sp": "ji", "r": "糸", "sc": 6, "l": 1, "f": 415},
  {"h": "整", "p": "zhěng", "sp": "zheng", "r": "攴", "sc": 16, "l": 3, "f": 416},
  {"h": "府", "p": "fǔ", "sp": "fu", "r": "广", "sc": 8, "l": 4, "f": 417},
  {"h": "离", "p": "lí", "sp": "li", "r": "禸", "sc": 10, "l": 2, "f": 418},
  {"h": "况", "p": "kuàng", "sp": "kuang", "r": "冫", "sc": 7, "l": 3, "f": 419},
  {"h": "亚", "p": "yà", "sp": "ya", "r": "一", "sc": 6, "l": 4, "f": 420},
  {"h": "请", "p": "qǐng", "sp": "qing", "r": "言", "sc": 10, "l": 1, "f": 421},
  {"h": "技", "p": "jì", "sp": "ji", "r": "手", "sc": 7, "l": 3, "f": 422},
  {"h": "际", "p": "jì", "sp": "ji", "r": "阜", "sc": 7, "l": 2, "f": 423},
  {"h": "约", "p": "yuē", "sp": "yue", "r": "糸", "sc": 6, "l": 3, "f": 424},
  {"h": "示", "p": "shì", "sp": "shi", "r": "示", "sc": 5, "l": 2, "f": 425},
  {"h": "复", "p": "fù", "sp": "fu", "r": "夊", "sc": 9, "l": 2, "f": 426},
  {"h": "病", "p": "bìng", "sp": "bing", "r": "疒", "sc": 10, "l": 1, "f": 427},
  {"h": "息", "p": "xī", "sp": "xi", "r": "心", "sc": 10, "l": 1, "f": 428},
  {"h": "究", "p": "jiū", "sp": "jiu", "r": "穴", "sc": 7, "l": 4, "f": 429},
  {"h": "线", "p": "xiàn", "sp": "xian", "r": "糸", "sc": 8, "l": 2, "f": 430},
  {"h": "似", "p": "shì", "sp": "shi", "r": "人", "sc": 6, "l": 3, "f": 431},
  {"h": "官", "p": "guān", "sp": "guan", "r": "宀", "sc": 8, "l": 4, "f": 432},
  {"h": "火", "p": "huǒ", "sp": "huo", "r": "火", "sc": 4, "l": 1, "f": 433},
  {"h": "断", "p": "duàn", "sp": "duan", "r": "斤", "sc": 11, "l": 3, "f": 434},
  {"h": "精", "p": "jīng", "sp": "jing", "r": "米", "sc": 14, "l": 3, "f": 435},
  {"h": "满", "p": "mǎn", "sp": "man", "r": "水", "sc": 13, "l": 2, "f": 436},
  {"h": "支", "p": "zhī", "sp": "zhi", "r": "支", "sc": 4, "l": 3, "f": 437},
  {"h": "视", "p": "shì", "sp": "shi", "r": "見", "sc": 8, "l": 1, "f": 438},
  {"h": "消", "p": "xiāo", "sp": "xiao", "r": "水", "sc": 10, "l": 3, "f": 439},
  {"h": "越", "p": "yuè", "sp": "yue", "r": "走", "sc": 12, "l": 2, "f": 440},
  {"h": "器", "p": "qì", "sp": "qi", "r": "口", "sc": 16, "l": 3, "f": 441},
  {"h": "容", "p": "róng", "sp": "rong", "r": "宀", "sc": 10, "l": 3, "f": 442},
  {"h": "照", "p": "zhào", "sp": "zhao", "r": "火", "sc": 13, "l": 2, "f": 443},
  {"h": "须", "p": "xū", "sp": "xu", "r": "彡", "sc": 9, "l": 2, "f": 444},
  {"h": "九", "p": "jiǔ", "sp": "jiu", "r": "乙", "sc": 2, "l": 1, "f": 445},
  {"h": "增", "p": "zēng", "sp": "zeng", "r": "土", "sc": 15, "l": 3, "f": 446},
  {"h": "研", "p": "yán", "sp": "yan", "r": "石", "sc": 9, "l": 4, "f": 447},
  {"h": "写", "p": "xiě", "sp": "xie", "r": "冖", "sc": 5, "l": 1, "f": 448},
  {"h": "称", "p": "chēng", "sp": "cheng", "r": "禾", "sc": 10, "l": 2, "f": 449},
  {"h": "企", "p": "qǐ", "sp": "qi", "r": "人", "sc": 6, "l": 4, "f": 450},
  {"h": "八", "p": "bā", "sp": "ba", "r": "八", "sc": 2, "l": 1, "f": 451},
  {"h": "功", "p": "gōng", "sp": "gong", "r": "力", "sc": 5, "l": 3, "f": 452},
  {"h": "吗", "p": "ma", "sp": "ma", "r": "口", "sc": 6, "l": 1, "f": 453},
  {"h": "包", "p": "bāo", "sp": "bao", "r": "勹", "sc": 5, "l": 1, "f": 454},
  {"h": "片", "p": "piàn", "sp": "pian", "r": "片", "sc": 4, "l": 2, "f": 455},
  {"h": "史", "p": "shǐ", "sp": "shi", "r": "口", "sc": 5, "l": 4, "f": 456},
  {"h": "委", "p": "wěi", "sp": "wei", "r": "女", "sc": 8, "l": 5, "f": 457},
  {"h": "乎", "p": "hū", "sp": "hu", "r": "丿", "sc": 5, "l": 3, "f": 458},
  {"h": "查", "p": "chá", "sp": "cha", "r": "木", "sc": 9, "l": 2, "f": 459},
  {"h": "轻", "p": "qīng", "sp": "qing", "r": "車", "sc": 9, "l": 2, "f": 460},
  {"h": "易", "p": "yì", "sp": "yi", "r": "日", "sc": 8, "l": 3, "f": 461},
  {"h": "早", "p": "zǎo", "sp": "zao", "r": "日", "sc": 6, "l": 1, "f": 462},
  {"h": "曾", "p": "céng", "sp": "ceng", "r": "曰", "sc": 12, "l": 3, "f": 463},
  {"h": "除", "p": "chú", "sp": "chu", "r": "阜", "sc": 9, "l": 3, "f": 464},
  {"h": "农", "p": "nóng", "sp": "nong", "r": "冖", "sc": 6, "l": 3, "f": 465},
  {"h": "找", "p": "zhǎo", "sp": "zhao", "r": "手", "sc": 7, "l": 1, "f": 466},
  {
    "h": "装",
    "p": "zhuāng",
    "sp": "zhuang",
    "r": "衣",
    "sc": 12,
    "l": 2,
    "f": 467
  },
  {"h": "广", "p": "guǎng", "sp": "guang", "r": "广", "sc": 3, "l": 2, "f": 468},
  {"h": "显", "p": "xiǎn", "sp": "xian", "r": "日", "sc": 9, "l": 3, "f": 469},
  {"h": "吧", "p": "ba", "sp": "ba", "r": "口", "sc": 7, "l": 1, "f": 470},
  {"h": "阿", "p": "ā", "sp": "a", "r": "阜", "sc": 7, "l": 3, "f": 471},
  {"h": "李", "p": "lǐ", "sp": "li", "r": "木", "sc": 7, "l": 2, "f": 472},
  {"h": "标", "p": "biāo", "sp": "biao", "r": "木", "sc": 9, "l": 3, "f": 473},
  {"h": "谈", "p": "tán", "sp": "tan", "r": "言", "sc": 10, "l": 3, "f": 474},
  {"h": "吃", "p": "chī", "sp": "chi", "r": "口", "sc": 6, "l": 1, "f": 475},
  {"h": "图", "p": "tú", "sp": "tu", "r": "囗", "sc": 8, "l": 1, "f": 476},
  {"h": "念", "p": "niàn", "sp": "nian", "r": "心", "sc": 8, "l": 3, "f": 477},
  {"h": "六", "p": "liù", "sp": "liu", "r": "八", "sc": 4, "l": 1, "f": 478},
  {"h": "引", "p": "yǐn", "sp": "yin", "r": "弓", "sc": 4, "l": 4, "f": 479},
  {"h": "历", "p": "lì", "sp": "li", "r": "厂", "sc": 4, "l": 3, "f": 480},
  {"h": "首", "p": "shǒu", "sp": "shou", "r": "首", "sc": 9, "l": 3, "f": 481},
  {"h": "医", "p": "yī", "sp": "yi", "r": "匸", "sc": 7, "l": 1, "f": 482},
  {"h": "局", "p": "jú", "sp": "ju", "r": "尸", "sc": 7, "l": 4, "f": 483},
  {"h": "突", "p": "tū", "sp": "tu", "r": "穴", "sc": 9, "l": 3, "f": 484},
  {"h": "专", "p": "zhuān", "sp": "zhuan", "r": "一", "sc": 4, "l": 3, "f": 485},
  {"h": "费", "p": "fèi", "sp": "fei", "r": "貝", "sc": 9, "l": 3, "f": 486},
  {"h": "号", "p": "hào", "sp": "hao", "r": "口", "sc": 5, "l": 1, "f": 487},
  {"h": "尽", "p": "jǐn", "sp": "jin", "r": "尸", "sc": 6, "l": 3, "f": 488},
  {"h": "另", "p": "lìng", "sp": "ling", "r": "口", "sc": 5, "l": 3, "f": 489},
  {"h": "周", "p": "zhōu", "sp": "zhou", "r": "口", "sc": 8, "l": 2, "f": 490},
  {"h": "较", "p": "jiào", "sp": "jiao", "r": "車", "sc": 10, "l": 3, "f": 491},
  {"h": "注", "p": "zhù", "sp": "zhu", "r": "水", "sc": 8, "l": 3, "f": 492},
  {"h": "语", "p": "yǔ", "sp": "yu", "r": "言", "sc": 9, "l": 1, "f": 493},
  {"h": "仅", "p": "jǐn", "sp": "jin", "r": "人", "sc": 4, "l": 3, "f": 494},
  {"h": "考", "p": "kǎo", "sp": "kao", "r": "老", "sc": 6, "l": 1, "f": 495},
  {"h": "落", "p": "luò", "sp": "luo", "r": "艸", "sc": 12, "l": 3, "f": 496},
  {"h": "青", "p": "qīng", "sp": "qing", "r": "青", "sc": 8, "l": 2, "f": 497},
  {"h": "随", "p": "suí", "sp": "sui", "r": "阜", "sc": 11, "l": 2, "f": 498},
  {"h": "选", "p": "xuǎn", "sp": "xuan", "r": "辵", "sc": 9, "l": 2, "f": 499},
  {"h": "列", "p": "liè", "sp": "lie", "r": "刀", "sc": 6, "l": 4, "f": 500},
  {"h": "武", "p": "wǔ", "sp": "wu", "r": "止", "sc": 8, "l": 3, "f": 501},
  {"h": "红", "p": "hóng", "sp": "hong", "r": "糸", "sc": 6, "l": 2, "f": 502},
  {"h": "响", "p": "xiǎng", "sp": "xiang", "r": "口", "sc": 9, "l": 2, "f": 503},
  {"h": "虽", "p": "suī", "sp": "sui", "r": "虫", "sc": 9, "l": 2, "f": 504},
  {"h": "推", "p": "tuī", "sp": "tui", "r": "手", "sc": 11, "l": 2, "f": 505},
  {"h": "势", "p": "shì", "sp": "shi", "r": "力", "sc": 8, "l": 3, "f": 506},
  {"h": "参", "p": "cān", "sp": "can", "r": "厶", "sc": 8, "l": 2, "f": 507},
  {"h": "希", "p": "xī", "sp": "xi", "r": "巾", "sc": 7, "l": 2, "f": 508},
  {"h": "古", "p": "gǔ", "sp": "gu", "r": "口", "sc": 5, "l": 3, "f": 509},
  {"h": "众", "p": "zhòng", "sp": "zhong", "r": "人", "sc": 6, "l": 3, "f": 510},
  {"h": "构", "p": "gòu", "sp": "gou", "r": "木", "sc": 8, "l": 4, "f": 511},
  {"h": "房", "p": "fáng", "sp": "fang", "r": "戶", "sc": 8, "l": 1, "f": 512},
  {"h": "半", "p": "bàn", "sp": "ban", "r": "十", "sc": 5, "l": 1, "f": 513},
  {"h": "节", "p": "jié", "sp": "jie", "r": "艸", "sc": 5, "l": 1, "f": 514},
  {"h": "土", "p": "tǔ", "sp": "tu", "r": "土", "sc": 3, "l": 4, "f": 515},
  {"h": "投", "p": "tóu", "sp": "tou", "r": "手", "sc": 7, "l": 4, "f": 516},
  {"h": "某", "p": "mǒu", "sp": "mou", "r": "木", "sc": 9, "l": 3, "f": 517},
  {"h": "案", "p": "àn", "sp": "an", "r": "木", "sc": 10, "l": 4, "f": 518},
  {"h": "黑", "p": "hēi", "sp": "hei", "r": "黑", "sc": 12, "l": 2, "f": 519},
  {"h": "维", "p": "wéi", "sp": "wei", "r": "糸", "sc": 11, "l": 4, "f": 520},
  {"h": "革", "p": "gé", "sp": "ge", "r": "革", "sc": 9, "l": 5, "f": 521},
  {"h": "划", "p": "huà", "sp": "hua", "r": "刀", "sc": 6, "l": 2, "f": 522},
  {"h": "敌", "p": "dí", "sp": "di", "r": "舌", "sc": 10, "l": 4, "f": 523},
  {"h": "致", "p": "zhì", "sp": "zhi", "r": "至", "sc": 10, "l": 4, "f": 524},
  {"h": "陈", "p": "chén", "sp": "chen", "r": "阜", "sc": 7, "l": 4, "f": 525},
  {"h": "律", "p": "lǜ", "sp": "lu", "r": "彳", "sc": 9, "l": 4, "f": 526},
  {"h": "足", "p": "zú", "sp": "zu", "r": "足", "sc": 7, "l": 2, "f": 527},
  {"h": "态", "p": "tài", "sp": "tai", "r": "心", "sc": 8, "l": 2, "f": 528},
  {"h": "护", "p": "hù", "sp": "hu", "r": "手", "sc": 7, "l": 2, "f": 529},
  {"h": "七", "p": "qī", "sp": "qi", "r": "一", "sc": 2, "l": 1, "f": 530},
  {"h": "兴", "p": "xìng", "sp": "xing", "r": "八", "sc": 6, "l": 1, "f": 531},
  {"h": "派", "p": "pài", "sp": "pai", "r": "水", "sc": 9, "l": 3, "f": 532},
  {"h": "孩", "p": "hái", "sp": "hai", "r": "子", "sc": 9, "l": 1, "f": 533},
  {"h": "验", "p": "yàn", "sp": "yan", "r": "馬", "sc": 10, "l": 3, "f": 534},
  {"h": "责", "p": "zé", "sp": "ze", "r": "貝", "sc": 8, "l": 3, "f": 535},
  {"h": "营", "p": "yíng", "sp": "ying", "r": "艸", "sc": 11, "l": 3, "f": 536},
  {"h": "星", "p": "xīng", "sp": "xing", "r": "日", "sc": 9, "l": 1, "f": 537},
  {"h": "够", "p": "gòu", "sp": "gou", "r": "夕", "sc": 11, "l": 2, "f": 538},
  {"h": "章", "p": "zhāng", "sp": "zhang", "r": "立", "sc": 11, "l": 3, "f": 539},
  {"h": "音", "p": "yīn", "sp": "yin", "r": "音", "sc": 9, "l": 2, "f": 540},
  {"h": "跟", "p": "gēn", "sp": "gen", "r": "足", "sc": 13, "l": 1, "f": 541},
  {"h": "志", "p": "zhì", "sp": "zhi", "r": "心", "sc": 7, "l": 3, "f": 542},
  {"h": "底", "p": "dǐ", "sp": "di", "r": "广", "sc": 8, "l": 3, "f": 543},
  {"h": "站", "p": "zhàn", "sp": "zhan", "r": "立", "sc": 10, "l": 1, "f": 544},
  {"h": "严", "p": "yán", "sp": "yan", "r": "一", "sc": 7, "l": 4, "f": 545},
  {"h": "巴", "p": "bā", "sp": "ba", "r": "己", "sc": 4, "l": 4, "f": 546},
  {"h": "例", "p": "lì", "sp": "li", "r": "人", "sc": 8, "l": 2, "f": 547},
  {"h": "防", "p": "fáng", "sp": "fang", "r": "阜", "sc": 6, "l": 3, "f": 548},
  {"h": "族", "p": "zú", "sp": "zu", "r": "方", "sc": 11, "l": 3, "f": 549},
  {"h": "供", "p": "gōng", "sp": "gong", "r": "人", "sc": 8, "l": 4, "f": 550},
  {"h": "效", "p": "xiào", "sp": "xiao", "r": "攴", "sc": 10, "l": 3, "f": 551},
  {"h": "续", "p": "xù", "sp": "xu", "r": "糸", "sc": 11, "l": 3, "f": 552},
  {"h": "施", "p": "shī", "sp": "shi", "r": "方", "sc": 9, "l": 4, "f": 553},
  {"h": "留", "p": "liú", "sp": "liu", "r": "田", "sc": 10, "l": 2, "f": 554},
  {"h": "讲", "p": "jiǎng", "sp": "jiang", "r": "言", "sc": 6, "l": 2, "f": 555},
  {"h": "型", "p": "xíng", "sp": "xing", "r": "土", "sc": 9, "l": 4, "f": 556},
  {"h": "料", "p": "liào", "sp": "liao", "r": "斗", "sc": 10, "l": 4, "f": 557},
  {"h": "终", "p": "zhōng", "sp": "zhong", "r": "糸", "sc": 8, "l": 3, "f": 558},
  {"h": "答", "p": "dá", "sp": "da", "r": "竹", "sc": 12, "l": 1, "f": 559},
  {"h": "紧", "p": "jǐn", "sp": "jin", "r": "糸", "sc": 10, "l": 3, "f": 560},
  {"h": "黄", "p": "huáng", "sp": "huang", "r": "黃", "sc": 11, "l": 2, "f": 561},
  {"h": "绝", "p": "jué", "sp": "jue", "r": "糸", "sc": 9, "l": 3, "f": 562},
  {"h": "奇", "p": "qí", "sp": "qi", "r": "大", "sc": 8, "l": 3, "f": 563},
  {"h": "察", "p": "chá", "sp": "cha", "r": "宀", "sc": 14, "l": 3, "f": 564},
  {"h": "母", "p": "mǔ", "sp": "mu", "r": "毋", "sc": 5, "l": 3, "f": 565},
  {"h": "京", "p": "jīng", "sp": "jing", "r": "亠", "sc": 8, "l": 1, "f": 566},
  {"h": "段", "p": "duàn", "sp": "duan", "r": "殳", "sc": 9, "l": 2, "f": 567},
  {"h": "依", "p": "yī", "sp": "yi", "r": "人", "sc": 8, "l": 4, "f": 568},
  {"h": "批", "p": "pī", "sp": "pi", "r": "手", "sc": 7, "l": 3, "f": 569},
  {"h": "群", "p": "qún", "sp": "qun", "r": "羊", "sc": 13, "l": 3, "f": 570},
  {"h": "项", "p": "xiàng", "sp": "xiang", "r": "頁", "sc": 9, "l": 4, "f": 571},
  {"h": "故", "p": "gù", "sp": "gu", "r": "攴", "sc": 9, "l": 2, "f": 572},
  {"h": "按", "p": "àn", "sp": "an", "r": "手", "sc": 9, "l": 3, "f": 573},
  {"h": "河", "p": "hé", "sp": "he", "r": "水", "sc": 8, "l": 2, "f": 574},
  {"h": "米", "p": "mǐ", "sp": "mi", "r": "米", "sc": 6, "l": 1, "f": 575},
  {"h": "围", "p": "wéi", "sp": "wei", "r": "囗", "sc": 7, "l": 3, "f": 576},
  {"h": "江", "p": "jiāng", "sp": "jiang", "r": "水", "sc": 6, "l": 4, "f": 577},
  {"h": "织", "p": "zhī", "sp": "zhi", "r": "糸", "sc": 8, "l": 5, "f": 578},
  {"h": "害", "p": "hài", "sp": "hai", "r": "宀", "sc": 10, "l": 3, "f": 579},
  {"h": "斗", "p": "dòu", "sp": "dou", "r": "斗", "sc": 4, "l": 4, "f": 580},
  {
    "h": "双",
    "p": "shuāng",
    "sp": "shuang",
    "r": "又",
    "sc": 4,
    "l": 2,
    "f": 581
  },
  {"h": "境", "p": "jìng", "sp": "jing", "r": "土", "sc": 14, "l": 3, "f": 582},
  {"h": "客", "p": "kè", "sp": "ke", "r": "宀", "sc": 9, "l": 1, "f": 583},
  {"h": "纪", "p": "jì", "sp": "ji", "r": "糸", "sc": 6, "l": 3, "f": 584},
  {"h": "采", "p": "cǎi", "sp": "cai", "r": "釆", "sc": 8, "l": 3, "f": 585},
  {"h": "举", "p": "jǔ", "sp": "ju", "r": "丶", "sc": 9, "l": 2, "f": 586},
  {"h": "杀", "p": "shā", "sp": "sha", "r": "木", "sc": 6, "l": 5, "f": 587},
  {"h": "攻", "p": "gōng", "sp": "gong", "r": "攴", "sc": 7, "l": 5, "f": 588},
  {"h": "父", "p": "fù", "sp": "fu", "r": "父", "sc": 4, "l": 3, "f": 589},
  {"h": "苏", "p": "sū", "sp": "su", "r": "艸", "sc": 7, "l": 5, "f": 590},
  {"h": "密", "p": "mì", "sp": "mi", "r": "宀", "sc": 11, "l": 4, "f": 591},
  {"h": "低", "p": "dī", "sp": "di", "r": "人", "sc": 7, "l": 2, "f": 592},
  {"h": "朝", "p": "cháo", "sp": "chao", "r": "月", "sc": 12, "l": 3, "f": 593},
  {"h": "友", "p": "yǒu", "sp": "you", "r": "又", "sc": 4, "l": 1, "f": 594},
  {"h": "诉", "p": "sù", "sp": "su", "r": "言", "sc": 7, "l": 1, "f": 595},
  {"h": "止", "p": "zhǐ", "sp": "zhi", "r": "止", "sc": 4, "l": 3, "f": 596},
  {"h": "细", "p": "xì", "sp": "xi", "r": "糸", "sc": 8, "l": 4, "f": 597},
  {"h": "愿", "p": "yuàn", "sp": "yuan", "r": "心", "sc": 14, "l": 2, "f": 598},
  {"h": "千", "p": "qiān", "sp": "qian", "r": "十", "sc": 3, "l": 2, "f": 599},
  {"h": "值", "p": "zhí", "sp": "zhi", "r": "人", "sc": 10, "l": 3, "f": 600},
  {"h": "仍", "p": "réng", "sp": "reng", "r": "人", "sc": 4, "l": 3, "f": 601},
  {"h": "男", "p": "nán", "sp": "nan", "r": "田", "sc": 7, "l": 1, "f": 602},
  {"h": "钱", "p": "qián", "sp": "qian", "r": "金", "sc": 10, "l": 1, "f": 603},
  {"h": "破", "p": "pò", "sp": "po", "r": "石", "sc": 10, "l": 3, "f": 604},
  {"h": "网", "p": "wǎng", "sp": "wang", "r": "网", "sc": 6, "l": 1, "f": 605},
  {"h": "热", "p": "rè", "sp": "re", "r": "火", "sc": 10, "l": 1, "f": 606},
  {"h": "助", "p": "zhù", "sp": "zhu", "r": "力", "sc": 7, "l": 1, "f": 607},
  {"h": "倒", "p": "dào", "sp": "dao", "r": "人", "sc": 10, "l": 2, "f": 608},
  {"h": "育", "p": "yù", "sp": "yu", "r": "肉", "sc": 8, "l": 2, "f": 609},
  {"h": "属", "p": "shǔ", "sp": "shu", "r": "尸", "sc": 12, "l": 3, "f": 610},
  {"h": "坐", "p": "zuò", "sp": "zuo", "r": "土", "sc": 7, "l": 1, "f": 611},
  {"h": "帝", "p": "dì", "sp": "di", "r": "巾", "sc": 9, "l": 5, "f": 612},
  {"h": "限", "p": "xiàn", "sp": "xian", "r": "阜", "sc": 8, "l": 4, "f": 613},
  {"h": "船", "p": "chuán", "sp": "chuan", "r": "舟", "sc": 11, "l": 2, "f": 614},
  {"h": "脸", "p": "liǎn", "sp": "lian", "r": "肉", "sc": 11, "l": 2, "f": 615},
  {"h": "职", "p": "zhí", "sp": "zhi", "r": "耳", "sc": 11, "l": 3, "f": 616},
  {"h": "速", "p": "sù", "sp": "su", "r": "辵", "sc": 10, "l": 3, "f": 617},
  {"h": "刻", "p": "kè", "sp": "ke", "r": "刀", "sc": 8, "l": 3, "f": 618},
  {"h": "乐", "p": "lè", "sp": "le", "r": "丿", "sc": 5, "l": 2, "f": 619},
  {"h": "否", "p": "fǒu", "sp": "fou", "r": "口", "sc": 7, "l": 3, "f": 620},
  {"h": "刚", "p": "gāng", "sp": "gang", "r": "刀", "sc": 6, "l": 2, "f": 621},
  {"h": "威", "p": "wēi", "sp": "wei", "r": "女", "sc": 9, "l": 5, "f": 622},
  {"h": "毛", "p": "máo", "sp": "mao", "r": "毛", "sc": 4, "l": 2, "f": 623},
  {
    "h": "状",
    "p": "zhuàng",
    "sp": "zhuang",
    "r": "犬",
    "sc": 7,
    "l": 3,
    "f": 624
  },
  {"h": "率", "p": "lǜ", "sp": "lu", "r": "玄", "sc": 11, "l": 4, "f": 625},
  {"h": "甚", "p": "shén", "sp": "shen", "r": "甘", "sc": 9, "l": 4, "f": 626},
  {"h": "独", "p": "dú", "sp": "du", "r": "犬", "sc": 9, "l": 4, "f": 627},
  {"h": "球", "p": "qiú", "sp": "qiu", "r": "玉", "sc": 11, "l": 1, "f": 628},
  {"h": "般", "p": "bān", "sp": "ban", "r": "舟", "sc": 10, "l": 2, "f": 629},
  {"h": "普", "p": "pǔ", "sp": "pu", "r": "日", "sc": 12, "l": 2, "f": 630},
  {"h": "怕", "p": "pà", "sp": "pa", "r": "心", "sc": 8, "l": 2, "f": 631},
  {"h": "弹", "p": "dàn", "sp": "dan", "r": "弓", "sc": 11, "l": 4, "f": 632},
  {"h": "校", "p": "xiào", "sp": "xiao", "r": "木", "sc": 10, "l": 1, "f": 633},
  {"h": "苦", "p": "kǔ", "sp": "ku", "r": "艸", "sc": 8, "l": 3, "f": 634},
  {
    "h": "创",
    "p": "chuàng",
    "sp": "chuang",
    "r": "刀",
    "sc": 6,
    "l": 3,
    "f": 635
  },
  {"h": "假", "p": "jiǎ", "sp": "jia", "r": "人", "sc": 11, "l": 1, "f": 636},
  {"h": "久", "p": "jiǔ", "sp": "jiu", "r": "丿", "sc": 3, "l": 2, "f": 637},
  {"h": "错", "p": "cuò", "sp": "cuo", "r": "金", "sc": 13, "l": 1, "f": 638},
  {"h": "承", "p": "chéng", "sp": "cheng", "r": "手", "sc": 8, "l": 4, "f": 639},
  {"h": "印", "p": "yìn", "sp": "yin", "r": "卩", "sc": 5, "l": 2, "f": 640},
  {"h": "晚", "p": "wǎn", "sp": "wan", "r": "日", "sc": 11, "l": 1, "f": 641},
  {"h": "兰", "p": "lán", "sp": "lan", "r": "八", "sc": 5, "l": 4, "f": 642},
  {"h": "试", "p": "shì", "sp": "shi", "r": "言", "sc": 8, "l": 1, "f": 643},
  {"h": "股", "p": "gǔ", "sp": "gu", "r": "肉", "sc": 8, "l": 5, "f": 644},
  {"h": "拿", "p": "ná", "sp": "na", "r": "手", "sc": 10, "l": 1, "f": 645},
  {"h": "脑", "p": "nǎo", "sp": "nao", "r": "肉", "sc": 10, "l": 1, "f": 646},
  {"h": "预", "p": "yù", "sp": "yu", "r": "頁", "sc": 10, "l": 2, "f": 647},
  {"h": "谁", "p": "shuí", "sp": "shui", "r": "言", "sc": 10, "l": 1, "f": 648},
  {"h": "益", "p": "yì", "sp": "yi", "r": "皿", "sc": 10, "l": 4, "f": 649},
  {"h": "阳", "p": "yáng", "sp": "yang", "r": "阜", "sc": 6, "l": 2, "f": 650},
  {"h": "若", "p": "ruò", "sp": "ruo", "r": "艸", "sc": 8, "l": 5, "f": 651},
  {"h": "哪", "p": "nǎ", "sp": "na", "r": "口", "sc": 9, "l": 1, "f": 652},
  {"h": "微", "p": "wēi", "sp": "wei", "r": "彳", "sc": 13, "l": 4, "f": 653},
  {"h": "尼", "p": "ní", "sp": "ni", "r": "尸", "sc": 5, "l": 5, "f": 654},
  {"h": "继", "p": "jì", "sp": "ji", "r": "糸", "sc": 10, "l": 3, "f": 655},
  {"h": "送", "p": "sòng", "sp": "song", "r": "辵", "sc": 9, "l": 1, "f": 656},
  {"h": "急", "p": "jí", "sp": "ji", "r": "心", "sc": 9, "l": 2, "f": 657},
  {"h": "血", "p": "xuè", "sp": "xue", "r": "血", "sc": 6, "l": 3, "f": 658},
  {"h": "惊", "p": "jīng", "sp": "jing", "r": "心", "sc": 11, "l": 4, "f": 659},
  {"h": "伤", "p": "shāng", "sp": "shang", "r": "人", "sc": 6, "l": 3, "f": 660},
  {"h": "素", "p": "sù", "sp": "su", "r": "糸", "sc": 10, "l": 5, "f": 661},
  {"h": "药", "p": "yào", "sp": "yao", "r": "艸", "sc": 9, "l": 2, "f": 662},
  {"h": "适", "p": "shì", "sp": "shi", "r": "辵", "sc": 9, "l": 2, "f": 663},
  {"h": "波", "p": "bō", "sp": "bo", "r": "水", "sc": 8, "l": 5, "f": 664},
  {"h": "夜", "p": "yè", "sp": "ye", "r": "夕", "sc": 8, "l": 2, "f": 665},
  {"h": "省", "p": "shěng", "sp": "sheng", "r": "目", "sc": 9, "l": 2, "f": 666},
  {"h": "初", "p": "chū", "sp": "chu", "r": "刀", "sc": 7, "l": 3, "f": 667},
  {"h": "喜", "p": "xǐ", "sp": "xi", "r": "口", "sc": 12, "l": 1, "f": 668},
  {"h": "卫", "p": "wèi", "sp": "wei", "r": "乙", "sc": 3, "l": 3, "f": 669},
  {"h": "源", "p": "yuán", "sp": "yuan", "r": "水", "sc": 13, "l": 4, "f": 670},
  {"h": "食", "p": "shí", "sp": "shi", "r": "食", "sc": 9, "l": 2, "f": 671},
  {"h": "险", "p": "xiǎn", "sp": "xian", "r": "阜", "sc": 9, "l": 3, "f": 672},
  {"h": "待", "p": "dài", "sp": "dai", "r": "彳", "sc": 9, "l": 3, "f": 673},
  {"h": "述", "p": "shù", "sp": "shu", "r": "辵", "sc": 8, "l": 4, "f": 674},
  {"h": "陆", "p": "lù", "sp": "lu", "r": "阜", "sc": 7, "l": 4, "f": 675},
  {"h": "习", "p": "xí", "sp": "xi", "r": "冫", "sc": 3, "l": 1, "f": 676},
  {"h": "置", "p": "zhì", "sp": "zhi", "r": "网", "sc": 13, "l": 4, "f": 677},
  {"h": "居", "p": "jū", "sp": "ju", "r": "尸", "sc": 8, "l": 3, "f": 678},
  {"h": "劳", "p": "láo", "sp": "lao", "r": "力", "sc": 7, "l": 5, "f": 679},
  {"h": "财", "p": "cái", "sp": "cai", "r": "貝", "sc": 7, "l": 4, "f": 680},
  {"h": "环", "p": "huán", "sp": "huan", "r": "玉", "sc": 8, "l": 3, "f": 681},
  {"h": "排", "p": "pái", "sp": "pai", "r": "手", "sc": 11, "l": 2, "f": 682},
  {"h": "福", "p": "fú", "sp": "fu", "r": "示", "sc": 13, "l": 3, "f": 683},
  {"h": "纳", "p": "nà", "sp": "na", "r": "糸", "sc": 7, "l": 5, "f": 684},
  {"h": "欢", "p": "huān", "sp": "huan", "r": "欠", "sc": 6, "l": 1, "f": 685},
  {"h": "雷", "p": "léi", "sp": "lei", "r": "雨", "sc": 13, "l": 4, "f": 686},
  {"h": "警", "p": "jǐng", "sp": "jing", "r": "言", "sc": 19, "l": 3, "f": 687},
  {"h": "获", "p": "huò", "sp": "huo", "r": "艸", "sc": 10, "l": 4, "f": 688},
  {"h": "模", "p": "mó", "sp": "mo", "r": "木", "sc": 14, "l": 4, "f": 689},
  {"h": "充", "p": "chōng", "sp": "chong", "r": "儿", "sc": 6, "l": 3, "f": 690},
  {"h": "负", "p": "fù", "sp": "fu", "r": "貝", "sc": 6, "l": 3, "f": 691},
  {"h": "云", "p": "yún", "sp": "yun", "r": "二", "sc": 4, "l": 2, "f": 692},
  {"h": "停", "p": "tíng", "sp": "ting", "r": "人", "sc": 11, "l": 2, "f": 693},
  {"h": "木", "p": "mù", "sp": "mu", "r": "木", "sc": 4, "l": 3, "f": 694},
  {"h": "游", "p": "yóu", "sp": "you", "r": "水", "sc": 12, "l": 2, "f": 695},
  {"h": "龙", "p": "lóng", "sp": "long", "r": "龍", "sc": 5, "l": 3, "f": 696},
  {"h": "树", "p": "shù", "sp": "shu", "r": "木", "sc": 9, "l": 1, "f": 697},
  {"h": "疑", "p": "yí", "sp": "yi", "r": "疋", "sc": 14, "l": 4, "f": 698},
  {"h": "层", "p": "céng", "sp": "ceng", "r": "尸", "sc": 7, "l": 2, "f": 699},
  {"h": "冷", "p": "lěng", "sp": "leng", "r": "冫", "sc": 7, "l": 1, "f": 700},
  {"h": "洲", "p": "zhōu", "sp": "zhou", "r": "水", "sc": 9, "l": 4, "f": 701},
  {"h": "冲", "p": "chōng", "sp": "chong", "r": "冫", "sc": 6, "l": 5, "f": 702},
  {"h": "射", "p": "shè", "sp": "she", "r": "寸", "sc": 10, "l": 5, "f": 703},
  {"h": "略", "p": "lüè", "sp": "lue", "r": "田", "sc": 11, "l": 5, "f": 704},
  {"h": "范", "p": "fàn", "sp": "fan", "r": "艸", "sc": 0, "l": 3, "f": 705},
  {"h": "竟", "p": "jìng", "sp": "jing", "r": "立", "sc": 11, "l": 4, "f": 706},
  {"h": "句", "p": "jù", "sp": "ju", "r": "口", "sc": 5, "l": 2, "f": 707},
  {"h": "室", "p": "shì", "sp": "shi", "r": "宀", "sc": 9, "l": 2, "f": 708},
  {"h": "异", "p": "yì", "sp": "yi", "r": "廾", "sc": 6, "l": 5, "f": 709},
  {"h": "激", "p": "jī", "sp": "ji", "r": "水", "sc": 16, "l": 4, "f": 710},
  {"h": "汉", "p": "hàn", "sp": "han", "r": "水", "sc": 5, "l": 1, "f": 711},
  {"h": "村", "p": "cūn", "sp": "cun", "r": "木", "sc": 7, "l": 3, "f": 712},
  {"h": "哈", "p": "hā", "sp": "ha", "r": "口", "sc": 9, "l": 3, "f": 713},
  {"h": "策", "p": "cè", "sp": "ce", "r": "竹", "sc": 12, "l": 5, "f": 714},
  {"h": "演", "p": "yǎn", "sp": "yan", "r": "水", "sc": 14, "l": 3, "f": 715},
  {"h": "简", "p": "jiǎn", "sp": "jian", "r": "竹", "sc": 13, "l": 3, "f": 716},
  {"h": "卡", "p": "kǎ", "sp": "ka", "r": "卜", "sc": 5, "l": 2, "f": 717},
  {"h": "罪", "p": "zuì", "sp": "zui", "r": "网", "sc": 13, "l": 5, "f": 718},
  {"h": "判", "p": "pàn", "sp": "pan", "r": "刀", "sc": 7, "l": 3, "f": 719},
  {"h": "担", "p": "dān", "sp": "dan", "r": "手", "sc": 8, "l": 3, "f": 720},
  {"h": "州", "p": "zhōu", "sp": "zhou", "r": "巛", "sc": 6, "l": 5, "f": 721},
  {"h": "静", "p": "jìng", "sp": "jing", "r": "青", "sc": 14, "l": 2, "f": 722},
  {"h": "退", "p": "tuì", "sp": "tui", "r": "辵", "sc": 9, "l": 3, "f": 723},
  {"h": "既", "p": "jì", "sp": "ji", "r": "无", "sc": 9, "l": 4, "f": 724},
  {"h": "衣", "p": "yī", "sp": "yi", "r": "衣", "sc": 6, "l": 1, "f": 725},
  {"h": "您", "p": "nín", "sp": "nin", "r": "心", "sc": 11, "l": 1, "f": 726},
  {"h": "宗", "p": "zōng", "sp": "zong", "r": "宀", "sc": 8, "l": 5, "f": 727},
  {"h": "积", "p": "jī", "sp": "ji", "r": "禾", "sc": 10, "l": 3, "f": 728},
  {"h": "余", "p": "yú", "sp": "yu", "r": "人", "sc": 7, "l": 4, "f": 729},
  {"h": "痛", "p": "tòng", "sp": "tong", "r": "疒", "sc": 12, "l": 3, "f": 730},
  {"h": "检", "p": "jiǎn", "sp": "jian", "r": "木", "sc": 11, "l": 2, "f": 731},
  {"h": "差", "p": "chà", "sp": "cha", "r": "工", "sc": 9, "l": 1, "f": 732},
  {"h": "富", "p": "fù", "sp": "fu", "r": "宀", "sc": 12, "l": 3, "f": 733},
  {"h": "灵", "p": "líng", "sp": "ling", "r": "彐", "sc": 7, "l": 5, "f": 734},
  {"h": "协", "p": "xié", "sp": "xie", "r": "十", "sc": 6, "l": 5, "f": 735},
  {"h": "角", "p": "jiǎo", "sp": "jiao", "r": "角", "sc": 7, "l": 2, "f": 736},
  {"h": "占", "p": "zhàn", "sp": "zhan", "r": "卜", "sc": 5, "l": 2, "f": 737},
  {"h": "配", "p": "pèi", "sp": "pei", "r": "酉", "sc": 10, "l": 3, "f": 738},
  {"h": "征", "p": "zhēng", "sp": "zheng", "r": "彳", "sc": 8, "l": 4, "f": 739},
  {"h": "修", "p": "xiū", "sp": "xiu", "r": "人", "sc": 9, "l": 3, "f": 740},
  {"h": "皮", "p": "pí", "sp": "pi", "r": "皮", "sc": 5, "l": 3, "f": 741},
  {"h": "挥", "p": "huī", "sp": "hui", "r": "手", "sc": 9, "l": 4, "f": 742},
  {"h": "胜", "p": "shèng", "sp": "sheng", "r": "肉", "sc": 9, "l": 3, "f": 743},
  {"h": "降", "p": "jiàng", "sp": "jiang", "r": "阜", "sc": 8, "l": 4, "f": 744},
  {"h": "阶", "p": "jiē", "sp": "jie", "r": "阜", "sc": 6, "l": 4, "f": 745},
  {"h": "审", "p": "shěn", "sp": "shen", "r": "宀", "sc": 8, "l": 5, "f": 746},
  {"h": "沉", "p": "chén", "sp": "chen", "r": "水", "sc": 7, "l": 4, "f": 747},
  {"h": "坚", "p": "jiān", "sp": "jian", "r": "土", "sc": 7, "l": 3, "f": 748},
  {"h": "善", "p": "shàn", "sp": "shan", "r": "口", "sc": 12, "l": 3, "f": 749},
  {"h": "妈", "p": "mā", "sp": "ma", "r": "女", "sc": 6, "l": 1, "f": 750},
  {"h": "刘", "p": "liú", "sp": "liu", "r": "刀", "sc": 6, "l": 2, "f": 751},
  {"h": "读", "p": "dú", "sp": "du", "r": "言", "sc": 10, "l": 1, "f": 752},
  {"h": "啊", "p": "a", "sp": "a", "r": "口", "sc": 10, "l": 2, "f": 753},
  {"h": "超", "p": "chāo", "sp": "chao", "r": "走", "sc": 12, "l": 2, "f": 754},
  {"h": "免", "p": "miǎn", "sp": "mian", "r": "儿", "sc": 7, "l": 4, "f": 755},
  {"h": "压", "p": "yā", "sp": "ya", "r": "厂", "sc": 6, "l": 3, "f": 756},
  {"h": "银", "p": "yín", "sp": "yin", "r": "金", "sc": 11, "l": 2, "f": 757},
  {"h": "买", "p": "mǎi", "sp": "mai", "r": "大", "sc": 6, "l": 1, "f": 758},
  {"h": "皇", "p": "huáng", "sp": "huang", "r": "白", "sc": 9, "l": 4, "f": 759},
  {"h": "养", "p": "yǎng", "sp": "yang", "r": "八", "sc": 9, "l": 2, "f": 760},
  {"h": "伊", "p": "yī", "sp": "yi", "r": "人", "sc": 6, "l": 5, "f": 761},
  {"h": "怀", "p": "huái", "sp": "huai", "r": "心", "sc": 7, "l": 4, "f": 762},
  {"h": "执", "p": "zhí", "sp": "zhi", "r": "手", "sc": 6, "l": 5, "f": 763},
  {"h": "副", "p": "fù", "sp": "fu", "r": "刀", "sc": 11, "l": 3, "f": 764},
  {"h": "乱", "p": "luàn", "sp": "luan", "r": "乙", "sc": 7, "l": 3, "f": 765},
  {"h": "抗", "p": "kàng", "sp": "kang", "r": "手", "sc": 7, "l": 5, "f": 766},
  {"h": "犯", "p": "fàn", "sp": "fan", "r": "犬", "sc": 5, "l": 5, "f": 767},
  {"h": "追", "p": "zhuī", "sp": "zhui", "r": "辵", "sc": 9, "l": 3, "f": 768},
  {"h": "帮", "p": "bāng", "sp": "bang", "r": "巾", "sc": 9, "l": 1, "f": 769},
  {"h": "宣", "p": "xuān", "sp": "xuan", "r": "宀", "sc": 9, "l": 3, "f": 770},
  {"h": "佛", "p": "fú", "sp": "fu", "r": "人", "sc": 7, "l": 5, "f": 771},
  {"h": "岁", "p": "suì", "sp": "sui", "r": "山", "sc": 6, "l": 1, "f": 772},
  {"h": "航", "p": "háng", "sp": "hang", "r": "舟", "sc": 10, "l": 4, "f": 773},
  {"h": "优", "p": "yōu", "sp": "you", "r": "人", "sc": 6, "l": 3, "f": 774},
  {"h": "怪", "p": "guài", "sp": "guai", "r": "心", "sc": 8, "l": 3, "f": 775},
  {"h": "香", "p": "xiāng", "sp": "xiang", "r": "香", "sc": 9, "l": 3, "f": 776},
  {"h": "著", "p": "zhe", "sp": "zhe", "r": "艸", "sc": 11, "l": 4, "f": 777},
  {"h": "田", "p": "tián", "sp": "tian", "r": "田", "sc": 5, "l": 5, "f": 778},
  {"h": "铁", "p": "tiě", "sp": "tie", "r": "金", "sc": 10, "l": 2, "f": 779},
  {"h": "控", "p": "kòng", "sp": "kong", "r": "手", "sc": 11, "l": 5, "f": 780},
  {"h": "税", "p": "shuì", "sp": "shui", "r": "禾", "sc": 12, "l": 5, "f": 781},
  {"h": "左", "p": "zuǒ", "sp": "zuo", "r": "工", "sc": 5, "l": 1, "f": 782},
  {"h": "右", "p": "yòu", "sp": "you", "r": "口", "sc": 5, "l": 1, "f": 783},
  {"h": "份", "p": "fèn", "sp": "fen", "r": "人", "sc": 6, "l": 2, "f": 784},
  {"h": "穿", "p": "chuān", "sp": "chuan", "r": "穴", "sc": 9, "l": 1, "f": 785},
  {"h": "艺", "p": "yì", "sp": "yi", "r": "艸", "sc": 4, "l": 3, "f": 786},
  {"h": "背", "p": "bèi", "sp": "bei", "r": "肉", "sc": 9, "l": 2, "f": 787},
  {"h": "阵", "p": "zhèn", "sp": "zhen", "r": "阜", "sc": 6, "l": 4, "f": 788},
  {"h": "草", "p": "cǎo", "sp": "cao", "r": "艸", "sc": 9, "l": 2, "f": 789},
  {"h": "脚", "p": "jiǎo", "sp": "jiao", "r": "肉", "sc": 11, "l": 2, "f": 790},
  {"h": "概", "p": "gài", "sp": "gai", "r": "木", "sc": 13, "l": 3, "f": 791},
  {"h": "恶", "p": "è", "sp": "e", "r": "心", "sc": 10, "l": 4, "f": 792},
  {"h": "块", "p": "kuài", "sp": "kuai", "r": "土", "sc": 7, "l": 1, "f": 793},
  {"h": "顿", "p": "dùn", "sp": "dun", "r": "頁", "sc": 10, "l": 3, "f": 794},
  {"h": "敢", "p": "gǎn", "sp": "gan", "r": "攴", "sc": 11, "l": 3, "f": 795},
  {"h": "守", "p": "shǒu", "sp": "shou", "r": "宀", "sc": 6, "l": 4, "f": 796},
  {"h": "酒", "p": "jiǔ", "sp": "jiu", "r": "酉", "sc": 10, "l": 2, "f": 797},
  {"h": "岛", "p": "dǎo", "sp": "dao", "r": "山", "sc": 7, "l": 4, "f": 798},
  {"h": "托", "p": "tuō", "sp": "tuo", "r": "手", "sc": 6, "l": 4, "f": 799},
  {"h": "央", "p": "yāng", "sp": "yang", "r": "大", "sc": 5, "l": 5, "f": 800},
  {"h": "户", "p": "hù", "sp": "hu", "r": "戶", "sc": 4, "l": 4, "f": 801},
  {"h": "烈", "p": "liè", "sp": "lie", "r": "火", "sc": 10, "l": 3, "f": 802},
  {"h": "洋", "p": "yáng", "sp": "yang", "r": "水", "sc": 9, "l": 4, "f": 803},
  {"h": "哥", "p": "gē", "sp": "ge", "r": "口", "sc": 10, "l": 1, "f": 804},
  {"h": "索", "p": "suǒ", "sp": "suo", "r": "糸", "sc": 10, "l": 5, "f": 805},
  {"h": "胡", "p": "hú", "sp": "hu", "r": "肉", "sc": 9, "l": 5, "f": 806},
  {"h": "款", "p": "kuǎn", "sp": "kuan", "r": "欠", "sc": 12, "l": 5, "f": 807},
  {"h": "靠", "p": "kào", "sp": "kao", "r": "非", "sc": 15, "l": 2, "f": 808},
  {"h": "评", "p": "píng", "sp": "ping", "r": "言", "sc": 7, "l": 3, "f": 809},
  {"h": "版", "p": "bǎn", "sp": "ban", "r": "片", "sc": 8, "l": 5, "f": 810},
  {"h": "宝", "p": "bǎo", "sp": "bao", "r": "宀", "sc": 8, "l": 4, "f": 811},
  {"h": "座", "p": "zuò", "sp": "zuo", "r": "广", "sc": 10, "l": 2, "f": 812},
  {"h": "释", "p": "shì", "sp": "shi", "r": "釆", "sc": 12, "l": 4, "f": 813},
  {"h": "景", "p": "jǐng", "sp": "jing", "r": "日", "sc": 12, "l": 3, "f": 814},
  {"h": "顾", "p": "gù", "sp": "gu", "r": "頁", "sc": 10, "l": 2, "f": 815},
  {"h": "弟", "p": "dì", "sp": "di", "r": "弓", "sc": 7, "l": 1, "f": 816},
  {"h": "登", "p": "dēng", "sp": "deng", "r": "癶", "sc": 12, "l": 4, "f": 817},
  {"h": "货", "p": "huò", "sp": "huo", "r": "貝", "sc": 8, "l": 4, "f": 818},
  {"h": "互", "p": "hù", "sp": "hu", "r": "二", "sc": 4, "l": 3, "f": 819},
  {"h": "付", "p": "fù", "sp": "fu", "r": "人", "sc": 5, "l": 3, "f": 820},
  {"h": "伯", "p": "bó", "sp": "bo", "r": "人", "sc": 7, "l": 5, "f": 821},
  {"h": "慢", "p": "màn", "sp": "man", "r": "心", "sc": 14, "l": 1, "f": 822},
  {"h": "欧", "p": "ōu", "sp": "ou", "r": "欠", "sc": 8, "l": 5, "f": 823},
  {"h": "换", "p": "huàn", "sp": "huan", "r": "手", "sc": 10, "l": 2, "f": 824},
  {"h": "闻", "p": "wén", "sp": "wen", "r": "門", "sc": 9, "l": 2, "f": 825},
  {"h": "危", "p": "wēi", "sp": "wei", "r": "卩", "sc": 6, "l": 3, "f": 826},
  {"h": "忙", "p": "máng", "sp": "mang", "r": "心", "sc": 6, "l": 1, "f": 827},
  {"h": "核", "p": "hé", "sp": "he", "r": "木", "sc": 10, "l": 5, "f": 828},
  {"h": "暗", "p": "àn", "sp": "an", "r": "日", "sc": 13, "l": 4, "f": 829},
  {"h": "姐", "p": "jiě", "sp": "jie", "r": "女", "sc": 8, "l": 1, "f": 830},
  {"h": "介", "p": "jiè", "sp": "jie", "r": "人", "sc": 4, "l": 1, "f": 831},
  {"h": "坏", "p": "huài", "sp": "huai", "r": "土", "sc": 7, "l": 1, "f": 832},
  {"h": "讨", "p": "tǎo", "sp": "tao", "r": "言", "sc": 5, "l": 2, "f": 833},
  {"h": "丽", "p": "lì", "sp": "li", "r": "丶", "sc": 7, "l": 3, "f": 834},
  {"h": "良", "p": "liáng", "sp": "liang", "r": "艮", "sc": 7, "l": 4, "f": 835},
  {"h": "序", "p": "xù", "sp": "xu", "r": "广", "sc": 7, "l": 4, "f": 836},
  {"h": "升", "p": "shēng", "sp": "sheng", "r": "十", "sc": 4, "l": 3, "f": 837},
  {"h": "监", "p": "jiān", "sp": "jian", "r": "皿", "sc": 10, "l": 5, "f": 838},
  {"h": "临", "p": "lín", "sp": "lin", "r": "丨", "sc": 9, "l": 4, "f": 839},
  {"h": "亮", "p": "liàng", "sp": "liang", "r": "亠", "sc": 9, "l": 1, "f": 840},
  {"h": "露", "p": "lù", "sp": "lu", "r": "雨", "sc": 21, "l": 5, "f": 841},
  {"h": "永", "p": "yǒng", "sp": "yong", "r": "水", "sc": 5, "l": 2, "f": 842},
  {"h": "呼", "p": "hū", "sp": "hu", "r": "口", "sc": 8, "l": 4, "f": 843},
  {"h": "味", "p": "wèi", "sp": "wei", "r": "口", "sc": 8, "l": 2, "f": 844},
  {"h": "野", "p": "yě", "sp": "ye", "r": "里", "sc": 11, "l": 4, "f": 845},
  {"h": "架", "p": "jià", "sp": "jia", "r": "木", "sc": 9, "l": 3, "f": 846},
  {"h": "域", "p": "yù", "sp": "yu", "r": "土", "sc": 11, "l": 5, "f": 847},
  {"h": "沙", "p": "shā", "sp": "sha", "r": "水", "sc": 7, "l": 3, "f": 848},
  {"h": "掉", "p": "diào", "sp": "diao", "r": "手", "sc": 11, "l": 2, "f": 849},
  {"h": "括", "p": "kuò", "sp": "kuo", "r": "手", "sc": 9, "l": 4, "f": 850},
  {"h": "舰", "p": "jiàn", "sp": "jian", "r": "舟", "sc": 10, "l": 5, "f": 851},
  {"h": "鱼", "p": "yú", "sp": "yu", "r": "魚", "sc": 8, "l": 2, "f": 852},
  {"h": "杂", "p": "zá", "sp": "za", "r": "木", "sc": 6, "l": 3, "f": 853},
  {"h": "误", "p": "wù", "sp": "wu", "r": "言", "sc": 9, "l": 3, "f": 854},
  {"h": "湾", "p": "wān", "sp": "wan", "r": "水", "sc": 12, "l": 3, "f": 855},
  {"h": "吉", "p": "jí", "sp": "ji", "r": "口", "sc": 6, "l": 5, "f": 856},
  {"h": "减", "p": "jiǎn", "sp": "jian", "r": "冫", "sc": 11, "l": 4, "f": 857},
  {"h": "编", "p": "biān", "sp": "bian", "r": "糸", "sc": 12, "l": 4, "f": 858},
  {"h": "楚", "p": "chǔ", "sp": "chu", "r": "木", "sc": 13, "l": 2, "f": 859},
  {"h": "肯", "p": "kěn", "sp": "ken", "r": "肉", "sc": 8, "l": 4, "f": 860},
  {"h": "测", "p": "cè", "sp": "ce", "r": "水", "sc": 9, "l": 4, "f": 861},
  {"h": "败", "p": "bài", "sp": "bai", "r": "貝", "sc": 8, "l": 4, "f": 862},
  {"h": "屋", "p": "wū", "sp": "wu", "r": "尸", "sc": 9, "l": 3, "f": 863},
  {"h": "跑", "p": "pǎo", "sp": "pao", "r": "足", "sc": 12, "l": 1, "f": 864},
  {"h": "梦", "p": "mèng", "sp": "meng", "r": "木", "sc": 11, "l": 3, "f": 865},
  {"h": "散", "p": "sàn", "sp": "san", "r": "攴", "sc": 12, "l": 3, "f": 866},
  {"h": "温", "p": "wēn", "sp": "wen", "r": "水", "sc": 12, "l": 2, "f": 867},
  {"h": "困", "p": "kùn", "sp": "kun", "r": "囗", "sc": 7, "l": 3, "f": 868},
  {"h": "剑", "p": "jiàn", "sp": "jian", "r": "刀", "sc": 9, "l": 5, "f": 869},
  {"h": "渐", "p": "jiàn", "sp": "jian", "r": "水", "sc": 11, "l": 4, "f": 870},
  {"h": "封", "p": "fēng", "sp": "feng", "r": "寸", "sc": 9, "l": 3, "f": 871},
  {"h": "救", "p": "jiù", "sp": "jiu", "r": "攴", "sc": 11, "l": 3, "f": 872},
  {"h": "贵", "p": "guì", "sp": "gui", "r": "貝", "sc": 9, "l": 1, "f": 873},
  {"h": "枪", "p": "qiāng", "sp": "qiang", "r": "木", "sc": 8, "l": 5, "f": 874},
  {"h": "缺", "p": "quē", "sp": "que", "r": "缶", "sc": 10, "l": 3, "f": 875},
  {"h": "楼", "p": "lóu", "sp": "lou", "r": "木", "sc": 13, "l": 1, "f": 876},
  {"h": "县", "p": "xiàn", "sp": "xian", "r": "厶", "sc": 7, "l": 4, "f": 877},
  {"h": "尚", "p": "shàng", "sp": "shang", "r": "小", "sc": 8, "l": 4, "f": 878},
  {"h": "毫", "p": "háo", "sp": "hao", "r": "毛", "sc": 11, "l": 4, "f": 879},
  {"h": "移", "p": "yí", "sp": "yi", "r": "禾", "sc": 11, "l": 4, "f": 880},
  {"h": "娘", "p": "niáng", "sp": "niang", "r": "女", "sc": 10, "l": 3, "f": 881},
  {"h": "朋", "p": "péng", "sp": "peng", "r": "月", "sc": 8, "l": 1, "f": 882},
  {"h": "画", "p": "huà", "sp": "hua", "r": "田", "sc": 8, "l": 2, "f": 883},
  {"h": "班", "p": "bān", "sp": "ban", "r": "玉", "sc": 10, "l": 1, "f": 884},
  {"h": "智", "p": "zhì", "sp": "zhi", "r": "日", "sc": 12, "l": 4, "f": 885},
  {"h": "亦", "p": "yì", "sp": "yi", "r": "亠", "sc": 6, "l": 5, "f": 886},
  {"h": "耳", "p": "ěr", "sp": "er", "r": "耳", "sc": 6, "l": 3, "f": 887},
  {"h": "恩", "p": "ēn", "sp": "en", "r": "心", "sc": 10, "l": 5, "f": 888},
  {"h": "短", "p": "duǎn", "sp": "duan", "r": "矢", "sc": 12, "l": 2, "f": 889},
  {"h": "掌", "p": "zhǎng", "sp": "zhang", "r": "手", "sc": 12, "l": 5, "f": 890},
  {"h": "恐", "p": "kǒng", "sp": "kong", "r": "心", "sc": 10, "l": 3, "f": 891},
  {"h": "遗", "p": "yí", "sp": "yi", "r": "辵", "sc": 12, "l": 4, "f": 892},
  {"h": "固", "p": "gù", "sp": "gu", "r": "囗", "sc": 8, "l": 4, "f": 893},
  {"h": "席", "p": "xí", "sp": "xi", "r": "巾", "sc": 10, "l": 4, "f": 894},
  {"h": "松", "p": "sōng", "sp": "song", "r": "木", "sc": 8, "l": 4, "f": 895},
  {"h": "秘", "p": "mì", "sp": "mi", "r": "禾", "sc": 10, "l": 4, "f": 896},
  {"h": "谢", "p": "xiè", "sp": "xie", "r": "言", "sc": 12, "l": 1, "f": 897},
  {"h": "鲁", "p": "lǔ", "sp": "lu", "r": "魚", "sc": 12, "l": 5, "f": 898},
  {"h": "遇", "p": "yù", "sp": "yu", "r": "辵", "sc": 12, "l": 4, "f": 899},
  {"h": "康", "p": "kāng", "sp": "kang", "r": "广", "sc": 11, "l": 2, "f": 900},
  {"h": "虑", "p": "lǜ", "sp": "lu", "r": "心", "sc": 10, "l": 4, "f": 901},
  {"h": "幸", "p": "xìng", "sp": "xing", "r": "干", "sc": 8, "l": 3, "f": 902},
  {"h": "均", "p": "jūn", "sp": "jun", "r": "土", "sc": 7, "l": 4, "f": 903},
  {"h": "销", "p": "xiāo", "sp": "xiao", "r": "金", "sc": 12, "l": 4, "f": 904},
  {"h": "钟", "p": "zhōng", "sp": "zhong", "r": "金", "sc": 9, "l": 1, "f": 905},
  {"h": "诗", "p": "shī", "sp": "shi", "r": "言", "sc": 8, "l": 4, "f": 906},
  {"h": "藏", "p": "cáng", "sp": "cang", "r": "艸", "sc": 17, "l": 5, "f": 907},
  {"h": "赶", "p": "gǎn", "sp": "gan", "r": "走", "sc": 10, "l": 3, "f": 908},
  {"h": "剧", "p": "jù", "sp": "ju", "r": "刀", "sc": 10, "l": 3, "f": 909},
  {"h": "票", "p": "piào", "sp": "piao", "r": "示", "sc": 11, "l": 1, "f": 910},
  {"h": "损", "p": "sǔn", "sp": "sun", "r": "手", "sc": 10, "l": 5, "f": 911},
  {"h": "忽", "p": "hū", "sp": "hu", "r": "心", "sc": 8, "l": 2, "f": 912},
  {"h": "巨", "p": "jù", "sp": "ju", "r": "工", "sc": 4, "l": 4, "f": 913},
  {"h": "炮", "p": "pào", "sp": "pao", "r": "火", "sc": 9, "l": 5, "f": 914},
  {"h": "旧", "p": "jiù", "sp": "jiu", "r": "日", "sc": 5, "l": 3, "f": 915},
  {"h": "端", "p": "duān", "sp": "duan", "r": "立", "sc": 14, "l": 5, "f": 916},
  {"h": "探", "p": "tàn", "sp": "tan", "r": "手", "sc": 11, "l": 5, "f": 917},
  {"h": "湖", "p": "hú", "sp": "hu", "r": "水", "sc": 12, "l": 2, "f": 918},
  {"h": "录", "p": "lù", "sp": "lu", "r": "彐", "sc": 8, "l": 3, "f": 919},
  {"h": "叶", "p": "yè", "sp": "ye", "r": "口", "sc": 5, "l": 4, "f": 920},
  {"h": "春", "p": "chūn", "sp": "chun", "r": "日", "sc": 9, "l": 2, "f": 921},
  {"h": "乡", "p": "xiāng", "sp": "xiang", "r": "幺", "sc": 3, "l": 3, "f": 922},
  {"h": "附", "p": "fù", "sp": "fu", "r": "阜", "sc": 7, "l": 3, "f": 923},
  {"h": "吸", "p": "xī", "sp": "xi", "r": "口", "sc": 6, "l": 4, "f": 924},
  {"h": "予", "p": "yǔ", "sp": "yu", "r": "亅", "sc": 4, "l": 5, "f": 925},
  {"h": "礼", "p": "lǐ", "sp": "li", "r": "示", "sc": 5, "l": 2, "f": 926},
  {"h": "港", "p": "gǎng", "sp": "gang", "r": "水", "sc": 12, "l": 5, "f": 927},
  {"h": "雨", "p": "yǔ", "sp": "yu", "r": "雨", "sc": 8, "l": 1, "f": 928},
  {"h": "呀", "p": "ya", "sp": "ya", "r": "口", "sc": 7, "l": 4, "f": 929},
  {"h": "板", "p": "bǎn", "sp": "ban", "r": "木", "sc": 8, "l": 2, "f": 930},
  {"h": "庭", "p": "tíng", "sp": "ting", "r": "广", "sc": 9, "l": 1, "f": 931},
  {"h": "妇", "p": "fù", "sp": "fu", "r": "女", "sc": 6, "l": 4, "f": 932},
  {"h": "归", "p": "guī", "sp": "gui", "r": "彐", "sc": 5, "l": 4, "f": 933},
  {"h": "睛", "p": "jīng", "sp": "jing", "r": "目", "sc": 13, "l": 2, "f": 934},
  {"h": "饭", "p": "fàn", "sp": "fan", "r": "食", "sc": 7, "l": 1, "f": 935},
  {"h": "额", "p": "é", "sp": "e", "r": "頁", "sc": 15, "l": 5, "f": 936},
  {"h": "含", "p": "hán", "sp": "han", "r": "口", "sc": 7, "l": 4, "f": 937},
  {"h": "顺", "p": "shùn", "sp": "shun", "r": "頁", "sc": 9, "l": 2, "f": 938},
  {"h": "输", "p": "shū", "sp": "shu", "r": "車", "sc": 13, "l": 3, "f": 939},
  {"h": "摇", "p": "yáo", "sp": "yao", "r": "手", "sc": 13, "l": 4, "f": 940},
  {"h": "招", "p": "zhāo", "sp": "zhao", "r": "手", "sc": 8, "l": 4, "f": 941},
  {"h": "婚", "p": "hūn", "sp": "hun", "r": "女", "sc": 11, "l": 3, "f": 942},
  {"h": "脱", "p": "tuō", "sp": "tuo", "r": "肉", "sc": 11, "l": 4, "f": 943},
  {"h": "补", "p": "bǔ", "sp": "bu", "r": "衣", "sc": 7, "l": 3, "f": 944},
  {"h": "谓", "p": "wèi", "sp": "wei", "r": "言", "sc": 11, "l": 4, "f": 945},
  {"h": "督", "p": "dū", "sp": "du", "r": "目", "sc": 13, "l": 5, "f": 946},
  {"h": "毒", "p": "dú", "sp": "du", "r": "毋", "sc": 9, "l": 5, "f": 947},
  {"h": "油", "p": "yóu", "sp": "you", "r": "水", "sc": 8, "l": 2, "f": 948},
  {"h": "疗", "p": "liáo", "sp": "liao", "r": "疒", "sc": 7, "l": 4, "f": 949},
  {"h": "旅", "p": "lǚ", "sp": "lu", "r": "方", "sc": 10, "l": 2, "f": 950},
  {"h": "泽", "p": "zé", "sp": "ze", "r": "水", "sc": 8, "l": 5, "f": 951},
  {"h": "材", "p": "cái", "sp": "cai", "r": "木", "sc": 7, "l": 3, "f": 952},
  {"h": "灭", "p": "miè", "sp": "mie", "r": "火", "sc": 5, "l": 5, "f": 953},
  {"h": "逐", "p": "zhú", "sp": "zhu", "r": "辵", "sc": 10, "l": 4, "f": 954},
  {"h": "莫", "p": "mò", "sp": "mo", "r": "艸", "sc": 10, "l": 5, "f": 955},
  {"h": "笔", "p": "bǐ", "sp": "bi", "r": "竹", "sc": 10, "l": 2, "f": 956},
  {"h": "亡", "p": "wáng", "sp": "wang", "r": "亠", "sc": 3, "l": 5, "f": 957},
  {"h": "鲜", "p": "xiān", "sp": "xian", "r": "魚", "sc": 14, "l": 4, "f": 958},
  {"h": "词", "p": "cí", "sp": "ci", "r": "言", "sc": 7, "l": 2, "f": 959},
  {"h": "圣", "p": "shèng", "sp": "sheng", "r": "土", "sc": 5, "l": 3, "f": 960},
  {"h": "择", "p": "zé", "sp": "ze", "r": "手", "sc": 8, "l": 4, "f": 961},
  {"h": "寻", "p": "xún", "sp": "xun", "r": "彐", "sc": 6, "l": 4, "f": 962},
  {"h": "厂", "p": "chǎng", "sp": "chang", "r": "厂", "sc": 2, "l": 3, "f": 963},
  {"h": "睡", "p": "shuì", "sp": "shui", "r": "目", "sc": 13, "l": 1, "f": 964},
  {"h": "博", "p": "bó", "sp": "bo", "r": "十", "sc": 12, "l": 5, "f": 965},
  {"h": "勒", "p": "lēi", "sp": "lei", "r": "力", "sc": 11, "l": 5, "f": 966},
  {"h": "烟", "p": "yān", "sp": "yan", "r": "火", "sc": 10, "l": 3, "f": 967},
  {"h": "授", "p": "shòu", "sp": "shou", "r": "手", "sc": 11, "l": 4, "f": 968},
  {"h": "诺", "p": "nuò", "sp": "nuo", "r": "言", "sc": 10, "l": 5, "f": 969},
  {"h": "伦", "p": "lún", "sp": "lun", "r": "人", "sc": 6, "l": 5, "f": 970},
  {"h": "岸", "p": "àn", "sp": "an", "r": "山", "sc": 8, "l": 5, "f": 971},
  {"h": "奥", "p": "ào", "sp": "ao", "r": "大", "sc": 12, "l": 5, "f": 972},
  {"h": "唐", "p": "táng", "sp": "tang", "r": "口", "sc": 10, "l": 5, "f": 973},
  {"h": "卖", "p": "mài", "sp": "mai", "r": "十", "sc": 8, "l": 2, "f": 974},
  {"h": "俄", "p": "é", "sp": "e", "r": "人", "sc": 9, "l": 5, "f": 975},
  {"h": "炸", "p": "zhà", "sp": "zha", "r": "火", "sc": 9, "l": 5, "f": 976},
  {"h": "载", "p": "zài", "sp": "zai", "r": "戈", "sc": 10, "l": 4, "f": 977},
  {"h": "洛", "p": "luò", "sp": "luo", "r": "水", "sc": 9, "l": 5, "f": 978},
  {"h": "健", "p": "jiàn", "sp": "jian", "r": "人", "sc": 10, "l": 2, "f": 979},
  {"h": "堂", "p": "táng", "sp": "tang", "r": "土", "sc": 11, "l": 2, "f": 980},
  {"h": "旁", "p": "páng", "sp": "pang", "r": "方", "sc": 10, "l": 1, "f": 981},
  {"h": "宫", "p": "gōng", "sp": "gong", "r": "宀", "sc": 9, "l": 4, "f": 982},
  {"h": "喝", "p": "hē", "sp": "he", "r": "口", "sc": 12, "l": 1, "f": 983},
  {"h": "借", "p": "jiè", "sp": "jie", "r": "人", "sc": 10, "l": 2, "f": 984},
  {"h": "君", "p": "jūn", "sp": "jun", "r": "口", "sc": 7, "l": 5, "f": 985},
  {"h": "禁", "p": "jìn", "sp": "jin", "r": "示", "sc": 13, "l": 4, "f": 986},
  {"h": "阴", "p": "yīn", "sp": "yin", "r": "阜", "sc": 6, "l": 2, "f": 987},
  {"h": "园", "p": "yuán", "sp": "yuan", "r": "囗", "sc": 7, "l": 2, "f": 988},
  {"h": "谋", "p": "móu", "sp": "mou", "r": "言", "sc": 11, "l": 5, "f": 989},
  {"h": "宋", "p": "sòng", "sp": "song", "r": "宀", "sc": 7, "l": 5, "f": 990},
  {"h": "避", "p": "bì", "sp": "bi", "r": "辵", "sc": 16, "l": 4, "f": 991},
  {"h": "抓", "p": "zhuā", "sp": "zhua", "r": "手", "sc": 7, "l": 3, "f": 992},
  {"h": "荣", "p": "róng", "sp": "rong", "r": "艸", "sc": 9, "l": 5, "f": 993},
  {"h": "姑", "p": "gū", "sp": "gu", "r": "女", "sc": 8, "l": 3, "f": 994},
  {"h": "孙", "p": "sūn", "sp": "sun", "r": "子", "sc": 6, "l": 4, "f": 995},
  {"h": "逃", "p": "táo", "sp": "tao", "r": "辵", "sc": 9, "l": 5, "f": 996},
  {"h": "牙", "p": "yá", "sp": "ya", "r": "牙", "sc": 4, "l": 2, "f": 997},
  {"h": "束", "p": "shù", "sp": "shu", "r": "木", "sc": 7, "l": 3, "f": 998},
  {"h": "跳", "p": "tiào", "sp": "tiao", "r": "足", "sc": 13, "l": 2, "f": 999},
  {"h": "顶", "p": "dǐng", "sp": "ding", "r": "頁", "sc": 8, "l": 4, "f": 1000},
  {"h": "玉", "p": "yù", "sp": "yu", "r": "玉", "sc": 5, "l": 4, "f": 1001},
  {"h": "镇", "p": "zhèn", "sp": "zhen", "r": "金", "sc": 15, "l": 5, "f": 1002},
  {"h": "雪", "p": "xuě", "sp": "xue", "r": "雨", "sc": 11, "l": 2, "f": 1003},
  {"h": "午", "p": "wǔ", "sp": "wu", "r": "十", "sc": 4, "l": 1, "f": 1004},
  {"h": "练", "p": "liàn", "sp": "lian", "r": "糸", "sc": 8, "l": 2, "f": 1005},
  {"h": "迫", "p": "pò", "sp": "po", "r": "辵", "sc": 8, "l": 4, "f": 1006},
  {"h": "爷", "p": "yé", "sp": "ye", "r": "父", "sc": 6, "l": 1, "f": 1007},
  {"h": "篇", "p": "piān", "sp": "pian", "r": "竹", "sc": 15, "l": 2, "f": 1008},
  {"h": "肉", "p": "ròu", "sp": "rou", "r": "肉", "sc": 6, "l": 1, "f": 1009},
  {"h": "嘴", "p": "zuǐ", "sp": "zui", "r": "口", "sc": 16, "l": 2, "f": 1010},
  {"h": "馆", "p": "guǎn", "sp": "guan", "r": "食", "sc": 11, "l": 1, "f": 1011},
  {"h": "遍", "p": "biàn", "sp": "bian", "r": "辵", "sc": 12, "l": 2, "f": 1012},
  {"h": "凡", "p": "fán", "sp": "fan", "r": "几", "sc": 3, "l": 5, "f": 1013},
  {"h": "础", "p": "chǔ", "sp": "chu", "r": "石", "sc": 10, "l": 3, "f": 1014},
  {"h": "洞", "p": "dòng", "sp": "dong", "r": "水", "sc": 9, "l": 5, "f": 1015},
  {"h": "卷", "p": "juǎn", "sp": "juan", "r": "卩", "sc": 8, "l": 4, "f": 1016},
  {"h": "坦", "p": "tǎn", "sp": "tan", "r": "土", "sc": 8, "l": 5, "f": 1017},
  {"h": "牛", "p": "niú", "sp": "niu", "r": "牛", "sc": 4, "l": 1, "f": 1018},
  {"h": "宁", "p": "níng", "sp": "ning", "r": "宀", "sc": 5, "l": 4, "f": 1019},
  {"h": "纸", "p": "zhǐ", "sp": "zhi", "r": "糸", "sc": 7, "l": 2, "f": 1020},
  {"h": "诸", "p": "zhū", "sp": "zhu", "r": "言", "sc": 10, "l": 5, "f": 1021},
  {"h": "训", "p": "xùn", "sp": "xun", "r": "言", "sc": 5, "l": 3, "f": 1022},
  {"h": "私", "p": "sī", "sp": "si", "r": "禾", "sc": 7, "l": 5, "f": 1023},
  {
    "h": "庄",
    "p": "zhuāng",
    "sp": "zhuang",
    "r": "广",
    "sc": 6,
    "l": 5,
    "f": 1024
  },
  {"h": "祖", "p": "zǔ", "sp": "zu", "r": "示", "sc": 9, "l": 5, "f": 1025},
  {"h": "丝", "p": "sī", "sp": "si", "r": "一", "sc": 5, "l": 5, "f": 1026},
  {"h": "翻", "p": "fān", "sp": "fan", "r": "羽", "sc": 18, "l": 4, "f": 1027},
  {"h": "暴", "p": "bào", "sp": "bao", "r": "日", "sc": 15, "l": 5, "f": 1028},
  {"h": "森", "p": "sēn", "sp": "sen", "r": "木", "sc": 12, "l": 4, "f": 1029},
  {"h": "塔", "p": "tǎ", "sp": "ta", "r": "土", "sc": 12, "l": 5, "f": 1030},
  {"h": "默", "p": "mò", "sp": "mo", "r": "黑", "sc": 16, "l": 4, "f": 1031},
  {"h": "握", "p": "wò", "sp": "wo", "r": "手", "sc": 12, "l": 3, "f": 1032},
  {"h": "戏", "p": "xì", "sp": "xi", "r": "戈", "sc": 6, "l": 3, "f": 1033},
  {"h": "隐", "p": "yǐn", "sp": "yin", "r": "阜", "sc": 11, "l": 5, "f": 1034},
  {"h": "熟", "p": "shú", "sp": "shu", "r": "火", "sc": 15, "l": 2, "f": 1035},
  {"h": "骨", "p": "gǔ", "sp": "gu", "r": "骨", "sc": 9, "l": 4, "f": 1036},
  {"h": "访", "p": "fǎng", "sp": "fang", "r": "言", "sc": 6, "l": 3, "f": 1037},
  {"h": "弱", "p": "ruò", "sp": "ruo", "r": "弓", "sc": 10, "l": 4, "f": 1038},
  {"h": "蒙", "p": "méng", "sp": "meng", "r": "艸", "sc": 13, "l": 5, "f": 1039},
  {"h": "歌", "p": "gē", "sp": "ge", "r": "欠", "sc": 14, "l": 1, "f": 1040},
  {"h": "店", "p": "diàn", "sp": "dian", "r": "广", "sc": 8, "l": 1, "f": 1041},
  {"h": "鬼", "p": "guǐ", "sp": "gui", "r": "鬼", "sc": 9, "l": 5, "f": 1042},
  {"h": "软", "p": "ruǎn", "sp": "ruan", "r": "車", "sc": 8, "l": 5, "f": 1043},
  {"h": "典", "p": "diǎn", "sp": "dian", "r": "八", "sc": 8, "l": 2, "f": 1044},
  {"h": "欲", "p": "yù", "sp": "yu", "r": "欠", "sc": 11, "l": 5, "f": 1045},
  {"h": "萨", "p": "sà", "sp": "sa", "r": "艸", "sc": 11, "l": 5, "f": 1046},
  {"h": "伙", "p": "huǒ", "sp": "huo", "r": "人", "sc": 6, "l": 4, "f": 1047},
  {"h": "遭", "p": "zāo", "sp": "zao", "r": "辵", "sc": 14, "l": 5, "f": 1048},
  {"h": "盘", "p": "pán", "sp": "pan", "r": "皿", "sc": 11, "l": 3, "f": 1049},
  {"h": "爸", "p": "bà", "sp": "ba", "r": "父", "sc": 8, "l": 1, "f": 1050},
  {"h": "扩", "p": "kuò", "sp": "kuo", "r": "手", "sc": 6, "l": 4, "f": 1051},
  {"h": "盖", "p": "gài", "sp": "gai", "r": "皿", "sc": 11, "l": 4, "f": 1052},
  {"h": "弄", "p": "nòng", "sp": "nong", "r": "廾", "sc": 7, "l": 2, "f": 1053},
  {
    "h": "雄",
    "p": "xióng",
    "sp": "xiong",
    "r": "隹",
    "sc": 12,
    "l": 5,
    "f": 1054
  },
  {"h": "稳", "p": "wěn", "sp": "wen", "r": "禾", "sc": 14, "l": 4, "f": 1055},
  {"h": "忘", "p": "wàng", "sp": "wang", "r": "心", "sc": 7, "l": 1, "f": 1056},
  {"h": "亿", "p": "yì", "sp": "yi", "r": "人", "sc": 3, "l": 2, "f": 1057},
  {"h": "刺", "p": "cì", "sp": "ci", "r": "刀", "sc": 8, "l": 4, "f": 1058},
  {"h": "拥", "p": "yōng", "sp": "yong", "r": "手", "sc": 8, "l": 5, "f": 1059},
  {"h": "徒", "p": "tú", "sp": "tu", "r": "彳", "sc": 10, "l": 5, "f": 1060},
  {"h": "姆", "p": "mǔ", "sp": "mu", "r": "女", "sc": 8, "l": 4, "f": 1061},
  {"h": "杨", "p": "yáng", "sp": "yang", "r": "木", "sc": 7, "l": 5, "f": 1062},
  {"h": "齐", "p": "qí", "sp": "qi", "r": "齊", "sc": 6, "l": 3, "f": 1063},
  {"h": "赛", "p": "sài", "sp": "sai", "r": "宀", "sc": 14, "l": 3, "f": 1064},
  {"h": "趣", "p": "qù", "sp": "qu", "r": "走", "sc": 15, "l": 3, "f": 1065},
  {"h": "曲", "p": "qū", "sp": "qu", "r": "曰", "sc": 6, "l": 5, "f": 1066},
  {"h": "刀", "p": "dāo", "sp": "dao", "r": "刀", "sc": 2, "l": 3, "f": 1067},
  {
    "h": "床",
    "p": "chuáng",
    "sp": "chuang",
    "r": "广",
    "sc": 7,
    "l": 1,
    "f": 1068
  },
  {"h": "迎", "p": "yíng", "sp": "ying", "r": "辵", "sc": 7, "l": 2, "f": 1069},
  {"h": "冰", "p": "bīng", "sp": "bing", "r": "冫", "sc": 6, "l": 3, "f": 1070},
  {"h": "虚", "p": "xū", "sp": "xu", "r": "虍", "sc": 11, "l": 5, "f": 1071},
  {"h": "玩", "p": "wán", "sp": "wan", "r": "玉", "sc": 8, "l": 1, "f": 1072},
  {"h": "析", "p": "xī", "sp": "xi", "r": "木", "sc": 8, "l": 5, "f": 1073},
  {
    "h": "窗",
    "p": "chuāng",
    "sp": "chuang",
    "r": "穴",
    "sc": 12,
    "l": 4,
    "f": 1074
  },
  {"h": "醒", "p": "xǐng", "sp": "xing", "r": "酉", "sc": 16, "l": 4, "f": 1075},
  {"h": "妻", "p": "qī", "sp": "qi", "r": "女", "sc": 8, "l": 4, "f": 1076},
  {"h": "透", "p": "tòu", "sp": "tou", "r": "辵", "sc": 10, "l": 4, "f": 1077},
  {"h": "购", "p": "gòu", "sp": "gou", "r": "貝", "sc": 8, "l": 4, "f": 1078},
  {"h": "替", "p": "tì", "sp": "ti", "r": "曰", "sc": 12, "l": 4, "f": 1079},
  {"h": "塞", "p": "sāi", "sp": "sai", "r": "土", "sc": 13, "l": 5, "f": 1080},
  {"h": "努", "p": "nǔ", "sp": "nu", "r": "力", "sc": 7, "l": 2, "f": 1081},
  {"h": "休", "p": "xiū", "sp": "xiu", "r": "人", "sc": 6, "l": 1, "f": 1082},
  {"h": "虎", "p": "hǔ", "sp": "hu", "r": "虍", "sc": 8, "l": 4, "f": 1083},
  {"h": "扬", "p": "yáng", "sp": "yang", "r": "手", "sc": 6, "l": 4, "f": 1084},
  {"h": "途", "p": "tú", "sp": "tu", "r": "辵", "sc": 10, "l": 4, "f": 1085},
  {"h": "侵", "p": "qīn", "sp": "qin", "r": "人", "sc": 9, "l": 5, "f": 1086},
  {"h": "刑", "p": "xíng", "sp": "xing", "r": "刀", "sc": 6, "l": 5, "f": 1087},
  {"h": "绿", "p": "lǜ", "sp": "lu", "r": "糸", "sc": 11, "l": 2, "f": 1088},
  {"h": "兄", "p": "xiōng", "sp": "xiong", "r": "儿", "sc": 5, "l": 4, "f": 1089},
  {"h": "迅", "p": "xùn", "sp": "xun", "r": "辵", "sc": 6, "l": 4, "f": 1090},
  {"h": "套", "p": "tào", "sp": "tao", "r": "大", "sc": 10, "l": 2, "f": 1091},
  {"h": "贸", "p": "mào", "sp": "mao", "r": "貝", "sc": 9, "l": 5, "f": 1092},
  {"h": "毕", "p": "bì", "sp": "bi", "r": "比", "sc": 6, "l": 4, "f": 1093},
  {"h": "唯", "p": "wéi", "sp": "wei", "r": "口", "sc": 11, "l": 5, "f": 1094},
  {"h": "谷", "p": "gǔ", "sp": "gu", "r": "谷", "sc": 7, "l": 5, "f": 1095},
  {"h": "轮", "p": "lún", "sp": "lun", "r": "車", "sc": 8, "l": 4, "f": 1096},
  {"h": "库", "p": "kù", "sp": "ku", "r": "广", "sc": 7, "l": 5, "f": 1097},
  {"h": "迹", "p": "jī", "sp": "ji", "r": "辵", "sc": 9, "l": 5, "f": 1098},
  {"h": "尤", "p": "yóu", "sp": "you", "r": "尢", "sc": 4, "l": 5, "f": 1099},
  {"h": "竞", "p": "jìng", "sp": "jing", "r": "立", "sc": 10, "l": 5, "f": 1100},
  {"h": "街", "p": "jiē", "sp": "jie", "r": "行", "sc": 12, "l": 2, "f": 1101},
  {"h": "促", "p": "cù", "sp": "cu", "r": "人", "sc": 9, "l": 4, "f": 1102},
  {"h": "延", "p": "yán", "sp": "yan", "r": "廴", "sc": 6, "l": 4, "f": 1103},
  {"h": "震", "p": "zhèn", "sp": "zhen", "r": "雨", "sc": 15, "l": 5, "f": 1104},
  {"h": "弃", "p": "qì", "sp": "qi", "r": "廾", "sc": 7, "l": 4, "f": 1105},
  {"h": "甲", "p": "jiǎ", "sp": "jia", "r": "田", "sc": 5, "l": 5, "f": 1106},
  {"h": "伟", "p": "wěi", "sp": "wei", "r": "人", "sc": 6, "l": 3, "f": 1107},
  {"h": "麻", "p": "má", "sp": "ma", "r": "麻", "sc": 11, "l": 3, "f": 1108},
  {"h": "川", "p": "chuān", "sp": "chuan", "r": "巛", "sc": 3, "l": 4, "f": 1109},
  {"h": "申", "p": "shēn", "sp": "shen", "r": "田", "sc": 5, "l": 4, "f": 1110},
  {"h": "缓", "p": "huǎn", "sp": "huan", "r": "糸", "sc": 12, "l": 4, "f": 1111},
  {"h": "潜", "p": "qián", "sp": "qian", "r": "水", "sc": 15, "l": 5, "f": 1112},
  {"h": "闪", "p": "shǎn", "sp": "shan", "r": "門", "sc": 5, "l": 4, "f": 1113},
  {"h": "售", "p": "shòu", "sp": "shou", "r": "口", "sc": 11, "l": 4, "f": 1114},
  {"h": "灯", "p": "dēng", "sp": "deng", "r": "火", "sc": 6, "l": 2, "f": 1115},
  {"h": "针", "p": "zhēn", "sp": "zhen", "r": "金", "sc": 7, "l": 4, "f": 1116},
  {"h": "哲", "p": "zhé", "sp": "zhe", "r": "口", "sc": 10, "l": 5, "f": 1117},
  {"h": "络", "p": "luò", "sp": "luo", "r": "糸", "sc": 9, "l": 4, "f": 1118},
  {"h": "抵", "p": "dǐ", "sp": "di", "r": "手", "sc": 8, "l": 5, "f": 1119},
  {"h": "朱", "p": "zhū", "sp": "zhu", "r": "木", "sc": 6, "l": 5, "f": 1120},
  {"h": "埃", "p": "āi", "sp": "ai", "r": "土", "sc": 10, "l": 5, "f": 1121},
  {"h": "抱", "p": "bào", "sp": "bao", "r": "手", "sc": 8, "l": 4, "f": 1122},
  {"h": "鼓", "p": "gǔ", "sp": "gu", "r": "鼓", "sc": 13, "l": 5, "f": 1123},
  {"h": "植", "p": "zhí", "sp": "zhi", "r": "木", "sc": 12, "l": 4, "f": 1124},
  {"h": "纯", "p": "chún", "sp": "chun", "r": "糸", "sc": 7, "l": 4, "f": 1125},
  {"h": "夏", "p": "xià", "sp": "xia", "r": "夊", "sc": 10, "l": 2, "f": 1126},
  {"h": "忍", "p": "rěn", "sp": "ren", "r": "心", "sc": 7, "l": 5, "f": 1127},
  {"h": "页", "p": "yè", "sp": "ye", "r": "頁", "sc": 6, "l": 1, "f": 1128},
  {"h": "杰", "p": "jié", "sp": "jie", "r": "木", "sc": 8, "l": 5, "f": 1129},
  {"h": "筑", "p": "zhù", "sp": "zhu", "r": "竹", "sc": 12, "l": 5, "f": 1130},
  {"h": "折", "p": "zhé", "sp": "zhe", "r": "手", "sc": 7, "l": 4, "f": 1131},
  {"h": "郑", "p": "zhèng", "sp": "zheng", "r": "邑", "sc": 8, "l": 5, "f": 1132},
  {"h": "贝", "p": "bèi", "sp": "bei", "r": "貝", "sc": 4, "l": 4, "f": 1133},
  {"h": "尊", "p": "zūn", "sp": "zun", "r": "寸", "sc": 12, "l": 5, "f": 1134},
  {"h": "吴", "p": "wú", "sp": "wu", "r": "口", "sc": 7, "l": 5, "f": 1135},
  {"h": "秀", "p": "xiù", "sp": "xiu", "r": "禾", "sc": 7, "l": 4, "f": 1136},
  {"h": "混", "p": "hùn", "sp": "hun", "r": "水", "sc": 11, "l": 4, "f": 1137},
  {"h": "臣", "p": "chén", "sp": "chen", "r": "臣", "sc": 6, "l": 5, "f": 1138},
  {"h": "雅", "p": "yǎ", "sp": "ya", "r": "隹", "sc": 12, "l": 5, "f": 1139},
  {"h": "振", "p": "zhèn", "sp": "zhen", "r": "手", "sc": 10, "l": 5, "f": 1140},
  {"h": "染", "p": "rǎn", "sp": "ran", "r": "木", "sc": 9, "l": 4, "f": 1141},
  {
    "h": "盛",
    "p": "shèng",
    "sp": "sheng",
    "r": "皿",
    "sc": 11,
    "l": 5,
    "f": 1142
  },
  {"h": "怒", "p": "nù", "sp": "nu", "r": "心", "sc": 9, "l": 5, "f": 1143},
  {"h": "舞", "p": "wǔ", "sp": "wu", "r": "舛", "sc": 14, "l": 2, "f": 1144},
  {"h": "圆", "p": "yuán", "sp": "yuan", "r": "囗", "sc": 10, "l": 4, "f": 1145},
  {"h": "搞", "p": "gǎo", "sp": "gao", "r": "手", "sc": 13, "l": 5, "f": 1146},
  {"h": "狂", "p": "kuáng", "sp": "kuang", "r": "犬", "sc": 7, "l": 5, "f": 1147},
  {"h": "措", "p": "cuò", "sp": "cuo", "r": "手", "sc": 11, "l": 4, "f": 1148},
  {"h": "姓", "p": "xìng", "sp": "xing", "r": "女", "sc": 8, "l": 2, "f": 1149},
  {"h": "残", "p": "cán", "sp": "can", "r": "歹", "sc": 9, "l": 5, "f": 1150},
  {"h": "秋", "p": "qiū", "sp": "qiu", "r": "禾", "sc": 9, "l": 2, "f": 1151},
  {"h": "培", "p": "péi", "sp": "pei", "r": "土", "sc": 11, "l": 4, "f": 1152},
  {"h": "迷", "p": "mí", "sp": "mi", "r": "辵", "sc": 9, "l": 3, "f": 1153},
  {"h": "诚", "p": "chéng", "sp": "cheng", "r": "言", "sc": 8, "l": 4, "f": 1154},
  {"h": "宽", "p": "kuān", "sp": "kuan", "r": "宀", "sc": 10, "l": 4, "f": 1155},
  {"h": "宇", "p": "yǔ", "sp": "yu", "r": "宀", "sc": 6, "l": 5, "f": 1156},
  {"h": "猛", "p": "měng", "sp": "meng", "r": "犬", "sc": 11, "l": 5, "f": 1157},
  {"h": "摆", "p": "bǎi", "sp": "bai", "r": "手", "sc": 13, "l": 4, "f": 1158},
  {"h": "梅", "p": "méi", "sp": "mei", "r": "木", "sc": 11, "l": 5, "f": 1159},
  {"h": "毁", "p": "huǐ", "sp": "hui", "r": "殳", "sc": 13, "l": 5, "f": 1160},
  {"h": "伸", "p": "shēn", "sp": "shen", "r": "人", "sc": 7, "l": 5, "f": 1161},
  {"h": "摩", "p": "mó", "sp": "mo", "r": "手", "sc": 15, "l": 4, "f": 1162},
  {"h": "盟", "p": "méng", "sp": "meng", "r": "皿", "sc": 13, "l": 5, "f": 1163},
  {"h": "末", "p": "mò", "sp": "mo", "r": "木", "sc": 5, "l": 2, "f": 1164},
  {"h": "乃", "p": "nǎi", "sp": "nai", "r": "丿", "sc": 2, "l": 5, "f": 1165},
  {"h": "悲", "p": "bēi", "sp": "bei", "r": "心", "sc": 12, "l": 5, "f": 1166},
  {"h": "拍", "p": "pāi", "sp": "pai", "r": "手", "sc": 8, "l": 3, "f": 1167},
  {"h": "丁", "p": "dīng", "sp": "ding", "r": "一", "sc": 2, "l": 5, "f": 1168},
  {"h": "赵", "p": "zhào", "sp": "zhao", "r": "走", "sc": 9, "l": 5, "f": 1169},
  {"h": "硬", "p": "yìng", "sp": "ying", "r": "石", "sc": 12, "l": 5, "f": 1170},
  {"h": "麦", "p": "mài", "sp": "mai", "r": "麥", "sc": 7, "l": 4, "f": 1171},
  {
    "h": "蒋",
    "p": "jiǎng",
    "sp": "jiang",
    "r": "艸",
    "sc": 12,
    "l": 5,
    "f": 1172
  },
  {"h": "操", "p": "cāo", "sp": "cao", "r": "手", "sc": 16, "l": 4, "f": 1173},
  {"h": "耶", "p": "yé", "sp": "ye", "r": "耳", "sc": 8, "l": 5, "f": 1174},
  {"h": "阻", "p": "zǔ", "sp": "zu", "r": "阜", "sc": 7, "l": 4, "f": 1175},
  {"h": "订", "p": "dìng", "sp": "ding", "r": "言", "sc": 4, "l": 3, "f": 1176},
  {"h": "彩", "p": "cǎi", "sp": "cai", "r": "彡", "sc": 11, "l": 3, "f": 1177},
  {"h": "抽", "p": "chōu", "sp": "chou", "r": "手", "sc": 8, "l": 4, "f": 1178},
  {"h": "赞", "p": "zàn", "sp": "zan", "r": "貝", "sc": 16, "l": 4, "f": 1179},
  {"h": "魔", "p": "mó", "sp": "mo", "r": "鬼", "sc": 20, "l": 5, "f": 1180},
  {"h": "纷", "p": "fēn", "sp": "fen", "r": "糸", "sc": 7, "l": 4, "f": 1181},
  {"h": "沿", "p": "yán", "sp": "yan", "r": "水", "sc": 8, "l": 5, "f": 1182},
  {"h": "喊", "p": "hǎn", "sp": "han", "r": "口", "sc": 12, "l": 2, "f": 1183},
  {"h": "违", "p": "wéi", "sp": "wei", "r": "辵", "sc": 7, "l": 5, "f": 1184},
  {"h": "妹", "p": "mèi", "sp": "mei", "r": "女", "sc": 8, "l": 1, "f": 1185},
  {"h": "浪", "p": "làng", "sp": "lang", "r": "水", "sc": 10, "l": 3, "f": 1186},
  {"h": "汇", "p": "huì", "sp": "hui", "r": "水", "sc": 5, "l": 4, "f": 1187},
  {"h": "币", "p": "bì", "sp": "bi", "r": "巾", "sc": 4, "l": 3, "f": 1188},
  {"h": "丰", "p": "fēng", "sp": "feng", "r": "丨", "sc": 4, "l": 3, "f": 1189},
  {"h": "蓝", "p": "lán", "sp": "lan", "r": "艸", "sc": 13, "l": 2, "f": 1190},
  {"h": "殊", "p": "shū", "sp": "shu", "r": "歹", "sc": 10, "l": 4, "f": 1191},
  {"h": "献", "p": "xiàn", "sp": "xian", "r": "犬", "sc": 13, "l": 5, "f": 1192},
  {"h": "桌", "p": "zhuō", "sp": "zhuo", "r": "木", "sc": 10, "l": 1, "f": 1193},
  {"h": "啦", "p": "la", "sp": "la", "r": "口", "sc": 11, "l": 5, "f": 1194},
  {"h": "瓦", "p": "wǎ", "sp": "wa", "r": "瓦", "sc": 4, "l": 5, "f": 1195},
  {"h": "莱", "p": "lái", "sp": "lai", "r": "艸", "sc": 10, "l": 5, "f": 1196},
  {"h": "援", "p": "yuán", "sp": "yuan", "r": "手", "sc": 12, "l": 5, "f": 1197},
  {"h": "译", "p": "yì", "sp": "yi", "r": "言", "sc": 7, "l": 4, "f": 1198},
  {"h": "夺", "p": "duó", "sp": "duo", "r": "大", "sc": 6, "l": 5, "f": 1199},
  {"h": "汽", "p": "qì", "sp": "qi", "r": "水", "sc": 7, "l": 1, "f": 1200},
  {"h": "烧", "p": "shāo", "sp": "shao", "r": "火", "sc": 10, "l": 4, "f": 1201},
  {"h": "距", "p": "jù", "sp": "ju", "r": "足", "sc": 11, "l": 4, "f": 1202},
  {"h": "裁", "p": "cái", "sp": "cai", "r": "衣", "sc": 12, "l": 5, "f": 1203},
  {"h": "偏", "p": "piān", "sp": "pian", "r": "人", "sc": 11, "l": 5, "f": 1204},
  {"h": "符", "p": "fú", "sp": "fu", "r": "竹", "sc": 11, "l": 4, "f": 1205},
  {"h": "勇", "p": "yǒng", "sp": "yong", "r": "力", "sc": 9, "l": 4, "f": 1206},
  {"h": "触", "p": "chù", "sp": "chu", "r": "角", "sc": 13, "l": 5, "f": 1207},
  {"h": "课", "p": "kè", "sp": "ke", "r": "言", "sc": 10, "l": 1, "f": 1208},
  {"h": "敬", "p": "jìng", "sp": "jing", "r": "攴", "sc": 12, "l": 5, "f": 1209},
  {"h": "哭", "p": "kū", "sp": "ku", "r": "口", "sc": 10, "l": 2, "f": 1210},
  {"h": "懂", "p": "dǒng", "sp": "dong", "r": "心", "sc": 15, "l": 1, "f": 1211},
  {
    "h": "墙",
    "p": "qiáng",
    "sp": "qiang",
    "r": "土",
    "sc": 14,
    "l": 2,
    "f": 1212
  },
  {"h": "袭", "p": "xí", "sp": "xi", "r": "龍", "sc": 11, "l": 5, "f": 1213},
  {"h": "召", "p": "zhào", "sp": "zhao", "r": "口", "sc": 5, "l": 4, "f": 1214},
  {"h": "罚", "p": "fá", "sp": "fa", "r": "网", "sc": 9, "l": 5, "f": 1215},
  {"h": "侠", "p": "xiá", "sp": "xia", "r": "人", "sc": 8, "l": 5, "f": 1216},
  {"h": "厅", "p": "tīng", "sp": "ting", "r": "厂", "sc": 4, "l": 4, "f": 1217},
  {"h": "拜", "p": "bài", "sp": "bai", "r": "手", "sc": 9, "l": 5, "f": 1218},
  {"h": "巧", "p": "qiǎo", "sp": "qiao", "r": "工", "sc": 5, "l": 3, "f": 1219},
  {"h": "侧", "p": "cè", "sp": "ce", "r": "人", "sc": 8, "l": 5, "f": 1220},
  {"h": "韩", "p": "hán", "sp": "han", "r": "韋", "sc": 12, "l": 5, "f": 1221},
  {"h": "冒", "p": "mào", "sp": "mao", "r": "冂", "sc": 9, "l": 3, "f": 1222},
  {"h": "债", "p": "zhài", "sp": "zhai", "r": "人", "sc": 10, "l": 5, "f": 1223},
  {"h": "曼", "p": "màn", "sp": "man", "r": "日", "sc": 11, "l": 5, "f": 1224},
  {"h": "融", "p": "róng", "sp": "rong", "r": "虫", "sc": 16, "l": 5, "f": 1225},
  {"h": "惯", "p": "guàn", "sp": "guan", "r": "心", "sc": 11, "l": 2, "f": 1226},
  {"h": "享", "p": "xiǎng", "sp": "xiang", "r": "亠", "sc": 8, "l": 5, "f": 1227},
  {"h": "戴", "p": "dài", "sp": "dai", "r": "戈", "sc": 17, "l": 4, "f": 1228},
  {"h": "童", "p": "tóng", "sp": "tong", "r": "立", "sc": 12, "l": 4, "f": 1229},
  {"h": "犹", "p": "yóu", "sp": "you", "r": "犬", "sc": 7, "l": 5, "f": 1230},
  {
    "h": "乘",
    "p": "chéng",
    "sp": "cheng",
    "r": "丿",
    "sc": 10,
    "l": 5,
    "f": 1231
  },
  {"h": "挂", "p": "guà", "sp": "gua", "r": "手", "sc": 9, "l": 3, "f": 1232},
  {"h": "奖", "p": "jiǎng", "sp": "jiang", "r": "大", "sc": 9, "l": 4, "f": 1233},
  {"h": "绍", "p": "shào", "sp": "shao", "r": "糸", "sc": 8, "l": 1, "f": 1234},
  {"h": "厚", "p": "hòu", "sp": "hou", "r": "厂", "sc": 9, "l": 4, "f": 1235},
  {"h": "纵", "p": "zòng", "sp": "zong", "r": "糸", "sc": 7, "l": 5, "f": 1236},
  {
    "h": "障",
    "p": "zhàng",
    "sp": "zhang",
    "r": "阜",
    "sc": 13,
    "l": 5,
    "f": 1237
  },
  {"h": "讯", "p": "xùn", "sp": "xun", "r": "言", "sc": 5, "l": 5, "f": 1238},
  {"h": "涉", "p": "shè", "sp": "she", "r": "水", "sc": 10, "l": 5, "f": 1239},
  {"h": "彻", "p": "chè", "sp": "che", "r": "彳", "sc": 7, "l": 4, "f": 1240},
  {"h": "刊", "p": "kān", "sp": "kan", "r": "刀", "sc": 5, "l": 5, "f": 1241},
  {"h": "丈", "p": "zhàng", "sp": "zhang", "r": "一", "sc": 3, "l": 4, "f": 1242},
  {"h": "爆", "p": "bào", "sp": "bao", "r": "火", "sc": 19, "l": 5, "f": 1243},
  {"h": "乌", "p": "wū", "sp": "wu", "r": "丿", "sc": 4, "l": 5, "f": 1244},
  {"h": "役", "p": "yì", "sp": "yi", "r": "彳", "sc": 7, "l": 5, "f": 1245},
  {"h": "描", "p": "miáo", "sp": "miao", "r": "手", "sc": 11, "l": 4, "f": 1246},
  {"h": "洗", "p": "xǐ", "sp": "xi", "r": "水", "sc": 9, "l": 1, "f": 1247},
  {"h": "玛", "p": "mǎ", "sp": "ma", "r": "玉", "sc": 7, "l": 5, "f": 1248},
  {"h": "患", "p": "huàn", "sp": "huan", "r": "心", "sc": 11, "l": 5, "f": 1249},
  {"h": "妙", "p": "miào", "sp": "miao", "r": "女", "sc": 7, "l": 5, "f": 1250},
  {"h": "镜", "p": "jìng", "sp": "jing", "r": "金", "sc": 16, "l": 3, "f": 1251},
  {
    "h": "唱",
    "p": "chàng",
    "sp": "chang",
    "r": "口",
    "sc": 11,
    "l": 1,
    "f": 1252
  },
  {"h": "烦", "p": "fán", "sp": "fan", "r": "火", "sc": 10, "l": 3, "f": 1253},
  {"h": "签", "p": "qiān", "sp": "qian", "r": "竹", "sc": 13, "l": 4, "f": 1254},
  {"h": "仙", "p": "xiān", "sp": "xian", "r": "人", "sc": 5, "l": 5, "f": 1255},
  {"h": "彼", "p": "bǐ", "sp": "bi", "r": "彳", "sc": 8, "l": 5, "f": 1256},
  {"h": "弗", "p": "fú", "sp": "fu", "r": "弓", "sc": 5, "l": 5, "f": 1257},
  {
    "h": "症",
    "p": "zhèng",
    "sp": "zheng",
    "r": "疒",
    "sc": 10,
    "l": 5,
    "f": 1258
  },
  {"h": "仿", "p": "fǎng", "sp": "fang", "r": "人", "sc": 6, "l": 5, "f": 1259},
  {"h": "倾", "p": "qīng", "sp": "qing", "r": "人", "sc": 10, "l": 5, "f": 1260},
  {"h": "牌", "p": "pái", "sp": "pai", "r": "片", "sc": 12, "l": 3, "f": 1261},
  {"h": "陷", "p": "xiàn", "sp": "xian", "r": "阜", "sc": 10, "l": 5, "f": 1262},
  {"h": "鸟", "p": "niǎo", "sp": "niao", "r": "鳥", "sc": 5, "l": 2, "f": 1263},
  {"h": "轰", "p": "hōng", "sp": "hong", "r": "車", "sc": 8, "l": 5, "f": 1264},
  {"h": "咱", "p": "zán", "sp": "zan", "r": "口", "sc": 9, "l": 2, "f": 1265},
  {"h": "菜", "p": "cài", "sp": "cai", "r": "艸", "sc": 11, "l": 1, "f": 1266},
  {"h": "闭", "p": "bì", "sp": "bi", "r": "門", "sc": 6, "l": 4, "f": 1267},
  {"h": "奋", "p": "fèn", "sp": "fen", "r": "大", "sc": 8, "l": 4, "f": 1268},
  {"h": "庆", "p": "qìng", "sp": "qing", "r": "广", "sc": 6, "l": 3, "f": 1269},
  {"h": "撤", "p": "chè", "sp": "che", "r": "手", "sc": 15, "l": 5, "f": 1270},
  {"h": "泪", "p": "lèi", "sp": "lei", "r": "水", "sc": 8, "l": 4, "f": 1271},
  {"h": "茶", "p": "chá", "sp": "cha", "r": "艸", "sc": 9, "l": 1, "f": 1272},
  {"h": "疾", "p": "jí", "sp": "ji", "r": "疒", "sc": 10, "l": 5, "f": 1273},
  {"h": "缘", "p": "yuán", "sp": "yuan", "r": "糸", "sc": 12, "l": 5, "f": 1274},
  {"h": "播", "p": "bō", "sp": "bo", "r": "手", "sc": 15, "l": 3, "f": 1275},
  {"h": "朗", "p": "lǎng", "sp": "lang", "r": "月", "sc": 10, "l": 5, "f": 1276},
  {"h": "杜", "p": "dù", "sp": "du", "r": "木", "sc": 7, "l": 5, "f": 1277},
  {"h": "奶", "p": "nǎi", "sp": "nai", "r": "女", "sc": 5, "l": 1, "f": 1278},
  {"h": "季", "p": "jì", "sp": "ji", "r": "子", "sc": 8, "l": 4, "f": 1279},
  {"h": "丹", "p": "dān", "sp": "dan", "r": "丶", "sc": 4, "l": 4, "f": 1280},
  {"h": "狗", "p": "gǒu", "sp": "gou", "r": "犬", "sc": 8, "l": 1, "f": 1281},
  {"h": "尾", "p": "wěi", "sp": "wei", "r": "尸", "sc": 7, "l": 4, "f": 1282},
  {"h": "仪", "p": "yí", "sp": "yi", "r": "人", "sc": 5, "l": 5, "f": 1283},
  {"h": "偷", "p": "tōu", "sp": "tou", "r": "人", "sc": 11, "l": 5, "f": 1284},
  {"h": "奔", "p": "bēn", "sp": "ben", "r": "大", "sc": 8, "l": 5, "f": 1285},
  {"h": "珠", "p": "zhū", "sp": "zhu", "r": "玉", "sc": 10, "l": 5, "f": 1286},
  {"h": "虫", "p": "chóng", "sp": "chong", "r": "虫", "sc": 6, "l": 4, "f": 1287},
  {"h": "驻", "p": "zhù", "sp": "zhu", "r": "馬", "sc": 8, "l": 5, "f": 1288},
  {"h": "孔", "p": "kǒng", "sp": "kong", "r": "子", "sc": 4, "l": 5, "f": 1289},
  {"h": "宜", "p": "yí", "sp": "yi", "r": "宀", "sc": 8, "l": 2, "f": 1290},
  {"h": "艾", "p": "ài", "sp": "ai", "r": "艸", "sc": 5, "l": 5, "f": 1291},
  {"h": "桥", "p": "qiáo", "sp": "qiao", "r": "木", "sc": 10, "l": 3, "f": 1292},
  {"h": "淡", "p": "dàn", "sp": "dan", "r": "水", "sc": 11, "l": 4, "f": 1293},
  {"h": "翼", "p": "yì", "sp": "yi", "r": "羽", "sc": 17, "l": 5, "f": 1294},
  {"h": "恨", "p": "hèn", "sp": "hen", "r": "心", "sc": 9, "l": 5, "f": 1295},
  {"h": "繁", "p": "fán", "sp": "fan", "r": "糸", "sc": 17, "l": 5, "f": 1296},
  {"h": "寒", "p": "hán", "sp": "han", "r": "宀", "sc": 12, "l": 4, "f": 1297},
  {"h": "伴", "p": "bàn", "sp": "ban", "r": "人", "sc": 7, "l": 4, "f": 1298},
  {"h": "叹", "p": "tàn", "sp": "tan", "r": "口", "sc": 5, "l": 5, "f": 1299},
  {"h": "旦", "p": "dàn", "sp": "dan", "r": "日", "sc": 5, "l": 5, "f": 1300},
  {"h": "愈", "p": "yù", "sp": "yu", "r": "心", "sc": 13, "l": 5, "f": 1301},
  {"h": "潮", "p": "cháo", "sp": "chao", "r": "水", "sc": 15, "l": 4, "f": 1302},
  {
    "h": "粮",
    "p": "liáng",
    "sp": "liang",
    "r": "米",
    "sc": 13,
    "l": 4,
    "f": 1303
  },
  {"h": "缩", "p": "suō", "sp": "suo", "r": "糸", "sc": 14, "l": 4, "f": 1304},
  {"h": "罢", "p": "bà", "sp": "ba", "r": "网", "sc": 10, "l": 5, "f": 1305},
  {"h": "聚", "p": "jù", "sp": "ju", "r": "耳", "sc": 14, "l": 4, "f": 1306},
  {"h": "径", "p": "jìng", "sp": "jing", "r": "彳", "sc": 8, "l": 5, "f": 1307},
  {"h": "恰", "p": "qià", "sp": "qia", "r": "心", "sc": 9, "l": 5, "f": 1308},
  {"h": "挑", "p": "tiāo", "sp": "tiao", "r": "手", "sc": 9, "l": 4, "f": 1309},
  {"h": "袋", "p": "dài", "sp": "dai", "r": "衣", "sc": 11, "l": 3, "f": 1310},
  {"h": "灰", "p": "huī", "sp": "hui", "r": "火", "sc": 6, "l": 5, "f": 1311},
  {"h": "捕", "p": "bǔ", "sp": "bu", "r": "手", "sc": 10, "l": 5, "f": 1312},
  {"h": "徐", "p": "xú", "sp": "xu", "r": "彳", "sc": 10, "l": 5, "f": 1313},
  {"h": "珍", "p": "zhēn", "sp": "zhen", "r": "玉", "sc": 9, "l": 5, "f": 1314},
  {"h": "幕", "p": "mù", "sp": "mu", "r": "巾", "sc": 13, "l": 5, "f": 1315},
  {"h": "映", "p": "yìng", "sp": "ying", "r": "日", "sc": 9, "l": 4, "f": 1316},
  {"h": "裂", "p": "liè", "sp": "lie", "r": "衣", "sc": 12, "l": 5, "f": 1317},
  {"h": "泰", "p": "tài", "sp": "tai", "r": "水", "sc": 10, "l": 5, "f": 1318},
  {"h": "隔", "p": "gé", "sp": "ge", "r": "阜", "sc": 12, "l": 4, "f": 1319},
  {"h": "启", "p": "qǐ", "sp": "qi", "r": "口", "sc": 7, "l": 5, "f": 1320},
  {"h": "尖", "p": "jiān", "sp": "jian", "r": "小", "sc": 6, "l": 5, "f": 1321},
  {"h": "忠", "p": "zhōng", "sp": "zhong", "r": "心", "sc": 8, "l": 5, "f": 1322},
  {"h": "累", "p": "lèi", "sp": "lei", "r": "糸", "sc": 11, "l": 1, "f": 1323},
  {"h": "炎", "p": "yán", "sp": "yan", "r": "火", "sc": 8, "l": 5, "f": 1324},
  {"h": "暂", "p": "zàn", "sp": "zan", "r": "日", "sc": 12, "l": 4, "f": 1325},
  {"h": "估", "p": "gū", "sp": "gu", "r": "人", "sc": 7, "l": 5, "f": 1326},
  {"h": "泛", "p": "fàn", "sp": "fan", "r": "水", "sc": 7, "l": 5, "f": 1327},
  {"h": "荒", "p": "huāng", "sp": "huang", "r": "艸", "sc": 9, "l": 5, "f": 1328},
  {
    "h": "偿",
    "p": "cháng",
    "sp": "chang",
    "r": "人",
    "sc": 11,
    "l": 5,
    "f": 1329
  },
  {"h": "横", "p": "héng", "sp": "heng", "r": "木", "sc": 15, "l": 5, "f": 1330},
  {"h": "拒", "p": "jù", "sp": "ju", "r": "手", "sc": 7, "l": 4, "f": 1331},
  {"h": "瑞", "p": "ruì", "sp": "rui", "r": "玉", "sc": 13, "l": 5, "f": 1332},
  {"h": "忆", "p": "yì", "sp": "yi", "r": "心", "sc": 4, "l": 5, "f": 1333},
  {"h": "孤", "p": "gū", "sp": "gu", "r": "子", "sc": 8, "l": 5, "f": 1334},
  {"h": "鼻", "p": "bí", "sp": "bi", "r": "鼻", "sc": 14, "l": 3, "f": 1335},
  {"h": "闹", "p": "nào", "sp": "nao", "r": "門", "sc": 8, "l": 4, "f": 1336},
  {"h": "羊", "p": "yáng", "sp": "yang", "r": "羊", "sc": 6, "l": 2, "f": 1337},
  {"h": "呆", "p": "dāi", "sp": "dai", "r": "口", "sc": 7, "l": 5, "f": 1338},
  {"h": "厉", "p": "lì", "sp": "li", "r": "厂", "sc": 5, "l": 4, "f": 1339},
  {"h": "衡", "p": "héng", "sp": "heng", "r": "行", "sc": 16, "l": 5, "f": 1340},
  {"h": "胞", "p": "bāo", "sp": "bao", "r": "肉", "sc": 9, "l": 5, "f": 1341},
  {"h": "零", "p": "líng", "sp": "ling", "r": "雨", "sc": 13, "l": 1, "f": 1342},
  {"h": "穷", "p": "qióng", "sp": "qiong", "r": "穴", "sc": 7, "l": 4, "f": 1343},
  {"h": "舍", "p": "shě", "sp": "she", "r": "舌", "sc": 8, "l": 5, "f": 1344},
  {"h": "码", "p": "mǎ", "sp": "ma", "r": "石", "sc": 8, "l": 4, "f": 1345},
  {"h": "赫", "p": "hè", "sp": "he", "r": "赤", "sc": 14, "l": 5, "f": 1346},
  {"h": "婆", "p": "pó", "sp": "po", "r": "女", "sc": 11, "l": 4, "f": 1347},
  {"h": "魂", "p": "hún", "sp": "hun", "r": "鬼", "sc": 13, "l": 5, "f": 1348},
  {"h": "灾", "p": "zāi", "sp": "zai", "r": "火", "sc": 7, "l": 5, "f": 1349},
  {"h": "洪", "p": "hóng", "sp": "hong", "r": "水", "sc": 9, "l": 5, "f": 1350},
  {"h": "腿", "p": "tuǐ", "sp": "tui", "r": "肉", "sc": 13, "l": 2, "f": 1351},
  {"h": "胆", "p": "dǎn", "sp": "dan", "r": "肉", "sc": 9, "l": 5, "f": 1352},
  {"h": "津", "p": "jīn", "sp": "jin", "r": "水", "sc": 9, "l": 4, "f": 1353},
  {"h": "俗", "p": "sú", "sp": "su", "r": "人", "sc": 9, "l": 4, "f": 1354},
  {"h": "辩", "p": "biàn", "sp": "bian", "r": "辛", "sc": 16, "l": 4, "f": 1355},
  {
    "h": "胸",
    "p": "xiōng",
    "sp": "xiong",
    "r": "肉",
    "sc": 10,
    "l": 4,
    "f": 1356
  },
  {"h": "晓", "p": "xiǎo", "sp": "xiao", "r": "日", "sc": 10, "l": 5, "f": 1357},
  {"h": "劲", "p": "jìn", "sp": "jin", "r": "力", "sc": 7, "l": 4, "f": 1358},
  {"h": "贫", "p": "pín", "sp": "pin", "r": "貝", "sc": 8, "l": 5, "f": 1359},
  {"h": "仁", "p": "rén", "sp": "ren", "r": "人", "sc": 4, "l": 5, "f": 1360},
  {"h": "偶", "p": "ǒu", "sp": "ou", "r": "人", "sc": 11, "l": 5, "f": 1361},
  {"h": "辑", "p": "jí", "sp": "ji", "r": "車", "sc": 13, "l": 5, "f": 1362},
  {"h": "邦", "p": "bāng", "sp": "bang", "r": "邑", "sc": 6, "l": 5, "f": 1363},
  {"h": "恢", "p": "huī", "sp": "hui", "r": "心", "sc": 9, "l": 5, "f": 1364},
  {"h": "赖", "p": "lài", "sp": "lai", "r": "貝", "sc": 13, "l": 5, "f": 1365},
  {"h": "圈", "p": "quān", "sp": "quan", "r": "囗", "sc": 11, "l": 4, "f": 1366},
  {"h": "摸", "p": "mō", "sp": "mo", "r": "手", "sc": 13, "l": 4, "f": 1367},
  {"h": "仰", "p": "yǎng", "sp": "yang", "r": "人", "sc": 6, "l": 5, "f": 1368},
  {"h": "润", "p": "rùn", "sp": "run", "r": "水", "sc": 10, "l": 5, "f": 1369},
  {"h": "堆", "p": "duī", "sp": "dui", "r": "土", "sc": 11, "l": 5, "f": 1370},
  {"h": "碰", "p": "pèng", "sp": "peng", "r": "石", "sc": 13, "l": 2, "f": 1371},
  {"h": "艇", "p": "tǐng", "sp": "ting", "r": "舟", "sc": 12, "l": 5, "f": 1372},
  {"h": "稍", "p": "shāo", "sp": "shao", "r": "禾", "sc": 12, "l": 5, "f": 1373},
  {"h": "迟", "p": "chí", "sp": "chi", "r": "辵", "sc": 7, "l": 3, "f": 1374},
  {
    "h": "辆",
    "p": "liàng",
    "sp": "liang",
    "r": "車",
    "sc": 11,
    "l": 2,
    "f": 1375
  },
  {"h": "废", "p": "fèi", "sp": "fei", "r": "广", "sc": 8, "l": 5, "f": 1376},
  {"h": "净", "p": "jìng", "sp": "jing", "r": "冫", "sc": 8, "l": 1, "f": 1377},
  {"h": "凶", "p": "xiōng", "sp": "xiong", "r": "凵", "sc": 4, "l": 5, "f": 1378},
  {"h": "署", "p": "shǔ", "sp": "shu", "r": "网", "sc": 13, "l": 5, "f": 1379},
  {"h": "壁", "p": "bì", "sp": "bi", "r": "土", "sc": 16, "l": 5, "f": 1380},
  {"h": "御", "p": "yù", "sp": "yu", "r": "彳", "sc": 12, "l": 5, "f": 1381},
  {"h": "奉", "p": "fèng", "sp": "feng", "r": "大", "sc": 8, "l": 5, "f": 1382},
  {"h": "旋", "p": "xuán", "sp": "xuan", "r": "方", "sc": 11, "l": 5, "f": 1383},
  {"h": "冬", "p": "dōng", "sp": "dong", "r": "冫", "sc": 5, "l": 2, "f": 1384},
  {"h": "矿", "p": "kuàng", "sp": "kuang", "r": "石", "sc": 8, "l": 4, "f": 1385},
  {"h": "抬", "p": "tái", "sp": "tai", "r": "手", "sc": 8, "l": 5, "f": 1386},
  {"h": "蛋", "p": "dàn", "sp": "dan", "r": "虫", "sc": 11, "l": 1, "f": 1387},
  {"h": "晨", "p": "chén", "sp": "chen", "r": "日", "sc": 11, "l": 2, "f": 1388},
  {"h": "伏", "p": "fú", "sp": "fu", "r": "人", "sc": 6, "l": 5, "f": 1389},
  {"h": "吹", "p": "chuī", "sp": "chui", "r": "口", "sc": 7, "l": 2, "f": 1390},
  {"h": "鸡", "p": "jī", "sp": "ji", "r": "鳥", "sc": 7, "l": 1, "f": 1391},
  {"h": "倍", "p": "bèi", "sp": "bei", "r": "人", "sc": 10, "l": 4, "f": 1392},
  {"h": "糊", "p": "hú", "sp": "hu", "r": "米", "sc": 15, "l": 5, "f": 1393},
  {"h": "秦", "p": "qín", "sp": "qin", "r": "禾", "sc": 10, "l": 5, "f": 1394},
  {"h": "盾", "p": "dùn", "sp": "dun", "r": "目", "sc": 9, "l": 4, "f": 1395},
  {"h": "杯", "p": "bēi", "sp": "bei", "r": "木", "sc": 8, "l": 1, "f": 1396},
  {"h": "租", "p": "zū", "sp": "zu", "r": "禾", "sc": 10, "l": 1, "f": 1397},
  {"h": "骑", "p": "qí", "sp": "qi", "r": "馬", "sc": 11, "l": 2, "f": 1398},
  {"h": "乏", "p": "fá", "sp": "fa", "r": "丿", "sc": 4, "l": 5, "f": 1399},
  {"h": "隆", "p": "lóng", "sp": "long", "r": "阜", "sc": 11, "l": 5, "f": 1400},
  {"h": "诊", "p": "zhěn", "sp": "zhen", "r": "言", "sc": 7, "l": 5, "f": 1401},
  {"h": "奴", "p": "nú", "sp": "nu", "r": "女", "sc": 5, "l": 5, "f": 1402},
  {"h": "摄", "p": "shè", "sp": "she", "r": "手", "sc": 13, "l": 5, "f": 1403},
  {"h": "丧", "p": "sàng", "sp": "sang", "r": "十", "sc": 8, "l": 5, "f": 1404},
  {"h": "污", "p": "wū", "sp": "wu", "r": "水", "sc": 6, "l": 4, "f": 1405},
  {"h": "渡", "p": "dù", "sp": "du", "r": "水", "sc": 12, "l": 5, "f": 1406},
  {"h": "旗", "p": "qí", "sp": "qi", "r": "方", "sc": 14, "l": 5, "f": 1407},
  {"h": "甘", "p": "gān", "sp": "gan", "r": "甘", "sc": 5, "l": 5, "f": 1408},
  {"h": "耐", "p": "nài", "sp": "nai", "r": "而", "sc": 9, "l": 5, "f": 1409},
  {"h": "凭", "p": "píng", "sp": "ping", "r": "几", "sc": 8, "l": 5, "f": 1410},
  {"h": "扎", "p": "zhā", "sp": "zha", "r": "手", "sc": 4, "l": 5, "f": 1411},
  {"h": "抢", "p": "qiǎng", "sp": "qiang", "r": "手", "sc": 7, "l": 5, "f": 1412},
  {"h": "绪", "p": "xù", "sp": "xu", "r": "糸", "sc": 11, "l": 5, "f": 1413},
  {"h": "粗", "p": "cū", "sp": "cu", "r": "米", "sc": 11, "l": 4, "f": 1414},
  {"h": "肩", "p": "jiān", "sp": "jian", "r": "肉", "sc": 8, "l": 5, "f": 1415},
  {
    "h": "梁",
    "p": "liáng",
    "sp": "liang",
    "r": "木",
    "sc": 11,
    "l": 5,
    "f": 1416
  },
  {"h": "幻", "p": "huàn", "sp": "huan", "r": "幺", "sc": 4, "l": 5, "f": 1417},
  {"h": "菲", "p": "fēi", "sp": "fei", "r": "艸", "sc": 11, "l": 5, "f": 1418},
  {"h": "皆", "p": "jiē", "sp": "jie", "r": "白", "sc": 9, "l": 5, "f": 1419},
  {"h": "碎", "p": "suì", "sp": "sui", "r": "石", "sc": 13, "l": 5, "f": 1420},
  {"h": "宙", "p": "zhòu", "sp": "zhou", "r": "宀", "sc": 8, "l": 5, "f": 1421},
  {"h": "叔", "p": "shū", "sp": "shu", "r": "又", "sc": 8, "l": 3, "f": 1422},
  {"h": "岩", "p": "yán", "sp": "yan", "r": "山", "sc": 8, "l": 5, "f": 1423},
  {"h": "荡", "p": "dàng", "sp": "dang", "r": "艸", "sc": 9, "l": 5, "f": 1424},
  {"h": "综", "p": "zōng", "sp": "zong", "r": "糸", "sc": 11, "l": 4, "f": 1425},
  {"h": "爬", "p": "pá", "sp": "pa", "r": "爪", "sc": 8, "l": 2, "f": 1426},
  {"h": "荷", "p": "hé", "sp": "he", "r": "艸", "sc": 10, "l": 4, "f": 1427},
  {"h": "悉", "p": "xī", "sp": "xi", "r": "心", "sc": 11, "l": 5, "f": 1428},
  {"h": "蒂", "p": "dì", "sp": "di", "r": "艸", "sc": 12, "l": 5, "f": 1429},
  {"h": "返", "p": "fǎn", "sp": "fan", "r": "辵", "sc": 7, "l": 5, "f": 1430},
  {"h": "井", "p": "jǐng", "sp": "jing", "r": "二", "sc": 4, "l": 5, "f": 1431},
  {
    "h": "壮",
    "p": "zhuàng",
    "sp": "zhuang",
    "r": "士",
    "sc": 6,
    "l": 5,
    "f": 1432
  },
  {"h": "薄", "p": "báo", "sp": "bao", "r": "艸", "sc": 16, "l": 4, "f": 1433},
  {"h": "悄", "p": "qiāo", "sp": "qiao", "r": "心", "sc": 10, "l": 5, "f": 1434},
  {"h": "扫", "p": "sǎo", "sp": "sao", "r": "手", "sc": 6, "l": 3, "f": 1435},
  {"h": "敏", "p": "mǐn", "sp": "min", "r": "攴", "sc": 11, "l": 5, "f": 1436},
  {"h": "碍", "p": "ài", "sp": "ai", "r": "石", "sc": 13, "l": 5, "f": 1437},
  {"h": "殖", "p": "zhí", "sp": "zhi", "r": "歹", "sc": 12, "l": 5, "f": 1438},
  {"h": "详", "p": "xiáng", "sp": "xiang", "r": "言", "sc": 8, "l": 5, "f": 1439},
  {"h": "迪", "p": "dí", "sp": "di", "r": "辵", "sc": 8, "l": 5, "f": 1440},
  {"h": "矛", "p": "máo", "sp": "mao", "r": "矛", "sc": 5, "l": 4, "f": 1441},
  {"h": "霍", "p": "huò", "sp": "huo", "r": "雨", "sc": 16, "l": 5, "f": 1442},
  {"h": "允", "p": "yǔn", "sp": "yun", "r": "儿", "sc": 4, "l": 5, "f": 1443},
  {"h": "幅", "p": "fú", "sp": "fu", "r": "巾", "sc": 12, "l": 3, "f": 1444},
  {"h": "撒", "p": "sā", "sp": "sa", "r": "手", "sc": 15, "l": 5, "f": 1445},
  {
    "h": "剩",
    "p": "shèng",
    "sp": "sheng",
    "r": "刀",
    "sc": 12,
    "l": 5,
    "f": 1446
  },
  {"h": "凯", "p": "kǎi", "sp": "kai", "r": "几", "sc": 8, "l": 5, "f": 1447},
  {"h": "颗", "p": "kē", "sp": "ke", "r": "頁", "sc": 14, "l": 5, "f": 1448},
  {"h": "骂", "p": "mà", "sp": "ma", "r": "口", "sc": 9, "l": 5, "f": 1449},
  {
    "h": "赏",
    "p": "shǎng",
    "sp": "shang",
    "r": "貝",
    "sc": 12,
    "l": 4,
    "f": 1450
  },
  {"h": "液", "p": "yè", "sp": "ye", "r": "水", "sc": 11, "l": 5, "f": 1451},
  {"h": "番", "p": "fān", "sp": "fan", "r": "田", "sc": 12, "l": 5, "f": 1452},
  {
    "h": "箱",
    "p": "xiāng",
    "sp": "xiang",
    "r": "竹",
    "sc": 15,
    "l": 3,
    "f": 1453
  },
  {"h": "贴", "p": "tiē", "sp": "tie", "r": "貝", "sc": 9, "l": 4, "f": 1454},
  {"h": "漫", "p": "màn", "sp": "man", "r": "水", "sc": 14, "l": 5, "f": 1455},
  {"h": "酸", "p": "suān", "sp": "suan", "r": "酉", "sc": 14, "l": 4, "f": 1456},
  {"h": "郎", "p": "láng", "sp": "lang", "r": "邑", "sc": 8, "l": 4, "f": 1457},
  {"h": "腰", "p": "yāo", "sp": "yao", "r": "肉", "sc": 13, "l": 3, "f": 1458},
  {"h": "舒", "p": "shū", "sp": "shu", "r": "舌", "sc": 12, "l": 2, "f": 1459},
  {"h": "眉", "p": "méi", "sp": "mei", "r": "目", "sc": 9, "l": 5, "f": 1460},
  {"h": "忧", "p": "yōu", "sp": "you", "r": "心", "sc": 7, "l": 5, "f": 1461},
  {"h": "浮", "p": "fú", "sp": "fu", "r": "水", "sc": 10, "l": 5, "f": 1462},
  {"h": "辛", "p": "xīn", "sp": "xin", "r": "辛", "sc": 7, "l": 5, "f": 1463},
  {"h": "恋", "p": "liàn", "sp": "lian", "r": "心", "sc": 10, "l": 5, "f": 1464},
  {"h": "餐", "p": "cān", "sp": "can", "r": "食", "sc": 16, "l": 2, "f": 1465},
  {"h": "吓", "p": "xià", "sp": "xia", "r": "口", "sc": 6, "l": 5, "f": 1466},
  {"h": "挺", "p": "tǐng", "sp": "ting", "r": "手", "sc": 9, "l": 2, "f": 1467},
  {"h": "励", "p": "lì", "sp": "li", "r": "力", "sc": 7, "l": 5, "f": 1468},
  {"h": "辞", "p": "cí", "sp": "ci", "r": "舌", "sc": 13, "l": 5, "f": 1469},
  {"h": "艘", "p": "sōu", "sp": "sou", "r": "舟", "sc": 15, "l": 5, "f": 1470},
  {"h": "键", "p": "jiàn", "sp": "jian", "r": "金", "sc": 13, "l": 4, "f": 1471},
  {"h": "伍", "p": "wǔ", "sp": "wu", "r": "人", "sc": 6, "l": 5, "f": 1472},
  {"h": "峰", "p": "fēng", "sp": "feng", "r": "山", "sc": 10, "l": 5, "f": 1473},
  {"h": "尺", "p": "chǐ", "sp": "chi", "r": "尸", "sc": 4, "l": 4, "f": 1474},
  {"h": "昨", "p": "zuó", "sp": "zuo", "r": "日", "sc": 9, "l": 1, "f": 1475},
  {"h": "黎", "p": "lí", "sp": "li", "r": "黍", "sc": 15, "l": 5, "f": 1476},
  {"h": "辈", "p": "bèi", "sp": "bei", "r": "非", "sc": 12, "l": 5, "f": 1477},
  {"h": "贯", "p": "guàn", "sp": "guan", "r": "貝", "sc": 8, "l": 5, "f": 1478},
  {"h": "侦", "p": "zhēn", "sp": "zhen", "r": "人", "sc": 8, "l": 5, "f": 1479},
  {"h": "滑", "p": "huá", "sp": "hua", "r": "水", "sc": 12, "l": 5, "f": 1480},
  {"h": "券", "p": "quàn", "sp": "quan", "r": "刀", "sc": 8, "l": 5, "f": 1481},
  {
    "h": "崇",
    "p": "chóng",
    "sp": "chong",
    "r": "山",
    "sc": 11,
    "l": 5,
    "f": 1482
  },
  {"h": "扰", "p": "rǎo", "sp": "rao", "r": "手", "sc": 7, "l": 4, "f": 1483},
  {"h": "宪", "p": "xiàn", "sp": "xian", "r": "宀", "sc": 9, "l": 5, "f": 1484},
  {"h": "绕", "p": "rào", "sp": "rao", "r": "糸", "sc": 9, "l": 5, "f": 1485},
  {"h": "趋", "p": "qū", "sp": "qu", "r": "走", "sc": 12, "l": 4, "f": 1486},
  {"h": "慈", "p": "cí", "sp": "ci", "r": "心", "sc": 13, "l": 5, "f": 1487},
  {"h": "乔", "p": "qiáo", "sp": "qiao", "r": "大", "sc": 6, "l": 5, "f": 1488},
  {"h": "阅", "p": "yuè", "sp": "yue", "r": "門", "sc": 10, "l": 4, "f": 1489},
  {"h": "汗", "p": "hàn", "sp": "han", "r": "水", "sc": 6, "l": 4, "f": 1490},
  {"h": "枝", "p": "zhī", "sp": "zhi", "r": "木", "sc": 8, "l": 5, "f": 1491},
  {"h": "拖", "p": "tuō", "sp": "tuo", "r": "手", "sc": 8, "l": 5, "f": 1492},
  {"h": "墨", "p": "mò", "sp": "mo", "r": "土", "sc": 15, "l": 4, "f": 1493},
  {"h": "胁", "p": "xié", "sp": "xie", "r": "肉", "sc": 8, "l": 5, "f": 1494},
  {"h": "插", "p": "chā", "sp": "cha", "r": "手", "sc": 12, "l": 5, "f": 1495},
  {"h": "箭", "p": "jiàn", "sp": "jian", "r": "竹", "sc": 15, "l": 5, "f": 1496},
  {"h": "腊", "p": "là", "sp": "la", "r": "肉", "sc": 12, "l": 4, "f": 1497},
  {"h": "粉", "p": "fěn", "sp": "fen", "r": "米", "sc": 10, "l": 4, "f": 1498},
  {"h": "泥", "p": "ní", "sp": "ni", "r": "水", "sc": 8, "l": 5, "f": 1499},
  {"h": "氏", "p": "shì", "sp": "shi", "r": "氏", "sc": 4, "l": 5, "f": 1500},
  {"h": "彭", "p": "péng", "sp": "peng", "r": "彡", "sc": 12, "l": 5, "f": 1501},
  {"h": "拔", "p": "bá", "sp": "ba", "r": "手", "sc": 8, "l": 5, "f": 1502},
  {"h": "骗", "p": "piàn", "sp": "pian", "r": "馬", "sc": 12, "l": 4, "f": 1503},
  {"h": "凤", "p": "fèng", "sp": "feng", "r": "几", "sc": 4, "l": 5, "f": 1504},
  {"h": "慧", "p": "huì", "sp": "hui", "r": "心", "sc": 15, "l": 5, "f": 1505},
  {"h": "媒", "p": "méi", "sp": "mei", "r": "女", "sc": 12, "l": 3, "f": 1506},
  {"h": "佩", "p": "pèi", "sp": "pei", "r": "人", "sc": 8, "l": 5, "f": 1507},
  {"h": "愤", "p": "fèn", "sp": "fen", "r": "心", "sc": 12, "l": 5, "f": 1508},
  {"h": "扑", "p": "pū", "sp": "pu", "r": "手", "sc": 5, "l": 5, "f": 1509},
  {"h": "龄", "p": "líng", "sp": "ling", "r": "齒", "sc": 13, "l": 5, "f": 1510},
  {"h": "驱", "p": "qū", "sp": "qu", "r": "馬", "sc": 7, "l": 5, "f": 1511},
  {"h": "惜", "p": "xī", "sp": "xi", "r": "心", "sc": 11, "l": 5, "f": 1512},
  {"h": "豪", "p": "háo", "sp": "hao", "r": "豕", "sc": 14, "l": 5, "f": 1513},
  {"h": "掩", "p": "yǎn", "sp": "yan", "r": "手", "sc": 11, "l": 5, "f": 1514},
  {"h": "兼", "p": "jiān", "sp": "jian", "r": "八", "sc": 10, "l": 5, "f": 1515},
  {"h": "跃", "p": "yuè", "sp": "yue", "r": "足", "sc": 11, "l": 5, "f": 1516},
  {"h": "尸", "p": "shī", "sp": "shi", "r": "尸", "sc": 3, "l": 5, "f": 1517},
  {"h": "肃", "p": "sù", "sp": "su", "r": "聿", "sc": 8, "l": 5, "f": 1518},
  {"h": "帕", "p": "pà", "sp": "pa", "r": "巾", "sc": 8, "l": 5, "f": 1519},
  {"h": "驶", "p": "shǐ", "sp": "shi", "r": "馬", "sc": 8, "l": 5, "f": 1520},
  {"h": "堡", "p": "bǎo", "sp": "bao", "r": "土", "sc": 12, "l": 5, "f": 1521},
  {"h": "届", "p": "jiè", "sp": "jie", "r": "尸", "sc": 8, "l": 5, "f": 1522},
  {"h": "欣", "p": "xīn", "sp": "xin", "r": "欠", "sc": 8, "l": 5, "f": 1523},
  {"h": "惠", "p": "huì", "sp": "hui", "r": "心", "sc": 12, "l": 5, "f": 1524},
  {"h": "册", "p": "cè", "sp": "ce", "r": "冂", "sc": 5, "l": 5, "f": 1525},
  {"h": "储", "p": "chǔ", "sp": "chu", "r": "人", "sc": 12, "l": 5, "f": 1526},
  {"h": "飘", "p": "piāo", "sp": "piao", "r": "風", "sc": 15, "l": 5, "f": 1527},
  {"h": "桑", "p": "sāng", "sp": "sang", "r": "木", "sc": 10, "l": 5, "f": 1528},
  {"h": "闲", "p": "xián", "sp": "xian", "r": "門", "sc": 7, "l": 5, "f": 1529},
  {"h": "惨", "p": "cǎn", "sp": "can", "r": "心", "sc": 11, "l": 5, "f": 1530},
  {"h": "洁", "p": "jié", "sp": "jie", "r": "水", "sc": 9, "l": 5, "f": 1531},
  {"h": "踪", "p": "zōng", "sp": "zong", "r": "足", "sc": 15, "l": 5, "f": 1532},
  {"h": "勃", "p": "bó", "sp": "bo", "r": "力", "sc": 9, "l": 5, "f": 1533},
  {"h": "宾", "p": "bīn", "sp": "bin", "r": "宀", "sc": 10, "l": 2, "f": 1534},
  {"h": "频", "p": "pín", "sp": "pin", "r": "頁", "sc": 13, "l": 5, "f": 1535},
  {"h": "仇", "p": "chóu", "sp": "chou", "r": "人", "sc": 4, "l": 5, "f": 1536},
  {"h": "磨", "p": "mó", "sp": "mo", "r": "石", "sc": 16, "l": 5, "f": 1537},
  {"h": "递", "p": "dì", "sp": "di", "r": "辵", "sc": 10, "l": 4, "f": 1538},
  {"h": "邪", "p": "xié", "sp": "xie", "r": "邑", "sc": 6, "l": 5, "f": 1539},
  {
    "h": "撞",
    "p": "zhuàng",
    "sp": "zhuang",
    "r": "手",
    "sc": 15,
    "l": 5,
    "f": 1540
  },
  {"h": "拟", "p": "nǐ", "sp": "ni", "r": "手", "sc": 7, "l": 5, "f": 1541},
  {"h": "滚", "p": "gǔn", "sp": "gun", "r": "水", "sc": 13, "l": 5, "f": 1542},
  {"h": "奏", "p": "zòu", "sp": "zou", "r": "大", "sc": 9, "l": 5, "f": 1543},
  {"h": "巡", "p": "xún", "sp": "xun", "r": "巛", "sc": 6, "l": 5, "f": 1544},
  {"h": "颜", "p": "yán", "sp": "yan", "r": "頁", "sc": 15, "l": 2, "f": 1545},
  {"h": "剂", "p": "jì", "sp": "ji", "r": "刀", "sc": 8, "l": 5, "f": 1546},
  {"h": "绩", "p": "jī", "sp": "ji", "r": "糸", "sc": 11, "l": 2, "f": 1547},
  {"h": "贡", "p": "gòng", "sp": "gong", "r": "貝", "sc": 7, "l": 5, "f": 1548},
  {"h": "疯", "p": "fēng", "sp": "feng", "r": "疒", "sc": 9, "l": 5, "f": 1549},
  {"h": "坡", "p": "pō", "sp": "po", "r": "土", "sc": 8, "l": 5, "f": 1550},
  {"h": "瞧", "p": "qiáo", "sp": "qiao", "r": "目", "sc": 17, "l": 5, "f": 1551},
  {"h": "截", "p": "jié", "sp": "jie", "r": "戈", "sc": 14, "l": 5, "f": 1552},
  {"h": "燃", "p": "rán", "sp": "ran", "r": "火", "sc": 16, "l": 4, "f": 1553},
  {"h": "焦", "p": "jiāo", "sp": "jiao", "r": "火", "sc": 12, "l": 5, "f": 1554},
  {"h": "殿", "p": "diàn", "sp": "dian", "r": "殳", "sc": 13, "l": 5, "f": 1555},
  {"h": "伪", "p": "wěi", "sp": "wei", "r": "人", "sc": 6, "l": 5, "f": 1556},
  {"h": "柳", "p": "liǔ", "sp": "liu", "r": "木", "sc": 9, "l": 5, "f": 1557},
  {"h": "锁", "p": "suǒ", "sp": "suo", "r": "金", "sc": 12, "l": 5, "f": 1558},
  {"h": "逼", "p": "bī", "sp": "bi", "r": "辵", "sc": 12, "l": 5, "f": 1559},
  {"h": "颇", "p": "pō", "sp": "po", "r": "頁", "sc": 11, "l": 5, "f": 1560},
  {"h": "昏", "p": "hūn", "sp": "hun", "r": "日", "sc": 8, "l": 5, "f": 1561},
  {"h": "劝", "p": "quàn", "sp": "quan", "r": "力", "sc": 4, "l": 5, "f": 1562},
  {"h": "呈", "p": "chéng", "sp": "cheng", "r": "口", "sc": 7, "l": 5, "f": 1563},
  {"h": "搜", "p": "sōu", "sp": "sou", "r": "手", "sc": 12, "l": 5, "f": 1564},
  {"h": "勤", "p": "qín", "sp": "qin", "r": "力", "sc": 13, "l": 5, "f": 1565},
  {"h": "戒", "p": "jiè", "sp": "jie", "r": "戈", "sc": 7, "l": 5, "f": 1566},
  {"h": "驾", "p": "jià", "sp": "jia", "r": "馬", "sc": 8, "l": 5, "f": 1567},
  {"h": "漂", "p": "piào", "sp": "piao", "r": "水", "sc": 14, "l": 1, "f": 1568},
  {"h": "饮", "p": "yǐn", "sp": "yin", "r": "食", "sc": 7, "l": 4, "f": 1569},
  {"h": "曹", "p": "cáo", "sp": "cao", "r": "曰", "sc": 11, "l": 5, "f": 1570},
  {"h": "朵", "p": "duǒ", "sp": "duo", "r": "木", "sc": 6, "l": 3, "f": 1571},
  {"h": "仔", "p": "zǐ", "sp": "zi", "r": "人", "sc": 5, "l": 5, "f": 1572},
  {"h": "柔", "p": "róu", "sp": "rou", "r": "木", "sc": 9, "l": 5, "f": 1573},
  {"h": "俩", "p": "liǎ", "sp": "lia", "r": "人", "sc": 9, "l": 4, "f": 1574},
  {"h": "孟", "p": "mèng", "sp": "meng", "r": "子", "sc": 8, "l": 5, "f": 1575},
  {"h": "腐", "p": "fǔ", "sp": "fu", "r": "肉", "sc": 14, "l": 4, "f": 1576},
  {"h": "幼", "p": "yòu", "sp": "you", "r": "幺", "sc": 5, "l": 4, "f": 1577},
  {"h": "践", "p": "jiàn", "sp": "jian", "r": "足", "sc": 12, "l": 5, "f": 1578},
  {"h": "籍", "p": "jí", "sp": "ji", "r": "竹", "sc": 20, "l": 5, "f": 1579},
  {"h": "牧", "p": "mù", "sp": "mu", "r": "牛", "sc": 8, "l": 5, "f": 1580},
  {
    "h": "凉",
    "p": "liáng",
    "sp": "liang",
    "r": "冫",
    "sc": 10,
    "l": 2,
    "f": 1581
  },
  {"h": "牲", "p": "shēng", "sp": "sheng", "r": "牛", "sc": 9, "l": 5, "f": 1582},
  {"h": "佳", "p": "jiā", "sp": "jia", "r": "人", "sc": 8, "l": 5, "f": 1583},
  {"h": "娜", "p": "nà", "sp": "na", "r": "女", "sc": 9, "l": 5, "f": 1584},
  {"h": "浓", "p": "nóng", "sp": "nong", "r": "水", "sc": 9, "l": 4, "f": 1585},
  {"h": "芳", "p": "fāng", "sp": "fang", "r": "艸", "sc": 7, "l": 5, "f": 1586},
  {"h": "稿", "p": "gǎo", "sp": "gao", "r": "禾", "sc": 15, "l": 5, "f": 1587},
  {"h": "竹", "p": "zhú", "sp": "zhu", "r": "竹", "sc": 6, "l": 5, "f": 1588},
  {"h": "腹", "p": "fù", "sp": "fu", "r": "肉", "sc": 13, "l": 5, "f": 1589},
  {"h": "跌", "p": "diē", "sp": "die", "r": "足", "sc": 12, "l": 5, "f": 1590},
  {"h": "逻", "p": "luó", "sp": "luo", "r": "辵", "sc": 11, "l": 5, "f": 1591},
  {"h": "垂", "p": "chuí", "sp": "chui", "r": "土", "sc": 8, "l": 5, "f": 1592},
  {"h": "遵", "p": "zūn", "sp": "zun", "r": "辵", "sc": 15, "l": 5, "f": 1593},
  {"h": "脉", "p": "mài", "sp": "mai", "r": "肉", "sc": 9, "l": 5, "f": 1594},
  {"h": "貌", "p": "mào", "sp": "mao", "r": "豸", "sc": 14, "l": 5, "f": 1595},
  {"h": "柏", "p": "bǎi", "sp": "bai", "r": "木", "sc": 9, "l": 5, "f": 1596},
  {"h": "狱", "p": "yù", "sp": "yu", "r": "犬", "sc": 9, "l": 5, "f": 1597},
  {"h": "猜", "p": "cāi", "sp": "cai", "r": "犬", "sc": 11, "l": 4, "f": 1598},
  {"h": "怜", "p": "lián", "sp": "lian", "r": "心", "sc": 8, "l": 4, "f": 1599},
  {"h": "惑", "p": "huò", "sp": "huo", "r": "心", "sc": 12, "l": 5, "f": 1600},
  {"h": "陶", "p": "táo", "sp": "tao", "r": "阜", "sc": 10, "l": 5, "f": 1601},
  {"h": "兽", "p": "shòu", "sp": "shou", "r": "八", "sc": 11, "l": 5, "f": 1602},
  {"h": "帐", "p": "zhàng", "sp": "zhang", "r": "巾", "sc": 7, "l": 5, "f": 1603},
  {"h": "饰", "p": "shì", "sp": "shi", "r": "食", "sc": 8, "l": 5, "f": 1604},
  {"h": "贷", "p": "dài", "sp": "dai", "r": "貝", "sc": 9, "l": 5, "f": 1605},
  {"h": "昌", "p": "chāng", "sp": "chang", "r": "日", "sc": 8, "l": 5, "f": 1606},
  {"h": "叙", "p": "xù", "sp": "xu", "r": "又", "sc": 9, "l": 5, "f": 1607},
  {"h": "躺", "p": "tǎng", "sp": "tang", "r": "身", "sc": 15, "l": 4, "f": 1608},
  {"h": "钢", "p": "gāng", "sp": "gang", "r": "金", "sc": 9, "l": 4, "f": 1609},
  {"h": "沟", "p": "gōu", "sp": "gou", "r": "水", "sc": 7, "l": 5, "f": 1610},
  {"h": "寄", "p": "jì", "sp": "ji", "r": "宀", "sc": 11, "l": 4, "f": 1611},
  {"h": "扶", "p": "fú", "sp": "fu", "r": "手", "sc": 7, "l": 5, "f": 1612},
  {"h": "铺", "p": "pù", "sp": "pu", "r": "金", "sc": 12, "l": 5, "f": 1613},
  {"h": "邓", "p": "dèng", "sp": "deng", "r": "邑", "sc": 4, "l": 5, "f": 1614},
  {"h": "寿", "p": "shòu", "sp": "shou", "r": "寸", "sc": 7, "l": 5, "f": 1615},
  {"h": "惧", "p": "jù", "sp": "ju", "r": "心", "sc": 11, "l": 5, "f": 1616},
  {"h": "询", "p": "xún", "sp": "xun", "r": "言", "sc": 8, "l": 5, "f": 1617},
  {"h": "汤", "p": "tāng", "sp": "tang", "r": "水", "sc": 6, "l": 3, "f": 1618},
  {"h": "盗", "p": "dào", "sp": "dao", "r": "皿", "sc": 11, "l": 5, "f": 1619},
  {"h": "肥", "p": "féi", "sp": "fei", "r": "肉", "sc": 8, "l": 4, "f": 1620},
  {"h": "尝", "p": "cháng", "sp": "chang", "r": "小", "sc": 9, "l": 4, "f": 1621},
  {"h": "匆", "p": "cōng", "sp": "cong", "r": "勹", "sc": 5, "l": 5, "f": 1622},
  {"h": "辉", "p": "huī", "sp": "hui", "r": "小", "sc": 12, "l": 5, "f": 1623},
  {"h": "奈", "p": "nài", "sp": "nai", "r": "大", "sc": 8, "l": 5, "f": 1624},
  {"h": "扣", "p": "kòu", "sp": "kou", "r": "手", "sc": 6, "l": 5, "f": 1625},
  {"h": "廷", "p": "tíng", "sp": "ting", "r": "廴", "sc": 6, "l": 5, "f": 1626},
  {"h": "澳", "p": "ào", "sp": "ao", "r": "水", "sc": 15, "l": 5, "f": 1627},
  {"h": "嘛", "p": "ma", "sp": "ma", "r": "口", "sc": 14, "l": 5, "f": 1628},
  {"h": "董", "p": "dǒng", "sp": "dong", "r": "艸", "sc": 12, "l": 5, "f": 1629},
  {"h": "迁", "p": "qiān", "sp": "qian", "r": "辵", "sc": 6, "l": 5, "f": 1630},
  {"h": "凝", "p": "níng", "sp": "ning", "r": "冫", "sc": 16, "l": 5, "f": 1631},
  {"h": "慰", "p": "wèi", "sp": "wei", "r": "心", "sc": 15, "l": 5, "f": 1632},
  {"h": "厌", "p": "yàn", "sp": "yan", "r": "厂", "sc": 6, "l": 4, "f": 1633},
  {"h": "脏", "p": "zàng", "sp": "zang", "r": "肉", "sc": 10, "l": 2, "f": 1634},
  {"h": "腾", "p": "téng", "sp": "teng", "r": "肉", "sc": 13, "l": 5, "f": 1635},
  {"h": "幽", "p": "yōu", "sp": "you", "r": "幺", "sc": 9, "l": 5, "f": 1636},
  {"h": "怨", "p": "yuàn", "sp": "yuan", "r": "心", "sc": 9, "l": 5, "f": 1637},
  {"h": "鞋", "p": "xié", "sp": "xie", "r": "革", "sc": 15, "l": 2, "f": 1638},
  {"h": "丢", "p": "diū", "sp": "diu", "r": "一", "sc": 6, "l": 4, "f": 1639},
  {"h": "埋", "p": "mái", "sp": "mai", "r": "土", "sc": 10, "l": 5, "f": 1640},
  {"h": "泉", "p": "quán", "sp": "quan", "r": "水", "sc": 9, "l": 4, "f": 1641},
  {"h": "涌", "p": "yǒng", "sp": "yong", "r": "水", "sc": 10, "l": 5, "f": 1642},
  {"h": "辖", "p": "xiá", "sp": "xia", "r": "車", "sc": 14, "l": 5, "f": 1643},
  {"h": "躲", "p": "duǒ", "sp": "duo", "r": "身", "sc": 13, "l": 5, "f": 1644},
  {"h": "晋", "p": "jìn", "sp": "jin", "r": "日", "sc": 10, "l": 5, "f": 1645},
  {"h": "紫", "p": "zǐ", "sp": "zi", "r": "糸", "sc": 12, "l": 5, "f": 1646},
  {"h": "艰", "p": "jiān", "sp": "jian", "r": "又", "sc": 8, "l": 5, "f": 1647},
  {"h": "魏", "p": "wèi", "sp": "wei", "r": "鬼", "sc": 17, "l": 5, "f": 1648},
  {"h": "吾", "p": "wú", "sp": "wu", "r": "口", "sc": 7, "l": 5, "f": 1649},
  {
    "h": "慌",
    "p": "huāng",
    "sp": "huang",
    "r": "心",
    "sc": 12,
    "l": 5,
    "f": 1650
  },
  {"h": "祝", "p": "zhù", "sp": "zhu", "r": "示", "sc": 9, "l": 3, "f": 1651},
  {"h": "邮", "p": "yóu", "sp": "you", "r": "邑", "sc": 7, "l": 3, "f": 1652},
  {"h": "吐", "p": "tǔ", "sp": "tu", "r": "口", "sc": 6, "l": 5, "f": 1653},
  {"h": "狠", "p": "hěn", "sp": "hen", "r": "犬", "sc": 9, "l": 5, "f": 1654},
  {"h": "鉴", "p": "jiàn", "sp": "jian", "r": "金", "sc": 13, "l": 5, "f": 1655},
  {"h": "曰", "p": "yuē", "sp": "yue", "r": "曰", "sc": 4, "l": 5, "f": 1656},
  {"h": "械", "p": "xiè", "sp": "xie", "r": "木", "sc": 11, "l": 5, "f": 1657},
  {"h": "咬", "p": "yǎo", "sp": "yao", "r": "口", "sc": 9, "l": 5, "f": 1658},
  {"h": "邻", "p": "lín", "sp": "lin", "r": "邑", "sc": 7, "l": 3, "f": 1659},
  {"h": "赤", "p": "chì", "sp": "chi", "r": "赤", "sc": 7, "l": 5, "f": 1660},
  {"h": "挤", "p": "jǐ", "sp": "ji", "r": "手", "sc": 9, "l": 5, "f": 1661},
  {"h": "弯", "p": "wān", "sp": "wan", "r": "弓", "sc": 9, "l": 4, "f": 1662},
  {"h": "椅", "p": "yǐ", "sp": "yi", "r": "木", "sc": 12, "l": 1, "f": 1663},
  {"h": "陪", "p": "péi", "sp": "pei", "r": "阜", "sc": 10, "l": 4, "f": 1664},
  {"h": "割", "p": "gē", "sp": "ge", "r": "刀", "sc": 12, "l": 5, "f": 1665},
  {"h": "揭", "p": "jiē", "sp": "jie", "r": "手", "sc": 12, "l": 5, "f": 1666},
  {"h": "韦", "p": "wéi", "sp": "wei", "r": "韋", "sc": 4, "l": 5, "f": 1667},
  {"h": "悟", "p": "wù", "sp": "wu", "r": "心", "sc": 10, "l": 5, "f": 1668},
  {"h": "聪", "p": "cōng", "sp": "cong", "r": "耳", "sc": 15, "l": 3, "f": 1669},
  {"h": "雾", "p": "wù", "sp": "wu", "r": "雨", "sc": 13, "l": 5, "f": 1670},
  {"h": "锋", "p": "fēng", "sp": "feng", "r": "金", "sc": 12, "l": 5, "f": 1671},
  {"h": "梯", "p": "tī", "sp": "ti", "r": "木", "sc": 11, "l": 3, "f": 1672},
  {"h": "猫", "p": "māo", "sp": "mao", "r": "犬", "sc": 11, "l": 1, "f": 1673},
  {
    "h": "祥",
    "p": "xiáng",
    "sp": "xiang",
    "r": "示",
    "sc": 10,
    "l": 5,
    "f": 1674
  },
  {"h": "阔", "p": "kuò", "sp": "kuo", "r": "門", "sc": 12, "l": 5, "f": 1675},
  {"h": "誉", "p": "yù", "sp": "yu", "r": "言", "sc": 13, "l": 5, "f": 1676},
  {"h": "筹", "p": "chóu", "sp": "chou", "r": "竹", "sc": 13, "l": 5, "f": 1677},
  {"h": "丛", "p": "cóng", "sp": "cong", "r": "一", "sc": 5, "l": 5, "f": 1678},
  {"h": "牵", "p": "qiān", "sp": "qian", "r": "大", "sc": 9, "l": 5, "f": 1679},
  {"h": "鸣", "p": "míng", "sp": "ming", "r": "口", "sc": 8, "l": 5, "f": 1680},
  {"h": "沈", "p": "chén", "sp": "chen", "r": "水", "sc": 7, "l": 5, "f": 1681},
  {"h": "阁", "p": "gé", "sp": "ge", "r": "門", "sc": 9, "l": 5, "f": 1682},
  {"h": "穆", "p": "mù", "sp": "mu", "r": "禾", "sc": 16, "l": 5, "f": 1683},
  {"h": "屈", "p": "qū", "sp": "qu", "r": "尸", "sc": 8, "l": 5, "f": 1684},
  {"h": "旨", "p": "zhǐ", "sp": "zhi", "r": "日", "sc": 6, "l": 5, "f": 1685},
  {"h": "袖", "p": "xiù", "sp": "xiu", "r": "衣", "sc": 10, "l": 5, "f": 1686},
  {"h": "猎", "p": "liè", "sp": "lie", "r": "犬", "sc": 11, "l": 5, "f": 1687},
  {"h": "臂", "p": "bì", "sp": "bi", "r": "肉", "sc": 17, "l": 5, "f": 1688},
  {"h": "蛇", "p": "shé", "sp": "she", "r": "虫", "sc": 11, "l": 5, "f": 1689},
  {"h": "贺", "p": "hè", "sp": "he", "r": "貝", "sc": 9, "l": 5, "f": 1690},
  {"h": "柱", "p": "zhù", "sp": "zhu", "r": "木", "sc": 9, "l": 5, "f": 1691},
  {"h": "抛", "p": "pāo", "sp": "pao", "r": "手", "sc": 7, "l": 5, "f": 1692},
  {"h": "鼠", "p": "shǔ", "sp": "shu", "r": "鼠", "sc": 13, "l": 5, "f": 1693},
  {"h": "瑟", "p": "sè", "sp": "se", "r": "玉", "sc": 13, "l": 5, "f": 1694},
  {"h": "戈", "p": "gē", "sp": "ge", "r": "戈", "sc": 4, "l": 5, "f": 1695},
  {"h": "牢", "p": "láo", "sp": "lao", "r": "牛", "sc": 7, "l": 5, "f": 1696},
  {"h": "逊", "p": "xùn", "sp": "xun", "r": "辵", "sc": 9, "l": 5, "f": 1697},
  {"h": "迈", "p": "mài", "sp": "mai", "r": "辵", "sc": 6, "l": 5, "f": 1698},
  {"h": "欺", "p": "qī", "sp": "qi", "r": "欠", "sc": 12, "l": 5, "f": 1699},
  {"h": "吨", "p": "dūn", "sp": "dun", "r": "口", "sc": 7, "l": 5, "f": 1700},
  {"h": "琴", "p": "qín", "sp": "qin", "r": "玉", "sc": 12, "l": 4, "f": 1701},
  {
    "h": "衰",
    "p": "shuāi",
    "sp": "shuai",
    "r": "衣",
    "sc": 10,
    "l": 5,
    "f": 1702
  },
  {"h": "瓶", "p": "píng", "sp": "ping", "r": "瓦", "sc": 10, "l": 2, "f": 1703},
  {"h": "恼", "p": "nǎo", "sp": "nao", "r": "心", "sc": 9, "l": 4, "f": 1704},
  {"h": "燕", "p": "yàn", "sp": "yan", "r": "火", "sc": 16, "l": 4, "f": 1705},
  {"h": "仲", "p": "zhòng", "sp": "zhong", "r": "人", "sc": 6, "l": 5, "f": 1706},
  {"h": "诱", "p": "yòu", "sp": "you", "r": "言", "sc": 9, "l": 5, "f": 1707},
  {"h": "狼", "p": "láng", "sp": "lang", "r": "犬", "sc": 10, "l": 5, "f": 1708},
  {"h": "池", "p": "chí", "sp": "chi", "r": "水", "sc": 6, "l": 5, "f": 1709},
  {"h": "疼", "p": "téng", "sp": "teng", "r": "疒", "sc": 10, "l": 2, "f": 1710},
  {"h": "卢", "p": "lú", "sp": "lu", "r": "卜", "sc": 5, "l": 5, "f": 1711},
  {"h": "仗", "p": "zhàng", "sp": "zhang", "r": "人", "sc": 5, "l": 5, "f": 1712},
  {"h": "冠", "p": "guān", "sp": "guan", "r": "冖", "sc": 9, "l": 5, "f": 1713},
  {"h": "粒", "p": "lì", "sp": "li", "r": "米", "sc": 11, "l": 5, "f": 1714},
  {"h": "遥", "p": "yáo", "sp": "yao", "r": "辵", "sc": 13, "l": 5, "f": 1715},
  {"h": "吕", "p": "lǚ", "sp": "lu", "r": "口", "sc": 6, "l": 5, "f": 1716},
  {"h": "玄", "p": "xuán", "sp": "xuan", "r": "玄", "sc": 5, "l": 5, "f": 1717},
  {"h": "尘", "p": "chén", "sp": "chen", "r": "小", "sc": 6, "l": 5, "f": 1718},
  {"h": "冯", "p": "féng", "sp": "feng", "r": "馬", "sc": 5, "l": 5, "f": 1719},
  {"h": "抚", "p": "fǔ", "sp": "fu", "r": "手", "sc": 7, "l": 5, "f": 1720},
  {"h": "浅", "p": "qiǎn", "sp": "qian", "r": "水", "sc": 8, "l": 4, "f": 1721},
  {"h": "敦", "p": "dūn", "sp": "dun", "r": "攴", "sc": 12, "l": 5, "f": 1722},
  {"h": "纠", "p": "jiū", "sp": "jiu", "r": "糸", "sc": 5, "l": 5, "f": 1723},
  {"h": "钻", "p": "zuān", "sp": "zuan", "r": "金", "sc": 10, "l": 5, "f": 1724},
  {"h": "晶", "p": "jīng", "sp": "jing", "r": "日", "sc": 12, "l": 5, "f": 1725},
  {"h": "岂", "p": "qǐ", "sp": "qi", "r": "山", "sc": 6, "l": 5, "f": 1726},
  {"h": "峡", "p": "xiá", "sp": "xia", "r": "山", "sc": 9, "l": 5, "f": 1727},
  {"h": "苍", "p": "cāng", "sp": "cang", "r": "艸", "sc": 7, "l": 5, "f": 1728},
  {"h": "喷", "p": "pēn", "sp": "pen", "r": "口", "sc": 12, "l": 5, "f": 1729},
  {"h": "耗", "p": "hào", "sp": "hao", "r": "耒", "sc": 10, "l": 5, "f": 1730},
  {"h": "凌", "p": "líng", "sp": "ling", "r": "冫", "sc": 10, "l": 5, "f": 1731},
  {"h": "敲", "p": "qiāo", "sp": "qiao", "r": "攴", "sc": 14, "l": 5, "f": 1732},
  {"h": "菌", "p": "jūn", "sp": "jun", "r": "艸", "sc": 11, "l": 5, "f": 1733},
  {"h": "赔", "p": "péi", "sp": "pei", "r": "貝", "sc": 12, "l": 5, "f": 1734},
  {"h": "涂", "p": "tú", "sp": "tu", "r": "水", "sc": 10, "l": 5, "f": 1735},
  {"h": "粹", "p": "cuì", "sp": "cui", "r": "米", "sc": 14, "l": 5, "f": 1736},
  {"h": "扁", "p": "biǎn", "sp": "bian", "r": "戶", "sc": 9, "l": 5, "f": 1737},
  {"h": "亏", "p": "kuī", "sp": "kui", "r": "二", "sc": 3, "l": 5, "f": 1738},
  {"h": "寂", "p": "jì", "sp": "ji", "r": "宀", "sc": 11, "l": 5, "f": 1739},
  {"h": "煤", "p": "méi", "sp": "mei", "r": "火", "sc": 13, "l": 5, "f": 1740},
  {
    "h": "熊",
    "p": "xióng",
    "sp": "xiong",
    "r": "火",
    "sc": 14,
    "l": 3,
    "f": 1741
  },
  {"h": "恭", "p": "gōng", "sp": "gong", "r": "心", "sc": 10, "l": 5, "f": 1742},
  {"h": "湿", "p": "shī", "sp": "shi", "r": "水", "sc": 12, "l": 4, "f": 1743},
  {"h": "循", "p": "xún", "sp": "xun", "r": "彳", "sc": 12, "l": 5, "f": 1744},
  {"h": "暖", "p": "nuǎn", "sp": "nuan", "r": "日", "sc": 13, "l": 3, "f": 1745},
  {"h": "糖", "p": "táng", "sp": "tang", "r": "米", "sc": 16, "l": 3, "f": 1746},
  {"h": "赋", "p": "fù", "sp": "fu", "r": "貝", "sc": 12, "l": 5, "f": 1747},
  {"h": "抑", "p": "yì", "sp": "yi", "r": "手", "sc": 7, "l": 5, "f": 1748},
  {"h": "秩", "p": "zhì", "sp": "zhi", "r": "禾", "sc": 10, "l": 5, "f": 1749},
  {"h": "帽", "p": "mào", "sp": "mao", "r": "巾", "sc": 12, "l": 3, "f": 1750},
  {"h": "哀", "p": "āi", "sp": "ai", "r": "口", "sc": 9, "l": 5, "f": 1751},
  {"h": "宿", "p": "sù", "sp": "su", "r": "宀", "sc": 11, "l": 5, "f": 1752},
  {"h": "踏", "p": "tà", "sp": "ta", "r": "足", "sc": 15, "l": 5, "f": 1753},
  {"h": "烂", "p": "làn", "sp": "lan", "r": "火", "sc": 9, "l": 5, "f": 1754},
  {"h": "袁", "p": "yuán", "sp": "yuan", "r": "衣", "sc": 10, "l": 5, "f": 1755},
  {"h": "侯", "p": "hóu", "sp": "hou", "r": "人", "sc": 9, "l": 5, "f": 1756},
  {"h": "抖", "p": "dǒu", "sp": "dou", "r": "手", "sc": 7, "l": 5, "f": 1757},
  {"h": "夹", "p": "jiā", "sp": "jia", "r": "一", "sc": 6, "l": 5, "f": 1758},
  {"h": "昆", "p": "kūn", "sp": "kun", "r": "日", "sc": 8, "l": 5, "f": 1759},
  {"h": "肝", "p": "gān", "sp": "gan", "r": "肉", "sc": 7, "l": 5, "f": 1760},
  {"h": "擦", "p": "cā", "sp": "ca", "r": "手", "sc": 17, "l": 4, "f": 1761},
  {"h": "猪", "p": "zhū", "sp": "zhu", "r": "犬", "sc": 11, "l": 3, "f": 1762},
  {"h": "炼", "p": "liàn", "sp": "lian", "r": "火", "sc": 9, "l": 4, "f": 1763},
  {"h": "恒", "p": "héng", "sp": "heng", "r": "心", "sc": 9, "l": 5, "f": 1764},
  {"h": "慎", "p": "shèn", "sp": "shen", "r": "心", "sc": 13, "l": 5, "f": 1765},
  {"h": "搬", "p": "bān", "sp": "ban", "r": "手", "sc": 13, "l": 3, "f": 1766},
  {"h": "纽", "p": "niǔ", "sp": "niu", "r": "糸", "sc": 7, "l": 5, "f": 1767},
  {"h": "纹", "p": "wén", "sp": "wen", "r": "糸", "sc": 7, "l": 5, "f": 1768},
  {"h": "玻", "p": "bō", "sp": "bo", "r": "玉", "sc": 9, "l": 5, "f": 1769},
  {"h": "渔", "p": "yú", "sp": "yu", "r": "水", "sc": 11, "l": 5, "f": 1770},
  {"h": "磁", "p": "cí", "sp": "ci", "r": "石", "sc": 14, "l": 5, "f": 1771},
  {"h": "铜", "p": "tóng", "sp": "tong", "r": "金", "sc": 11, "l": 5, "f": 1772},
  {"h": "齿", "p": "chǐ", "sp": "chi", "r": "齒", "sc": 8, "l": 5, "f": 1773},
  {"h": "跨", "p": "kuà", "sp": "kua", "r": "足", "sc": 13, "l": 5, "f": 1774},
  {"h": "押", "p": "yā", "sp": "ya", "r": "手", "sc": 8, "l": 5, "f": 1775},
  {"h": "怖", "p": "bù", "sp": "bu", "r": "心", "sc": 8, "l": 5, "f": 1776},
  {"h": "漠", "p": "mò", "sp": "mo", "r": "水", "sc": 13, "l": 5, "f": 1777},
  {"h": "疲", "p": "pí", "sp": "pi", "r": "疒", "sc": 10, "l": 5, "f": 1778},
  {"h": "叛", "p": "pàn", "sp": "pan", "r": "又", "sc": 9, "l": 5, "f": 1779},
  {"h": "遣", "p": "qiǎn", "sp": "qian", "r": "辵", "sc": 13, "l": 5, "f": 1780},
  {"h": "兹", "p": "zī", "sp": "zi", "r": "艸", "sc": 9, "l": 5, "f": 1781},
  {"h": "祭", "p": "jì", "sp": "ji", "r": "示", "sc": 11, "l": 5, "f": 1782},
  {"h": "醉", "p": "zuì", "sp": "zui", "r": "酉", "sc": 15, "l": 5, "f": 1783},
  {"h": "拳", "p": "quán", "sp": "quan", "r": "手", "sc": 10, "l": 5, "f": 1784},
  {"h": "弥", "p": "mí", "sp": "mi", "r": "弓", "sc": 8, "l": 5, "f": 1785},
  {"h": "斜", "p": "xié", "sp": "xie", "r": "斗", "sc": 11, "l": 5, "f": 1786},
  {"h": "档", "p": "dàng", "sp": "dang", "r": "木", "sc": 10, "l": 5, "f": 1787},
  {"h": "稀", "p": "xī", "sp": "xi", "r": "禾", "sc": 12, "l": 5, "f": 1788},
  {"h": "捷", "p": "jié", "sp": "jie", "r": "手", "sc": 11, "l": 5, "f": 1789},
  {"h": "肤", "p": "fū", "sp": "fu", "r": "肉", "sc": 8, "l": 5, "f": 1790},
  {"h": "疫", "p": "yì", "sp": "yi", "r": "疒", "sc": 9, "l": 5, "f": 1791},
  {"h": "肿", "p": "zhǒng", "sp": "zhong", "r": "肉", "sc": 8, "l": 5, "f": 1792},
  {"h": "豆", "p": "dòu", "sp": "dou", "r": "豆", "sc": 7, "l": 4, "f": 1793},
  {"h": "削", "p": "xuē", "sp": "xue", "r": "刀", "sc": 9, "l": 5, "f": 1794},
  {"h": "岗", "p": "gǎng", "sp": "gang", "r": "山", "sc": 7, "l": 5, "f": 1795},
  {
    "h": "晃",
    "p": "huǎng",
    "sp": "huang",
    "r": "日",
    "sc": 10,
    "l": 5,
    "f": 1796
  },
  {"h": "吞", "p": "tūn", "sp": "tun", "r": "口", "sc": 7, "l": 5, "f": 1797},
  {"h": "宏", "p": "hóng", "sp": "hong", "r": "宀", "sc": 7, "l": 5, "f": 1798},
  {"h": "癌", "p": "ái", "sp": "ai", "r": "疒", "sc": 17, "l": 5, "f": 1799},
  {"h": "肚", "p": "dù", "sp": "du", "r": "肉", "sc": 7, "l": 4, "f": 1800},
  {"h": "隶", "p": "lì", "sp": "li", "r": "隶", "sc": 8, "l": 5, "f": 1801},
  {"h": "履", "p": "lǚ", "sp": "lu", "r": "尸", "sc": 15, "l": 5, "f": 1802},
  {
    "h": "涨",
    "p": "zhǎng",
    "sp": "zhang",
    "r": "水",
    "sc": 10,
    "l": 4,
    "f": 1803
  },
  {"h": "耀", "p": "yào", "sp": "yao", "r": "羽", "sc": 20, "l": 5, "f": 1804},
  {"h": "扭", "p": "niǔ", "sp": "niu", "r": "手", "sc": 7, "l": 5, "f": 1805},
  {"h": "坛", "p": "tán", "sp": "tan", "r": "土", "sc": 7, "l": 5, "f": 1806},
  {"h": "拨", "p": "bō", "sp": "bo", "r": "手", "sc": 8, "l": 5, "f": 1807},
  {"h": "沃", "p": "wò", "sp": "wo", "r": "水", "sc": 7, "l": 5, "f": 1808},
  {"h": "绘", "p": "huì", "sp": "hui", "r": "糸", "sc": 9, "l": 5, "f": 1809},
  {"h": "伐", "p": "fá", "sp": "fa", "r": "人", "sc": 6, "l": 5, "f": 1810},
  {"h": "堪", "p": "kān", "sp": "kan", "r": "土", "sc": 12, "l": 5, "f": 1811},
  {"h": "仆", "p": "pū", "sp": "pu", "r": "人", "sc": 4, "l": 5, "f": 1812},
  {"h": "郭", "p": "guō", "sp": "guo", "r": "邑", "sc": 10, "l": 5, "f": 1813},
  {"h": "牺", "p": "xī", "sp": "xi", "r": "牛", "sc": 10, "l": 5, "f": 1814},
  {"h": "歼", "p": "jiān", "sp": "jian", "r": "歹", "sc": 7, "l": 5, "f": 1815},
  {"h": "墓", "p": "mù", "sp": "mu", "r": "土", "sc": 13, "l": 5, "f": 1816},
  {"h": "雇", "p": "gù", "sp": "gu", "r": "隹", "sc": 12, "l": 5, "f": 1817},
  {"h": "廉", "p": "lián", "sp": "lian", "r": "广", "sc": 13, "l": 5, "f": 1818},
  {"h": "契", "p": "qì", "sp": "qi", "r": "大", "sc": 9, "l": 5, "f": 1819},
  {"h": "拼", "p": "pīn", "sp": "pin", "r": "手", "sc": 9, "l": 5, "f": 1820},
  {
    "h": "惩",
    "p": "chéng",
    "sp": "cheng",
    "r": "心",
    "sc": 12,
    "l": 5,
    "f": 1821
  },
  {"h": "捉", "p": "zhuō", "sp": "zhuo", "r": "手", "sc": 10, "l": 5, "f": 1822},
  {"h": "覆", "p": "fù", "sp": "fu", "r": "襾", "sc": 18, "l": 5, "f": 1823},
  {"h": "刷", "p": "shuā", "sp": "shua", "r": "刀", "sc": 8, "l": 3, "f": 1824},
  {"h": "劫", "p": "jié", "sp": "jie", "r": "力", "sc": 7, "l": 5, "f": 1825},
  {"h": "嫌", "p": "xián", "sp": "xian", "r": "女", "sc": 13, "l": 5, "f": 1826},
  {"h": "瓜", "p": "guā", "sp": "gua", "r": "瓜", "sc": 5, "l": 2, "f": 1827},
  {"h": "歇", "p": "xiē", "sp": "xie", "r": "欠", "sc": 13, "l": 5, "f": 1828},
  {"h": "雕", "p": "diāo", "sp": "diao", "r": "隹", "sc": 16, "l": 5, "f": 1829},
  {"h": "闷", "p": "mèn", "sp": "men", "r": "門", "sc": 7, "l": 5, "f": 1830},
  {"h": "乳", "p": "rǔ", "sp": "ru", "r": "乙", "sc": 8, "l": 5, "f": 1831},
  {"h": "串", "p": "chuàn", "sp": "chuan", "r": "丨", "sc": 7, "l": 5, "f": 1832},
  {"h": "娃", "p": "wá", "sp": "wa", "r": "女", "sc": 9, "l": 5, "f": 1833},
  {"h": "缴", "p": "jiǎo", "sp": "jiao", "r": "糸", "sc": 16, "l": 5, "f": 1834},
  {"h": "唤", "p": "huàn", "sp": "huan", "r": "口", "sc": 10, "l": 5, "f": 1835},
  {"h": "赢", "p": "yíng", "sp": "ying", "r": "亠", "sc": 17, "l": 3, "f": 1836},
  {"h": "莲", "p": "lián", "sp": "lian", "r": "艸", "sc": 10, "l": 5, "f": 1837},
  {"h": "霸", "p": "bà", "sp": "ba", "r": "雨", "sc": 21, "l": 5, "f": 1838},
  {"h": "桃", "p": "táo", "sp": "tao", "r": "木", "sc": 10, "l": 5, "f": 1839},
  {"h": "妥", "p": "tuǒ", "sp": "tuo", "r": "女", "sc": 7, "l": 5, "f": 1840},
  {"h": "瘦", "p": "shòu", "sp": "shou", "r": "疒", "sc": 14, "l": 3, "f": 1841},
  {"h": "搭", "p": "dā", "sp": "da", "r": "手", "sc": 12, "l": 5, "f": 1842},
  {"h": "赴", "p": "fù", "sp": "fu", "r": "走", "sc": 9, "l": 5, "f": 1843},
  {"h": "岳", "p": "yuè", "sp": "yue", "r": "山", "sc": 8, "l": 5, "f": 1844},
  {"h": "嘉", "p": "jiā", "sp": "jia", "r": "口", "sc": 14, "l": 5, "f": 1845},
  {"h": "舱", "p": "cāng", "sp": "cang", "r": "舟", "sc": 10, "l": 5, "f": 1846},
  {"h": "俊", "p": "jùn", "sp": "jun", "r": "人", "sc": 9, "l": 5, "f": 1847},
  {"h": "址", "p": "zhǐ", "sp": "zhi", "r": "土", "sc": 7, "l": 4, "f": 1848},
  {"h": "庞", "p": "páng", "sp": "pang", "r": "广", "sc": 8, "l": 5, "f": 1849},
  {"h": "耕", "p": "gēng", "sp": "geng", "r": "耒", "sc": 10, "l": 5, "f": 1850},
  {"h": "锐", "p": "ruì", "sp": "rui", "r": "金", "sc": 12, "l": 5, "f": 1851},
  {"h": "缝", "p": "fèng", "sp": "feng", "r": "糸", "sc": 13, "l": 5, "f": 1852},
  {"h": "悔", "p": "huǐ", "sp": "hui", "r": "心", "sc": 10, "l": 4, "f": 1853},
  {"h": "邀", "p": "yāo", "sp": "yao", "r": "辵", "sc": 16, "l": 4, "f": 1854},
  {"h": "玲", "p": "líng", "sp": "ling", "r": "玉", "sc": 9, "l": 5, "f": 1855},
  {"h": "惟", "p": "wéi", "sp": "wei", "r": "心", "sc": 11, "l": 5, "f": 1856},
  {"h": "斥", "p": "chì", "sp": "chi", "r": "斤", "sc": 5, "l": 5, "f": 1857},
  {"h": "宅", "p": "zhái", "sp": "zhai", "r": "宀", "sc": 6, "l": 5, "f": 1858},
  {"h": "添", "p": "tiān", "sp": "tian", "r": "水", "sc": 11, "l": 5, "f": 1859},
  {"h": "挖", "p": "wā", "sp": "wa", "r": "手", "sc": 9, "l": 5, "f": 1860},
  {"h": "呵", "p": "ā", "sp": "a", "r": "口", "sc": 8, "l": 5, "f": 1861},
  {"h": "讼", "p": "sòng", "sp": "song", "r": "言", "sc": 6, "l": 5, "f": 1862},
  {"h": "氧", "p": "yǎng", "sp": "yang", "r": "气", "sc": 10, "l": 5, "f": 1863},
  {"h": "浩", "p": "hào", "sp": "hao", "r": "水", "sc": 10, "l": 5, "f": 1864},
  {"h": "羽", "p": "yǔ", "sp": "yu", "r": "羽", "sc": 6, "l": 5, "f": 1865},
  {"h": "斤", "p": "jīn", "sp": "jin", "r": "斤", "sc": 4, "l": 2, "f": 1866},
  {"h": "酷", "p": "kù", "sp": "ku", "r": "酉", "sc": 14, "l": 5, "f": 1867},
  {"h": "掠", "p": "lüè", "sp": "lue", "r": "手", "sc": 11, "l": 5, "f": 1868},
  {"h": "妖", "p": "yāo", "sp": "yao", "r": "女", "sc": 7, "l": 5, "f": 1869},
  {"h": "祸", "p": "huò", "sp": "huo", "r": "示", "sc": 11, "l": 5, "f": 1870},
  {"h": "侍", "p": "shì", "sp": "shi", "r": "人", "sc": 8, "l": 5, "f": 1871},
  {"h": "乙", "p": "yǐ", "sp": "yi", "r": "乙", "sc": 1, "l": 5, "f": 1872},
  {"h": "妨", "p": "fáng", "sp": "fang", "r": "女", "sc": 7, "l": 5, "f": 1873},
  {"h": "贪", "p": "tān", "sp": "tan", "r": "貝", "sc": 8, "l": 5, "f": 1874},
  {"h": "挣", "p": "zhēng", "sp": "zheng", "r": "手", "sc": 9, "l": 5, "f": 1875},
  {"h": "汪", "p": "wāng", "sp": "wang", "r": "水", "sc": 7, "l": 5, "f": 1876},
  {"h": "尿", "p": "niào", "sp": "niao", "r": "尸", "sc": 7, "l": 5, "f": 1877},
  {"h": "莉", "p": "lì", "sp": "li", "r": "艸", "sc": 10, "l": 5, "f": 1878},
  {"h": "悬", "p": "xuán", "sp": "xuan", "r": "心", "sc": 11, "l": 5, "f": 1879},
  {"h": "唇", "p": "chún", "sp": "chun", "r": "口", "sc": 10, "l": 5, "f": 1880},
  {"h": "翰", "p": "hàn", "sp": "han", "r": "羽", "sc": 16, "l": 5, "f": 1881},
  {"h": "仓", "p": "cāng", "sp": "cang", "r": "人", "sc": 4, "l": 5, "f": 1882},
  {"h": "轨", "p": "guǐ", "sp": "gui", "r": "車", "sc": 6, "l": 5, "f": 1883},
  {"h": "枚", "p": "méi", "sp": "mei", "r": "木", "sc": 8, "l": 5, "f": 1884},
  {"h": "盐", "p": "yán", "sp": "yan", "r": "皿", "sc": 10, "l": 4, "f": 1885},
  {"h": "览", "p": "lǎn", "sp": "lan", "r": "見", "sc": 9, "l": 5, "f": 1886},
  {"h": "傅", "p": "fù", "sp": "fu", "r": "人", "sc": 12, "l": 5, "f": 1887},
  {"h": "帅", "p": "shuài", "sp": "shuai", "r": "巾", "sc": 5, "l": 4, "f": 1888},
  {"h": "庙", "p": "miào", "sp": "miao", "r": "广", "sc": 8, "l": 5, "f": 1889},
  {"h": "芬", "p": "fēn", "sp": "fen", "r": "艸", "sc": 7, "l": 5, "f": 1890},
  {"h": "屏", "p": "píng", "sp": "ping", "r": "尸", "sc": 9, "l": 5, "f": 1891},
  {"h": "寺", "p": "sì", "sp": "si", "r": "寸", "sc": 6, "l": 5, "f": 1892},
  {"h": "胖", "p": "pàng", "sp": "pang", "r": "肉", "sc": 9, "l": 3, "f": 1893},
  {"h": "璃", "p": "lí", "sp": "li", "r": "玉", "sc": 14, "l": 5, "f": 1894},
  {"h": "愚", "p": "yú", "sp": "yu", "r": "心", "sc": 13, "l": 5, "f": 1895},
  {"h": "滴", "p": "dī", "sp": "di", "r": "水", "sc": 14, "l": 5, "f": 1896},
  {"h": "疏", "p": "shū", "sp": "shu", "r": "疋", "sc": 12, "l": 5, "f": 1897},
  {"h": "萧", "p": "xiāo", "sp": "xiao", "r": "艸", "sc": 11, "l": 5, "f": 1898},
  {"h": "姿", "p": "zī", "sp": "zi", "r": "女", "sc": 9, "l": 5, "f": 1899},
  {"h": "颤", "p": "chàn", "sp": "chan", "r": "頁", "sc": 19, "l": 5, "f": 1900},
  {"h": "丑", "p": "chǒu", "sp": "chou", "r": "一", "sc": 4, "l": 3, "f": 1901},
  {"h": "劣", "p": "liè", "sp": "lie", "r": "力", "sc": 6, "l": 5, "f": 1902},
  {"h": "柯", "p": "kē", "sp": "ke", "r": "木", "sc": 9, "l": 5, "f": 1903},
  {"h": "寸", "p": "cùn", "sp": "cun", "r": "寸", "sc": 3, "l": 4, "f": 1904},
  {"h": "扔", "p": "rēng", "sp": "reng", "r": "手", "sc": 5, "l": 4, "f": 1905},
  {"h": "盯", "p": "dīng", "sp": "ding", "r": "目", "sc": 7, "l": 5, "f": 1906},
  {"h": "辱", "p": "rǔ", "sp": "ru", "r": "辰", "sc": 10, "l": 5, "f": 1907},
  {"h": "匹", "p": "pǐ", "sp": "pi", "r": "匸", "sc": 4, "l": 5, "f": 1908},
  {"h": "俱", "p": "jù", "sp": "ju", "r": "人", "sc": 10, "l": 5, "f": 1909},
  {"h": "辨", "p": "biàn", "sp": "bian", "r": "辛", "sc": 16, "l": 5, "f": 1910},
  {"h": "饿", "p": "è", "sp": "e", "r": "食", "sc": 10, "l": 1, "f": 1911},
  {"h": "蜂", "p": "fēng", "sp": "feng", "r": "虫", "sc": 13, "l": 5, "f": 1912},
  {"h": "哦", "p": "ó", "sp": "o", "r": "口", "sc": 10, "l": 5, "f": 1913},
  {
    "h": "腔",
    "p": "qiāng",
    "sp": "qiang",
    "r": "肉",
    "sc": 12,
    "l": 5,
    "f": 1914
  },
  {"h": "郁", "p": "yù", "sp": "yu", "r": "邑", "sc": 8, "l": 5, "f": 1915},
  {"h": "溃", "p": "kuì", "sp": "kui", "r": "水", "sc": 12, "l": 5, "f": 1916},
  {"h": "谨", "p": "jǐn", "sp": "jin", "r": "言", "sc": 13, "l": 5, "f": 1917},
  {"h": "糟", "p": "zāo", "sp": "zao", "r": "米", "sc": 17, "l": 5, "f": 1918},
  {"h": "葛", "p": "gé", "sp": "ge", "r": "艸", "sc": 12, "l": 5, "f": 1919},
  {"h": "苗", "p": "miáo", "sp": "miao", "r": "艸", "sc": 8, "l": 5, "f": 1920},
  {"h": "肠", "p": "cháng", "sp": "chang", "r": "肉", "sc": 7, "l": 5, "f": 1921},
  {"h": "忌", "p": "jì", "sp": "ji", "r": "心", "sc": 7, "l": 5, "f": 1922},
  {"h": "溜", "p": "liū", "sp": "liu", "r": "水", "sc": 13, "l": 5, "f": 1923},
  {"h": "鸿", "p": "hóng", "sp": "hong", "r": "鳥", "sc": 11, "l": 5, "f": 1924},
  {"h": "爵", "p": "jué", "sp": "jue", "r": "爪", "sc": 17, "l": 5, "f": 1925},
  {"h": "鹏", "p": "péng", "sp": "peng", "r": "鳥", "sc": 13, "l": 5, "f": 1926},
  {"h": "鹰", "p": "yīng", "sp": "ying", "r": "鳥", "sc": 18, "l": 5, "f": 1927},
  {"h": "笼", "p": "lóng", "sp": "long", "r": "竹", "sc": 11, "l": 5, "f": 1928},
  {"h": "丘", "p": "qiū", "sp": "qiu", "r": "一", "sc": 5, "l": 5, "f": 1929},
  {"h": "桂", "p": "guì", "sp": "gui", "r": "木", "sc": 10, "l": 5, "f": 1930},
  {"h": "滋", "p": "zī", "sp": "zi", "r": "水", "sc": 12, "l": 5, "f": 1931},
  {"h": "聊", "p": "liáo", "sp": "liao", "r": "耳", "sc": 11, "l": 3, "f": 1932},
  {"h": "挡", "p": "dǎng", "sp": "dang", "r": "手", "sc": 9, "l": 5, "f": 1933},
  {"h": "纲", "p": "gāng", "sp": "gang", "r": "糸", "sc": 7, "l": 5, "f": 1934},
  {"h": "肌", "p": "jī", "sp": "ji", "r": "肉", "sc": 6, "l": 5, "f": 1935},
  {"h": "茨", "p": "cí", "sp": "ci", "r": "艸", "sc": 9, "l": 5, "f": 1936},
  {"h": "壳", "p": "ké", "sp": "ke", "r": "士", "sc": 7, "l": 5, "f": 1937},
  {"h": "痕", "p": "hén", "sp": "hen", "r": "疒", "sc": 11, "l": 5, "f": 1938},
  {"h": "碗", "p": "wǎn", "sp": "wan", "r": "石", "sc": 13, "l": 2, "f": 1939},
  {"h": "穴", "p": "xué", "sp": "xue", "r": "穴", "sc": 5, "l": 5, "f": 1940},
  {"h": "膀", "p": "bǎng", "sp": "bang", "r": "肉", "sc": 14, "l": 5, "f": 1941},
  {"h": "卓", "p": "zhuō", "sp": "zhuo", "r": "十", "sc": 8, "l": 5, "f": 1942},
  {"h": "贤", "p": "xián", "sp": "xian", "r": "貝", "sc": 8, "l": 5, "f": 1943},
  {"h": "卧", "p": "wò", "sp": "wo", "r": "臣", "sc": 8, "l": 5, "f": 1944},
  {"h": "膜", "p": "mó", "sp": "mo", "r": "肉", "sc": 14, "l": 5, "f": 1945},
  {"h": "毅", "p": "yì", "sp": "yi", "r": "殳", "sc": 15, "l": 5, "f": 1946},
  {"h": "锦", "p": "jǐn", "sp": "jin", "r": "金", "sc": 13, "l": 5, "f": 1947},
  {"h": "欠", "p": "qiàn", "sp": "qian", "r": "欠", "sc": 4, "l": 5, "f": 1948},
  {"h": "哩", "p": "lī", "sp": "li", "r": "口", "sc": 10, "l": 5, "f": 1949},
  {"h": "函", "p": "hán", "sp": "han", "r": "凵", "sc": 8, "l": 5, "f": 1950},
  {"h": "茫", "p": "máng", "sp": "mang", "r": "艸", "sc": 9, "l": 5, "f": 1951},
  {"h": "昂", "p": "áng", "sp": "ang", "r": "日", "sc": 8, "l": 5, "f": 1952},
  {"h": "薛", "p": "xuē", "sp": "xue", "r": "艸", "sc": 16, "l": 5, "f": 1953},
  {"h": "皱", "p": "zhòu", "sp": "zhou", "r": "皮", "sc": 10, "l": 5, "f": 1954},
  {"h": "夸", "p": "kuā", "sp": "kua", "r": "大", "sc": 6, "l": 5, "f": 1955},
  {"h": "豫", "p": "yù", "sp": "yu", "r": "豕", "sc": 15, "l": 5, "f": 1956},
  {"h": "胃", "p": "wèi", "sp": "wei", "r": "肉", "sc": 9, "l": 5, "f": 1957},
  {"h": "舌", "p": "shé", "sp": "she", "r": "舌", "sc": 6, "l": 5, "f": 1958},
  {"h": "剥", "p": "bō", "sp": "bo", "r": "刀", "sc": 10, "l": 5, "f": 1959},
  {"h": "傲", "p": "ào", "sp": "ao", "r": "人", "sc": 12, "l": 5, "f": 1960},
  {"h": "拾", "p": "shí", "sp": "shi", "r": "手", "sc": 9, "l": 5, "f": 1961},
  {"h": "窝", "p": "wō", "sp": "wo", "r": "穴", "sc": 12, "l": 5, "f": 1962},
  {
    "h": "睁",
    "p": "zhēng",
    "sp": "zheng",
    "r": "目",
    "sc": 11,
    "l": 5,
    "f": 1963
  },
  {"h": "携", "p": "xié", "sp": "xie", "r": "手", "sc": 13, "l": 5, "f": 1964},
  {"h": "陵", "p": "líng", "sp": "ling", "r": "阜", "sc": 10, "l": 5, "f": 1965},
  {"h": "哼", "p": "hēng", "sp": "heng", "r": "口", "sc": 10, "l": 5, "f": 1966},
  {"h": "棉", "p": "mián", "sp": "mian", "r": "木", "sc": 12, "l": 5, "f": 1967},
  {"h": "晴", "p": "qíng", "sp": "qing", "r": "日", "sc": 12, "l": 2, "f": 1968},
  {"h": "铃", "p": "líng", "sp": "ling", "r": "金", "sc": 10, "l": 5, "f": 1969},
  {"h": "填", "p": "tián", "sp": "tian", "r": "土", "sc": 13, "l": 4, "f": 1970},
  {"h": "饲", "p": "sì", "sp": "si", "r": "食", "sc": 8, "l": 5, "f": 1971},
  {"h": "渴", "p": "kě", "sp": "ke", "r": "水", "sc": 12, "l": 1, "f": 1972},
  {"h": "吻", "p": "wěn", "sp": "wen", "r": "口", "sc": 7, "l": 5, "f": 1973},
  {"h": "扮", "p": "bàn", "sp": "ban", "r": "手", "sc": 7, "l": 4, "f": 1974},
  {"h": "逆", "p": "nì", "sp": "ni", "r": "辵", "sc": 9, "l": 5, "f": 1975},
  {"h": "脆", "p": "cuì", "sp": "cui", "r": "肉", "sc": 10, "l": 5, "f": 1976},
  {
    "h": "喘",
    "p": "chuǎn",
    "sp": "chuan",
    "r": "口",
    "sc": 12,
    "l": 5,
    "f": 1977
  },
  {"h": "罩", "p": "zhào", "sp": "zhao", "r": "网", "sc": 13, "l": 5, "f": 1978},
  {"h": "卜", "p": "bo", "sp": "bo", "r": "卜", "sc": 2, "l": 5, "f": 1979},
  {"h": "炉", "p": "lú", "sp": "lu", "r": "火", "sc": 8, "l": 5, "f": 1980},
  {"h": "柴", "p": "chái", "sp": "chai", "r": "木", "sc": 10, "l": 5, "f": 1981},
  {"h": "愉", "p": "yú", "sp": "yu", "r": "心", "sc": 12, "l": 4, "f": 1982},
  {
    "h": "绳",
    "p": "shéng",
    "sp": "sheng",
    "r": "糸",
    "sc": 11,
    "l": 5,
    "f": 1983
  },
  {"h": "胎", "p": "tāi", "sp": "tai", "r": "肉", "sc": 9, "l": 5, "f": 1984},
  {"h": "蓄", "p": "xù", "sp": "xu", "r": "艸", "sc": 13, "l": 5, "f": 1985},
  {"h": "眠", "p": "mián", "sp": "mian", "r": "目", "sc": 10, "l": 5, "f": 1986},
  {"h": "竭", "p": "jié", "sp": "jie", "r": "立", "sc": 14, "l": 5, "f": 1987},
  {"h": "喂", "p": "wèi", "sp": "wei", "r": "口", "sc": 12, "l": 1, "f": 1988},
  {"h": "傻", "p": "shǎ", "sp": "sha", "r": "人", "sc": 13, "l": 5, "f": 1989},
  {"h": "慕", "p": "mù", "sp": "mu", "r": "心", "sc": 14, "l": 5, "f": 1990},
  {"h": "浑", "p": "hún", "sp": "hun", "r": "水", "sc": 9, "l": 5, "f": 1991},
  {"h": "奸", "p": "jiān", "sp": "jian", "r": "女", "sc": 6, "l": 5, "f": 1992},
  {"h": "扇", "p": "shàn", "sp": "shan", "r": "戶", "sc": 10, "l": 5, "f": 1993},
  {"h": "柜", "p": "guì", "sp": "gui", "r": "木", "sc": 8, "l": 5, "f": 1994},
  {"h": "悦", "p": "yuè", "sp": "yue", "r": "心", "sc": 10, "l": 5, "f": 1995},
  {"h": "拦", "p": "lán", "sp": "lan", "r": "手", "sc": 8, "l": 5, "f": 1996},
  {"h": "诞", "p": "dàn", "sp": "dan", "r": "言", "sc": 8, "l": 5, "f": 1997},
  {"h": "饱", "p": "bǎo", "sp": "bao", "r": "食", "sc": 8, "l": 2, "f": 1998},
  {"h": "乾", "p": "gān", "sp": "gan", "r": "乙", "sc": 11, "l": 5, "f": 1999},
  {"h": "泡", "p": "pào", "sp": "pao", "r": "水", "sc": 8, "l": 5, "f": 2000},
  {"h": "贼", "p": "zéi", "sp": "zei", "r": "貝", "sc": 10, "l": 5, "f": 2001},
  {"h": "亭", "p": "tíng", "sp": "ting", "r": "亠", "sc": 9, "l": 5, "f": 2002},
  {"h": "夕", "p": "xī", "sp": "xi", "r": "夕", "sc": 3, "l": 5, "f": 2003},
  {"h": "爹", "p": "diē", "sp": "die", "r": "父", "sc": 10, "l": 5, "f": 2004},
  {"h": "酬", "p": "chóu", "sp": "chou", "r": "酉", "sc": 13, "l": 5, "f": 2005},
  {"h": "儒", "p": "rú", "sp": "ru", "r": "人", "sc": 16, "l": 5, "f": 2006},
  {"h": "姻", "p": "yīn", "sp": "yin", "r": "女", "sc": 9, "l": 5, "f": 2007},
  {"h": "卵", "p": "luǎn", "sp": "luan", "r": "卩", "sc": 7, "l": 5, "f": 2008},
  {"h": "氛", "p": "fēn", "sp": "fen", "r": "气", "sc": 8, "l": 5, "f": 2009},
  {"h": "泄", "p": "xiè", "sp": "xie", "r": "水", "sc": 8, "l": 5, "f": 2010},
  {"h": "杆", "p": "gān", "sp": "gan", "r": "木", "sc": 7, "l": 5, "f": 2011},
  {"h": "挨", "p": "āi", "sp": "ai", "r": "手", "sc": 10, "l": 5, "f": 2012},
  {"h": "僧", "p": "sēng", "sp": "seng", "r": "人", "sc": 14, "l": 5, "f": 2013},
  {"h": "蜜", "p": "mì", "sp": "mi", "r": "虫", "sc": 14, "l": 5, "f": 2014},
  {"h": "吟", "p": "yín", "sp": "yin", "r": "口", "sc": 7, "l": 5, "f": 2015},
  {"h": "猩", "p": "xīng", "sp": "xing", "r": "犬", "sc": 12, "l": 5, "f": 2016},
  {"h": "遂", "p": "suì", "sp": "sui", "r": "辵", "sc": 12, "l": 5, "f": 2017},
  {"h": "狭", "p": "xiá", "sp": "xia", "r": "犬", "sc": 9, "l": 5, "f": 2018},
  {"h": "肖", "p": "xiào", "sp": "xiao", "r": "肉", "sc": 7, "l": 5, "f": 2019},
  {"h": "甜", "p": "tián", "sp": "tian", "r": "甘", "sc": 11, "l": 3, "f": 2020},
  {"h": "霞", "p": "xiá", "sp": "xia", "r": "雨", "sc": 17, "l": 5, "f": 2021},
  {"h": "驳", "p": "bó", "sp": "bo", "r": "馬", "sc": 7, "l": 5, "f": 2022},
  {"h": "裕", "p": "yù", "sp": "yu", "r": "衣", "sc": 12, "l": 5, "f": 2023},
  {"h": "顽", "p": "wán", "sp": "wan", "r": "頁", "sc": 10, "l": 5, "f": 2024},
  {"h": "於", "p": "yú", "sp": "yu", "r": "方", "sc": 8, "l": 5, "f": 2025},
  {"h": "摘", "p": "zhāi", "sp": "zhai", "r": "手", "sc": 14, "l": 5, "f": 2026},
  {"h": "矮", "p": "ǎi", "sp": "ai", "r": "矢", "sc": 13, "l": 2, "f": 2027},
  {"h": "秒", "p": "miǎo", "sp": "miao", "r": "禾", "sc": 9, "l": 5, "f": 2028},
  {"h": "卿", "p": "qīng", "sp": "qing", "r": "卩", "sc": 10, "l": 5, "f": 2029},
  {"h": "畜", "p": "chù", "sp": "chu", "r": "田", "sc": 10, "l": 5, "f": 2030},
  {"h": "咽", "p": "yàn", "sp": "yan", "r": "口", "sc": 9, "l": 5, "f": 2031},
  {"h": "披", "p": "pī", "sp": "pi", "r": "手", "sc": 8, "l": 5, "f": 2032},
  {"h": "辅", "p": "fǔ", "sp": "fu", "r": "車", "sc": 11, "l": 5, "f": 2033},
  {"h": "勾", "p": "gōu", "sp": "gou", "r": "勹", "sc": 4, "l": 5, "f": 2034},
  {"h": "盆", "p": "pén", "sp": "pen", "r": "皿", "sc": 9, "l": 5, "f": 2035},
  {
    "h": "疆",
    "p": "jiāng",
    "sp": "jiang",
    "r": "田",
    "sc": 19,
    "l": 5,
    "f": 2036
  },
  {"h": "赌", "p": "dǔ", "sp": "du", "r": "貝", "sc": 12, "l": 5, "f": 2037},
  {"h": "塑", "p": "sù", "sp": "su", "r": "土", "sc": 13, "l": 4, "f": 2038},
  {"h": "畏", "p": "wèi", "sp": "wei", "r": "田", "sc": 9, "l": 5, "f": 2039},
  {"h": "吵", "p": "chǎo", "sp": "chao", "r": "口", "sc": 7, "l": 3, "f": 2040},
  {"h": "囊", "p": "náng", "sp": "nang", "r": "口", "sc": 22, "l": 5, "f": 2041},
  {"h": "嗯", "p": "ń", "sp": "ń", "r": "口", "sc": 13, "l": 5, "f": 2042},
  {"h": "泊", "p": "pō", "sp": "po", "r": "水", "sc": 8, "l": 5, "f": 2043},
  {"h": "肺", "p": "fèi", "sp": "fei", "r": "肉", "sc": 8, "l": 5, "f": 2044},
  {"h": "骤", "p": "zhòu", "sp": "zhou", "r": "馬", "sc": 17, "l": 5, "f": 2045},
  {"h": "缠", "p": "chán", "sp": "chan", "r": "糸", "sc": 13, "l": 5, "f": 2046},
  {"h": "冈", "p": "gāng", "sp": "gang", "r": "冂", "sc": 4, "l": 5, "f": 2047},
  {"h": "羞", "p": "xiū", "sp": "xiu", "r": "羊", "sc": 10, "l": 5, "f": 2048},
  {"h": "瞪", "p": "dèng", "sp": "deng", "r": "目", "sc": 17, "l": 5, "f": 2049},
  {"h": "吊", "p": "diào", "sp": "diao", "r": "口", "sc": 6, "l": 5, "f": 2050},
  {"h": "贾", "p": "jiǎ", "sp": "jia", "r": "貝", "sc": 10, "l": 5, "f": 2051},
  {"h": "漏", "p": "lòu", "sp": "lou", "r": "水", "sc": 14, "l": 5, "f": 2052},
  {"h": "斑", "p": "bān", "sp": "ban", "r": "文", "sc": 12, "l": 5, "f": 2053},
  {"h": "涛", "p": "tāo", "sp": "tao", "r": "水", "sc": 10, "l": 5, "f": 2054},
  {"h": "悠", "p": "yōu", "sp": "you", "r": "心", "sc": 11, "l": 5, "f": 2055},
  {"h": "鹿", "p": "lù", "sp": "lu", "r": "鹿", "sc": 11, "l": 5, "f": 2056},
  {"h": "俘", "p": "fú", "sp": "fu", "r": "人", "sc": 9, "l": 5, "f": 2057},
  {"h": "锡", "p": "xī", "sp": "xi", "r": "金", "sc": 13, "l": 5, "f": 2058},
  {"h": "卑", "p": "bēi", "sp": "bei", "r": "十", "sc": 8, "l": 5, "f": 2059},
  {"h": "葬", "p": "zàng", "sp": "zang", "r": "艸", "sc": 12, "l": 5, "f": 2060},
  {"h": "铭", "p": "míng", "sp": "ming", "r": "金", "sc": 11, "l": 5, "f": 2061},
  {"h": "滩", "p": "tān", "sp": "tan", "r": "水", "sc": 13, "l": 5, "f": 2062},
  {"h": "嫁", "p": "jià", "sp": "jia", "r": "女", "sc": 13, "l": 5, "f": 2063},
  {"h": "催", "p": "cuī", "sp": "cui", "r": "人", "sc": 13, "l": 5, "f": 2064},
  {"h": "璇", "p": "xuán", "sp": "xuan", "r": "玉", "sc": 15, "l": 5, "f": 2065},
  {"h": "翅", "p": "chì", "sp": "chi", "r": "羽", "sc": 10, "l": 5, "f": 2066},
  {"h": "盒", "p": "hé", "sp": "he", "r": "皿", "sc": 11, "l": 5, "f": 2067},
  {"h": "蛮", "p": "mán", "sp": "man", "r": "虫", "sc": 12, "l": 5, "f": 2068},
  {"h": "矣", "p": "yǐ", "sp": "yi", "r": "矢", "sc": 7, "l": 5, "f": 2069},
  {"h": "潘", "p": "pān", "sp": "pan", "r": "水", "sc": 15, "l": 5, "f": 2070},
  {"h": "歧", "p": "qí", "sp": "qi", "r": "止", "sc": 8, "l": 5, "f": 2071},
  {"h": "赐", "p": "cì", "sp": "ci", "r": "貝", "sc": 12, "l": 5, "f": 2072},
  {"h": "鲍", "p": "bào", "sp": "bao", "r": "魚", "sc": 13, "l": 5, "f": 2073},
  {"h": "锅", "p": "guō", "sp": "guo", "r": "金", "sc": 12, "l": 5, "f": 2074},
  {"h": "廊", "p": "láng", "sp": "lang", "r": "广", "sc": 11, "l": 5, "f": 2075},
  {"h": "拆", "p": "chāi", "sp": "chai", "r": "手", "sc": 8, "l": 5, "f": 2076},
  {"h": "灌", "p": "guàn", "sp": "guan", "r": "水", "sc": 20, "l": 5, "f": 2077},
  {"h": "勉", "p": "miǎn", "sp": "mian", "r": "力", "sc": 9, "l": 5, "f": 2078},
  {"h": "盲", "p": "máng", "sp": "mang", "r": "目", "sc": 8, "l": 5, "f": 2079},
  {"h": "宰", "p": "zǎi", "sp": "zai", "r": "宀", "sc": 10, "l": 5, "f": 2080},
  {"h": "佐", "p": "zuǒ", "sp": "zuo", "r": "人", "sc": 7, "l": 5, "f": 2081},
  {"h": "啥", "p": "shà", "sp": "sha", "r": "口", "sc": 11, "l": 5, "f": 2082},
  {"h": "胀", "p": "zhàng", "sp": "zhang", "r": "肉", "sc": 8, "l": 5, "f": 2083},
  {"h": "扯", "p": "chě", "sp": "che", "r": "手", "sc": 7, "l": 5, "f": 2084},
  {"h": "禧", "p": "xǐ", "sp": "xi", "r": "示", "sc": 16, "l": 5, "f": 2085},
  {"h": "辽", "p": "liáo", "sp": "liao", "r": "辵", "sc": 5, "l": 5, "f": 2086},
  {"h": "抹", "p": "mǒ", "sp": "mo", "r": "手", "sc": 8, "l": 5, "f": 2087},
  {"h": "筒", "p": "tǒng", "sp": "tong", "r": "竹", "sc": 12, "l": 5, "f": 2088},
  {"h": "棋", "p": "qí", "sp": "qi", "r": "木", "sc": 12, "l": 5, "f": 2089},
  {"h": "裤", "p": "kù", "sp": "ku", "r": "衣", "sc": 12, "l": 3, "f": 2090},
  {"h": "唉", "p": "āi", "sp": "ai", "r": "口", "sc": 10, "l": 5, "f": 2091},
  {"h": "朴", "p": "pǔ", "sp": "pu", "r": "木", "sc": 6, "l": 5, "f": 2092},
  {"h": "咐", "p": "fù", "sp": "fu", "r": "口", "sc": 8, "l": 5, "f": 2093},
  {"h": "孕", "p": "yùn", "sp": "yun", "r": "子", "sc": 5, "l": 5, "f": 2094},
  {"h": "誓", "p": "shì", "sp": "shi", "r": "言", "sc": 14, "l": 5, "f": 2095},
  {"h": "喉", "p": "hóu", "sp": "hou", "r": "口", "sc": 12, "l": 5, "f": 2096},
  {"h": "妄", "p": "wàng", "sp": "wang", "r": "女", "sc": 6, "l": 5, "f": 2097},
  {"h": "拘", "p": "jū", "sp": "ju", "r": "手", "sc": 8, "l": 5, "f": 2098},
  {"h": "链", "p": "liàn", "sp": "lian", "r": "金", "sc": 12, "l": 5, "f": 2099},
  {"h": "驰", "p": "chí", "sp": "chi", "r": "馬", "sc": 6, "l": 5, "f": 2100},
  {"h": "栏", "p": "lán", "sp": "lan", "r": "木", "sc": 9, "l": 5, "f": 2101},
  {"h": "逝", "p": "shì", "sp": "shi", "r": "辵", "sc": 10, "l": 5, "f": 2102},
  {"h": "窃", "p": "qiè", "sp": "qie", "r": "穴", "sc": 9, "l": 5, "f": 2103},
  {"h": "艳", "p": "yàn", "sp": "yan", "r": "色", "sc": 10, "l": 5, "f": 2104},
  {"h": "臭", "p": "chòu", "sp": "chou", "r": "自", "sc": 10, "l": 5, "f": 2105},
  {"h": "纤", "p": "xiān", "sp": "xian", "r": "糸", "sc": 6, "l": 5, "f": 2106},
  {"h": "玑", "p": "jī", "sp": "ji", "r": "玉", "sc": 6, "l": 5, "f": 2107},
  {"h": "棵", "p": "kē", "sp": "ke", "r": "木", "sc": 12, "l": 4, "f": 2108},
  {"h": "趁", "p": "chèn", "sp": "chen", "r": "走", "sc": 12, "l": 5, "f": 2109},
  {"h": "匠", "p": "jiàng", "sp": "jiang", "r": "匚", "sc": 6, "l": 5, "f": 2110},
  {"h": "盈", "p": "yíng", "sp": "ying", "r": "皿", "sc": 9, "l": 5, "f": 2111},
  {"h": "翁", "p": "wēng", "sp": "weng", "r": "羽", "sc": 10, "l": 5, "f": 2112},
  {"h": "愁", "p": "chóu", "sp": "chou", "r": "心", "sc": 13, "l": 5, "f": 2113},
  {"h": "瞬", "p": "shùn", "sp": "shun", "r": "目", "sc": 17, "l": 5, "f": 2114},
  {"h": "婴", "p": "yīng", "sp": "ying", "r": "女", "sc": 11, "l": 5, "f": 2115},
  {"h": "孝", "p": "xiào", "sp": "xiao", "r": "子", "sc": 7, "l": 5, "f": 2116},
  {"h": "颈", "p": "jǐng", "sp": "jing", "r": "頁", "sc": 11, "l": 5, "f": 2117},
  {"h": "倘", "p": "tǎng", "sp": "tang", "r": "人", "sc": 10, "l": 5, "f": 2118},
  {"h": "浙", "p": "zhè", "sp": "zhe", "r": "水", "sc": 10, "l": 5, "f": 2119},
  {
    "h": "谅",
    "p": "liàng",
    "sp": "liang",
    "r": "言",
    "sc": 10,
    "l": 4,
    "f": 2120
  },
  {"h": "蔽", "p": "bì", "sp": "bi", "r": "艸", "sc": 14, "l": 5, "f": 2121},
  {"h": "畅", "p": "chàng", "sp": "chang", "r": "日", "sc": 8, "l": 5, "f": 2122},
  {"h": "赠", "p": "zèng", "sp": "zeng", "r": "貝", "sc": 16, "l": 5, "f": 2123},
  {"h": "妮", "p": "nī", "sp": "ni", "r": "女", "sc": 8, "l": 5, "f": 2124},
  {"h": "莎", "p": "shā", "sp": "sha", "r": "艸", "sc": 10, "l": 5, "f": 2125},
  {"h": "尉", "p": "wèi", "sp": "wei", "r": "寸", "sc": 11, "l": 5, "f": 2126},
  {"h": "冻", "p": "dòng", "sp": "dong", "r": "冫", "sc": 7, "l": 5, "f": 2127},
  {"h": "跪", "p": "guì", "sp": "gui", "r": "足", "sc": 13, "l": 5, "f": 2128},
  {
    "h": "闯",
    "p": "chuǎng",
    "sp": "chuang",
    "r": "門",
    "sc": 6,
    "l": 5,
    "f": 2129
  },
  {"h": "葡", "p": "pú", "sp": "pu", "r": "艸", "sc": 12, "l": 5, "f": 2130},
  {"h": "後", "p": "hòu", "sp": "hou", "r": "彳", "sc": 9, "l": 5, "f": 2131},
  {"h": "厨", "p": "chú", "sp": "chu", "r": "厂", "sc": 12, "l": 3, "f": 2132},
  {"h": "鸭", "p": "yā", "sp": "ya", "r": "鳥", "sc": 10, "l": 4, "f": 2133},
  {"h": "颠", "p": "diān", "sp": "dian", "r": "頁", "sc": 16, "l": 5, "f": 2134},
  {"h": "遮", "p": "zhē", "sp": "zhe", "r": "辵", "sc": 14, "l": 5, "f": 2135},
  {"h": "谊", "p": "yì", "sp": "yi", "r": "言", "sc": 10, "l": 5, "f": 2136},
  {"h": "圳", "p": "zhèn", "sp": "zhen", "r": "土", "sc": 6, "l": 5, "f": 2137},
  {"h": "吁", "p": "xū", "sp": "xu", "r": "口", "sc": 6, "l": 5, "f": 2138},
  {"h": "仑", "p": "lún", "sp": "lun", "r": "人", "sc": 4, "l": 5, "f": 2139},
  {"h": "辟", "p": "pì", "sp": "pi", "r": "辛", "sc": 13, "l": 5, "f": 2140},
  {"h": "瘤", "p": "liú", "sp": "liu", "r": "疒", "sc": 15, "l": 5, "f": 2141},
  {"h": "嫂", "p": "sǎo", "sp": "sao", "r": "女", "sc": 12, "l": 5, "f": 2142},
  {"h": "陀", "p": "tuó", "sp": "tuo", "r": "阜", "sc": 7, "l": 5, "f": 2143},
  {
    "h": "框",
    "p": "kuāng",
    "sp": "kuang",
    "r": "木",
    "sc": 10,
    "l": 5,
    "f": 2144
  },
  {"h": "谭", "p": "tán", "sp": "tan", "r": "言", "sc": 14, "l": 5, "f": 2145},
  {"h": "亨", "p": "hēng", "sp": "heng", "r": "亠", "sc": 7, "l": 5, "f": 2146},
  {"h": "钦", "p": "qīn", "sp": "qin", "r": "金", "sc": 9, "l": 5, "f": 2147},
  {"h": "庸", "p": "yōng", "sp": "yong", "r": "广", "sc": 11, "l": 5, "f": 2148},
  {"h": "歉", "p": "qiàn", "sp": "qian", "r": "欠", "sc": 14, "l": 4, "f": 2149},
  {"h": "芝", "p": "zhī", "sp": "zhi", "r": "艸", "sc": 6, "l": 5, "f": 2150},
  {"h": "吼", "p": "hǒu", "sp": "hou", "r": "口", "sc": 7, "l": 5, "f": 2151},
  {"h": "甫", "p": "fǔ", "sp": "fu", "r": "用", "sc": 7, "l": 5, "f": 2152},
  {"h": "衫", "p": "shān", "sp": "shan", "r": "衣", "sc": 8, "l": 3, "f": 2153},
  {"h": "摊", "p": "tān", "sp": "tan", "r": "手", "sc": 13, "l": 5, "f": 2154},
  {"h": "宴", "p": "yàn", "sp": "yan", "r": "宀", "sc": 10, "l": 5, "f": 2155},
  {"h": "嘱", "p": "zhǔ", "sp": "zhu", "r": "口", "sc": 15, "l": 5, "f": 2156},
  {
    "h": "衷",
    "p": "zhōng",
    "sp": "zhong",
    "r": "衣",
    "sc": 10,
    "l": 5,
    "f": 2157
  },
  {"h": "娇", "p": "jiāo", "sp": "jiao", "r": "女", "sc": 9, "l": 5, "f": 2158},
  {"h": "陕", "p": "shǎn", "sp": "shan", "r": "阜", "sc": 8, "l": 5, "f": 2159},
  {"h": "矩", "p": "jǔ", "sp": "ju", "r": "矢", "sc": 9, "l": 5, "f": 2160},
  {"h": "浦", "p": "pǔ", "sp": "pu", "r": "水", "sc": 10, "l": 5, "f": 2161},
  {"h": "讶", "p": "yà", "sp": "ya", "r": "言", "sc": 6, "l": 5, "f": 2162},
  {"h": "耸", "p": "sǒng", "sp": "song", "r": "耳", "sc": 10, "l": 5, "f": 2163},
  {"h": "裸", "p": "luǒ", "sp": "luo", "r": "衣", "sc": 13, "l": 5, "f": 2164},
  {"h": "碧", "p": "bì", "sp": "bi", "r": "石", "sc": 14, "l": 5, "f": 2165},
  {"h": "摧", "p": "cuī", "sp": "cui", "r": "手", "sc": 14, "l": 5, "f": 2166},
  {"h": "薪", "p": "xīn", "sp": "xin", "r": "艸", "sc": 16, "l": 5, "f": 2167},
  {"h": "淋", "p": "lín", "sp": "lin", "r": "水", "sc": 11, "l": 5, "f": 2168},
  {"h": "耻", "p": "chǐ", "sp": "chi", "r": "耳", "sc": 10, "l": 5, "f": 2169},
  {"h": "胶", "p": "jiāo", "sp": "jiao", "r": "肉", "sc": 10, "l": 5, "f": 2170},
  {"h": "屠", "p": "tú", "sp": "tu", "r": "尸", "sc": 11, "l": 5, "f": 2171},
  {"h": "鹅", "p": "é", "sp": "e", "r": "鳥", "sc": 12, "l": 5, "f": 2172},
  {"h": "饥", "p": "jī", "sp": "ji", "r": "食", "sc": 5, "l": 5, "f": 2173},
  {"h": "盼", "p": "pàn", "sp": "pan", "r": "目", "sc": 9, "l": 5, "f": 2174},
  {"h": "脖", "p": "bó", "sp": "bo", "r": "肉", "sc": 11, "l": 5, "f": 2175},
  {"h": "虹", "p": "hóng", "sp": "hong", "r": "虫", "sc": 9, "l": 5, "f": 2176},
  {"h": "翠", "p": "cuì", "sp": "cui", "r": "羽", "sc": 14, "l": 5, "f": 2177},
  {"h": "崩", "p": "bēng", "sp": "beng", "r": "山", "sc": 11, "l": 5, "f": 2178},
  {"h": "账", "p": "zhàng", "sp": "zhang", "r": "貝", "sc": 8, "l": 5, "f": 2179},
  {"h": "萍", "p": "píng", "sp": "ping", "r": "艸", "sc": 11, "l": 5, "f": 2180},
  {"h": "逢", "p": "féng", "sp": "feng", "r": "辵", "sc": 10, "l": 5, "f": 2181},
  {
    "h": "赚",
    "p": "zhuàn",
    "sp": "zhuan",
    "r": "貝",
    "sc": 14,
    "l": 5,
    "f": 2182
  },
  {
    "h": "撑",
    "p": "chēng",
    "sp": "cheng",
    "r": "手",
    "sc": 15,
    "l": 5,
    "f": 2183
  },
  {
    "h": "翔",
    "p": "xiáng",
    "sp": "xiang",
    "r": "羽",
    "sc": 12,
    "l": 5,
    "f": 2184
  },
  {
    "h": "倡",
    "p": "chàng",
    "sp": "chang",
    "r": "人",
    "sc": 10,
    "l": 5,
    "f": 2185
  },
  {"h": "绵", "p": "mián", "sp": "mian", "r": "糸", "sc": 11, "l": 5, "f": 2186},
  {"h": "猴", "p": "hóu", "sp": "hou", "r": "犬", "sc": 12, "l": 5, "f": 2187},
  {"h": "枯", "p": "kū", "sp": "ku", "r": "木", "sc": 9, "l": 5, "f": 2188},
  {"h": "巫", "p": "wū", "sp": "wu", "r": "工", "sc": 7, "l": 5, "f": 2189},
  {"h": "昭", "p": "zhāo", "sp": "zhao", "r": "日", "sc": 9, "l": 5, "f": 2190},
  {"h": "怔", "p": "zhēng", "sp": "zheng", "r": "心", "sc": 8, "l": 5, "f": 2191},
  {"h": "渊", "p": "yuān", "sp": "yuan", "r": "水", "sc": 11, "l": 5, "f": 2192},
  {"h": "凑", "p": "còu", "sp": "cou", "r": "冫", "sc": 11, "l": 5, "f": 2193},
  {"h": "溪", "p": "xī", "sp": "xi", "r": "水", "sc": 13, "l": 5, "f": 2194},
  {"h": "蠢", "p": "chǔn", "sp": "chun", "r": "虫", "sc": 21, "l": 5, "f": 2195},
  {"h": "禅", "p": "chán", "sp": "chan", "r": "示", "sc": 12, "l": 5, "f": 2196},
  {"h": "阐", "p": "chǎn", "sp": "chan", "r": "門", "sc": 11, "l": 5, "f": 2197},
  {"h": "旺", "p": "wàng", "sp": "wang", "r": "日", "sc": 8, "l": 5, "f": 2198},
  {"h": "寓", "p": "yù", "sp": "yu", "r": "宀", "sc": 12, "l": 5, "f": 2199},
  {"h": "藤", "p": "téng", "sp": "teng", "r": "艸", "sc": 18, "l": 5, "f": 2200},
  {"h": "啸", "p": "xiào", "sp": "xiao", "r": "口", "sc": 11, "l": 5, "f": 2301},
  {"h": "滞", "p": "zhì", "sp": "zhi", "r": "水", "sc": 12, "l": 5, "f": 2302},
  {"h": "玫", "p": "méi", "sp": "mei", "r": "玉", "sc": 8, "l": 5, "f": 2303},
  {"h": "硕", "p": "shuò", "sp": "shuo", "r": "石", "sc": 11, "l": 5, "f": 2304},
  {"h": "钓", "p": "diào", "sp": "diao", "r": "金", "sc": 8, "l": 5, "f": 2305},
  {"h": "蝶", "p": "dié", "sp": "die", "r": "虫", "sc": 15, "l": 5, "f": 2306},
  {"h": "膝", "p": "xī", "sp": "xi", "r": "肉", "sc": 15, "l": 5, "f": 2307},
  {"h": "姚", "p": "yáo", "sp": "yao", "r": "女", "sc": 9, "l": 5, "f": 2308},
  {"h": "茂", "p": "mào", "sp": "mao", "r": "艸", "sc": 8, "l": 5, "f": 2309},
  {"h": "躯", "p": "qū", "sp": "qu", "r": "身", "sc": 11, "l": 5, "f": 2310},
  {"h": "吏", "p": "lì", "sp": "li", "r": "口", "sc": 6, "l": 5, "f": 2311},
  {"h": "猿", "p": "yuán", "sp": "yuan", "r": "犬", "sc": 13, "l": 5, "f": 2312},
  {"h": "寨", "p": "zhài", "sp": "zhai", "r": "宀", "sc": 14, "l": 5, "f": 2313},
  {"h": "恕", "p": "shù", "sp": "shu", "r": "心", "sc": 10, "l": 5, "f": 2314},
  {"h": "渠", "p": "qú", "sp": "qu", "r": "水", "sc": 11, "l": 5, "f": 2315},
  {"h": "戚", "p": "qī", "sp": "qi", "r": "戈", "sc": 11, "l": 5, "f": 2316},
  {"h": "辰", "p": "chén", "sp": "chen", "r": "辰", "sc": 7, "l": 5, "f": 2317},
  {"h": "舶", "p": "bó", "sp": "bo", "r": "舟", "sc": 11, "l": 5, "f": 2318},
  {"h": "颁", "p": "bān", "sp": "ban", "r": "頁", "sc": 10, "l": 5, "f": 2319},
  {
    "h": "惶",
    "p": "huáng",
    "sp": "huang",
    "r": "心",
    "sc": 12,
    "l": 5,
    "f": 2320
  },
  {"h": "狐", "p": "hú", "sp": "hu", "r": "犬", "sc": 8, "l": 5, "f": 2321},
  {"h": "讽", "p": "fěng", "sp": "feng", "r": "言", "sc": 6, "l": 5, "f": 2322},
  {"h": "笨", "p": "bèn", "sp": "ben", "r": "竹", "sc": 11, "l": 4, "f": 2323},
  {"h": "袍", "p": "páo", "sp": "pao", "r": "衣", "sc": 10, "l": 5, "f": 2324},
  {"h": "嘲", "p": "cháo", "sp": "chao", "r": "口", "sc": 15, "l": 5, "f": 2325},
  {"h": "啡", "p": "fēi", "sp": "fei", "r": "口", "sc": 11, "l": 2, "f": 2326},
  {"h": "泼", "p": "pō", "sp": "po", "r": "水", "sc": 8, "l": 4, "f": 2327},
  {"h": "衔", "p": "xián", "sp": "xian", "r": "彳", "sc": 11, "l": 5, "f": 2328},
  {"h": "倦", "p": "juàn", "sp": "juan", "r": "人", "sc": 10, "l": 5, "f": 2329},
  {"h": "涵", "p": "hán", "sp": "han", "r": "水", "sc": 11, "l": 5, "f": 2330},
  {"h": "雀", "p": "què", "sp": "que", "r": "隹", "sc": 11, "l": 5, "f": 2331},
  {"h": "旬", "p": "xún", "sp": "xun", "r": "日", "sc": 6, "l": 5, "f": 2332},
  {
    "h": "僵",
    "p": "jiāng",
    "sp": "jiang",
    "r": "人",
    "sc": 15,
    "l": 5,
    "f": 2333
  },
  {"h": "撕", "p": "sī", "sp": "si", "r": "手", "sc": 15, "l": 5, "f": 2334},
  {"h": "肢", "p": "zhī", "sp": "zhi", "r": "肉", "sc": 8, "l": 5, "f": 2335},
  {"h": "垄", "p": "lǒng", "sp": "long", "r": "龍", "sc": 8, "l": 5, "f": 2336},
  {"h": "夷", "p": "yí", "sp": "yi", "r": "大", "sc": 6, "l": 5, "f": 2337},
  {"h": "逸", "p": "yì", "sp": "yi", "r": "辵", "sc": 11, "l": 5, "f": 2338},
  {"h": "茅", "p": "máo", "sp": "mao", "r": "艸", "sc": 8, "l": 5, "f": 2339},
  {"h": "侨", "p": "qiáo", "sp": "qiao", "r": "人", "sc": 8, "l": 5, "f": 2340},
  {"h": "舆", "p": "yú", "sp": "yu", "r": "八", "sc": 14, "l": 5, "f": 2341},
  {"h": "窑", "p": "yáo", "sp": "yao", "r": "穴", "sc": 11, "l": 5, "f": 2342},
  {"h": "涅", "p": "niè", "sp": "nie", "r": "水", "sc": 10, "l": 5, "f": 2343},
  {"h": "蒲", "p": "pú", "sp": "pu", "r": "艸", "sc": 13, "l": 5, "f": 2344},
  {"h": "谦", "p": "qiān", "sp": "qian", "r": "言", "sc": 12, "l": 5, "f": 2345},
  {"h": "杭", "p": "háng", "sp": "hang", "r": "木", "sc": 8, "l": 5, "f": 2346},
  {"h": "噢", "p": "ō", "sp": "o", "r": "口", "sc": 15, "l": 5, "f": 2347},
  {"h": "弊", "p": "bì", "sp": "bi", "r": "廾", "sc": 14, "l": 5, "f": 2348},
  {"h": "勋", "p": "xūn", "sp": "xun", "r": "力", "sc": 9, "l": 5, "f": 2349},
  {"h": "刮", "p": "guā", "sp": "gua", "r": "刀", "sc": 8, "l": 5, "f": 2350},
  {"h": "郊", "p": "jiāo", "sp": "jiao", "r": "邑", "sc": 8, "l": 5, "f": 2351},
  {"h": "凄", "p": "qī", "sp": "qi", "r": "冫", "sc": 10, "l": 5, "f": 2352},
  {"h": "捧", "p": "pěng", "sp": "peng", "r": "手", "sc": 11, "l": 5, "f": 2353},
  {"h": "浸", "p": "jìn", "sp": "jin", "r": "水", "sc": 10, "l": 5, "f": 2354},
  {"h": "砖", "p": "zhuān", "sp": "zhuan", "r": "石", "sc": 9, "l": 5, "f": 2355},
  {"h": "鼎", "p": "dǐng", "sp": "ding", "r": "鼎", "sc": 12, "l": 5, "f": 2356},
  {"h": "篮", "p": "lán", "sp": "lan", "r": "竹", "sc": 16, "l": 2, "f": 2357},
  {
    "h": "蒸",
    "p": "zhēng",
    "sp": "zheng",
    "r": "艸",
    "sc": 13,
    "l": 5,
    "f": 2358
  },
  {"h": "饼", "p": "bǐng", "sp": "bing", "r": "食", "sc": 9, "l": 4, "f": 2359},
  {"h": "亩", "p": "mǔ", "sp": "mu", "r": "亠", "sc": 7, "l": 5, "f": 2360},
  {"h": "肾", "p": "shèn", "sp": "shen", "r": "肉", "sc": 8, "l": 5, "f": 2361},
  {"h": "陡", "p": "dǒu", "sp": "dou", "r": "阜", "sc": 9, "l": 5, "f": 2362},
  {"h": "爪", "p": "zhǎo", "sp": "zhao", "r": "爪", "sc": 4, "l": 5, "f": 2363},
  {"h": "兔", "p": "tù", "sp": "tu", "r": "儿", "sc": 8, "l": 5, "f": 2364},
  {"h": "殷", "p": "yīn", "sp": "yin", "r": "殳", "sc": 10, "l": 5, "f": 2365},
  {"h": "贞", "p": "zhēn", "sp": "zhen", "r": "卜", "sc": 6, "l": 5, "f": 2366},
  {"h": "荐", "p": "jiàn", "sp": "jian", "r": "艸", "sc": 9, "l": 5, "f": 2367},
  {"h": "哑", "p": "yǎ", "sp": "ya", "r": "口", "sc": 9, "l": 5, "f": 2368},
  {"h": "炭", "p": "tàn", "sp": "tan", "r": "火", "sc": 9, "l": 5, "f": 2369},
  {"h": "坟", "p": "fén", "sp": "fen", "r": "土", "sc": 7, "l": 5, "f": 2370},
  {"h": "眨", "p": "zhǎ", "sp": "zha", "r": "目", "sc": 9, "l": 5, "f": 2371},
  {"h": "搏", "p": "bó", "sp": "bo", "r": "手", "sc": 13, "l": 5, "f": 2372},
  {"h": "咳", "p": "hāi", "sp": "hai", "r": "口", "sc": 9, "l": 4, "f": 2373},
  {"h": "拢", "p": "lǒng", "sp": "long", "r": "手", "sc": 8, "l": 5, "f": 2374},
  {"h": "舅", "p": "jiù", "sp": "jiu", "r": "臼", "sc": 13, "l": 5, "f": 2375},
  {"h": "昧", "p": "mèi", "sp": "mei", "r": "日", "sc": 9, "l": 5, "f": 2376},
  {"h": "擅", "p": "shàn", "sp": "shan", "r": "手", "sc": 16, "l": 5, "f": 2377},
  {
    "h": "爽",
    "p": "shuǎng",
    "sp": "shuang",
    "r": "爻",
    "sc": 11,
    "l": 5,
    "f": 2378
  },
  {"h": "咖", "p": "kā", "sp": "ka", "r": "口", "sc": 8, "l": 2, "f": 2379},
  {"h": "搁", "p": "gē", "sp": "ge", "r": "手", "sc": 12, "l": 5, "f": 2380},
  {"h": "禄", "p": "lù", "sp": "lu", "r": "示", "sc": 12, "l": 5, "f": 2381},
  {"h": "雌", "p": "cí", "sp": "ci", "r": "隹", "sc": 14, "l": 5, "f": 2382},
  {"h": "哨", "p": "shào", "sp": "shao", "r": "口", "sc": 10, "l": 5, "f": 2383},
  {"h": "巩", "p": "gǒng", "sp": "gong", "r": "工", "sc": 6, "l": 5, "f": 2384},
  {"h": "绢", "p": "juàn", "sp": "juan", "r": "糸", "sc": 10, "l": 5, "f": 2385},
  {"h": "螺", "p": "luó", "sp": "luo", "r": "虫", "sc": 17, "l": 5, "f": 2386},
  {"h": "裹", "p": "guǒ", "sp": "guo", "r": "衣", "sc": 14, "l": 4, "f": 2387},
  {"h": "昔", "p": "xī", "sp": "xi", "r": "日", "sc": 8, "l": 5, "f": 2388},
  {"h": "轩", "p": "xuān", "sp": "xuan", "r": "車", "sc": 7, "l": 5, "f": 2389},
  {"h": "谬", "p": "miù", "sp": "miu", "r": "言", "sc": 13, "l": 5, "f": 2390},
  {"h": "谍", "p": "dié", "sp": "die", "r": "言", "sc": 11, "l": 5, "f": 2391},
  {"h": "龟", "p": "guī", "sp": "gui", "r": "龜", "sc": 7, "l": 5, "f": 2392},
  {"h": "媳", "p": "xí", "sp": "xi", "r": "女", "sc": 13, "l": 5, "f": 2393},
  {"h": "姜", "p": "jiāng", "sp": "jiang", "r": "女", "sc": 9, "l": 5, "f": 2394},
  {"h": "瞎", "p": "xiā", "sp": "xia", "r": "目", "sc": 15, "l": 5, "f": 2395},
  {"h": "冤", "p": "yuān", "sp": "yuan", "r": "冖", "sc": 10, "l": 5, "f": 2396},
  {"h": "鸦", "p": "yā", "sp": "ya", "r": "鳥", "sc": 9, "l": 5, "f": 2397},
  {"h": "蓬", "p": "péng", "sp": "peng", "r": "艸", "sc": 13, "l": 5, "f": 2398},
  {"h": "巷", "p": "xiàng", "sp": "xiang", "r": "己", "sc": 9, "l": 5, "f": 2399},
  {"h": "琳", "p": "lín", "sp": "lin", "r": "玉", "sc": 12, "l": 4, "f": 2400},
  {"h": "栽", "p": "zāi", "sp": "zai", "r": "木", "sc": 10, "l": 5, "f": 2401},
  {"h": "沾", "p": "zhān", "sp": "zhan", "r": "水", "sc": 8, "l": 5, "f": 2402},
  {"h": "诈", "p": "zhà", "sp": "zha", "r": "言", "sc": 7, "l": 5, "f": 2403},
  {"h": "斋", "p": "zhāi", "sp": "zhai", "r": "文", "sc": 10, "l": 5, "f": 2404},
  {"h": "瞒", "p": "mán", "sp": "man", "r": "目", "sc": 15, "l": 5, "f": 2405},
  {"h": "彪", "p": "biāo", "sp": "biao", "r": "彡", "sc": 11, "l": 5, "f": 2406},
  {"h": "厄", "p": "è", "sp": "e", "r": "厂", "sc": 4, "l": 5, "f": 2407},
  {"h": "咨", "p": "zī", "sp": "zi", "r": "口", "sc": 9, "l": 5, "f": 2408},
  {"h": "纺", "p": "fǎng", "sp": "fang", "r": "糸", "sc": 7, "l": 5, "f": 2409},
  {"h": "罐", "p": "guàn", "sp": "guan", "r": "缶", "sc": 23, "l": 5, "f": 2410},
  {"h": "桶", "p": "tǒng", "sp": "tong", "r": "木", "sc": 11, "l": 4, "f": 2411},
  {"h": "壤", "p": "rǎng", "sp": "rang", "r": "土", "sc": 20, "l": 5, "f": 2412},
  {"h": "糕", "p": "gāo", "sp": "gao", "r": "米", "sc": 16, "l": 3, "f": 2413},
  {"h": "颂", "p": "sòng", "sp": "song", "r": "頁", "sc": 10, "l": 5, "f": 2414},
  {"h": "膨", "p": "péng", "sp": "peng", "r": "肉", "sc": 16, "l": 5, "f": 2415},
  {"h": "谐", "p": "xié", "sp": "xie", "r": "言", "sc": 11, "l": 5, "f": 2416},
  {"h": "垒", "p": "lěi", "sp": "lei", "r": "土", "sc": 9, "l": 5, "f": 2417},
  {"h": "咕", "p": "gū", "sp": "gu", "r": "口", "sc": 8, "l": 5, "f": 2418},
  {"h": "隙", "p": "xì", "sp": "xi", "r": "阜", "sc": 12, "l": 5, "f": 2419},
  {"h": "辣", "p": "là", "sp": "la", "r": "辛", "sc": 14, "l": 4, "f": 2420},
  {"h": "绑", "p": "bǎng", "sp": "bang", "r": "糸", "sc": 9, "l": 5, "f": 2421},
  {"h": "宠", "p": "chǒng", "sp": "chong", "r": "宀", "sc": 8, "l": 5, "f": 2422},
  {"h": "嘿", "p": "hēi", "sp": "hei", "r": "口", "sc": 15, "l": 5, "f": 2423},
  {"h": "兑", "p": "duì", "sp": "dui", "r": "儿", "sc": 7, "l": 5, "f": 2424},
  {"h": "霉", "p": "méi", "sp": "mei", "r": "雨", "sc": 15, "l": 5, "f": 2425},
  {"h": "挫", "p": "cuò", "sp": "cuo", "r": "手", "sc": 10, "l": 5, "f": 2426},
  {"h": "稽", "p": "jī", "sp": "ji", "r": "禾", "sc": 15, "l": 5, "f": 2427},
  {"h": "辐", "p": "fú", "sp": "fu", "r": "車", "sc": 13, "l": 5, "f": 2428},
  {"h": "乞", "p": "qǐ", "sp": "qi", "r": "乙", "sc": 3, "l": 5, "f": 2429},
  {"h": "纱", "p": "shā", "sp": "sha", "r": "糸", "sc": 7, "l": 5, "f": 2430},
  {"h": "裙", "p": "qún", "sp": "qun", "r": "衣", "sc": 12, "l": 3, "f": 2431},
  {"h": "嘻", "p": "xī", "sp": "xi", "r": "口", "sc": 15, "l": 5, "f": 2432},
  {"h": "哇", "p": "wa", "sp": "wa", "r": "口", "sc": 9, "l": 5, "f": 2433},
  {"h": "绣", "p": "xiù", "sp": "xiu", "r": "糸", "sc": 10, "l": 5, "f": 2434},
  {"h": "杖", "p": "zhàng", "sp": "zhang", "r": "木", "sc": 7, "l": 5, "f": 2435},
  {"h": "塘", "p": "táng", "sp": "tang", "r": "土", "sc": 13, "l": 5, "f": 2436},
  {"h": "衍", "p": "yǎn", "sp": "yan", "r": "行", "sc": 9, "l": 5, "f": 2437},
  {"h": "轴", "p": "zhóu", "sp": "zhou", "r": "車", "sc": 9, "l": 5, "f": 2438},
  {"h": "攀", "p": "pān", "sp": "pan", "r": "手", "sc": 19, "l": 5, "f": 2439},
  {"h": "膊", "p": "bó", "sp": "bo", "r": "肉", "sc": 14, "l": 5, "f": 2440},
  {"h": "譬", "p": "pì", "sp": "pi", "r": "言", "sc": 20, "l": 5, "f": 2441},
  {"h": "斌", "p": "bīn", "sp": "bin", "r": "文", "sc": 12, "l": 5, "f": 2442},
  {"h": "祈", "p": "qí", "sp": "qi", "r": "示", "sc": 8, "l": 5, "f": 2443},
  {"h": "踢", "p": "tī", "sp": "ti", "r": "足", "sc": 15, "l": 2, "f": 2444},
  {"h": "肆", "p": "sì", "sp": "si", "r": "聿", "sc": 13, "l": 5, "f": 2445},
  {"h": "坎", "p": "kǎn", "sp": "kan", "r": "土", "sc": 7, "l": 5, "f": 2446},
  {"h": "轿", "p": "jiào", "sp": "jiao", "r": "車", "sc": 10, "l": 5, "f": 2447},
  {"h": "棚", "p": "péng", "sp": "peng", "r": "木", "sc": 12, "l": 5, "f": 2448},
  {"h": "泣", "p": "qì", "sp": "qi", "r": "水", "sc": 8, "l": 5, "f": 2449},
  {"h": "屡", "p": "lǚ", "sp": "lu", "r": "尸", "sc": 12, "l": 5, "f": 2450},
  {"h": "躁", "p": "zào", "sp": "zao", "r": "足", "sc": 20, "l": 5, "f": 2451},
  {"h": "邱", "p": "qiū", "sp": "qiu", "r": "邑", "sc": 7, "l": 5, "f": 2452},
  {
    "h": "凰",
    "p": "huáng",
    "sp": "huang",
    "r": "几",
    "sc": 11,
    "l": 5,
    "f": 2453
  },
  {"h": "溢", "p": "yì", "sp": "yi", "r": "水", "sc": 13, "l": 5, "f": 2454},
  {"h": "椎", "p": "chuí", "sp": "chui", "r": "木", "sc": 12, "l": 5, "f": 2455},
  {"h": "砸", "p": "zá", "sp": "za", "r": "石", "sc": 10, "l": 5, "f": 2456},
  {"h": "趟", "p": "tàng", "sp": "tang", "r": "走", "sc": 15, "l": 5, "f": 2457},
  {"h": "帘", "p": "lián", "sp": "lian", "r": "巾", "sc": 8, "l": 5, "f": 2458},
  {"h": "帆", "p": "fān", "sp": "fan", "r": "巾", "sc": 6, "l": 5, "f": 2459},
  {"h": "栖", "p": "qī", "sp": "qi", "r": "木", "sc": 10, "l": 5, "f": 2460},
  {"h": "窜", "p": "cuàn", "sp": "cuan", "r": "穴", "sc": 12, "l": 5, "f": 2461},
  {"h": "丸", "p": "wán", "sp": "wan", "r": "丶", "sc": 3, "l": 5, "f": 2462},
  {"h": "斩", "p": "zhǎn", "sp": "zhan", "r": "車", "sc": 8, "l": 5, "f": 2463},
  {"h": "堤", "p": "dī", "sp": "di", "r": "土", "sc": 12, "l": 5, "f": 2464},
  {"h": "塌", "p": "tā", "sp": "ta", "r": "土", "sc": 13, "l": 5, "f": 2465},
  {"h": "贩", "p": "fàn", "sp": "fan", "r": "貝", "sc": 8, "l": 5, "f": 2466},
  {
    "h": "厢",
    "p": "xiāng",
    "sp": "xiang",
    "r": "厂",
    "sc": 11,
    "l": 5,
    "f": 2467
  },
  {"h": "掀", "p": "xiān", "sp": "xian", "r": "手", "sc": 11, "l": 5, "f": 2468},
  {"h": "喀", "p": "kā", "sp": "ka", "r": "口", "sc": 12, "l": 5, "f": 2469},
  {"h": "乖", "p": "guāi", "sp": "guai", "r": "丿", "sc": 8, "l": 5, "f": 2470},
  {"h": "谜", "p": "mí", "sp": "mi", "r": "言", "sc": 11, "l": 5, "f": 2471},
  {"h": "捏", "p": "niē", "sp": "nie", "r": "手", "sc": 10, "l": 5, "f": 2472},
  {"h": "阎", "p": "yán", "sp": "yan", "r": "門", "sc": 11, "l": 5, "f": 2473},
  {"h": "滨", "p": "bīn", "sp": "bin", "r": "水", "sc": 13, "l": 5, "f": 2474},
  {"h": "虏", "p": "lǔ", "sp": "lu", "r": "虍", "sc": 8, "l": 5, "f": 2475},
  {"h": "匙", "p": "shi", "sp": "shi", "r": "匕", "sc": 11, "l": 4, "f": 2476},
  {"h": "芦", "p": "lú", "sp": "lu", "r": "艸", "sc": 7, "l": 5, "f": 2477},
  {"h": "苹", "p": "píng", "sp": "ping", "r": "艸", "sc": 8, "l": 1, "f": 2478},
  {"h": "卸", "p": "xiè", "sp": "xie", "r": "卩", "sc": 9, "l": 5, "f": 2479},
  {"h": "沼", "p": "zhǎo", "sp": "zhao", "r": "水", "sc": 8, "l": 5, "f": 2480},
  {"h": "钥", "p": "yào", "sp": "yao", "r": "金", "sc": 9, "l": 4, "f": 2481},
  {"h": "株", "p": "zhū", "sp": "zhu", "r": "木", "sc": 10, "l": 5, "f": 2482},
  {"h": "祷", "p": "dǎo", "sp": "dao", "r": "示", "sc": 11, "l": 5, "f": 2483},
  {"h": "剖", "p": "pōu", "sp": "pou", "r": "刀", "sc": 10, "l": 5, "f": 2484},
  {"h": "熙", "p": "xī", "sp": "xi", "r": "火", "sc": 14, "l": 5, "f": 2485},
  {"h": "哗", "p": "huā", "sp": "hua", "r": "口", "sc": 9, "l": 5, "f": 2486},
  {"h": "劈", "p": "pī", "sp": "pi", "r": "刀", "sc": 15, "l": 5, "f": 2487},
  {"h": "怯", "p": "qiè", "sp": "qie", "r": "心", "sc": 8, "l": 5, "f": 2488},
  {"h": "棠", "p": "táng", "sp": "tang", "r": "木", "sc": 12, "l": 5, "f": 2489},
  {"h": "胳", "p": "gē", "sp": "ge", "r": "肉", "sc": 10, "l": 5, "f": 2490},
  {
    "h": "桩",
    "p": "zhuāng",
    "sp": "zhuang",
    "r": "木",
    "sc": 10,
    "l": 5,
    "f": 2491
  },
  {"h": "瑰", "p": "guī", "sp": "gui", "r": "玉", "sc": 13, "l": 5, "f": 2492},
  {"h": "娱", "p": "yú", "sp": "yu", "r": "女", "sc": 10, "l": 5, "f": 2493},
  {"h": "娶", "p": "qǔ", "sp": "qu", "r": "女", "sc": 11, "l": 5, "f": 2494},
  {"h": "沫", "p": "mò", "sp": "mo", "r": "水", "sc": 8, "l": 5, "f": 2495},
  {"h": "嗓", "p": "sǎng", "sp": "sang", "r": "口", "sc": 13, "l": 5, "f": 2496},
  {"h": "蹲", "p": "dūn", "sp": "dun", "r": "足", "sc": 19, "l": 5, "f": 2497},
  {"h": "焚", "p": "fén", "sp": "fen", "r": "火", "sc": 12, "l": 5, "f": 2498},
  {"h": "淘", "p": "táo", "sp": "tao", "r": "水", "sc": 11, "l": 5, "f": 2499},
  {"h": "嫩", "p": "nèn", "sp": "nen", "r": "女", "sc": 14, "l": 5, "f": 2500},
  {"h": "韵", "p": "yùn", "sp": "yun", "r": "音", "sc": 13, "l": 5, "f": 2501},
  {"h": "衬", "p": "chèn", "sp": "chen", "r": "衣", "sc": 8, "l": 3, "f": 2502},
  {"h": "匈", "p": "xiōng", "sp": "xiong", "r": "勹", "sc": 6, "l": 5, "f": 2503},
  {"h": "钧", "p": "jūn", "sp": "jun", "r": "金", "sc": 9, "l": 5, "f": 2504},
  {"h": "竖", "p": "shù", "sp": "shu", "r": "立", "sc": 9, "l": 5, "f": 2505},
  {"h": "峻", "p": "jùn", "sp": "jun", "r": "山", "sc": 10, "l": 5, "f": 2506},
  {"h": "豹", "p": "bào", "sp": "bao", "r": "豸", "sc": 10, "l": 5, "f": 2507},
  {"h": "捞", "p": "lāo", "sp": "lao", "r": "手", "sc": 10, "l": 5, "f": 2508},
  {"h": "菊", "p": "jú", "sp": "ju", "r": "艸", "sc": 11, "l": 5, "f": 2509},
  {"h": "鄙", "p": "bǐ", "sp": "bi", "r": "邑", "sc": 13, "l": 5, "f": 2510},
  {"h": "魄", "p": "pò", "sp": "po", "r": "鬼", "sc": 14, "l": 5, "f": 2511},
  {"h": "兜", "p": "dōu", "sp": "dou", "r": "儿", "sc": 11, "l": 5, "f": 2512},
  {"h": "哄", "p": "hōng", "sp": "hong", "r": "口", "sc": 9, "l": 5, "f": 2513},
  {"h": "颖", "p": "yǐng", "sp": "ying", "r": "頁", "sc": 13, "l": 5, "f": 2514},
  {"h": "镑", "p": "bàng", "sp": "bang", "r": "金", "sc": 15, "l": 5, "f": 2515},
  {"h": "屑", "p": "xiè", "sp": "xie", "r": "尸", "sc": 10, "l": 5, "f": 2516},
  {"h": "蚁", "p": "yǐ", "sp": "yi", "r": "虫", "sc": 9, "l": 5, "f": 2517},
  {"h": "壶", "p": "hú", "sp": "hu", "r": "士", "sc": 10, "l": 5, "f": 2518},
  {"h": "怡", "p": "yí", "sp": "yi", "r": "心", "sc": 8, "l": 5, "f": 2519},
  {"h": "渗", "p": "shèn", "sp": "shen", "r": "水", "sc": 11, "l": 5, "f": 2520},
  {"h": "秃", "p": "tū", "sp": "tu", "r": "禾", "sc": 7, "l": 5, "f": 2521},
  {"h": "迦", "p": "jiā", "sp": "jia", "r": "辵", "sc": 8, "l": 5, "f": 2522},
  {"h": "旱", "p": "hàn", "sp": "han", "r": "日", "sc": 7, "l": 5, "f": 2523},
  {"h": "哟", "p": "yō", "sp": "yo", "r": "口", "sc": 9, "l": 5, "f": 2524},
  {"h": "咸", "p": "xián", "sp": "xian", "r": "口", "sc": 9, "l": 4, "f": 2525},
  {"h": "焉", "p": "yān", "sp": "yan", "r": "火", "sc": 11, "l": 5, "f": 2526},
  {"h": "谴", "p": "qiǎn", "sp": "qian", "r": "言", "sc": 15, "l": 5, "f": 2527},
  {"h": "宛", "p": "wǎn", "sp": "wan", "r": "宀", "sc": 8, "l": 5, "f": 2528},
  {"h": "稻", "p": "dào", "sp": "dao", "r": "禾", "sc": 15, "l": 5, "f": 2529},
  {"h": "铸", "p": "zhù", "sp": "zhu", "r": "金", "sc": 12, "l": 5, "f": 2530},
  {"h": "锻", "p": "duàn", "sp": "duan", "r": "金", "sc": 14, "l": 4, "f": 2531},
  {"h": "伽", "p": "jiā", "sp": "jia", "r": "人", "sc": 7, "l": 5, "f": 2532},
  {"h": "詹", "p": "zhān", "sp": "zhan", "r": "言", "sc": 13, "l": 5, "f": 2533},
  {"h": "毙", "p": "bì", "sp": "bi", "r": "比", "sc": 10, "l": 5, "f": 2534},
  {"h": "恍", "p": "huǎng", "sp": "huang", "r": "心", "sc": 9, "l": 5, "f": 2535},
  {"h": "贬", "p": "biǎn", "sp": "bian", "r": "貝", "sc": 8, "l": 5, "f": 2536},
  {"h": "烛", "p": "zhú", "sp": "zhu", "r": "火", "sc": 10, "l": 5, "f": 2537},
  {"h": "骇", "p": "hài", "sp": "hai", "r": "馬", "sc": 9, "l": 5, "f": 2538},
  {"h": "芯", "p": "xīn", "sp": "xin", "r": "艸", "sc": 7, "l": 5, "f": 2539},
  {"h": "汁", "p": "zhī", "sp": "zhi", "r": "水", "sc": 5, "l": 3, "f": 2540},
  {"h": "桓", "p": "huán", "sp": "huan", "r": "木", "sc": 10, "l": 5, "f": 2541},
  {"h": "坊", "p": "fāng", "sp": "fang", "r": "土", "sc": 7, "l": 5, "f": 2542},
  {"h": "驴", "p": "lǘ", "sp": "lu", "r": "馬", "sc": 7, "l": 5, "f": 2543},
  {"h": "朽", "p": "xiǔ", "sp": "xiu", "r": "木", "sc": 6, "l": 5, "f": 2544},
  {"h": "靖", "p": "jìng", "sp": "jing", "r": "青", "sc": 13, "l": 5, "f": 2545},
  {"h": "佣", "p": "yōng", "sp": "yong", "r": "人", "sc": 7, "l": 5, "f": 2546},
  {"h": "汝", "p": "rǔ", "sp": "ru", "r": "水", "sc": 6, "l": 5, "f": 2547},
  {"h": "碌", "p": "lù", "sp": "lu", "r": "石", "sc": 13, "l": 5, "f": 2548},
  {"h": "迄", "p": "qì", "sp": "qi", "r": "辵", "sc": 6, "l": 5, "f": 2549},
  {"h": "冀", "p": "jì", "sp": "ji", "r": "八", "sc": 16, "l": 5, "f": 2550},
  {"h": "荆", "p": "jīng", "sp": "jing", "r": "刀", "sc": 9, "l": 5, "f": 2551},
  {"h": "崔", "p": "cuī", "sp": "cui", "r": "山", "sc": 11, "l": 5, "f": 2552},
  {"h": "雁", "p": "yàn", "sp": "yan", "r": "隹", "sc": 12, "l": 5, "f": 2553},
  {"h": "绅", "p": "shēn", "sp": "shen", "r": "糸", "sc": 8, "l": 5, "f": 2554},
  {"h": "珊", "p": "shān", "sp": "shan", "r": "玉", "sc": 9, "l": 5, "f": 2555},
  {"h": "榜", "p": "bǎng", "sp": "bang", "r": "木", "sc": 14, "l": 5, "f": 2556},
  {"h": "诵", "p": "sòng", "sp": "song", "r": "言", "sc": 9, "l": 5, "f": 2557},
  {"h": "傍", "p": "bàng", "sp": "bang", "r": "人", "sc": 12, "l": 5, "f": 2558},
  {"h": "彦", "p": "yàn", "sp": "yan", "r": "彡", "sc": 9, "l": 5, "f": 2559},
  {"h": "醇", "p": "chún", "sp": "chun", "r": "酉", "sc": 15, "l": 5, "f": 2560},
  {"h": "笛", "p": "dí", "sp": "di", "r": "竹", "sc": 11, "l": 5, "f": 2561},
  {"h": "禽", "p": "qín", "sp": "qin", "r": "禸", "sc": 12, "l": 5, "f": 2562},
  {"h": "勿", "p": "wù", "sp": "wu", "r": "勹", "sc": 4, "l": 5, "f": 2563},
  {"h": "娟", "p": "juān", "sp": "juan", "r": "女", "sc": 10, "l": 5, "f": 2564},
  {"h": "瞄", "p": "miáo", "sp": "miao", "r": "目", "sc": 13, "l": 5, "f": 2565},
  {
    "h": "幢",
    "p": "chuáng",
    "sp": "chuang",
    "r": "巾",
    "sc": 15,
    "l": 5,
    "f": 2566
  },
  {"h": "寇", "p": "kòu", "sp": "kou", "r": "宀", "sc": 11, "l": 5, "f": 2567},
  {"h": "睹", "p": "dǔ", "sp": "du", "r": "目", "sc": 13, "l": 5, "f": 2568},
  {"h": "贿", "p": "huì", "sp": "hui", "r": "貝", "sc": 10, "l": 5, "f": 2569},
  {"h": "踩", "p": "cǎi", "sp": "cai", "r": "足", "sc": 15, "l": 5, "f": 2570},
  {"h": "霆", "p": "tíng", "sp": "ting", "r": "雨", "sc": 14, "l": 5, "f": 2571},
  {"h": "呜", "p": "wū", "sp": "wu", "r": "口", "sc": 7, "l": 5, "f": 2572},
  {"h": "拱", "p": "gǒng", "sp": "gong", "r": "手", "sc": 9, "l": 5, "f": 2573},
  {"h": "妃", "p": "fēi", "sp": "fei", "r": "女", "sc": 6, "l": 5, "f": 2574},
  {"h": "蔑", "p": "miè", "sp": "mie", "r": "艸", "sc": 14, "l": 5, "f": 2575},
  {"h": "谕", "p": "yù", "sp": "yu", "r": "言", "sc": 11, "l": 5, "f": 2576},
  {"h": "缚", "p": "fù", "sp": "fu", "r": "糸", "sc": 13, "l": 5, "f": 2577},
  {"h": "诡", "p": "guǐ", "sp": "gui", "r": "言", "sc": 8, "l": 5, "f": 2578},
  {"h": "篷", "p": "péng", "sp": "peng", "r": "竹", "sc": 16, "l": 5, "f": 2579},
  {"h": "淹", "p": "yān", "sp": "yan", "r": "水", "sc": 11, "l": 5, "f": 2580},
  {"h": "腕", "p": "wàn", "sp": "wan", "r": "肉", "sc": 12, "l": 5, "f": 2581},
  {"h": "煮", "p": "zhǔ", "sp": "zhu", "r": "火", "sc": 12, "l": 5, "f": 2582},
  {"h": "倩", "p": "qiàn", "sp": "qian", "r": "人", "sc": 10, "l": 5, "f": 2583},
  {"h": "卒", "p": "zú", "sp": "zu", "r": "十", "sc": 8, "l": 5, "f": 2584},
  {"h": "勘", "p": "kān", "sp": "kan", "r": "力", "sc": 11, "l": 5, "f": 2585},
  {"h": "馨", "p": "xīn", "sp": "xin", "r": "香", "sc": 20, "l": 5, "f": 2586},
  {"h": "逗", "p": "dòu", "sp": "dou", "r": "辵", "sc": 10, "l": 5, "f": 2587},
  {"h": "甸", "p": "diān", "sp": "dian", "r": "田", "sc": 7, "l": 5, "f": 2588},
  {"h": "贱", "p": "jiàn", "sp": "jian", "r": "貝", "sc": 9, "l": 5, "f": 2589},
  {"h": "炒", "p": "chǎo", "sp": "chao", "r": "火", "sc": 8, "l": 1, "f": 2590},
  {"h": "灿", "p": "càn", "sp": "can", "r": "火", "sc": 7, "l": 5, "f": 2591},
  {
    "h": "敞",
    "p": "chǎng",
    "sp": "chang",
    "r": "攴",
    "sc": 12,
    "l": 5,
    "f": 2592
  },
  {"h": "蜡", "p": "là", "sp": "la", "r": "虫", "sc": 14, "l": 5, "f": 2593},
  {"h": "囚", "p": "qiú", "sp": "qiu", "r": "囗", "sc": 5, "l": 5, "f": 2594},
  {"h": "栗", "p": "lì", "sp": "li", "r": "木", "sc": 10, "l": 5, "f": 2595},
  {"h": "辜", "p": "gū", "sp": "gu", "r": "辛", "sc": 12, "l": 5, "f": 2596},
  {"h": "垫", "p": "diàn", "sp": "dian", "r": "土", "sc": 9, "l": 5, "f": 2597},
  {"h": "妒", "p": "dù", "sp": "du", "r": "女", "sc": 7, "l": 5, "f": 2598},
  {"h": "魁", "p": "kuí", "sp": "kui", "r": "鬼", "sc": 13, "l": 5, "f": 2599},
  {"h": "谣", "p": "yáo", "sp": "yao", "r": "言", "sc": 12, "l": 5, "f": 2600},
  {"h": "寞", "p": "mò", "sp": "mo", "r": "宀", "sc": 13, "l": 5, "f": 2601},
  {"h": "蜀", "p": "shǔ", "sp": "shu", "r": "虫", "sc": 13, "l": 5, "f": 2602},
  {"h": "甩", "p": "shuǎi", "sp": "shuai", "r": "用", "sc": 5, "l": 5, "f": 2603},
  {"h": "涯", "p": "yá", "sp": "ya", "r": "水", "sc": 11, "l": 5, "f": 2604},
  {"h": "枕", "p": "zhěn", "sp": "zhen", "r": "木", "sc": 8, "l": 5, "f": 2605},
  {"h": "丐", "p": "gài", "sp": "gai", "r": "一", "sc": 4, "l": 5, "f": 2606},
  {"h": "泳", "p": "yǒng", "sp": "yong", "r": "水", "sc": 8, "l": 2, "f": 2607},
  {"h": "奎", "p": "kuí", "sp": "kui", "r": "大", "sc": 9, "l": 5, "f": 2608},
  {"h": "泌", "p": "mì", "sp": "mi", "r": "水", "sc": 8, "l": 5, "f": 2609},
  {"h": "逾", "p": "yú", "sp": "yu", "r": "辵", "sc": 12, "l": 5, "f": 2610},
  {"h": "叮", "p": "dīng", "sp": "ding", "r": "口", "sc": 5, "l": 5, "f": 2611},
  {"h": "黛", "p": "dài", "sp": "dai", "r": "黑", "sc": 17, "l": 5, "f": 2612},
  {"h": "燥", "p": "zào", "sp": "zao", "r": "火", "sc": 17, "l": 5, "f": 2613},
  {"h": "掷", "p": "zhì", "sp": "zhi", "r": "手", "sc": 11, "l": 5, "f": 2614},
  {"h": "藉", "p": "jí", "sp": "ji", "r": "艸", "sc": 17, "l": 5, "f": 2615},
  {"h": "枢", "p": "shū", "sp": "shu", "r": "木", "sc": 8, "l": 5, "f": 2616},
  {"h": "憎", "p": "zēng", "sp": "zeng", "r": "心", "sc": 15, "l": 5, "f": 2617},
  {"h": "鲸", "p": "jīng", "sp": "jing", "r": "魚", "sc": 16, "l": 5, "f": 2618},
  {"h": "弘", "p": "hóng", "sp": "hong", "r": "弓", "sc": 5, "l": 5, "f": 2619},
  {"h": "倚", "p": "yǐ", "sp": "yi", "r": "人", "sc": 10, "l": 5, "f": 2620},
  {"h": "侮", "p": "wǔ", "sp": "wu", "r": "人", "sc": 9, "l": 5, "f": 2621},
  {"h": "藩", "p": "fān", "sp": "fan", "r": "艸", "sc": 18, "l": 5, "f": 2622},
  {"h": "拂", "p": "fú", "sp": "fu", "r": "手", "sc": 8, "l": 5, "f": 2623},
  {"h": "鹤", "p": "hè", "sp": "he", "r": "鳥", "sc": 15, "l": 5, "f": 2624},
  {"h": "蚀", "p": "shí", "sp": "shi", "r": "食", "sc": 9, "l": 5, "f": 2625},
  {
    "h": "浆",
    "p": "jiāng",
    "sp": "jiang",
    "r": "水",
    "sc": 10,
    "l": 5,
    "f": 2626
  },
  {"h": "芙", "p": "fú", "sp": "fu", "r": "艸", "sc": 7, "l": 5, "f": 2627},
  {"h": "垃", "p": "lā", "sp": "la", "r": "土", "sc": 8, "l": 4, "f": 2628},
  {"h": "烤", "p": "kǎo", "sp": "kao", "r": "火", "sc": 10, "l": 4, "f": 2629},
  {"h": "晒", "p": "shài", "sp": "shai", "r": "日", "sc": 10, "l": 4, "f": 2630},
  {
    "h": "霜",
    "p": "shuāng",
    "sp": "shuang",
    "r": "雨",
    "sc": 17,
    "l": 5,
    "f": 2631
  },
  {"h": "剿", "p": "jiǎo", "sp": "jiao", "r": "刀", "sc": 13, "l": 5, "f": 2632},
  {"h": "蕴", "p": "yùn", "sp": "yun", "r": "艸", "sc": 15, "l": 5, "f": 2633},
  {"h": "圾", "p": "jī", "sp": "ji", "r": "土", "sc": 6, "l": 4, "f": 2634},
  {"h": "绸", "p": "chóu", "sp": "chou", "r": "糸", "sc": 11, "l": 5, "f": 2635},
  {"h": "屿", "p": "yǔ", "sp": "yu", "r": "山", "sc": 6, "l": 5, "f": 2636},
  {"h": "氢", "p": "qīng", "sp": "qing", "r": "气", "sc": 9, "l": 5, "f": 2637},
  {"h": "驼", "p": "tuó", "sp": "tuo", "r": "馬", "sc": 8, "l": 5, "f": 2638},
  {
    "h": "妆",
    "p": "zhuāng",
    "sp": "zhuang",
    "r": "女",
    "sc": 6,
    "l": 5,
    "f": 2639
  },
  {"h": "捆", "p": "kǔn", "sp": "kun", "r": "手", "sc": 10, "l": 5, "f": 2640},
  {"h": "铅", "p": "qiān", "sp": "qian", "r": "金", "sc": 10, "l": 2, "f": 2641},
  {
    "h": "逛",
    "p": "guàng",
    "sp": "guang",
    "r": "辵",
    "sc": 10,
    "l": 4,
    "f": 2642
  },
  {"h": "淑", "p": "shū", "sp": "shu", "r": "水", "sc": 11, "l": 5, "f": 2643},
  {"h": "榴", "p": "liú", "sp": "liu", "r": "木", "sc": 14, "l": 5, "f": 2644},
  {"h": "丙", "p": "bǐng", "sp": "bing", "r": "一", "sc": 5, "l": 5, "f": 2645},
  {"h": "痒", "p": "yǎng", "sp": "yang", "r": "疒", "sc": 11, "l": 5, "f": 2646},
  {"h": "钞", "p": "chāo", "sp": "chao", "r": "金", "sc": 9, "l": 5, "f": 2647},
  {"h": "蹄", "p": "tí", "sp": "ti", "r": "足", "sc": 16, "l": 5, "f": 2648},
  {"h": "犬", "p": "quǎn", "sp": "quan", "r": "犬", "sc": 4, "l": 5, "f": 2649},
  {"h": "躬", "p": "gōng", "sp": "gong", "r": "身", "sc": 10, "l": 5, "f": 2650},
  {"h": "昼", "p": "zhòu", "sp": "zhou", "r": "尸", "sc": 9, "l": 5, "f": 2651},
  {"h": "藻", "p": "zǎo", "sp": "zao", "r": "艸", "sc": 19, "l": 5, "f": 2652},
  {"h": "蛛", "p": "zhū", "sp": "zhu", "r": "虫", "sc": 12, "l": 5, "f": 2653},
  {"h": "褐", "p": "hè", "sp": "he", "r": "衣", "sc": 14, "l": 5, "f": 2654},
  {"h": "颊", "p": "jiá", "sp": "jia", "r": "頁", "sc": 12, "l": 5, "f": 2655},
  {"h": "奠", "p": "diàn", "sp": "dian", "r": "大", "sc": 12, "l": 5, "f": 2656},
  {"h": "募", "p": "mù", "sp": "mu", "r": "力", "sc": 12, "l": 5, "f": 2657},
  {"h": "耽", "p": "dān", "sp": "dan", "r": "耳", "sc": 10, "l": 5, "f": 2658},
  {"h": "蹈", "p": "dǎo", "sp": "dao", "r": "足", "sc": 17, "l": 5, "f": 2659},
  {"h": "陋", "p": "lòu", "sp": "lou", "r": "阜", "sc": 8, "l": 5, "f": 2660},
  {"h": "侣", "p": "lǚ", "sp": "lu", "r": "人", "sc": 8, "l": 5, "f": 2661},
  {"h": "魅", "p": "mèi", "sp": "mei", "r": "鬼", "sc": 14, "l": 5, "f": 2662},
  {"h": "岚", "p": "lán", "sp": "lan", "r": "山", "sc": 7, "l": 5, "f": 2663},
  {"h": "侄", "p": "zhí", "sp": "zhi", "r": "人", "sc": 8, "l": 5, "f": 2664},
  {"h": "虐", "p": "nüè", "sp": "nue", "r": "虍", "sc": 9, "l": 5, "f": 2665},
  {"h": "堕", "p": "duò", "sp": "duo", "r": "土", "sc": 11, "l": 5, "f": 2666},
  {"h": "陛", "p": "bì", "sp": "bi", "r": "阜", "sc": 9, "l": 5, "f": 2667},
  {"h": "莹", "p": "yíng", "sp": "ying", "r": "艸", "sc": 10, "l": 5, "f": 2668},
  {"h": "荫", "p": "yīn", "sp": "yin", "r": "艸", "sc": 9, "l": 5, "f": 2669},
  {"h": "狡", "p": "jiǎo", "sp": "jiao", "r": "犬", "sc": 9, "l": 5, "f": 2670},
  {"h": "阀", "p": "fá", "sp": "fa", "r": "門", "sc": 9, "l": 5, "f": 2671},
  {"h": "绞", "p": "jiǎo", "sp": "jiao", "r": "糸", "sc": 9, "l": 5, "f": 2672},
  {"h": "膏", "p": "gāo", "sp": "gao", "r": "肉", "sc": 14, "l": 4, "f": 2673},
  {"h": "垮", "p": "kuǎ", "sp": "kua", "r": "土", "sc": 9, "l": 5, "f": 2674},
  {"h": "茎", "p": "jīng", "sp": "jing", "r": "艸", "sc": 8, "l": 5, "f": 2675},
  {"h": "缅", "p": "miǎn", "sp": "mian", "r": "糸", "sc": 12, "l": 5, "f": 2676},
  {"h": "喇", "p": "lǎ", "sp": "la", "r": "口", "sc": 12, "l": 5, "f": 2677},
  {"h": "绒", "p": "róng", "sp": "rong", "r": "糸", "sc": 9, "l": 5, "f": 2678},
  {"h": "搅", "p": "jiǎo", "sp": "jiao", "r": "手", "sc": 12, "l": 5, "f": 2679},
  {"h": "凳", "p": "dèng", "sp": "deng", "r": "几", "sc": 14, "l": 5, "f": 2680},
  {"h": "梭", "p": "suō", "sp": "suo", "r": "木", "sc": 11, "l": 5, "f": 2681},
  {"h": "丫", "p": "yā", "sp": "ya", "r": "丨", "sc": 3, "l": 5, "f": 2682},
  {"h": "姬", "p": "jī", "sp": "ji", "r": "女", "sc": 10, "l": 5, "f": 2683},
  {"h": "诏", "p": "zhào", "sp": "zhao", "r": "言", "sc": 7, "l": 5, "f": 2684},
  {"h": "钮", "p": "niǔ", "sp": "niu", "r": "金", "sc": 9, "l": 5, "f": 2685},
  {"h": "棺", "p": "guān", "sp": "guan", "r": "木", "sc": 12, "l": 5, "f": 2686},
  {"h": "耿", "p": "gěng", "sp": "geng", "r": "耳", "sc": 10, "l": 5, "f": 2687},
  {"h": "缔", "p": "dì", "sp": "di", "r": "糸", "sc": 12, "l": 5, "f": 2688},
  {"h": "懈", "p": "xiè", "sp": "xie", "r": "心", "sc": 16, "l": 5, "f": 2689},
  {"h": "嫉", "p": "jí", "sp": "ji", "r": "女", "sc": 13, "l": 5, "f": 2690},
  {"h": "灶", "p": "zào", "sp": "zao", "r": "火", "sc": 7, "l": 5, "f": 2691},
  {"h": "匀", "p": "yún", "sp": "yun", "r": "勹", "sc": 4, "l": 5, "f": 2692},
  {"h": "嗣", "p": "sì", "sp": "si", "r": "口", "sc": 13, "l": 5, "f": 2693},
  {"h": "鸽", "p": "gē", "sp": "ge", "r": "鳥", "sc": 11, "l": 5, "f": 2694},
  {"h": "澡", "p": "zǎo", "sp": "zao", "r": "水", "sc": 16, "l": 2, "f": 2695},
  {"h": "凿", "p": "záo", "sp": "zao", "r": "凵", "sc": 12, "l": 5, "f": 2696},
  {"h": "纬", "p": "wěi", "sp": "wei", "r": "糸", "sc": 7, "l": 5, "f": 2697},
  {"h": "沸", "p": "fèi", "sp": "fei", "r": "水", "sc": 8, "l": 5, "f": 2698},
  {"h": "畴", "p": "chóu", "sp": "chou", "r": "田", "sc": 12, "l": 5, "f": 2699},
  {"h": "刃", "p": "rèn", "sp": "ren", "r": "刀", "sc": 3, "l": 5, "f": 2700},
  {"h": "遏", "p": "è", "sp": "e", "r": "辵", "sc": 12, "l": 5, "f": 2701},
  {"h": "烁", "p": "shuò", "sp": "shuo", "r": "火", "sc": 9, "l": 5, "f": 2702},
  {"h": "嗅", "p": "xiù", "sp": "xiu", "r": "口", "sc": 13, "l": 5, "f": 2703},
  {"h": "叭", "p": "bā", "sp": "ba", "r": "口", "sc": 5, "l": 5, "f": 2704},
  {"h": "熬", "p": "áo", "sp": "ao", "r": "火", "sc": 14, "l": 5, "f": 2705},
  {"h": "瞥", "p": "piē", "sp": "pie", "r": "目", "sc": 16, "l": 5, "f": 2706},
  {"h": "骸", "p": "hái", "sp": "hai", "r": "骨", "sc": 15, "l": 5, "f": 2707},
  {"h": "奢", "p": "shē", "sp": "she", "r": "大", "sc": 11, "l": 5, "f": 2708},
  {"h": "拙", "p": "zhuō", "sp": "zhuo", "r": "手", "sc": 8, "l": 5, "f": 2709},
  {"h": "栋", "p": "dòng", "sp": "dong", "r": "木", "sc": 9, "l": 5, "f": 2710},
  {"h": "毯", "p": "tǎn", "sp": "tan", "r": "毛", "sc": 12, "l": 5, "f": 2711},
  {"h": "桐", "p": "tóng", "sp": "tong", "r": "木", "sc": 10, "l": 5, "f": 2712},
  {"h": "砂", "p": "shā", "sp": "sha", "r": "石", "sc": 9, "l": 5, "f": 2713},
  {"h": "莽", "p": "mǎng", "sp": "mang", "r": "艸", "sc": 10, "l": 5, "f": 2714},
  {"h": "泻", "p": "xiè", "sp": "xie", "r": "水", "sc": 8, "l": 5, "f": 2715},
  {"h": "坪", "p": "píng", "sp": "ping", "r": "土", "sc": 8, "l": 5, "f": 2716},
  {"h": "梳", "p": "shū", "sp": "shu", "r": "木", "sc": 11, "l": 5, "f": 2717},
  {"h": "杉", "p": "shān", "sp": "shan", "r": "木", "sc": 7, "l": 5, "f": 2718},
  {"h": "晤", "p": "wù", "sp": "wu", "r": "日", "sc": 11, "l": 5, "f": 2719},
  {"h": "稚", "p": "zhì", "sp": "zhi", "r": "禾", "sc": 13, "l": 5, "f": 2720},
  {"h": "蔬", "p": "shū", "sp": "shu", "r": "艸", "sc": 15, "l": 1, "f": 2721},
  {"h": "蝇", "p": "yíng", "sp": "ying", "r": "虫", "sc": 14, "l": 5, "f": 2722},
  {"h": "捣", "p": "dǎo", "sp": "dao", "r": "手", "sc": 10, "l": 5, "f": 2723},
  {"h": "顷", "p": "qǐng", "sp": "qing", "r": "頁", "sc": 8, "l": 5, "f": 2724},
  {"h": "麽", "p": "mó", "sp": "mo", "r": "麻", "sc": 14, "l": 5, "f": 2725},
  {"h": "尴", "p": "gān", "sp": "gan", "r": "尢", "sc": 13, "l": 5, "f": 2726},
  {"h": "镖", "p": "biāo", "sp": "biao", "r": "金", "sc": 16, "l": 5, "f": 2727},
  {"h": "诧", "p": "chà", "sp": "cha", "r": "言", "sc": 8, "l": 5, "f": 2728},
  {"h": "尬", "p": "gà", "sp": "ga", "r": "尢", "sc": 7, "l": 5, "f": 2729},
  {"h": "硫", "p": "liú", "sp": "liu", "r": "石", "sc": 12, "l": 5, "f": 2730},
  {"h": "嚼", "p": "jué", "sp": "jue", "r": "口", "sc": 20, "l": 5, "f": 2731},
  {"h": "羡", "p": "xiàn", "sp": "xian", "r": "羊", "sc": 12, "l": 5, "f": 2732},
  {"h": "沦", "p": "lún", "sp": "lun", "r": "水", "sc": 7, "l": 5, "f": 2733},
  {"h": "沪", "p": "hù", "sp": "hu", "r": "水", "sc": 7, "l": 5, "f": 2734},
  {"h": "旷", "p": "kuàng", "sp": "kuang", "r": "日", "sc": 7, "l": 5, "f": 2735},
  {"h": "彬", "p": "bīn", "sp": "bin", "r": "彡", "sc": 11, "l": 5, "f": 2736},
  {"h": "芽", "p": "yá", "sp": "ya", "r": "艸", "sc": 7, "l": 5, "f": 2737},
  {"h": "狸", "p": "lí", "sp": "li", "r": "犬", "sc": 10, "l": 5, "f": 2738},
  {"h": "冥", "p": "míng", "sp": "ming", "r": "冖", "sc": 10, "l": 5, "f": 2739},
  {"h": "碳", "p": "tàn", "sp": "tan", "r": "石", "sc": 14, "l": 5, "f": 2740},
  {"h": "咧", "p": "liě", "sp": "lie", "r": "口", "sc": 9, "l": 5, "f": 2741},
  {"h": "惕", "p": "tì", "sp": "ti", "r": "心", "sc": 11, "l": 5, "f": 2742},
  {"h": "暑", "p": "shǔ", "sp": "shu", "r": "日", "sc": 12, "l": 4, "f": 2743},
  {"h": "咯", "p": "gē", "sp": "ge", "r": "口", "sc": 9, "l": 5, "f": 2744},
  {"h": "萝", "p": "luó", "sp": "luo", "r": "艸", "sc": 11, "l": 5, "f": 2745},
  {"h": "汹", "p": "xiōng", "sp": "xiong", "r": "水", "sc": 7, "l": 5, "f": 2746},
  {"h": "腥", "p": "xīng", "sp": "xing", "r": "肉", "sc": 13, "l": 5, "f": 2747},
  {"h": "窥", "p": "kuī", "sp": "kui", "r": "穴", "sc": 13, "l": 5, "f": 2748},
  {"h": "俺", "p": "ǎn", "sp": "an", "r": "人", "sc": 10, "l": 5, "f": 2749},
  {"h": "潭", "p": "tán", "sp": "tan", "r": "水", "sc": 15, "l": 5, "f": 2750},
  {"h": "崎", "p": "qí", "sp": "qi", "r": "山", "sc": 11, "l": 5, "f": 2751},
  {"h": "麟", "p": "lín", "sp": "lin", "r": "鹿", "sc": 23, "l": 5, "f": 2752},
  {"h": "捡", "p": "jiǎn", "sp": "jian", "r": "手", "sc": 10, "l": 5, "f": 2753},
  {"h": "拯", "p": "zhěng", "sp": "zheng", "r": "手", "sc": 9, "l": 5, "f": 2754},
  {"h": "厥", "p": "jué", "sp": "jue", "r": "厂", "sc": 12, "l": 5, "f": 2755},
  {
    "h": "澄",
    "p": "chéng",
    "sp": "cheng",
    "r": "水",
    "sc": 15,
    "l": 5,
    "f": 2756
  },
  {"h": "萎", "p": "wēi", "sp": "wei", "r": "艸", "sc": 11, "l": 5, "f": 2757},
  {"h": "哉", "p": "zāi", "sp": "zai", "r": "口", "sc": 9, "l": 5, "f": 2758},
  {"h": "涡", "p": "wō", "sp": "wo", "r": "水", "sc": 10, "l": 5, "f": 2759},
  {"h": "滔", "p": "tāo", "sp": "tao", "r": "水", "sc": 13, "l": 5, "f": 2760},
  {"h": "暇", "p": "xiá", "sp": "xia", "r": "日", "sc": 13, "l": 5, "f": 2761},
  {"h": "溯", "p": "sù", "sp": "su", "r": "水", "sc": 13, "l": 5, "f": 2762},
  {"h": "鳞", "p": "lín", "sp": "lin", "r": "魚", "sc": 20, "l": 5, "f": 2763},
  {
    "h": "酿",
    "p": "niàng",
    "sp": "niang",
    "r": "酉",
    "sc": 14,
    "l": 5,
    "f": 2764
  },
  {"h": "茵", "p": "yīn", "sp": "yin", "r": "艸", "sc": 9, "l": 5, "f": 2765},
  {"h": "愕", "p": "è", "sp": "e", "r": "心", "sc": 12, "l": 5, "f": 2766},
  {"h": "瞅", "p": "chǒu", "sp": "chou", "r": "目", "sc": 14, "l": 5, "f": 2767},
  {"h": "暮", "p": "mù", "sp": "mu", "r": "日", "sc": 14, "l": 5, "f": 2768},
  {"h": "衙", "p": "yá", "sp": "ya", "r": "行", "sc": 13, "l": 5, "f": 2769},
  {"h": "诫", "p": "jiè", "sp": "jie", "r": "言", "sc": 9, "l": 5, "f": 2770},
  {"h": "斧", "p": "fǔ", "sp": "fu", "r": "斤", "sc": 8, "l": 5, "f": 2771},
  {"h": "兮", "p": "xī", "sp": "xi", "r": "八", "sc": 4, "l": 5, "f": 2772},
  {"h": "焕", "p": "huàn", "sp": "huan", "r": "火", "sc": 11, "l": 5, "f": 2773},
  {"h": "棕", "p": "zōng", "sp": "zong", "r": "木", "sc": 12, "l": 5, "f": 2774},
  {"h": "佑", "p": "yòu", "sp": "you", "r": "人", "sc": 7, "l": 5, "f": 2775},
  {"h": "嘶", "p": "sī", "sp": "si", "r": "口", "sc": 15, "l": 5, "f": 2776},
  {"h": "妓", "p": "jì", "sp": "ji", "r": "女", "sc": 7, "l": 5, "f": 2777},
  {"h": "喧", "p": "xuān", "sp": "xuan", "r": "口", "sc": 12, "l": 5, "f": 2778},
  {"h": "蓉", "p": "róng", "sp": "rong", "r": "艸", "sc": 13, "l": 5, "f": 2779},
  {"h": "删", "p": "shān", "sp": "shan", "r": "刀", "sc": 7, "l": 5, "f": 2780},
  {"h": "樱", "p": "yīng", "sp": "ying", "r": "木", "sc": 15, "l": 5, "f": 2781},
  {"h": "伺", "p": "cì", "sp": "ci", "r": "人", "sc": 7, "l": 5, "f": 2782},
  {"h": "嗡", "p": "wēng", "sp": "weng", "r": "口", "sc": 13, "l": 5, "f": 2783},
  {"h": "娥", "p": "é", "sp": "e", "r": "女", "sc": 10, "l": 5, "f": 2784},
  {"h": "梢", "p": "shāo", "sp": "shao", "r": "木", "sc": 11, "l": 5, "f": 2785},
  {"h": "坝", "p": "bà", "sp": "ba", "r": "土", "sc": 7, "l": 5, "f": 2786},
  {"h": "蚕", "p": "cán", "sp": "can", "r": "虫", "sc": 10, "l": 5, "f": 2787},
  {"h": "敷", "p": "fū", "sp": "fu", "r": "攴", "sc": 15, "l": 5, "f": 2788},
  {"h": "澜", "p": "lán", "sp": "lan", "r": "水", "sc": 15, "l": 5, "f": 2789},
  {"h": "杏", "p": "xìng", "sp": "xing", "r": "木", "sc": 7, "l": 5, "f": 2790},
  {"h": "绥", "p": "suí", "sp": "sui", "r": "糸", "sc": 10, "l": 5, "f": 2791},
  {"h": "冶", "p": "yě", "sp": "ye", "r": "冫", "sc": 7, "l": 5, "f": 2792},
  {"h": "庇", "p": "bì", "sp": "bi", "r": "广", "sc": 7, "l": 5, "f": 2793},
  {"h": "挠", "p": "náo", "sp": "nao", "r": "手", "sc": 9, "l": 5, "f": 2794},
  {"h": "搂", "p": "lǒu", "sp": "lou", "r": "手", "sc": 12, "l": 5, "f": 2795},
  {"h": "倏", "p": "shū", "sp": "shu", "r": "人", "sc": 10, "l": 5, "f": 2796},
  {"h": "聂", "p": "niè", "sp": "nie", "r": "耳", "sc": 10, "l": 5, "f": 2797},
  {"h": "婉", "p": "wǎn", "sp": "wan", "r": "女", "sc": 11, "l": 5, "f": 2798},
  {"h": "噪", "p": "zào", "sp": "zao", "r": "口", "sc": 16, "l": 5, "f": 2799},
  {"h": "稼", "p": "jià", "sp": "jia", "r": "禾", "sc": 15, "l": 5, "f": 2800},
  {"h": "鳍", "p": "qí", "sp": "qi", "r": "魚", "sc": 18, "l": 5, "f": 2801},
  {"h": "菱", "p": "líng", "sp": "ling", "r": "艸", "sc": 11, "l": 5, "f": 2802},
  {"h": "盏", "p": "zhǎn", "sp": "zhan", "r": "皿", "sc": 10, "l": 5, "f": 2803},
  {"h": "匿", "p": "nì", "sp": "ni", "r": "匸", "sc": 10, "l": 5, "f": 2804},
  {"h": "吱", "p": "zhī", "sp": "zhi", "r": "口", "sc": 7, "l": 5, "f": 2805},
  {"h": "寝", "p": "qǐn", "sp": "qin", "r": "宀", "sc": 13, "l": 5, "f": 2806},
  {"h": "揽", "p": "lǎn", "sp": "lan", "r": "手", "sc": 12, "l": 5, "f": 2807},
  {"h": "髓", "p": "suǐ", "sp": "sui", "r": "骨", "sc": 21, "l": 5, "f": 2808},
  {"h": "秉", "p": "bǐng", "sp": "bing", "r": "禾", "sc": 8, "l": 5, "f": 2809},
  {"h": "哺", "p": "bǔ", "sp": "bu", "r": "口", "sc": 10, "l": 5, "f": 2810},
  {"h": "矢", "p": "shǐ", "sp": "shi", "r": "矢", "sc": 5, "l": 5, "f": 2811},
  {"h": "啪", "p": "pā", "sp": "pa", "r": "口", "sc": 11, "l": 5, "f": 2812},
  {"h": "帜", "p": "zhì", "sp": "zhi", "r": "巾", "sc": 8, "l": 5, "f": 2813},
  {"h": "邵", "p": "shào", "sp": "shao", "r": "邑", "sc": 7, "l": 5, "f": 2814},
  {"h": "嗽", "p": "sòu", "sp": "sou", "r": "口", "sc": 14, "l": 4, "f": 2815},
  {"h": "挟", "p": "xié", "sp": "xie", "r": "手", "sc": 9, "l": 5, "f": 2816},
  {"h": "缸", "p": "gāng", "sp": "gang", "r": "缶", "sc": 9, "l": 5, "f": 2817},
  {"h": "揉", "p": "róu", "sp": "rou", "r": "手", "sc": 12, "l": 5, "f": 2818},
  {"h": "腻", "p": "nì", "sp": "ni", "r": "肉", "sc": 13, "l": 5, "f": 2819},
  {"h": "驯", "p": "xún", "sp": "xun", "r": "馬", "sc": 6, "l": 5, "f": 2820},
  {"h": "缆", "p": "lǎn", "sp": "lan", "r": "糸", "sc": 12, "l": 5, "f": 2821},
  {
    "h": "晌",
    "p": "shǎng",
    "sp": "shang",
    "r": "日",
    "sc": 10,
    "l": 5,
    "f": 2822
  },
  {"h": "瘫", "p": "tān", "sp": "tan", "r": "疒", "sc": 15, "l": 5, "f": 2823},
  {"h": "贮", "p": "zhù", "sp": "zhu", "r": "貝", "sc": 8, "l": 5, "f": 2824},
  {"h": "觅", "p": "mì", "sp": "mi", "r": "爪", "sc": 8, "l": 5, "f": 2825},
  {"h": "朦", "p": "méng", "sp": "meng", "r": "月", "sc": 17, "l": 5, "f": 2826},
  {"h": "僻", "p": "pì", "sp": "pi", "r": "人", "sc": 15, "l": 5, "f": 2827},
  {"h": "隋", "p": "suí", "sp": "sui", "r": "阜", "sc": 11, "l": 5, "f": 2828},
  {"h": "蔓", "p": "màn", "sp": "man", "r": "艸", "sc": 14, "l": 5, "f": 2829},
  {"h": "咋", "p": "zǎ", "sp": "za", "r": "口", "sc": 8, "l": 5, "f": 2830},
  {"h": "嵌", "p": "qiàn", "sp": "qian", "r": "山", "sc": 12, "l": 5, "f": 2831},
  {"h": "虔", "p": "qián", "sp": "qian", "r": "虍", "sc": 10, "l": 5, "f": 2832},
  {"h": "畔", "p": "pàn", "sp": "pan", "r": "田", "sc": 10, "l": 5, "f": 2833},
  {"h": "琐", "p": "suǒ", "sp": "suo", "r": "玉", "sc": 11, "l": 5, "f": 2834},
  {"h": "碟", "p": "dié", "sp": "die", "r": "石", "sc": 14, "l": 5, "f": 2835},
  {"h": "涩", "p": "sè", "sp": "se", "r": "水", "sc": 10, "l": 5, "f": 2836},
  {"h": "胧", "p": "lóng", "sp": "long", "r": "肉", "sc": 9, "l": 5, "f": 2837},
  {"h": "嘟", "p": "dū", "sp": "du", "r": "口", "sc": 13, "l": 5, "f": 2838},
  {"h": "蹦", "p": "bèng", "sp": "beng", "r": "足", "sc": 18, "l": 5, "f": 2839},
  {
    "h": "冢",
    "p": "zhǒng",
    "sp": "zhong",
    "r": "冖",
    "sc": 10,
    "l": 5,
    "f": 2840
  },
  {"h": "浏", "p": "liú", "sp": "liu", "r": "水", "sc": 9, "l": 5, "f": 2841},
  {"h": "裔", "p": "yì", "sp": "yi", "r": "衣", "sc": 13, "l": 5, "f": 2842},
  {"h": "襟", "p": "jīn", "sp": "jin", "r": "衣", "sc": 18, "l": 5, "f": 2843},
  {"h": "叨", "p": "dāo", "sp": "dao", "r": "口", "sc": 5, "l": 5, "f": 2844},
  {"h": "诀", "p": "jué", "sp": "jue", "r": "言", "sc": 6, "l": 5, "f": 2845},
  {"h": "旭", "p": "xù", "sp": "xu", "r": "日", "sc": 6, "l": 5, "f": 2846},
  {"h": "虾", "p": "xiā", "sp": "xia", "r": "虫", "sc": 9, "l": 5, "f": 2847},
  {"h": "簿", "p": "bù", "sp": "bu", "r": "竹", "sc": 19, "l": 5, "f": 2848},
  {"h": "啤", "p": "pí", "sp": "pi", "r": "口", "sc": 11, "l": 3, "f": 2849},
  {"h": "擒", "p": "qín", "sp": "qin", "r": "手", "sc": 15, "l": 5, "f": 2850},
  {"h": "枣", "p": "zǎo", "sp": "zao", "r": "木", "sc": 8, "l": 5, "f": 2851},
  {"h": "嘎", "p": "gā", "sp": "ga", "r": "口", "sc": 14, "l": 5, "f": 2852},
  {"h": "苑", "p": "yuàn", "sp": "yuan", "r": "艸", "sc": 8, "l": 5, "f": 2853},
  {"h": "牟", "p": "móu", "sp": "mou", "r": "牛", "sc": 6, "l": 5, "f": 2854},
  {"h": "呕", "p": "ǒu", "sp": "ou", "r": "口", "sc": 7, "l": 5, "f": 2855},
  {"h": "骆", "p": "luò", "sp": "luo", "r": "馬", "sc": 9, "l": 5, "f": 2856},
  {"h": "凸", "p": "tū", "sp": "tu", "r": "凵", "sc": 5, "l": 5, "f": 2857},
  {"h": "熄", "p": "xī", "sp": "xi", "r": "火", "sc": 14, "l": 5, "f": 2858},
  {"h": "兀", "p": "wù", "sp": "wu", "r": "儿", "sc": 3, "l": 5, "f": 2859},
  {"h": "喔", "p": "ō", "sp": "o", "r": "口", "sc": 12, "l": 5, "f": 2860},
  {
    "h": "裳",
    "p": "shang",
    "sp": "shang",
    "r": "衣",
    "sc": 14,
    "l": 5,
    "f": 2861
  },
  {"h": "凹", "p": "āo", "sp": "ao", "r": "凵", "sc": 5, "l": 5, "f": 2862},
  {"h": "赎", "p": "shú", "sp": "shu", "r": "貝", "sc": 12, "l": 5, "f": 2863},
  {"h": "屯", "p": "tún", "sp": "tun", "r": "屮", "sc": 4, "l": 5, "f": 2864},
  {"h": "膛", "p": "táng", "sp": "tang", "r": "肉", "sc": 15, "l": 5, "f": 2865},
  {"h": "浇", "p": "jiāo", "sp": "jiao", "r": "水", "sc": 9, "l": 5, "f": 2866},
  {"h": "灼", "p": "zhuó", "sp": "zhuo", "r": "火", "sc": 7, "l": 5, "f": 2867},
  {"h": "裘", "p": "qiú", "sp": "qiu", "r": "衣", "sc": 13, "l": 5, "f": 2868},
  {"h": "砰", "p": "pēng", "sp": "peng", "r": "石", "sc": 10, "l": 5, "f": 2869},
  {"h": "棘", "p": "jí", "sp": "ji", "r": "木", "sc": 12, "l": 5, "f": 2870},
  {
    "h": "橡",
    "p": "xiàng",
    "sp": "xiang",
    "r": "木",
    "sc": 15,
    "l": 5,
    "f": 2871
  },
  {"h": "碱", "p": "jiǎn", "sp": "jian", "r": "石", "sc": 14, "l": 5, "f": 2872},
  {"h": "聋", "p": "lóng", "sp": "long", "r": "耳", "sc": 11, "l": 5, "f": 2873},
  {"h": "姥", "p": "lǎo", "sp": "lao", "r": "女", "sc": 9, "l": 5, "f": 2874},
  {"h": "瑜", "p": "yú", "sp": "yu", "r": "玉", "sc": 13, "l": 5, "f": 2875},
  {"h": "毋", "p": "wú", "sp": "wu", "r": "毋", "sc": 4, "l": 5, "f": 2876},
  {"h": "娅", "p": "yà", "sp": "ya", "r": "女", "sc": 9, "l": 5, "f": 2877},
  {"h": "沮", "p": "jǔ", "sp": "ju", "r": "水", "sc": 8, "l": 5, "f": 2878},
  {"h": "萌", "p": "méng", "sp": "meng", "r": "艸", "sc": 11, "l": 5, "f": 2879},
  {"h": "俏", "p": "qiào", "sp": "qiao", "r": "人", "sc": 9, "l": 5, "f": 2880},
  {"h": "黯", "p": "àn", "sp": "an", "r": "音", "sc": 21, "l": 5, "f": 2881},
  {"h": "撇", "p": "piē", "sp": "pie", "r": "手", "sc": 14, "l": 5, "f": 2882},
  {"h": "粟", "p": "sù", "sp": "su", "r": "米", "sc": 12, "l": 5, "f": 2883},
  {"h": "粪", "p": "fèn", "sp": "fen", "r": "米", "sc": 12, "l": 5, "f": 2884},
  {"h": "尹", "p": "yǐn", "sp": "yin", "r": "尸", "sc": 4, "l": 5, "f": 2885},
  {"h": "苟", "p": "gǒu", "sp": "gou", "r": "艸", "sc": 8, "l": 5, "f": 2886},
  {"h": "癫", "p": "diān", "sp": "dian", "r": "疒", "sc": 21, "l": 5, "f": 2887},
  {"h": "蚂", "p": "mǎ", "sp": "ma", "r": "虫", "sc": 9, "l": 5, "f": 2888},
  {"h": "禹", "p": "yǔ", "sp": "yu", "r": "禸", "sc": 9, "l": 5, "f": 2889},
  {"h": "廖", "p": "liào", "sp": "liao", "r": "广", "sc": 14, "l": 5, "f": 2890},
  {"h": "俭", "p": "jiǎn", "sp": "jian", "r": "人", "sc": 9, "l": 5, "f": 2891},
  {"h": "帖", "p": "tiē", "sp": "tie", "r": "巾", "sc": 8, "l": 5, "f": 2892},
  {"h": "煎", "p": "jiān", "sp": "jian", "r": "火", "sc": 13, "l": 5, "f": 2893},
  {"h": "缕", "p": "lǚ", "sp": "lu", "r": "糸", "sc": 12, "l": 5, "f": 2894},
  {"h": "窦", "p": "dòu", "sp": "dou", "r": "穴", "sc": 13, "l": 5, "f": 2895},
  {"h": "簇", "p": "cù", "sp": "cu", "r": "竹", "sc": 17, "l": 5, "f": 2896},
  {"h": "棱", "p": "léng", "sp": "leng", "r": "木", "sc": 12, "l": 5, "f": 2897},
  {"h": "叩", "p": "kòu", "sp": "kou", "r": "口", "sc": 5, "l": 5, "f": 2898},
  {"h": "呐", "p": "nà", "sp": "na", "r": "口", "sc": 7, "l": 5, "f": 2899},
  {"h": "瑶", "p": "yáo", "sp": "yao", "r": "玉", "sc": 14, "l": 5, "f": 2900},
  {"h": "墅", "p": "shù", "sp": "shu", "r": "土", "sc": 14, "l": 5, "f": 2901},
  {"h": "莺", "p": "yīng", "sp": "ying", "r": "艸", "sc": 10, "l": 5, "f": 2902},
  {"h": "烫", "p": "tàng", "sp": "tang", "r": "火", "sc": 10, "l": 5, "f": 2903},
  {"h": "蛙", "p": "wā", "sp": "wa", "r": "虫", "sc": 12, "l": 5, "f": 2904},
  {"h": "歹", "p": "dǎi", "sp": "dai", "r": "歹", "sc": 4, "l": 5, "f": 2905},
  {"h": "伶", "p": "líng", "sp": "ling", "r": "人", "sc": 7, "l": 5, "f": 2906},
  {"h": "葱", "p": "cōng", "sp": "cong", "r": "艸", "sc": 12, "l": 5, "f": 2907},
  {"h": "哮", "p": "xiāo", "sp": "xiao", "r": "口", "sc": 10, "l": 5, "f": 2908},
  {"h": "眩", "p": "xuàn", "sp": "xuan", "r": "目", "sc": 10, "l": 5, "f": 2909},
  {"h": "坤", "p": "kūn", "sp": "kun", "r": "土", "sc": 8, "l": 5, "f": 2910},
  {"h": "廓", "p": "kuò", "sp": "kuo", "r": "广", "sc": 13, "l": 5, "f": 2911},
  {"h": "讳", "p": "huì", "sp": "hui", "r": "言", "sc": 6, "l": 5, "f": 2912},
  {"h": "啼", "p": "tí", "sp": "ti", "r": "口", "sc": 12, "l": 5, "f": 2913},
  {"h": "乍", "p": "zhà", "sp": "zha", "r": "丿", "sc": 5, "l": 5, "f": 2914},
  {"h": "瓣", "p": "bàn", "sp": "ban", "r": "瓜", "sc": 19, "l": 5, "f": 2915},
  {"h": "矫", "p": "jiǎo", "sp": "jiao", "r": "矢", "sc": 11, "l": 5, "f": 2916},
  {"h": "跋", "p": "bá", "sp": "ba", "r": "足", "sc": 12, "l": 5, "f": 2917},
  {"h": "枉", "p": "wǎng", "sp": "wang", "r": "木", "sc": 8, "l": 5, "f": 2918},
  {"h": "梗", "p": "gěng", "sp": "geng", "r": "木", "sc": 11, "l": 5, "f": 2919},
  {"h": "厕", "p": "cè", "sp": "ce", "r": "厂", "sc": 8, "l": 4, "f": 2920},
  {"h": "琢", "p": "zuó", "sp": "zuo", "r": "玉", "sc": 12, "l": 5, "f": 2921},
  {"h": "讥", "p": "jī", "sp": "ji", "r": "言", "sc": 4, "l": 5, "f": 2922},
  {"h": "釉", "p": "yòu", "sp": "you", "r": "釆", "sc": 12, "l": 5, "f": 2923},
  {"h": "窟", "p": "kū", "sp": "ku", "r": "穴", "sc": 13, "l": 5, "f": 2924},
  {"h": "敛", "p": "liǎn", "sp": "lian", "r": "攴", "sc": 11, "l": 5, "f": 2925},
  {"h": "轼", "p": "shì", "sp": "shi", "r": "車", "sc": 10, "l": 5, "f": 2926},
  {"h": "庐", "p": "lú", "sp": "lu", "r": "广", "sc": 7, "l": 5, "f": 2927},
  {"h": "胚", "p": "pēi", "sp": "pei", "r": "肉", "sc": 9, "l": 5, "f": 2928},
  {"h": "呻", "p": "shēn", "sp": "shen", "r": "口", "sc": 8, "l": 5, "f": 2929},
  {"h": "绰", "p": "chuò", "sp": "chuo", "r": "糸", "sc": 11, "l": 5, "f": 2930},
  {"h": "扼", "p": "è", "sp": "e", "r": "手", "sc": 7, "l": 5, "f": 2931},
  {"h": "懿", "p": "yì", "sp": "yi", "r": "心", "sc": 22, "l": 5, "f": 2932},
  {"h": "炯", "p": "jiǒng", "sp": "jiong", "r": "火", "sc": 9, "l": 5, "f": 2933},
  {"h": "竿", "p": "gān", "sp": "gan", "r": "竹", "sc": 9, "l": 5, "f": 2934},
  {"h": "慷", "p": "kāng", "sp": "kang", "r": "心", "sc": 14, "l": 5, "f": 2935},
  {"h": "虞", "p": "yú", "sp": "yu", "r": "虍", "sc": 13, "l": 5, "f": 2936},
  {"h": "锤", "p": "chuí", "sp": "chui", "r": "金", "sc": 13, "l": 5, "f": 2937},
  {
    "h": "栓",
    "p": "shuān",
    "sp": "shuan",
    "r": "木",
    "sc": 10,
    "l": 5,
    "f": 2938
  },
  {
    "h": "桨",
    "p": "jiǎng",
    "sp": "jiang",
    "r": "木",
    "sc": 10,
    "l": 5,
    "f": 2939
  },
  {"h": "蚊", "p": "wén", "sp": "wen", "r": "虫", "sc": 10, "l": 5, "f": 2940},
  {"h": "磅", "p": "bàng", "sp": "bang", "r": "石", "sc": 15, "l": 5, "f": 2941},
  {"h": "孽", "p": "niè", "sp": "nie", "r": "子", "sc": 19, "l": 5, "f": 2942},
  {"h": "惭", "p": "cán", "sp": "can", "r": "心", "sc": 11, "l": 5, "f": 2943},
  {"h": "戳", "p": "chuō", "sp": "chuo", "r": "戈", "sc": 18, "l": 5, "f": 2944},
  {"h": "禀", "p": "bǐng", "sp": "bing", "r": "示", "sc": 13, "l": 5, "f": 2945},
  {"h": "鄂", "p": "è", "sp": "e", "r": "邑", "sc": 11, "l": 5, "f": 2946},
  {"h": "馈", "p": "kuì", "sp": "kui", "r": "食", "sc": 12, "l": 5, "f": 2947},
  {"h": "垣", "p": "yuán", "sp": "yuan", "r": "土", "sc": 9, "l": 5, "f": 2948},
  {"h": "溅", "p": "jiàn", "sp": "jian", "r": "水", "sc": 12, "l": 5, "f": 2949},
  {"h": "咚", "p": "dōng", "sp": "dong", "r": "口", "sc": 8, "l": 5, "f": 2950},
  {"h": "钙", "p": "gài", "sp": "gai", "r": "金", "sc": 9, "l": 5, "f": 2951},
  {"h": "礁", "p": "jiāo", "sp": "jiao", "r": "石", "sc": 17, "l": 5, "f": 2952},
  {
    "h": "彰",
    "p": "zhāng",
    "sp": "zhang",
    "r": "彡",
    "sc": 14,
    "l": 5,
    "f": 2953
  },
  {"h": "豁", "p": "huō", "sp": "huo", "r": "谷", "sc": 17, "l": 5, "f": 2954},
  {"h": "眯", "p": "mī", "sp": "mi", "r": "目", "sc": 11, "l": 5, "f": 2955},
  {"h": "磷", "p": "lín", "sp": "lin", "r": "石", "sc": 17, "l": 5, "f": 2956},
  {"h": "雯", "p": "wén", "sp": "wen", "r": "雨", "sc": 12, "l": 5, "f": 2957},
  {"h": "墟", "p": "xū", "sp": "xu", "r": "土", "sc": 14, "l": 5, "f": 2958},
  {"h": "迂", "p": "yū", "sp": "yu", "r": "辵", "sc": 6, "l": 5, "f": 2959},
  {"h": "瞻", "p": "zhān", "sp": "zhan", "r": "目", "sc": 18, "l": 5, "f": 2960},
  {"h": "颅", "p": "lú", "sp": "lu", "r": "頁", "sc": 11, "l": 5, "f": 2961},
  {"h": "琉", "p": "liú", "sp": "liu", "r": "玉", "sc": 11, "l": 5, "f": 2962},
  {"h": "悼", "p": "dào", "sp": "dao", "r": "心", "sc": 11, "l": 5, "f": 2963},
  {"h": "蝴", "p": "hú", "sp": "hu", "r": "虫", "sc": 15, "l": 5, "f": 2964},
  {"h": "拣", "p": "jiǎn", "sp": "jian", "r": "手", "sc": 8, "l": 5, "f": 2965},
  {"h": "渺", "p": "miǎo", "sp": "miao", "r": "水", "sc": 12, "l": 5, "f": 2966},
  {"h": "眷", "p": "juàn", "sp": "juan", "r": "目", "sc": 11, "l": 5, "f": 2967},
  {"h": "悯", "p": "mǐn", "sp": "min", "r": "心", "sc": 10, "l": 5, "f": 2968},
  {"h": "汰", "p": "tài", "sp": "tai", "r": "水", "sc": 7, "l": 5, "f": 2969},
  {"h": "慑", "p": "shè", "sp": "she", "r": "心", "sc": 13, "l": 5, "f": 2970},
  {"h": "婶", "p": "shěn", "sp": "shen", "r": "女", "sc": 11, "l": 5, "f": 2971},
  {"h": "斐", "p": "fěi", "sp": "fei", "r": "文", "sc": 12, "l": 5, "f": 2972},
  {"h": "嘘", "p": "xū", "sp": "xu", "r": "口", "sc": 14, "l": 5, "f": 2973},
  {
    "h": "镶",
    "p": "xiāng",
    "sp": "xiang",
    "r": "金",
    "sc": 22,
    "l": 5,
    "f": 2974
  },
  {"h": "炕", "p": "kàng", "sp": "kang", "r": "火", "sc": 8, "l": 5, "f": 2975},
  {"h": "宦", "p": "huàn", "sp": "huan", "r": "宀", "sc": 9, "l": 5, "f": 2976},
  {"h": "趴", "p": "pā", "sp": "pa", "r": "足", "sc": 9, "l": 5, "f": 2977},
  {"h": "绷", "p": "běng", "sp": "beng", "r": "糸", "sc": 11, "l": 5, "f": 2978},
  {
    "h": "窘",
    "p": "jiǒng",
    "sp": "jiong",
    "r": "穴",
    "sc": 12,
    "l": 5,
    "f": 2979
  },
  {
    "h": "襄",
    "p": "xiāng",
    "sp": "xiang",
    "r": "衣",
    "sc": 17,
    "l": 5,
    "f": 2980
  },
  {"h": "珀", "p": "pò", "sp": "po", "r": "玉", "sc": 9, "l": 5, "f": 2981},
  {"h": "嚣", "p": "xiāo", "sp": "xiao", "r": "口", "sc": 18, "l": 5, "f": 2982},
  {"h": "拚", "p": "pàn", "sp": "pan", "r": "手", "sc": 8, "l": 5, "f": 2983},
  {"h": "酌", "p": "zhuó", "sp": "zhuo", "r": "酉", "sc": 10, "l": 5, "f": 2984},
  {"h": "浊", "p": "zhuó", "sp": "zhuo", "r": "水", "sc": 9, "l": 5, "f": 2985},
  {"h": "毓", "p": "yù", "sp": "yu", "r": "毋", "sc": 14, "l": 5, "f": 2986},
  {"h": "撼", "p": "hàn", "sp": "han", "r": "手", "sc": 16, "l": 5, "f": 2987},
  {"h": "嗜", "p": "shì", "sp": "shi", "r": "口", "sc": 13, "l": 5, "f": 2988},
  {"h": "扛", "p": "káng", "sp": "kang", "r": "手", "sc": 6, "l": 5, "f": 2989},
  {"h": "峭", "p": "qiào", "sp": "qiao", "r": "山", "sc": 10, "l": 5, "f": 2990},
  {"h": "磕", "p": "kē", "sp": "ke", "r": "石", "sc": 15, "l": 5, "f": 2991},
  {"h": "翘", "p": "qiào", "sp": "qiao", "r": "羽", "sc": 12, "l": 5, "f": 2992},
  {"h": "槽", "p": "cáo", "sp": "cao", "r": "木", "sc": 15, "l": 5, "f": 2993},
  {"h": "淌", "p": "tǎng", "sp": "tang", "r": "水", "sc": 11, "l": 5, "f": 2994},
  {"h": "栅", "p": "zhà", "sp": "zha", "r": "木", "sc": 9, "l": 5, "f": 2995},
  {"h": "颓", "p": "tuí", "sp": "tui", "r": "頁", "sc": 13, "l": 5, "f": 2996},
  {"h": "熏", "p": "xūn", "sp": "xun", "r": "火", "sc": 14, "l": 5, "f": 2997},
  {"h": "瑛", "p": "yīng", "sp": "ying", "r": "玉", "sc": 12, "l": 5, "f": 2998},
  {"h": "颐", "p": "yí", "sp": "yi", "r": "頁", "sc": 13, "l": 5, "f": 2999},
  {"h": "忖", "p": "cǔn", "sp": "cun", "r": "心", "sc": 6, "l": 5, "f": 3000},
  {"h": "牡", "p": "mǔ", "sp": "mu", "r": "牛", "sc": 7, "l": 5, "f": 3001},
  {"h": "缀", "p": "zhuì", "sp": "zhui", "r": "糸", "sc": 11, "l": 5, "f": 3002},
  {"h": "徊", "p": "huái", "sp": "huai", "r": "彳", "sc": 9, "l": 5, "f": 3003},
  {"h": "梨", "p": "lí", "sp": "li", "r": "木", "sc": 11, "l": 5, "f": 3004},
  {"h": "肪", "p": "fáng", "sp": "fang", "r": "肉", "sc": 8, "l": 5, "f": 3005},
  {"h": "涕", "p": "tì", "sp": "ti", "r": "水", "sc": 10, "l": 5, "f": 3006},
  {"h": "惫", "p": "bèi", "sp": "bei", "r": "夂", "sc": 12, "l": 5, "f": 3007},
  {"h": "摹", "p": "mó", "sp": "mo", "r": "手", "sc": 14, "l": 5, "f": 3008},
  {"h": "踱", "p": "duó", "sp": "duo", "r": "足", "sc": 16, "l": 5, "f": 3009},
  {"h": "肘", "p": "zhǒu", "sp": "zhou", "r": "肉", "sc": 7, "l": 5, "f": 3010},
  {"h": "熔", "p": "róng", "sp": "rong", "r": "火", "sc": 14, "l": 5, "f": 3011},
  {"h": "挚", "p": "zhì", "sp": "zhi", "r": "手", "sc": 10, "l": 5, "f": 3012},
  {"h": "氯", "p": "lǜ", "sp": "lu", "r": "气", "sc": 12, "l": 5, "f": 3013},
  {"h": "凛", "p": "lǐn", "sp": "lin", "r": "冫", "sc": 15, "l": 5, "f": 3014},
  {"h": "绎", "p": "yì", "sp": "yi", "r": "糸", "sc": 8, "l": 5, "f": 3015},
  {"h": "庶", "p": "shù", "sp": "shu", "r": "广", "sc": 11, "l": 5, "f": 3016},
  {"h": "脯", "p": "pú", "sp": "pu", "r": "肉", "sc": 11, "l": 5, "f": 3017},
  {"h": "迭", "p": "dié", "sp": "die", "r": "辵", "sc": 8, "l": 5, "f": 3018},
  {"h": "睦", "p": "mù", "sp": "mu", "r": "目", "sc": 13, "l": 5, "f": 3019},
  {"h": "窍", "p": "qiào", "sp": "qiao", "r": "穴", "sc": 10, "l": 5, "f": 3020},
  {"h": "粥", "p": "zhōu", "sp": "zhou", "r": "米", "sc": 12, "l": 5, "f": 3021},
  {"h": "庵", "p": "ān", "sp": "an", "r": "广", "sc": 11, "l": 5, "f": 3022},
  {"h": "沧", "p": "cāng", "sp": "cang", "r": "水", "sc": 7, "l": 5, "f": 3023},
  {"h": "怠", "p": "dài", "sp": "dai", "r": "心", "sc": 9, "l": 5, "f": 3024},
  {"h": "沁", "p": "qìn", "sp": "qin", "r": "水", "sc": 7, "l": 5, "f": 3025},
  {"h": "奕", "p": "yì", "sp": "yi", "r": "大", "sc": 9, "l": 5, "f": 3026},
  {"h": "咙", "p": "lóng", "sp": "long", "r": "口", "sc": 8, "l": 5, "f": 3027},
  {"h": "氨", "p": "ān", "sp": "an", "r": "气", "sc": 10, "l": 5, "f": 3028},
  {"h": "矗", "p": "chù", "sp": "chu", "r": "目", "sc": 24, "l": 5, "f": 3029},
  {"h": "盔", "p": "kuī", "sp": "kui", "r": "皿", "sc": 11, "l": 5, "f": 3030},
  {"h": "拇", "p": "mǔ", "sp": "mu", "r": "手", "sc": 8, "l": 5, "f": 3031},
  {"h": "沛", "p": "pèi", "sp": "pei", "r": "水", "sc": 7, "l": 5, "f": 3032},
  {"h": "榻", "p": "tà", "sp": "ta", "r": "木", "sc": 14, "l": 5, "f": 3033},
  {
    "h": "揣",
    "p": "chuāi",
    "sp": "chuai",
    "r": "手",
    "sc": 12,
    "l": 5,
    "f": 3034
  },
  {"h": "崭", "p": "zhǎn", "sp": "zhan", "r": "山", "sc": 11, "l": 5, "f": 3035},
  {"h": "鞘", "p": "qiào", "sp": "qiao", "r": "革", "sc": 16, "l": 5, "f": 3036},
  {"h": "鞠", "p": "jū", "sp": "ju", "r": "革", "sc": 17, "l": 5, "f": 3037},
  {"h": "垦", "p": "kěn", "sp": "ken", "r": "土", "sc": 9, "l": 5, "f": 3038},
  {"h": "洽", "p": "qià", "sp": "qia", "r": "水", "sc": 9, "l": 5, "f": 3039},
  {"h": "唾", "p": "tuò", "sp": "tuo", "r": "口", "sc": 11, "l": 5, "f": 3040},
  {"h": "橱", "p": "chú", "sp": "chu", "r": "木", "sc": 16, "l": 5, "f": 3041},
  {"h": "仕", "p": "shì", "sp": "shi", "r": "人", "sc": 5, "l": 5, "f": 3042},
  {"h": "蜘", "p": "zhī", "sp": "zhi", "r": "虫", "sc": 14, "l": 5, "f": 3043},
  {"h": "痰", "p": "tán", "sp": "tan", "r": "疒", "sc": 13, "l": 5, "f": 3044},
  {"h": "袜", "p": "wà", "sp": "wa", "r": "衣", "sc": 10, "l": 4, "f": 3045},
  {"h": "峙", "p": "zhì", "sp": "zhi", "r": "山", "sc": 9, "l": 5, "f": 3046},
  {"h": "柬", "p": "jiǎn", "sp": "jian", "r": "木", "sc": 9, "l": 5, "f": 3047},
  {"h": "蝉", "p": "chán", "sp": "chan", "r": "虫", "sc": 14, "l": 5, "f": 3048},
  {"h": "蟹", "p": "xiè", "sp": "xie", "r": "虫", "sc": 19, "l": 5, "f": 3049},
  {"h": "谏", "p": "jiàn", "sp": "jian", "r": "言", "sc": 11, "l": 5, "f": 3050},
  {"h": "鹃", "p": "juān", "sp": "juan", "r": "鳥", "sc": 12, "l": 5, "f": 3051},
  {"h": "擎", "p": "qíng", "sp": "qing", "r": "手", "sc": 16, "l": 5, "f": 3052},
  {"h": "皓", "p": "hào", "sp": "hao", "r": "白", "sc": 12, "l": 5, "f": 3053},
  {"h": "朕", "p": "zhèn", "sp": "zhen", "r": "月", "sc": 10, "l": 5, "f": 3054},
  {"h": "疤", "p": "bā", "sp": "ba", "r": "疒", "sc": 9, "l": 5, "f": 3055},
  {"h": "禺", "p": "yú", "sp": "yu", "r": "禸", "sc": 9, "l": 5, "f": 3056},
  {"h": "铲", "p": "chǎn", "sp": "chan", "r": "金", "sc": 11, "l": 5, "f": 3057},
  {"h": "酶", "p": "méi", "sp": "mei", "r": "酉", "sc": 14, "l": 5, "f": 3058},
  {"h": "钝", "p": "dùn", "sp": "dun", "r": "金", "sc": 9, "l": 5, "f": 3059},
  {"h": "氓", "p": "máng", "sp": "mang", "r": "氏", "sc": 8, "l": 5, "f": 3060},
  {"h": "匣", "p": "xiá", "sp": "xia", "r": "匚", "sc": 7, "l": 5, "f": 3061},
  {"h": "弧", "p": "hú", "sp": "hu", "r": "弓", "sc": 8, "l": 5, "f": 3062},
  {"h": "峨", "p": "é", "sp": "e", "r": "山", "sc": 10, "l": 5, "f": 3063},
  {"h": "锥", "p": "zhuī", "sp": "zhui", "r": "金", "sc": 13, "l": 5, "f": 3064},
  {"h": "揪", "p": "jiū", "sp": "jiu", "r": "手", "sc": 12, "l": 5, "f": 3065},
  {"h": "杠", "p": "gāng", "sp": "gang", "r": "木", "sc": 7, "l": 5, "f": 3066},
  {"h": "吭", "p": "kēng", "sp": "keng", "r": "口", "sc": 7, "l": 5, "f": 3067},
  {"h": "崛", "p": "jué", "sp": "jue", "r": "山", "sc": 11, "l": 5, "f": 3068},
  {"h": "诬", "p": "wū", "sp": "wu", "r": "言", "sc": 9, "l": 5, "f": 3069},
  {"h": "冉", "p": "rǎn", "sp": "ran", "r": "冂", "sc": 5, "l": 5, "f": 3070},
  {"h": "抒", "p": "shū", "sp": "shu", "r": "手", "sc": 7, "l": 5, "f": 3071},
  {"h": "庚", "p": "gēng", "sp": "geng", "r": "广", "sc": 8, "l": 5, "f": 3072},
  {"h": "悍", "p": "hàn", "sp": "han", "r": "心", "sc": 10, "l": 5, "f": 3073},
  {"h": "靡", "p": "mí", "sp": "mi", "r": "非", "sc": 19, "l": 5, "f": 3074},
  {"h": "晦", "p": "huì", "sp": "hui", "r": "日", "sc": 11, "l": 5, "f": 3075},
  {"h": "醋", "p": "cù", "sp": "cu", "r": "酉", "sc": 15, "l": 5, "f": 3076},
  {"h": "壕", "p": "háo", "sp": "hao", "r": "土", "sc": 17, "l": 5, "f": 3077},
  {"h": "锯", "p": "jù", "sp": "ju", "r": "金", "sc": 13, "l": 5, "f": 3078},
  {"h": "夭", "p": "yāo", "sp": "yao", "r": "大", "sc": 4, "l": 5, "f": 3079},
  {"h": "咦", "p": "yí", "sp": "yi", "r": "口", "sc": 9, "l": 5, "f": 3080},
  {"h": "侈", "p": "chǐ", "sp": "chi", "r": "人", "sc": 8, "l": 5, "f": 3081},
  {"h": "婢", "p": "bì", "sp": "bi", "r": "女", "sc": 11, "l": 5, "f": 3082},
  {"h": "猾", "p": "huá", "sp": "hua", "r": "犬", "sc": 12, "l": 5, "f": 3083},
  {"h": "徘", "p": "pái", "sp": "pai", "r": "彳", "sc": 11, "l": 5, "f": 3084},
  {"h": "硝", "p": "xiāo", "sp": "xiao", "r": "石", "sc": 12, "l": 5, "f": 3085},
  {"h": "煽", "p": "shān", "sp": "shan", "r": "火", "sc": 14, "l": 5, "f": 3086},
  {"h": "皂", "p": "zào", "sp": "zao", "r": "白", "sc": 7, "l": 5, "f": 3087},
  {"h": "舵", "p": "duò", "sp": "duo", "r": "舟", "sc": 11, "l": 5, "f": 3088},
  {"h": "嗦", "p": "suo", "sp": "suo", "r": "口", "sc": 13, "l": 5, "f": 3089},
  {"h": "狈", "p": "bèi", "sp": "bei", "r": "犬", "sc": 7, "l": 5, "f": 3090},
  {"h": "靴", "p": "xuē", "sp": "xue", "r": "革", "sc": 13, "l": 5, "f": 3091},
  {"h": "捂", "p": "wǔ", "sp": "wu", "r": "手", "sc": 10, "l": 5, "f": 3092},
  {
    "h": "疮",
    "p": "chuāng",
    "sp": "chuang",
    "r": "疒",
    "sc": 9,
    "l": 5,
    "f": 3093
  },
  {"h": "郝", "p": "hǎo", "sp": "hao", "r": "邑", "sc": 9, "l": 5, "f": 3094},
  {"h": "苛", "p": "kē", "sp": "ke", "r": "艸", "sc": 8, "l": 5, "f": 3095},
  {"h": "秽", "p": "huì", "sp": "hui", "r": "禾", "sc": 11, "l": 5, "f": 3096},
  {"h": "茜", "p": "qiàn", "sp": "qian", "r": "艸", "sc": 9, "l": 5, "f": 3097},
  {"h": "搓", "p": "cuō", "sp": "cuo", "r": "手", "sc": 12, "l": 5, "f": 3098},
  {"h": "芸", "p": "yún", "sp": "yun", "r": "艸", "sc": 7, "l": 5, "f": 3099},
  {
    "h": "酱",
    "p": "jiàng",
    "sp": "jiang",
    "r": "酉",
    "sc": 13,
    "l": 5,
    "f": 3100
  },
  {"h": "赁", "p": "lìn", "sp": "lin", "r": "貝", "sc": 10, "l": 5, "f": 3101},
  {"h": "檐", "p": "yán", "sp": "yan", "r": "木", "sc": 17, "l": 5, "f": 3102},
  {"h": "饷", "p": "xiǎng", "sp": "xiang", "r": "食", "sc": 9, "l": 5, "f": 3103},
  {"h": "蕉", "p": "jiāo", "sp": "jiao", "r": "艸", "sc": 15, "l": 3, "f": 3104},
  {"h": "铀", "p": "yóu", "sp": "you", "r": "金", "sc": 10, "l": 5, "f": 3105},
  {"h": "苔", "p": "tái", "sp": "tai", "r": "艸", "sc": 8, "l": 5, "f": 3106},
  {"h": "赦", "p": "shè", "sp": "she", "r": "赤", "sc": 11, "l": 5, "f": 3107},
  {"h": "缎", "p": "duàn", "sp": "duan", "r": "糸", "sc": 12, "l": 5, "f": 3108},
  {"h": "舷", "p": "xián", "sp": "xian", "r": "舟", "sc": 11, "l": 5, "f": 3109},
  {"h": "筷", "p": "kuài", "sp": "kuai", "r": "竹", "sc": 13, "l": 2, "f": 3110},
  {"h": "朔", "p": "shuò", "sp": "shuo", "r": "月", "sc": 10, "l": 5, "f": 3111},
  {"h": "婪", "p": "lán", "sp": "lan", "r": "女", "sc": 11, "l": 5, "f": 3112},
  {"h": "紊", "p": "wěn", "sp": "wen", "r": "糸", "sc": 10, "l": 5, "f": 3113},
  {"h": "厮", "p": "sī", "sp": "si", "r": "厂", "sc": 14, "l": 5, "f": 3114},
  {"h": "婿", "p": "xù", "sp": "xu", "r": "女", "sc": 12, "l": 5, "f": 3115},
  {"h": "寥", "p": "liáo", "sp": "liao", "r": "宀", "sc": 14, "l": 5, "f": 3116},
  {"h": "兢", "p": "jīng", "sp": "jing", "r": "儿", "sc": 14, "l": 5, "f": 3117},
  {"h": "糙", "p": "cāo", "sp": "cao", "r": "米", "sc": 16, "l": 5, "f": 3118},
  {"h": "卦", "p": "guà", "sp": "gua", "r": "卜", "sc": 8, "l": 5, "f": 3119},
  {"h": "槐", "p": "huái", "sp": "huai", "r": "木", "sc": 13, "l": 5, "f": 3120},
  {"h": "扒", "p": "bā", "sp": "ba", "r": "手", "sc": 5, "l": 5, "f": 3121},
  {"h": "裴", "p": "péi", "sp": "pei", "r": "衣", "sc": 14, "l": 5, "f": 3122},
  {"h": "祀", "p": "sì", "sp": "si", "r": "示", "sc": 7, "l": 5, "f": 3123},
  {"h": "埔", "p": "bù", "sp": "bu", "r": "土", "sc": 10, "l": 5, "f": 3124},
  {"h": "絮", "p": "xù", "sp": "xu", "r": "糸", "sc": 12, "l": 5, "f": 3125},
  {"h": "芭", "p": "bā", "sp": "ba", "r": "艸", "sc": 7, "l": 5, "f": 3126},
  {"h": "屉", "p": "tì", "sp": "ti", "r": "尸", "sc": 8, "l": 5, "f": 3127},
  {"h": "痪", "p": "huàn", "sp": "huan", "r": "疒", "sc": 12, "l": 5, "f": 3128},
  {"h": "霄", "p": "xiāo", "sp": "xiao", "r": "雨", "sc": 15, "l": 5, "f": 3129},
  {"h": "绽", "p": "zhàn", "sp": "zhan", "r": "糸", "sc": 11, "l": 5, "f": 3130},
  {"h": "宵", "p": "xiāo", "sp": "xiao", "r": "宀", "sc": 10, "l": 5, "f": 3131},
  {"h": "邑", "p": "yì", "sp": "yi", "r": "邑", "sc": 7, "l": 5, "f": 3132},
  {"h": "霖", "p": "lín", "sp": "lin", "r": "雨", "sc": 16, "l": 5, "f": 3133},
  {"h": "岔", "p": "chà", "sp": "cha", "r": "山", "sc": 7, "l": 5, "f": 3134},
  {"h": "饵", "p": "ěr", "sp": "er", "r": "食", "sc": 9, "l": 5, "f": 3135},
  {"h": "茄", "p": "jiā", "sp": "jia", "r": "艸", "sc": 8, "l": 5, "f": 3136},
  {"h": "韧", "p": "rèn", "sp": "ren", "r": "韋", "sc": 7, "l": 5, "f": 3137},
  {"h": "琪", "p": "qí", "sp": "qi", "r": "玉", "sc": 12, "l": 5, "f": 3138},
  {"h": "邹", "p": "zōu", "sp": "zou", "r": "邑", "sc": 7, "l": 5, "f": 3139},
  {"h": "瑚", "p": "hú", "sp": "hu", "r": "玉", "sc": 13, "l": 5, "f": 3140},
  {"h": "憋", "p": "biē", "sp": "bie", "r": "心", "sc": 15, "l": 5, "f": 3141},
  {"h": "殆", "p": "dài", "sp": "dai", "r": "歹", "sc": 9, "l": 5, "f": 3142},
  {"h": "噜", "p": "lū", "sp": "lu", "r": "口", "sc": 15, "l": 5, "f": 3143},
  {"h": "忒", "p": "tè", "sp": "te", "r": "心", "sc": 7, "l": 5, "f": 3144},
  {"h": "忿", "p": "fèn", "sp": "fen", "r": "心", "sc": 8, "l": 5, "f": 3145},
  {"h": "衅", "p": "xìn", "sp": "xin", "r": "血", "sc": 11, "l": 5, "f": 3146},
  {"h": "淳", "p": "chún", "sp": "chun", "r": "水", "sc": 11, "l": 5, "f": 3147},
  {"h": "悖", "p": "bèi", "sp": "bei", "r": "心", "sc": 10, "l": 5, "f": 3148},
  {"h": "髦", "p": "máo", "sp": "mao", "r": "髟", "sc": 14, "l": 5, "f": 3149},
  {"h": "孜", "p": "zī", "sp": "zi", "r": "子", "sc": 7, "l": 5, "f": 3150},
  {"h": "粤", "p": "yuè", "sp": "yue", "r": "米", "sc": 12, "l": 5, "f": 3151},
  {"h": "隘", "p": "ài", "sp": "ai", "r": "阜", "sc": 12, "l": 5, "f": 3152},
  {"h": "濒", "p": "bīn", "sp": "bin", "r": "水", "sc": 16, "l": 5, "f": 3153},
  {
    "h": "铮",
    "p": "zhēng",
    "sp": "zheng",
    "r": "金",
    "sc": 11,
    "l": 5,
    "f": 3154
  },
  {"h": "畸", "p": "jī", "sp": "ji", "r": "田", "sc": 13, "l": 5, "f": 3155},
  {"h": "剔", "p": "tī", "sp": "ti", "r": "刀", "sc": 10, "l": 5, "f": 3156},
  {"h": "坞", "p": "wù", "sp": "wu", "r": "土", "sc": 7, "l": 5, "f": 3157},
  {"h": "篱", "p": "lí", "sp": "li", "r": "竹", "sc": 16, "l": 5, "f": 3158},
  {"h": "淀", "p": "diàn", "sp": "dian", "r": "水", "sc": 11, "l": 5, "f": 3159},
  {"h": "蓦", "p": "mò", "sp": "mo", "r": "馬", "sc": 13, "l": 5, "f": 3160},
  {"h": "唬", "p": "hǔ", "sp": "hu", "r": "口", "sc": 11, "l": 5, "f": 3161},
  {"h": "锣", "p": "luó", "sp": "luo", "r": "金", "sc": 13, "l": 5, "f": 3162},
  {"h": "汀", "p": "tīng", "sp": "ting", "r": "水", "sc": 5, "l": 5, "f": 3163},
  {"h": "趾", "p": "zhǐ", "sp": "zhi", "r": "足", "sc": 11, "l": 5, "f": 3164},
  {"h": "缉", "p": "jī", "sp": "ji", "r": "糸", "sc": 12, "l": 5, "f": 3165},
  {
    "h": "嫦",
    "p": "cháng",
    "sp": "chang",
    "r": "女",
    "sc": 14,
    "l": 5,
    "f": 3166
  },
  {"h": "斟", "p": "zhēn", "sp": "zhen", "r": "斗", "sc": 13, "l": 5, "f": 3167},
  {"h": "鞍", "p": "ān", "sp": "an", "r": "革", "sc": 15, "l": 5, "f": 3168},
  {"h": "扳", "p": "bān", "sp": "ban", "r": "手", "sc": 7, "l": 5, "f": 3169},
  {"h": "拴", "p": "shuān", "sp": "shuan", "r": "手", "sc": 9, "l": 5, "f": 3170},
  {"h": "诅", "p": "zǔ", "sp": "zu", "r": "言", "sc": 7, "l": 5, "f": 3171},
  {"h": "谟", "p": "mó", "sp": "mo", "r": "言", "sc": 12, "l": 5, "f": 3172},
  {"h": "呃", "p": "è", "sp": "e", "r": "口", "sc": 7, "l": 5, "f": 3173},
  {"h": "懦", "p": "nuò", "sp": "nuo", "r": "心", "sc": 17, "l": 5, "f": 3174},
  {
    "h": "逞",
    "p": "chěng",
    "sp": "cheng",
    "r": "辵",
    "sc": 10,
    "l": 5,
    "f": 3175
  },
  {"h": "犁", "p": "lí", "sp": "li", "r": "牛", "sc": 11, "l": 5, "f": 3176},
  {"h": "忏", "p": "chàn", "sp": "chan", "r": "心", "sc": 6, "l": 5, "f": 3177},
  {"h": "拧", "p": "níng", "sp": "ning", "r": "手", "sc": 8, "l": 5, "f": 3178},
  {"h": "亥", "p": "hài", "sp": "hai", "r": "亠", "sc": 6, "l": 5, "f": 3179},
  {"h": "佟", "p": "tóng", "sp": "tong", "r": "人", "sc": 7, "l": 5, "f": 3180},
  {"h": "叱", "p": "chì", "sp": "chi", "r": "口", "sc": 5, "l": 5, "f": 3181},
  {"h": "舜", "p": "shùn", "sp": "shun", "r": "舛", "sc": 12, "l": 5, "f": 3182},
  {"h": "绊", "p": "bàn", "sp": "ban", "r": "糸", "sc": 8, "l": 5, "f": 3183},
  {"h": "龚", "p": "gōng", "sp": "gong", "r": "龍", "sc": 11, "l": 5, "f": 3184},
  {"h": "腮", "p": "sāi", "sp": "sai", "r": "肉", "sc": 13, "l": 5, "f": 3185},
  {"h": "邸", "p": "dǐ", "sp": "di", "r": "邑", "sc": 7, "l": 5, "f": 3186},
  {"h": "椒", "p": "jiāo", "sp": "jiao", "r": "木", "sc": 12, "l": 5, "f": 3187},
  {"h": "蔚", "p": "wèi", "sp": "wei", "r": "艸", "sc": 14, "l": 5, "f": 3188},
  {"h": "湛", "p": "zhàn", "sp": "zhan", "r": "水", "sc": 12, "l": 5, "f": 3189},
  {"h": "狩", "p": "shòu", "sp": "shou", "r": "犬", "sc": 9, "l": 5, "f": 3190},
  {
    "h": "眶",
    "p": "kuàng",
    "sp": "kuang",
    "r": "目",
    "sc": 11,
    "l": 5,
    "f": 3191
  },
  {"h": "栈", "p": "zhàn", "sp": "zhan", "r": "木", "sc": 9, "l": 5, "f": 3192},
  {"h": "薇", "p": "wēi", "sp": "wei", "r": "艸", "sc": 16, "l": 5, "f": 3193},
  {"h": "肮", "p": "āng", "sp": "ang", "r": "肉", "sc": 8, "l": 5, "f": 3194},
  {"h": "瀑", "p": "pù", "sp": "pu", "r": "水", "sc": 18, "l": 5, "f": 3195},
  {"h": "渣", "p": "zhā", "sp": "zha", "r": "水", "sc": 12, "l": 5, "f": 3196},
  {"h": "褂", "p": "guà", "sp": "gua", "r": "衣", "sc": 13, "l": 5, "f": 3197},
  {"h": "叽", "p": "jī", "sp": "ji", "r": "口", "sc": 5, "l": 5, "f": 3198},
  {"h": "臀", "p": "tún", "sp": "tun", "r": "肉", "sc": 17, "l": 5, "f": 3199},
  {"h": "妞", "p": "niū", "sp": "niu", "r": "女", "sc": 7, "l": 5, "f": 3200},
  {"h": "巍", "p": "wēi", "sp": "wei", "r": "山", "sc": 20, "l": 5, "f": 3201},
  {"h": "唔", "p": "wú", "sp": "wu", "r": "口", "sc": 10, "l": 5, "f": 3202},
  {"h": "疚", "p": "jiù", "sp": "jiu", "r": "疒", "sc": 8, "l": 5, "f": 3203},
  {"h": "鲤", "p": "lǐ", "sp": "li", "r": "魚", "sc": 15, "l": 5, "f": 3204},
  {"h": "戎", "p": "róng", "sp": "rong", "r": "戈", "sc": 6, "l": 5, "f": 3205},
  {"h": "肇", "p": "zhào", "sp": "zhao", "r": "聿", "sc": 14, "l": 5, "f": 3206},
  {"h": "笃", "p": "dǔ", "sp": "du", "r": "竹", "sc": 9, "l": 5, "f": 3207},
  {"h": "辙", "p": "zhé", "sp": "zhe", "r": "車", "sc": 16, "l": 5, "f": 3208},
  {"h": "娴", "p": "xián", "sp": "xian", "r": "女", "sc": 10, "l": 5, "f": 3209},
  {"h": "阮", "p": "ruǎn", "sp": "ruan", "r": "阜", "sc": 6, "l": 5, "f": 3210},
  {"h": "札", "p": "zhá", "sp": "zha", "r": "木", "sc": 5, "l": 5, "f": 3211},
  {"h": "懊", "p": "ào", "sp": "ao", "r": "心", "sc": 15, "l": 5, "f": 3212},
  {"h": "焘", "p": "dào", "sp": "dao", "r": "火", "sc": 11, "l": 5, "f": 3213},
  {"h": "恤", "p": "xù", "sp": "xu", "r": "心", "sc": 9, "l": 5, "f": 3214},
  {"h": "疹", "p": "zhěn", "sp": "zhen", "r": "疒", "sc": 10, "l": 5, "f": 3215},
  {"h": "潇", "p": "xiāo", "sp": "xiao", "r": "水", "sc": 14, "l": 5, "f": 3216},
  {"h": "铝", "p": "lǚ", "sp": "lu", "r": "金", "sc": 11, "l": 5, "f": 3217},
  {"h": "涤", "p": "dí", "sp": "di", "r": "水", "sc": 10, "l": 5, "f": 3218},
  {"h": "恃", "p": "shì", "sp": "shi", "r": "心", "sc": 9, "l": 5, "f": 3219},
  {"h": "喽", "p": "lóu", "sp": "lou", "r": "口", "sc": 12, "l": 5, "f": 3220},
  {"h": "砌", "p": "qì", "sp": "qi", "r": "石", "sc": 9, "l": 5, "f": 3221},
  {"h": "遁", "p": "dùn", "sp": "dun", "r": "辵", "sc": 12, "l": 5, "f": 3222},
  {"h": "楞", "p": "léng", "sp": "leng", "r": "木", "sc": 13, "l": 5, "f": 3223},
  {"h": "阱", "p": "jǐng", "sp": "jing", "r": "阜", "sc": 6, "l": 5, "f": 3224},
  {"h": "咎", "p": "jiù", "sp": "jiu", "r": "口", "sc": 8, "l": 5, "f": 3225},
  {"h": "洼", "p": "wā", "sp": "wa", "r": "水", "sc": 9, "l": 5, "f": 3226},
  {"h": "炳", "p": "bǐng", "sp": "bing", "r": "火", "sc": 9, "l": 5, "f": 3227},
  {"h": "噬", "p": "shì", "sp": "shi", "r": "口", "sc": 16, "l": 5, "f": 3228},
  {"h": "枫", "p": "fēng", "sp": "feng", "r": "木", "sc": 8, "l": 5, "f": 3229},
  {"h": "拷", "p": "kǎo", "sp": "kao", "r": "手", "sc": 9, "l": 5, "f": 3230},
  {"h": "哆", "p": "duō", "sp": "duo", "r": "口", "sc": 9, "l": 5, "f": 3231},
  {"h": "矶", "p": "jī", "sp": "ji", "r": "石", "sc": 7, "l": 5, "f": 3232},
  {"h": "苇", "p": "wěi", "sp": "wei", "r": "艸", "sc": 7, "l": 5, "f": 3233},
  {"h": "翩", "p": "piān", "sp": "pian", "r": "羽", "sc": 15, "l": 5, "f": 3234},
  {"h": "窒", "p": "zhì", "sp": "zhi", "r": "穴", "sc": 11, "l": 5, "f": 3235},
  {"h": "侬", "p": "nóng", "sp": "nong", "r": "人", "sc": 8, "l": 5, "f": 3236},
  {"h": "靶", "p": "bǎ", "sp": "ba", "r": "革", "sc": 13, "l": 5, "f": 3237},
  {"h": "胰", "p": "yí", "sp": "yi", "r": "肉", "sc": 10, "l": 5, "f": 3238},
  {"h": "芜", "p": "wú", "sp": "wu", "r": "艸", "sc": 7, "l": 5, "f": 3239},
  {"h": "辫", "p": "biàn", "sp": "bian", "r": "辛", "sc": 17, "l": 5, "f": 3240},
  {"h": "嚎", "p": "háo", "sp": "hao", "r": "口", "sc": 17, "l": 5, "f": 3241},
  {"h": "妾", "p": "qiè", "sp": "qie", "r": "女", "sc": 8, "l": 5, "f": 3242},
  {
    "h": "幌",
    "p": "huǎng",
    "sp": "huang",
    "r": "巾",
    "sc": 13,
    "l": 5,
    "f": 3243
  },
  {
    "h": "踉",
    "p": "liáng",
    "sp": "liang",
    "r": "足",
    "sc": 14,
    "l": 5,
    "f": 3244
  },
  {"h": "佃", "p": "diàn", "sp": "dian", "r": "人", "sc": 7, "l": 5, "f": 3245},
  {"h": "葫", "p": "hú", "sp": "hu", "r": "艸", "sc": 12, "l": 5, "f": 3246},
  {"h": "皖", "p": "wǎn", "sp": "wan", "r": "白", "sc": 12, "l": 5, "f": 3247},
  {"h": "拽", "p": "zhuāi", "sp": "zhuai", "r": "手", "sc": 9, "l": 5, "f": 3248},
  {"h": "滤", "p": "lǜ", "sp": "lu", "r": "水", "sc": 13, "l": 5, "f": 3249},
  {"h": "睬", "p": "cǎi", "sp": "cai", "r": "目", "sc": 13, "l": 5, "f": 3250},
  {"h": "俞", "p": "yú", "sp": "yu", "r": "人", "sc": 9, "l": 5, "f": 3251},
  {"h": "匕", "p": "bǐ", "sp": "bi", "r": "匕", "sc": 2, "l": 5, "f": 3252},
  {"h": "谤", "p": "bàng", "sp": "bang", "r": "言", "sc": 12, "l": 5, "f": 3253},
  {"h": "嗤", "p": "chī", "sp": "chi", "r": "口", "sc": 13, "l": 5, "f": 3254},
  {"h": "捍", "p": "hàn", "sp": "han", "r": "手", "sc": 10, "l": 5, "f": 3255},
  {"h": "孵", "p": "fū", "sp": "fu", "r": "子", "sc": 14, "l": 5, "f": 3256},
  {"h": "倪", "p": "ní", "sp": "ni", "r": "人", "sc": 10, "l": 5, "f": 3257},
  {"h": "瘾", "p": "yǐn", "sp": "yin", "r": "疒", "sc": 16, "l": 5, "f": 3258},
  {"h": "敝", "p": "bì", "sp": "bi", "r": "攴", "sc": 11, "l": 5, "f": 3259},
  {"h": "匡", "p": "kuāng", "sp": "kuang", "r": "匚", "sc": 6, "l": 5, "f": 3260},
  {"h": "磋", "p": "cuō", "sp": "cuo", "r": "石", "sc": 14, "l": 5, "f": 3261},
  {"h": "绫", "p": "líng", "sp": "ling", "r": "糸", "sc": 11, "l": 5, "f": 3262},
  {"h": "淆", "p": "xiáo", "sp": "xiao", "r": "水", "sc": 11, "l": 5, "f": 3263},
  {"h": "尧", "p": "yáo", "sp": "yao", "r": "儿", "sc": 6, "l": 5, "f": 3264},
  {"h": "蕊", "p": "ruǐ", "sp": "rui", "r": "艸", "sc": 15, "l": 5, "f": 3265},
  {"h": "烘", "p": "hōng", "sp": "hong", "r": "火", "sc": 10, "l": 5, "f": 3266},
  {
    "h": "璋",
    "p": "zhāng",
    "sp": "zhang",
    "r": "玉",
    "sc": 15,
    "l": 5,
    "f": 3267
  },
  {"h": "亢", "p": "kàng", "sp": "kang", "r": "亠", "sc": 4, "l": 5, "f": 3268},
  {"h": "轧", "p": "yà", "sp": "ya", "r": "車", "sc": 5, "l": 5, "f": 3269},
  {"h": "赂", "p": "lù", "sp": "lu", "r": "貝", "sc": 10, "l": 5, "f": 3270},
  {
    "h": "蝗",
    "p": "huáng",
    "sp": "huang",
    "r": "虫",
    "sc": 15,
    "l": 5,
    "f": 3271
  },
  {"h": "榆", "p": "yú", "sp": "yu", "r": "木", "sc": 13, "l": 5, "f": 3272},
  {"h": "骏", "p": "jùn", "sp": "jun", "r": "馬", "sc": 10, "l": 5, "f": 3273},
  {"h": "诛", "p": "zhū", "sp": "zhu", "r": "言", "sc": 8, "l": 5, "f": 3274},
  {"h": "勺", "p": "sháo", "sp": "shao", "r": "勹", "sc": 3, "l": 5, "f": 3275},
  {"h": "梵", "p": "fàn", "sp": "fan", "r": "木", "sc": 11, "l": 5, "f": 3276},
  {"h": "炽", "p": "chì", "sp": "chi", "r": "火", "sc": 9, "l": 5, "f": 3277},
  {"h": "笠", "p": "lì", "sp": "li", "r": "竹", "sc": 11, "l": 5, "f": 3278},
  {"h": "颌", "p": "hé", "sp": "he", "r": "頁", "sc": 12, "l": 5, "f": 3279},
  {"h": "闸", "p": "zhá", "sp": "zha", "r": "門", "sc": 8, "l": 5, "f": 3280},
  {"h": "狒", "p": "fèi", "sp": "fei", "r": "犬", "sc": 8, "l": 5, "f": 3281},
  {"h": "樊", "p": "fán", "sp": "fan", "r": "木", "sc": 15, "l": 5, "f": 3282},
  {"h": "镕", "p": "róng", "sp": "rong", "r": "金", "sc": 15, "l": 5, "f": 3283},
  {"h": "垢", "p": "gòu", "sp": "gou", "r": "土", "sc": 9, "l": 5, "f": 3284},
  {"h": "瘟", "p": "wēn", "sp": "wen", "r": "疒", "sc": 14, "l": 5, "f": 3285},
  {"h": "缪", "p": "móu", "sp": "mou", "r": "糸", "sc": 14, "l": 5, "f": 3286},
  {"h": "菇", "p": "gū", "sp": "gu", "r": "艸", "sc": 11, "l": 5, "f": 3287},
  {"h": "琦", "p": "qí", "sp": "qi", "r": "玉", "sc": 12, "l": 5, "f": 3288},
  {"h": "剃", "p": "tì", "sp": "ti", "r": "刀", "sc": 9, "l": 5, "f": 3289},
  {"h": "迸", "p": "bèng", "sp": "beng", "r": "辵", "sc": 9, "l": 5, "f": 3290},
  {"h": "溺", "p": "nì", "sp": "ni", "r": "水", "sc": 13, "l": 5, "f": 3291},
  {"h": "炫", "p": "xuàn", "sp": "xuan", "r": "火", "sc": 9, "l": 5, "f": 3292},
  {"h": "惚", "p": "hū", "sp": "hu", "r": "心", "sc": 11, "l": 5, "f": 3293},
  {"h": "嗨", "p": "hāi", "sp": "hai", "r": "口", "sc": 13, "l": 5, "f": 3294},
  {"h": "陨", "p": "yǔn", "sp": "yun", "r": "阜", "sc": 9, "l": 5, "f": 3295},
  {"h": "赃", "p": "zāng", "sp": "zang", "r": "貝", "sc": 10, "l": 5, "f": 3296},
  {"h": "羁", "p": "jī", "sp": "ji", "r": "网", "sc": 17, "l": 5, "f": 3297},
  {"h": "臻", "p": "zhēn", "sp": "zhen", "r": "至", "sc": 16, "l": 5, "f": 3298},
  {"h": "嘀", "p": "dí", "sp": "di", "r": "口", "sc": 14, "l": 5, "f": 3299},
  {"h": "膳", "p": "shàn", "sp": "shan", "r": "肉", "sc": 16, "l": 5, "f": 3300},
  {"h": "赣", "p": "gàn", "sp": "gan", "r": "貝", "sc": 21, "l": 5, "f": 3301},
  {"h": "踌", "p": "chóu", "sp": "chou", "r": "足", "sc": 14, "l": 5, "f": 3302},
  {"h": "殉", "p": "xùn", "sp": "xun", "r": "歹", "sc": 10, "l": 5, "f": 3303},
  {"h": "桔", "p": "jú", "sp": "ju", "r": "木", "sc": 10, "l": 5, "f": 3304},
  {"h": "瞿", "p": "qú", "sp": "qu", "r": "目", "sc": 18, "l": 5, "f": 3305},
  {"h": "闽", "p": "mǐn", "sp": "min", "r": "門", "sc": 9, "l": 5, "f": 3306},
  {"h": "豚", "p": "tún", "sp": "tun", "r": "豕", "sc": 11, "l": 5, "f": 3307},
  {"h": "掺", "p": "càn", "sp": "can", "r": "手", "sc": 11, "l": 5, "f": 3308},
  {"h": "沌", "p": "dùn", "sp": "dun", "r": "水", "sc": 7, "l": 5, "f": 3309},
  {"h": "惰", "p": "duò", "sp": "duo", "r": "心", "sc": 12, "l": 5, "f": 3310},
  {"h": "喳", "p": "zhā", "sp": "zha", "r": "口", "sc": 12, "l": 5, "f": 3311},
  {"h": "椭", "p": "tuǒ", "sp": "tuo", "r": "木", "sc": 12, "l": 5, "f": 3312},
  {"h": "咪", "p": "mī", "sp": "mi", "r": "口", "sc": 9, "l": 5, "f": 3313},
  {"h": "霎", "p": "shà", "sp": "sha", "r": "雨", "sc": 16, "l": 5, "f": 3314},
  {"h": "侃", "p": "kǎn", "sp": "kan", "r": "人", "sc": 8, "l": 5, "f": 3315},
  {"h": "猝", "p": "cù", "sp": "cu", "r": "犬", "sc": 11, "l": 5, "f": 3316},
  {"h": "窖", "p": "jiào", "sp": "jiao", "r": "穴", "sc": 12, "l": 5, "f": 3317},
  {"h": "戮", "p": "lù", "sp": "lu", "r": "戈", "sc": 15, "l": 5, "f": 3318},
  {"h": "祠", "p": "cí", "sp": "ci", "r": "示", "sc": 9, "l": 5, "f": 3319},
  {"h": "瞩", "p": "zhǔ", "sp": "zhu", "r": "目", "sc": 17, "l": 5, "f": 3320},
  {"h": "菁", "p": "jīng", "sp": "jing", "r": "艸", "sc": 11, "l": 5, "f": 3321},
  {"h": "躇", "p": "chú", "sp": "chu", "r": "足", "sc": 18, "l": 5, "f": 3322},
  {"h": "佬", "p": "lǎo", "sp": "lao", "r": "人", "sc": 8, "l": 5, "f": 3323},
  {"h": "肋", "p": "lē", "sp": "le", "r": "肉", "sc": 6, "l": 5, "f": 3324},
  {"h": "咄", "p": "duō", "sp": "duo", "r": "口", "sc": 8, "l": 5, "f": 3325},
  {"h": "忡", "p": "chōng", "sp": "chong", "r": "心", "sc": 7, "l": 5, "f": 3326},
  {"h": "雍", "p": "yōng", "sp": "yong", "r": "隹", "sc": 13, "l": 5, "f": 3327},
  {"h": "忱", "p": "chén", "sp": "chen", "r": "心", "sc": 7, "l": 5, "f": 3328},
  {"h": "蕾", "p": "lěi", "sp": "lei", "r": "艸", "sc": 16, "l": 5, "f": 3329},
  {
    "h": "跄",
    "p": "qiāng",
    "sp": "qiang",
    "r": "足",
    "sc": 11,
    "l": 5,
    "f": 3330
  },
  {"h": "硅", "p": "guī", "sp": "gui", "r": "石", "sc": 11, "l": 5, "f": 3331},
  {"h": "伎", "p": "jì", "sp": "ji", "r": "人", "sc": 6, "l": 5, "f": 3332},
  {"h": "炊", "p": "chuī", "sp": "chui", "r": "火", "sc": 8, "l": 5, "f": 3333},
  {"h": "钊", "p": "zhāo", "sp": "zhao", "r": "金", "sc": 7, "l": 5, "f": 3334},
  {"h": "蝠", "p": "fú", "sp": "fu", "r": "虫", "sc": 15, "l": 5, "f": 3335},
  {"h": "屎", "p": "shǐ", "sp": "shi", "r": "尸", "sc": 9, "l": 5, "f": 3336},
  {"h": "拭", "p": "shì", "sp": "shi", "r": "手", "sc": 9, "l": 5, "f": 3337},
  {"h": "谛", "p": "dì", "sp": "di", "r": "言", "sc": 11, "l": 5, "f": 3338},
  {"h": "褪", "p": "tuì", "sp": "tui", "r": "衣", "sc": 14, "l": 5, "f": 3339},
  {"h": "丞", "p": "chéng", "sp": "cheng", "r": "一", "sc": 6, "l": 5, "f": 3340},
  {"h": "卉", "p": "huì", "sp": "hui", "r": "十", "sc": 5, "l": 5, "f": 3341},
  {"h": "隧", "p": "suì", "sp": "sui", "r": "阜", "sc": 14, "l": 5, "f": 3342},
  {"h": "茸", "p": "rōng", "sp": "rong", "r": "艸", "sc": 9, "l": 5, "f": 3343},
  {"h": "钳", "p": "qián", "sp": "qian", "r": "金", "sc": 10, "l": 5, "f": 3344},
  {"h": "啃", "p": "kěn", "sp": "ken", "r": "口", "sc": 11, "l": 5, "f": 3345},
  {"h": "伢", "p": "yá", "sp": "ya", "r": "人", "sc": 6, "l": 5, "f": 3346},
  {"h": "闺", "p": "guī", "sp": "gui", "r": "門", "sc": 9, "l": 5, "f": 3347},
  {"h": "舔", "p": "tiǎn", "sp": "tian", "r": "舌", "sc": 14, "l": 5, "f": 3348},
  {"h": "蹬", "p": "dēng", "sp": "deng", "r": "足", "sc": 19, "l": 5, "f": 3349},
  {"h": "挛", "p": "luán", "sp": "luan", "r": "手", "sc": 10, "l": 5, "f": 3350},
  {"h": "眺", "p": "tiào", "sp": "tiao", "r": "目", "sc": 11, "l": 5, "f": 3351},
  {"h": "袱", "p": "fú", "sp": "fu", "r": "衣", "sc": 11, "l": 5, "f": 3352},
  {"h": "陇", "p": "lǒng", "sp": "long", "r": "阜", "sc": 7, "l": 5, "f": 3353},
  {"h": "殴", "p": "ōu", "sp": "ou", "r": "殳", "sc": 8, "l": 5, "f": 3354},
  {"h": "柿", "p": "shì", "sp": "shi", "r": "木", "sc": 9, "l": 4, "f": 3355},
  {"h": "梧", "p": "wú", "sp": "wu", "r": "木", "sc": 11, "l": 5, "f": 3356},
  {"h": "惺", "p": "xīng", "sp": "xing", "r": "心", "sc": 12, "l": 5, "f": 3357},
  {"h": "弛", "p": "chí", "sp": "chi", "r": "弓", "sc": 6, "l": 5, "f": 3358},
  {"h": "侥", "p": "jiǎo", "sp": "jiao", "r": "人", "sc": 8, "l": 5, "f": 3359},
  {"h": "琛", "p": "chēn", "sp": "chen", "r": "玉", "sc": 12, "l": 5, "f": 3360},
  {"h": "捅", "p": "tǒng", "sp": "tong", "r": "手", "sc": 10, "l": 5, "f": 3361},
  {"h": "酝", "p": "yùn", "sp": "yun", "r": "酉", "sc": 11, "l": 5, "f": 3362},
  {"h": "薯", "p": "shǔ", "sp": "shu", "r": "艸", "sc": 16, "l": 5, "f": 3363},
  {"h": "曳", "p": "yè", "sp": "ye", "r": "曰", "sc": 6, "l": 5, "f": 3364},
  {"h": "澈", "p": "chè", "sp": "che", "r": "水", "sc": 15, "l": 5, "f": 3365},
  {"h": "锈", "p": "xiù", "sp": "xiu", "r": "金", "sc": 12, "l": 5, "f": 3366},
  {"h": "稠", "p": "chóu", "sp": "chou", "r": "禾", "sc": 13, "l": 5, "f": 3367},
  {"h": "眸", "p": "móu", "sp": "mou", "r": "目", "sc": 11, "l": 5, "f": 3368},
  {"h": "咆", "p": "páo", "sp": "pao", "r": "口", "sc": 8, "l": 5, "f": 3369},
  {
    "h": "簧",
    "p": "huáng",
    "sp": "huang",
    "r": "竹",
    "sc": 17,
    "l": 5,
    "f": 3370
  },
  {"h": "鸥", "p": "ōu", "sp": "ou", "r": "鳥", "sc": 9, "l": 5, "f": 3371},
  {"h": "疡", "p": "yáng", "sp": "yang", "r": "疒", "sc": 8, "l": 5, "f": 3372},
  {"h": "渎", "p": "dú", "sp": "du", "r": "水", "sc": 11, "l": 5, "f": 3373},
  {"h": "汲", "p": "jí", "sp": "ji", "r": "水", "sc": 6, "l": 5, "f": 3374},
  {"h": "嬉", "p": "xī", "sp": "xi", "r": "女", "sc": 15, "l": 5, "f": 3375},
  {"h": "脓", "p": "nóng", "sp": "nong", "r": "肉", "sc": 10, "l": 5, "f": 3376},
  {"h": "骡", "p": "luó", "sp": "luo", "r": "馬", "sc": 14, "l": 5, "f": 3377},
  {"h": "穗", "p": "suì", "sp": "sui", "r": "禾", "sc": 17, "l": 5, "f": 3378},
  {"h": "槛", "p": "kǎn", "sp": "kan", "r": "木", "sc": 14, "l": 5, "f": 3379},
  {"h": "拎", "p": "līn", "sp": "lin", "r": "手", "sc": 8, "l": 5, "f": 3380},
  {"h": "巳", "p": "sì", "sp": "si", "r": "己", "sc": 3, "l": 5, "f": 3381},
  {"h": "邢", "p": "xíng", "sp": "xing", "r": "邑", "sc": 6, "l": 5, "f": 3382},
  {"h": "廿", "p": "niàn", "sp": "nian", "r": "廾", "sc": 4, "l": 5, "f": 3383},
  {"h": "搀", "p": "chān", "sp": "chan", "r": "手", "sc": 12, "l": 5, "f": 3384},
  {"h": "曙", "p": "shǔ", "sp": "shu", "r": "日", "sc": 17, "l": 5, "f": 3385},
  {"h": "樵", "p": "qiáo", "sp": "qiao", "r": "木", "sc": 16, "l": 5, "f": 3386},
  {"h": "隅", "p": "yú", "sp": "yu", "r": "阜", "sc": 11, "l": 5, "f": 3387},
  {"h": "筛", "p": "shāi", "sp": "shai", "r": "竹", "sc": 12, "l": 5, "f": 3388},
  {"h": "谒", "p": "yè", "sp": "ye", "r": "言", "sc": 11, "l": 5, "f": 3389},
  {"h": "倭", "p": "wō", "sp": "wo", "r": "人", "sc": 10, "l": 5, "f": 3390},
  {"h": "痹", "p": "bì", "sp": "bi", "r": "疒", "sc": 13, "l": 5, "f": 3391},
  {
    "h": "猖",
    "p": "chāng",
    "sp": "chang",
    "r": "犬",
    "sc": 11,
    "l": 5,
    "f": 3392
  },
  {"h": "佯", "p": "yáng", "sp": "yang", "r": "人", "sc": 8, "l": 5, "f": 3393},
  {"h": "肛", "p": "gāng", "sp": "gang", "r": "肉", "sc": 7, "l": 5, "f": 3394},
  {"h": "奚", "p": "xī", "sp": "xi", "r": "大", "sc": 10, "l": 5, "f": 3395},
  {"h": "甭", "p": "béng", "sp": "beng", "r": "用", "sc": 9, "l": 5, "f": 3396},
  {"h": "抨", "p": "pēng", "sp": "peng", "r": "手", "sc": 8, "l": 5, "f": 3397},
  {"h": "蛾", "p": "é", "sp": "e", "r": "虫", "sc": 13, "l": 5, "f": 3398},
  {"h": "唠", "p": "láo", "sp": "lao", "r": "口", "sc": 10, "l": 5, "f": 3399},
  {"h": "荧", "p": "yíng", "sp": "ying", "r": "艸", "sc": 9, "l": 5, "f": 3400},
  {"h": "嵩", "p": "sōng", "sp": "song", "r": "山", "sc": 13, "l": 5, "f": 3401},
  {"h": "漱", "p": "shù", "sp": "shu", "r": "水", "sc": 14, "l": 5, "f": 3402},
  {"h": "酋", "p": "qiú", "sp": "qiu", "r": "酉", "sc": 9, "l": 5, "f": 3403},
  {"h": "攘", "p": "rǎng", "sp": "rang", "r": "手", "sc": 20, "l": 5, "f": 3404},
  {"h": "诘", "p": "jí", "sp": "ji", "r": "言", "sc": 8, "l": 5, "f": 3405},
  {"h": "篡", "p": "cuàn", "sp": "cuan", "r": "竹", "sc": 16, "l": 5, "f": 3406},
  {"h": "睿", "p": "ruì", "sp": "rui", "r": "目", "sc": 14, "l": 5, "f": 3407},
  {"h": "噩", "p": "è", "sp": "e", "r": "口", "sc": 16, "l": 5, "f": 3408},
  {"h": "怅", "p": "chàng", "sp": "chang", "r": "心", "sc": 7, "l": 5, "f": 3409},
  {"h": "盎", "p": "àng", "sp": "ang", "r": "皿", "sc": 10, "l": 5, "f": 3410},
  {"h": "徙", "p": "xǐ", "sp": "xi", "r": "彳", "sc": 11, "l": 5, "f": 3411},
  {"h": "鞅", "p": "yāng", "sp": "yang", "r": "革", "sc": 14, "l": 5, "f": 3412},
  {"h": "漓", "p": "lí", "sp": "li", "r": "水", "sc": 13, "l": 5, "f": 3413},
  {"h": "祟", "p": "suì", "sp": "sui", "r": "示", "sc": 10, "l": 5, "f": 3414},
  {"h": "睫", "p": "jié", "sp": "jie", "r": "目", "sc": 13, "l": 5, "f": 3415},
  {"h": "攸", "p": "yōu", "sp": "you", "r": "攴", "sc": 7, "l": 5, "f": 3416},
  {"h": "翎", "p": "líng", "sp": "ling", "r": "羽", "sc": 11, "l": 5, "f": 3417},
  {"h": "呛", "p": "qiāng", "sp": "qiang", "r": "口", "sc": 7, "l": 5, "f": 3418},
  {
    "h": "筐",
    "p": "kuāng",
    "sp": "kuang",
    "r": "竹",
    "sc": 12,
    "l": 5,
    "f": 3419
  },
  {"h": "堑", "p": "qiàn", "sp": "qian", "r": "土", "sc": 11, "l": 5, "f": 3420},
  {"h": "檀", "p": "tán", "sp": "tan", "r": "木", "sc": 17, "l": 5, "f": 3421},
  {"h": "寅", "p": "yín", "sp": "yin", "r": "宀", "sc": 11, "l": 5, "f": 3422},
  {"h": "磊", "p": "lěi", "sp": "lei", "r": "石", "sc": 15, "l": 5, "f": 3423},
  {"h": "驭", "p": "yù", "sp": "yu", "r": "馬", "sc": 5, "l": 5, "f": 3424},
  {"h": "惘", "p": "wǎng", "sp": "wang", "r": "心", "sc": 11, "l": 5, "f": 3425},
  {"h": "吠", "p": "fèi", "sp": "fei", "r": "口", "sc": 7, "l": 5, "f": 3426},
  {"h": "驮", "p": "tuó", "sp": "tuo", "r": "馬", "sc": 6, "l": 5, "f": 3427},
  {"h": "瑙", "p": "nǎo", "sp": "nao", "r": "玉", "sc": 13, "l": 5, "f": 3428},
  {"h": "炬", "p": "jù", "sp": "ju", "r": "火", "sc": 8, "l": 5, "f": 3429},
  {"h": "痉", "p": "jìng", "sp": "jing", "r": "疒", "sc": 10, "l": 5, "f": 3430},
  {"h": "曝", "p": "pù", "sp": "pu", "r": "日", "sc": 19, "l": 5, "f": 3431},
  {"h": "恺", "p": "kǎi", "sp": "kai", "r": "心", "sc": 9, "l": 5, "f": 3432},
  {"h": "胺", "p": "àn", "sp": "an", "r": "肉", "sc": 10, "l": 5, "f": 3433},
  {"h": "萤", "p": "yíng", "sp": "ying", "r": "艸", "sc": 11, "l": 5, "f": 3434},
  {"h": "敕", "p": "chì", "sp": "chi", "r": "攴", "sc": 11, "l": 5, "f": 3435},
  {
    "h": "筝",
    "p": "zhēng",
    "sp": "zheng",
    "r": "竹",
    "sc": 12,
    "l": 5,
    "f": 3436
  },
  {"h": "幡", "p": "fān", "sp": "fan", "r": "巾", "sc": 15, "l": 5, "f": 3437},
  {"h": "霹", "p": "pī", "sp": "pi", "r": "雨", "sc": 21, "l": 5, "f": 3438},
  {"h": "竺", "p": "zhú", "sp": "zhu", "r": "竹", "sc": 8, "l": 5, "f": 3439},
  {"h": "烙", "p": "lào", "sp": "lao", "r": "火", "sc": 10, "l": 5, "f": 3440},
  {"h": "毗", "p": "pí", "sp": "pi", "r": "田", "sc": 9, "l": 5, "f": 3441},
  {"h": "鸠", "p": "jiū", "sp": "jiu", "r": "鳥", "sc": 7, "l": 5, "f": 3442},
  {"h": "埠", "p": "bù", "sp": "bu", "r": "土", "sc": 11, "l": 5, "f": 3443},
  {"h": "蒜", "p": "suàn", "sp": "suan", "r": "艸", "sc": 13, "l": 5, "f": 3444},
  {"h": "阜", "p": "fù", "sp": "fu", "r": "阜", "sc": 8, "l": 5, "f": 3445},
  {"h": "嘈", "p": "cáo", "sp": "cao", "r": "口", "sc": 14, "l": 5, "f": 3446},
  {"h": "乒", "p": "pīng", "sp": "ping", "r": "丿", "sc": 6, "l": 4, "f": 3447},
  {"h": "帷", "p": "wéi", "sp": "wei", "r": "巾", "sc": 11, "l": 5, "f": 3448},
  {"h": "啄", "p": "zhuó", "sp": "zhuo", "r": "口", "sc": 11, "l": 5, "f": 3449},
  {"h": "鳌", "p": "áo", "sp": "ao", "r": "魚", "sc": 18, "l": 5, "f": 3450},
  {"h": "毡", "p": "zhān", "sp": "zhan", "r": "毛", "sc": 9, "l": 5, "f": 3451},
  {"h": "阙", "p": "quē", "sp": "que", "r": "門", "sc": 13, "l": 5, "f": 3452},
  {"h": "褥", "p": "rù", "sp": "ru", "r": "衣", "sc": 15, "l": 5, "f": 3453},
  {"h": "搔", "p": "sāo", "sp": "sao", "r": "手", "sc": 12, "l": 5, "f": 3454},
  {"h": "笋", "p": "sǔn", "sp": "sun", "r": "竹", "sc": 10, "l": 5, "f": 3455},
  {"h": "冕", "p": "miǎn", "sp": "mian", "r": "冂", "sc": 11, "l": 5, "f": 3456},
  {"h": "狞", "p": "níng", "sp": "ning", "r": "犬", "sc": 8, "l": 5, "f": 3457},
  {"h": "韶", "p": "sháo", "sp": "shao", "r": "音", "sc": 14, "l": 5, "f": 3458},
  {"h": "骼", "p": "gé", "sp": "ge", "r": "骨", "sc": 15, "l": 5, "f": 3459},
  {"h": "蔼", "p": "ǎi", "sp": "ai", "r": "艸", "sc": 14, "l": 5, "f": 3460},
  {"h": "烹", "p": "pēng", "sp": "peng", "r": "火", "sc": 11, "l": 5, "f": 3461},
  {"h": "奄", "p": "yǎn", "sp": "yan", "r": "大", "sc": 8, "l": 5, "f": 3462},
  {"h": "嫖", "p": "piáo", "sp": "piao", "r": "女", "sc": 14, "l": 5, "f": 3463},
  {"h": "沐", "p": "mù", "sp": "mu", "r": "水", "sc": 7, "l": 5, "f": 3464},
  {"h": "噗", "p": "pū", "sp": "pu", "r": "口", "sc": 15, "l": 5, "f": 3465},
  {"h": "岑", "p": "cén", "sp": "cen", "r": "山", "sc": 7, "l": 5, "f": 3466},
  {"h": "蛟", "p": "jiāo", "sp": "jiao", "r": "虫", "sc": 12, "l": 5, "f": 3467},
  {"h": "掳", "p": "lǔ", "sp": "lu", "r": "手", "sc": 11, "l": 5, "f": 3468},
  {"h": "咏", "p": "yǒng", "sp": "yong", "r": "口", "sc": 8, "l": 5, "f": 3469},
  {"h": "弩", "p": "nǔ", "sp": "nu", "r": "弓", "sc": 8, "l": 5, "f": 3470},
  {"h": "捻", "p": "niǎn", "sp": "nian", "r": "手", "sc": 11, "l": 5, "f": 3471},
  {"h": "圃", "p": "pǔ", "sp": "pu", "r": "囗", "sc": 10, "l": 5, "f": 3472},
  {"h": "孚", "p": "fú", "sp": "fu", "r": "子", "sc": 7, "l": 5, "f": 3473},
  {"h": "悴", "p": "cuì", "sp": "cui", "r": "心", "sc": 11, "l": 5, "f": 3474},
  {"h": "诣", "p": "yì", "sp": "yi", "r": "言", "sc": 8, "l": 5, "f": 3475},
  {"h": "呱", "p": "gū", "sp": "gu", "r": "口", "sc": 8, "l": 5, "f": 3476},
  {"h": "祁", "p": "qí", "sp": "qi", "r": "示", "sc": 6, "l": 5, "f": 3477},
  {"h": "捶", "p": "chuí", "sp": "chui", "r": "手", "sc": 11, "l": 5, "f": 3478},
  {"h": "钠", "p": "nà", "sp": "na", "r": "金", "sc": 9, "l": 5, "f": 3479},
  {"h": "袄", "p": "ǎo", "sp": "ao", "r": "衣", "sc": 9, "l": 5, "f": 3480},
  {"h": "澎", "p": "pēng", "sp": "peng", "r": "水", "sc": 15, "l": 5, "f": 3481},
  {"h": "氮", "p": "dàn", "sp": "dan", "r": "气", "sc": 12, "l": 5, "f": 3482},
  {"h": "恪", "p": "kè", "sp": "ke", "r": "心", "sc": 9, "l": 5, "f": 3483},
  {"h": "雏", "p": "chú", "sp": "chu", "r": "隹", "sc": 13, "l": 5, "f": 3484},
  {"h": "撮", "p": "cuō", "sp": "cuo", "r": "手", "sc": 15, "l": 5, "f": 3485},
  {"h": "堰", "p": "yàn", "sp": "yan", "r": "土", "sc": 12, "l": 5, "f": 3486},
  {"h": "彷", "p": "fǎng", "sp": "fang", "r": "彳", "sc": 7, "l": 5, "f": 3487},
  {"h": "鹦", "p": "yīng", "sp": "ying", "r": "鳥", "sc": 16, "l": 5, "f": 3488},
  {"h": "晖", "p": "huī", "sp": "hui", "r": "日", "sc": 10, "l": 5, "f": 3489},
  {"h": "犀", "p": "xī", "sp": "xi", "r": "牛", "sc": 12, "l": 5, "f": 3490},
  {"h": "腑", "p": "fǔ", "sp": "fu", "r": "肉", "sc": 12, "l": 5, "f": 3491},
  {"h": "沽", "p": "gū", "sp": "gu", "r": "水", "sc": 8, "l": 5, "f": 3492},
  {"h": "橄", "p": "gǎn", "sp": "gan", "r": "木", "sc": 15, "l": 5, "f": 3493},
  {"h": "掐", "p": "qiā", "sp": "qia", "r": "手", "sc": 11, "l": 5, "f": 3494},
  {"h": "亵", "p": "xiè", "sp": "xie", "r": "亠", "sc": 12, "l": 5, "f": 3495},
  {"h": "龋", "p": "qǔ", "sp": "qu", "r": "齒", "sc": 17, "l": 5, "f": 3496},
  {"h": "嗒", "p": "dā", "sp": "da", "r": "口", "sc": 12, "l": 5, "f": 3497},
  {"h": "咀", "p": "jǔ", "sp": "ju", "r": "口", "sc": 8, "l": 5, "f": 3498},
  {"h": "祺", "p": "qí", "sp": "qi", "r": "示", "sc": 12, "l": 5, "f": 3499},
  {"h": "锚", "p": "máo", "sp": "mao", "r": "金", "sc": 13, "l": 5, "f": 3500},
  {"h": "匾", "p": "biǎn", "sp": "bian", "r": "匸", "sc": 11, "l": 5, "f": 3501},
  {"h": "乓", "p": "pāng", "sp": "pang", "r": "丿", "sc": 6, "l": 4, "f": 3502},
  {"h": "萃", "p": "cuì", "sp": "cui", "r": "艸", "sc": 11, "l": 5, "f": 3503},
  {"h": "贻", "p": "yí", "sp": "yi", "r": "貝", "sc": 9, "l": 5, "f": 3504},
  {"h": "揖", "p": "yī", "sp": "yi", "r": "手", "sc": 12, "l": 5, "f": 3505},
  {"h": "觑", "p": "qù", "sp": "qu", "r": "見", "sc": 15, "l": 5, "f": 3506},
  {"h": "吝", "p": "lìn", "sp": "lin", "r": "口", "sc": 7, "l": 5, "f": 3507},
  {"h": "憔", "p": "qiáo", "sp": "qiao", "r": "心", "sc": 15, "l": 5, "f": 3508},
  {"h": "羌", "p": "qiāng", "sp": "qiang", "r": "羊", "sc": 7, "l": 5, "f": 3509},
  {"h": "诲", "p": "huì", "sp": "hui", "r": "言", "sc": 9, "l": 5, "f": 3510},
  {"h": "砾", "p": "lì", "sp": "li", "r": "石", "sc": 10, "l": 5, "f": 3511},
  {"h": "蠕", "p": "rú", "sp": "ru", "r": "虫", "sc": 20, "l": 5, "f": 3512},
  {"h": "肴", "p": "yáo", "sp": "yao", "r": "肉", "sc": 8, "l": 5, "f": 3513},
  {"h": "撩", "p": "liāo", "sp": "liao", "r": "手", "sc": 15, "l": 5, "f": 3514},
  {"h": "坍", "p": "tān", "sp": "tan", "r": "土", "sc": 7, "l": 5, "f": 3515},
  {"h": "酥", "p": "sū", "sp": "su", "r": "酉", "sc": 12, "l": 5, "f": 3516},
  {"h": "袅", "p": "niǎo", "sp": "niao", "r": "衣", "sc": 10, "l": 5, "f": 3517},
  {"h": "黝", "p": "yǒu", "sp": "you", "r": "黑", "sc": 17, "l": 5, "f": 3518},
  {"h": "俾", "p": "bǐ", "sp": "bi", "r": "人", "sc": 10, "l": 5, "f": 3519},
  {"h": "嫣", "p": "yān", "sp": "yan", "r": "女", "sc": 14, "l": 5, "f": 3520},
  {"h": "穹", "p": "qióng", "sp": "qiong", "r": "穴", "sc": 8, "l": 5, "f": 3521},
  {"h": "秧", "p": "yāng", "sp": "yang", "r": "禾", "sc": 10, "l": 5, "f": 3522},
  {"h": "妊", "p": "rèn", "sp": "ren", "r": "女", "sc": 7, "l": 5, "f": 3523},
  {"h": "溉", "p": "gài", "sp": "gai", "r": "水", "sc": 12, "l": 5, "f": 3524},
  {"h": "鹊", "p": "què", "sp": "que", "r": "鳥", "sc": 13, "l": 5, "f": 3525},
  {"h": "聿", "p": "yù", "sp": "yu", "r": "聿", "sc": 6, "l": 5, "f": 3526},
  {"h": "疙", "p": "gē", "sp": "ge", "r": "疒", "sc": 8, "l": 5, "f": 3527},
  {"h": "蘑", "p": "mó", "sp": "mo", "r": "艸", "sc": 19, "l": 5, "f": 3528},
  {"h": "睾", "p": "gāo", "sp": "gao", "r": "目", "sc": 14, "l": 5, "f": 3529},
  {"h": "楷", "p": "kǎi", "sp": "kai", "r": "木", "sc": 13, "l": 5, "f": 3530},
  {"h": "酵", "p": "jiào", "sp": "jiao", "r": "酉", "sc": 14, "l": 5, "f": 3531},
  {"h": "茹", "p": "rú", "sp": "ru", "r": "艸", "sc": 9, "l": 5, "f": 3532},
  {"h": "锌", "p": "xīn", "sp": "xin", "r": "金", "sc": 12, "l": 5, "f": 3533},
  {"h": "滇", "p": "diān", "sp": "dian", "r": "水", "sc": 13, "l": 5, "f": 3534},
  {"h": "辗", "p": "niǎn", "sp": "nian", "r": "車", "sc": 14, "l": 5, "f": 3535},
  {"h": "纂", "p": "zuǎn", "sp": "zuan", "r": "糸", "sc": 20, "l": 5, "f": 3536},
  {"h": "圭", "p": "guī", "sp": "gui", "r": "土", "sc": 6, "l": 5, "f": 3537},
  {"h": "幔", "p": "màn", "sp": "man", "r": "巾", "sc": 14, "l": 5, "f": 3538},
  {"h": "褒", "p": "bāo", "sp": "bao", "r": "衣", "sc": 15, "l": 5, "f": 3539},
  {"h": "揍", "p": "zòu", "sp": "zou", "r": "手", "sc": 12, "l": 5, "f": 3540},
  {"h": "诽", "p": "fěi", "sp": "fei", "r": "言", "sc": 10, "l": 5, "f": 3541},
  {"h": "倔", "p": "jué", "sp": "jue", "r": "人", "sc": 10, "l": 5, "f": 3542},
  {"h": "腓", "p": "féi", "sp": "fei", "r": "肉", "sc": 12, "l": 5, "f": 3543},
  {"h": "颉", "p": "jié", "sp": "jie", "r": "頁", "sc": 12, "l": 5, "f": 3544},
  {"h": "锄", "p": "chú", "sp": "chu", "r": "金", "sc": 12, "l": 5, "f": 3545},
  {"h": "嗔", "p": "chēn", "sp": "chen", "r": "口", "sc": 13, "l": 5, "f": 3546},
  {
    "h": "磺",
    "p": "huáng",
    "sp": "huang",
    "r": "石",
    "sc": 16,
    "l": 5,
    "f": 3547
  },
  {"h": "攒", "p": "zǎn", "sp": "zan", "r": "手", "sc": 19, "l": 5, "f": 3548},
  {"h": "瘩", "p": "da", "sp": "da", "r": "疒", "sc": 14, "l": 5, "f": 3549},
  {"h": "雳", "p": "lì", "sp": "li", "r": "雨", "sc": 12, "l": 5, "f": 3550},
  {"h": "吆", "p": "yāo", "sp": "yao", "r": "口", "sc": 6, "l": 5, "f": 3551},
  {"h": "悚", "p": "sǒng", "sp": "song", "r": "心", "sc": 10, "l": 5, "f": 3552},
  {"h": "墩", "p": "dūn", "sp": "dun", "r": "土", "sc": 15, "l": 5, "f": 3553},
  {"h": "彝", "p": "yí", "sp": "yi", "r": "彐", "sc": 18, "l": 5, "f": 3554},
  {"h": "囱", "p": "cōng", "sp": "cong", "r": "囗", "sc": 7, "l": 5, "f": 3555},
  {"h": "逍", "p": "xiāo", "sp": "xiao", "r": "辵", "sc": 10, "l": 5, "f": 3556},
  {"h": "辄", "p": "zhé", "sp": "zhe", "r": "車", "sc": 11, "l": 5, "f": 3557},
  {"h": "桅", "p": "wéi", "sp": "wei", "r": "木", "sc": 10, "l": 5, "f": 3558},
  {"h": "俨", "p": "yǎn", "sp": "yan", "r": "人", "sc": 9, "l": 5, "f": 3559},
  {"h": "纶", "p": "lún", "sp": "lun", "r": "糸", "sc": 7, "l": 5, "f": 3560},
  {"h": "悸", "p": "jì", "sp": "ji", "r": "心", "sc": 11, "l": 5, "f": 3561},
  {"h": "殃", "p": "yāng", "sp": "yang", "r": "歹", "sc": 9, "l": 5, "f": 3562},
  {"h": "帧", "p": "zhèng", "sp": "zheng", "r": "巾", "sc": 9, "l": 5, "f": 3563},
  {"h": "俐", "p": "lì", "sp": "li", "r": "人", "sc": 9, "l": 5, "f": 3564},
  {"h": "绮", "p": "qǐ", "sp": "qi", "r": "糸", "sc": 11, "l": 5, "f": 3565},
  {"h": "袒", "p": "tǎn", "sp": "tan", "r": "衣", "sc": 10, "l": 5, "f": 3566},
  {"h": "籽", "p": "zǐ", "sp": "zi", "r": "米", "sc": 9, "l": 5, "f": 3567},
  {"h": "孰", "p": "shú", "sp": "shu", "r": "子", "sc": 11, "l": 5, "f": 3568},
  {"h": "愫", "p": "sù", "sp": "su", "r": "心", "sc": 13, "l": 5, "f": 3569},
  {"h": "拌", "p": "bàn", "sp": "ban", "r": "手", "sc": 8, "l": 5, "f": 3570},
  {
    "h": "橙",
    "p": "chéng",
    "sp": "cheng",
    "r": "木",
    "sc": 16,
    "l": 5,
    "f": 3571
  },
  {"h": "暨", "p": "jì", "sp": "ji", "r": "日", "sc": 14, "l": 5, "f": 3572},
  {"h": "敖", "p": "áo", "sp": "ao", "r": "攴", "sc": 10, "l": 5, "f": 3573},
  {"h": "赘", "p": "zhuì", "sp": "zhui", "r": "貝", "sc": 14, "l": 5, "f": 3574},
  {"h": "抉", "p": "jué", "sp": "jue", "r": "手", "sc": 7, "l": 5, "f": 3575},
  {"h": "淤", "p": "yū", "sp": "yu", "r": "水", "sc": 11, "l": 5, "f": 3576},
  {"h": "剌", "p": "lá", "sp": "la", "r": "刀", "sc": 9, "l": 5, "f": 3577},
  {
    "h": "娼",
    "p": "chāng",
    "sp": "chang",
    "r": "女",
    "sc": 11,
    "l": 5,
    "f": 3578
  },
  {"h": "顼", "p": "xū", "sp": "xu", "r": "玉", "sc": 10, "l": 5, "f": 3579},
  {"h": "葵", "p": "kuí", "sp": "kui", "r": "艸", "sc": 12, "l": 5, "f": 3580},
  {"h": "哝", "p": "nóng", "sp": "nong", "r": "口", "sc": 9, "l": 5, "f": 3581},
  {"h": "酣", "p": "hān", "sp": "han", "r": "酉", "sc": 12, "l": 5, "f": 3582},
  {"h": "麓", "p": "lù", "sp": "lu", "r": "鹿", "sc": 19, "l": 5, "f": 3583},
  {"h": "钵", "p": "bō", "sp": "bo", "r": "金", "sc": 10, "l": 5, "f": 3584},
  {"h": "琅", "p": "láng", "sp": "lang", "r": "玉", "sc": 11, "l": 5, "f": 3585},
  {"h": "簸", "p": "bǒ", "sp": "bo", "r": "竹", "sc": 19, "l": 5, "f": 3586},
  {"h": "禾", "p": "hé", "sp": "he", "r": "禾", "sc": 5, "l": 5, "f": 3587},
  {"h": "铢", "p": "zhū", "sp": "zhu", "r": "金", "sc": 11, "l": 5, "f": 3588},
  {"h": "璧", "p": "bì", "sp": "bi", "r": "玉", "sc": 18, "l": 5, "f": 3589},
  {"h": "娠", "p": "shēn", "sp": "shen", "r": "女", "sc": 10, "l": 5, "f": 3590},
  {"h": "彗", "p": "huì", "sp": "hui", "r": "彐", "sc": 11, "l": 5, "f": 3591},
  {"h": "惋", "p": "wǎn", "sp": "wan", "r": "心", "sc": 11, "l": 5, "f": 3592},
  {"h": "腋", "p": "yè", "sp": "ye", "r": "肉", "sc": 12, "l": 5, "f": 3593},
  {"h": "螂", "p": "láng", "sp": "lang", "r": "虫", "sc": 14, "l": 5, "f": 3594},
  {"h": "阪", "p": "bǎn", "sp": "ban", "r": "阜", "sc": 6, "l": 5, "f": 3595},
  {"h": "掣", "p": "chè", "sp": "che", "r": "手", "sc": 12, "l": 5, "f": 3596},
  {"h": "劾", "p": "hé", "sp": "he", "r": "力", "sc": 8, "l": 5, "f": 3597},
  {"h": "沥", "p": "lì", "sp": "li", "r": "水", "sc": 7, "l": 5, "f": 3598},
  {
    "h": "粱",
    "p": "liáng",
    "sp": "liang",
    "r": "米",
    "sc": 13,
    "l": 5,
    "f": 3599
  },
  {"h": "嚓", "p": "cā", "sp": "ca", "r": "口", "sc": 17, "l": 5, "f": 3600},
  {"h": "惮", "p": "dàn", "sp": "dan", "r": "心", "sc": 11, "l": 5, "f": 3601},
  {"h": "氖", "p": "nǎi", "sp": "nai", "r": "气", "sc": 6, "l": 5, "f": 3602},
  {"h": "捎", "p": "shāo", "sp": "shao", "r": "手", "sc": 10, "l": 5, "f": 3603},
  {"h": "羔", "p": "gāo", "sp": "gao", "r": "羊", "sc": 10, "l": 5, "f": 3604},
  {"h": "俟", "p": "qí", "sp": "qi", "r": "人", "sc": 9, "l": 5, "f": 3605},
  {"h": "渲", "p": "xuàn", "sp": "xuan", "r": "水", "sc": 12, "l": 5, "f": 3606},
  {"h": "榄", "p": "lǎn", "sp": "lan", "r": "木", "sc": 13, "l": 5, "f": 3607},
  {"h": "茧", "p": "jiǎn", "sp": "jian", "r": "艸", "sc": 9, "l": 5, "f": 3608},
  {"h": "霓", "p": "ní", "sp": "ni", "r": "雨", "sc": 16, "l": 5, "f": 3609},
  {"h": "鹉", "p": "wǔ", "sp": "wu", "r": "鳥", "sc": 13, "l": 5, "f": 3610},
  {"h": "胥", "p": "xū", "sp": "xu", "r": "肉", "sc": 9, "l": 5, "f": 3611},
  {"h": "琶", "p": "pá", "sp": "pa", "r": "玉", "sc": 12, "l": 5, "f": 3612},
  {"h": "撬", "p": "qiào", "sp": "qiao", "r": "手", "sc": 15, "l": 5, "f": 3613},
  {"h": "橘", "p": "jú", "sp": "ju", "r": "木", "sc": 16, "l": 5, "f": 3614},
  {"h": "醫", "p": "yī", "sp": "yi", "r": "酉", "sc": 18, "l": 5, "f": 3615},
  {"h": "拈", "p": "niān", "sp": "nian", "r": "手", "sc": 8, "l": 5, "f": 3616},
  {"h": "笆", "p": "bā", "sp": "ba", "r": "竹", "sc": 10, "l": 5, "f": 3617},
  {"h": "痊", "p": "quán", "sp": "quan", "r": "疒", "sc": 11, "l": 5, "f": 3618},
  {"h": "亟", "p": "jí", "sp": "ji", "r": "二", "sc": 8, "l": 5, "f": 3619},
  {"h": "渭", "p": "wèi", "sp": "wei", "r": "水", "sc": 12, "l": 5, "f": 3620},
  {"h": "狙", "p": "jū", "sp": "ju", "r": "犬", "sc": 8, "l": 5, "f": 3621},
  {"h": "珂", "p": "kē", "sp": "ke", "r": "玉", "sc": 9, "l": 5, "f": 3622},
  {"h": "刨", "p": "páo", "sp": "pao", "r": "刀", "sc": 7, "l": 5, "f": 3623},
  {"h": "蜕", "p": "tuì", "sp": "tui", "r": "虫", "sc": 13, "l": 5, "f": 3624},
  {"h": "谚", "p": "yàn", "sp": "yan", "r": "言", "sc": 11, "l": 5, "f": 3625},
  {
    "h": "憧",
    "p": "chōng",
    "sp": "chong",
    "r": "心",
    "sc": 15,
    "l": 5,
    "f": 3626
  },
  {"h": "瞟", "p": "piǎo", "sp": "piao", "r": "目", "sc": 16, "l": 5, "f": 3627},
  {"h": "馒", "p": "mán", "sp": "man", "r": "食", "sc": 14, "l": 5, "f": 3628},
  {"h": "拗", "p": "ǎo", "sp": "ao", "r": "手", "sc": 8, "l": 5, "f": 3629},
  {"h": "帚", "p": "zhǒu", "sp": "zhou", "r": "巾", "sc": 8, "l": 5, "f": 3630},
  {"h": "钗", "p": "chāi", "sp": "chai", "r": "金", "sc": 8, "l": 5, "f": 3631},
  {"h": "哧", "p": "chī", "sp": "chi", "r": "口", "sc": 10, "l": 5, "f": 3632},
  {"h": "喋", "p": "dié", "sp": "die", "r": "口", "sc": 12, "l": 5, "f": 3633},
  {"h": "箫", "p": "xiāo", "sp": "xiao", "r": "竹", "sc": 14, "l": 5, "f": 3634},
  {"h": "刁", "p": "diāo", "sp": "diao", "r": "刀", "sc": 2, "l": 5, "f": 3635},
  {"h": "怦", "p": "pēng", "sp": "peng", "r": "心", "sc": 8, "l": 5, "f": 3636},
  {"h": "缭", "p": "liáo", "sp": "liao", "r": "糸", "sc": 15, "l": 5, "f": 3637},
  {"h": "迥", "p": "jiǒng", "sp": "jiong", "r": "辵", "sc": 8, "l": 4, "f": 3638},
  {"h": "湄", "p": "méi", "sp": "mei", "r": "水", "sc": 12, "l": 5, "f": 3639},
  {"h": "磐", "p": "pán", "sp": "pan", "r": "石", "sc": 15, "l": 5, "f": 3640},
  {"h": "渝", "p": "yú", "sp": "yu", "r": "水", "sc": 12, "l": 5, "f": 3641},
  {"h": "冗", "p": "rǒng", "sp": "rong", "r": "冖", "sc": 4, "l": 5, "f": 3642},
  {"h": "闵", "p": "mǐn", "sp": "min", "r": "門", "sc": 7, "l": 5, "f": 3643},
  {"h": "噶", "p": "gá", "sp": "ga", "r": "口", "sc": 15, "l": 5, "f": 3644},
  {"h": "黏", "p": "nián", "sp": "nian", "r": "黍", "sc": 17, "l": 5, "f": 3645},
  {"h": "蕃", "p": "fān", "sp": "fan", "r": "艸", "sc": 15, "l": 5, "f": 3646},
  {"h": "弼", "p": "bì", "sp": "bi", "r": "弓", "sc": 12, "l": 5, "f": 3647},
  {"h": "驿", "p": "yì", "sp": "yi", "r": "馬", "sc": 8, "l": 5, "f": 3648},
  {"h": "淄", "p": "zī", "sp": "zi", "r": "水", "sc": 11, "l": 5, "f": 3649},
  {"h": "饺", "p": "jiǎo", "sp": "jiao", "r": "食", "sc": 9, "l": 2, "f": 3650},
  {"h": "踞", "p": "jù", "sp": "ju", "r": "足", "sc": 15, "l": 5, "f": 3651},
  {"h": "韬", "p": "tāo", "sp": "tao", "r": "韋", "sc": 14, "l": 5, "f": 3652},
  {"h": "婷", "p": "tíng", "sp": "ting", "r": "女", "sc": 12, "l": 5, "f": 3653},
  {"h": "唆", "p": "suō", "sp": "suo", "r": "口", "sc": 10, "l": 5, "f": 3654},
  {"h": "蜒", "p": "yán", "sp": "yan", "r": "虫", "sc": 12, "l": 5, "f": 3655},
  {"h": "偎", "p": "wēi", "sp": "wei", "r": "人", "sc": 11, "l": 5, "f": 3656},
  {"h": "榨", "p": "zhà", "sp": "zha", "r": "木", "sc": 14, "l": 5, "f": 3657},
  {"h": "漉", "p": "lù", "sp": "lu", "r": "水", "sc": 14, "l": 5, "f": 3658},
  {"h": "碉", "p": "diāo", "sp": "diao", "r": "石", "sc": 13, "l": 5, "f": 3659},
  {"h": "皈", "p": "guī", "sp": "gui", "r": "白", "sc": 9, "l": 5, "f": 3660},
  {"h": "矜", "p": "jīn", "sp": "jin", "r": "矛", "sc": 9, "l": 5, "f": 3661},
  {"h": "笈", "p": "jí", "sp": "ji", "r": "竹", "sc": 9, "l": 5, "f": 3662},
  {"h": "枷", "p": "jiā", "sp": "jia", "r": "木", "sc": 9, "l": 5, "f": 3663},
  {"h": "鲨", "p": "shā", "sp": "sha", "r": "魚", "sc": 15, "l": 5, "f": 3664},
  {"h": "蹑", "p": "niè", "sp": "nie", "r": "足", "sc": 17, "l": 5, "f": 3665},
  {"h": "瀚", "p": "hàn", "sp": "han", "r": "水", "sc": 19, "l": 5, "f": 3666},
  {"h": "酪", "p": "lào", "sp": "lao", "r": "酉", "sc": 13, "l": 5, "f": 3667},
  {"h": "谑", "p": "xuè", "sp": "xue", "r": "言", "sc": 11, "l": 5, "f": 3668},
  {"h": "癖", "p": "pǐ", "sp": "pi", "r": "疒", "sc": 18, "l": 5, "f": 3669},
  {"h": "烬", "p": "jìn", "sp": "jin", "r": "火", "sc": 10, "l": 5, "f": 3670},
  {"h": "揩", "p": "kāi", "sp": "kai", "r": "手", "sc": 12, "l": 5, "f": 3671},
  {"h": "炙", "p": "zhì", "sp": "zhi", "r": "火", "sc": 8, "l": 5, "f": 3672},
  {"h": "蜷", "p": "quán", "sp": "quan", "r": "虫", "sc": 14, "l": 5, "f": 3673},
  {"h": "侏", "p": "zhū", "sp": "zhu", "r": "人", "sc": 8, "l": 5, "f": 3674},
  {"h": "凋", "p": "diāo", "sp": "diao", "r": "冫", "sc": 10, "l": 5, "f": 3675},
  {"h": "漪", "p": "yī", "sp": "yi", "r": "水", "sc": 14, "l": 5, "f": 3676},
  {"h": "悻", "p": "xìng", "sp": "xing", "r": "心", "sc": 11, "l": 5, "f": 3677},
  {"h": "蹋", "p": "tà", "sp": "ta", "r": "足", "sc": 17, "l": 5, "f": 3678},
  {"h": "讪", "p": "shàn", "sp": "shan", "r": "言", "sc": 5, "l": 5, "f": 3679},
  {"h": "搐", "p": "chù", "sp": "chu", "r": "手", "sc": 13, "l": 5, "f": 3680},
  {"h": "碘", "p": "diǎn", "sp": "dian", "r": "石", "sc": 13, "l": 5, "f": 3681},
  {"h": "帛", "p": "bó", "sp": "bo", "r": "巾", "sc": 8, "l": 5, "f": 3682},
  {"h": "诠", "p": "quán", "sp": "quan", "r": "言", "sc": 8, "l": 5, "f": 3683},
  {"h": "碾", "p": "niǎn", "sp": "nian", "r": "石", "sc": 15, "l": 5, "f": 3684},
  {"h": "擂", "p": "léi", "sp": "lei", "r": "手", "sc": 16, "l": 5, "f": 3685},
  {"h": "苯", "p": "běn", "sp": "ben", "r": "艸", "sc": 8, "l": 5, "f": 3686},
  {"h": "诃", "p": "hē", "sp": "he", "r": "言", "sc": 7, "l": 5, "f": 3687},
  {"h": "铎", "p": "duó", "sp": "duo", "r": "金", "sc": 10, "l": 5, "f": 3688},
  {"h": "戊", "p": "wù", "sp": "wu", "r": "戈", "sc": 5, "l": 5, "f": 3689},
  {"h": "荀", "p": "xún", "sp": "xun", "r": "艸", "sc": 9, "l": 5, "f": 3690},
  {"h": "驹", "p": "jū", "sp": "ju", "r": "馬", "sc": 8, "l": 5, "f": 3691},
  {"h": "攫", "p": "jué", "sp": "jue", "r": "手", "sc": 23, "l": 5, "f": 3692},
  {"h": "憬", "p": "jǐng", "sp": "jing", "r": "心", "sc": 15, "l": 5, "f": 3693},
  {"h": "哽", "p": "gěng", "sp": "geng", "r": "口", "sc": 10, "l": 5, "f": 3694},
  {
    "h": "踵",
    "p": "zhǒng",
    "sp": "zhong",
    "r": "足",
    "sc": 16,
    "l": 5,
    "f": 3695
  },
  {"h": "蟒", "p": "mǎng", "sp": "mang", "r": "虫", "sc": 16, "l": 5, "f": 3696},
  {"h": "漾", "p": "yàng", "sp": "yang", "r": "水", "sc": 14, "l": 5, "f": 3697},
  {"h": "啧", "p": "zé", "sp": "ze", "r": "口", "sc": 11, "l": 5, "f": 3698},
  {"h": "吮", "p": "shǔn", "sp": "shun", "r": "口", "sc": 7, "l": 5, "f": 3699},
  {"h": "楠", "p": "nán", "sp": "nan", "r": "木", "sc": 13, "l": 5, "f": 3700},
  {"h": "氟", "p": "fú", "sp": "fu", "r": "气", "sc": 9, "l": 5, "f": 3701},
  {"h": "怂", "p": "sǒng", "sp": "song", "r": "心", "sc": 8, "l": 5, "f": 3702},
  {"h": "叼", "p": "diāo", "sp": "diao", "r": "口", "sc": 5, "l": 5, "f": 3703},
  {"h": "竣", "p": "jùn", "sp": "jun", "r": "立", "sc": 12, "l": 5, "f": 3704},
  {"h": "偕", "p": "xié", "sp": "xie", "r": "人", "sc": 11, "l": 5, "f": 3705},
  {"h": "漩", "p": "xuán", "sp": "xuan", "r": "水", "sc": 14, "l": 5, "f": 3706},
  {"h": "蹭", "p": "cèng", "sp": "ceng", "r": "足", "sc": 19, "l": 5, "f": 3707},
  {"h": "翌", "p": "yì", "sp": "yi", "r": "羽", "sc": 11, "l": 5, "f": 3708},
  {"h": "臆", "p": "yì", "sp": "yi", "r": "肉", "sc": 17, "l": 5, "f": 3709},
  {"h": "挝", "p": "wō", "sp": "wo", "r": "手", "sc": 9, "l": 5, "f": 3710},
  {"h": "绚", "p": "xuàn", "sp": "xuan", "r": "糸", "sc": 9, "l": 5, "f": 3711},
  {"h": "崽", "p": "zǎi", "sp": "zai", "r": "山", "sc": 12, "l": 5, "f": 3712},
  {"h": "糜", "p": "mí", "sp": "mi", "r": "米", "sc": 17, "l": 5, "f": 3713},
  {"h": "瘢", "p": "bān", "sp": "ban", "r": "疒", "sc": 15, "l": 5, "f": 3714},
  {"h": "跤", "p": "jiāo", "sp": "jiao", "r": "足", "sc": 13, "l": 5, "f": 3715},
  {"h": "阑", "p": "lán", "sp": "lan", "r": "門", "sc": 12, "l": 5, "f": 3716},
  {"h": "恬", "p": "tián", "sp": "tian", "r": "心", "sc": 9, "l": 5, "f": 3717},
  {"h": "豢", "p": "huàn", "sp": "huan", "r": "豕", "sc": 13, "l": 5, "f": 3718},
  {"h": "汶", "p": "wèn", "sp": "wen", "r": "水", "sc": 7, "l": 5, "f": 3719},
  {"h": "跷", "p": "qiāo", "sp": "qiao", "r": "足", "sc": 13, "l": 5, "f": 3720},
  {"h": "琵", "p": "pí", "sp": "pi", "r": "玉", "sc": 12, "l": 5, "f": 3721},
  {"h": "憨", "p": "hān", "sp": "han", "r": "心", "sc": 15, "l": 5, "f": 3722},
  {"h": "蜗", "p": "wō", "sp": "wo", "r": "虫", "sc": 13, "l": 5, "f": 3723},
  {"h": "螅", "p": "xī", "sp": "xi", "r": "虫", "sc": 16, "l": 5, "f": 3724},
  {"h": "惴", "p": "zhuì", "sp": "zhui", "r": "心", "sc": 12, "l": 5, "f": 3725},
  {"h": "戟", "p": "jǐ", "sp": "ji", "r": "戈", "sc": 12, "l": 5, "f": 3726},
  {"h": "匮", "p": "kuì", "sp": "kui", "r": "匚", "sc": 11, "l": 5, "f": 3727},
  {"h": "恙", "p": "yàng", "sp": "yang", "r": "心", "sc": 10, "l": 5, "f": 3728},
  {"h": "抿", "p": "mǐn", "sp": "min", "r": "手", "sc": 8, "l": 5, "f": 3729},
  {"h": "桢", "p": "zhēn", "sp": "zhen", "r": "木", "sc": 10, "l": 5, "f": 3730},
  {"h": "笺", "p": "jiān", "sp": "jian", "r": "竹", "sc": 11, "l": 5, "f": 3731},
  {"h": "蛤", "p": "há", "sp": "ha", "r": "虫", "sc": 12, "l": 5, "f": 3732},
  {"h": "瞳", "p": "tóng", "sp": "tong", "r": "目", "sc": 17, "l": 5, "f": 3733},
  {"h": "藥", "p": "yào", "sp": "yao", "r": "艸", "sc": 18, "l": 5, "f": 3734},
  {"h": "瓢", "p": "piáo", "sp": "piao", "r": "瓜", "sc": 16, "l": 5, "f": 3735},
  {"h": "衹", "p": "zhǐ", "sp": "zhi", "r": "衣", "sc": 9, "l": 5, "f": 3736},
  {
    "h": "秤",
    "p": "chèng",
    "sp": "cheng",
    "r": "禾",
    "sc": 10,
    "l": 5,
    "f": 3737
  },
  {"h": "跺", "p": "duò", "sp": "duo", "r": "足", "sc": 13, "l": 5, "f": 3738},
  {"h": "潦", "p": "lǎo", "sp": "lao", "r": "水", "sc": 15, "l": 5, "f": 3739},
  {"h": "芹", "p": "qín", "sp": "qin", "r": "艸", "sc": 7, "l": 5, "f": 3740},
  {"h": "哒", "p": "dā", "sp": "da", "r": "口", "sc": 9, "l": 5, "f": 3741},
  {"h": "饬", "p": "chì", "sp": "chi", "r": "食", "sc": 7, "l": 5, "f": 3742},
  {"h": "栩", "p": "xǔ", "sp": "xu", "r": "木", "sc": 10, "l": 5, "f": 3743},
  {"h": "曦", "p": "xī", "sp": "xi", "r": "日", "sc": 20, "l": 5, "f": 3744},
  {"h": "骷", "p": "kū", "sp": "ku", "r": "骨", "sc": 14, "l": 5, "f": 3745},
  {"h": "嫡", "p": "dí", "sp": "di", "r": "女", "sc": 14, "l": 5, "f": 3746},
  {"h": "卤", "p": "lǔ", "sp": "lu", "r": "卜", "sc": 7, "l": 5, "f": 3747},
  {"h": "丕", "p": "pī", "sp": "pi", "r": "一", "sc": 5, "l": 5, "f": 3748},
  {"h": "鬓", "p": "bìn", "sp": "bin", "r": "髟", "sc": 20, "l": 5, "f": 3749},
  {"h": "梓", "p": "zǐ", "sp": "zi", "r": "木", "sc": 11, "l": 5, "f": 3750},
  {"h": "嗖", "p": "sōu", "sp": "sou", "r": "口", "sc": 12, "l": 5, "f": 3751},
  {"h": "惦", "p": "diàn", "sp": "dian", "r": "心", "sc": 11, "l": 5, "f": 3752},
  {"h": "浚", "p": "jùn", "sp": "jun", "r": "水", "sc": 10, "l": 5, "f": 3753},
  {"h": "咔", "p": "kā", "sp": "ka", "r": "口", "sc": 8, "l": 5, "f": 3754},
  {"h": "藐", "p": "miǎo", "sp": "miao", "r": "艸", "sc": 17, "l": 5, "f": 3755},
  {"h": "荃", "p": "quán", "sp": "quan", "r": "艸", "sc": 9, "l": 5, "f": 3756},
  {"h": "唧", "p": "jī", "sp": "ji", "r": "口", "sc": 10, "l": 5, "f": 3757},
  {"h": "玺", "p": "xǐ", "sp": "xi", "r": "玉", "sc": 10, "l": 5, "f": 3758},
  {"h": "汛", "p": "xùn", "sp": "xun", "r": "水", "sc": 6, "l": 5, "f": 3759},
  {"h": "铐", "p": "kào", "sp": "kao", "r": "金", "sc": 11, "l": 5, "f": 3760},
  {"h": "髅", "p": "lóu", "sp": "lou", "r": "骨", "sc": 18, "l": 5, "f": 3761},
  {"h": "渤", "p": "bó", "sp": "bo", "r": "水", "sc": 12, "l": 5, "f": 3762},
  {"h": "皿", "p": "mǐn", "sp": "min", "r": "皿", "sc": 5, "l": 5, "f": 3763},
  {"h": "箍", "p": "gū", "sp": "gu", "r": "竹", "sc": 14, "l": 5, "f": 3764},
  {"h": "馅", "p": "xiàn", "sp": "xian", "r": "食", "sc": 11, "l": 5, "f": 3765},
  {"h": "汾", "p": "fén", "sp": "fen", "r": "水", "sc": 7, "l": 5, "f": 3766},
  {"h": "戍", "p": "shù", "sp": "shu", "r": "戈", "sc": 6, "l": 5, "f": 3767},
  {"h": "痔", "p": "zhì", "sp": "zhi", "r": "疒", "sc": 11, "l": 5, "f": 3768},
  {"h": "褶", "p": "zhě", "sp": "zhe", "r": "衣", "sc": 16, "l": 5, "f": 3769},
  {"h": "聆", "p": "líng", "sp": "ling", "r": "耳", "sc": 11, "l": 5, "f": 3770},
  {"h": "涎", "p": "xián", "sp": "xian", "r": "水", "sc": 9, "l": 5, "f": 3771},
  {"h": "汞", "p": "gǒng", "sp": "gong", "r": "水", "sc": 7, "l": 5, "f": 3772},
  {"h": "渍", "p": "zì", "sp": "zi", "r": "水", "sc": 11, "l": 5, "f": 3773},
  {"h": "奂", "p": "huàn", "sp": "huan", "r": "大", "sc": 7, "l": 5, "f": 3774},
  {"h": "巅", "p": "diān", "sp": "dian", "r": "山", "sc": 19, "l": 5, "f": 3775},
  {"h": "疣", "p": "yóu", "sp": "you", "r": "疒", "sc": 9, "l": 5, "f": 3776},
  {"h": "傩", "p": "nuó", "sp": "nuo", "r": "人", "sc": 12, "l": 5, "f": 3777},
  {"h": "逵", "p": "kuí", "sp": "kui", "r": "辵", "sc": 11, "l": 5, "f": 3778},
  {"h": "耆", "p": "qí", "sp": "qi", "r": "老", "sc": 10, "l": 5, "f": 3779},
  {"h": "蟋", "p": "xī", "sp": "xi", "r": "虫", "sc": 17, "l": 5, "f": 3780},
  {"h": "鳄", "p": "è", "sp": "e", "r": "魚", "sc": 17, "l": 5, "f": 3781},
  {"h": "讹", "p": "é", "sp": "e", "r": "言", "sc": 6, "l": 5, "f": 3782},
  {"h": "膺", "p": "yīng", "sp": "ying", "r": "肉", "sc": 17, "l": 5, "f": 3783},
  {"h": "蹿", "p": "cuān", "sp": "cuan", "r": "足", "sc": 19, "l": 5, "f": 3784},
  {"h": "筏", "p": "fá", "sp": "fa", "r": "竹", "sc": 12, "l": 5, "f": 3785},
  {"h": "釜", "p": "fǔ", "sp": "fu", "r": "金", "sc": 10, "l": 5, "f": 3786},
  {"h": "沂", "p": "yí", "sp": "yi", "r": "水", "sc": 7, "l": 5, "f": 3787},
  {"h": "坯", "p": "pī", "sp": "pi", "r": "土", "sc": 8, "l": 5, "f": 3788},
  {"h": "峦", "p": "luán", "sp": "luan", "r": "山", "sc": 9, "l": 5, "f": 3789},
  {"h": "茬", "p": "chá", "sp": "cha", "r": "艸", "sc": 9, "l": 5, "f": 3790},
  {"h": "摒", "p": "bǐng", "sp": "bing", "r": "手", "sc": 12, "l": 5, "f": 3791},
  {
    "h": "蟀",
    "p": "shuài",
    "sp": "shuai",
    "r": "虫",
    "sc": 17,
    "l": 5,
    "f": 3792
  },
  {"h": "撵", "p": "niǎn", "sp": "nian", "r": "手", "sc": 15, "l": 5, "f": 3793},
  {"h": "浒", "p": "hǔ", "sp": "hu", "r": "水", "sc": 9, "l": 5, "f": 3794},
  {"h": "缤", "p": "bīn", "sp": "bin", "r": "糸", "sc": 13, "l": 5, "f": 3795},
  {"h": "嵋", "p": "méi", "sp": "mei", "r": "山", "sc": 12, "l": 5, "f": 3796},
  {"h": "珑", "p": "lóng", "sp": "long", "r": "玉", "sc": 9, "l": 5, "f": 3797},
  {"h": "苞", "p": "bāo", "sp": "bao", "r": "艸", "sc": 8, "l": 5, "f": 3798},
  {"h": "瑾", "p": "jǐn", "sp": "jin", "r": "玉", "sc": 15, "l": 5, "f": 3799},
  {"h": "泵", "p": "bèng", "sp": "beng", "r": "水", "sc": 9, "l": 5, "f": 3800},
  {"h": "钾", "p": "jiǎ", "sp": "jia", "r": "金", "sc": 10, "l": 5, "f": 3801},
  {"h": "暧", "p": "ài", "sp": "ai", "r": "日", "sc": 14, "l": 5, "f": 3802},
  {"h": "赓", "p": "gēng", "sp": "geng", "r": "广", "sc": 12, "l": 5, "f": 3803},
  {"h": "叟", "p": "sǒu", "sp": "sou", "r": "又", "sc": 9, "l": 5, "f": 3804},
  {"h": "佚", "p": "yì", "sp": "yi", "r": "人", "sc": 7, "l": 5, "f": 3805},
  {"h": "沓", "p": "dá", "sp": "da", "r": "水", "sc": 8, "l": 5, "f": 3806},
  {"h": "撂", "p": "liào", "sp": "liao", "r": "手", "sc": 14, "l": 5, "f": 3807},
  {"h": "蛊", "p": "gǔ", "sp": "gu", "r": "虫", "sc": 11, "l": 5, "f": 3808},
  {
    "h": "甥",
    "p": "shēng",
    "sp": "sheng",
    "r": "生",
    "sc": 12,
    "l": 5,
    "f": 3809
  },
  {"h": "璐", "p": "lù", "sp": "lu", "r": "玉", "sc": 17, "l": 5, "f": 3810},
  {"h": "晏", "p": "yàn", "sp": "yan", "r": "日", "sc": 10, "l": 5, "f": 3811},
  {"h": "瘪", "p": "biě", "sp": "bie", "r": "疒", "sc": 15, "l": 5, "f": 3812},
  {
    "h": "漳",
    "p": "zhāng",
    "sp": "zhang",
    "r": "水",
    "sc": 14,
    "l": 5,
    "f": 3813
  },
  {"h": "阉", "p": "yān", "sp": "yan", "r": "門", "sc": 11, "l": 5, "f": 3814},
  {"h": "蹂", "p": "róu", "sp": "rou", "r": "足", "sc": 16, "l": 5, "f": 3815},
  {"h": "鳃", "p": "sāi", "sp": "sai", "r": "魚", "sc": 17, "l": 5, "f": 3816},
  {"h": "琏", "p": "liǎn", "sp": "lian", "r": "玉", "sc": 11, "l": 5, "f": 3817},
  {"h": "湃", "p": "pài", "sp": "pai", "r": "水", "sc": 12, "l": 5, "f": 3818},
  {"h": "辘", "p": "lù", "sp": "lu", "r": "車", "sc": 15, "l": 5, "f": 3819},
  {"h": "僭", "p": "jiàn", "sp": "jian", "r": "人", "sc": 14, "l": 5, "f": 3820},
  {"h": "躏", "p": "lìn", "sp": "lin", "r": "足", "sc": 21, "l": 5, "f": 3821},
  {"h": "鼾", "p": "hān", "sp": "han", "r": "鼻", "sc": 17, "l": 5, "f": 3822},
  {"h": "懵", "p": "měng", "sp": "meng", "r": "心", "sc": 18, "l": 5, "f": 3823},
  {"h": "镰", "p": "lián", "sp": "lian", "r": "金", "sc": 18, "l": 5, "f": 3824},
  {"h": "寐", "p": "mèi", "sp": "mei", "r": "宀", "sc": 12, "l": 5, "f": 3825},
  {"h": "褚", "p": "chǔ", "sp": "chu", "r": "衣", "sc": 13, "l": 5, "f": 3826},
  {"h": "攥", "p": "zuàn", "sp": "zuan", "r": "手", "sc": 23, "l": 5, "f": 3827},
  {"h": "涧", "p": "jiàn", "sp": "jian", "r": "水", "sc": 10, "l": 5, "f": 3828},
  {"h": "蝙", "p": "biān", "sp": "bian", "r": "虫", "sc": 15, "l": 5, "f": 3829},
  {"h": "脐", "p": "qí", "sp": "qi", "r": "肉", "sc": 10, "l": 5, "f": 3830},
  {"h": "辕", "p": "yuán", "sp": "yuan", "r": "車", "sc": 14, "l": 5, "f": 3831},
  {"h": "涣", "p": "huàn", "sp": "huan", "r": "水", "sc": 10, "l": 5, "f": 3832},
  {"h": "杞", "p": "qǐ", "sp": "qi", "r": "木", "sc": 7, "l": 5, "f": 3833},
  {"h": "煜", "p": "yù", "sp": "yu", "r": "火", "sc": 13, "l": 5, "f": 3834},
  {"h": "骥", "p": "jì", "sp": "ji", "r": "馬", "sc": 19, "l": 5, "f": 3835},
  {"h": "傣", "p": "dǎi", "sp": "dai", "r": "人", "sc": 12, "l": 5, "f": 3836},
  {"h": "嗳", "p": "āi", "sp": "ai", "r": "口", "sc": 13, "l": 5, "f": 3837},
  {"h": "祯", "p": "zhēn", "sp": "zhen", "r": "示", "sc": 10, "l": 5, "f": 3838},
  {"h": "酉", "p": "yǒu", "sp": "you", "r": "酉", "sc": 7, "l": 5, "f": 3839},
  {"h": "秸", "p": "jiē", "sp": "jie", "r": "禾", "sc": 11, "l": 5, "f": 3840},
  {"h": "捺", "p": "nà", "sp": "na", "r": "手", "sc": 11, "l": 5, "f": 3841},
  {"h": "瑕", "p": "xiá", "sp": "xia", "r": "玉", "sc": 13, "l": 5, "f": 3842},
  {"h": "鑫", "p": "xīn", "sp": "xin", "r": "金", "sc": 24, "l": 5, "f": 3843},
  {"h": "馋", "p": "chán", "sp": "chan", "r": "食", "sc": 12, "l": 5, "f": 3844},
  {"h": "窿", "p": "lóng", "sp": "long", "r": "穴", "sc": 16, "l": 5, "f": 3845},
  {"h": "楔", "p": "xiē", "sp": "xie", "r": "木", "sc": 13, "l": 5, "f": 3846},
  {
    "h": "胱",
    "p": "guāng",
    "sp": "guang",
    "r": "肉",
    "sc": 10,
    "l": 5,
    "f": 3847
  },
  {"h": "荔", "p": "lì", "sp": "li", "r": "艸", "sc": 9, "l": 5, "f": 3848},
  {"h": "蟆", "p": "má", "sp": "ma", "r": "虫", "sc": 16, "l": 5, "f": 3849},
  {"h": "湍", "p": "tuān", "sp": "tuan", "r": "水", "sc": 12, "l": 5, "f": 3850},
  {"h": "屹", "p": "yì", "sp": "yi", "r": "山", "sc": 6, "l": 5, "f": 3851},
  {"h": "遐", "p": "xiá", "sp": "xia", "r": "辵", "sc": 12, "l": 5, "f": 3852},
  {"h": "轲", "p": "kē", "sp": "ke", "r": "車", "sc": 9, "l": 5, "f": 3853},
  {"h": "镯", "p": "zhuó", "sp": "zhuo", "r": "金", "sc": 18, "l": 5, "f": 3854},
  {
    "h": "缰",
    "p": "jiāng",
    "sp": "jiang",
    "r": "糸",
    "sc": 16,
    "l": 5,
    "f": 3855
  },
  {"h": "桦", "p": "huà", "sp": "hua", "r": "木", "sc": 10, "l": 5, "f": 3856},
  {"h": "炖", "p": "dùn", "sp": "dun", "r": "火", "sc": 8, "l": 5, "f": 3857},
  {"h": "钡", "p": "bèi", "sp": "bei", "r": "金", "sc": 9, "l": 5, "f": 3858},
  {"h": "羚", "p": "líng", "sp": "ling", "r": "羊", "sc": 11, "l": 5, "f": 3859},
  {"h": "啬", "p": "sè", "sp": "se", "r": "十", "sc": 11, "l": 5, "f": 3860},
  {"h": "诩", "p": "xǔ", "sp": "xu", "r": "言", "sc": 8, "l": 5, "f": 3861},
  {"h": "绯", "p": "fēi", "sp": "fei", "r": "糸", "sc": 11, "l": 5, "f": 3862},
  {"h": "掖", "p": "yē", "sp": "ye", "r": "手", "sc": 11, "l": 5, "f": 3863},
  {"h": "箓", "p": "lù", "sp": "lu", "r": "竹", "sc": 14, "l": 5, "f": 3864},
  {"h": "涸", "p": "hé", "sp": "he", "r": "水", "sc": 11, "l": 5, "f": 3865},
  {"h": "鸳", "p": "yuān", "sp": "yuan", "r": "鳥", "sc": 10, "l": 5, "f": 3866},
  {"h": "塾", "p": "shú", "sp": "shu", "r": "土", "sc": 14, "l": 5, "f": 3867},
  {"h": "呸", "p": "pēi", "sp": "pei", "r": "口", "sc": 8, "l": 5, "f": 3868},
  {"h": "抡", "p": "lūn", "sp": "lun", "r": "手", "sc": 7, "l": 5, "f": 3869},
  {"h": "擞", "p": "sǒu", "sp": "sou", "r": "手", "sc": 16, "l": 5, "f": 3870},
  {"h": "熹", "p": "xī", "sp": "xi", "r": "火", "sc": 16, "l": 5, "f": 3871},
  {"h": "坷", "p": "kě", "sp": "ke", "r": "土", "sc": 8, "l": 5, "f": 3872},
  {"h": "瓮", "p": "wèng", "sp": "weng", "r": "瓦", "sc": 8, "l": 5, "f": 3873},
  {"h": "亘", "p": "gèn", "sp": "gen", "r": "二", "sc": 6, "l": 5, "f": 3874},
  {"h": "嗟", "p": "jiē", "sp": "jie", "r": "口", "sc": 12, "l": 5, "f": 3875},
  {"h": "筵", "p": "yán", "sp": "yan", "r": "竹", "sc": 12, "l": 5, "f": 3876},
  {"h": "跛", "p": "bǒ", "sp": "bo", "r": "足", "sc": 12, "l": 5, "f": 3877},
  {"h": "汕", "p": "shàn", "sp": "shan", "r": "水", "sc": 6, "l": 5, "f": 3878},
  {"h": "欤", "p": "yú", "sp": "yu", "r": "欠", "sc": 7, "l": 5, "f": 3879},
  {"h": "壑", "p": "hè", "sp": "he", "r": "土", "sc": 17, "l": 5, "f": 3880},
  {"h": "颍", "p": "yǐng", "sp": "ying", "r": "水", "sc": 12, "l": 5, "f": 3881},
  {"h": "溥", "p": "pǔ", "sp": "pu", "r": "水", "sc": 13, "l": 5, "f": 3882},
  {"h": "姗", "p": "shān", "sp": "shan", "r": "女", "sc": 8, "l": 5, "f": 3883},
  {"h": "踊", "p": "yǒng", "sp": "yong", "r": "足", "sc": 14, "l": 5, "f": 3884},
  {"h": "枭", "p": "xiāo", "sp": "xiao", "r": "木", "sc": 8, "l": 5, "f": 3885},
  {"h": "暄", "p": "xuān", "sp": "xuan", "r": "日", "sc": 13, "l": 5, "f": 3886},
  {"h": "稷", "p": "jì", "sp": "ji", "r": "禾", "sc": 15, "l": 5, "f": 3887},
  {"h": "跚", "p": "shān", "sp": "shan", "r": "足", "sc": 12, "l": 5, "f": 3888},
  {"h": "涟", "p": "lián", "sp": "lian", "r": "水", "sc": 10, "l": 5, "f": 3889},
  {"h": "瀛", "p": "yíng", "sp": "ying", "r": "水", "sc": 19, "l": 5, "f": 3890},
  {
    "h": "笙",
    "p": "shēng",
    "sp": "sheng",
    "r": "竹",
    "sc": 11,
    "l": 5,
    "f": 3891
  },
  {"h": "滕", "p": "téng", "sp": "teng", "r": "水", "sc": 15, "l": 5, "f": 3892},
  {"h": "踝", "p": "huái", "sp": "huai", "r": "足", "sc": 15, "l": 5, "f": 3893},
  {"h": "贰", "p": "èr", "sp": "er", "r": "弋", "sc": 9, "l": 5, "f": 3894},
  {"h": "瞰", "p": "kàn", "sp": "kan", "r": "目", "sc": 16, "l": 5, "f": 3895},
  {"h": "恻", "p": "cè", "sp": "ce", "r": "心", "sc": 9, "l": 5, "f": 3896},
  {"h": "嚏", "p": "tì", "sp": "ti", "r": "口", "sc": 17, "l": 5, "f": 3897},
  {"h": "迢", "p": "tiáo", "sp": "tiao", "r": "辵", "sc": 8, "l": 5, "f": 3898},
  {"h": "獗", "p": "jué", "sp": "jue", "r": "犬", "sc": 15, "l": 5, "f": 3899},
  {"h": "邯", "p": "hán", "sp": "han", "r": "邑", "sc": 7, "l": 5, "f": 3900},
  {"h": "睑", "p": "jiǎn", "sp": "jian", "r": "目", "sc": 12, "l": 5, "f": 3901},
  {"h": "赡", "p": "shàn", "sp": "shan", "r": "貝", "sc": 17, "l": 5, "f": 3902},
  {"h": "萦", "p": "yíng", "sp": "ying", "r": "艸", "sc": 11, "l": 5, "f": 3903},
  {"h": "珥", "p": "ěr", "sp": "er", "r": "玉", "sc": 10, "l": 5, "f": 3904},
  {"h": "酮", "p": "tóng", "sp": "tong", "r": "酉", "sc": 13, "l": 5, "f": 3905},
  {"h": "璞", "p": "pú", "sp": "pu", "r": "玉", "sc": 16, "l": 5, "f": 3906},
  {"h": "羹", "p": "gēng", "sp": "geng", "r": "羊", "sc": 19, "l": 5, "f": 3907},
  {"h": "缄", "p": "jiān", "sp": "jian", "r": "糸", "sc": 12, "l": 5, "f": 3908},
  {
    "h": "晾",
    "p": "liàng",
    "sp": "liang",
    "r": "日",
    "sc": 12,
    "l": 5,
    "f": 3909
  },
  {"h": "俸", "p": "fèng", "sp": "feng", "r": "人", "sc": 10, "l": 5, "f": 3910},
  {"h": "媲", "p": "pì", "sp": "pi", "r": "女", "sc": 13, "l": 5, "f": 3911},
  {"h": "鸾", "p": "luán", "sp": "luan", "r": "鳥", "sc": 11, "l": 5, "f": 3912},
  {"h": "恿", "p": "yǒng", "sp": "yong", "r": "心", "sc": 11, "l": 5, "f": 3913},
  {"h": "蜿", "p": "wān", "sp": "wan", "r": "虫", "sc": 14, "l": 5, "f": 3914},
  {"h": "犊", "p": "dú", "sp": "du", "r": "牛", "sc": 12, "l": 5, "f": 3915},
  {"h": "讷", "p": "nè", "sp": "ne", "r": "言", "sc": 6, "l": 5, "f": 3916},
  {"h": "扈", "p": "hù", "sp": "hu", "r": "戶", "sc": 11, "l": 5, "f": 3917},
  {"h": "蜈", "p": "wú", "sp": "wu", "r": "虫", "sc": 13, "l": 5, "f": 3918},
  {"h": "翟", "p": "dí", "sp": "di", "r": "羽", "sc": 14, "l": 5, "f": 3919},
  {"h": "藕", "p": "ǒu", "sp": "ou", "r": "艸", "sc": 18, "l": 5, "f": 3920},
  {"h": "戌", "p": "xū", "sp": "xu", "r": "戈", "sc": 6, "l": 5, "f": 3921},
  {"h": "蓓", "p": "bèi", "sp": "bei", "r": "艸", "sc": 13, "l": 5, "f": 3922},
  {"h": "鋆", "p": "yún", "sp": "yun", "r": "金", "sc": 15, "l": 5, "f": 3923},
  {"h": "谩", "p": "mán", "sp": "man", "r": "言", "sc": 13, "l": 5, "f": 3924},
  {"h": "谀", "p": "yú", "sp": "yu", "r": "言", "sc": 10, "l": 5, "f": 3925},
  {"h": "卯", "p": "mǎo", "sp": "mao", "r": "卩", "sc": 5, "l": 5, "f": 3926},
  {"h": "谙", "p": "ān", "sp": "an", "r": "言", "sc": 11, "l": 5, "f": 3927},
  {"h": "岐", "p": "qí", "sp": "qi", "r": "山", "sc": 7, "l": 5, "f": 3928},
  {"h": "蝎", "p": "xiē", "sp": "xie", "r": "虫", "sc": 15, "l": 5, "f": 3929},
  {"h": "荼", "p": "tú", "sp": "tu", "r": "艸", "sc": 10, "l": 5, "f": 3930},
  {"h": "镀", "p": "dù", "sp": "du", "r": "金", "sc": 14, "l": 5, "f": 3931},
  {"h": "椰", "p": "yē", "sp": "ye", "r": "木", "sc": 12, "l": 5, "f": 3932},
  {"h": "甄", "p": "zhēn", "sp": "zhen", "r": "瓦", "sc": 13, "l": 5, "f": 3933},
  {"h": "蟾", "p": "chán", "sp": "chan", "r": "虫", "sc": 19, "l": 5, "f": 3934},
  {"h": "蹊", "p": "qī", "sp": "qi", "r": "足", "sc": 17, "l": 5, "f": 3935},
  {"h": "泞", "p": "nìng", "sp": "ning", "r": "水", "sc": 8, "l": 5, "f": 3936},
  {"h": "撸", "p": "lū", "sp": "lu", "r": "手", "sc": 15, "l": 5, "f": 3937},
  {"h": "螃", "p": "páng", "sp": "pang", "r": "虫", "sc": 16, "l": 5, "f": 3938},
  {"h": "檬", "p": "méng", "sp": "meng", "r": "木", "sc": 17, "l": 5, "f": 3939},
  {"h": "猓", "p": "guǒ", "sp": "guo", "r": "犬", "sc": 11, "l": 5, "f": 3940},
  {
    "h": "蔷",
    "p": "qiáng",
    "sp": "qiang",
    "r": "艸",
    "sc": 14,
    "l": 5,
    "f": 3941
  },
  {"h": "羲", "p": "xī", "sp": "xi", "r": "羊", "sc": 16, "l": 5, "f": 3942},
  {"h": "瘸", "p": "qué", "sp": "que", "r": "疒", "sc": 16, "l": 5, "f": 3943},
  {"h": "蘸", "p": "zhàn", "sp": "zhan", "r": "艸", "sc": 22, "l": 5, "f": 3944},
  {"h": "蔗", "p": "zhè", "sp": "zhe", "r": "艸", "sc": 14, "l": 5, "f": 3945},
  {"h": "傀", "p": "guī", "sp": "gui", "r": "人", "sc": 11, "l": 5, "f": 3946},
  {"h": "蚌", "p": "bàng", "sp": "bang", "r": "虫", "sc": 10, "l": 5, "f": 3947},
  {"h": "锢", "p": "gù", "sp": "gu", "r": "金", "sc": 13, "l": 5, "f": 3948},
  {"h": "遽", "p": "jù", "sp": "ju", "r": "辵", "sc": 16, "l": 5, "f": 3949},
  {"h": "邃", "p": "suì", "sp": "sui", "r": "辵", "sc": 17, "l": 5, "f": 3950},
  {"h": "恚", "p": "huì", "sp": "hui", "r": "心", "sc": 10, "l": 5, "f": 3951},
  {"h": "皑", "p": "ái", "sp": "ai", "r": "白", "sc": 11, "l": 5, "f": 3952},
  {
    "h": "锵",
    "p": "qiāng",
    "sp": "qiang",
    "r": "金",
    "sc": 14,
    "l": 5,
    "f": 3953
  },
  {"h": "簌", "p": "sù", "sp": "su", "r": "竹", "sc": 17, "l": 5, "f": 3954},
  {"h": "焙", "p": "bèi", "sp": "bei", "r": "火", "sc": 12, "l": 5, "f": 3955},
  {"h": "昊", "p": "hào", "sp": "hao", "r": "日", "sc": 8, "l": 5, "f": 3956},
  {"h": "鹳", "p": "guàn", "sp": "guan", "r": "鳥", "sc": 22, "l": 5, "f": 3957},
  {"h": "睽", "p": "kuí", "sp": "kui", "r": "目", "sc": 14, "l": 5, "f": 3958},
  {"h": "刽", "p": "guì", "sp": "gui", "r": "刀", "sc": 8, "l": 5, "f": 3959},
  {"h": "鳖", "p": "biē", "sp": "bie", "r": "魚", "sc": 19, "l": 5, "f": 3960},
  {"h": "噎", "p": "yē", "sp": "ye", "r": "口", "sc": 15, "l": 5, "f": 3961},
  {"h": "呗", "p": "bei", "sp": "bei", "r": "口", "sc": 7, "l": 5, "f": 3962},
  {"h": "寰", "p": "huán", "sp": "huan", "r": "宀", "sc": 16, "l": 5, "f": 3963},
  {"h": "唷", "p": "yō", "sp": "yo", "r": "口", "sc": 11, "l": 5, "f": 3964},
  {"h": "殡", "p": "bìn", "sp": "bin", "r": "歹", "sc": 14, "l": 5, "f": 3965},
  {"h": "淖", "p": "nào", "sp": "nao", "r": "水", "sc": 11, "l": 5, "f": 3966},
  {"h": "诰", "p": "gào", "sp": "gao", "r": "言", "sc": 9, "l": 5, "f": 3967},
  {"h": "恣", "p": "zì", "sp": "zi", "r": "心", "sc": 10, "l": 5, "f": 3968},
  {"h": "睐", "p": "lài", "sp": "lai", "r": "目", "sc": 12, "l": 5, "f": 3969},
  {"h": "婵", "p": "chán", "sp": "chan", "r": "女", "sc": 11, "l": 5, "f": 3970},
  {"h": "榈", "p": "lǘ", "sp": "lu", "r": "木", "sc": 13, "l": 5, "f": 3971},
  {"h": "氦", "p": "hài", "sp": "hai", "r": "气", "sc": 10, "l": 5, "f": 3972},
  {"h": "靳", "p": "jìn", "sp": "jin", "r": "革", "sc": 13, "l": 5, "f": 3973},
  {"h": "蛹", "p": "yǒng", "sp": "yong", "r": "虫", "sc": 13, "l": 5, "f": 3974},
  {"h": "鸯", "p": "yāng", "sp": "yang", "r": "鳥", "sc": 10, "l": 5, "f": 3975},
  {"h": "惬", "p": "qiè", "sp": "qie", "r": "心", "sc": 11, "l": 5, "f": 3976},
  {"h": "蹙", "p": "cù", "sp": "cu", "r": "足", "sc": 18, "l": 5, "f": 3977},
  {"h": "诙", "p": "huī", "sp": "hui", "r": "言", "sc": 8, "l": 5, "f": 3978},
  {"h": "眈", "p": "dān", "sp": "dan", "r": "目", "sc": 9, "l": 5, "f": 3979},
  {"h": "罡", "p": "gāng", "sp": "gang", "r": "网", "sc": 10, "l": 5, "f": 3980},
  {"h": "缮", "p": "shàn", "sp": "shan", "r": "糸", "sc": 15, "l": 5, "f": 3981},
  {"h": "胤", "p": "yìn", "sp": "yin", "r": "肉", "sc": 9, "l": 5, "f": 3982},
  {"h": "皋", "p": "gāo", "sp": "gao", "r": "白", "sc": 10, "l": 5, "f": 3983},
  {"h": "蛀", "p": "zhù", "sp": "zhu", "r": "虫", "sc": 11, "l": 5, "f": 3984},
  {"h": "偌", "p": "ruò", "sp": "ruo", "r": "人", "sc": 10, "l": 5, "f": 3985},
  {"h": "疵", "p": "cī", "sp": "ci", "r": "疒", "sc": 11, "l": 5, "f": 3986},
  {"h": "绛", "p": "jiàng", "sp": "jiang", "r": "糸", "sc": 9, "l": 5, "f": 3987},
  {"h": "葆", "p": "bǎo", "sp": "bao", "r": "艸", "sc": 12, "l": 5, "f": 3988},
  {"h": "黔", "p": "qián", "sp": "qian", "r": "黑", "sc": 16, "l": 5, "f": 3989},
  {"h": "喙", "p": "huì", "sp": "hui", "r": "口", "sc": 12, "l": 5, "f": 3990},
  {"h": "烽", "p": "fēng", "sp": "feng", "r": "火", "sc": 11, "l": 5, "f": 3991},
  {"h": "儡", "p": "lěi", "sp": "lei", "r": "人", "sc": 17, "l": 5, "f": 3992},
  {"h": "佼", "p": "jiǎo", "sp": "jiao", "r": "人", "sc": 8, "l": 5, "f": 3993},
  {"h": "斓", "p": "lán", "sp": "lan", "r": "文", "sc": 16, "l": 5, "f": 3994},
  {"h": "嫔", "p": "pín", "sp": "pin", "r": "女", "sc": 13, "l": 5, "f": 3995},
  {"h": "颚", "p": "è", "sp": "e", "r": "頁", "sc": 15, "l": 5, "f": 3996},
  {"h": "龈", "p": "kěn", "sp": "ken", "r": "齒", "sc": 14, "l": 5, "f": 3997},
  {"h": "盅", "p": "zhōng", "sp": "zhong", "r": "皿", "sc": 9, "l": 5, "f": 3998},
  {"h": "娓", "p": "wěi", "sp": "wei", "r": "女", "sc": 10, "l": 5, "f": 3999},
  {"h": "坂", "p": "bǎn", "sp": "ban", "r": "土", "sc": 7, "l": 5, "f": 4000},
  {"h": "町", "p": "tīng", "sp": "ting", "r": "田", "sc": 7, "l": 5, "f": 4001},
  {"h": "芥", "p": "jiè", "sp": "jie", "r": "艸", "sc": 7, "l": 5, "f": 4002},
  {"h": "瘠", "p": "jí", "sp": "ji", "r": "疒", "sc": 15, "l": 5, "f": 4003},
  {"h": "阂", "p": "hé", "sp": "he", "r": "門", "sc": 9, "l": 5, "f": 4004},
  {"h": "挎", "p": "kuà", "sp": "kua", "r": "手", "sc": 9, "l": 5, "f": 4005},
  {"h": "橇", "p": "qiāo", "sp": "qiao", "r": "木", "sc": 16, "l": 5, "f": 4006},
  {"h": "荟", "p": "huì", "sp": "hui", "r": "艸", "sc": 9, "l": 5, "f": 4007},
  {
    "h": "啜",
    "p": "chuài",
    "sp": "chuai",
    "r": "口",
    "sc": 11,
    "l": 5,
    "f": 4008
  },
  {"h": "垛", "p": "duǒ", "sp": "duo", "r": "土", "sc": 9, "l": 5, "f": 4009},
  {"h": "淇", "p": "qí", "sp": "qi", "r": "水", "sc": 11, "l": 5, "f": 4010},
  {"h": "瓒", "p": "zàn", "sp": "zan", "r": "玉", "sc": 20, "l": 5, "f": 4011},
  {"h": "篓", "p": "lǒu", "sp": "lou", "r": "竹", "sc": 15, "l": 5, "f": 4012},
  {"h": "虱", "p": "shī", "sp": "shi", "r": "虫", "sc": 8, "l": 5, "f": 4013},
  {"h": "跻", "p": "jī", "sp": "ji", "r": "足", "sc": 13, "l": 5, "f": 4014},
  {"h": "龛", "p": "kān", "sp": "kan", "r": "龍", "sc": 11, "l": 5, "f": 4015},
  {"h": "蹒", "p": "pán", "sp": "pan", "r": "足", "sc": 17, "l": 5, "f": 4016},
  {"h": "髯", "p": "rán", "sp": "ran", "r": "髟", "sc": 15, "l": 5, "f": 4017},
  {
    "h": "瞠",
    "p": "chēng",
    "sp": "cheng",
    "r": "目",
    "sc": 16,
    "l": 5,
    "f": 4018
  },
  {"h": "痫", "p": "xián", "sp": "xian", "r": "疒", "sc": 12, "l": 5, "f": 4019},
  {"h": "掂", "p": "diān", "sp": "dian", "r": "手", "sc": 11, "l": 5, "f": 4020},
  {"h": "潼", "p": "tóng", "sp": "tong", "r": "水", "sc": 15, "l": 5, "f": 4021},
  {"h": "酰", "p": "xiān", "sp": "xian", "r": "酉", "sc": 13, "l": 5, "f": 4022},
  {"h": "镁", "p": "měi", "sp": "mei", "r": "金", "sc": 14, "l": 5, "f": 4023},
  {"h": "灸", "p": "jiǔ", "sp": "jiu", "r": "火", "sc": 7, "l": 5, "f": 4024},
  {"h": "腆", "p": "tiǎn", "sp": "tian", "r": "肉", "sc": 12, "l": 5, "f": 4025},
  {"h": "筱", "p": "xiǎo", "sp": "xiao", "r": "竹", "sc": 13, "l": 5, "f": 4026},
  {"h": "谆", "p": "zhūn", "sp": "zhun", "r": "言", "sc": 10, "l": 5, "f": 4027},
  {
    "h": "骋",
    "p": "chěng",
    "sp": "cheng",
    "r": "馬",
    "sc": 10,
    "l": 5,
    "f": 4028
  },
  {"h": "壬", "p": "rén", "sp": "ren", "r": "士", "sc": 4, "l": 5, "f": 4029},
  {"h": "茗", "p": "míng", "sp": "ming", "r": "艸", "sc": 9, "l": 5, "f": 4030},
  {
    "h": "椋",
    "p": "liáng",
    "sp": "liang",
    "r": "木",
    "sc": 12,
    "l": 5,
    "f": 4031
  },
  {"h": "蛔", "p": "huí", "sp": "hui", "r": "虫", "sc": 12, "l": 5, "f": 4032},
  {"h": "潺", "p": "chán", "sp": "chan", "r": "水", "sc": 15, "l": 5, "f": 4033},
  {"h": "扉", "p": "fēi", "sp": "fei", "r": "戶", "sc": 12, "l": 5, "f": 4034},
  {"h": "耘", "p": "yún", "sp": "yun", "r": "耒", "sc": 10, "l": 5, "f": 4035},
  {"h": "槟", "p": "bīn", "sp": "bin", "r": "木", "sc": 14, "l": 5, "f": 4036},
  {"h": "雹", "p": "báo", "sp": "bao", "r": "雨", "sc": 13, "l": 5, "f": 4037},
  {"h": "甬", "p": "yǒng", "sp": "yong", "r": "用", "sc": 7, "l": 5, "f": 4038},
  {"h": "谥", "p": "shì", "sp": "shi", "r": "言", "sc": 12, "l": 5, "f": 4039},
  {"h": "淞", "p": "sōng", "sp": "song", "r": "水", "sc": 11, "l": 5, "f": 4040},
  {"h": "燎", "p": "liáo", "sp": "liao", "r": "火", "sc": 16, "l": 5, "f": 4041},
  {"h": "蕙", "p": "huì", "sp": "hui", "r": "艸", "sc": 15, "l": 5, "f": 4042},
  {"h": "蚪", "p": "dǒu", "sp": "dou", "r": "虫", "sc": 10, "l": 5, "f": 4043},
  {"h": "蜻", "p": "qīng", "sp": "qing", "r": "虫", "sc": 14, "l": 5, "f": 4044},
  {"h": "郸", "p": "dān", "sp": "dan", "r": "邑", "sc": 10, "l": 5, "f": 4045},
  {"h": "轶", "p": "yì", "sp": "yi", "r": "車", "sc": 9, "l": 5, "f": 4046},
  {"h": "狰", "p": "zhēng", "sp": "zheng", "r": "犬", "sc": 9, "l": 5, "f": 4047},
  {"h": "楣", "p": "méi", "sp": "mei", "r": "木", "sc": 13, "l": 5, "f": 4048},
  {"h": "捋", "p": "lǚ", "sp": "lu", "r": "手", "sc": 10, "l": 5, "f": 4049},
  {"h": "涓", "p": "juān", "sp": "juan", "r": "水", "sc": 10, "l": 5, "f": 4050},
  {"h": "荪", "p": "sūn", "sp": "sun", "r": "艸", "sc": 9, "l": 5, "f": 4051},
  {"h": "娄", "p": "lóu", "sp": "lou", "r": "女", "sc": 9, "l": 5, "f": 4052},
  {"h": "麝", "p": "shè", "sp": "she", "r": "鹿", "sc": 21, "l": 5, "f": 4053},
  {"h": "蚤", "p": "zǎo", "sp": "zao", "r": "虫", "sc": 9, "l": 5, "f": 4054},
  {"h": "薰", "p": "xūn", "sp": "xun", "r": "艸", "sc": 17, "l": 5, "f": 4055},
  {"h": "醮", "p": "jiào", "sp": "jiao", "r": "酉", "sc": 19, "l": 5, "f": 4056},
  {"h": "搪", "p": "táng", "sp": "tang", "r": "手", "sc": 13, "l": 5, "f": 4057},
  {"h": "谧", "p": "mì", "sp": "mi", "r": "言", "sc": 12, "l": 5, "f": 4058},
  {"h": "湮", "p": "yān", "sp": "yan", "r": "水", "sc": 12, "l": 5, "f": 4059},
  {"h": "辍", "p": "chuò", "sp": "chuo", "r": "車", "sc": 12, "l": 5, "f": 4060},
  {"h": "瞌", "p": "kē", "sp": "ke", "r": "目", "sc": 15, "l": 5, "f": 4061},
  {"h": "梆", "p": "bāng", "sp": "bang", "r": "木", "sc": 10, "l": 5, "f": 4062},
  {
    "h": "樟",
    "p": "zhāng",
    "sp": "zhang",
    "r": "木",
    "sc": 15,
    "l": 5,
    "f": 4063
  },
  {"h": "茉", "p": "mò", "sp": "mo", "r": "艸", "sc": 8, "l": 5, "f": 4064},
  {"h": "岖", "p": "qū", "sp": "qu", "r": "山", "sc": 7, "l": 5, "f": 4065},
  {"h": "臼", "p": "jiù", "sp": "jiu", "r": "臼", "sc": 6, "l": 5, "f": 4066},
  {"h": "癣", "p": "xuǎn", "sp": "xuan", "r": "疒", "sc": 19, "l": 5, "f": 4067},
  {"h": "穑", "p": "sè", "sp": "se", "r": "禾", "sc": 16, "l": 5, "f": 4068},
  {"h": "玷", "p": "diàn", "sp": "dian", "r": "玉", "sc": 9, "l": 5, "f": 4069},
  {"h": "馍", "p": "mó", "sp": "mo", "r": "食", "sc": 13, "l": 5, "f": 4070},
  {"h": "呷", "p": "gā", "sp": "ga", "r": "口", "sc": 8, "l": 5, "f": 4071},
  {"h": "萼", "p": "è", "sp": "e", "r": "艸", "sc": 12, "l": 5, "f": 4072},
  {"h": "妩", "p": "wǔ", "sp": "wu", "r": "女", "sc": 7, "l": 5, "f": 4073},
  {"h": "伫", "p": "zhù", "sp": "zhu", "r": "人", "sc": 6, "l": 5, "f": 4074},
  {"h": "彤", "p": "tóng", "sp": "tong", "r": "彡", "sc": 7, "l": 5, "f": 4075},
  {"h": "莓", "p": "méi", "sp": "mei", "r": "艸", "sc": 10, "l": 5, "f": 4076},
  {"h": "岬", "p": "jiǎ", "sp": "jia", "r": "山", "sc": 8, "l": 5, "f": 4077},
  {"h": "媛", "p": "yuàn", "sp": "yuan", "r": "女", "sc": 12, "l": 5, "f": 4078},
  {"h": "惆", "p": "chóu", "sp": "chou", "r": "心", "sc": 11, "l": 5, "f": 4079},
  {"h": "鳎", "p": "tǎ", "sp": "ta", "r": "魚", "sc": 18, "l": 5, "f": 4080},
  {"h": "啾", "p": "jiū", "sp": "jiu", "r": "口", "sc": 12, "l": 5, "f": 4081},
  {"h": "囔", "p": "nāng", "sp": "nang", "r": "口", "sc": 25, "l": 5, "f": 4082},
  {"h": "蜓", "p": "tíng", "sp": "ting", "r": "虫", "sc": 12, "l": 5, "f": 4083},
  {"h": "孺", "p": "rú", "sp": "ru", "r": "子", "sc": 17, "l": 5, "f": 4084},
  {"h": "徇", "p": "xùn", "sp": "xun", "r": "彳", "sc": 9, "l": 5, "f": 4085},
  {
    "h": "徵",
    "p": "zhēng",
    "sp": "zheng",
    "r": "彳",
    "sc": 15,
    "l": 5,
    "f": 4086
  },
  {"h": "焊", "p": "hàn", "sp": "han", "r": "火", "sc": 11, "l": 5, "f": 4087},
  {"h": "岱", "p": "dài", "sp": "dai", "r": "山", "sc": 8, "l": 5, "f": 4088},
  {"h": "昵", "p": "nì", "sp": "ni", "r": "日", "sc": 9, "l": 5, "f": 4089},
  {"h": "卅", "p": "sà", "sp": "sa", "r": "十", "sc": 4, "l": 5, "f": 4090},
  {"h": "飙", "p": "biāo", "sp": "biao", "r": "風", "sc": 16, "l": 5, "f": 4091},
  {"h": "邙", "p": "máng", "sp": "mang", "r": "邑", "sc": 5, "l": 5, "f": 4092},
  {"h": "痞", "p": "pǐ", "sp": "pi", "r": "疒", "sc": 12, "l": 5, "f": 4093},
  {"h": "隼", "p": "sǔn", "sp": "sun", "r": "隹", "sc": 10, "l": 5, "f": 4094},
  {"h": "恫", "p": "dòng", "sp": "dong", "r": "心", "sc": 9, "l": 5, "f": 4095},
  {
    "h": "怆",
    "p": "chuàng",
    "sp": "chuang",
    "r": "心",
    "sc": 7,
    "l": 5,
    "f": 4096
  },
  {"h": "桀", "p": "jié", "sp": "jie", "r": "木", "sc": 10, "l": 5, "f": 4097},
  {"h": "绶", "p": "shòu", "sp": "shou", "r": "糸", "sc": 11, "l": 5, "f": 4098},
  {"h": "裆", "p": "dāng", "sp": "dang", "r": "衣", "sc": 11, "l": 5, "f": 4099},
  {"h": "盂", "p": "yú", "sp": "yu", "r": "皿", "sc": 8, "l": 5, "f": 4100},
  {"h": "桧", "p": "guì", "sp": "gui", "r": "木", "sc": 10, "l": 5, "f": 4101},
  {"h": "蚓", "p": "yǐn", "sp": "yin", "r": "虫", "sc": 10, "l": 5, "f": 4102},
  {"h": "抠", "p": "kōu", "sp": "kou", "r": "手", "sc": 7, "l": 5, "f": 4103},
  {"h": "嗷", "p": "áo", "sp": "ao", "r": "口", "sc": 13, "l": 5, "f": 4104},
  {"h": "槌", "p": "chuí", "sp": "chui", "r": "木", "sc": 13, "l": 5, "f": 4105},
  {"h": "痘", "p": "dòu", "sp": "dou", "r": "疒", "sc": 12, "l": 5, "f": 4106},
  {"h": "痢", "p": "lì", "sp": "li", "r": "疒", "sc": 12, "l": 5, "f": 4107},
  {"h": "芮", "p": "ruì", "sp": "rui", "r": "艸", "sc": 7, "l": 5, "f": 4108},
  {"h": "蚣", "p": "gōng", "sp": "gong", "r": "虫", "sc": 10, "l": 5, "f": 4109},
  {"h": "闩", "p": "shuān", "sp": "shuan", "r": "門", "sc": 4, "l": 5, "f": 4110},
  {"h": "铿", "p": "kēng", "sp": "keng", "r": "金", "sc": 12, "l": 5, "f": 4111},
  {"h": "飓", "p": "jù", "sp": "ju", "r": "風", "sc": 12, "l": 5, "f": 4112},
  {"h": "疱", "p": "pào", "sp": "pao", "r": "疒", "sc": 10, "l": 5, "f": 4113},
  {"h": "蝌", "p": "kē", "sp": "ke", "r": "虫", "sc": 15, "l": 5, "f": 4114},
  {"h": "撅", "p": "juē", "sp": "jue", "r": "手", "sc": 15, "l": 5, "f": 4115},
  {"h": "蚯", "p": "qiū", "sp": "qiu", "r": "虫", "sc": 11, "l": 5, "f": 4116},
  {"h": "斡", "p": "wò", "sp": "wo", "r": "斗", "sc": 14, "l": 5, "f": 4117},
  {"h": "窠", "p": "kē", "sp": "ke", "r": "穴", "sc": 13, "l": 5, "f": 4118},
  {"h": "荚", "p": "jiá", "sp": "jia", "r": "艸", "sc": 9, "l": 5, "f": 4119},
  {"h": "耷", "p": "dā", "sp": "da", "r": "耳", "sc": 9, "l": 5, "f": 4120},
  {"h": "砚", "p": "yàn", "sp": "yan", "r": "石", "sc": 9, "l": 5, "f": 4121},
  {"h": "牒", "p": "dié", "sp": "die", "r": "片", "sc": 13, "l": 5, "f": 4122},
  {"h": "赈", "p": "zhèn", "sp": "zhen", "r": "貝", "sc": 11, "l": 5, "f": 4123},
  {"h": "煦", "p": "xù", "sp": "xu", "r": "火", "sc": 13, "l": 5, "f": 4124},
  {"h": "嗫", "p": "niè", "sp": "nie", "r": "口", "sc": 13, "l": 5, "f": 4125},
  {"h": "耙", "p": "bà", "sp": "ba", "r": "耒", "sc": 10, "l": 5, "f": 4126},
  {"h": "榕", "p": "róng", "sp": "rong", "r": "木", "sc": 14, "l": 5, "f": 4127},
  {"h": "鞑", "p": "dá", "sp": "da", "r": "革", "sc": 15, "l": 5, "f": 4128},
  {"h": "袤", "p": "mào", "sp": "mao", "r": "衣", "sc": 11, "l": 5, "f": 4129},
  {"h": "谌", "p": "chén", "sp": "chen", "r": "言", "sc": 11, "l": 5, "f": 4130},
  {"h": "醺", "p": "xūn", "sp": "xun", "r": "酉", "sc": 21, "l": 5, "f": 4131},
  {"h": "秆", "p": "gǎn", "sp": "gan", "r": "禾", "sc": 8, "l": 5, "f": 4132},
  {
    "h": "徨",
    "p": "huáng",
    "sp": "huang",
    "r": "彳",
    "sc": 12,
    "l": 5,
    "f": 4133
  },
  {"h": "橹", "p": "lǔ", "sp": "lu", "r": "木", "sc": 16, "l": 5, "f": 4134},
  {"h": "翡", "p": "fěi", "sp": "fei", "r": "羽", "sc": 14, "l": 5, "f": 4135},
  {"h": "缨", "p": "yīng", "sp": "ying", "r": "糸", "sc": 14, "l": 5, "f": 4136},
  {"h": "锹", "p": "qiāo", "sp": "qiao", "r": "金", "sc": 14, "l": 5, "f": 4137},
  {"h": "嵇", "p": "jī", "sp": "ji", "r": "山", "sc": 12, "l": 5, "f": 4138},
  {"h": "圪", "p": "gē", "sp": "ge", "r": "土", "sc": 6, "l": 5, "f": 4139},
  {"h": "髻", "p": "jì", "sp": "ji", "r": "髟", "sc": 16, "l": 5, "f": 4140},
  {"h": "嗬", "p": "hē", "sp": "he", "r": "口", "sc": 13, "l": 5, "f": 4141},
  {"h": "辎", "p": "zī", "sp": "zi", "r": "車", "sc": 12, "l": 5, "f": 4142},
  {"h": "痣", "p": "zhì", "sp": "zhi", "r": "疒", "sc": 12, "l": 5, "f": 4143},
  {"h": "娩", "p": "miǎn", "sp": "mian", "r": "女", "sc": 10, "l": 5, "f": 4144},
  {"h": "谄", "p": "chǎn", "sp": "chan", "r": "言", "sc": 10, "l": 5, "f": 4145},
  {"h": "蛐", "p": "qū", "sp": "qu", "r": "虫", "sc": 12, "l": 5, "f": 4146},
  {"h": "鹞", "p": "yào", "sp": "yao", "r": "鳥", "sc": 15, "l": 5, "f": 4147},
  {"h": "翱", "p": "áo", "sp": "ao", "r": "羽", "sc": 16, "l": 5, "f": 4148},
  {"h": "庖", "p": "páo", "sp": "pao", "r": "广", "sc": 8, "l": 5, "f": 4149},
  {"h": "籁", "p": "lài", "sp": "lai", "r": "竹", "sc": 19, "l": 5, "f": 4150},
  {"h": "蓿", "p": "xu", "sp": "xu", "r": "艸", "sc": 14, "l": 5, "f": 4151},
  {"h": "鳗", "p": "mán", "sp": "man", "r": "魚", "sc": 19, "l": 5, "f": 4152},
  {"h": "疟", "p": "nüè", "sp": "nue", "r": "疒", "sc": 8, "l": 5, "f": 4153},
  {"h": "鲇", "p": "nián", "sp": "nian", "r": "魚", "sc": 13, "l": 5, "f": 4154},
  {"h": "這", "p": "zhè", "sp": "zhe", "r": "辵", "sc": 10, "l": 5, "f": 4155},
  {"h": "嚅", "p": "rú", "sp": "ru", "r": "口", "sc": 17, "l": 5, "f": 4156},
  {"h": "瘀", "p": "yū", "sp": "yu", "r": "疒", "sc": 13, "l": 5, "f": 4157},
  {"h": "颔", "p": "hàn", "sp": "han", "r": "頁", "sc": 13, "l": 5, "f": 4158},
  {"h": "黜", "p": "chù", "sp": "chu", "r": "黑", "sc": 17, "l": 5, "f": 4159},
  {"h": "黠", "p": "xiá", "sp": "xia", "r": "黑", "sc": 18, "l": 5, "f": 4160},
  {"h": "濑", "p": "lài", "sp": "lai", "r": "水", "sc": 16, "l": 5, "f": 4161},
  {"h": "馁", "p": "něi", "sp": "nei", "r": "食", "sc": 10, "l": 5, "f": 4162},
  {"h": "洵", "p": "xún", "sp": "xun", "r": "水", "sc": 9, "l": 5, "f": 4163},
  {"h": "忐", "p": "tǎn", "sp": "tan", "r": "心", "sc": 7, "l": 5, "f": 4164},
  {"h": "忑", "p": "tè", "sp": "te", "r": "心", "sc": 7, "l": 5, "f": 4165},
  {"h": "砥", "p": "dǐ", "sp": "di", "r": "石", "sc": 10, "l": 5, "f": 4166},
  {"h": "咂", "p": "zā", "sp": "za", "r": "口", "sc": 8, "l": 5, "f": 4167},
  {"h": "罹", "p": "lí", "sp": "li", "r": "网", "sc": 16, "l": 5, "f": 4168},
  {"h": "糠", "p": "kāng", "sp": "kang", "r": "米", "sc": 17, "l": 5, "f": 4169},
  {"h": "匝", "p": "zā", "sp": "za", "r": "匚", "sc": 5, "l": 5, "f": 4170},
  {"h": "偃", "p": "yǎn", "sp": "yan", "r": "人", "sc": 11, "l": 5, "f": 4171},
  {"h": "淙", "p": "cóng", "sp": "cong", "r": "水", "sc": 11, "l": 5, "f": 4172},
  {"h": "纫", "p": "rèn", "sp": "ren", "r": "糸", "sc": 6, "l": 5, "f": 4173},
  {"h": "喏", "p": "nuò", "sp": "nuo", "r": "口", "sc": 11, "l": 5, "f": 4174},
  {"h": "闾", "p": "lǘ", "sp": "lu", "r": "門", "sc": 9, "l": 5, "f": 4175},
  {"h": "祛", "p": "qū", "sp": "qu", "r": "示", "sc": 9, "l": 5, "f": 4176},
  {"h": "蛰", "p": "zhé", "sp": "zhe", "r": "虫", "sc": 12, "l": 5, "f": 4177},
  {"h": "腼", "p": "miǎn", "sp": "mian", "r": "肉", "sc": 13, "l": 5, "f": 4178},
  {"h": "涝", "p": "lào", "sp": "lao", "r": "水", "sc": 10, "l": 5, "f": 4179},
  {"h": "曜", "p": "yào", "sp": "yao", "r": "日", "sc": 18, "l": 5, "f": 4180},
  {"h": "厩", "p": "jiù", "sp": "jiu", "r": "厂", "sc": 11, "l": 5, "f": 4181},
  {"h": "疽", "p": "jū", "sp": "ju", "r": "疒", "sc": 10, "l": 5, "f": 4182},
  {"h": "闰", "p": "rùn", "sp": "run", "r": "門", "sc": 7, "l": 5, "f": 4183},
  {"h": "洄", "p": "huí", "sp": "hui", "r": "水", "sc": 9, "l": 5, "f": 4184},
  {"h": "煊", "p": "xuān", "sp": "xuan", "r": "火", "sc": 13, "l": 5, "f": 4185},
  {"h": "汐", "p": "xī", "sp": "xi", "r": "水", "sc": 6, "l": 5, "f": 4186},
  {"h": "藓", "p": "xiǎn", "sp": "xian", "r": "艸", "sc": 17, "l": 5, "f": 4187},
  {
    "h": "璜",
    "p": "huáng",
    "sp": "huang",
    "r": "玉",
    "sc": 15,
    "l": 5,
    "f": 4188
  },
  {"h": "铬", "p": "gè", "sp": "ge", "r": "金", "sc": 11, "l": 5, "f": 4189},
  {"h": "經", "p": "jīng", "sp": "jing", "r": "糸", "sc": 13, "l": 5, "f": 4190},
  {"h": "渥", "p": "wò", "sp": "wo", "r": "水", "sc": 12, "l": 5, "f": 4191},
  {"h": "靼", "p": "dá", "sp": "da", "r": "革", "sc": 14, "l": 5, "f": 4192},
  {"h": "酗", "p": "xù", "sp": "xu", "r": "酉", "sc": 11, "l": 5, "f": 4193},
  {"h": "苓", "p": "líng", "sp": "ling", "r": "艸", "sc": 8, "l": 5, "f": 4194},
  {"h": "噤", "p": "jìn", "sp": "jin", "r": "口", "sc": 16, "l": 5, "f": 4195},
  {"h": "咫", "p": "zhǐ", "sp": "zhi", "r": "口", "sc": 9, "l": 5, "f": 4196},
  {"h": "椿", "p": "chūn", "sp": "chun", "r": "木", "sc": 13, "l": 5, "f": 4197},
  {"h": "鲫", "p": "jì", "sp": "ji", "r": "魚", "sc": 15, "l": 5, "f": 4198},
  {"h": "锭", "p": "dìng", "sp": "ding", "r": "金", "sc": 13, "l": 5, "f": 4199},
  {"h": "罔", "p": "wǎng", "sp": "wang", "r": "网", "sc": 8, "l": 5, "f": 4200},
  {
    "h": "锺",
    "p": "zhōng",
    "sp": "zhong",
    "r": "金",
    "sc": 14,
    "l": 5,
    "f": 4201
  },
  {"h": "匍", "p": "pú", "sp": "pu", "r": "勹", "sc": 9, "l": 5, "f": 4202},
  {"h": "祗", "p": "zhī", "sp": "zhi", "r": "示", "sc": 9, "l": 5, "f": 4203},
  {"h": "锰", "p": "měng", "sp": "meng", "r": "金", "sc": 13, "l": 5, "f": 4204},
  {"h": "岌", "p": "jí", "sp": "ji", "r": "山", "sc": 6, "l": 5, "f": 4205},
  {"h": "馀", "p": "yú", "sp": "yu", "r": "食", "sc": 10, "l": 5, "f": 4206},
  {"h": "畹", "p": "wǎn", "sp": "wan", "r": "田", "sc": 13, "l": 5, "f": 4207},
  {"h": "糯", "p": "nuò", "sp": "nuo", "r": "米", "sc": 20, "l": 5, "f": 4208},
  {"h": "胫", "p": "jìng", "sp": "jing", "r": "肉", "sc": 9, "l": 5, "f": 4209},
  {"h": "熠", "p": "yì", "sp": "yi", "r": "火", "sc": 15, "l": 5, "f": 4210},
  {"h": "銮", "p": "luán", "sp": "luan", "r": "金", "sc": 14, "l": 5, "f": 4211},
  {"h": "沅", "p": "yuán", "sp": "yuan", "r": "水", "sc": 7, "l": 5, "f": 4212},
  {"h": "棣", "p": "dì", "sp": "di", "r": "木", "sc": 12, "l": 5, "f": 4213},
  {"h": "旌", "p": "jīng", "sp": "jing", "r": "方", "sc": 11, "l": 5, "f": 4214},
  {"h": "豌", "p": "wān", "sp": "wan", "r": "豆", "sc": 15, "l": 5, "f": 4215},
  {"h": "孢", "p": "bāo", "sp": "bao", "r": "子", "sc": 8, "l": 5, "f": 4216},
  {"h": "镭", "p": "léi", "sp": "lei", "r": "金", "sc": 18, "l": 5, "f": 4217},
  {"h": "驸", "p": "fù", "sp": "fu", "r": "馬", "sc": 8, "l": 5, "f": 4218},
  {"h": "腌", "p": "yān", "sp": "yan", "r": "肉", "sc": 12, "l": 5, "f": 4219},
  {"h": "盹", "p": "dǔn", "sp": "dun", "r": "目", "sc": 9, "l": 5, "f": 4220},
  {
    "h": "熵",
    "p": "shāng",
    "sp": "shang",
    "r": "火",
    "sc": 15,
    "l": 5,
    "f": 4221
  },
  {"h": "镐", "p": "gǎo", "sp": "gao", "r": "金", "sc": 15, "l": 5, "f": 4222},
  {"h": "馐", "p": "xiū", "sp": "xiu", "r": "食", "sc": 13, "l": 5, "f": 4223},
  {"h": "嘤", "p": "yīng", "sp": "ying", "r": "口", "sc": 14, "l": 5, "f": 4224},
  {"h": "癞", "p": "lài", "sp": "lai", "r": "疒", "sc": 18, "l": 5, "f": 4225},
  {"h": "骰", "p": "tóu", "sp": "tou", "r": "骨", "sc": 13, "l": 5, "f": 4226},
  {"h": "韭", "p": "jiǔ", "sp": "jiu", "r": "韭", "sc": 9, "l": 5, "f": 4227},
  {"h": "阖", "p": "hé", "sp": "he", "r": "門", "sc": 13, "l": 5, "f": 4228},
  {"h": "瞑", "p": "míng", "sp": "ming", "r": "目", "sc": 15, "l": 5, "f": 4229},
  {"h": "裨", "p": "bì", "sp": "bi", "r": "衣", "sc": 13, "l": 5, "f": 4230},
  {"h": "宕", "p": "dàng", "sp": "dang", "r": "宀", "sc": 8, "l": 5, "f": 4231},
  {"h": "戾", "p": "lì", "sp": "li", "r": "戶", "sc": 8, "l": 5, "f": 4232},
  {"h": "镌", "p": "juān", "sp": "juan", "r": "金", "sc": 15, "l": 5, "f": 4233},
  {"h": "溟", "p": "míng", "sp": "ming", "r": "水", "sc": 13, "l": 5, "f": 4234},
  {"h": "牍", "p": "dú", "sp": "du", "r": "片", "sc": 12, "l": 5, "f": 4235},
  {"h": "隽", "p": "juàn", "sp": "juan", "r": "隹", "sc": 10, "l": 5, "f": 4236},
  {"h": "婊", "p": "biǎo", "sp": "biao", "r": "女", "sc": 11, "l": 5, "f": 4237},
  {"h": "鹄", "p": "gǔ", "sp": "gu", "r": "鳥", "sc": 12, "l": 5, "f": 4238},
  {"h": "埂", "p": "gěng", "sp": "geng", "r": "土", "sc": 10, "l": 5, "f": 4239},
  {"h": "拄", "p": "zhǔ", "sp": "zhu", "r": "手", "sc": 8, "l": 5, "f": 4240},
  {"h": "娲", "p": "wā", "sp": "wa", "r": "女", "sc": 10, "l": 5, "f": 4241},
  {"h": "虬", "p": "qiú", "sp": "qiu", "r": "虫", "sc": 7, "l": 5, "f": 4242},
  {"h": "萱", "p": "xuān", "sp": "xuan", "r": "艸", "sc": 12, "l": 5, "f": 4243},
  {"h": "啵", "p": "bo", "sp": "bo", "r": "口", "sc": 11, "l": 5, "f": 4244},
  {"h": "蠡", "p": "lí", "sp": "li", "r": "虫", "sc": 21, "l": 5, "f": 4245},
  {"h": "芋", "p": "yù", "sp": "yu", "r": "艸", "sc": 6, "l": 5, "f": 4246},
  {"h": "胭", "p": "yān", "sp": "yan", "r": "肉", "sc": 10, "l": 5, "f": 4247},
  {"h": "豺", "p": "chái", "sp": "chai", "r": "豸", "sc": 10, "l": 5, "f": 4248},
  {"h": "啻", "p": "chì", "sp": "chi", "r": "口", "sc": 12, "l": 5, "f": 4249},
  {"h": "褛", "p": "lǚ", "sp": "lu", "r": "衣", "sc": 14, "l": 5, "f": 4250},
  {"h": "蛆", "p": "qū", "sp": "qu", "r": "虫", "sc": 11, "l": 5, "f": 4251},
  {"h": "柠", "p": "níng", "sp": "ning", "r": "木", "sc": 9, "l": 5, "f": 4252},
  {"h": "掰", "p": "bāi", "sp": "bai", "r": "手", "sc": 12, "l": 5, "f": 4253},
  {
    "h": "篆",
    "p": "zhuàn",
    "sp": "zhuan",
    "r": "竹",
    "sc": 15,
    "l": 5,
    "f": 4254
  },
  {"h": "倌", "p": "guān", "sp": "guan", "r": "人", "sc": 10, "l": 5, "f": 4255},
  {"h": "咛", "p": "níng", "sp": "ning", "r": "口", "sc": 8, "l": 5, "f": 4256},
  {"h": "蛭", "p": "zhì", "sp": "zhi", "r": "虫", "sc": 12, "l": 5, "f": 4257},
  {"h": "谡", "p": "sù", "sp": "su", "r": "言", "sc": 12, "l": 5, "f": 4258},
  {"h": "荨", "p": "xún", "sp": "xun", "r": "艸", "sc": 9, "l": 5, "f": 4259},
  {"h": "莞", "p": "guǎn", "sp": "guan", "r": "艸", "sc": 10, "l": 5, "f": 4260},
  {"h": "澹", "p": "dàn", "sp": "dan", "r": "水", "sc": 16, "l": 5, "f": 4261},
  {"h": "纭", "p": "yún", "sp": "yun", "r": "糸", "sc": 7, "l": 5, "f": 4262},
  {"h": "潞", "p": "lù", "sp": "lu", "r": "水", "sc": 16, "l": 5, "f": 4263},
  {"h": "郅", "p": "zhì", "sp": "zhi", "r": "邑", "sc": 8, "l": 5, "f": 4264},
  {"h": "弋", "p": "yì", "sp": "yi", "r": "弋", "sc": 3, "l": 5, "f": 4265},
  {"h": "飕", "p": "sōu", "sp": "sou", "r": "風", "sc": 13, "l": 5, "f": 4266},
  {"h": "螳", "p": "táng", "sp": "tang", "r": "虫", "sc": 17, "l": 5, "f": 4267},
  {"h": "胄", "p": "zhòu", "sp": "zhou", "r": "肉", "sc": 9, "l": 5, "f": 4268},
  {
    "h": "蟑",
    "p": "zhāng",
    "sp": "zhang",
    "r": "虫",
    "sc": 17,
    "l": 5,
    "f": 4269
  },
  {"h": "猥", "p": "wěi", "sp": "wei", "r": "犬", "sc": 12, "l": 5, "f": 4270},
  {"h": "宓", "p": "mì", "sp": "mi", "r": "宀", "sc": 8, "l": 5, "f": 4271},
  {"h": "昙", "p": "tán", "sp": "tan", "r": "日", "sc": 8, "l": 5, "f": 4272},
  {"h": "锏", "p": "jiǎn", "sp": "jian", "r": "金", "sc": 12, "l": 5, "f": 4273},
  {"h": "蟠", "p": "pán", "sp": "pan", "r": "虫", "sc": 18, "l": 5, "f": 4274},
  {"h": "過", "p": "guò", "sp": "guo", "r": "辵", "sc": 11, "l": 5, "f": 4275},
  {"h": "柑", "p": "gān", "sp": "gan", "r": "木", "sc": 9, "l": 5, "f": 4276},
  {"h": "烯", "p": "xī", "sp": "xi", "r": "火", "sc": 11, "l": 5, "f": 4277},
  {"h": "匐", "p": "fú", "sp": "fu", "r": "勹", "sc": 11, "l": 5, "f": 4278},
  {"h": "濮", "p": "pú", "sp": "pu", "r": "水", "sc": 17, "l": 5, "f": 4279},
  {"h": "蟮", "p": "shàn", "sp": "shan", "r": "虫", "sc": 18, "l": 5, "f": 4280},
  {"h": "祐", "p": "yòu", "sp": "you", "r": "示", "sc": 9, "l": 5, "f": 4281},
  {"h": "仄", "p": "zè", "sp": "ze", "r": "人", "sc": 4, "l": 5, "f": 4282},
  {"h": "偈", "p": "jì", "sp": "ji", "r": "人", "sc": 11, "l": 5, "f": 4283},
  {"h": "蜃", "p": "shèn", "sp": "shen", "r": "虫", "sc": 13, "l": 5, "f": 4284},
  {"h": "箴", "p": "zhēn", "sp": "zhen", "r": "竹", "sc": 15, "l": 5, "f": 4285},
  {"h": "粼", "p": "lín", "sp": "lin", "r": "米", "sc": 14, "l": 5, "f": 4286},
  {"h": "嗥", "p": "háo", "sp": "hao", "r": "口", "sc": 13, "l": 5, "f": 4287},
  {"h": "褴", "p": "lán", "sp": "lan", "r": "衣", "sc": 15, "l": 5, "f": 4288},
  {"h": "蕨", "p": "jué", "sp": "jue", "r": "艸", "sc": 15, "l": 5, "f": 4289},
  {"h": "蓟", "p": "jì", "sp": "ji", "r": "艸", "sc": 13, "l": 5, "f": 4290},
  {"h": "圩", "p": "wéi", "sp": "wei", "r": "土", "sc": 6, "l": 5, "f": 4291},
  {"h": "孪", "p": "luán", "sp": "luan", "r": "子", "sc": 9, "l": 5, "f": 4292},
  {"h": "杳", "p": "yǎo", "sp": "yao", "r": "木", "sc": 8, "l": 5, "f": 4293},
  {"h": "魇", "p": "yǎn", "sp": "yan", "r": "鬼", "sc": 15, "l": 5, "f": 4294},
  {"h": "荤", "p": "hūn", "sp": "hun", "r": "艸", "sc": 9, "l": 5, "f": 4295},
  {"h": "诿", "p": "wěi", "sp": "wei", "r": "言", "sc": 10, "l": 5, "f": 4296},
  {"h": "簪", "p": "zān", "sp": "zan", "r": "竹", "sc": 18, "l": 5, "f": 4297},
  {"h": "氲", "p": "yūn", "sp": "yun", "r": "气", "sc": 13, "l": 5, "f": 4298},
  {"h": "摞", "p": "luò", "sp": "luo", "r": "手", "sc": 14, "l": 5, "f": 4299},
  {"h": "飒", "p": "sà", "sp": "sa", "r": "風", "sc": 9, "l": 5, "f": 4300},
  {"h": "镂", "p": "lòu", "sp": "lou", "r": "金", "sc": 14, "l": 5, "f": 4301},
  {"h": "舀", "p": "yǎo", "sp": "yao", "r": "臼", "sc": 10, "l": 5, "f": 4302},
  {"h": "夙", "p": "sù", "sp": "su", "r": "夕", "sc": 6, "l": 5, "f": 4303},
  {"h": "臧", "p": "zāng", "sp": "zang", "r": "臣", "sc": 14, "l": 5, "f": 4304},
  {"h": "蒿", "p": "hāo", "sp": "hao", "r": "艸", "sc": 13, "l": 5, "f": 4305},
  {"h": "貂", "p": "diāo", "sp": "diao", "r": "豸", "sc": 12, "l": 5, "f": 4306},
  {"h": "蜥", "p": "xī", "sp": "xi", "r": "虫", "sc": 14, "l": 5, "f": 4307},
  {"h": "蹩", "p": "bié", "sp": "bie", "r": "足", "sc": 18, "l": 5, "f": 4308},
  {"h": "噼", "p": "pī", "sp": "pi", "r": "口", "sc": 16, "l": 5, "f": 4309},
  {"h": "钛", "p": "tài", "sp": "tai", "r": "金", "sc": 9, "l": 5, "f": 4310},
  {"h": "钚", "p": "bù", "sp": "bu", "r": "金", "sc": 9, "l": 5, "f": 4311},
  {"h": "獾", "p": "huān", "sp": "huan", "r": "犬", "sc": 20, "l": 5, "f": 4312},
  {"h": "濂", "p": "lián", "sp": "lian", "r": "水", "sc": 16, "l": 5, "f": 4313},
  {"h": "铠", "p": "kǎi", "sp": "kai", "r": "金", "sc": 11, "l": 5, "f": 4314},
  {"h": "皙", "p": "xī", "sp": "xi", "r": "白", "sc": 13, "l": 5, "f": 4315},
  {"h": "霭", "p": "ǎi", "sp": "ai", "r": "雨", "sc": 19, "l": 5, "f": 4316},
  {"h": "鲈", "p": "lú", "sp": "lu", "r": "魚", "sc": 13, "l": 5, "f": 4317},
  {"h": "叵", "p": "pǒ", "sp": "po", "r": "口", "sc": 5, "l": 5, "f": 4318},
  {"h": "霾", "p": "mái", "sp": "mai", "r": "雨", "sc": 22, "l": 5, "f": 4319},
  {"h": "泯", "p": "mǐn", "sp": "min", "r": "水", "sc": 8, "l": 5, "f": 4320},
  {"h": "碴", "p": "chá", "sp": "cha", "r": "石", "sc": 14, "l": 5, "f": 4321},
  {"h": "鸵", "p": "tuó", "sp": "tuo", "r": "鳥", "sc": 10, "l": 5, "f": 4322},
  {"h": "峪", "p": "yù", "sp": "yu", "r": "山", "sc": 10, "l": 5, "f": 4323},
  {"h": "饕", "p": "tāo", "sp": "tao", "r": "食", "sc": 22, "l": 5, "f": 4324},
  {"h": "瘁", "p": "cuì", "sp": "cui", "r": "疒", "sc": 13, "l": 5, "f": 4325},
  {"h": "睢", "p": "suī", "sp": "sui", "r": "目", "sc": 13, "l": 5, "f": 4326},
  {"h": "鬃", "p": "zōng", "sp": "zong", "r": "髟", "sc": 18, "l": 5, "f": 4327},
  {"h": "迩", "p": "ěr", "sp": "er", "r": "辵", "sc": 8, "l": 5, "f": 4328},
  {"h": "纣", "p": "zhòu", "sp": "zhou", "r": "糸", "sc": 6, "l": 5, "f": 4329},
  {"h": "夔", "p": "kuí", "sp": "kui", "r": "夊", "sc": 21, "l": 5, "f": 4330},
  {"h": "垠", "p": "yín", "sp": "yin", "r": "土", "sc": 9, "l": 5, "f": 4331},
  {"h": "饨", "p": "tún", "sp": "tun", "r": "食", "sc": 7, "l": 5, "f": 4332},
  {"h": "榭", "p": "xiè", "sp": "xie", "r": "木", "sc": 14, "l": 5, "f": 4333},
  {
    "h": "隍",
    "p": "huáng",
    "sp": "huang",
    "r": "阜",
    "sc": 11,
    "l": 5,
    "f": 4334
  },
  {"h": "娑", "p": "suō", "sp": "suo", "r": "女", "sc": 10, "l": 5, "f": 4335},
  {"h": "篝", "p": "gōu", "sp": "gou", "r": "竹", "sc": 16, "l": 5, "f": 4336},
  {"h": "榔", "p": "láng", "sp": "lang", "r": "木", "sc": 12, "l": 5, "f": 4337},
  {"h": "洌", "p": "liè", "sp": "lie", "r": "水", "sc": 9, "l": 5, "f": 4338},
  {"h": "浜", "p": "bāng", "sp": "bang", "r": "水", "sc": 10, "l": 5, "f": 4339},
  {"h": "鲑", "p": "guī", "sp": "gui", "r": "魚", "sc": 14, "l": 5, "f": 4340},
  {"h": "谔", "p": "è", "sp": "e", "r": "言", "sc": 11, "l": 5, "f": 4341},
  {"h": "汩", "p": "gǔ", "sp": "gu", "r": "水", "sc": 7, "l": 5, "f": 4342},
  {"h": "浣", "p": "huàn", "sp": "huan", "r": "水", "sc": 10, "l": 5, "f": 4343},
  {"h": "舐", "p": "shì", "sp": "shi", "r": "舌", "sc": 10, "l": 5, "f": 4344},
  {"h": "瞭", "p": "liǎo", "sp": "liao", "r": "目", "sc": 17, "l": 5, "f": 4345},
  {"h": "忻", "p": "xīn", "sp": "xin", "r": "心", "sc": 7, "l": 5, "f": 4346},
  {"h": "咻", "p": "xiū", "sp": "xiu", "r": "口", "sc": 9, "l": 5, "f": 4347},
  {"h": "鹑", "p": "chún", "sp": "chun", "r": "鳥", "sc": 13, "l": 5, "f": 4348},
  {"h": "唑", "p": "zuò", "sp": "zuo", "r": "口", "sc": 10, "l": 5, "f": 4349},
  {"h": "懋", "p": "mào", "sp": "mao", "r": "心", "sc": 17, "l": 5, "f": 4350},
  {"h": "皎", "p": "jiǎo", "sp": "jiao", "r": "白", "sc": 11, "l": 5, "f": 4351},
  {"h": "诒", "p": "yí", "sp": "yi", "r": "言", "sc": 7, "l": 5, "f": 4352},
  {"h": "麾", "p": "huī", "sp": "hui", "r": "麻", "sc": 15, "l": 5, "f": 4353},
  {"h": "辏", "p": "còu", "sp": "cou", "r": "車", "sc": 13, "l": 5, "f": 4354},
  {"h": "氐", "p": "dī", "sp": "di", "r": "氏", "sc": 5, "l": 5, "f": 4355},
  {"h": "冽", "p": "liè", "sp": "lie", "r": "冫", "sc": 8, "l": 5, "f": 4356},
  {"h": "箕", "p": "jī", "sp": "ji", "r": "竹", "sc": 14, "l": 5, "f": 4357},
  {"h": "俚", "p": "lǐ", "sp": "li", "r": "人", "sc": 9, "l": 5, "f": 4358},
  {"h": "汴", "p": "biàn", "sp": "bian", "r": "水", "sc": 7, "l": 5, "f": 4359},
  {"h": "宸", "p": "chén", "sp": "chen", "r": "宀", "sc": 10, "l": 5, "f": 4360},
  {"h": "芍", "p": "sháo", "sp": "shao", "r": "艸", "sc": 6, "l": 5, "f": 4361},
  {"h": "捱", "p": "ái", "sp": "ai", "r": "手", "sc": 11, "l": 5, "f": 4362},
  {"h": "摈", "p": "bìn", "sp": "bin", "r": "手", "sc": 13, "l": 5, "f": 4363},
  {"h": "摺", "p": "zhé", "sp": "zhe", "r": "手", "sc": 14, "l": 5, "f": 4364},
  {"h": "簦", "p": "dēng", "sp": "deng", "r": "竹", "sc": 18, "l": 5, "f": 4365},
  {"h": "箔", "p": "bó", "sp": "bo", "r": "竹", "sc": 14, "l": 5, "f": 4366},
  {"h": "咝", "p": "sī", "sp": "si", "r": "口", "sc": 8, "l": 5, "f": 4367},
  {
    "h": "孀",
    "p": "shuāng",
    "sp": "shuang",
    "r": "女",
    "sc": 20,
    "l": 5,
    "f": 4368
  },
  {"h": "怏", "p": "yàng", "sp": "yang", "r": "心", "sc": 8, "l": 5, "f": 4369},
  {"h": "谝", "p": "piǎn", "sp": "pian", "r": "言", "sc": 11, "l": 5, "f": 4370},
  {"h": "砧", "p": "zhēn", "sp": "zhen", "r": "石", "sc": 10, "l": 5, "f": 4371},
  {"h": "馕", "p": "náng", "sp": "nang", "r": "食", "sc": 25, "l": 5, "f": 4372},
  {"h": "耄", "p": "mào", "sp": "mao", "r": "老", "sc": 10, "l": 5, "f": 4373},
  {"h": "罂", "p": "yīng", "sp": "ying", "r": "缶", "sc": 14, "l": 5, "f": 4374},
  {"h": "漕", "p": "cáo", "sp": "cao", "r": "水", "sc": 14, "l": 5, "f": 4375},
  {"h": "沣", "p": "fēng", "sp": "feng", "r": "水", "sc": 7, "l": 5, "f": 4376},
  {"h": "栾", "p": "luán", "sp": "luan", "r": "木", "sc": 10, "l": 5, "f": 4377},
  {"h": "榘", "p": "jǔ", "sp": "ju", "r": "木", "sc": 13, "l": 5, "f": 4378},
  {"h": "烷", "p": "wán", "sp": "wan", "r": "火", "sc": 11, "l": 5, "f": 4379},
  {"h": "榷", "p": "què", "sp": "que", "r": "木", "sc": 14, "l": 5, "f": 4380},
  {"h": "俑", "p": "yǒng", "sp": "yong", "r": "人", "sc": 9, "l": 5, "f": 4381},
  {"h": "沱", "p": "tuó", "sp": "tuo", "r": "水", "sc": 8, "l": 5, "f": 4382},
  {"h": "缜", "p": "zhěn", "sp": "zhen", "r": "糸", "sc": 13, "l": 5, "f": 4383},
  {"h": "鹫", "p": "jiù", "sp": "jiu", "r": "鳥", "sc": 17, "l": 5, "f": 4384},
  {"h": "蛳", "p": "sī", "sp": "si", "r": "虫", "sc": 12, "l": 5, "f": 4385},
  {"h": "剽", "p": "piāo", "sp": "piao", "r": "刀", "sc": 13, "l": 5, "f": 4386},
  {"h": "衢", "p": "qú", "sp": "qu", "r": "行", "sc": 24, "l": 5, "f": 4387},
  {"h": "泗", "p": "sì", "sp": "si", "r": "水", "sc": 8, "l": 5, "f": 4388},
  {"h": "臊", "p": "sāo", "sp": "sao", "r": "肉", "sc": 17, "l": 5, "f": 4389},
  {
    "h": "瘴",
    "p": "zhàng",
    "sp": "zhang",
    "r": "疒",
    "sc": 16,
    "l": 5,
    "f": 4390
  },
  {"h": "酚", "p": "fēn", "sp": "fen", "r": "酉", "sc": 11, "l": 5, "f": 4391},
  {"h": "纾", "p": "shū", "sp": "shu", "r": "糸", "sc": 7, "l": 5, "f": 4392},
  {"h": "晁", "p": "cháo", "sp": "chao", "r": "日", "sc": 10, "l": 5, "f": 4393},
  {"h": "孛", "p": "bèi", "sp": "bei", "r": "子", "sc": 7, "l": 5, "f": 4394},
  {"h": "炀", "p": "yáng", "sp": "yang", "r": "火", "sc": 7, "l": 5, "f": 4395},
  {"h": "叁", "p": "sān", "sp": "san", "r": "厶", "sc": 8, "l": 5, "f": 4396},
  {"h": "憩", "p": "qì", "sp": "qi", "r": "心", "sc": 16, "l": 5, "f": 4397},
  {"h": "掬", "p": "jū", "sp": "ju", "r": "手", "sc": 11, "l": 5, "f": 4398},
  {"h": "椤", "p": "luó", "sp": "luo", "r": "木", "sc": 12, "l": 5, "f": 4399},
  {"h": "啮", "p": "niè", "sp": "nie", "r": "口", "sc": 11, "l": 5, "f": 4400},
  {"h": "畿", "p": "jī", "sp": "ji", "r": "田", "sc": 15, "l": 5, "f": 4401},
  {"h": "掸", "p": "dǎn", "sp": "dan", "r": "手", "sc": 11, "l": 5, "f": 4402},
  {"h": "镣", "p": "liào", "sp": "liao", "r": "金", "sc": 17, "l": 5, "f": 4403},
  {"h": "骁", "p": "xiāo", "sp": "xiao", "r": "馬", "sc": 9, "l": 5, "f": 4404},
  {
    "h": "椽",
    "p": "chuán",
    "sp": "chuan",
    "r": "木",
    "sc": 13,
    "l": 5,
    "f": 4405
  },
  {"h": "侗", "p": "dòng", "sp": "dong", "r": "人", "sc": 8, "l": 5, "f": 4406},
  {"h": "滦", "p": "luán", "sp": "luan", "r": "水", "sc": 13, "l": 5, "f": 4407},
  {"h": "荩", "p": "jìn", "sp": "jin", "r": "艸", "sc": 9, "l": 5, "f": 4408},
  {"h": "泓", "p": "hóng", "sp": "hong", "r": "水", "sc": 8, "l": 5, "f": 4409},
  {"h": "蚱", "p": "zhà", "sp": "zha", "r": "虫", "sc": 11, "l": 5, "f": 4410},
  {"h": "癜", "p": "diàn", "sp": "dian", "r": "疒", "sc": 18, "l": 5, "f": 4411},
  {"h": "酯", "p": "zhǐ", "sp": "zhi", "r": "酉", "sc": 13, "l": 5, "f": 4412},
  {"h": "體", "p": "tǐ", "sp": "ti", "r": "骨", "sc": 22, "l": 5, "f": 4413},
  {"h": "癸", "p": "guǐ", "sp": "gui", "r": "癶", "sc": 9, "l": 5, "f": 4414},
  {"h": "蚜", "p": "yá", "sp": "ya", "r": "虫", "sc": 10, "l": 5, "f": 4415},
  {"h": "扪", "p": "mén", "sp": "men", "r": "手", "sc": 6, "l": 5, "f": 4416},
  {"h": "庑", "p": "wǔ", "sp": "wu", "r": "广", "sc": 7, "l": 5, "f": 4417},
  {"h": "歆", "p": "xīn", "sp": "xin", "r": "欠", "sc": 13, "l": 5, "f": 4418},
  {"h": "蝮", "p": "fù", "sp": "fu", "r": "虫", "sc": 15, "l": 5, "f": 4419},
  {"h": "蹶", "p": "jué", "sp": "jue", "r": "足", "sc": 19, "l": 5, "f": 4420},
  {"h": "弈", "p": "yì", "sp": "yi", "r": "廾", "sc": 9, "l": 5, "f": 4421},
  {"h": "庋", "p": "guǐ", "sp": "gui", "r": "广", "sc": 7, "l": 5, "f": 4422},
  {"h": "喟", "p": "kuì", "sp": "kui", "r": "口", "sc": 12, "l": 5, "f": 4423},
  {"h": "滂", "p": "pāng", "sp": "pang", "r": "水", "sc": 13, "l": 5, "f": 4424},
  {"h": "啕", "p": "táo", "sp": "tao", "r": "口", "sc": 11, "l": 5, "f": 4425},
  {"h": "蛎", "p": "lì", "sp": "li", "r": "虫", "sc": 11, "l": 5, "f": 4426},
  {"h": "獭", "p": "tǎ", "sp": "ta", "r": "犬", "sc": 16, "l": 5, "f": 4427},
  {"h": "槁", "p": "gǎo", "sp": "gao", "r": "木", "sc": 14, "l": 5, "f": 4428},
  {"h": "翊", "p": "yì", "sp": "yi", "r": "羽", "sc": 11, "l": 5, "f": 4429},
  {"h": "龊", "p": "chuò", "sp": "chuo", "r": "齒", "sc": 15, "l": 5, "f": 4430},
  {"h": "邺", "p": "yè", "sp": "ye", "r": "邑", "sc": 7, "l": 5, "f": 4431},
  {"h": "莘", "p": "shēn", "sp": "shen", "r": "艸", "sc": 10, "l": 5, "f": 4432},
  {"h": "燮", "p": "xiè", "sp": "xie", "r": "火", "sc": 17, "l": 5, "f": 4433},
  {"h": "剁", "p": "duò", "sp": "duo", "r": "刀", "sc": 8, "l": 5, "f": 4434},
  {"h": "觐", "p": "jìn", "sp": "jin", "r": "見", "sc": 15, "l": 5, "f": 4435},
  {"h": "铛", "p": "dāng", "sp": "dang", "r": "金", "sc": 11, "l": 5, "f": 4436},
  {"h": "谗", "p": "chán", "sp": "chan", "r": "言", "sc": 11, "l": 5, "f": 4437},
  {"h": "镍", "p": "niè", "sp": "nie", "r": "金", "sc": 15, "l": 5, "f": 4438},
  {"h": "臃", "p": "yōng", "sp": "yong", "r": "肉", "sc": 17, "l": 5, "f": 4439},
  {
    "h": "墒",
    "p": "shāng",
    "sp": "shang",
    "r": "土",
    "sc": 14,
    "l": 5,
    "f": 4440
  },
  {"h": "晔", "p": "yè", "sp": "ye", "r": "日", "sc": 10, "l": 5, "f": 4441},
  {"h": "燔", "p": "fán", "sp": "fan", "r": "火", "sc": 16, "l": 5, "f": 4442},
  {"h": "嘭", "p": "pēng", "sp": "peng", "r": "口", "sc": 15, "l": 5, "f": 4443},
  {"h": "涿", "p": "zhuō", "sp": "zhuo", "r": "水", "sc": 11, "l": 5, "f": 4444},
  {"h": "醯", "p": "xī", "sp": "xi", "r": "酉", "sc": 19, "l": 5, "f": 4445},
  {"h": "箩", "p": "luó", "sp": "luo", "r": "竹", "sc": 14, "l": 5, "f": 4446},
  {"h": "鄱", "p": "pó", "sp": "po", "r": "邑", "sc": 14, "l": 5, "f": 4447},
  {"h": "睨", "p": "nì", "sp": "ni", "r": "目", "sc": 13, "l": 5, "f": 4448},
  {"h": "诤", "p": "zhēng", "sp": "zheng", "r": "言", "sc": 8, "l": 5, "f": 4449},
  {"h": "坳", "p": "ào", "sp": "ao", "r": "土", "sc": 8, "l": 5, "f": 4450},
  {"h": "鹭", "p": "lù", "sp": "lu", "r": "鳥", "sc": 18, "l": 5, "f": 4451},
  {"h": "砷", "p": "shēn", "sp": "shen", "r": "石", "sc": 10, "l": 5, "f": 4452},
  {"h": "唏", "p": "xī", "sp": "xi", "r": "口", "sc": 10, "l": 5, "f": 4453},
  {"h": "伲", "p": "nì", "sp": "ni", "r": "人", "sc": 7, "l": 5, "f": 4454},
  {"h": "猬", "p": "wèi", "sp": "wei", "r": "犬", "sc": 12, "l": 5, "f": 4455},
  {"h": "琥", "p": "hǔ", "sp": "hu", "r": "玉", "sc": 12, "l": 5, "f": 4456},
  {"h": "殁", "p": "mò", "sp": "mo", "r": "歹", "sc": 8, "l": 5, "f": 4457},
  {"h": "蚩", "p": "chī", "sp": "chi", "r": "虫", "sc": 10, "l": 5, "f": 4458},
  {"h": "泾", "p": "jīng", "sp": "jing", "r": "水", "sc": 8, "l": 5, "f": 4459},
  {"h": "缥", "p": "piāo", "sp": "piao", "r": "糸", "sc": 14, "l": 5, "f": 4460},
  {"h": "殓", "p": "liàn", "sp": "lian", "r": "歹", "sc": 11, "l": 5, "f": 4461},
  {"h": "鳅", "p": "qiū", "sp": "qiu", "r": "魚", "sc": 17, "l": 5, "f": 4462},
  {"h": "氰", "p": "qíng", "sp": "qing", "r": "气", "sc": 12, "l": 5, "f": 4463},
  {"h": "诋", "p": "dǐ", "sp": "di", "r": "言", "sc": 7, "l": 5, "f": 4464},
  {"h": "刍", "p": "chú", "sp": "chu", "r": "彐", "sc": 5, "l": 5, "f": 4465},
  {"h": "芷", "p": "zhǐ", "sp": "zhi", "r": "艸", "sc": 7, "l": 5, "f": 4466},
  {"h": "嶙", "p": "lín", "sp": "lin", "r": "山", "sc": 15, "l": 5, "f": 4467},
  {"h": "逅", "p": "hòu", "sp": "hou", "r": "辵", "sc": 9, "l": 5, "f": 4468},
  {"h": "舫", "p": "fǎng", "sp": "fang", "r": "舟", "sc": 10, "l": 5, "f": 4469},
  {"h": "呓", "p": "yì", "sp": "yi", "r": "口", "sc": 7, "l": 5, "f": 4470},
  {"h": "唰", "p": "shuā", "sp": "shua", "r": "口", "sc": 11, "l": 5, "f": 4471},
  {"h": "茁", "p": "zhuó", "sp": "zhuo", "r": "艸", "sc": 8, "l": 5, "f": 4472},
  {"h": "馑", "p": "jǐn", "sp": "jin", "r": "食", "sc": 14, "l": 5, "f": 4473},
  {"h": "妫", "p": "guī", "sp": "gui", "r": "女", "sc": 7, "l": 5, "f": 4474},
  {
    "h": "骧",
    "p": "xiāng",
    "sp": "xiang",
    "r": "馬",
    "sc": 20,
    "l": 5,
    "f": 4475
  },
  {"h": "苷", "p": "gān", "sp": "gan", "r": "艸", "sc": 8, "l": 5, "f": 4476},
  {"h": "擢", "p": "zhuó", "sp": "zhuo", "r": "手", "sc": 17, "l": 5, "f": 4477},
  {"h": "峋", "p": "xún", "sp": "xun", "r": "山", "sc": 9, "l": 5, "f": 4478},
  {"h": "袂", "p": "mèi", "sp": "mei", "r": "衣", "sc": 9, "l": 5, "f": 4479},
  {"h": "懑", "p": "mèn", "sp": "men", "r": "心", "sc": 17, "l": 5, "f": 4480},
  {"h": "蓑", "p": "suō", "sp": "suo", "r": "艸", "sc": 13, "l": 5, "f": 4481},
  {"h": "與", "p": "yǔ", "sp": "yu", "r": "臼", "sc": 13, "l": 5, "f": 4482},
  {"h": "涞", "p": "lái", "sp": "lai", "r": "水", "sc": 10, "l": 5, "f": 4483},
  {"h": "祉", "p": "zhǐ", "sp": "zhi", "r": "示", "sc": 8, "l": 5, "f": 4484},
  {
    "h": "踹",
    "p": "chuài",
    "sp": "chuai",
    "r": "足",
    "sc": 16,
    "l": 5,
    "f": 4485
  },
  {"h": "掇", "p": "duō", "sp": "duo", "r": "手", "sc": 11, "l": 5, "f": 4486},
  {"h": "沏", "p": "qī", "sp": "qi", "r": "水", "sc": 7, "l": 5, "f": 4487},
  {"h": "诳", "p": "kuáng", "sp": "kuang", "r": "言", "sc": 9, "l": 5, "f": 4488},
  {"h": "噫", "p": "yī", "sp": "yi", "r": "口", "sc": 16, "l": 5, "f": 4489},
  {"h": "饽", "p": "bō", "sp": "bo", "r": "食", "sc": 10, "l": 5, "f": 4490},
  {"h": "饪", "p": "rèn", "sp": "ren", "r": "食", "sc": 7, "l": 5, "f": 4491},
  {"h": "绺", "p": "liǔ", "sp": "liu", "r": "糸", "sc": 11, "l": 5, "f": 4492},
  {"h": "谘", "p": "zī", "sp": "zi", "r": "言", "sc": 11, "l": 5, "f": 4493},
  {"h": "飧", "p": "sūn", "sp": "sun", "r": "食", "sc": 12, "l": 5, "f": 4494},
  {"h": "迳", "p": "jìng", "sp": "jing", "r": "辵", "sc": 8, "l": 5, "f": 4495},
  {"h": "铡", "p": "zhá", "sp": "zha", "r": "金", "sc": 11, "l": 5, "f": 4496},
  {"h": "枞", "p": "cōng", "sp": "cong", "r": "木", "sc": 8, "l": 5, "f": 4497},
  {"h": "熨", "p": "yùn", "sp": "yun", "r": "火", "sc": 15, "l": 5, "f": 4498},
  {"h": "鋈", "p": "wù", "sp": "wu", "r": "金", "sc": 15, "l": 5, "f": 4499},
  {"h": "荭", "p": "hóng", "sp": "hong", "r": "艸", "sc": 9, "l": 5, "f": 4500},
  {"h": "赊", "p": "shē", "sp": "she", "r": "貝", "sc": 11, "l": 5, "f": 4501},
  {"h": "俦", "p": "chóu", "sp": "chou", "r": "人", "sc": 9, "l": 5, "f": 4502},
  {"h": "戛", "p": "jiá", "sp": "jia", "r": "戈", "sc": 11, "l": 5, "f": 4503},
  {"h": "湎", "p": "miǎn", "sp": "mian", "r": "水", "sc": 12, "l": 5, "f": 4504},
  {"h": "幺", "p": "yāo", "sp": "yao", "r": "幺", "sc": 3, "l": 5, "f": 4505},
  {"h": "凇", "p": "sōng", "sp": "song", "r": "冫", "sc": 10, "l": 5, "f": 4506},
  {"h": "芪", "p": "qí", "sp": "qi", "r": "艸", "sc": 7, "l": 5, "f": 4507},
  {"h": "觯", "p": "zhì", "sp": "zhi", "r": "角", "sc": 15, "l": 5, "f": 4508},
  {"h": "龌", "p": "wò", "sp": "wo", "r": "齒", "sc": 17, "l": 5, "f": 4509},
  {"h": "挞", "p": "tà", "sp": "ta", "r": "手", "sc": 9, "l": 5, "f": 4510},
  {"h": "嬴", "p": "yíng", "sp": "ying", "r": "女", "sc": 16, "l": 5, "f": 4511},
  {"h": "苻", "p": "fú", "sp": "fu", "r": "艸", "sc": 8, "l": 5, "f": 4512},
  {"h": "嘁", "p": "qī", "sp": "qi", "r": "口", "sc": 14, "l": 5, "f": 4513},
  {"h": "鞯", "p": "jiān", "sp": "jian", "r": "革", "sc": 18, "l": 5, "f": 4514},
  {"h": "肽", "p": "tài", "sp": "tai", "r": "肉", "sc": 8, "l": 5, "f": 4515},
  {"h": "恸", "p": "tòng", "sp": "tong", "r": "心", "sc": 9, "l": 5, "f": 4516},
  {"h": "迨", "p": "dài", "sp": "dai", "r": "辵", "sc": 8, "l": 5, "f": 4517},
  {"h": "钰", "p": "yù", "sp": "yu", "r": "金", "sc": 10, "l": 5, "f": 4518},
  {"h": "儆", "p": "jǐng", "sp": "jing", "r": "人", "sc": 14, "l": 5, "f": 4519},
  {"h": "觎", "p": "yú", "sp": "yu", "r": "見", "sc": 13, "l": 5, "f": 4520},
  {"h": "讫", "p": "qì", "sp": "qi", "r": "言", "sc": 5, "l": 5, "f": 4521},
  {"h": "滓", "p": "zǐ", "sp": "zi", "r": "水", "sc": 13, "l": 5, "f": 4522},
  {"h": "僮", "p": "tóng", "sp": "tong", "r": "人", "sc": 14, "l": 5, "f": 4523},
  {"h": "媾", "p": "gòu", "sp": "gou", "r": "女", "sc": 13, "l": 5, "f": 4524},
  {"h": "龇", "p": "zī", "sp": "zi", "r": "齒", "sc": 14, "l": 5, "f": 4525},
  {"h": "胯", "p": "kuà", "sp": "kua", "r": "肉", "sc": 10, "l": 5, "f": 4526},
  {
    "h": "涮",
    "p": "shuàn",
    "sp": "shuan",
    "r": "水",
    "sc": 11,
    "l": 5,
    "f": 4527
  },
  {"h": "绾", "p": "wǎn", "sp": "wan", "r": "糸", "sc": 11, "l": 5, "f": 4528},
  {"h": "杈", "p": "chā", "sp": "cha", "r": "木", "sc": 7, "l": 5, "f": 4529},
  {"h": "赳", "p": "jiū", "sp": "jiu", "r": "走", "sc": 9, "l": 5, "f": 4530},
  {"h": "斛", "p": "hú", "sp": "hu", "r": "斗", "sc": 11, "l": 5, "f": 4531},
  {"h": "觥", "p": "gōng", "sp": "gong", "r": "角", "sc": 13, "l": 5, "f": 4532},
  {"h": "疸", "p": "dǎn", "sp": "dan", "r": "疒", "sc": 10, "l": 5, "f": 4533},
  {"h": "卞", "p": "biàn", "sp": "bian", "r": "卜", "sc": 4, "l": 5, "f": 4534},
  {"h": "愠", "p": "yùn", "sp": "yun", "r": "心", "sc": 12, "l": 5, "f": 4535},
  {"h": "拮", "p": "jié", "sp": "jie", "r": "手", "sc": 9, "l": 5, "f": 4536},
  {"h": "庠", "p": "xiáng", "sp": "xiang", "r": "广", "sc": 9, "l": 5, "f": 4537},
  {"h": "烨", "p": "yè", "sp": "ye", "r": "火", "sc": 10, "l": 5, "f": 4538},
  {"h": "龢", "p": "hé", "sp": "he", "r": "龠", "sc": 22, "l": 5, "f": 4539},
  {"h": "菠", "p": "bō", "sp": "bo", "r": "艸", "sc": 11, "l": 5, "f": 4540},
  {"h": "窈", "p": "yǎo", "sp": "yao", "r": "穴", "sc": 10, "l": 5, "f": 4541},
  {"h": "罄", "p": "qìng", "sp": "qing", "r": "缶", "sc": 17, "l": 5, "f": 4542},
  {"h": "囤", "p": "dùn", "sp": "dun", "r": "囗", "sc": 7, "l": 5, "f": 4543},
  {"h": "弁", "p": "biàn", "sp": "bian", "r": "廾", "sc": 5, "l": 5, "f": 4544},
  {"h": "奘", "p": "zàng", "sp": "zang", "r": "大", "sc": 10, "l": 5, "f": 4545},
  {"h": "咣", "p": "guāng", "sp": "guang", "r": "口", "sc": 9, "l": 5, "f": 4546},
  {"h": "缫", "p": "sāo", "sp": "sao", "r": "糸", "sc": 14, "l": 5, "f": 4547},
  {"h": "腴", "p": "yú", "sp": "yu", "r": "肉", "sc": 12, "l": 5, "f": 4548},
  {"h": "缈", "p": "miǎo", "sp": "miao", "r": "糸", "sc": 12, "l": 5, "f": 4549},
  {"h": "喵", "p": "miāo", "sp": "miao", "r": "口", "sc": 11, "l": 5, "f": 4550},
  {
    "h": "潢",
    "p": "huáng",
    "sp": "huang",
    "r": "水",
    "sc": 14,
    "l": 5,
    "f": 4551
  },
  {"h": "遛", "p": "liú", "sp": "liu", "r": "辵", "sc": 13, "l": 5, "f": 4552},
  {"h": "柚", "p": "yòu", "sp": "you", "r": "木", "sc": 9, "l": 5, "f": 4553},
  {"h": "郏", "p": "jiá", "sp": "jia", "r": "邑", "sc": 8, "l": 5, "f": 4554},
  {"h": "荻", "p": "dí", "sp": "di", "r": "艸", "sc": 10, "l": 5, "f": 4555},
  {"h": "藜", "p": "lí", "sp": "li", "r": "艸", "sc": 18, "l": 5, "f": 4556},
  {"h": "琨", "p": "kūn", "sp": "kun", "r": "玉", "sc": 12, "l": 5, "f": 4557},
  {"h": "镳", "p": "biāo", "sp": "biao", "r": "金", "sc": 20, "l": 5, "f": 4558},
  {"h": "雉", "p": "zhì", "sp": "zhi", "r": "隹", "sc": 13, "l": 5, "f": 4559},
  {"h": "橐", "p": "tuó", "sp": "tuo", "r": "木", "sc": 16, "l": 5, "f": 4560},
  {"h": "陽", "p": "yáng", "sp": "yang", "r": "阜", "sc": 11, "l": 5, "f": 4561},
  {"h": "骈", "p": "pián", "sp": "pian", "r": "馬", "sc": 9, "l": 5, "f": 4562},
  {"h": "蛉", "p": "líng", "sp": "ling", "r": "虫", "sc": 11, "l": 5, "f": 4563},
  {"h": "艮", "p": "gěn", "sp": "gen", "r": "艮", "sc": 6, "l": 5, "f": 4564},
  {"h": "搽", "p": "chá", "sp": "cha", "r": "手", "sc": 12, "l": 5, "f": 4565},
  {"h": "濡", "p": "rú", "sp": "ru", "r": "水", "sc": 17, "l": 5, "f": 4566},
  {"h": "寮", "p": "liáo", "sp": "liao", "r": "宀", "sc": 15, "l": 5, "f": 4567},
  {"h": "柩", "p": "jiù", "sp": "jiu", "r": "木", "sc": 9, "l": 5, "f": 4568},
  {"h": "佗", "p": "tuó", "sp": "tuo", "r": "人", "sc": 7, "l": 5, "f": 4569},
  {"h": "啷", "p": "lāng", "sp": "lang", "r": "口", "sc": 11, "l": 5, "f": 4570},
  {"h": "诜", "p": "shēn", "sp": "shen", "r": "言", "sc": 8, "l": 5, "f": 4571},
  {"h": "視", "p": "shì", "sp": "shi", "r": "見", "sc": 11, "l": 5, "f": 4572},
  {"h": "偻", "p": "lóu", "sp": "lou", "r": "人", "sc": 11, "l": 5, "f": 4573},
  {"h": "夯", "p": "hāng", "sp": "hang", "r": "大", "sc": 5, "l": 5, "f": 4574},
  {"h": "闱", "p": "wéi", "sp": "wei", "r": "門", "sc": 7, "l": 5, "f": 4575},
  {"h": "谖", "p": "xuān", "sp": "xuan", "r": "言", "sc": 11, "l": 5, "f": 4576},
  {"h": "夥", "p": "huǒ", "sp": "huo", "r": "夕", "sc": 14, "l": 5, "f": 4577},
  {"h": "枸", "p": "gǒu", "sp": "gou", "r": "木", "sc": 9, "l": 5, "f": 4578},
  {"h": "膑", "p": "bìn", "sp": "bin", "r": "肉", "sc": 14, "l": 5, "f": 4579},
  {"h": "虻", "p": "méng", "sp": "meng", "r": "虫", "sc": 9, "l": 5, "f": 4580},
  {"h": "筠", "p": "yún", "sp": "yun", "r": "竹", "sc": 13, "l": 5, "f": 4581},
  {"h": "埽", "p": "sào", "sp": "sao", "r": "土", "sc": 11, "l": 5, "f": 4582},
  {"h": "笞", "p": "chī", "sp": "chi", "r": "竹", "sc": 11, "l": 5, "f": 4583},
  {"h": "臾", "p": "yú", "sp": "yu", "r": "臼", "sc": 8, "l": 5, "f": 4584},
  {"h": "婀", "p": "ē", "sp": "e", "r": "女", "sc": 10, "l": 5, "f": 4585},
  {"h": "珞", "p": "luò", "sp": "luo", "r": "玉", "sc": 10, "l": 5, "f": 4586},
  {"h": "粑", "p": "bā", "sp": "ba", "r": "米", "sc": 10, "l": 5, "f": 4587},
  {"h": "怵", "p": "chù", "sp": "chu", "r": "心", "sc": 8, "l": 5, "f": 4588},
  {"h": "绻", "p": "quǎn", "sp": "quan", "r": "糸", "sc": 11, "l": 5, "f": 4589},
  {"h": "殒", "p": "yǔn", "sp": "yun", "r": "歹", "sc": 11, "l": 5, "f": 4590},
  {"h": "觊", "p": "jì", "sp": "ji", "r": "見", "sc": 10, "l": 5, "f": 4591},
  {"h": "崂", "p": "láo", "sp": "lao", "r": "山", "sc": 10, "l": 5, "f": 4592},
  {"h": "颧", "p": "quán", "sp": "quan", "r": "頁", "sc": 23, "l": 5, "f": 4593},
  {"h": "嗑", "p": "kē", "sp": "ke", "r": "口", "sc": 13, "l": 5, "f": 4594},
  {"h": "榛", "p": "zhēn", "sp": "zhen", "r": "木", "sc": 14, "l": 5, "f": 4595},
  {"h": "昱", "p": "yù", "sp": "yu", "r": "日", "sc": 9, "l": 5, "f": 4596},
  {"h": "蜴", "p": "yì", "sp": "yi", "r": "虫", "sc": 14, "l": 5, "f": 4597},
  {"h": "鳝", "p": "shàn", "sp": "shan", "r": "魚", "sc": 20, "l": 5, "f": 4598},
  {"h": "噙", "p": "qín", "sp": "qin", "r": "口", "sc": 15, "l": 5, "f": 4599},
  {"h": "淼", "p": "miǎo", "sp": "miao", "r": "水", "sc": 12, "l": 5, "f": 4600},
  {"h": "矾", "p": "fán", "sp": "fan", "r": "石", "sc": 8, "l": 5, "f": 4601},
  {"h": "硼", "p": "péng", "sp": "peng", "r": "石", "sc": 13, "l": 5, "f": 4602},
  {"h": "囿", "p": "yòu", "sp": "you", "r": "囗", "sc": 9, "l": 5, "f": 4603},
  {"h": "泅", "p": "qiú", "sp": "qiu", "r": "水", "sc": 8, "l": 5, "f": 4604},
  {"h": "邂", "p": "xiè", "sp": "xie", "r": "辵", "sc": 16, "l": 5, "f": 4605},
  {"h": "钜", "p": "jù", "sp": "ju", "r": "金", "sc": 9, "l": 5, "f": 4606},
  {"h": "蠹", "p": "dù", "sp": "du", "r": "虫", "sc": 24, "l": 5, "f": 4607},
  {"h": "垩", "p": "è", "sp": "e", "r": "土", "sc": 9, "l": 5, "f": 4608},
  {"h": "乩", "p": "jī", "sp": "ji", "r": "乙", "sc": 6, "l": 5, "f": 4609},
  {"h": "嗝", "p": "gé", "sp": "ge", "r": "口", "sc": 13, "l": 5, "f": 4610},
  {"h": "淦", "p": "gàn", "sp": "gan", "r": "水", "sc": 11, "l": 5, "f": 4611},
  {"h": "樽", "p": "zūn", "sp": "zun", "r": "木", "sc": 16, "l": 5, "f": 4612},
  {"h": "诮", "p": "qiào", "sp": "qiao", "r": "言", "sc": 9, "l": 5, "f": 4613},
  {"h": "揆", "p": "kuí", "sp": "kui", "r": "手", "sc": 12, "l": 5, "f": 4614},
  {"h": "啐", "p": "cuì", "sp": "cui", "r": "口", "sc": 11, "l": 5, "f": 4615},
  {"h": "淅", "p": "xī", "sp": "xi", "r": "水", "sc": 11, "l": 5, "f": 4616},
  {"h": "榉", "p": "jǔ", "sp": "ju", "r": "木", "sc": 13, "l": 5, "f": 4617},
  {"h": "馗", "p": "kuí", "sp": "kui", "r": "首", "sc": 11, "l": 5, "f": 4618},
  {"h": "辔", "p": "pèi", "sp": "pei", "r": "口", "sc": 13, "l": 5, "f": 4619},
  {"h": "暹", "p": "xiān", "sp": "xian", "r": "日", "sc": 15, "l": 5, "f": 4620},
  {"h": "骛", "p": "wù", "sp": "wu", "r": "馬", "sc": 12, "l": 5, "f": 4621},
  {"h": "鱿", "p": "yóu", "sp": "you", "r": "魚", "sc": 12, "l": 5, "f": 4622},
  {"h": "苫", "p": "shān", "sp": "shan", "r": "艸", "sc": 8, "l": 5, "f": 4623},
  {"h": "犷", "p": "guǎng", "sp": "guang", "r": "犬", "sc": 6, "l": 5, "f": 4624},
  {"h": "獠", "p": "liáo", "sp": "liao", "r": "犬", "sc": 15, "l": 5, "f": 4625},
  {"h": "詈", "p": "lì", "sp": "li", "r": "言", "sc": 12, "l": 5, "f": 4626},
  {"h": "竦", "p": "sǒng", "sp": "song", "r": "立", "sc": 12, "l": 5, "f": 4627},
  {"h": "篙", "p": "gāo", "sp": "gao", "r": "竹", "sc": 16, "l": 5, "f": 4628},
  {"h": "诨", "p": "hùn", "sp": "hun", "r": "言", "sc": 8, "l": 5, "f": 4629},
  {"h": "铰", "p": "jiǎo", "sp": "jiao", "r": "金", "sc": 11, "l": 5, "f": 4630},
  {"h": "馄", "p": "hún", "sp": "hun", "r": "食", "sc": 11, "l": 5, "f": 4631},
  {"h": "蜚", "p": "fēi", "sp": "fei", "r": "虫", "sc": 14, "l": 5, "f": 4632},
  {"h": "峒", "p": "dòng", "sp": "dong", "r": "山", "sc": 9, "l": 5, "f": 4633},
  {"h": "滢", "p": "yíng", "sp": "ying", "r": "水", "sc": 13, "l": 5, "f": 4634},
  {"h": "琬", "p": "wǎn", "sp": "wan", "r": "玉", "sc": 12, "l": 5, "f": 4635},
  {"h": "靓", "p": "jìng", "sp": "jing", "r": "青", "sc": 12, "l": 5, "f": 4636},
  {"h": "狻", "p": "suān", "sp": "suan", "r": "犬", "sc": 10, "l": 5, "f": 4637},
  {"h": "璨", "p": "càn", "sp": "can", "r": "玉", "sc": 17, "l": 5, "f": 4638},
  {
    "h": "犟",
    "p": "jiàng",
    "sp": "jiang",
    "r": "牛",
    "sc": 16,
    "l": 5,
    "f": 4639
  },
  {"h": "鸬", "p": "lú", "sp": "lu", "r": "鳥", "sc": 10, "l": 5, "f": 4640},
  {"h": "螨", "p": "mǎn", "sp": "man", "r": "虫", "sc": 16, "l": 5, "f": 4641},
  {"h": "芩", "p": "qín", "sp": "qin", "r": "艸", "sc": 7, "l": 5, "f": 4642},
  {"h": "嘹", "p": "liáo", "sp": "liao", "r": "口", "sc": 15, "l": 5, "f": 4643},
  {"h": "锟", "p": "kūn", "sp": "kun", "r": "金", "sc": 13, "l": 5, "f": 4644},
  {"h": "蜇", "p": "zhē", "sp": "zhe", "r": "虫", "sc": 13, "l": 5, "f": 4645},
  {"h": "洹", "p": "huán", "sp": "huan", "r": "水", "sc": 9, "l": 5, "f": 4646},
  {"h": "栉", "p": "zhì", "sp": "zhi", "r": "木", "sc": 9, "l": 5, "f": 4647},
  {"h": "俪", "p": "lì", "sp": "li", "r": "人", "sc": 9, "l": 5, "f": 4648},
  {"h": "钍", "p": "tǔ", "sp": "tu", "r": "金", "sc": 8, "l": 5, "f": 4649},
  {"h": "锨", "p": "xiān", "sp": "xian", "r": "金", "sc": 13, "l": 5, "f": 4650},
  {"h": "瑁", "p": "mào", "sp": "mao", "r": "玉", "sc": 13, "l": 5, "f": 4651},
  {"h": "壹", "p": "yī", "sp": "yi", "r": "士", "sc": 12, "l": 5, "f": 4652},
  {"h": "痿", "p": "wěi", "sp": "wei", "r": "疒", "sc": 13, "l": 5, "f": 4653},
  {"h": "竑", "p": "hóng", "sp": "hong", "r": "立", "sc": 9, "l": 5, "f": 4654},
  {"h": "粕", "p": "pò", "sp": "po", "r": "米", "sc": 11, "l": 5, "f": 4655},
  {"h": "犄", "p": "jī", "sp": "ji", "r": "牛", "sc": 12, "l": 5, "f": 4656},
  {"h": "瘙", "p": "sào", "sp": "sao", "r": "疒", "sc": 14, "l": 5, "f": 4657},
  {"h": "饯", "p": "jiàn", "sp": "jian", "r": "食", "sc": 8, "l": 5, "f": 4658},
  {"h": "抟", "p": "tuán", "sp": "tuan", "r": "手", "sc": 7, "l": 5, "f": 4659},
  {"h": "衲", "p": "nà", "sp": "na", "r": "衣", "sc": 9, "l": 5, "f": 4660},
  {"h": "踮", "p": "diǎn", "sp": "dian", "r": "足", "sc": 15, "l": 5, "f": 4661},
  {"h": "龅", "p": "bāo", "sp": "bao", "r": "齒", "sc": 13, "l": 5, "f": 4662},
  {"h": "愎", "p": "bì", "sp": "bi", "r": "心", "sc": 12, "l": 5, "f": 4663},
  {"h": "馥", "p": "fù", "sp": "fu", "r": "香", "sc": 18, "l": 5, "f": 4664},
  {"h": "梏", "p": "gù", "sp": "gu", "r": "木", "sc": 11, "l": 5, "f": 4665},
  {"h": "讣", "p": "fù", "sp": "fu", "r": "言", "sc": 4, "l": 5, "f": 4666},
  {"h": "邝", "p": "kuàng", "sp": "kuang", "r": "邑", "sc": 5, "l": 5, "f": 4667},
  {"h": "艿", "p": "nǎi", "sp": "nai", "r": "艸", "sc": 5, "l": 5, "f": 4668},
  {"h": "趺", "p": "fū", "sp": "fu", "r": "足", "sc": 11, "l": 5, "f": 4669},
  {"h": "鲟", "p": "xún", "sp": "xun", "r": "魚", "sc": 14, "l": 5, "f": 4670},
  {"h": "剜", "p": "wān", "sp": "wan", "r": "刀", "sc": 10, "l": 5, "f": 4671},
  {"h": "绉", "p": "zhòu", "sp": "zhou", "r": "糸", "sc": 8, "l": 5, "f": 4672},
  {"h": "罅", "p": "xià", "sp": "xia", "r": "缶", "sc": 17, "l": 5, "f": 4673},
  {"h": "笥", "p": "sì", "sp": "si", "r": "竹", "sc": 11, "l": 5, "f": 4674},
  {"h": "衩", "p": "chǎ", "sp": "cha", "r": "衣", "sc": 8, "l": 5, "f": 4675},
  {"h": "姣", "p": "jiāo", "sp": "jiao", "r": "女", "sc": 9, "l": 5, "f": 4676},
  {"h": "斫", "p": "zhuó", "sp": "zhuo", "r": "斤", "sc": 9, "l": 5, "f": 4677},
  {"h": "鹗", "p": "è", "sp": "e", "r": "鳥", "sc": 14, "l": 5, "f": 4678},
  {"h": "腎", "p": "shèn", "sp": "shen", "r": "肉", "sc": 12, "l": 5, "f": 4679},
  {"h": "爻", "p": "yáo", "sp": "yao", "r": "爻", "sc": 4, "l": 5, "f": 4680},
  {"h": "猕", "p": "mí", "sp": "mi", "r": "犬", "sc": 11, "l": 5, "f": 4681},
  {"h": "晗", "p": "hán", "sp": "han", "r": "日", "sc": 11, "l": 5, "f": 4682},
  {"h": "铩", "p": "shā", "sp": "sha", "r": "金", "sc": 11, "l": 5, "f": 4683},
  {"h": "窕", "p": "tiǎo", "sp": "tiao", "r": "穴", "sc": 11, "l": 5, "f": 4684},
  {"h": "仨", "p": "sā", "sp": "sa", "r": "人", "sc": 5, "l": 5, "f": 4685},
  {"h": "搡", "p": "sǎng", "sp": "sang", "r": "手", "sc": 13, "l": 5, "f": 4686},
  {"h": "崴", "p": "wǎi", "sp": "wai", "r": "山", "sc": 12, "l": 5, "f": 4687},
  {"h": "酢", "p": "cù", "sp": "cu", "r": "酉", "sc": 12, "l": 5, "f": 4688},
  {"h": "檄", "p": "xí", "sp": "xi", "r": "木", "sc": 17, "l": 5, "f": 4689},
  {"h": "佞", "p": "nìng", "sp": "ning", "r": "人", "sc": 7, "l": 5, "f": 4690},
  {"h": "孑", "p": "jié", "sp": "jie", "r": "子", "sc": 3, "l": 5, "f": 4691},
  {"h": "璀", "p": "cuǐ", "sp": "cui", "r": "玉", "sc": 15, "l": 5, "f": 4692},
  {"h": "岷", "p": "mín", "sp": "min", "r": "山", "sc": 8, "l": 5, "f": 4693},
  {"h": "舛", "p": "chuǎn", "sp": "chuan", "r": "舛", "sc": 6, "l": 5, "f": 4694},
  {"h": "邕", "p": "yōng", "sp": "yong", "r": "邑", "sc": 10, "l": 5, "f": 4695},
  {"h": "闿", "p": "kǎi", "sp": "kai", "r": "門", "sc": 9, "l": 5, "f": 4696},
  {"h": "铂", "p": "bó", "sp": "bo", "r": "金", "sc": 10, "l": 5, "f": 4697},
  {"h": "霁", "p": "jì", "sp": "ji", "r": "雨", "sc": 14, "l": 5, "f": 4698},
  {"h": "犒", "p": "kào", "sp": "kao", "r": "牛", "sc": 14, "l": 5, "f": 4699},
  {"h": "馏", "p": "liú", "sp": "liu", "r": "食", "sc": 13, "l": 5, "f": 4700},
  {"h": "阈", "p": "yù", "sp": "yu", "r": "門", "sc": 11, "l": 5, "f": 4701},
  {"h": "麋", "p": "mí", "sp": "mi", "r": "鹿", "sc": 17, "l": 5, "f": 4702},
  {"h": "麒", "p": "qí", "sp": "qi", "r": "鹿", "sc": 19, "l": 5, "f": 4703},
  {"h": "苁", "p": "cōng", "sp": "cong", "r": "艸", "sc": 7, "l": 5, "f": 4704},
  {"h": "摁", "p": "èn", "sp": "en", "r": "手", "sc": 13, "l": 5, "f": 4705},
  {"h": "涔", "p": "cén", "sp": "cen", "r": "水", "sc": 10, "l": 5, "f": 4706},
  {"h": "宥", "p": "yòu", "sp": "you", "r": "宀", "sc": 9, "l": 5, "f": 4707},
  {"h": "妍", "p": "yán", "sp": "yan", "r": "女", "sc": 7, "l": 5, "f": 4708},
  {"h": "铤", "p": "dìng", "sp": "ding", "r": "金", "sc": 11, "l": 5, "f": 4709},
  {"h": "锷", "p": "è", "sp": "e", "r": "金", "sc": 14, "l": 5, "f": 4710},
  {"h": "嗲", "p": "diǎ", "sp": "dia", "r": "口", "sc": 13, "l": 5, "f": 4711},
  {"h": "恽", "p": "yùn", "sp": "yun", "r": "心", "sc": 9, "l": 5, "f": 4712},
  {"h": "麂", "p": "jǐ", "sp": "ji", "r": "鹿", "sc": 13, "l": 5, "f": 4713},
  {"h": "赝", "p": "yàn", "sp": "yan", "r": "厂", "sc": 16, "l": 5, "f": 4714},
  {"h": "胛", "p": "jiǎ", "sp": "jia", "r": "肉", "sc": 9, "l": 5, "f": 4715},
  {"h": "哂", "p": "shěn", "sp": "shen", "r": "口", "sc": 9, "l": 5, "f": 4716},
  {"h": "撷", "p": "xié", "sp": "xie", "r": "手", "sc": 15, "l": 5, "f": 4717},
  {"h": "呶", "p": "náo", "sp": "nao", "r": "口", "sc": 8, "l": 5, "f": 4718},
  {"h": "噘", "p": "juē", "sp": "jue", "r": "口", "sc": 15, "l": 5, "f": 4719},
  {"h": "懔", "p": "lǐn", "sp": "lin", "r": "心", "sc": 16, "l": 5, "f": 4720},
  {"h": "栎", "p": "lì", "sp": "li", "r": "木", "sc": 9, "l": 5, "f": 4721},
  {"h": "桎", "p": "zhì", "sp": "zhi", "r": "木", "sc": 10, "l": 5, "f": 4722},
  {"h": "霰", "p": "sǎn", "sp": "san", "r": "雨", "sc": 20, "l": 5, "f": 4723},
  {
    "h": "飨",
    "p": "xiǎng",
    "sp": "xiang",
    "r": "食",
    "sc": 12,
    "l": 5,
    "f": 4724
  },
  {"h": "揄", "p": "yú", "sp": "yu", "r": "手", "sc": 12, "l": 5, "f": 4725},
  {"h": "噔", "p": "dēng", "sp": "deng", "r": "口", "sc": 15, "l": 5, "f": 4726},
  {"h": "娣", "p": "dì", "sp": "di", "r": "女", "sc": 10, "l": 5, "f": 4727},
  {"h": "薏", "p": "yì", "sp": "yi", "r": "艸", "sc": 16, "l": 5, "f": 4728},
  {"h": "忝", "p": "tiǎn", "sp": "tian", "r": "心", "sc": 8, "l": 5, "f": 4729},
  {"h": "咤", "p": "zhà", "sp": "zha", "r": "口", "sc": 9, "l": 5, "f": 4730},
  {"h": "嗵", "p": "tōng", "sp": "tong", "r": "口", "sc": 13, "l": 5, "f": 4731},
  {"h": "迤", "p": "yí", "sp": "yi", "r": "辵", "sc": 8, "l": 5, "f": 4732},
  {"h": "贲", "p": "bēn", "sp": "ben", "r": "貝", "sc": 9, "l": 5, "f": 4733},
  {"h": "胪", "p": "lú", "sp": "lu", "r": "肉", "sc": 9, "l": 5, "f": 4734},
  {"h": "鍪", "p": "móu", "sp": "mou", "r": "金", "sc": 17, "l": 5, "f": 4735},
  {"h": "泸", "p": "lú", "sp": "lu", "r": "水", "sc": 8, "l": 5, "f": 4736},
  {"h": "蔫", "p": "niān", "sp": "nian", "r": "艸", "sc": 14, "l": 5, "f": 4737},
  {"h": "刈", "p": "yì", "sp": "yi", "r": "刀", "sc": 4, "l": 5, "f": 4738},
  {"h": "僖", "p": "xī", "sp": "xi", "r": "人", "sc": 14, "l": 5, "f": 4739},
  {"h": "咿", "p": "yī", "sp": "yi", "r": "口", "sc": 9, "l": 5, "f": 4740},
  {"h": "鹌", "p": "ān", "sp": "an", "r": "鳥", "sc": 13, "l": 5, "f": 4741},
  {"h": "嗪", "p": "qín", "sp": "qin", "r": "口", "sc": 13, "l": 5, "f": 4742},
  {"h": "茏", "p": "lóng", "sp": "long", "r": "艸", "sc": 8, "l": 5, "f": 4743},
  {"h": "茯", "p": "fú", "sp": "fu", "r": "艸", "sc": 9, "l": 5, "f": 4744},
  {"h": "岫", "p": "xiù", "sp": "xiu", "r": "山", "sc": 8, "l": 5, "f": 4745},
  {"h": "嵘", "p": "róng", "sp": "rong", "r": "山", "sc": 12, "l": 5, "f": 4746},
  {"h": "轱", "p": "gū", "sp": "gu", "r": "車", "sc": 9, "l": 5, "f": 4747},
  {"h": "怼", "p": "duì", "sp": "dui", "r": "心", "sc": 9, "l": 5, "f": 4748},
  {"h": "铨", "p": "quán", "sp": "quan", "r": "金", "sc": 11, "l": 5, "f": 4749},
  {"h": "昕", "p": "xīn", "sp": "xin", "r": "日", "sc": 8, "l": 5, "f": 4750},
  {"h": "郢", "p": "yǐng", "sp": "ying", "r": "邑", "sc": 9, "l": 5, "f": 4751},
  {"h": "咩", "p": "miē", "sp": "mie", "r": "口", "sc": 9, "l": 5, "f": 4752},
  {"h": "馊", "p": "sōu", "sp": "sou", "r": "食", "sc": 12, "l": 5, "f": 4753},
  {"h": "髡", "p": "kūn", "sp": "kun", "r": "髟", "sc": 13, "l": 5, "f": 4754},
  {"h": "澧", "p": "lǐ", "sp": "li", "r": "水", "sc": 16, "l": 5, "f": 4755},
  {"h": "苣", "p": "jù", "sp": "ju", "r": "艸", "sc": 7, "l": 5, "f": 4756},
  {"h": "濯", "p": "zhuó", "sp": "zhuo", "r": "水", "sc": 17, "l": 5, "f": 4757},
  {"h": "盥", "p": "guàn", "sp": "guan", "r": "皿", "sc": 16, "l": 5, "f": 4758},
  {"h": "囡", "p": "nān", "sp": "nan", "r": "囗", "sc": 6, "l": 5, "f": 4759},
  {"h": "砺", "p": "lì", "sp": "li", "r": "石", "sc": 10, "l": 5, "f": 4760},
  {"h": "佘", "p": "shé", "sp": "she", "r": "人", "sc": 7, "l": 5, "f": 4761},
  {"h": "谶", "p": "chèn", "sp": "chen", "r": "言", "sc": 19, "l": 5, "f": 4762},
  {"h": "弑", "p": "shì", "sp": "shi", "r": "弋", "sc": 12, "l": 5, "f": 4763},
  {"h": "楂", "p": "zhā", "sp": "zha", "r": "木", "sc": 13, "l": 5, "f": 4764},
  {"h": "翦", "p": "jiǎn", "sp": "jian", "r": "羽", "sc": 15, "l": 5, "f": 4765},
  {"h": "怩", "p": "ní", "sp": "ni", "r": "心", "sc": 8, "l": 5, "f": 4766},
  {"h": "蠼", "p": "qú", "sp": "qu", "r": "虫", "sc": 26, "l": 5, "f": 4767},
  {"h": "霏", "p": "fēi", "sp": "fei", "r": "雨", "sc": 16, "l": 5, "f": 4768},
  {"h": "楹", "p": "yíng", "sp": "ying", "r": "木", "sc": 13, "l": 5, "f": 4769},
  {"h": "讴", "p": "ōu", "sp": "ou", "r": "言", "sc": 6, "l": 5, "f": 4770},
  {"h": "锲", "p": "qiè", "sp": "qie", "r": "金", "sc": 14, "l": 5, "f": 4771},
  {"h": "慵", "p": "yōng", "sp": "yong", "r": "心", "sc": 14, "l": 5, "f": 4772},
  {"h": "胝", "p": "zhī", "sp": "zhi", "r": "肉", "sc": 9, "l": 5, "f": 4773},
  {"h": "砭", "p": "biān", "sp": "bian", "r": "石", "sc": 9, "l": 5, "f": 4774},
  {"h": "潍", "p": "wéi", "sp": "wei", "r": "水", "sc": 14, "l": 5, "f": 4775},
  {"h": "杵", "p": "chǔ", "sp": "chu", "r": "木", "sc": 8, "l": 5, "f": 4776},
  {"h": "樾", "p": "yuè", "sp": "yue", "r": "木", "sc": 16, "l": 5, "f": 4777},
  {"h": "帼", "p": "guó", "sp": "guo", "r": "巾", "sc": 11, "l": 5, "f": 4778},
  {"h": "碣", "p": "jié", "sp": "jie", "r": "石", "sc": 14, "l": 5, "f": 4779},
  {"h": "诌", "p": "zhōu", "sp": "zhou", "r": "言", "sc": 7, "l": 5, "f": 4780},
  {"h": "徕", "p": "lái", "sp": "lai", "r": "彳", "sc": 10, "l": 5, "f": 4781},
  {"h": "胴", "p": "dòng", "sp": "dong", "r": "肉", "sc": 10, "l": 5, "f": 4782},
  {"h": "钴", "p": "gǔ", "sp": "gu", "r": "金", "sc": 10, "l": 5, "f": 4783},
  {"h": "裟", "p": "shā", "sp": "sha", "r": "衣", "sc": 13, "l": 5, "f": 4784},
  {"h": "啶", "p": "dìng", "sp": "ding", "r": "口", "sc": 11, "l": 5, "f": 4785},
  {"h": "铣", "p": "xǐ", "sp": "xi", "r": "金", "sc": 11, "l": 5, "f": 4786},
  {"h": "铱", "p": "yī", "sp": "yi", "r": "金", "sc": 11, "l": 5, "f": 4787},
  {"h": "楫", "p": "jí", "sp": "ji", "r": "木", "sc": 13, "l": 5, "f": 4788},
  {"h": "赭", "p": "zhě", "sp": "zhe", "r": "赤", "sc": 15, "l": 5, "f": 4789},
  {"h": "碛", "p": "qì", "sp": "qi", "r": "石", "sc": 13, "l": 5, "f": 4790},
  {"h": "酊", "p": "dīng", "sp": "ding", "r": "酉", "sc": 9, "l": 5, "f": 4791},
  {"h": "魑", "p": "chī", "sp": "chi", "r": "鬼", "sc": 19, "l": 5, "f": 4792},
  {"h": "醛", "p": "quán", "sp": "quan", "r": "酉", "sc": 16, "l": 5, "f": 4793},
  {"h": "剐", "p": "guǎ", "sp": "gua", "r": "刀", "sc": 9, "l": 5, "f": 4794},
  {"h": "畦", "p": "qí", "sp": "qi", "r": "田", "sc": 11, "l": 5, "f": 4795},
  {"h": "陂", "p": "bēi", "sp": "bei", "r": "阜", "sc": 7, "l": 5, "f": 4796},
  {"h": "闶", "p": "kāng", "sp": "kang", "r": "門", "sc": 7, "l": 5, "f": 4797},
  {"h": "阄", "p": "jiū", "sp": "jiu", "r": "門", "sc": 10, "l": 5, "f": 4798},
  {"h": "祚", "p": "zuò", "sp": "zuo", "r": "示", "sc": 9, "l": 5, "f": 4799},
  {"h": "鹘", "p": "gǔ", "sp": "gu", "r": "鳥", "sc": 14, "l": 5, "f": 4800},
  {"h": "泱", "p": "yāng", "sp": "yang", "r": "水", "sc": 8, "l": 5, "f": 4801},
  {"h": "趄", "p": "jū", "sp": "ju", "r": "走", "sc": 12, "l": 5, "f": 4802},
  {"h": "骅", "p": "huá", "sp": "hua", "r": "馬", "sc": 9, "l": 5, "f": 4803},
  {"h": "陲", "p": "chuí", "sp": "chui", "r": "阜", "sc": 10, "l": 5, "f": 4804},
  {"h": "郧", "p": "yún", "sp": "yun", "r": "邑", "sc": 9, "l": 5, "f": 4805},
  {"h": "倜", "p": "tì", "sp": "ti", "r": "人", "sc": 10, "l": 5, "f": 4806},
  {"h": "呤", "p": "lìng", "sp": "ling", "r": "口", "sc": 8, "l": 5, "f": 4807},
  {"h": "燧", "p": "suì", "sp": "sui", "r": "火", "sc": 16, "l": 5, "f": 4808},
  {"h": "铉", "p": "xuàn", "sp": "xuan", "r": "金", "sc": 10, "l": 5, "f": 4809},
  {"h": "粲", "p": "càn", "sp": "can", "r": "米", "sc": 13, "l": 5, "f": 4810},
  {"h": "骶", "p": "dǐ", "sp": "di", "r": "骨", "sc": 14, "l": 5, "f": 4811},
  {"h": "峁", "p": "mǎo", "sp": "mao", "r": "山", "sc": 8, "l": 5, "f": 4812},
  {"h": "忸", "p": "niǔ", "sp": "niu", "r": "心", "sc": 7, "l": 5, "f": 4813},
  {"h": "渌", "p": "lù", "sp": "lu", "r": "水", "sc": 11, "l": 5, "f": 4814},
  {"h": "骞", "p": "qiān", "sp": "qian", "r": "馬", "sc": 13, "l": 5, "f": 4815},
  {"h": "髭", "p": "zī", "sp": "zi", "r": "髟", "sc": 16, "l": 5, "f": 4816},
  {"h": "戡", "p": "kān", "sp": "kan", "r": "戈", "sc": 13, "l": 5, "f": 4817},
  {"h": "钨", "p": "wū", "sp": "wu", "r": "金", "sc": 9, "l": 5, "f": 4818},
  {"h": "谲", "p": "jué", "sp": "jue", "r": "言", "sc": 14, "l": 5, "f": 4819},
  {"h": "苋", "p": "xiàn", "sp": "xian", "r": "艸", "sc": 7, "l": 5, "f": 4820},
  {"h": "锃", "p": "zèng", "sp": "zeng", "r": "金", "sc": 12, "l": 5, "f": 4821},
  {"h": "蜊", "p": "lí", "sp": "li", "r": "虫", "sc": 13, "l": 5, "f": 4822},
  {"h": "幄", "p": "wò", "sp": "wo", "r": "巾", "sc": 12, "l": 5, "f": 4823},
  {"h": "闼", "p": "tà", "sp": "ta", "r": "門", "sc": 9, "l": 5, "f": 4824},
  {"h": "戕", "p": "qiāng", "sp": "qiang", "r": "戈", "sc": 8, "l": 5, "f": 4825},
  {"h": "骊", "p": "lí", "sp": "li", "r": "馬", "sc": 10, "l": 5, "f": 4826},
  {"h": "虢", "p": "guó", "sp": "guo", "r": "虍", "sc": 15, "l": 5, "f": 4827},
  {"h": "烩", "p": "huì", "sp": "hui", "r": "火", "sc": 10, "l": 5, "f": 4828},
  {"h": "傥", "p": "tǎng", "sp": "tang", "r": "人", "sc": 12, "l": 5, "f": 4829},
  {"h": "妲", "p": "dá", "sp": "da", "r": "女", "sc": 8, "l": 5, "f": 4830},
  {"h": "绌", "p": "chù", "sp": "chu", "r": "糸", "sc": 8, "l": 5, "f": 4831},
  {"h": "桠", "p": "yā", "sp": "ya", "r": "木", "sc": 10, "l": 5, "f": 4832},
  {"h": "袈", "p": "jiā", "sp": "jia", "r": "衣", "sc": 11, "l": 5, "f": 4833},
  {
    "h": "鎗",
    "p": "qiāng",
    "sp": "qiang",
    "r": "金",
    "sc": 18,
    "l": 5,
    "f": 4834
  },
  {"h": "薮", "p": "sǒu", "sp": "sou", "r": "艸", "sc": 16, "l": 5, "f": 4835},
  {"h": "揿", "p": "qìn", "sp": "qin", "r": "手", "sc": 12, "l": 5, "f": 4836},
  {"h": "杲", "p": "gǎo", "sp": "gao", "r": "木", "sc": 8, "l": 5, "f": 4837},
  {"h": "肓", "p": "huāng", "sp": "huang", "r": "肉", "sc": 7, "l": 5, "f": 4838},
  {"h": "厝", "p": "cuò", "sp": "cuo", "r": "厂", "sc": 10, "l": 5, "f": 4839},
  {"h": "莅", "p": "lì", "sp": "li", "r": "艸", "sc": 10, "l": 5, "f": 4840},
  {"h": "氤", "p": "yīn", "sp": "yin", "r": "气", "sc": 10, "l": 5, "f": 4841},
  {"h": "缙", "p": "jìn", "sp": "jin", "r": "糸", "sc": 13, "l": 5, "f": 4842},
  {"h": "衮", "p": "gǔn", "sp": "gun", "r": "衣", "sc": 10, "l": 5, "f": 4843},
  {"h": "诟", "p": "gòu", "sp": "gou", "r": "言", "sc": 8, "l": 5, "f": 4844},
  {"h": "旖", "p": "yǐ", "sp": "yi", "r": "方", "sc": 14, "l": 5, "f": 4845},
  {"h": "硒", "p": "xī", "sp": "xi", "r": "石", "sc": 11, "l": 5, "f": 4846},
  {"h": "唁", "p": "yàn", "sp": "yan", "r": "口", "sc": 10, "l": 5, "f": 4847},
  {"h": "嬗", "p": "shàn", "sp": "shan", "r": "女", "sc": 16, "l": 5, "f": 4848},
  {"h": "硎", "p": "xíng", "sp": "xing", "r": "石", "sc": 11, "l": 5, "f": 4849},
  {"h": "裱", "p": "biǎo", "sp": "biao", "r": "衣", "sc": 13, "l": 5, "f": 4850},
  {"h": "颦", "p": "pín", "sp": "pin", "r": "十", "sc": 21, "l": 5, "f": 4851},
  {"h": "質", "p": "zhì", "sp": "zhi", "r": "貝", "sc": 15, "l": 5, "f": 4852},
  {"h": "靥", "p": "yè", "sp": "ye", "r": "厂", "sc": 15, "l": 5, "f": 4853},
  {"h": "纥", "p": "gē", "sp": "ge", "r": "糸", "sc": 6, "l": 5, "f": 4854},
  {"h": "煨", "p": "wēi", "sp": "wei", "r": "火", "sc": 13, "l": 5, "f": 4855},
  {"h": "礴", "p": "bó", "sp": "bo", "r": "石", "sc": 21, "l": 5, "f": 4856},
  {"h": "鏖", "p": "áo", "sp": "ao", "r": "金", "sc": 19, "l": 5, "f": 4857},
  {"h": "蝈", "p": "guō", "sp": "guo", "r": "虫", "sc": 14, "l": 5, "f": 4858},
  {"h": "笏", "p": "hù", "sp": "hu", "r": "竹", "sc": 10, "l": 5, "f": 4859},
  {"h": "羿", "p": "yì", "sp": "yi", "r": "羽", "sc": 9, "l": 5, "f": 4860},
  {"h": "鼐", "p": "nài", "sp": "nai", "r": "鼎", "sc": 14, "l": 5, "f": 4861},
  {
    "h": "湟",
    "p": "huáng",
    "sp": "huang",
    "r": "水",
    "sc": 12,
    "l": 5,
    "f": 4862
  },
  {"h": "甑", "p": "zèng", "sp": "zeng", "r": "瓦", "sc": 16, "l": 5, "f": 4863},
  {"h": "炜", "p": "wěi", "sp": "wei", "r": "火", "sc": 8, "l": 5, "f": 4864},
  {"h": "煲", "p": "bāo", "sp": "bao", "r": "火", "sc": 13, "l": 5, "f": 4865},
  {"h": "锉", "p": "cuò", "sp": "cuo", "r": "金", "sc": 12, "l": 5, "f": 4866},
  {"h": "笕", "p": "jiǎn", "sp": "jian", "r": "竹", "sc": 10, "l": 5, "f": 4867},
  {"h": "喑", "p": "yīn", "sp": "yin", "r": "口", "sc": 12, "l": 5, "f": 4868},
  {
    "h": "嶂",
    "p": "zhàng",
    "sp": "zhang",
    "r": "山",
    "sc": 14,
    "l": 5,
    "f": 4869
  },
  {"h": "浔", "p": "xún", "sp": "xun", "r": "水", "sc": 9, "l": 5, "f": 4870},
  {"h": "弭", "p": "mǐ", "sp": "mi", "r": "弓", "sc": 9, "l": 5, "f": 4871},
  {"h": "妪", "p": "yù", "sp": "yu", "r": "女", "sc": 7, "l": 5, "f": 4872},
  {"h": "锂", "p": "lǐ", "sp": "li", "r": "金", "sc": 12, "l": 5, "f": 4873},
  {"h": "苡", "p": "yǐ", "sp": "yi", "r": "艸", "sc": 7, "l": 5, "f": 4874},
  {"h": "孳", "p": "zī", "sp": "zi", "r": "子", "sc": 12, "l": 5, "f": 4875},
  {"h": "颏", "p": "kē", "sp": "ke", "r": "頁", "sc": 12, "l": 5, "f": 4876},
  {"h": "醴", "p": "lǐ", "sp": "li", "r": "酉", "sc": 20, "l": 5, "f": 4877},
  {"h": "間", "p": "jiān", "sp": "jian", "r": "門", "sc": 12, "l": 5, "f": 4878},
  {"h": "渚", "p": "zhǔ", "sp": "zhu", "r": "水", "sc": 11, "l": 5, "f": 4879},
  {"h": "轭", "p": "è", "sp": "e", "r": "車", "sc": 8, "l": 5, "f": 4880},
  {"h": "鹬", "p": "yù", "sp": "yu", "r": "鳥", "sc": 17, "l": 5, "f": 4881},
  {"h": "蚝", "p": "háo", "sp": "hao", "r": "虫", "sc": 10, "l": 5, "f": 4882},
  {
    "h": "黃",
    "p": "huáng",
    "sp": "huang",
    "r": "黃",
    "sc": 12,
    "l": 5,
    "f": 4883
  },
  {"h": "膘", "p": "biāo", "sp": "biao", "r": "肉", "sc": 15, "l": 5, "f": 4884},
  {"h": "邛", "p": "qióng", "sp": "qiong", "r": "邑", "sc": 5, "l": 5, "f": 4885},
  {"h": "痨", "p": "láo", "sp": "lao", "r": "疒", "sc": 12, "l": 5, "f": 4886},
  {"h": "褡", "p": "dā", "sp": "da", "r": "衣", "sc": 14, "l": 5, "f": 4887},
  {"h": "耦", "p": "ǒu", "sp": "ou", "r": "耒", "sc": 15, "l": 5, "f": 4888},
  {"h": "覃", "p": "tán", "sp": "tan", "r": "襾", "sc": 12, "l": 5, "f": 4889},
  {"h": "虛", "p": "xū", "sp": "xu", "r": "虍", "sc": 12, "l": 5, "f": 4890},
  {
    "h": "馔",
    "p": "zhuàn",
    "sp": "zhuan",
    "r": "食",
    "sc": 15,
    "l": 5,
    "f": 4891
  },
  {"h": "篾", "p": "miè", "sp": "mie", "r": "竹", "sc": 17, "l": 5, "f": 4892},
  {"h": "兖", "p": "yǎn", "sp": "yan", "r": "儿", "sc": 8, "l": 5, "f": 4893},
  {"h": "阋", "p": "xì", "sp": "xi", "r": "門", "sc": 11, "l": 5, "f": 4894},
  {"h": "遨", "p": "áo", "sp": "ao", "r": "辵", "sc": 13, "l": 5, "f": 4895},
  {"h": "爰", "p": "yuán", "sp": "yuan", "r": "爪", "sc": 9, "l": 5, "f": 4896},
  {"h": "痂", "p": "jiā", "sp": "jia", "r": "疒", "sc": 10, "l": 5, "f": 4897},
  {"h": "艄", "p": "shāo", "sp": "shao", "r": "舟", "sc": 13, "l": 5, "f": 4898},
  {"h": "耨", "p": "nòu", "sp": "nou", "r": "耒", "sc": 16, "l": 5, "f": 4899},
  {"h": "沤", "p": "ōu", "sp": "ou", "r": "水", "sc": 7, "l": 5, "f": 4900},
  {"h": "邋", "p": "lā", "sp": "la", "r": "辵", "sc": 18, "l": 5, "f": 4901},
  {"h": "焓", "p": "hán", "sp": "han", "r": "火", "sc": 11, "l": 5, "f": 4902},
  {"h": "秣", "p": "mò", "sp": "mo", "r": "禾", "sc": 10, "l": 5, "f": 4903},
  {"h": "昶", "p": "chǎng", "sp": "chang", "r": "日", "sc": 9, "l": 5, "f": 4904},
  {
    "h": "種",
    "p": "zhǒng",
    "sp": "zhong",
    "r": "禾",
    "sc": 14,
    "l": 5,
    "f": 4905
  },
  {"h": "變", "p": "biàn", "sp": "bian", "r": "言", "sc": 23, "l": 5, "f": 4906},
  {"h": "窣", "p": "sū", "sp": "su", "r": "穴", "sc": 13, "l": 5, "f": 4907},
  {"h": "绦", "p": "tāo", "sp": "tao", "r": "糸", "sc": 10, "l": 5, "f": 4908},
  {"h": "俎", "p": "zǔ", "sp": "zu", "r": "人", "sc": 9, "l": 5, "f": 4909},
  {"h": "榫", "p": "sǔn", "sp": "sun", "r": "木", "sc": 14, "l": 5, "f": 4910},
  {"h": "蟪", "p": "huì", "sp": "hui", "r": "虫", "sc": 18, "l": 5, "f": 4911},
  {"h": "稗", "p": "bài", "sp": "bai", "r": "禾", "sc": 13, "l": 5, "f": 4912},
  {"h": "謇", "p": "jiǎn", "sp": "jian", "r": "言", "sc": 17, "l": 5, "f": 4913},
  {"h": "氩", "p": "yà", "sp": "ya", "r": "气", "sc": 10, "l": 5, "f": 4914},
  {"h": "類", "p": "lèi", "sp": "lei", "r": "頁", "sc": 19, "l": 5, "f": 4915},
  {"h": "锴", "p": "kǎi", "sp": "kai", "r": "金", "sc": 14, "l": 5, "f": 4916},
  {"h": "龉", "p": "yǔ", "sp": "yu", "r": "齒", "sc": 15, "l": 5, "f": 4917},
  {"h": "烃", "p": "tīng", "sp": "ting", "r": "火", "sc": 9, "l": 5, "f": 4918},
  {"h": "俣", "p": "yǔ", "sp": "yu", "r": "人", "sc": 9, "l": 5, "f": 4919},
  {"h": "嬷", "p": "mā", "sp": "ma", "r": "女", "sc": 17, "l": 5, "f": 4920},
  {"h": "肱", "p": "gōng", "sp": "gong", "r": "肉", "sc": 8, "l": 5, "f": 4921},
  {"h": "鸢", "p": "yuān", "sp": "yuan", "r": "鳥", "sc": 8, "l": 5, "f": 4922},
  {"h": "笫", "p": "zǐ", "sp": "zi", "r": "竹", "sc": 10, "l": 5, "f": 4923},
  {"h": "痤", "p": "cuó", "sp": "cuo", "r": "疒", "sc": 12, "l": 5, "f": 4924},
  {"h": "陰", "p": "yīn", "sp": "yin", "r": "阜", "sc": 10, "l": 5, "f": 4925},
  {"h": "菏", "p": "hé", "sp": "he", "r": "艸", "sc": 11, "l": 5, "f": 4926},
  {"h": "莆", "p": "pú", "sp": "pu", "r": "艸", "sc": 10, "l": 5, "f": 4927},
  {"h": "芨", "p": "jī", "sp": "ji", "r": "艸", "sc": 6, "l": 5, "f": 4928},
  {"h": "阕", "p": "què", "sp": "que", "r": "門", "sc": 12, "l": 5, "f": 4929},
  {"h": "砣", "p": "tuó", "sp": "tuo", "r": "石", "sc": 10, "l": 5, "f": 4930},
  {"h": "碜", "p": "chěn", "sp": "chen", "r": "石", "sc": 13, "l": 5, "f": 4931},
  {"h": "鼹", "p": "yǎn", "sp": "yan", "r": "鼠", "sc": 23, "l": 5, "f": 4932},
  {"h": "長", "p": "zhǎng", "sp": "zhang", "r": "長", "sc": 8, "l": 5, "f": 4933},
  {"h": "猷", "p": "yóu", "sp": "you", "r": "犬", "sc": 13, "l": 5, "f": 4934},
  {"h": "竽", "p": "yú", "sp": "yu", "r": "竹", "sc": 9, "l": 5, "f": 4935},
  {"h": "舸", "p": "gě", "sp": "ge", "r": "舟", "sc": 11, "l": 5, "f": 4936},
  {"h": "诓", "p": "kuāng", "sp": "kuang", "r": "言", "sc": 8, "l": 5, "f": 4937},
  {"h": "錾", "p": "zàn", "sp": "zan", "r": "金", "sc": 16, "l": 5, "f": 4938},
  {"h": "淬", "p": "cuì", "sp": "cui", "r": "水", "sc": 11, "l": 5, "f": 4939},
  {"h": "隗", "p": "kuí", "sp": "kui", "r": "阜", "sc": 11, "l": 5, "f": 4940},
  {"h": "悌", "p": "tì", "sp": "ti", "r": "心", "sc": 10, "l": 5, "f": 4941},
  {"h": "姘", "p": "pīn", "sp": "pin", "r": "女", "sc": 9, "l": 5, "f": 4942},
  {"h": "槭", "p": "qī", "sp": "qi", "r": "木", "sc": 15, "l": 5, "f": 4943},
  {"h": "邈", "p": "miǎo", "sp": "miao", "r": "辵", "sc": 17, "l": 5, "f": 4944},
  {"h": "婕", "p": "jié", "sp": "jie", "r": "女", "sc": 11, "l": 5, "f": 4945},
  {"h": "歙", "p": "shè", "sp": "she", "r": "欠", "sc": 16, "l": 5, "f": 4946},
  {"h": "稹", "p": "zhěn", "sp": "zhen", "r": "禾", "sc": 15, "l": 5, "f": 4947},
  {"h": "蹴", "p": "cù", "sp": "cu", "r": "足", "sc": 19, "l": 5, "f": 4948},
  {"h": "砒", "p": "pī", "sp": "pi", "r": "石", "sc": 9, "l": 5, "f": 4949},
  {"h": "痈", "p": "yōng", "sp": "yong", "r": "疒", "sc": 10, "l": 5, "f": 4950},
  {"h": "镏", "p": "liú", "sp": "liu", "r": "金", "sc": 15, "l": 5, "f": 4951},
  {"h": "羯", "p": "jié", "sp": "jie", "r": "羊", "sc": 15, "l": 5, "f": 4952},
  {"h": "豕", "p": "shǐ", "sp": "shi", "r": "豕", "sc": 7, "l": 5, "f": 4953},
  {"h": "鲂", "p": "fáng", "sp": "fang", "r": "魚", "sc": 12, "l": 5, "f": 4954},
  {"h": "蓖", "p": "bì", "sp": "bi", "r": "艸", "sc": 13, "l": 5, "f": 4955},
  {"h": "匦", "p": "guǐ", "sp": "gui", "r": "匚", "sc": 8, "l": 5, "f": 4956},
  {"h": "笤", "p": "tiáo", "sp": "tiao", "r": "竹", "sc": 11, "l": 5, "f": 4957},
  {"h": "峥", "p": "zhēng", "sp": "zheng", "r": "山", "sc": 9, "l": 5, "f": 4958},
  {"h": "徭", "p": "yáo", "sp": "yao", "r": "彳", "sc": 13, "l": 5, "f": 4959},
  {"h": "浃", "p": "jiā", "sp": "jia", "r": "水", "sc": 9, "l": 5, "f": 4960},
  {"h": "烊", "p": "yáng", "sp": "yang", "r": "火", "sc": 10, "l": 5, "f": 4961},
  {"h": "補", "p": "bǔ", "sp": "bu", "r": "衣", "sc": 12, "l": 5, "f": 4962},
  {"h": "窸", "p": "xī", "sp": "xi", "r": "穴", "sc": 16, "l": 5, "f": 4963},
  {"h": "酆", "p": "fēng", "sp": "feng", "r": "邑", "sc": 20, "l": 5, "f": 4964},
  {"h": "缢", "p": "yì", "sp": "yi", "r": "糸", "sc": 13, "l": 5, "f": 4965},
  {"h": "褓", "p": "bǎo", "sp": "bao", "r": "衣", "sc": 14, "l": 5, "f": 4966},
  {"h": "蚨", "p": "fú", "sp": "fu", "r": "虫", "sc": 10, "l": 5, "f": 4967},
  {"h": "翳", "p": "yì", "sp": "yi", "r": "羽", "sc": 17, "l": 5, "f": 4968},
  {"h": "趔", "p": "liè", "sp": "lie", "r": "走", "sc": 13, "l": 5, "f": 4969},
  {"h": "炔", "p": "guì", "sp": "gui", "r": "火", "sc": 8, "l": 5, "f": 4970},
  {"h": "誊", "p": "téng", "sp": "teng", "r": "言", "sc": 13, "l": 5, "f": 4971},
  {"h": "赜", "p": "zé", "sp": "ze", "r": "匚", "sc": 15, "l": 5, "f": 4972},
  {"h": "仃", "p": "dīng", "sp": "ding", "r": "人", "sc": 4, "l": 5, "f": 4973},
  {"h": "勖", "p": "xù", "sp": "xu", "r": "力", "sc": 11, "l": 5, "f": 4974},
  {"h": "葺", "p": "qì", "sp": "qi", "r": "艸", "sc": 12, "l": 5, "f": 4975},
  {"h": "蚴", "p": "yòu", "sp": "you", "r": "虫", "sc": 11, "l": 5, "f": 4976},
  {"h": "泷", "p": "lóng", "sp": "long", "r": "水", "sc": 8, "l": 5, "f": 4977},
  {"h": "蛴", "p": "qí", "sp": "qi", "r": "虫", "sc": 12, "l": 5, "f": 4978},
  {"h": "結", "p": "jié", "sp": "jie", "r": "糸", "sc": 12, "l": 5, "f": 4979},
  {"h": "媸", "p": "chī", "sp": "chi", "r": "女", "sc": 13, "l": 5, "f": 4980},
  {"h": "俳", "p": "pái", "sp": "pai", "r": "人", "sc": 10, "l": 5, "f": 4981},
  {"h": "诖", "p": "guà", "sp": "gua", "r": "言", "sc": 8, "l": 5, "f": 4982},
  {"h": "茑", "p": "niǎo", "sp": "niao", "r": "艸", "sc": 8, "l": 5, "f": 4983},
  {"h": "逡", "p": "qūn", "sp": "qun", "r": "辵", "sc": 10, "l": 5, "f": 4984},
  {"h": "孱", "p": "càn", "sp": "can", "r": "子", "sc": 12, "l": 5, "f": 4985},
  {"h": "砦", "p": "zhài", "sp": "zhai", "r": "石", "sc": 11, "l": 5, "f": 4986},
  {"h": "跸", "p": "bì", "sp": "bi", "r": "足", "sc": 13, "l": 5, "f": 4987},
  {"h": "祜", "p": "hù", "sp": "hu", "r": "示", "sc": 9, "l": 5, "f": 4988},
  {"h": "伉", "p": "kàng", "sp": "kang", "r": "人", "sc": 6, "l": 5, "f": 4989},
  {"h": "溴", "p": "xiù", "sp": "xiu", "r": "水", "sc": 13, "l": 5, "f": 4990},
  {"h": "屐", "p": "jī", "sp": "ji", "r": "尸", "sc": 10, "l": 5, "f": 4991},
  {"h": "飚", "p": "biāo", "sp": "biao", "r": "風", "sc": 16, "l": 5, "f": 4992},
  {"h": "蛞", "p": "kuò", "sp": "kuo", "r": "虫", "sc": 12, "l": 5, "f": 4993},
  {"h": "鏡", "p": "jìng", "sp": "jing", "r": "金", "sc": 19, "l": 5, "f": 4994},
  {"h": "掮", "p": "qián", "sp": "qian", "r": "手", "sc": 11, "l": 5, "f": 4995},
  {"h": "崆", "p": "kōng", "sp": "kong", "r": "山", "sc": 11, "l": 5, "f": 4996},
  {"h": "庾", "p": "yǔ", "sp": "yu", "r": "广", "sc": 11, "l": 5, "f": 4997},
  {"h": "橛", "p": "jué", "sp": "jue", "r": "木", "sc": 16, "l": 5, "f": 4998},
  {"h": "矸", "p": "gān", "sp": "gan", "r": "石", "sc": 8, "l": 5, "f": 4999},
  {"h": "鸨", "p": "bǎo", "sp": "bao", "r": "鳥", "sc": 9, "l": 5, "f": 5000},
  {"h": "圻", "p": "qí", "sp": "qi", "r": "土", "sc": 7, "l": 5, "f": 5001},
  {"h": "缂", "p": "kè", "sp": "ke", "r": "糸", "sc": 12, "l": 5, "f": 5002},
  {"h": "蒯", "p": "kuǎi", "sp": "kuai", "r": "艸", "sc": 13, "l": 5, "f": 5003},
  {"h": "诹", "p": "zōu", "sp": "zou", "r": "言", "sc": 10, "l": 5, "f": 5004},
  {
    "h": "啭",
    "p": "zhuàn",
    "sp": "zhuan",
    "r": "口",
    "sc": 11,
    "l": 5,
    "f": 5005
  },
  {"h": "饧", "p": "táng", "sp": "tang", "r": "食", "sc": 6, "l": 5, "f": 5006},
  {"h": "镉", "p": "gé", "sp": "ge", "r": "金", "sc": 15, "l": 5, "f": 5007},
  {"h": "鸪", "p": "gū", "sp": "gu", "r": "鳥", "sc": 10, "l": 5, "f": 5008},
  {
    "h": "蛩",
    "p": "qióng",
    "sp": "qiong",
    "r": "虫",
    "sc": 12,
    "l": 5,
    "f": 5009
  },
  {"h": "蠖", "p": "huò", "sp": "huo", "r": "虫", "sc": 19, "l": 5, "f": 5010},
  {"h": "說", "p": "shuō", "sp": "shuo", "r": "言", "sc": 14, "l": 5, "f": 5011},
  {"h": "劭", "p": "shào", "sp": "shao", "r": "力", "sc": 7, "l": 5, "f": 5012},
  {"h": "哐", "p": "kuāng", "sp": "kuang", "r": "口", "sc": 9, "l": 5, "f": 5013},
  {"h": "崧", "p": "sōng", "sp": "song", "r": "山", "sc": 11, "l": 5, "f": 5014},
  {"h": "杼", "p": "zhù", "sp": "zhu", "r": "木", "sc": 8, "l": 5, "f": 5015},
  {"h": "棂", "p": "líng", "sp": "ling", "r": "木", "sc": 11, "l": 5, "f": 5016},
  {"h": "螫", "p": "shì", "sp": "shi", "r": "虫", "sc": 17, "l": 5, "f": 5017},
  {"h": "龃", "p": "jǔ", "sp": "ju", "r": "齒", "sc": 13, "l": 5, "f": 5018},
  {"h": "饔", "p": "yōng", "sp": "yong", "r": "食", "sc": 22, "l": 5, "f": 5019},
  {
    "h": "遑",
    "p": "huáng",
    "sp": "huang",
    "r": "辵",
    "sc": 12,
    "l": 5,
    "f": 5020
  },
  {"h": "颢", "p": "hào", "sp": "hao", "r": "頁", "sc": 18, "l": 5, "f": 5021},
  {"h": "腱", "p": "jiàn", "sp": "jian", "r": "肉", "sc": 12, "l": 5, "f": 5022},
  {
    "h": "襁",
    "p": "qiǎng",
    "sp": "qiang",
    "r": "衣",
    "sc": 17,
    "l": 5,
    "f": 5023
  },
  {"h": "忾", "p": "kài", "sp": "kai", "r": "心", "sc": 7, "l": 5, "f": 5024},
  {"h": "濠", "p": "háo", "sp": "hao", "r": "水", "sc": 17, "l": 5, "f": 5025},
  {"h": "牝", "p": "pìn", "sp": "pin", "r": "牛", "sc": 6, "l": 5, "f": 5026},
  {"h": "蛄", "p": "gū", "sp": "gu", "r": "虫", "sc": 11, "l": 5, "f": 5027},
  {"h": "鲆", "p": "píng", "sp": "ping", "r": "魚", "sc": 13, "l": 5, "f": 5028},
  {"h": "嗄", "p": "á", "sp": "a", "r": "口", "sc": 13, "l": 5, "f": 5029},
  {"h": "灏", "p": "hào", "sp": "hao", "r": "水", "sc": 21, "l": 5, "f": 5030},
  {"h": "疥", "p": "jiè", "sp": "jie", "r": "疒", "sc": 9, "l": 5, "f": 5031},
  {"h": "苜", "p": "mù", "sp": "mu", "r": "艸", "sc": 8, "l": 5, "f": 5032},
  {"h": "荞", "p": "qiáo", "sp": "qiao", "r": "艸", "sc": 9, "l": 5, "f": 5033},
  {"h": "嘣", "p": "bēng", "sp": "beng", "r": "口", "sc": 14, "l": 5, "f": 5034},
  {"h": "夤", "p": "yín", "sp": "yin", "r": "夕", "sc": 14, "l": 5, "f": 5035},
  {"h": "砝", "p": "fá", "sp": "fa", "r": "石", "sc": 10, "l": 5, "f": 5036},
  {"h": "颞", "p": "niè", "sp": "nie", "r": "頁", "sc": 16, "l": 5, "f": 5037},
  {"h": "開", "p": "kāi", "sp": "kai", "r": "門", "sc": 12, "l": 5, "f": 5038},
  {"h": "忤", "p": "wǔ", "sp": "wu", "r": "心", "sc": 7, "l": 5, "f": 5039},
  {"h": "遢", "p": "tà", "sp": "ta", "r": "辵", "sc": 13, "l": 5, "f": 5040},
  {"h": "旎", "p": "nǐ", "sp": "ni", "r": "方", "sc": 11, "l": 5, "f": 5041},
  {"h": "瘛", "p": "chì", "sp": "chi", "r": "疒", "sc": 15, "l": 5, "f": 5042},
  {
    "h": "魉",
    "p": "liǎng",
    "sp": "liang",
    "r": "鬼",
    "sc": 16,
    "l": 5,
    "f": 5043
  },
  {"h": "辇", "p": "niǎn", "sp": "nian", "r": "車", "sc": 12, "l": 5, "f": 5044},
  {"h": "見", "p": "jiàn", "sp": "jian", "r": "見", "sc": 7, "l": 5, "f": 5045},
  {"h": "瓤", "p": "ráng", "sp": "rang", "r": "瓜", "sc": 22, "l": 5, "f": 5046},
  {"h": "荥", "p": "xíng", "sp": "xing", "r": "艸", "sc": 9, "l": 5, "f": 5047},
  {"h": "涫", "p": "guàn", "sp": "guan", "r": "水", "sc": 11, "l": 5, "f": 5048},
  {"h": "娌", "p": "lǐ", "sp": "li", "r": "女", "sc": 10, "l": 5, "f": 5049},
  {"h": "氚", "p": "chuān", "sp": "chuan", "r": "气", "sc": 7, "l": 5, "f": 5050},
  {"h": "臁", "p": "lián", "sp": "lian", "r": "肉", "sc": 17, "l": 5, "f": 5051},
  {"h": "毂", "p": "gǔ", "sp": "gu", "r": "殳", "sc": 13, "l": 5, "f": 5052},
  {"h": "碇", "p": "dìng", "sp": "ding", "r": "石", "sc": 13, "l": 5, "f": 5053},
  {"h": "毖", "p": "bì", "sp": "bi", "r": "比", "sc": 9, "l": 5, "f": 5054},
  {"h": "壅", "p": "yōng", "sp": "yong", "r": "土", "sc": 16, "l": 5, "f": 5055},
  {"h": "吡", "p": "bǐ", "sp": "bi", "r": "口", "sc": 7, "l": 5, "f": 5056},
  {"h": "缛", "p": "rù", "sp": "ru", "r": "糸", "sc": 13, "l": 5, "f": 5057},
  {"h": "玮", "p": "wěi", "sp": "wei", "r": "玉", "sc": 8, "l": 5, "f": 5058},
  {
    "h": "羟",
    "p": "qiǎng",
    "sp": "qiang",
    "r": "羊",
    "sc": 11,
    "l": 5,
    "f": 5059
  },
  {"h": "還", "p": "hái", "sp": "hai", "r": "辵", "sc": 16, "l": 5, "f": 5060},
  {"h": "珈", "p": "jiā", "sp": "jia", "r": "玉", "sc": 9, "l": 5, "f": 5061},
  {"h": "颀", "p": "qí", "sp": "qi", "r": "斤", "sc": 10, "l": 5, "f": 5062},
  {"h": "虼", "p": "gè", "sp": "ge", "r": "虫", "sc": 9, "l": 5, "f": 5063},
  {"h": "祇", "p": "qí", "sp": "qi", "r": "示", "sc": 8, "l": 5, "f": 5064},
  {"h": "佝", "p": "gōu", "sp": "gou", "r": "人", "sc": 7, "l": 5, "f": 5065},
  {"h": "翕", "p": "xī", "sp": "xi", "r": "羽", "sc": 12, "l": 5, "f": 5066},
  {"h": "遴", "p": "lín", "sp": "lin", "r": "辵", "sc": 15, "l": 5, "f": 5067},
  {"h": "珏", "p": "jué", "sp": "jue", "r": "玉", "sc": 9, "l": 5, "f": 5068},
  {"h": "郛", "p": "fú", "sp": "fu", "r": "邑", "sc": 9, "l": 5, "f": 5069},
  {"h": "較", "p": "jiào", "sp": "jiao", "r": "車", "sc": 13, "l": 5, "f": 5070},
  {"h": "驗", "p": "yàn", "sp": "yan", "r": "馬", "sc": 23, "l": 5, "f": 5071},
  {"h": "玖", "p": "jiǔ", "sp": "jiu", "r": "玉", "sc": 7, "l": 5, "f": 5072},
  {"h": "蹇", "p": "jiǎn", "sp": "jian", "r": "足", "sc": 17, "l": 5, "f": 5073},
  {"h": "逋", "p": "bū", "sp": "bu", "r": "辵", "sc": 10, "l": 5, "f": 5074},
  {
    "h": "氅",
    "p": "chǎng",
    "sp": "chang",
    "r": "毛",
    "sc": 16,
    "l": 5,
    "f": 5075
  },
  {"h": "粽", "p": "zòng", "sp": "zong", "r": "米", "sc": 14, "l": 5, "f": 5076},
  {"h": "诂", "p": "gǔ", "sp": "gu", "r": "言", "sc": 7, "l": 5, "f": 5077},
  {"h": "岢", "p": "kě", "sp": "ke", "r": "山", "sc": 8, "l": 5, "f": 5078},
  {"h": "聒", "p": "guā", "sp": "gua", "r": "耳", "sc": 12, "l": 5, "f": 5079},
  {"h": "髁", "p": "kē", "sp": "ke", "r": "骨", "sc": 17, "l": 5, "f": 5080},
  {"h": "黍", "p": "shǔ", "sp": "shu", "r": "黍", "sc": 12, "l": 5, "f": 5081},
  {"h": "芾", "p": "fèi", "sp": "fei", "r": "艸", "sc": 7, "l": 5, "f": 5082},
  {"h": "淝", "p": "féi", "sp": "fei", "r": "水", "sc": 11, "l": 5, "f": 5083},
  {"h": "鲎", "p": "hòu", "sp": "hou", "r": "魚", "sc": 13, "l": 5, "f": 5084},
  {"h": "鞣", "p": "róu", "sp": "rou", "r": "革", "sc": 18, "l": 5, "f": 5085},
  {"h": "髋", "p": "kuān", "sp": "kuan", "r": "骨", "sc": 19, "l": 5, "f": 5086},
  {"h": "闳", "p": "hóng", "sp": "hong", "r": "門", "sc": 7, "l": 5, "f": 5087},
  {"h": "潆", "p": "yíng", "sp": "ying", "r": "水", "sc": 14, "l": 5, "f": 5088},
  {"h": "汨", "p": "mì", "sp": "mi", "r": "水", "sc": 7, "l": 5, "f": 5089},
  {"h": "胍", "p": "guā", "sp": "gua", "r": "肉", "sc": 9, "l": 5, "f": 5090},
  {"h": "阏", "p": "è", "sp": "e", "r": "門", "sc": 11, "l": 5, "f": 5091},
  {"h": "钤", "p": "qián", "sp": "qian", "r": "金", "sc": 9, "l": 5, "f": 5092},
  {"h": "鹜", "p": "wù", "sp": "wu", "r": "鳥", "sc": 14, "l": 5, "f": 5093},
  {"h": "鬈", "p": "quán", "sp": "quan", "r": "髟", "sc": 18, "l": 5, "f": 5094},
  {"h": "铵", "p": "ǎn", "sp": "an", "r": "金", "sc": 11, "l": 5, "f": 5095},
  {"h": "戬", "p": "jiǎn", "sp": "jian", "r": "戈", "sc": 14, "l": 5, "f": 5096},
  {"h": "點", "p": "diǎn", "sp": "dian", "r": "黑", "sc": 17, "l": 5, "f": 5097},
  {"h": "崮", "p": "gù", "sp": "gu", "r": "山", "sc": 11, "l": 5, "f": 5098},
  {"h": "枰", "p": "píng", "sp": "ping", "r": "木", "sc": 9, "l": 5, "f": 5099},
  {
    "h": "樯",
    "p": "qiáng",
    "sp": "qiang",
    "r": "木",
    "sc": 15,
    "l": 5,
    "f": 5100
  },
  {"h": "脍", "p": "kuài", "sp": "kuai", "r": "肉", "sc": 10, "l": 5, "f": 5101},
  {"h": "畲", "p": "shē", "sp": "she", "r": "田", "sc": 12, "l": 5, "f": 5102},
  {"h": "衾", "p": "qīn", "sp": "qin", "r": "衣", "sc": 10, "l": 5, "f": 5103},
  {"h": "蹼", "p": "pǔ", "sp": "pu", "r": "足", "sc": 19, "l": 5, "f": 5104},
  {"h": "題", "p": "tí", "sp": "ti", "r": "頁", "sc": 18, "l": 5, "f": 5105},
  {"h": "劬", "p": "qú", "sp": "qu", "r": "力", "sc": 7, "l": 5, "f": 5106},
  {"h": "咭", "p": "jī", "sp": "ji", "r": "口", "sc": 9, "l": 5, "f": 5107},
  {"h": "囫", "p": "hú", "sp": "hu", "r": "囗", "sc": 7, "l": 5, "f": 5108},
  {"h": "洱", "p": "ěr", "sp": "er", "r": "水", "sc": 9, "l": 5, "f": 5109},
  {"h": "刎", "p": "wěn", "sp": "wen", "r": "刀", "sc": 6, "l": 5, "f": 5110},
  {"h": "芏", "p": "dù", "sp": "du", "r": "艸", "sc": 6, "l": 5, "f": 5111},
  {"h": "琊", "p": "yá", "sp": "ya", "r": "玉", "sc": 10, "l": 5, "f": 5112},
  {"h": "碚", "p": "bèi", "sp": "bei", "r": "石", "sc": 13, "l": 5, "f": 5113},
  {"h": "鳕", "p": "xuě", "sp": "xue", "r": "魚", "sc": 19, "l": 5, "f": 5114},
  {"h": "谪", "p": "zhé", "sp": "zhe", "r": "言", "sc": 13, "l": 5, "f": 5115},
  {"h": "芎", "p": "qiōng", "sp": "qiong", "r": "艸", "sc": 6, "l": 5, "f": 5116},
  {"h": "恂", "p": "xún", "sp": "xun", "r": "心", "sc": 9, "l": 5, "f": 5117},
  {"h": "槿", "p": "jǐn", "sp": "jin", "r": "木", "sc": 15, "l": 5, "f": 5118},
  {"h": "鲢", "p": "lián", "sp": "lian", "r": "魚", "sc": 15, "l": 5, "f": 5119},
  {"h": "鲧", "p": "gǔn", "sp": "gun", "r": "魚", "sc": 15, "l": 5, "f": 5120},
  {"h": "嘧", "p": "mì", "sp": "mi", "r": "口", "sc": 14, "l": 5, "f": 5121},
  {"h": "绀", "p": "gàn", "sp": "gan", "r": "糸", "sc": 8, "l": 5, "f": 5122},
  {"h": "郦", "p": "lì", "sp": "li", "r": "邑", "sc": 9, "l": 5, "f": 5123},
  {"h": "噱", "p": "jué", "sp": "jue", "r": "口", "sc": 16, "l": 5, "f": 5124},
  {"h": "浠", "p": "xī", "sp": "xi", "r": "水", "sc": 10, "l": 5, "f": 5125},
  {"h": "潸", "p": "shān", "sp": "shan", "r": "水", "sc": 15, "l": 5, "f": 5126},
  {"h": "跏", "p": "jiā", "sp": "jia", "r": "足", "sc": 12, "l": 5, "f": 5127},
  {"h": "鲶", "p": "nián", "sp": "nian", "r": "魚", "sc": 16, "l": 5, "f": 5128},
  {"h": "矍", "p": "jué", "sp": "jue", "r": "目", "sc": 20, "l": 5, "f": 5129},
  {"h": "苌", "p": "cháng", "sp": "chang", "r": "艸", "sc": 7, "l": 5, "f": 5130},
  {"h": "抻", "p": "chēn", "sp": "chen", "r": "手", "sc": 8, "l": 5, "f": 5131},
  {"h": "琰", "p": "yǎn", "sp": "yan", "r": "玉", "sc": 12, "l": 5, "f": 5132},
  {"h": "鹚", "p": "cí", "sp": "ci", "r": "鳥", "sc": 14, "l": 5, "f": 5133},
  {"h": "龆", "p": "tiáo", "sp": "tiao", "r": "齒", "sc": 13, "l": 5, "f": 5134},
  {"h": "臬", "p": "niè", "sp": "nie", "r": "自", "sc": 10, "l": 5, "f": 5135},
  {"h": "芄", "p": "wán", "sp": "wan", "r": "艸", "sc": 6, "l": 5, "f": 5136},
  {"h": "呔", "p": "dāi", "sp": "dai", "r": "口", "sc": 7, "l": 5, "f": 5137},
  {"h": "雒", "p": "luò", "sp": "luo", "r": "隹", "sc": 14, "l": 5, "f": 5138},
  {
    "h": "觞",
    "p": "shāng",
    "sp": "shang",
    "r": "角",
    "sc": 12,
    "l": 5,
    "f": 5139
  },
  {"h": "钒", "p": "fán", "sp": "fan", "r": "金", "sc": 8, "l": 5, "f": 5140},
  {"h": "饫", "p": "yù", "sp": "yu", "r": "食", "sc": 7, "l": 5, "f": 5141},
  {"h": "阒", "p": "qù", "sp": "qu", "r": "門", "sc": 12, "l": 5, "f": 5142},
  {"h": "槎", "p": "chá", "sp": "cha", "r": "木", "sc": 13, "l": 5, "f": 5143},
  {"h": "鸩", "p": "zhèn", "sp": "zhen", "r": "鳥", "sc": 9, "l": 5, "f": 5144},
  {
    "h": "舂",
    "p": "chōng",
    "sp": "chong",
    "r": "臼",
    "sc": 11,
    "l": 5,
    "f": 5145
  },
  {"h": "谠", "p": "dǎng", "sp": "dang", "r": "言", "sc": 12, "l": 5, "f": 5146},
  {"h": "阡", "p": "qiān", "sp": "qian", "r": "阜", "sc": 5, "l": 5, "f": 5147},
  {"h": "莒", "p": "jǔ", "sp": "ju", "r": "艸", "sc": 9, "l": 5, "f": 5148},
  {"h": "萸", "p": "yú", "sp": "yu", "r": "艸", "sc": 11, "l": 5, "f": 5149},
  {"h": "妗", "p": "jìn", "sp": "jin", "r": "女", "sc": 7, "l": 5, "f": 5150},
  {"h": "稔", "p": "rěn", "sp": "ren", "r": "禾", "sc": 13, "l": 5, "f": 5151},
  {"h": "穰", "p": "ráng", "sp": "rang", "r": "禾", "sc": 22, "l": 5, "f": 5152},
  {"h": "蚧", "p": "jiè", "sp": "jie", "r": "虫", "sc": 10, "l": 5, "f": 5153},
  {"h": "餍", "p": "yàn", "sp": "yan", "r": "食", "sc": 15, "l": 5, "f": 5154},
  {"h": "谯", "p": "qiáo", "sp": "qiao", "r": "言", "sc": 14, "l": 5, "f": 5155},
  {"h": "芗", "p": "xiāng", "sp": "xiang", "r": "艸", "sc": 6, "l": 5, "f": 5156},
  {"h": "菸", "p": "yān", "sp": "yan", "r": "艸", "sc": 11, "l": 5, "f": 5157},
  {"h": "葩", "p": "pā", "sp": "pa", "r": "艸", "sc": 12, "l": 5, "f": 5158},
  {"h": "踔", "p": "chuō", "sp": "chuo", "r": "足", "sc": 15, "l": 5, "f": 5159},
  {"h": "厣", "p": "yǎn", "sp": "yan", "r": "厂", "sc": 11, "l": 5, "f": 5160},
  {"h": "佻", "p": "tiāo", "sp": "tiao", "r": "人", "sc": 8, "l": 5, "f": 5161},
  {"h": "嘌", "p": "piào", "sp": "piao", "r": "口", "sc": 14, "l": 5, "f": 5162},
  {"h": "饩", "p": "xì", "sp": "xi", "r": "食", "sc": 7, "l": 5, "f": 5163},
  {"h": "钏", "p": "chuàn", "sp": "chuan", "r": "金", "sc": 8, "l": 5, "f": 5164},
  {"h": "蠓", "p": "měng", "sp": "meng", "r": "虫", "sc": 19, "l": 5, "f": 5165},
  {"h": "黩", "p": "dú", "sp": "du", "r": "黑", "sc": 20, "l": 5, "f": 5166},
  {"h": "倨", "p": "jù", "sp": "ju", "r": "人", "sc": 10, "l": 5, "f": 5167},
  {
    "h": "腸",
    "p": "cháng",
    "sp": "chang",
    "r": "肉",
    "sc": 13,
    "l": 5,
    "f": 5168
  },
  {"h": "缬", "p": "xié", "sp": "xie", "r": "糸", "sc": 15, "l": 5, "f": 5169},
  {"h": "殚", "p": "dān", "sp": "dan", "r": "歹", "sc": 12, "l": 5, "f": 5170},
  {"h": "钿", "p": "diàn", "sp": "dian", "r": "金", "sc": 10, "l": 5, "f": 5171},
  {"h": "鎏", "p": "liú", "sp": "liu", "r": "金", "sc": 18, "l": 5, "f": 5172},
  {"h": "恁", "p": "nèn", "sp": "nen", "r": "心", "sc": 10, "l": 5, "f": 5173},
  {"h": "藿", "p": "huò", "sp": "huo", "r": "艸", "sc": 19, "l": 5, "f": 5174},
  {"h": "囟", "p": "xìn", "sp": "xin", "r": "囗", "sc": 6, "l": 5, "f": 5175},
  {
    "h": "鄣",
    "p": "zhāng",
    "sp": "zhang",
    "r": "邑",
    "sc": 13,
    "l": 5,
    "f": 5176
  },
  {"h": "呋", "p": "fū", "sp": "fu", "r": "口", "sc": 7, "l": 5, "f": 5177},
  {"h": "婺", "p": "wù", "sp": "wu", "r": "女", "sc": 12, "l": 5, "f": 5178},
  {
    "h": "绱",
    "p": "shàng",
    "sp": "shang",
    "r": "糸",
    "sc": 11,
    "l": 5,
    "f": 5179
  },
  {"h": "瓯", "p": "ōu", "sp": "ou", "r": "瓦", "sc": 8, "l": 5, "f": 5180},
  {"h": "旃", "p": "zhān", "sp": "zhan", "r": "方", "sc": 10, "l": 5, "f": 5181},
  {"h": "锶", "p": "sī", "sp": "si", "r": "金", "sc": 14, "l": 5, "f": 5182},
  {"h": "酩", "p": "mǐng", "sp": "ming", "r": "酉", "sc": 13, "l": 5, "f": 5183},
  {"h": "恹", "p": "yān", "sp": "yan", "r": "心", "sc": 9, "l": 5, "f": 5184},
  {"h": "逶", "p": "wēi", "sp": "wei", "r": "辵", "sc": 11, "l": 5, "f": 5185},
  {"h": "缦", "p": "màn", "sp": "man", "r": "糸", "sc": 14, "l": 5, "f": 5186},
  {"h": "鸹", "p": "guā", "sp": "gua", "r": "鳥", "sc": 11, "l": 5, "f": 5187},
  {"h": "螟", "p": "míng", "sp": "ming", "r": "虫", "sc": 16, "l": 5, "f": 5188},
  {"h": "菟", "p": "tú", "sp": "tu", "r": "艸", "sc": 11, "l": 5, "f": 5189},
  {"h": "阗", "p": "tián", "sp": "tian", "r": "門", "sc": 13, "l": 5, "f": 5190},
  {"h": "濉", "p": "suī", "sp": "sui", "r": "水", "sc": 16, "l": 5, "f": 5191},
  {"h": "篑", "p": "kuì", "sp": "kui", "r": "竹", "sc": 15, "l": 5, "f": 5192},
  {"h": "醪", "p": "láo", "sp": "lao", "r": "酉", "sc": 18, "l": 5, "f": 5193},
  {"h": "鲛", "p": "jiāo", "sp": "jiao", "r": "魚", "sc": 14, "l": 5, "f": 5194},
  {"h": "讦", "p": "jié", "sp": "jie", "r": "言", "sc": 5, "l": 5, "f": 5195},
  {"h": "媪", "p": "ǎo", "sp": "ao", "r": "女", "sc": 12, "l": 5, "f": 5196},
  {"h": "邬", "p": "wū", "sp": "wu", "r": "邑", "sc": 6, "l": 5, "f": 5197},
  {"h": "殇", "p": "shāng", "sp": "shang", "r": "歹", "sc": 9, "l": 5, "f": 5198},
  {"h": "鄯", "p": "shàn", "sp": "shan", "r": "邑", "sc": 14, "l": 5, "f": 5199},
  {"h": "芡", "p": "qiàn", "sp": "qian", "r": "艸", "sc": 7, "l": 5, "f": 5200},
  {"h": "嫠", "p": "lí", "sp": "li", "r": "女", "sc": 14, "l": 5, "f": 5201},
  {"h": "肼", "p": "jǐng", "sp": "jing", "r": "肉", "sc": 8, "l": 5, "f": 5202},
  {"h": "峤", "p": "jiào", "sp": "jiao", "r": "山", "sc": 9, "l": 5, "f": 5203},
  {"h": "矽", "p": "xì", "sp": "xi", "r": "石", "sc": 8, "l": 5, "f": 5204},
  {"h": "讧", "p": "hòng", "sp": "hong", "r": "言", "sc": 5, "l": 5, "f": 5205},
  {"h": "掼", "p": "guàn", "sp": "guan", "r": "手", "sc": 11, "l": 5, "f": 5206},
  {"h": "焖", "p": "mèn", "sp": "men", "r": "火", "sc": 11, "l": 5, "f": 5207},
  {"h": "愆", "p": "qiān", "sp": "qian", "r": "心", "sc": 13, "l": 5, "f": 5208},
  {"h": "聩", "p": "kuì", "sp": "kui", "r": "耳", "sc": 15, "l": 5, "f": 5209},
  {"h": "岘", "p": "xiàn", "sp": "xian", "r": "山", "sc": 7, "l": 5, "f": 5210},
  {"h": "靛", "p": "diàn", "sp": "dian", "r": "青", "sc": 16, "l": 5, "f": 5211},
  {
    "h": "菖",
    "p": "chāng",
    "sp": "chang",
    "r": "艸",
    "sc": 11,
    "l": 5,
    "f": 5212
  },
  {"h": "卟", "p": "bǔ", "sp": "bu", "r": "口", "sc": 5, "l": 5, "f": 5213},
  {"h": "姒", "p": "sì", "sp": "si", "r": "女", "sc": 7, "l": 5, "f": 5214},
  {"h": "杷", "p": "pá", "sp": "pa", "r": "木", "sc": 8, "l": 5, "f": 5215},
  {"h": "砉", "p": "huò", "sp": "huo", "r": "石", "sc": 9, "l": 5, "f": 5216},
  {"h": "袢", "p": "pàn", "sp": "pan", "r": "衣", "sc": 10, "l": 5, "f": 5217},
  {"h": "蚋", "p": "ruì", "sp": "rui", "r": "虫", "sc": 10, "l": 5, "f": 5218},
  {"h": "笳", "p": "jiā", "sp": "jia", "r": "竹", "sc": 11, "l": 5, "f": 5219},
  {"h": "挈", "p": "qiè", "sp": "qie", "r": "手", "sc": 10, "l": 5, "f": 5220},
  {"h": "關", "p": "guān", "sp": "guan", "r": "門", "sc": 19, "l": 5, "f": 5221},
  {"h": "踽", "p": "jǔ", "sp": "ju", "r": "足", "sc": 16, "l": 5, "f": 5222},
  {"h": "黾", "p": "miǎn", "sp": "mian", "r": "黽", "sc": 8, "l": 5, "f": 5223},
  {"h": "麼", "p": "me", "sp": "me", "r": "麻", "sc": 14, "l": 5, "f": 5224},
  {"h": "侩", "p": "kuài", "sp": "kuai", "r": "人", "sc": 8, "l": 5, "f": 5225},
  {"h": "凫", "p": "fú", "sp": "fu", "r": "几", "sc": 6, "l": 5, "f": 5226},
  {"h": "诔", "p": "lěi", "sp": "lei", "r": "言", "sc": 8, "l": 5, "f": 5227},
  {"h": "郯", "p": "tán", "sp": "tan", "r": "邑", "sc": 10, "l": 5, "f": 5228},
  {"h": "韪", "p": "wěi", "sp": "wei", "r": "韋", "sc": 13, "l": 5, "f": 5229},
  {"h": "挲", "p": "sā", "sp": "sa", "r": "手", "sc": 11, "l": 5, "f": 5230},
  {"h": "笪", "p": "dá", "sp": "da", "r": "竹", "sc": 11, "l": 5, "f": 5231},
  {"h": "鼋", "p": "yuán", "sp": "yuan", "r": "龜", "sc": 12, "l": 5, "f": 5232},
  {"h": "莜", "p": "yóu", "sp": "you", "r": "艸", "sc": 10, "l": 5, "f": 5233},
  {"h": "風", "p": "fēng", "sp": "feng", "r": "風", "sc": 9, "l": 5, "f": 5234},
  {"h": "菅", "p": "jiān", "sp": "jian", "r": "艸", "sc": 11, "l": 5, "f": 5235},
  {
    "h": "嵊",
    "p": "shèng",
    "sp": "sheng",
    "r": "山",
    "sc": 13,
    "l": 5,
    "f": 5236
  },
  {"h": "裢", "p": "lián", "sp": "lian", "r": "衣", "sc": 12, "l": 5, "f": 5237},
  {"h": "趿", "p": "tā", "sp": "ta", "r": "足", "sc": 10, "l": 5, "f": 5238},
  {"h": "箸", "p": "zhù", "sp": "zhu", "r": "竹", "sc": 14, "l": 5, "f": 5239},
  {"h": "莴", "p": "wō", "sp": "wo", "r": "艸", "sc": 10, "l": 5, "f": 5240},
  {"h": "莠", "p": "yǒu", "sp": "you", "r": "艸", "sc": 10, "l": 5, "f": 5241},
  {"h": "阌", "p": "wén", "sp": "wen", "r": "門", "sc": 11, "l": 5, "f": 5242},
  {"h": "旯", "p": "lá", "sp": "la", "r": "日", "sc": 6, "l": 5, "f": 5243},
  {"h": "圜", "p": "huán", "sp": "huan", "r": "囗", "sc": 16, "l": 5, "f": 5244},
  {"h": "涪", "p": "fú", "sp": "fu", "r": "水", "sc": 11, "l": 5, "f": 5245},
  {"h": "赍", "p": "jī", "sp": "ji", "r": "貝", "sc": 12, "l": 5, "f": 5246},
  {"h": "柞", "p": "zhà", "sp": "zha", "r": "木", "sc": 9, "l": 5, "f": 5247},
  {"h": "嗍", "p": "suō", "sp": "suo", "r": "口", "sc": 13, "l": 5, "f": 5248},
  {"h": "囵", "p": "lún", "sp": "lun", "r": "囗", "sc": 7, "l": 5, "f": 5249},
  {"h": "榧", "p": "fěi", "sp": "fei", "r": "木", "sc": 14, "l": 5, "f": 5250},
  {"h": "裰", "p": "duō", "sp": "duo", "r": "衣", "sc": 13, "l": 5, "f": 5251},
  {"h": "笾", "p": "biān", "sp": "bian", "r": "竹", "sc": 11, "l": 5, "f": 5252},
  {"h": "簟", "p": "diàn", "sp": "dian", "r": "竹", "sc": 18, "l": 5, "f": 5253},
  {"h": "跎", "p": "tuó", "sp": "tuo", "r": "足", "sc": 12, "l": 5, "f": 5254},
  {"h": "巽", "p": "xùn", "sp": "xun", "r": "己", "sc": 12, "l": 5, "f": 5255},
  {"h": "曷", "p": "hé", "sp": "he", "r": "曰", "sc": 9, "l": 5, "f": 5256},
  {"h": "逖", "p": "tì", "sp": "ti", "r": "辵", "sc": 10, "l": 5, "f": 5257},
  {"h": "骓", "p": "zhuī", "sp": "zhui", "r": "馬", "sc": 11, "l": 5, "f": 5258},
  {"h": "绔", "p": "kù", "sp": "ku", "r": "糸", "sc": 9, "l": 5, "f": 5259},
  {"h": "枋", "p": "fāng", "sp": "fang", "r": "木", "sc": 8, "l": 5, "f": 5260},
  {"h": "镒", "p": "yì", "sp": "yi", "r": "金", "sc": 15, "l": 5, "f": 5261},
  {"h": "魃", "p": "bá", "sp": "ba", "r": "鬼", "sc": 14, "l": 5, "f": 5262},
  {"h": "餮", "p": "tiè", "sp": "tie", "r": "食", "sc": 18, "l": 5, "f": 5263},
  {"h": "讵", "p": "jù", "sp": "ju", "r": "言", "sc": 6, "l": 5, "f": 5264},
  {"h": "乜", "p": "miē", "sp": "mie", "r": "乙", "sc": 2, "l": 5, "f": 5265},
  {"h": "鄢", "p": "yān", "sp": "yan", "r": "邑", "sc": 13, "l": 5, "f": 5266},
  {"h": "瑭", "p": "táng", "sp": "tang", "r": "玉", "sc": 14, "l": 5, "f": 5267},
  {"h": "踅", "p": "xué", "sp": "xue", "r": "足", "sc": 14, "l": 5, "f": 5268},
  {"h": "馓", "p": "sǎn", "sp": "san", "r": "食", "sc": 15, "l": 5, "f": 5269},
  {"h": "蟛", "p": "péng", "sp": "peng", "r": "虫", "sc": 18, "l": 5, "f": 5270},
  {"h": "鳟", "p": "zūn", "sp": "zun", "r": "魚", "sc": 20, "l": 5, "f": 5271},
  {"h": "荛", "p": "ráo", "sp": "rao", "r": "艸", "sc": 9, "l": 5, "f": 5272},
  {"h": "菬", "p": "qiáo", "sp": "qiao", "r": "艸", "sc": 11, "l": 5, "f": 5273},
  {"h": "忪", "p": "sōng", "sp": "song", "r": "心", "sc": 7, "l": 5, "f": 5274},
  {"h": "阍", "p": "hūn", "sp": "hun", "r": "門", "sc": 11, "l": 5, "f": 5275},
  {"h": "姹", "p": "chà", "sp": "cha", "r": "女", "sc": 9, "l": 5, "f": 5276},
  {"h": "纰", "p": "pī", "sp": "pi", "r": "糸", "sc": 7, "l": 5, "f": 5277},
  {"h": "桉", "p": "ān", "sp": "an", "r": "木", "sc": 10, "l": 5, "f": 5278},
  {"h": "氪", "p": "kè", "sp": "ke", "r": "气", "sc": 11, "l": 5, "f": 5279},
  {"h": "氘", "p": "dāo", "sp": "dao", "r": "气", "sc": 6, "l": 5, "f": 5280},
  {"h": "垅", "p": "lǒng", "sp": "long", "r": "土", "sc": 8, "l": 5, "f": 5281},
  {"h": "郃", "p": "hé", "sp": "he", "r": "邑", "sc": 8, "l": 5, "f": 5282},
  {"h": "汊", "p": "chà", "sp": "cha", "r": "水", "sc": 6, "l": 5, "f": 5283},
  {"h": "娉", "p": "pīng", "sp": "ping", "r": "女", "sc": 10, "l": 5, "f": 5284},
  {"h": "纡", "p": "yū", "sp": "yu", "r": "糸", "sc": 6, "l": 5, "f": 5285},
  {"h": "缟", "p": "gǎo", "sp": "gao", "r": "糸", "sc": 13, "l": 5, "f": 5286},
  {"h": "旮", "p": "gā", "sp": "ga", "r": "日", "sc": 6, "l": 5, "f": 5287},
  {"h": "镢", "p": "jué", "sp": "jue", "r": "金", "sc": 17, "l": 5, "f": 5288},
  {"h": "傈", "p": "lì", "sp": "li", "r": "人", "sc": 12, "l": 5, "f": 5289},
  {"h": "堋", "p": "péng", "sp": "peng", "r": "土", "sc": 11, "l": 5, "f": 5290},
  {"h": "蔺", "p": "lìn", "sp": "lin", "r": "艸", "sc": 14, "l": 5, "f": 5291},
  {"h": "庥", "p": "xiū", "sp": "xiu", "r": "广", "sc": 9, "l": 5, "f": 5292},
  {"h": "枥", "p": "lì", "sp": "li", "r": "木", "sc": 8, "l": 5, "f": 5293},
  {"h": "腭", "p": "è", "sp": "e", "r": "肉", "sc": 13, "l": 5, "f": 5294},
  {"h": "鹕", "p": "hú", "sp": "hu", "r": "鳥", "sc": 14, "l": 5, "f": 5295},
  {"h": "笮", "p": "zé", "sp": "ze", "r": "竹", "sc": 11, "l": 5, "f": 5296},
  {"h": "髂", "p": "qià", "sp": "qia", "r": "骨", "sc": 18, "l": 5, "f": 5297},
  {"h": "魍", "p": "wǎng", "sp": "wang", "r": "鬼", "sc": 17, "l": 5, "f": 5298},
  {"h": "缁", "p": "zī", "sp": "zi", "r": "糸", "sc": 11, "l": 5, "f": 5299},
  {"h": "槊", "p": "shuò", "sp": "shuo", "r": "木", "sc": 14, "l": 5, "f": 5300},
  {"h": "跞", "p": "lì", "sp": "li", "r": "足", "sc": 12, "l": 5, "f": 5301},
  {"h": "醚", "p": "mí", "sp": "mi", "r": "酉", "sc": 16, "l": 5, "f": 5302},
  {"h": "吒", "p": "zhā", "sp": "zha", "r": "口", "sc": 6, "l": 5, "f": 5303},
  {"h": "枳", "p": "zhǐ", "sp": "zhi", "r": "木", "sc": 9, "l": 5, "f": 5304},
  {"h": "搿", "p": "gé", "sp": "ge", "r": "手", "sc": 14, "l": 5, "f": 5305},
  {"h": "鹧", "p": "zhè", "sp": "zhe", "r": "鳥", "sc": 16, "l": 5, "f": 5306},
  {"h": "蜍", "p": "chú", "sp": "chu", "r": "虫", "sc": 13, "l": 5, "f": 5307},
  {"h": "舻", "p": "lú", "sp": "lu", "r": "舟", "sc": 11, "l": 5, "f": 5308},
  {"h": "鏊", "p": "ào", "sp": "ao", "r": "金", "sc": 18, "l": 5, "f": 5309},
  {"h": "禳", "p": "ráng", "sp": "rang", "r": "示", "sc": 21, "l": 5, "f": 5310},
  {"h": "蒺", "p": "jí", "sp": "ji", "r": "艸", "sc": 13, "l": 5, "f": 5311},
  {"h": "钹", "p": "bó", "sp": "bo", "r": "金", "sc": 10, "l": 5, "f": 5312},
  {"h": "蜢", "p": "měng", "sp": "meng", "r": "虫", "sc": 14, "l": 5, "f": 5313},
  {"h": "鬻", "p": "yù", "sp": "yu", "r": "鬲", "sc": 22, "l": 5, "f": 5314},
  {"h": "珩", "p": "háng", "sp": "hang", "r": "玉", "sc": 10, "l": 5, "f": 5315},
  {"h": "卮", "p": "zhī", "sp": "zhi", "r": "卩", "sc": 5, "l": 5, "f": 5316},
  {"h": "垭", "p": "yā", "sp": "ya", "r": "土", "sc": 9, "l": 5, "f": 5317},
  {"h": "苄", "p": "biàn", "sp": "bian", "r": "艸", "sc": 7, "l": 5, "f": 5318},
  {"h": "苕", "p": "sháo", "sp": "shao", "r": "艸", "sc": 8, "l": 5, "f": 5319},
  {"h": "菀", "p": "wǎn", "sp": "wan", "r": "艸", "sc": 11, "l": 5, "f": 5320},
  {"h": "骠", "p": "biāo", "sp": "biao", "r": "馬", "sc": 14, "l": 5, "f": 5321},
  {"h": "袷", "p": "jiá", "sp": "jia", "r": "衣", "sc": 11, "l": 5, "f": 5322},
  {"h": "跹", "p": "xiān", "sp": "xian", "r": "足", "sc": 13, "l": 5, "f": 5323},
  {"h": "瘘", "p": "lòu", "sp": "lou", "r": "疒", "sc": 14, "l": 5, "f": 5324},
  {"h": "騔", "p": "gé", "sp": "ge", "r": "馬", "sc": 19, "l": 5, "f": 5325},
  {"h": "論", "p": "lùn", "sp": "lun", "r": "言", "sc": 15, "l": 5, "f": 5326},
  {"h": "磬", "p": "qìng", "sp": "qing", "r": "石", "sc": 16, "l": 5, "f": 5327},
  {"h": "缶", "p": "fǒu", "sp": "fou", "r": "缶", "sc": 6, "l": 5, "f": 5328},
  {"h": "笸", "p": "pǒ", "sp": "po", "r": "竹", "sc": 11, "l": 5, "f": 5329},
  {"h": "鸷", "p": "zhì", "sp": "zhi", "r": "鳥", "sc": 11, "l": 5, "f": 5330},
  {"h": "頭", "p": "tóu", "sp": "tou", "r": "頁", "sc": 16, "l": 5, "f": 5331},
  {"h": "芰", "p": "jì", "sp": "ji", "r": "艸", "sc": 7, "l": 5, "f": 5332},
  {"h": "蕲", "p": "qí", "sp": "qi", "r": "艸", "sc": 15, "l": 5, "f": 5333},
  {"h": "阆", "p": "láng", "sp": "lang", "r": "門", "sc": 10, "l": 5, "f": 5334},
  {"h": "纨", "p": "wán", "sp": "wan", "r": "糸", "sc": 6, "l": 5, "f": 5335},
  {"h": "琮", "p": "cóng", "sp": "cong", "r": "玉", "sc": 12, "l": 5, "f": 5336},
  {"h": "牦", "p": "máo", "sp": "mao", "r": "牛", "sc": 8, "l": 5, "f": 5337},
  {"h": "砩", "p": "fú", "sp": "fu", "r": "石", "sc": 10, "l": 5, "f": 5338},
  {"h": "蠲", "p": "juān", "sp": "juan", "r": "虫", "sc": 23, "l": 5, "f": 5339},
  {"h": "锒", "p": "láng", "sp": "lang", "r": "金", "sc": 12, "l": 5, "f": 5340},
  {"h": "锕", "p": "ā", "sp": "a", "r": "金", "sc": 12, "l": 5, "f": 5341},
  {"h": "郓", "p": "yùn", "sp": "yun", "r": "邑", "sc": 8, "l": 5, "f": 5342},
  {"h": "妯", "p": "zhóu", "sp": "zhou", "r": "女", "sc": 8, "l": 5, "f": 5343},
  {"h": "驷", "p": "sì", "sp": "si", "r": "馬", "sc": 8, "l": 5, "f": 5344},
  {"h": "鹩", "p": "liáo", "sp": "liao", "r": "鳥", "sc": 17, "l": 5, "f": 5345},
  {"h": "舢", "p": "shān", "sp": "shan", "r": "舟", "sc": 9, "l": 5, "f": 5346},
  {"h": "趸", "p": "dǔn", "sp": "dun", "r": "足", "sc": 10, "l": 5, "f": 5347},
  {
    "h": "證",
    "p": "zhèng",
    "sp": "zheng",
    "r": "言",
    "sc": 19,
    "l": 5,
    "f": 5348
  },
  {"h": "養", "p": "yǎng", "sp": "yang", "r": "食", "sc": 14, "l": 5, "f": 5349},
  {"h": "芫", "p": "yán", "sp": "yan", "r": "艸", "sc": 7, "l": 5, "f": 5350},
  {"h": "嗉", "p": "sù", "sp": "su", "r": "口", "sc": 13, "l": 5, "f": 5351},
  {"h": "蠊", "p": "lián", "sp": "lian", "r": "虫", "sc": 19, "l": 5, "f": 5352},
  {"h": "笊", "p": "zhào", "sp": "zhao", "r": "竹", "sc": 10, "l": 5, "f": 5353},
  {"h": "莸", "p": "yóu", "sp": "you", "r": "艸", "sc": 10, "l": 5, "f": 5354},
  {"h": "饴", "p": "yí", "sp": "yi", "r": "食", "sc": 8, "l": 5, "f": 5355},
  {"h": "阃", "p": "kǔn", "sp": "kun", "r": "門", "sc": 10, "l": 5, "f": 5356},
  {"h": "浯", "p": "wú", "sp": "wu", "r": "水", "sc": 10, "l": 5, "f": 5357},
  {"h": "枇", "p": "pí", "sp": "pi", "r": "木", "sc": 8, "l": 5, "f": 5358},
  {"h": "焱", "p": "yàn", "sp": "yan", "r": "火", "sc": 12, "l": 5, "f": 5359},
  {"h": "铆", "p": "mǎo", "sp": "mao", "r": "金", "sc": 10, "l": 5, "f": 5360},
  {"h": "擤", "p": "xǐng", "sp": "xing", "r": "手", "sc": 17, "l": 5, "f": 5361},
  {"h": "柢", "p": "dǐ", "sp": "di", "r": "木", "sc": 9, "l": 5, "f": 5362},
  {"h": "醢", "p": "hǎi", "sp": "hai", "r": "酉", "sc": 17, "l": 5, "f": 5363},
  {"h": "呲", "p": "cī", "sp": "ci", "r": "口", "sc": 9, "l": 5, "f": 5364},
  {"h": "崾", "p": "yǎo", "sp": "yao", "r": "山", "sc": 12, "l": 5, "f": 5365},
  {"h": "溆", "p": "xù", "sp": "xu", "r": "水", "sc": 12, "l": 5, "f": 5366},
  {"h": "潴", "p": "zhū", "sp": "zhu", "r": "水", "sc": 14, "l": 5, "f": 5367},
  {"h": "牖", "p": "yǒu", "sp": "you", "r": "片", "sc": 15, "l": 5, "f": 5368},
  {"h": "硪", "p": "wò", "sp": "wo", "r": "石", "sc": 12, "l": 5, "f": 5369},
  {"h": "碓", "p": "duì", "sp": "dui", "r": "石", "sc": 13, "l": 5, "f": 5370},
  {"h": "鹆", "p": "yù", "sp": "yu", "r": "鳥", "sc": 12, "l": 5, "f": 5371},
  {"h": "鬣", "p": "liè", "sp": "lie", "r": "髟", "sc": 25, "l": 5, "f": 5372},
  {"h": "堀", "p": "kū", "sp": "ku", "r": "土", "sc": 11, "l": 5, "f": 5373},
  {"h": "帙", "p": "zhì", "sp": "zhi", "r": "巾", "sc": 8, "l": 5, "f": 5374},
  {"h": "雱", "p": "pāng", "sp": "pang", "r": "雨", "sc": 12, "l": 5, "f": 5375},
  {"h": "須", "p": "xū", "sp": "xu", "r": "頁", "sc": 12, "l": 5, "f": 5376},
  {"h": "進", "p": "jìn", "sp": "jin", "r": "辵", "sc": 11, "l": 5, "f": 5377},
  {"h": "诎", "p": "qū", "sp": "qu", "r": "言", "sc": 7, "l": 5, "f": 5378},
  {
    "h": "獐",
    "p": "zhāng",
    "sp": "zhang",
    "r": "犬",
    "sc": 14,
    "l": 5,
    "f": 5379
  },
  {"h": "桁", "p": "héng", "sp": "heng", "r": "木", "sc": 10, "l": 5, "f": 5380},
  {"h": "蛱", "p": "jiá", "sp": "jia", "r": "虫", "sc": 12, "l": 5, "f": 5381},
  {"h": "鳏", "p": "guān", "sp": "guan", "r": "魚", "sc": 18, "l": 5, "f": 5382},
  {"h": "郴", "p": "chēn", "sp": "chen", "r": "邑", "sc": 10, "l": 5, "f": 5383},
  {"h": "幂", "p": "mì", "sp": "mi", "r": "冖", "sc": 12, "l": 5, "f": 5384},
  {"h": "箝", "p": "qián", "sp": "qian", "r": "竹", "sc": 14, "l": 5, "f": 5385},
  {"h": "僳", "p": "sù", "sp": "su", "r": "人", "sc": 14, "l": 5, "f": 5386},
  {"h": "疝", "p": "shàn", "sp": "shan", "r": "疒", "sc": 8, "l": 5, "f": 5387},
  {"h": "茴", "p": "huí", "sp": "hui", "r": "艸", "sc": 9, "l": 5, "f": 5388},
  {"h": "揶", "p": "yé", "sp": "ye", "r": "手", "sc": 11, "l": 5, "f": 5389},
  {"h": "呦", "p": "yōu", "sp": "you", "r": "口", "sc": 8, "l": 5, "f": 5390},
  {"h": "嗌", "p": "ài", "sp": "ai", "r": "口", "sc": 13, "l": 5, "f": 5391},
  {"h": "囹", "p": "líng", "sp": "ling", "r": "囗", "sc": 8, "l": 5, "f": 5392},
  {"h": "螈", "p": "yuán", "sp": "yuan", "r": "虫", "sc": 16, "l": 5, "f": 5393},
  {"h": "脲", "p": "niào", "sp": "niao", "r": "肉", "sc": 11, "l": 5, "f": 5394},
  {"h": "镊", "p": "niè", "sp": "nie", "r": "金", "sc": 15, "l": 5, "f": 5395},
  {"h": "锑", "p": "tī", "sp": "ti", "r": "金", "sc": 12, "l": 5, "f": 5396},
  {"h": "胨", "p": "dòng", "sp": "dong", "r": "肉", "sc": 9, "l": 5, "f": 5397},
  {"h": "膈", "p": "gé", "sp": "ge", "r": "肉", "sc": 14, "l": 5, "f": 5398},
  {"h": "痼", "p": "gù", "sp": "gu", "r": "疒", "sc": 13, "l": 5, "f": 5399},
  {"h": "鳊", "p": "biān", "sp": "bian", "r": "魚", "sc": 17, "l": 5, "f": 5400},
  {"h": "赅", "p": "gāi", "sp": "gai", "r": "貝", "sc": 10, "l": 5, "f": 5401},
  {"h": "贽", "p": "zhì", "sp": "zhi", "r": "貝", "sc": 10, "l": 5, "f": 5402},
  {"h": "處", "p": "chù", "sp": "chu", "r": "虍", "sc": 11, "l": 5, "f": 5403},
  {"h": "苤", "p": "piě", "sp": "pie", "r": "艸", "sc": 8, "l": 5, "f": 5404},
  {"h": "峄", "p": "yì", "sp": "yi", "r": "山", "sc": 8, "l": 5, "f": 5405},
  {"h": "桡", "p": "ráo", "sp": "rao", "r": "木", "sc": 10, "l": 5, "f": 5406},
  {"h": "雎", "p": "jū", "sp": "ju", "r": "隹", "sc": 13, "l": 5, "f": 5407},
  {"h": "鲋", "p": "fù", "sp": "fu", "r": "魚", "sc": 13, "l": 5, "f": 5408},
  {"h": "鞫", "p": "jū", "sp": "ju", "r": "革", "sc": 18, "l": 5, "f": 5409},
  {"h": "鼬", "p": "yòu", "sp": "you", "r": "鼠", "sc": 18, "l": 5, "f": 5410},
  {"h": "獯", "p": "xūn", "sp": "xun", "r": "犬", "sc": 17, "l": 5, "f": 5411},
  {"h": "昀", "p": "yún", "sp": "yun", "r": "日", "sc": 8, "l": 5, "f": 5412},
  {"h": "痍", "p": "yí", "sp": "yi", "r": "疒", "sc": 11, "l": 5, "f": 5413},
  {"h": "蟊", "p": "máo", "sp": "mao", "r": "虫", "sc": 17, "l": 5, "f": 5414},
  {"h": "鞴", "p": "bèi", "sp": "bei", "r": "革", "sc": 19, "l": 5, "f": 5415},
  {"h": "疖", "p": "jiē", "sp": "jie", "r": "疒", "sc": 7, "l": 5, "f": 5416},
  {"h": "熘", "p": "liū", "sp": "liu", "r": "火", "sc": 14, "l": 5, "f": 5417},
  {"h": "乇", "p": "tuō", "sp": "tuo", "r": "丿", "sc": 3, "l": 5, "f": 5418},
  {"h": "羸", "p": "léi", "sp": "lei", "r": "羊", "sc": 19, "l": 5, "f": 5419},
  {"h": "嵴", "p": "jǐ", "sp": "ji", "r": "山", "sc": 13, "l": 5, "f": 5420},
  {"h": "栀", "p": "zhī", "sp": "zhi", "r": "木", "sc": 9, "l": 5, "f": 5421},
  {"h": "槲", "p": "hú", "sp": "hu", "r": "木", "sc": 15, "l": 5, "f": 5422},
  {"h": "炝", "p": "qiàng", "sp": "qiang", "r": "火", "sc": 8, "l": 5, "f": 5423},
  {"h": "炷", "p": "zhù", "sp": "zhu", "r": "火", "sc": 9, "l": 5, "f": 5424},
  {"h": "硐", "p": "dòng", "sp": "dong", "r": "石", "sc": 11, "l": 5, "f": 5425},
  {"h": "锸", "p": "chā", "sp": "cha", "r": "金", "sc": 14, "l": 5, "f": 5426},
  {"h": "鹂", "p": "lí", "sp": "li", "r": "鳥", "sc": 12, "l": 5, "f": 5427},
  {"h": "裾", "p": "jū", "sp": "ju", "r": "衣", "sc": 13, "l": 5, "f": 5428},
  {"h": "侪", "p": "chái", "sp": "chai", "r": "人", "sc": 8, "l": 5, "f": 5429},
  {"h": "診", "p": "zhěn", "sp": "zhen", "r": "言", "sc": 12, "l": 5, "f": 5430},
  {"h": "調", "p": "diào", "sp": "diao", "r": "言", "sc": 15, "l": 5, "f": 5431},
  {"h": "珐", "p": "fà", "sp": "fa", "r": "玉", "sc": 9, "l": 5, "f": 5432},
  {"h": "縯", "p": "yǎn", "sp": "yan", "r": "糸", "sc": 17, "l": 5, "f": 5433},
  {"h": "哔", "p": "bì", "sp": "bi", "r": "口", "sc": 9, "l": 5, "f": 5434},
  {"h": "屙", "p": "ē", "sp": "e", "r": "尸", "sc": 10, "l": 5, "f": 5435},
  {"h": "旆", "p": "pèi", "sp": "pei", "r": "方", "sc": 10, "l": 5, "f": 5436},
  {"h": "佰", "p": "bǎi", "sp": "bai", "r": "人", "sc": 8, "l": 5, "f": 5437},
  {"h": "僦", "p": "jiù", "sp": "jiu", "r": "人", "sc": 14, "l": 5, "f": 5438},
  {"h": "牯", "p": "gǔ", "sp": "gu", "r": "牛", "sc": 9, "l": 5, "f": 5439},
  {"h": "钪", "p": "kàng", "sp": "kang", "r": "金", "sc": 9, "l": 5, "f": 5440},
  {"h": "掾", "p": "yuàn", "sp": "yuan", "r": "手", "sc": 12, "l": 5, "f": 5441},
  {"h": "針", "p": "zhēn", "sp": "zhen", "r": "金", "sc": 10, "l": 5, "f": 5442},
  {"h": "仟", "p": "qiān", "sp": "qian", "r": "人", "sc": 5, "l": 5, "f": 5443},
  {"h": "圮", "p": "pǐ", "sp": "pi", "r": "土", "sc": 6, "l": 5, "f": 5444},
  {"h": "芟", "p": "shān", "sp": "shan", "r": "艸", "sc": 7, "l": 5, "f": 5445},
  {"h": "崃", "p": "lái", "sp": "lai", "r": "山", "sc": 10, "l": 5, "f": 5446},
  {"h": "廪", "p": "lǐn", "sp": "lin", "r": "广", "sc": 16, "l": 5, "f": 5447},
  {"h": "擘", "p": "bāi", "sp": "bai", "r": "手", "sc": 17, "l": 5, "f": 5448},
  {"h": "笱", "p": "gǒu", "sp": "gou", "r": "竹", "sc": 11, "l": 5, "f": 5449},
  {"h": "跗", "p": "fū", "sp": "fu", "r": "足", "sc": 12, "l": 5, "f": 5450},
  {"h": "鲅", "p": "bà", "sp": "ba", "r": "魚", "sc": 13, "l": 5, "f": 5451},
  {"h": "硷", "p": "jiǎn", "sp": "jian", "r": "石", "sc": 12, "l": 5, "f": 5452},
  {"h": "苎", "p": "zhù", "sp": "zhu", "r": "艸", "sc": 7, "l": 5, "f": 5453},
  {"h": "匏", "p": "páo", "sp": "pao", "r": "勹", "sc": 11, "l": 5, "f": 5454},
  {"h": "嗾", "p": "sǒu", "sp": "sou", "r": "口", "sc": 14, "l": 5, "f": 5455},
  {"h": "圄", "p": "yǔ", "sp": "yu", "r": "囗", "sc": 10, "l": 5, "f": 5456},
  {"h": "彀", "p": "gòu", "sp": "gou", "r": "弓", "sc": 13, "l": 5, "f": 5457},
  {"h": "粳", "p": "jīng", "sp": "jing", "r": "米", "sc": 13, "l": 5, "f": 5458},
  {"h": "卣", "p": "yǒu", "sp": "you", "r": "卜", "sc": 7, "l": 5, "f": 5459},
  {"h": "勐", "p": "měng", "sp": "meng", "r": "力", "sc": 10, "l": 5, "f": 5460},
  {"h": "掴", "p": "guāi", "sp": "guai", "r": "手", "sc": 11, "l": 5, "f": 5461},
  {"h": "涑", "p": "sù", "sp": "su", "r": "水", "sc": 10, "l": 5, "f": 5462},
  {"h": "浞", "p": "zhuó", "sp": "zhuo", "r": "水", "sc": 10, "l": 5, "f": 5463},
  {"h": "玳", "p": "dài", "sp": "dai", "r": "玉", "sc": 9, "l": 5, "f": 5464},
  {"h": "愍", "p": "mǐn", "sp": "min", "r": "心", "sc": 13, "l": 5, "f": 5465},
  {"h": "畛", "p": "zhěn", "sp": "zhen", "r": "田", "sc": 10, "l": 5, "f": 5466},
  {"h": "赧", "p": "nǎn", "sp": "nan", "r": "赤", "sc": 11, "l": 5, "f": 5467},
  {"h": "貉", "p": "háo", "sp": "hao", "r": "豸", "sc": 13, "l": 5, "f": 5468},
  {"h": "擀", "p": "gǎn", "sp": "gan", "r": "手", "sc": 16, "l": 5, "f": 5469},
  {"h": "湫", "p": "jiǎo", "sp": "jiao", "r": "水", "sc": 12, "l": 5, "f": 5470},
  {"h": "逦", "p": "lǐ", "sp": "li", "r": "辵", "sc": 10, "l": 5, "f": 5471},
  {"h": "椴", "p": "duàn", "sp": "duan", "r": "木", "sc": 13, "l": 5, "f": 5472},
  {"h": "铄", "p": "shuò", "sp": "shuo", "r": "金", "sc": 10, "l": 5, "f": 5473},
  {"h": "箧", "p": "qiè", "sp": "qie", "r": "竹", "sc": 14, "l": 5, "f": 5474},
  {"h": "刖", "p": "yuè", "sp": "yue", "r": "刀", "sc": 6, "l": 5, "f": 5475},
  {"h": "鲮", "p": "líng", "sp": "ling", "r": "魚", "sc": 16, "l": 5, "f": 5476},
  {"h": "訇", "p": "hōng", "sp": "hong", "r": "言", "sc": 9, "l": 5, "f": 5477},
  {"h": "茱", "p": "zhū", "sp": "zhu", "r": "艸", "sc": 9, "l": 5, "f": 5478},
  {"h": "啖", "p": "dàn", "sp": "dan", "r": "口", "sc": 11, "l": 5, "f": 5479},
  {"h": "悭", "p": "qiān", "sp": "qian", "r": "心", "sc": 10, "l": 5, "f": 5480},
  {"h": "愀", "p": "qiǎo", "sp": "qiao", "r": "心", "sc": 12, "l": 5, "f": 5481},
  {"h": "朐", "p": "qú", "sp": "qu", "r": "月", "sc": 9, "l": 5, "f": 5482},
  {"h": "畈", "p": "fàn", "sp": "fan", "r": "田", "sc": 9, "l": 5, "f": 5483},
  {"h": "鹨", "p": "liù", "sp": "liu", "r": "鳥", "sc": 16, "l": 5, "f": 5484},
  {"h": "蛘", "p": "yáng", "sp": "yang", "r": "虫", "sc": 12, "l": 5, "f": 5485},
  {"h": "佶", "p": "jí", "sp": "ji", "r": "人", "sc": 8, "l": 5, "f": 5486},
  {
    "h": "缃",
    "p": "xiāng",
    "sp": "xiang",
    "r": "糸",
    "sc": 12,
    "l": 5,
    "f": 5487
  },
  {
    "h": "晟",
    "p": "chéng",
    "sp": "cheng",
    "r": "日",
    "sc": 10,
    "l": 5,
    "f": 5488
  },
  {"h": "鲱", "p": "fēi", "sp": "fei", "r": "魚", "sc": 16, "l": 5, "f": 5489},
  {"h": "凼", "p": "dàng", "sp": "dang", "r": "水", "sc": 6, "l": 5, "f": 5490},
  {"h": "苴", "p": "jū", "sp": "ju", "r": "艸", "sc": 8, "l": 5, "f": 5491},
  {
    "h": "颛",
    "p": "zhuān",
    "sp": "zhuan",
    "r": "頁",
    "sc": 15,
    "l": 5,
    "f": 5492
  },
  {"h": "厍", "p": "shè", "sp": "she", "r": "厂", "sc": 6, "l": 5, "f": 5493},
  {"h": "匚", "p": "fāng", "sp": "fang", "r": "匚", "sc": 2, "l": 5, "f": 5494},
  {"h": "徉", "p": "yáng", "sp": "yang", "r": "彳", "sc": 9, "l": 5, "f": 5495},
  {"h": "洙", "p": "zhū", "sp": "zhu", "r": "水", "sc": 9, "l": 5, "f": 5496},
  {"h": "氡", "p": "dōng", "sp": "dong", "r": "气", "sc": 9, "l": 5, "f": 5497},
  {"h": "胗", "p": "zhēn", "sp": "zhen", "r": "肉", "sc": 9, "l": 5, "f": 5498},
  {"h": "癯", "p": "qú", "sp": "qu", "r": "疒", "sc": 23, "l": 5, "f": 5499},
  {"h": "鞒", "p": "qiáo", "sp": "qiao", "r": "革", "sc": 15, "l": 5, "f": 5500},
  {"h": "锆", "p": "gào", "sp": "gao", "r": "金", "sc": 12, "l": 5, "f": 5501},
  {"h": "佤", "p": "wǎ", "sp": "wa", "r": "人", "sc": 6, "l": 5, "f": 5502},
  {"h": "錢", "p": "qián", "sp": "qian", "r": "金", "sc": 16, "l": 5, "f": 5503},
  {"h": "飲", "p": "yǐn", "sp": "yin", "r": "食", "sc": 12, "l": 5, "f": 5504},
  {"h": "細", "p": "xì", "sp": "xi", "r": "糸", "sc": 11, "l": 5, "f": 5505},
  {"h": "勰", "p": "xié", "sp": "xie", "r": "力", "sc": 15, "l": 5, "f": 5506},
  {"h": "钺", "p": "yuè", "sp": "yue", "r": "金", "sc": 10, "l": 5, "f": 5507},
  {"h": "繇", "p": "yáo", "sp": "yao", "r": "糸", "sc": 17, "l": 5, "f": 5508},
  {"h": "螭", "p": "chī", "sp": "chi", "r": "虫", "sc": 16, "l": 5, "f": 5509},
  {"h": "嵬", "p": "wéi", "sp": "wei", "r": "山", "sc": 12, "l": 5, "f": 5510},
  {"h": "轸", "p": "zhěn", "sp": "zhen", "r": "車", "sc": 9, "l": 5, "f": 5511},
  {"h": "肟", "p": "wò", "sp": "wo", "r": "肉", "sc": 7, "l": 5, "f": 5512},
  {"h": "肫", "p": "zhūn", "sp": "zhun", "r": "肉", "sc": 8, "l": 5, "f": 5513},
  {"h": "邨", "p": "cūn", "sp": "cun", "r": "邑", "sc": 6, "l": 5, "f": 5514},
  {"h": "瘿", "p": "yǐng", "sp": "ying", "r": "疒", "sc": 16, "l": 5, "f": 5515},
  {"h": "仞", "p": "rèn", "sp": "ren", "r": "人", "sc": 5, "l": 5, "f": 5516},
  {"h": "奁", "p": "lián", "sp": "lian", "r": "大", "sc": 7, "l": 5, "f": 5517},
  {"h": "宄", "p": "guǐ", "sp": "gui", "r": "宀", "sc": 5, "l": 5, "f": 5518},
  {"h": "轳", "p": "lú", "sp": "lu", "r": "車", "sc": 9, "l": 5, "f": 5519},
  {"h": "熳", "p": "màn", "sp": "man", "r": "火", "sc": 15, "l": 5, "f": 5520},
  {"h": "睇", "p": "dì", "sp": "di", "r": "目", "sc": 12, "l": 5, "f": 5521},
  {"h": "钼", "p": "mù", "sp": "mu", "r": "金", "sc": 10, "l": 5, "f": 5522},
  {"h": "蝼", "p": "lóu", "sp": "lou", "r": "虫", "sc": 15, "l": 5, "f": 5523},
  {"h": "跆", "p": "tái", "sp": "tai", "r": "足", "sc": 12, "l": 5, "f": 5524},
  {"h": "樗", "p": "chū", "sp": "chu", "r": "木", "sc": 15, "l": 5, "f": 5525},
  {"h": "鲰", "p": "zōu", "sp": "zou", "r": "魚", "sc": 16, "l": 5, "f": 5526},
  {"h": "節", "p": "jié", "sp": "jie", "r": "竹", "sc": 13, "l": 5, "f": 5527},
  {"h": "诶", "p": "éi", "sp": "ei", "r": "言", "sc": 9, "l": 5, "f": 5528},
  {"h": "薜", "p": "bì", "sp": "bi", "r": "艸", "sc": 16, "l": 5, "f": 5529},
  {"h": "铧", "p": "huá", "sp": "hua", "r": "金", "sc": 11, "l": 5, "f": 5530},
  {"h": "裥", "p": "jiǎn", "sp": "jian", "r": "衣", "sc": 12, "l": 5, "f": 5531},
  {"h": "榇", "p": "chèn", "sp": "chen", "r": "木", "sc": 13, "l": 5, "f": 5532},
  {"h": "馃", "p": "guǒ", "sp": "guo", "r": "食", "sc": 11, "l": 5, "f": 5533},
  {"h": "術", "p": "shù", "sp": "shu", "r": "行", "sc": 11, "l": 5, "f": 5534},
  {"h": "蹚", "p": "tāng", "sp": "tang", "r": "足", "sc": 18, "l": 5, "f": 5535},
  {"h": "怄", "p": "òu", "sp": "ou", "r": "心", "sc": 7, "l": 5, "f": 5536},
  {"h": "寤", "p": "wù", "sp": "wu", "r": "宀", "sc": 14, "l": 5, "f": 5537},
  {"h": "缗", "p": "mín", "sp": "min", "r": "糸", "sc": 12, "l": 5, "f": 5538},
  {"h": "硗", "p": "qiāo", "sp": "qiao", "r": "石", "sc": 11, "l": 5, "f": 5539},
  {"h": "碡", "p": "dú", "sp": "du", "r": "石", "sc": 14, "l": 5, "f": 5540},
  {"h": "矬", "p": "cuó", "sp": "cuo", "r": "矢", "sc": 12, "l": 5, "f": 5541},
  {"h": "鸱", "p": "chī", "sp": "chi", "r": "鳥", "sc": 10, "l": 5, "f": 5542},
  {"h": "虺", "p": "huī", "sp": "hui", "r": "虫", "sc": 9, "l": 5, "f": 5543},
  {"h": "糅", "p": "róu", "sp": "rou", "r": "米", "sc": 15, "l": 5, "f": 5544},
  {"h": "雠", "p": "chóu", "sp": "chou", "r": "隹", "sc": 18, "l": 5, "f": 5545},
  {"h": "帑", "p": "tǎng", "sp": "tang", "r": "巾", "sc": 8, "l": 5, "f": 5546},
  {"h": "镧", "p": "lán", "sp": "lan", "r": "金", "sc": 17, "l": 5, "f": 5547},
  {"h": "埙", "p": "xūn", "sp": "xun", "r": "土", "sc": 10, "l": 5, "f": 5548},
  {"h": "啁", "p": "zhāo", "sp": "zhao", "r": "口", "sc": 11, "l": 5, "f": 5549},
  {"h": "悒", "p": "yì", "sp": "yi", "r": "心", "sc": 10, "l": 5, "f": 5550},
  {"h": "犍", "p": "jiān", "sp": "jian", "r": "牛", "sc": 12, "l": 5, "f": 5551},
  {"h": "硌", "p": "gè", "sp": "ge", "r": "石", "sc": 11, "l": 5, "f": 5552},
  {"h": "锩", "p": "juǎn", "sp": "juan", "r": "金", "sc": 13, "l": 5, "f": 5553},
  {"h": "虿", "p": "chài", "sp": "chai", "r": "虫", "sc": 9, "l": 5, "f": 5554},
  {"h": "蛑", "p": "móu", "sp": "mou", "r": "虫", "sc": 12, "l": 5, "f": 5555},
  {"h": "艉", "p": "wěi", "sp": "wei", "r": "舟", "sc": 13, "l": 5, "f": 5556},
  {"h": "钅", "p": "jīn", "sp": "jin", "r": "金", "sc": 5, "l": 5, "f": 5557},
  {"h": "咴", "p": "huī", "sp": "hui", "r": "口", "sc": 9, "l": 5, "f": 5558},
  {"h": "筮", "p": "shì", "sp": "shi", "r": "竹", "sc": 13, "l": 5, "f": 5559},
  {"h": "艏", "p": "shǒu", "sp": "shou", "r": "舟", "sc": 15, "l": 5, "f": 5560},
  {"h": "糁", "p": "sǎn", "sp": "san", "r": "米", "sc": 14, "l": 5, "f": 5561},
  {"h": "鼍", "p": "tuó", "sp": "tuo", "r": "黽", "sc": 20, "l": 5, "f": 5562},
  {"h": "肄", "p": "yì", "sp": "yi", "r": "聿", "sc": 13, "l": 5, "f": 5563},
  {"h": "籴", "p": "dí", "sp": "di", "r": "米", "sc": 8, "l": 5, "f": 5564},
  {"h": "骜", "p": "ào", "sp": "ao", "r": "馬", "sc": 13, "l": 5, "f": 5565},
  {"h": "砻", "p": "lóng", "sp": "long", "r": "龍", "sc": 10, "l": 5, "f": 5566},
  {"h": "蜮", "p": "yù", "sp": "yu", "r": "虫", "sc": 14, "l": 5, "f": 5567},
  {"h": "龀", "p": "chèn", "sp": "chen", "r": "齒", "sc": 10, "l": 5, "f": 5568},
  {"h": "黢", "p": "qū", "sp": "qu", "r": "黑", "sc": 19, "l": 5, "f": 5569},
  {"h": "劢", "p": "mài", "sp": "mai", "r": "力", "sc": 5, "l": 5, "f": 5570},
  {
    "h": "腫",
    "p": "zhǒng",
    "sp": "zhong",
    "r": "肉",
    "sc": 13,
    "l": 5,
    "f": 5571
  },
  {"h": "耪", "p": "pǎng", "sp": "pang", "r": "耒", "sc": 16, "l": 5, "f": 5572},
  {
    "h": "鬯",
    "p": "chàng",
    "sp": "chang",
    "r": "鬯",
    "sc": 10,
    "l": 5,
    "f": 5573
  },
  {"h": "畚", "p": "běn", "sp": "ben", "r": "田", "sc": 10, "l": 5, "f": 5574},
  {"h": "觳", "p": "hú", "sp": "hu", "r": "角", "sc": 17, "l": 5, "f": 5575},
  {"h": "稞", "p": "kē", "sp": "ke", "r": "禾", "sc": 13, "l": 5, "f": 5576},
  {"h": "鹁", "p": "bó", "sp": "bo", "r": "鳥", "sc": 12, "l": 5, "f": 5577},
  {"h": "鲲", "p": "kūn", "sp": "kun", "r": "魚", "sc": 16, "l": 5, "f": 5578},
  {
    "h": "稱",
    "p": "chēng",
    "sp": "cheng",
    "r": "禾",
    "sc": 14,
    "l": 5,
    "f": 5579
  },
  {"h": "捌", "p": "bā", "sp": "ba", "r": "手", "sc": 10, "l": 5, "f": 5580},
  {"h": "菔", "p": "fú", "sp": "fu", "r": "艸", "sc": 11, "l": 5, "f": 5581},
  {"h": "獬", "p": "xiè", "sp": "xie", "r": "犬", "sc": 16, "l": 5, "f": 5582},
  {"h": "柘", "p": "zhè", "sp": "zhe", "r": "木", "sc": 9, "l": 5, "f": 5583},
  {"h": "娆", "p": "ráo", "sp": "rao", "r": "女", "sc": 9, "l": 5, "f": 5584},
  {"h": "篪", "p": "chí", "sp": "chi", "r": "竹", "sc": 16, "l": 5, "f": 5585},
  {"h": "鲀", "p": "tún", "sp": "tun", "r": "魚", "sc": 12, "l": 5, "f": 5586},
  {"h": "谰", "p": "lán", "sp": "lan", "r": "言", "sc": 14, "l": 5, "f": 5587},
  {"h": "孬", "p": "nāo", "sp": "nao", "r": "子", "sc": 10, "l": 5, "f": 5588},
  {"h": "伥", "p": "chāng", "sp": "chang", "r": "人", "sc": 6, "l": 5, "f": 5589},
  {"h": "谇", "p": "suì", "sp": "sui", "r": "言", "sc": 10, "l": 5, "f": 5590},
  {"h": "鄄", "p": "juàn", "sp": "juan", "r": "邑", "sc": 11, "l": 5, "f": 5591},
  {"h": "狎", "p": "xiá", "sp": "xia", "r": "犬", "sc": 8, "l": 5, "f": 5592},
  {"h": "闫", "p": "yàn", "sp": "yan", "r": "門", "sc": 6, "l": 5, "f": 5593},
  {"h": "滟", "p": "yàn", "sp": "yan", "r": "水", "sc": 13, "l": 5, "f": 5594},
  {"h": "齑", "p": "jī", "sp": "ji", "r": "齊", "sc": 15, "l": 5, "f": 5595},
  {"h": "遒", "p": "qiú", "sp": "qiu", "r": "辵", "sc": 12, "l": 5, "f": 5596},
  {"h": "磔", "p": "zhé", "sp": "zhe", "r": "石", "sc": 15, "l": 5, "f": 5597},
  {"h": "聃", "p": "dān", "sp": "dan", "r": "耳", "sc": 11, "l": 5, "f": 5598},
  {"h": "綦", "p": "qí", "sp": "qi", "r": "糸", "sc": 14, "l": 5, "f": 5599},
  {"h": "鲡", "p": "lí", "sp": "li", "r": "魚", "sc": 15, "l": 5, "f": 5600},
  {"h": "蔻", "p": "kòu", "sp": "kou", "r": "艸", "sc": 14, "l": 5, "f": 5601},
  {"h": "泠", "p": "líng", "sp": "ling", "r": "水", "sc": 8, "l": 5, "f": 5602},
  {"h": "砗", "p": "chē", "sp": "che", "r": "石", "sc": 9, "l": 5, "f": 5603},
  {"h": "钕", "p": "nǚ", "sp": "nu", "r": "金", "sc": 8, "l": 5, "f": 5604},
  {"h": "镫", "p": "dèng", "sp": "deng", "r": "金", "sc": 17, "l": 5, "f": 5605},
  {"h": "菹", "p": "jū", "sp": "ju", "r": "艸", "sc": 11, "l": 5, "f": 5606},
  {"h": "胂", "p": "shèn", "sp": "shen", "r": "肉", "sc": 9, "l": 5, "f": 5607},
  {"h": "煅", "p": "duàn", "sp": "duan", "r": "火", "sc": 13, "l": 5, "f": 5608},
  {"h": "煸", "p": "biān", "sp": "bian", "r": "火", "sc": 13, "l": 5, "f": 5609},
  {"h": "螯", "p": "áo", "sp": "ao", "r": "虫", "sc": 16, "l": 5, "f": 5610},
  {"h": "躅", "p": "zhú", "sp": "zhu", "r": "足", "sc": 20, "l": 5, "f": 5611},
  {"h": "鲠", "p": "gěng", "sp": "geng", "r": "魚", "sc": 15, "l": 5, "f": 5612},
  {"h": "佥", "p": "qiān", "sp": "qian", "r": "人", "sc": 7, "l": 5, "f": 5613},
  {"h": "罘", "p": "fú", "sp": "fu", "r": "网", "sc": 9, "l": 5, "f": 5614},
  {"h": "嶝", "p": "dèng", "sp": "deng", "r": "山", "sc": 15, "l": 5, "f": 5615},
  {"h": "適", "p": "shì", "sp": "shi", "r": "辵", "sc": 14, "l": 5, "f": 5616},
  {"h": "坨", "p": "tuó", "sp": "tuo", "r": "土", "sc": 8, "l": 5, "f": 5617},
  {"h": "菽", "p": "shū", "sp": "shu", "r": "艸", "sc": 11, "l": 5, "f": 5618},
  {"h": "哞", "p": "mōu", "sp": "mou", "r": "口", "sc": 9, "l": 5, "f": 5619},
  {
    "h": "徜",
    "p": "cháng",
    "sp": "chang",
    "r": "彳",
    "sc": 11,
    "l": 5,
    "f": 5620
  },
  {"h": "慊", "p": "qiàn", "sp": "qian", "r": "心", "sc": 13, "l": 5, "f": 5621},
  {"h": "洳", "p": "rù", "sp": "ru", "r": "水", "sc": 9, "l": 5, "f": 5622},
  {"h": "渑", "p": "miǎn", "sp": "mian", "r": "水", "sc": 11, "l": 5, "f": 5623},
  {"h": "灞", "p": "bà", "sp": "ba", "r": "水", "sc": 24, "l": 5, "f": 5624},
  {"h": "盍", "p": "hé", "sp": "he", "r": "皿", "sc": 10, "l": 5, "f": 5625},
  {"h": "钋", "p": "pō", "sp": "po", "r": "金", "sc": 7, "l": 5, "f": 5626},
  {"h": "鸫", "p": "dōng", "sp": "dong", "r": "鳥", "sc": 10, "l": 5, "f": 5627},
  {"h": "踯", "p": "zhí", "sp": "zhi", "r": "足", "sc": 15, "l": 5, "f": 5628},
  {"h": "縻", "p": "mí", "sp": "mi", "r": "糸", "sc": 17, "l": 5, "f": 5629},
  {"h": "萘", "p": "nài", "sp": "nai", "r": "艸", "sc": 11, "l": 5, "f": 5630},
  {"h": "褫", "p": "chǐ", "sp": "chi", "r": "衣", "sc": 15, "l": 5, "f": 5631},
  {"h": "羰", "p": "tāng", "sp": "tang", "r": "羊", "sc": 15, "l": 5, "f": 5632},
  {"h": "腦", "p": "nǎo", "sp": "nao", "r": "肉", "sc": 13, "l": 5, "f": 5633},
  {"h": "俅", "p": "qiú", "sp": "qiu", "r": "人", "sc": 9, "l": 5, "f": 5634},
  {"h": "芤", "p": "kōu", "sp": "kou", "r": "艸", "sc": 7, "l": 5, "f": 5635},
  {"h": "隳", "p": "huī", "sp": "hui", "r": "阜", "sc": 17, "l": 5, "f": 5636},
  {"h": "洮", "p": "táo", "sp": "tao", "r": "水", "sc": 9, "l": 5, "f": 5637},
  {"h": "胼", "p": "pián", "sp": "pian", "r": "肉", "sc": 10, "l": 5, "f": 5638},
  {"h": "罴", "p": "pí", "sp": "pi", "r": "网", "sc": 14, "l": 5, "f": 5639},
  {"h": "镛", "p": "yōng", "sp": "yong", "r": "金", "sc": 16, "l": 5, "f": 5640},
  {"h": "怛", "p": "dá", "sp": "da", "r": "心", "sc": 8, "l": 5, "f": 5641},
  {"h": "芊", "p": "qiān", "sp": "qian", "r": "艸", "sc": 6, "l": 5, "f": 5642},
  {"h": "啉", "p": "lín", "sp": "lin", "r": "口", "sc": 11, "l": 5, "f": 5643},
  {"h": "噌", "p": "cēng", "sp": "ceng", "r": "口", "sc": 15, "l": 5, "f": 5644},
  {
    "h": "嫱",
    "p": "qiáng",
    "sp": "qiang",
    "r": "女",
    "sc": 14,
    "l": 5,
    "f": 5645
  },
  {"h": "绲", "p": "gǔn", "sp": "gun", "r": "糸", "sc": 11, "l": 5, "f": 5646},
  {"h": "膻", "p": "shān", "sp": "shan", "r": "肉", "sc": 17, "l": 5, "f": 5647},
  {"h": "焐", "p": "wù", "sp": "wu", "r": "火", "sc": 11, "l": 5, "f": 5648},
  {"h": "裡", "p": "lǐ", "sp": "li", "r": "衣", "sc": 12, "l": 5, "f": 5649},
  {"h": "葎", "p": "lǜ", "sp": "lu", "r": "艸", "sc": 12, "l": 5, "f": 5650},
  {"h": "亓", "p": "qí", "sp": "qi", "r": "二", "sc": 4, "l": 5, "f": 5651},
  {"h": "倮", "p": "luǒ", "sp": "luo", "r": "人", "sc": 10, "l": 5, "f": 5652},
  {"h": "莼", "p": "chún", "sp": "chun", "r": "艸", "sc": 10, "l": 5, "f": 5653},
  {"h": "蘅", "p": "héng", "sp": "heng", "r": "艸", "sc": 19, "l": 5, "f": 5654},
  {"h": "嘞", "p": "lei", "sp": "lei", "r": "口", "sc": 14, "l": 5, "f": 5655},
  {"h": "缒", "p": "zhuì", "sp": "zhui", "r": "糸", "sc": 12, "l": 5, "f": 5656},
  {"h": "镆", "p": "mò", "sp": "mo", "r": "金", "sc": 15, "l": 5, "f": 5657},
  {"h": "網", "p": "wǎng", "sp": "wang", "r": "糸", "sc": 14, "l": 5, "f": 5658},
  {"h": "伧", "p": "cāng", "sp": "cang", "r": "人", "sc": 6, "l": 5, "f": 5659},
  {"h": "荏", "p": "rěn", "sp": "ren", "r": "艸", "sc": 9, "l": 5, "f": 5660},
  {"h": "唳", "p": "lì", "sp": "li", "r": "口", "sc": 11, "l": 5, "f": 5661},
  {"h": "檩", "p": "lǐn", "sp": "lin", "r": "木", "sc": 17, "l": 5, "f": 5662},
  {"h": "鸶", "p": "sī", "sp": "si", "r": "鳥", "sc": 10, "l": 5, "f": 5663},
  {"h": "蚬", "p": "xiǎn", "sp": "xian", "r": "虫", "sc": 10, "l": 5, "f": 5664},
  {"h": "骱", "p": "jiè", "sp": "jie", "r": "骨", "sc": 13, "l": 5, "f": 5665},
  {"h": "蘖", "p": "niè", "sp": "nie", "r": "艸", "sc": 20, "l": 5, "f": 5666},
  {"h": "澍", "p": "shù", "sp": "shu", "r": "水", "sc": 15, "l": 5, "f": 5667},
  {"h": "韫", "p": "yùn", "sp": "yun", "r": "韋", "sc": 13, "l": 5, "f": 5668},
  {
    "h": "颎",
    "p": "jiǒng",
    "sp": "jiong",
    "r": "火",
    "sc": 12,
    "l": 5,
    "f": 5669
  },
  {"h": "嘏", "p": "gǔ", "sp": "gu", "r": "口", "sc": 14, "l": 5, "f": 5670},
  {"h": "垡", "p": "fá", "sp": "fa", "r": "土", "sc": 9, "l": 5, "f": 5671},
  {"h": "腚", "p": "dìng", "sp": "ding", "r": "肉", "sc": 12, "l": 5, "f": 5672},
  {"h": "焯", "p": "chāo", "sp": "chao", "r": "火", "sc": 12, "l": 5, "f": 5673},
  {"h": "繻", "p": "xū", "sp": "xu", "r": "糸", "sc": 20, "l": 5, "f": 5674},
  {"h": "怙", "p": "hù", "sp": "hu", "r": "心", "sc": 8, "l": 5, "f": 5675},
  {"h": "羧", "p": "suō", "sp": "suo", "r": "羊", "sc": 13, "l": 5, "f": 5676},
  {"h": "鼙", "p": "pí", "sp": "pi", "r": "鼓", "sc": 21, "l": 5, "f": 5677},
  {"h": "倥", "p": "kōng", "sp": "kong", "r": "人", "sc": 10, "l": 5, "f": 5678},
  {"h": "亳", "p": "bó", "sp": "bo", "r": "亠", "sc": 10, "l": 5, "f": 5679},
  {"h": "艽", "p": "jiāo", "sp": "jiao", "r": "艸", "sc": 5, "l": 5, "f": 5680},
  {"h": "荠", "p": "jì", "sp": "ji", "r": "艸", "sc": 9, "l": 5, "f": 5681},
  {"h": "昴", "p": "mǎo", "sp": "mao", "r": "日", "sc": 9, "l": 5, "f": 5682},
  {"h": "舨", "p": "bǎn", "sp": "ban", "r": "舟", "sc": 10, "l": 5, "f": 5683},
  {"h": "魈", "p": "xiāo", "sp": "xiao", "r": "鬼", "sc": 16, "l": 5, "f": 5684},
  {"h": "醣", "p": "táng", "sp": "tang", "r": "酉", "sc": 17, "l": 5, "f": 5685},
  {"h": "枵", "p": "xiāo", "sp": "xiao", "r": "木", "sc": 9, "l": 5, "f": 5686},
  {"h": "粜", "p": "tiào", "sp": "tiao", "r": "米", "sc": 11, "l": 5, "f": 5687},
  {"h": "甙", "p": "dài", "sp": "dai", "r": "弋", "sc": 8, "l": 5, "f": 5688},
  {"h": "珲", "p": "huī", "sp": "hui", "r": "玉", "sc": 10, "l": 5, "f": 5689},
  {"h": "杓", "p": "biāo", "sp": "biao", "r": "木", "sc": 7, "l": 5, "f": 5690},
  {"h": "楸", "p": "qiū", "sp": "qiu", "r": "木", "sc": 13, "l": 5, "f": 5691},
  {"h": "楦", "p": "xuàn", "sp": "xuan", "r": "木", "sc": 13, "l": 5, "f": 5692},
  {"h": "疃", "p": "tuǎn", "sp": "tuan", "r": "田", "sc": 17, "l": 5, "f": 5693},
  {
    "h": "蛏",
    "p": "chēng",
    "sp": "cheng",
    "r": "虫",
    "sc": 11,
    "l": 5,
    "f": 5694
  },
  {"h": "蠛", "p": "miè", "sp": "mie", "r": "虫", "sc": 20, "l": 5, "f": 5695},
  {"h": "髌", "p": "bìn", "sp": "bin", "r": "骨", "sc": 19, "l": 5, "f": 5696},
  {"h": "茔", "p": "yíng", "sp": "ying", "r": "艸", "sc": 8, "l": 5, "f": 5697},
  {"h": "臨", "p": "lín", "sp": "lin", "r": "臣", "sc": 17, "l": 5, "f": 5698},
  {"h": "诼", "p": "zhuó", "sp": "zhuo", "r": "言", "sc": 10, "l": 5, "f": 5699},
  {"h": "嬖", "p": "bì", "sp": "bi", "r": "女", "sc": 16, "l": 5, "f": 5700},
  {"h": "耒", "p": "lěi", "sp": "lei", "r": "耒", "sc": 6, "l": 5, "f": 5701},
  {
    "h": "蜣",
    "p": "qiāng",
    "sp": "qiang",
    "r": "虫",
    "sc": 13,
    "l": 5,
    "f": 5702
  },
  {"h": "笄", "p": "jī", "sp": "ji", "r": "竹", "sc": 10, "l": 5, "f": 5703},
  {"h": "跣", "p": "xiǎn", "sp": "xian", "r": "足", "sc": 13, "l": 5, "f": 5704},
  {"h": "钣", "p": "bǎn", "sp": "ban", "r": "金", "sc": 9, "l": 5, "f": 5705},
  {"h": "戆", "p": "gàng", "sp": "gang", "r": "心", "sc": 25, "l": 5, "f": 5706},
  {"h": "蜉", "p": "fú", "sp": "fu", "r": "虫", "sc": 13, "l": 5, "f": 5707},
  {"h": "喾", "p": "kù", "sp": "ku", "r": "口", "sc": 12, "l": 5, "f": 5708},
  {"h": "铍", "p": "pī", "sp": "pi", "r": "金", "sc": 10, "l": 5, "f": 5709},
  {"h": "陉", "p": "xíng", "sp": "xing", "r": "阜", "sc": 7, "l": 5, "f": 5710},
  {"h": "薹", "p": "tái", "sp": "tai", "r": "艸", "sc": 17, "l": 5, "f": 5711},
  {"h": "肷", "p": "qiǎn", "sp": "qian", "r": "肉", "sc": 8, "l": 5, "f": 5712},
  {"h": "岵", "p": "hù", "sp": "hu", "r": "山", "sc": 8, "l": 5, "f": 5713},
  {"h": "瓴", "p": "líng", "sp": "ling", "r": "瓦", "sc": 9, "l": 5, "f": 5714},
  {"h": "荽", "p": "suī", "sp": "sui", "r": "艸", "sc": 10, "l": 5, "f": 5715},
  {"h": "怫", "p": "fú", "sp": "fu", "r": "心", "sc": 8, "l": 5, "f": 5716},
  {"h": "钭", "p": "tǒu", "sp": "tou", "r": "金", "sc": 9, "l": 5, "f": 5717},
  {"h": "窀", "p": "zhūn", "sp": "zhun", "r": "穴", "sc": 9, "l": 5, "f": 5718},
  {"h": "缯", "p": "zēng", "sp": "zeng", "r": "糸", "sc": 15, "l": 5, "f": 5719},
  {"h": "倬", "p": "zhuō", "sp": "zhuo", "r": "人", "sc": 10, "l": 5, "f": 5720},
  {"h": "摭", "p": "zhí", "sp": "zhi", "r": "手", "sc": 14, "l": 5, "f": 5721},
  {"h": "帔", "p": "pèi", "sp": "pei", "r": "巾", "sc": 8, "l": 5, "f": 5722},
  {"h": "楝", "p": "liàn", "sp": "lian", "r": "木", "sc": 13, "l": 5, "f": 5723},
  {"h": "痱", "p": "fèi", "sp": "fei", "r": "疒", "sc": 13, "l": 5, "f": 5724},
  {"h": "蚶", "p": "hān", "sp": "han", "r": "虫", "sc": 11, "l": 5, "f": 5725},
  {"h": "螬", "p": "cáo", "sp": "cao", "r": "虫", "sc": 17, "l": 5, "f": 5726},
  {"h": "髑", "p": "dú", "sp": "du", "r": "骨", "sc": 22, "l": 5, "f": 5727},
  {"h": "紅", "p": "hóng", "sp": "hong", "r": "糸", "sc": 9, "l": 5, "f": 5728},
  {"h": "鼗", "p": "táo", "sp": "tao", "r": "鼓", "sc": 19, "l": 5, "f": 5729},
  {"h": "狷", "p": "juàn", "sp": "juan", "r": "犬", "sc": 10, "l": 5, "f": 5730},
  {"h": "殛", "p": "jí", "sp": "ji", "r": "歹", "sc": 12, "l": 5, "f": 5731},
  {"h": "裉", "p": "kèn", "sp": "ken", "r": "衣", "sc": 11, "l": 5, "f": 5732},
  {"h": "粝", "p": "lì", "sp": "li", "r": "米", "sc": 11, "l": 5, "f": 5733},
  {"h": "萋", "p": "qī", "sp": "qi", "r": "艸", "sc": 11, "l": 5, "f": 5734},
  {"h": "葭", "p": "jiā", "sp": "jia", "r": "艸", "sc": 12, "l": 5, "f": 5735},
  {"h": "衽", "p": "rèn", "sp": "ren", "r": "衣", "sc": 9, "l": 5, "f": 5736},
  {"h": "鳢", "p": "lǐ", "sp": "li", "r": "魚", "sc": 21, "l": 5, "f": 5737},
  {"h": "傧", "p": "bīn", "sp": "bin", "r": "人", "sc": 12, "l": 5, "f": 5738},
  {"h": "喁", "p": "yóng", "sp": "yong", "r": "口", "sc": 12, "l": 5, "f": 5739},
  {"h": "嫘", "p": "léi", "sp": "lei", "r": "女", "sc": 14, "l": 5, "f": 5740},
  {"h": "罟", "p": "gǔ", "sp": "gu", "r": "网", "sc": 10, "l": 5, "f": 5741},
  {"h": "钌", "p": "liǎo", "sp": "liao", "r": "金", "sc": 7, "l": 5, "f": 5742},
  {"h": "裼", "p": "tì", "sp": "ti", "r": "衣", "sc": 13, "l": 5, "f": 5743},
  {"h": "愦", "p": "kuì", "sp": "kui", "r": "心", "sc": 12, "l": 5, "f": 5744},
  {"h": "蝽", "p": "chūn", "sp": "chun", "r": "虫", "sc": 15, "l": 5, "f": 5745},
  {"h": "雖", "p": "suī", "sp": "sui", "r": "隹", "sc": 17, "l": 5, "f": 5746},
  {"h": "锗", "p": "zhě", "sp": "zhe", "r": "金", "sc": 13, "l": 5, "f": 5747},
  {"h": "衿", "p": "jīn", "sp": "jin", "r": "衣", "sc": 9, "l": 5, "f": 5748},
  {"h": "粢", "p": "zī", "sp": "zi", "r": "米", "sc": 12, "l": 5, "f": 5749},
  {"h": "醵", "p": "jù", "sp": "ju", "r": "酉", "sc": 20, "l": 5, "f": 5750},
  {
    "h": "跫",
    "p": "qióng",
    "sp": "qiong",
    "r": "足",
    "sc": 13,
    "l": 5,
    "f": 5751
  },
  {"h": "鐾", "p": "bèi", "sp": "bei", "r": "金", "sc": 21, "l": 5, "f": 5752},
  {"h": "廛", "p": "chán", "sp": "chan", "r": "广", "sc": 15, "l": 5, "f": 5753},
  {"h": "墉", "p": "yōng", "sp": "yong", "r": "土", "sc": 14, "l": 5, "f": 5754},
  {"h": "哌", "p": "pài", "sp": "pai", "r": "口", "sc": 9, "l": 5, "f": 5755},
  {"h": "輕", "p": "qīng", "sp": "qing", "r": "車", "sc": 14, "l": 5, "f": 5756},
  {"h": "扦", "p": "qiān", "sp": "qian", "r": "手", "sc": 6, "l": 5, "f": 5757},
  {"h": "堇", "p": "jǐn", "sp": "jin", "r": "土", "sc": 11, "l": 5, "f": 5758},
  {"h": "婧", "p": "jìng", "sp": "jing", "r": "女", "sc": 11, "l": 5, "f": 5759},
  {"h": "暌", "p": "kuí", "sp": "kui", "r": "日", "sc": 13, "l": 5, "f": 5760},
  {"h": "罱", "p": "lǎn", "sp": "lan", "r": "网", "sc": 14, "l": 5, "f": 5761},
  {"h": "镞", "p": "zú", "sp": "zu", "r": "金", "sc": 16, "l": 5, "f": 5762},
  {"h": "蹰", "p": "chú", "sp": "chu", "r": "足", "sc": 19, "l": 5, "f": 5763},
  {"h": "陟", "p": "zhì", "sp": "zhi", "r": "阜", "sc": 9, "l": 5, "f": 5764},
  {"h": "鳔", "p": "biào", "sp": "biao", "r": "魚", "sc": 19, "l": 5, "f": 5765},
  {"h": "脘", "p": "wǎn", "sp": "wan", "r": "肉", "sc": 11, "l": 5, "f": 5766},
  {"h": "臟", "p": "zàng", "sp": "zang", "r": "肉", "sc": 21, "l": 5, "f": 5767},
  {"h": "岿", "p": "kuī", "sp": "kui", "r": "山", "sc": 8, "l": 5, "f": 5768},
  {"h": "侔", "p": "móu", "sp": "mou", "r": "人", "sc": 8, "l": 5, "f": 5769},
  {"h": "郾", "p": "yǎn", "sp": "yan", "r": "邑", "sc": 11, "l": 5, "f": 5770},
  {"h": "唿", "p": "hū", "sp": "hu", "r": "口", "sc": 11, "l": 5, "f": 5771},
  {"h": "砹", "p": "ài", "sp": "ai", "r": "石", "sc": 10, "l": 5, "f": 5772},
  {"h": "疴", "p": "kē", "sp": "ke", "r": "疒", "sc": 10, "l": 5, "f": 5773},
  {"h": "麸", "p": "fū", "sp": "fu", "r": "麥", "sc": 11, "l": 5, "f": 5774},
  {"h": "薨", "p": "hōng", "sp": "hong", "r": "艸", "sc": 16, "l": 5, "f": 5775},
  {"h": "綝", "p": "chēn", "sp": "chen", "r": "糸", "sc": 14, "l": 5, "f": 5776},
  {"h": "滁", "p": "chú", "sp": "chu", "r": "水", "sc": 12, "l": 5, "f": 5777},
  {"h": "偾", "p": "fèn", "sp": "fen", "r": "人", "sc": 11, "l": 5, "f": 5778},
  {"h": "拊", "p": "fǔ", "sp": "fu", "r": "手", "sc": 8, "l": 5, "f": 5779},
  {"h": "撺", "p": "cuān", "sp": "cuan", "r": "手", "sc": 15, "l": 5, "f": 5780},
  {"h": "呒", "p": "fǔ", "sp": "fu", "r": "口", "sc": 7, "l": 5, "f": 5781},
  {"h": "狯", "p": "kuài", "sp": "kuai", "r": "犬", "sc": 9, "l": 5, "f": 5782},
  {"h": "猢", "p": "hú", "sp": "hu", "r": "犬", "sc": 12, "l": 5, "f": 5783},
  {"h": "椁", "p": "guǒ", "sp": "guo", "r": "木", "sc": 12, "l": 5, "f": 5784},
  {"h": "榱", "p": "cuī", "sp": "cui", "r": "木", "sc": 14, "l": 5, "f": 5785},
  {"h": "罾", "p": "zēng", "sp": "zeng", "r": "网", "sc": 17, "l": 5, "f": 5786},
  {
    "h": "铳",
    "p": "chòng",
    "sp": "chong",
    "r": "金",
    "sc": 11,
    "l": 5,
    "f": 5787
  },
  {
    "h": "裎",
    "p": "chéng",
    "sp": "cheng",
    "r": "衣",
    "sc": 12,
    "l": 5,
    "f": 5788
  },
  {"h": "鳚", "p": "wèi", "sp": "wei", "r": "魚", "sc": 19, "l": 5, "f": 5789},
  {"h": "眦", "p": "zì", "sp": "zi", "r": "目", "sc": 11, "l": 5, "f": 5790},
  {"h": "璎", "p": "yīng", "sp": "ying", "r": "玉", "sc": 15, "l": 5, "f": 5791},
  {"h": "認", "p": "rèn", "sp": "ren", "r": "言", "sc": 14, "l": 5, "f": 5792},
  {"h": "崞", "p": "guō", "sp": "guo", "r": "山", "sc": 11, "l": 5, "f": 5793},
  {"h": "缇", "p": "tí", "sp": "ti", "r": "糸", "sc": 12, "l": 5, "f": 5794},
  {"h": "蝣", "p": "yóu", "sp": "you", "r": "虫", "sc": 15, "l": 5, "f": 5795},
  {"h": "萑", "p": "huán", "sp": "huan", "r": "艸", "sc": 11, "l": 5, "f": 5796},
  {"h": "狲", "p": "sūn", "sp": "sun", "r": "犬", "sc": 9, "l": 5, "f": 5797},
  {"h": "缱", "p": "qiǎn", "sp": "qian", "r": "糸", "sc": 16, "l": 5, "f": 5798},
  {"h": "晷", "p": "guǐ", "sp": "gui", "r": "日", "sc": 12, "l": 5, "f": 5799},
  {"h": "冼", "p": "xiǎn", "sp": "xian", "r": "冫", "sc": 8, "l": 5, "f": 5800},
  {"h": "痧", "p": "shā", "sp": "sha", "r": "疒", "sc": 12, "l": 5, "f": 5801},
  {"h": "統", "p": "tǒng", "sp": "tong", "r": "糸", "sc": 12, "l": 5, "f": 5802},
  {"h": "蕖", "p": "qú", "sp": "qu", "r": "艸", "sc": 14, "l": 5, "f": 5803},
  {"h": "狍", "p": "páo", "sp": "pao", "r": "犬", "sc": 8, "l": 5, "f": 5804},
  {"h": "憷", "p": "chù", "sp": "chu", "r": "心", "sc": 16, "l": 5, "f": 5805},
  {"h": "锛", "p": "bēn", "sp": "ben", "r": "金", "sc": 13, "l": 5, "f": 5806},
  {"h": "窨", "p": "xūn", "sp": "xun", "r": "穴", "sc": 14, "l": 5, "f": 5807},
  {"h": "袼", "p": "gē", "sp": "ge", "r": "衣", "sc": 11, "l": 5, "f": 5808},
  {"h": "帏", "p": "wéi", "sp": "wei", "r": "巾", "sc": 7, "l": 5, "f": 5809},
  {"h": "儋", "p": "dān", "sp": "dan", "r": "人", "sc": 15, "l": 5, "f": 5810},
  {"h": "绨", "p": "tí", "sp": "ti", "r": "糸", "sc": 10, "l": 5, "f": 5811},
  {"h": "疠", "p": "lì", "sp": "li", "r": "疒", "sc": 8, "l": 5, "f": 5812},
  {"h": "蘩", "p": "fán", "sp": "fan", "r": "艸", "sc": 20, "l": 5, "f": 5813},
  {"h": "嵝", "p": "lǒu", "sp": "lou", "r": "山", "sc": 12, "l": 5, "f": 5814},
  {"h": "庀", "p": "pǐ", "sp": "pi", "r": "广", "sc": 5, "l": 5, "f": 5815},
  {"h": "汜", "p": "sì", "sp": "si", "r": "水", "sc": 6, "l": 5, "f": 5816},
  {"h": "炅", "p": "jiǒng", "sp": "jiong", "r": "火", "sc": 8, "l": 5, "f": 5817},
  {"h": "煳", "p": "hú", "sp": "hu", "r": "火", "sc": 13, "l": 5, "f": 5818},
  {"h": "泶", "p": "xué", "sp": "xue", "r": "水", "sc": 9, "l": 5, "f": 5819},
  {"h": "瓠", "p": "hù", "sp": "hu", "r": "瓜", "sc": 11, "l": 5, "f": 5820},
  {"h": "窳", "p": "yǔ", "sp": "yu", "r": "穴", "sc": 15, "l": 5, "f": 5821},
  {"h": "虮", "p": "jǐ", "sp": "ji", "r": "虫", "sc": 8, "l": 5, "f": 5822},
  {"h": "蚰", "p": "yóu", "sp": "you", "r": "虫", "sc": 11, "l": 5, "f": 5823},
  {"h": "邰", "p": "tái", "sp": "tai", "r": "邑", "sc": 7, "l": 5, "f": 5824},
  {"h": "苊", "p": "è", "sp": "e", "r": "艸", "sc": 7, "l": 5, "f": 5825},
  {"h": "砀", "p": "dàng", "sp": "dang", "r": "石", "sc": 8, "l": 5, "f": 5826},
  {"h": "捩", "p": "liè", "sp": "lie", "r": "手", "sc": 11, "l": 5, "f": 5827},
  {"h": "蹉", "p": "cuō", "sp": "cuo", "r": "足", "sc": 16, "l": 5, "f": 5828},
  {"h": "莪", "p": "é", "sp": "e", "r": "艸", "sc": 10, "l": 5, "f": 5829},
  {
    "h": "螽",
    "p": "zhōng",
    "sp": "zhong",
    "r": "虫",
    "sc": 17,
    "l": 5,
    "f": 5830
  },
  {"h": "覺", "p": "jué", "sp": "jue", "r": "見", "sc": 20, "l": 5, "f": 5831},
  {"h": "蘼", "p": "mí", "sp": "mi", "r": "艸", "sc": 22, "l": 5, "f": 5832},
  {"h": "槔", "p": "gāo", "sp": "gao", "r": "木", "sc": 14, "l": 5, "f": 5833},
  {"h": "曛", "p": "xūn", "sp": "xun", "r": "日", "sc": 18, "l": 5, "f": 5834},
  {"h": "蛲", "p": "náo", "sp": "nao", "r": "虫", "sc": 12, "l": 5, "f": 5835},
  {"h": "鹾", "p": "cuó", "sp": "cuo", "r": "鹵", "sc": 16, "l": 5, "f": 5836},
  {"h": "隹", "p": "zhuī", "sp": "zhui", "r": "隹", "sc": 8, "l": 5, "f": 5837},
  {"h": "犸", "p": "mà", "sp": "ma", "r": "犬", "sc": 6, "l": 5, "f": 5838},
  {"h": "衄", "p": "nǜ", "sp": "nu", "r": "血", "sc": 10, "l": 5, "f": 5839},
  {"h": "觀", "p": "guān", "sp": "guan", "r": "見", "sc": 24, "l": 5, "f": 5840},
  {
    "h": "轉",
    "p": "zhuǎn",
    "sp": "zhuan",
    "r": "車",
    "sc": 18,
    "l": 5,
    "f": 5841
  },
  {
    "h": "銎",
    "p": "qióng",
    "sp": "qiong",
    "r": "金",
    "sc": 14,
    "l": 5,
    "f": 5842
  },
  {"h": "泫", "p": "xuàn", "sp": "xuan", "r": "水", "sc": 8, "l": 5, "f": 5843},
  {"h": "玢", "p": "bīn", "sp": "bin", "r": "玉", "sc": 8, "l": 5, "f": 5844},
  {"h": "辊", "p": "gǔn", "sp": "gun", "r": "車", "sc": 12, "l": 5, "f": 5845},
  {"h": "瞋", "p": "chēn", "sp": "chen", "r": "目", "sc": 15, "l": 5, "f": 5846},
  {"h": "墀", "p": "chí", "sp": "chi", "r": "土", "sc": 15, "l": 5, "f": 5847},
  {"h": "酐", "p": "gān", "sp": "gan", "r": "酉", "sc": 10, "l": 5, "f": 5848},
  {"h": "隱", "p": "yǐn", "sp": "yin", "r": "阜", "sc": 16, "l": 5, "f": 5849},
  {"h": "堞", "p": "dié", "sp": "die", "r": "土", "sc": 12, "l": 5, "f": 5850},
  {"h": "尥", "p": "liào", "sp": "liao", "r": "尢", "sc": 6, "l": 5, "f": 5851},
  {"h": "嚯", "p": "huò", "sp": "huo", "r": "口", "sc": 19, "l": 5, "f": 5852},
  {"h": "猗", "p": "yī", "sp": "yi", "r": "犬", "sc": 11, "l": 5, "f": 5853},
  {"h": "逑", "p": "qiú", "sp": "qiu", "r": "辵", "sc": 10, "l": 5, "f": 5854},
  {"h": "逯", "p": "lù", "sp": "lu", "r": "辵", "sc": 11, "l": 5, "f": 5855},
  {"h": "硖", "p": "xiá", "sp": "xia", "r": "石", "sc": 11, "l": 5, "f": 5856},
  {"h": "噻", "p": "sāi", "sp": "sai", "r": "口", "sc": 16, "l": 5, "f": 5857},
  {"h": "嵛", "p": "yú", "sp": "yu", "r": "山", "sc": 12, "l": 5, "f": 5858},
  {"h": "畀", "p": "bì", "sp": "bi", "r": "田", "sc": 8, "l": 5, "f": 5859},
  {"h": "運", "p": "yùn", "sp": "yun", "r": "辵", "sc": 12, "l": 5, "f": 5860},
  {"h": "鲃", "p": "bā", "sp": "ba", "r": "魚", "sc": 12, "l": 5, "f": 5861},
  {"h": "偬", "p": "zǒng", "sp": "zong", "r": "人", "sc": 11, "l": 5, "f": 5862},
  {"h": "鄞", "p": "yín", "sp": "yin", "r": "邑", "sc": 13, "l": 5, "f": 5863},
  {"h": "呖", "p": "lì", "sp": "li", "r": "口", "sc": 7, "l": 5, "f": 5864},
  {"h": "溧", "p": "lì", "sp": "li", "r": "水", "sc": 13, "l": 5, "f": 5865},
  {"h": "嬲", "p": "niǎo", "sp": "niao", "r": "女", "sc": 17, "l": 5, "f": 5866},
  {"h": "肭", "p": "nà", "sp": "na", "r": "肉", "sc": 8, "l": 5, "f": 5867},
  {"h": "鹈", "p": "tí", "sp": "ti", "r": "鳥", "sc": 12, "l": 5, "f": 5868},
  {"h": "鹱", "p": "hù", "sp": "hu", "r": "鳥", "sc": 18, "l": 5, "f": 5869},
  {"h": "窭", "p": "jù", "sp": "ju", "r": "穴", "sc": 14, "l": 5, "f": 5870},
  {"h": "黧", "p": "lí", "sp": "li", "r": "黑", "sc": 20, "l": 5, "f": 5871},
  {"h": "谵", "p": "zhān", "sp": "zhan", "r": "言", "sc": 15, "l": 5, "f": 5872},
  {"h": "沆", "p": "hàng", "sp": "hang", "r": "水", "sc": 7, "l": 5, "f": 5873},
  {"h": "嫒", "p": "ài", "sp": "ai", "r": "女", "sc": 13, "l": 5, "f": 5874},
  {"h": "塬", "p": "yuán", "sp": "yuan", "r": "土", "sc": 13, "l": 5, "f": 5875},
  {"h": "缣", "p": "jiān", "sp": "jian", "r": "糸", "sc": 13, "l": 5, "f": 5876},
  {"h": "篯", "p": "jiān", "sp": "jian", "r": "竹", "sc": 16, "l": 5, "f": 5877},
  {"h": "酃", "p": "líng", "sp": "ling", "r": "邑", "sc": 19, "l": 5, "f": 5878},
  {"h": "喱", "p": "lí", "sp": "li", "r": "口", "sc": 12, "l": 5, "f": 5879},
  {"h": "泔", "p": "gān", "sp": "gan", "r": "水", "sc": 8, "l": 5, "f": 5880},
  {"h": "溘", "p": "kè", "sp": "ke", "r": "水", "sc": 13, "l": 5, "f": 5881},
  {"h": "迕", "p": "wù", "sp": "wu", "r": "辵", "sc": 7, "l": 5, "f": 5882},
  {"h": "肀", "p": "yù", "sp": "yu", "r": "聿", "sc": 4, "l": 5, "f": 5883},
  {"h": "秫", "p": "shú", "sp": "shu", "r": "禾", "sc": 10, "l": 5, "f": 5884},
  {"h": "裣", "p": "liǎn", "sp": "lian", "r": "衣", "sc": 12, "l": 5, "f": 5885},
  {"h": "铋", "p": "bì", "sp": "bi", "r": "金", "sc": 10, "l": 5, "f": 5886},
  {"h": "蒌", "p": "lóu", "sp": "lou", "r": "艸", "sc": 12, "l": 5, "f": 5887},
  {"h": "曩", "p": "nǎng", "sp": "nang", "r": "日", "sc": 21, "l": 5, "f": 5888},
  {"h": "赀", "p": "zī", "sp": "zi", "r": "貝", "sc": 10, "l": 5, "f": 5889},
  {"h": "箪", "p": "dān", "sp": "dan", "r": "竹", "sc": 14, "l": 5, "f": 5890},
  {"h": "朊", "p": "ruǎn", "sp": "ruan", "r": "肉", "sc": 8, "l": 5, "f": 5891},
  {"h": "鳙", "p": "yōng", "sp": "yong", "r": "魚", "sc": 19, "l": 5, "f": 5892},
  {"h": "仫", "p": "mù", "sp": "mu", "r": "人", "sc": 5, "l": 5, "f": 5893},
  {"h": "钎", "p": "qiān", "sp": "qian", "r": "金", "sc": 8, "l": 5, "f": 5894},
  {"h": "芑", "p": "qǐ", "sp": "qi", "r": "艸", "sc": 6, "l": 5, "f": 5895},
  {"h": "胙", "p": "zuò", "sp": "zuo", "r": "肉", "sc": 9, "l": 5, "f": 5896},
  {"h": "盱", "p": "xū", "sp": "xu", "r": "目", "sc": 8, "l": 5, "f": 5897},
  {"h": "糇", "p": "hóu", "sp": "hou", "r": "米", "sc": 15, "l": 5, "f": 5898},
  {"h": "挹", "p": "yì", "sp": "yi", "r": "手", "sc": 10, "l": 5, "f": 5899},
  {"h": "捭", "p": "bǎi", "sp": "bai", "r": "手", "sc": 11, "l": 5, "f": 5900},
  {"h": "悱", "p": "fěi", "sp": "fei", "r": "心", "sc": 11, "l": 5, "f": 5901},
  {"h": "鬟", "p": "huán", "sp": "huan", "r": "髟", "sc": 23, "l": 5, "f": 5902},
  {"h": "緩", "p": "huǎn", "sp": "huan", "r": "糸", "sc": 15, "l": 5, "f": 5903},
  {"h": "請", "p": "qǐng", "sp": "qing", "r": "言", "sc": 15, "l": 5, "f": 5904},
  {"h": "崤", "p": "xiáo", "sp": "xiao", "r": "山", "sc": 11, "l": 5, "f": 5905},
  {"h": "澶", "p": "chán", "sp": "chan", "r": "水", "sc": 16, "l": 5, "f": 5906},
  {"h": "甾", "p": "zāi", "sp": "zai", "r": "田", "sc": 8, "l": 5, "f": 5907},
  {"h": "欹", "p": "yī", "sp": "yi", "r": "欠", "sc": 12, "l": 5, "f": 5908},
  {"h": "瞽", "p": "gǔ", "sp": "gu", "r": "目", "sc": 18, "l": 5, "f": 5909},
  {"h": "钇", "p": "yǐ", "sp": "yi", "r": "金", "sc": 6, "l": 5, "f": 5910},
  {"h": "鹪", "p": "jiāo", "sp": "jiao", "r": "鳥", "sc": 17, "l": 5, "f": 5911},
  {"h": "鞔", "p": "mán", "sp": "man", "r": "革", "sc": 16, "l": 5, "f": 5912},
  {"h": "缡", "p": "lí", "sp": "li", "r": "糸", "sc": 13, "l": 5, "f": 5913},
  {"h": "铯", "p": "sè", "sp": "se", "r": "金", "sc": 11, "l": 5, "f": 5914},
  {"h": "鲚", "p": "jì", "sp": "ji", "r": "魚", "sc": 14, "l": 5, "f": 5915},
  {"h": "組", "p": "zǔ", "sp": "zu", "r": "糸", "sc": 11, "l": 5, "f": 5916},
  {
    "h": "嘬",
    "p": "chuài",
    "sp": "chuai",
    "r": "口",
    "sc": 15,
    "l": 5,
    "f": 5917
  },
  {"h": "庹", "p": "tuǒ", "sp": "tuo", "r": "广", "sc": 11, "l": 5, "f": 5918},
  {"h": "渖", "p": "shěn", "sp": "shen", "r": "水", "sc": 11, "l": 5, "f": 5919},
  {"h": "湔", "p": "jiān", "sp": "jian", "r": "水", "sc": 12, "l": 5, "f": 5920},
  {"h": "玎", "p": "dīng", "sp": "ding", "r": "玉", "sc": 6, "l": 5, "f": 5921},
  {"h": "锜", "p": "qí", "sp": "qi", "r": "金", "sc": 13, "l": 5, "f": 5922},
  {"h": "锊", "p": "lüè", "sp": "lue", "r": "金", "sc": 12, "l": 5, "f": 5923},
  {"h": "舾", "p": "xī", "sp": "xi", "r": "舟", "sc": 12, "l": 5, "f": 5924},
  {"h": "籼", "p": "xiān", "sp": "xian", "r": "米", "sc": 9, "l": 5, "f": 5925},
  {
    "h": "阊",
    "p": "chāng",
    "sp": "chang",
    "r": "門",
    "sc": 11,
    "l": 5,
    "f": 5926
  },
  {"h": "祕", "p": "mì", "sp": "mi", "r": "示", "sc": 9, "l": 5, "f": 5927},
  {"h": "猊", "p": "ní", "sp": "ni", "r": "犬", "sc": 11, "l": 5, "f": 5928},
  {"h": "燹", "p": "xiǎn", "sp": "xian", "r": "火", "sc": 18, "l": 5, "f": 5929},
  {"h": "葑", "p": "fēng", "sp": "feng", "r": "艸", "sc": 12, "l": 5, "f": 5930},
  {"h": "蓼", "p": "liǎo", "sp": "liao", "r": "艸", "sc": 14, "l": 5, "f": 5931},
  {
    "h": "幛",
    "p": "zhàng",
    "sp": "zhang",
    "r": "巾",
    "sc": 14,
    "l": 5,
    "f": 5932
  },
  {"h": "岣", "p": "gǒu", "sp": "gou", "r": "山", "sc": 8, "l": 5, "f": 5933},
  {"h": "浼", "p": "měi", "sp": "mei", "r": "水", "sc": 10, "l": 5, "f": 5934},
  {"h": "甯", "p": "níng", "sp": "ning", "r": "用", "sc": 12, "l": 5, "f": 5935},
  {"h": "瑷", "p": "ài", "sp": "ai", "r": "玉", "sc": 14, "l": 5, "f": 5936},
  {"h": "敫", "p": "jiǎo", "sp": "jiao", "r": "攴", "sc": 13, "l": 5, "f": 5937},
  {"h": "钔", "p": "mén", "sp": "men", "r": "金", "sc": 8, "l": 5, "f": 5938},
  {"h": "钫", "p": "fāng", "sp": "fang", "r": "金", "sc": 9, "l": 5, "f": 5939},
  {"h": "锼", "p": "sōu", "sp": "sou", "r": "金", "sc": 14, "l": 5, "f": 5940},
  {"h": "锿", "p": "āi", "sp": "ai", "r": "金", "sc": 14, "l": 5, "f": 5941},
  {"h": "癔", "p": "yì", "sp": "yi", "r": "疒", "sc": 18, "l": 5, "f": 5942},
  {"h": "穸", "p": "xī", "sp": "xi", "r": "穴", "sc": 8, "l": 5, "f": 5943},
  {"h": "褊", "p": "biǎn", "sp": "bian", "r": "衣", "sc": 14, "l": 5, "f": 5944},
  {"h": "蚍", "p": "pí", "sp": "pi", "r": "虫", "sc": 10, "l": 5, "f": 5945},
  {"h": "篦", "p": "bì", "sp": "bi", "r": "竹", "sc": 16, "l": 5, "f": 5946},
  {"h": "麇", "p": "jūn", "sp": "jun", "r": "鹿", "sc": 16, "l": 5, "f": 5947},
  {"h": "樘", "p": "táng", "sp": "tang", "r": "木", "sc": 15, "l": 5, "f": 5948},
  {"h": "钯", "p": "bǎ", "sp": "ba", "r": "金", "sc": 9, "l": 5, "f": 5949},
  {"h": "禇", "p": "zhě", "sp": "zhe", "r": "示", "sc": 12, "l": 5, "f": 5950},
  {"h": "铒", "p": "ěr", "sp": "er", "r": "金", "sc": 11, "l": 5, "f": 5951},
  {"h": "續", "p": "xù", "sp": "xu", "r": "糸", "sc": 21, "l": 5, "f": 5952},
  {"h": "莩", "p": "fú", "sp": "fu", "r": "艸", "sc": 10, "l": 5, "f": 5953},
  {"h": "嵯", "p": "cuó", "sp": "cuo", "r": "山", "sc": 12, "l": 5, "f": 5954},
  {"h": "逭", "p": "huàn", "sp": "huan", "r": "辵", "sc": 11, "l": 5, "f": 5955},
  {
    "h": "遄",
    "p": "chuán",
    "sp": "chuan",
    "r": "辵",
    "sc": 12,
    "l": 5,
    "f": 5956
  },
  {"h": "戗", "p": "qiāng", "sp": "qiang", "r": "戈", "sc": 8, "l": 5, "f": 5957},
  {"h": "睃", "p": "suō", "sp": "suo", "r": "目", "sc": 12, "l": 5, "f": 5958},
  {"h": "鮈", "p": "jū", "sp": "ju", "r": "魚", "sc": 16, "l": 5, "f": 5959},
  {"h": "瀣", "p": "xiè", "sp": "xie", "r": "水", "sc": 19, "l": 5, "f": 5960},
  {"h": "皴", "p": "cūn", "sp": "cun", "r": "皮", "sc": 12, "l": 5, "f": 5961},
  {"h": "泮", "p": "pàn", "sp": "pan", "r": "水", "sc": 8, "l": 5, "f": 5962},
  {"h": "轫", "p": "rèn", "sp": "ren", "r": "車", "sc": 7, "l": 5, "f": 5963},
  {"h": "褰", "p": "qiān", "sp": "qian", "r": "衣", "sc": 16, "l": 5, "f": 5964},
  {"h": "炱", "p": "tái", "sp": "tai", "r": "火", "sc": 9, "l": 5, "f": 5965},
  {"h": "醍", "p": "tí", "sp": "ti", "r": "酉", "sc": 16, "l": 5, "f": 5966},
  {"h": "锱", "p": "zī", "sp": "zi", "r": "金", "sc": 13, "l": 5, "f": 5967},
  {
    "h": "篁",
    "p": "huáng",
    "sp": "huang",
    "r": "竹",
    "sc": 15,
    "l": 5,
    "f": 5968
  },
  {"h": "葚", "p": "rèn", "sp": "ren", "r": "艸", "sc": 12, "l": 5, "f": 5969},
  {"h": "難", "p": "nán", "sp": "nan", "r": "隹", "sc": 19, "l": 5, "f": 5970},
  {"h": "矇", "p": "méng", "sp": "meng", "r": "目", "sc": 18, "l": 5, "f": 5971},
  {"h": "驽", "p": "nú", "sp": "nu", "r": "馬", "sc": 8, "l": 5, "f": 5972},
  {"h": "辚", "p": "lín", "sp": "lin", "r": "車", "sc": 16, "l": 5, "f": 5973},
  {"h": "睥", "p": "pì", "sp": "pi", "r": "目", "sc": 13, "l": 5, "f": 5974},
  {"h": "鸺", "p": "xiū", "sp": "xiu", "r": "鳥", "sc": 11, "l": 5, "f": 5975},
  {
    "h": "筇",
    "p": "qióng",
    "sp": "qiong",
    "r": "竹",
    "sc": 11,
    "l": 5,
    "f": 5976
  },
  {"h": "戥", "p": "děng", "sp": "deng", "r": "戈", "sc": 13, "l": 5, "f": 5977},
  {"h": "髀", "p": "bì", "sp": "bi", "r": "骨", "sc": 17, "l": 5, "f": 5978},
  {"h": "驺", "p": "zōu", "sp": "zou", "r": "馬", "sc": 8, "l": 5, "f": 5979},
  {"h": "頸", "p": "jǐng", "sp": "jing", "r": "頁", "sc": 16, "l": 5, "f": 5980},
  {"h": "哙", "p": "kuài", "sp": "kuai", "r": "口", "sc": 9, "l": 5, "f": 5981},
  {"h": "濞", "p": "bì", "sp": "bi", "r": "水", "sc": 17, "l": 5, "f": 5982},
  {"h": "逄", "p": "páng", "sp": "pang", "r": "辵", "sc": 9, "l": 5, "f": 5983},
  {"h": "桤", "p": "qī", "sp": "qi", "r": "木", "sc": 10, "l": 5, "f": 5984},
  {"h": "攵", "p": "pū", "sp": "pu", "r": "攴", "sc": 4, "l": 5, "f": 5985},
  {"h": "炻", "p": "shí", "sp": "shi", "r": "火", "sc": 9, "l": 5, "f": 5986},
  {"h": "磙", "p": "gǔn", "sp": "gun", "r": "石", "sc": 15, "l": 5, "f": 5987},
  {"h": "疳", "p": "gān", "sp": "gan", "r": "疒", "sc": 10, "l": 5, "f": 5988},
  {"h": "醭", "p": "bú", "sp": "bu", "r": "酉", "sc": 19, "l": 5, "f": 5989},
  {
    "h": "鳇",
    "p": "huáng",
    "sp": "huang",
    "r": "魚",
    "sc": 17,
    "l": 5,
    "f": 5990
  },
  {"h": "鹮", "p": "huán", "sp": "huan", "r": "鳥", "sc": 18, "l": 5, "f": 5991},
  {"h": "迓", "p": "yà", "sp": "ya", "r": "辵", "sc": 7, "l": 5, "f": 5992},
  {"h": "眇", "p": "miǎo", "sp": "miao", "r": "目", "sc": 9, "l": 5, "f": 5993},
  {"h": "楮", "p": "chǔ", "sp": "chu", "r": "木", "sc": 12, "l": 5, "f": 5994},
  {"h": "砜", "p": "fēng", "sp": "feng", "r": "石", "sc": 9, "l": 5, "f": 5995},
  {"h": "謝", "p": "xiè", "sp": "xie", "r": "言", "sc": 17, "l": 5, "f": 5996},
  {"h": "離", "p": "lí", "sp": "li", "r": "隹", "sc": 18, "l": 5, "f": 5997},
  {"h": "約", "p": "yuē", "sp": "yue", "r": "糸", "sc": 9, "l": 5, "f": 5998},
  {"h": "菘", "p": "sōng", "sp": "song", "r": "艸", "sc": 11, "l": 5, "f": 5999},
  {"h": "馇", "p": "chā", "sp": "cha", "r": "食", "sc": 12, "l": 5, "f": 6000},
  {"h": "栌", "p": "lú", "sp": "lu", "r": "木", "sc": 9, "l": 5, "f": 6001},
  {"h": "醐", "p": "hú", "sp": "hu", "r": "酉", "sc": 16, "l": 5, "f": 6002},
  {"h": "唢", "p": "suǒ", "sp": "suo", "r": "口", "sc": 10, "l": 5, "f": 6003},
  {"h": "獍", "p": "jìng", "sp": "jing", "r": "犬", "sc": 14, "l": 5, "f": 6004},
  {"h": "殳", "p": "shū", "sp": "shu", "r": "殳", "sc": 4, "l": 5, "f": 6005},
  {"h": "仡", "p": "gē", "sp": "ge", "r": "人", "sc": 5, "l": 5, "f": 6006},
  {"h": "軟", "p": "ruǎn", "sp": "ruan", "r": "車", "sc": 11, "l": 5, "f": 6007},
  {"h": "刿", "p": "guì", "sp": "gui", "r": "刀", "sc": 8, "l": 5, "f": 6008},
  {"h": "葜", "p": "qiā", "sp": "qia", "r": "艸", "sc": 12, "l": 5, "f": 6009},
  {"h": "薅", "p": "hāo", "sp": "hao", "r": "艸", "sc": 16, "l": 5, "f": 6010},
  {"h": "湓", "p": "pén", "sp": "pen", "r": "水", "sc": 12, "l": 5, "f": 6011},
  {"h": "搴", "p": "qiān", "sp": "qian", "r": "手", "sc": 14, "l": 5, "f": 6012},
  {"h": "尕", "p": "gǎ", "sp": "ga", "r": "小", "sc": 5, "l": 5, "f": 6013},
  {"h": "磴", "p": "dèng", "sp": "deng", "r": "石", "sc": 17, "l": 5, "f": 6014},
  {"h": "锝", "p": "dé", "sp": "de", "r": "金", "sc": 13, "l": 5, "f": 6015},
  {"h": "镅", "p": "méi", "sp": "mei", "r": "金", "sc": 14, "l": 5, "f": 6016},
  {"h": "艚", "p": "cáo", "sp": "cao", "r": "舟", "sc": 17, "l": 5, "f": 6017},
  {"h": "殄", "p": "tiǎn", "sp": "tian", "r": "歹", "sc": 9, "l": 5, "f": 6018},
  {"h": "暝", "p": "míng", "sp": "ming", "r": "日", "sc": 14, "l": 5, "f": 6019},
  {"h": "祢", "p": "mí", "sp": "mi", "r": "示", "sc": 9, "l": 5, "f": 6020},
  {"h": "豉", "p": "shì", "sp": "shi", "r": "豆", "sc": 11, "l": 5, "f": 6021},
  {"h": "垴", "p": "nǎo", "sp": "nao", "r": "土", "sc": 9, "l": 5, "f": 6022},
  {"h": "維", "p": "wéi", "sp": "wei", "r": "糸", "sc": 14, "l": 5, "f": 6023},
  {"h": "瓿", "p": "bù", "sp": "bu", "r": "瓦", "sc": 12, "l": 5, "f": 6024},
  {"h": "煺", "p": "tuì", "sp": "tui", "r": "火", "sc": 13, "l": 5, "f": 6025},
  {
    "h": "蟥",
    "p": "huáng",
    "sp": "huang",
    "r": "虫",
    "sc": 17,
    "l": 5,
    "f": 6026
  },
  {"h": "蹀", "p": "dié", "sp": "die", "r": "足", "sc": 16, "l": 5, "f": 6027},
  {"h": "門", "p": "mén", "sp": "men", "r": "門", "sc": 8, "l": 5, "f": 6028},
  {"h": "黼", "p": "fǔ", "sp": "fu", "r": "黹", "sc": 19, "l": 5, "f": 6029},
  {"h": "氙", "p": "xiān", "sp": "xian", "r": "气", "sc": 7, "l": 5, "f": 6030},
  {"h": "铑", "p": "lǎo", "sp": "lao", "r": "金", "sc": 11, "l": 5, "f": 6031},
  {"h": "積", "p": "jī", "sp": "ji", "r": "禾", "sc": 16, "l": 5, "f": 6032},
  {"h": "線", "p": "xiàn", "sp": "xian", "r": "糸", "sc": 15, "l": 5, "f": 6033},
  {"h": "蒡", "p": "bàng", "sp": "bang", "r": "艸", "sc": 13, "l": 5, "f": 6034},
  {"h": "杪", "p": "miǎo", "sp": "miao", "r": "木", "sc": 8, "l": 5, "f": 6035},
  {"h": "肜", "p": "róng", "sp": "rong", "r": "肉", "sc": 7, "l": 5, "f": 6036},
  {"h": "膂", "p": "lǚ", "sp": "lu", "r": "肉", "sc": 14, "l": 5, "f": 6037},
  {"h": "砬", "p": "lá", "sp": "la", "r": "石", "sc": 10, "l": 5, "f": 6038},
  {"h": "硭", "p": "máng", "sp": "mang", "r": "石", "sc": 11, "l": 5, "f": 6039},
  {"h": "酽", "p": "yàn", "sp": "yan", "r": "酉", "sc": 14, "l": 5, "f": 6040},
  {"h": "踟", "p": "chí", "sp": "chi", "r": "足", "sc": 15, "l": 5, "f": 6041},
  {"h": "鲷", "p": "diāo", "sp": "diao", "r": "魚", "sc": 16, "l": 5, "f": 6042},
  {"h": "蒴", "p": "shuò", "sp": "shuo", "r": "艸", "sc": 13, "l": 5, "f": 6043},
  {"h": "藁", "p": "gǎo", "sp": "gao", "r": "艸", "sc": 17, "l": 5, "f": 6044},
  {"h": "猡", "p": "luó", "sp": "luo", "r": "犬", "sc": 11, "l": 5, "f": 6045},
  {"h": "旒", "p": "liú", "sp": "liu", "r": "方", "sc": 13, "l": 5, "f": 6046},
  {"h": "硇", "p": "náo", "sp": "nao", "r": "石", "sc": 11, "l": 5, "f": 6047},
  {"h": "锍", "p": "liǔ", "sp": "liu", "r": "金", "sc": 12, "l": 5, "f": 6048},
  {"h": "鞲", "p": "gōu", "sp": "gou", "r": "革", "sc": 19, "l": 5, "f": 6049},
  {"h": "哿", "p": "gě", "sp": "ge", "r": "口", "sc": 10, "l": 5, "f": 6050},
  {"h": "皤", "p": "pó", "sp": "po", "r": "白", "sc": 17, "l": 5, "f": 6051},
  {"h": "哏", "p": "gén", "sp": "gen", "r": "口", "sc": 9, "l": 5, "f": 6052},
  {"h": "殍", "p": "piǎo", "sp": "piao", "r": "歹", "sc": 11, "l": 5, "f": 6053},
  {"h": "漯", "p": "luò", "sp": "luo", "r": "水", "sc": 14, "l": 5, "f": 6054},
  {"h": "確", "p": "què", "sp": "que", "r": "石", "sc": 15, "l": 5, "f": 6055},
  {"h": "織", "p": "zhī", "sp": "zhi", "r": "糸", "sc": 18, "l": 5, "f": 6056},
  {"h": "連", "p": "lián", "sp": "lian", "r": "辵", "sc": 10, "l": 5, "f": 6057},
  {"h": "谂", "p": "shěn", "sp": "shen", "r": "言", "sc": 10, "l": 5, "f": 6058},
  {"h": "跶", "p": "tà", "sp": "ta", "r": "足", "sc": 13, "l": 5, "f": 6059},
  {"h": "橥", "p": "zhū", "sp": "zhu", "r": "木", "sc": 15, "l": 5, "f": 6060},
  {"h": "镬", "p": "huò", "sp": "huo", "r": "金", "sc": 18, "l": 5, "f": 6061},
  {"h": "鸲", "p": "qú", "sp": "qu", "r": "鳥", "sc": 10, "l": 5, "f": 6062},
  {"h": "姝", "p": "shū", "sp": "shu", "r": "女", "sc": 9, "l": 5, "f": 6063},
  {"h": "茭", "p": "jiāo", "sp": "jiao", "r": "艸", "sc": 9, "l": 5, "f": 6064},
  {"h": "窆", "p": "biǎn", "sp": "bian", "r": "穴", "sc": 9, "l": 5, "f": 6065},
  {"h": "髟", "p": "biāo", "sp": "biao", "r": "髟", "sc": 10, "l": 5, "f": 6066},
  {"h": "畋", "p": "tián", "sp": "tian", "r": "田", "sc": 9, "l": 5, "f": 6067},
  {"h": "溏", "p": "táng", "sp": "tang", "r": "水", "sc": 13, "l": 5, "f": 6068},
  {"h": "鲔", "p": "wěi", "sp": "wei", "r": "魚", "sc": 14, "l": 5, "f": 6069},
  {"h": "顯", "p": "xiǎn", "sp": "xian", "r": "頁", "sc": 23, "l": 5, "f": 6070},
  {"h": "喹", "p": "kuí", "sp": "kui", "r": "口", "sc": 12, "l": 5, "f": 6071},
  {"h": "紹", "p": "shào", "sp": "shao", "r": "糸", "sc": 11, "l": 5, "f": 6072},
  {"h": "吣", "p": "qìn", "sp": "qin", "r": "口", "sc": 7, "l": 5, "f": 6073},
  {"h": "羼", "p": "chàn", "sp": "chan", "r": "羊", "sc": 21, "l": 5, "f": 6074},
  {"h": "缧", "p": "léi", "sp": "lei", "r": "糸", "sc": 14, "l": 5, "f": 6075},
  {"h": "窬", "p": "yú", "sp": "yu", "r": "穴", "sc": 14, "l": 5, "f": 6076},
  {"h": "骘", "p": "zhì", "sp": "zhi", "r": "馬", "sc": 12, "l": 5, "f": 6077},
  {"h": "赉", "p": "lài", "sp": "lai", "r": "貝", "sc": 11, "l": 5, "f": 6078},
  {"h": "镓", "p": "jiā", "sp": "jia", "r": "金", "sc": 15, "l": 5, "f": 6079},
  {"h": "邳", "p": "pī", "sp": "pi", "r": "邑", "sc": 7, "l": 5, "f": 6080},
  {"h": "璩", "p": "qú", "sp": "qu", "r": "玉", "sc": 17, "l": 5, "f": 6081},
  {"h": "鸸", "p": "ér", "sp": "er", "r": "鳥", "sc": 11, "l": 5, "f": 6082},
  {"h": "涠", "p": "wéi", "sp": "wei", "r": "水", "sc": 10, "l": 5, "f": 6083},
  {"h": "溲", "p": "sōu", "sp": "sou", "r": "水", "sc": 12, "l": 5, "f": 6084},
  {"h": "孥", "p": "nú", "sp": "nu", "r": "子", "sc": 8, "l": 5, "f": 6085},
  {"h": "敉", "p": "mǐ", "sp": "mi", "r": "攴", "sc": 10, "l": 5, "f": 6086},
  {"h": "筌", "p": "quán", "sp": "quan", "r": "竹", "sc": 12, "l": 5, "f": 6087},
  {"h": "䜣", "p": "xīn", "sp": "xin", "r": "", "sc": 11, "l": 5, "f": 6088},
  {"h": "義", "p": "yì", "sp": "yi", "r": "羊", "sc": 13, "l": 5, "f": 6089},
  {"h": "衛", "p": "wèi", "sp": "wei", "r": "行", "sc": 15, "l": 5, "f": 6090},
  {
    "h": "響",
    "p": "xiǎng",
    "sp": "xiang",
    "r": "音",
    "sc": 20,
    "l": 5,
    "f": 6091
  },
  {"h": "蹓", "p": "liū", "sp": "liu", "r": "足", "sc": 17, "l": 5, "f": 6092},
  {"h": "垓", "p": "gāi", "sp": "gai", "r": "土", "sc": 9, "l": 5, "f": 6093},
  {"h": "苒", "p": "rǎn", "sp": "ran", "r": "艸", "sc": 8, "l": 5, "f": 6094},
  {"h": "荸", "p": "bí", "sp": "bi", "r": "艸", "sc": 10, "l": 5, "f": 6095},
  {"h": "甓", "p": "pì", "sp": "pi", "r": "瓦", "sc": 17, "l": 5, "f": 6096},
  {"h": "锪", "p": "huō", "sp": "huo", "r": "金", "sc": 13, "l": 5, "f": 6097},
  {"h": "镄", "p": "fèi", "sp": "fei", "r": "金", "sc": 14, "l": 5, "f": 6098},
  {"h": "脔", "p": "luán", "sp": "luan", "r": "肉", "sc": 12, "l": 5, "f": 6099},
  {
    "h": "艟",
    "p": "chōng",
    "sp": "chong",
    "r": "舟",
    "sc": 18,
    "l": 5,
    "f": 6100
  },
  {"h": "選", "p": "xuǎn", "sp": "xuan", "r": "辵", "sc": 15, "l": 5, "f": 6101},
  {"h": "達", "p": "dá", "sp": "da", "r": "辵", "sc": 12, "l": 5, "f": 6102},
  {"h": "眬", "p": "lóng", "sp": "long", "r": "目", "sc": 10, "l": 5, "f": 6103},
  {"h": "彘", "p": "zhì", "sp": "zhi", "r": "彐", "sc": 12, "l": 5, "f": 6104},
  {"h": "檠", "p": "qíng", "sp": "qing", "r": "木", "sc": 16, "l": 5, "f": 6105},
  {"h": "胩", "p": "kǎ", "sp": "ka", "r": "肉", "sc": 9, "l": 5, "f": 6106},
  {"h": "旄", "p": "máo", "sp": "mao", "r": "方", "sc": 10, "l": 5, "f": 6107},
  {"h": "祆", "p": "xiān", "sp": "xian", "r": "示", "sc": 8, "l": 5, "f": 6108},
  {"h": "伛", "p": "yǔ", "sp": "yu", "r": "人", "sc": 6, "l": 5, "f": 6109},
  {"h": "鲥", "p": "shí", "sp": "shi", "r": "魚", "sc": 15, "l": 5, "f": 6110},
  {"h": "荦", "p": "luò", "sp": "luo", "r": "艸", "sc": 9, "l": 5, "f": 6111},
  {"h": "暾", "p": "tūn", "sp": "tun", "r": "日", "sc": 16, "l": 5, "f": 6112},
  {"h": "脈", "p": "mài", "sp": "mai", "r": "肉", "sc": 10, "l": 5, "f": 6113},
  {"h": "納", "p": "nà", "sp": "na", "r": "糸", "sc": 10, "l": 5, "f": 6114},
  {"h": "缌", "p": "sī", "sp": "si", "r": "糸", "sc": 12, "l": 5, "f": 6115},
  {"h": "飏", "p": "yáng", "sp": "yang", "r": "風", "sc": 7, "l": 5, "f": 6116},
  {"h": "锞", "p": "kè", "sp": "ke", "r": "金", "sc": 13, "l": 5, "f": 6117},
  {"h": "趵", "p": "bào", "sp": "bao", "r": "足", "sc": 10, "l": 5, "f": 6118},
  {"h": "跖", "p": "zhí", "sp": "zhi", "r": "足", "sc": 12, "l": 5, "f": 6119},
  {"h": "跬", "p": "kuǐ", "sp": "kui", "r": "足", "sc": 13, "l": 5, "f": 6120},
  {"h": "僬", "p": "jiāo", "sp": "jiao", "r": "人", "sc": 14, "l": 5, "f": 6121},
  {"h": "皲", "p": "jūn", "sp": "jun", "r": "皮", "sc": 11, "l": 5, "f": 6122},
  {"h": "翥", "p": "zhù", "sp": "zhu", "r": "羽", "sc": 14, "l": 5, "f": 6123},
  {"h": "祎", "p": "yī", "sp": "yi", "r": "示", "sc": 8, "l": 5, "f": 6124},
  {"h": "貢", "p": "gòng", "sp": "gong", "r": "貝", "sc": 10, "l": 5, "f": 6125},
  {"h": "習", "p": "xí", "sp": "xi", "r": "羽", "sc": 11, "l": 5, "f": 6126},
  {"h": "閉", "p": "bì", "sp": "bi", "r": "門", "sc": 11, "l": 5, "f": 6127},
  {"h": "芴", "p": "wù", "sp": "wu", "r": "艸", "sc": 7, "l": 5, "f": 6128},
  {"h": "菰", "p": "gū", "sp": "gu", "r": "艸", "sc": 11, "l": 5, "f": 6129},
  {"h": "眙", "p": "yí", "sp": "yi", "r": "目", "sc": 10, "l": 5, "f": 6130},
  {"h": "锓", "p": "qǐn", "sp": "qin", "r": "金", "sc": 12, "l": 5, "f": 6131},
  {"h": "訾", "p": "zī", "sp": "zi", "r": "言", "sc": 13, "l": 5, "f": 6132},
  {"h": "莳", "p": "shí", "sp": "shi", "r": "艸", "sc": 10, "l": 5, "f": 6133},
  {"h": "菪", "p": "dàng", "sp": "dang", "r": "艸", "sc": 11, "l": 5, "f": 6134},
  {"h": "槠", "p": "zhū", "sp": "zhu", "r": "木", "sc": 14, "l": 5, "f": 6135},
  {"h": "榍", "p": "xiè", "sp": "xie", "r": "木", "sc": 14, "l": 5, "f": 6136},
  {"h": "俜", "p": "pīng", "sp": "ping", "r": "人", "sc": 9, "l": 5, "f": 6137},
  {"h": "觚", "p": "gū", "sp": "gu", "r": "角", "sc": 12, "l": 5, "f": 6138},
  {"h": "撄", "p": "yīng", "sp": "ying", "r": "手", "sc": 14, "l": 5, "f": 6139},
  {"h": "嵫", "p": "zī", "sp": "zi", "r": "山", "sc": 12, "l": 5, "f": 6140},
  {"h": "桴", "p": "fú", "sp": "fu", "r": "木", "sc": 11, "l": 5, "f": 6141},
  {"h": "晡", "p": "bū", "sp": "bu", "r": "日", "sc": 11, "l": 5, "f": 6142},
  {"h": "爨", "p": "cuàn", "sp": "cuan", "r": "火", "sc": 30, "l": 5, "f": 6143},
  {"h": "悫", "p": "què", "sp": "que", "r": "心", "sc": 11, "l": 5, "f": 6144},
  {"h": "锇", "p": "é", "sp": "e", "r": "金", "sc": 12, "l": 5, "f": 6145},
  {
    "h": "锖",
    "p": "qiāng",
    "sp": "qiang",
    "r": "金",
    "sc": 13,
    "l": 5,
    "f": 6146
  },
  {"h": "锘", "p": "nuò", "sp": "nuo", "r": "金", "sc": 13, "l": 5, "f": 6147},
  {"h": "蝾", "p": "róng", "sp": "rong", "r": "虫", "sc": 15, "l": 5, "f": 6148},
  {"h": "糌", "p": "zān", "sp": "zan", "r": "米", "sc": 15, "l": 5, "f": 6149},
  {"h": "柙", "p": "xiá", "sp": "xia", "r": "木", "sc": 9, "l": 5, "f": 6150},
  {"h": "滹", "p": "hū", "sp": "hu", "r": "水", "sc": 14, "l": 5, "f": 6151},
  {"h": "睚", "p": "yá", "sp": "ya", "r": "目", "sc": 13, "l": 5, "f": 6152},
  {"h": "辦", "p": "bàn", "sp": "ban", "r": "辛", "sc": 16, "l": 5, "f": 6153},
  {"h": "莨", "p": "làng", "sp": "lang", "r": "艸", "sc": 10, "l": 5, "f": 6154},
  {"h": "尻", "p": "kāo", "sp": "kao", "r": "尸", "sc": 5, "l": 5, "f": 6155},
  {"h": "罨", "p": "yǎn", "sp": "yan", "r": "网", "sc": 13, "l": 5, "f": 6156},
  {"h": "锔", "p": "jū", "sp": "ju", "r": "金", "sc": 12, "l": 5, "f": 6157},
  {"h": "筲", "p": "shāo", "sp": "shao", "r": "竹", "sc": 13, "l": 5, "f": 6158},
  {"h": "妣", "p": "bǐ", "sp": "bi", "r": "女", "sc": 7, "l": 5, "f": 6159},
  {"h": "砘", "p": "dùn", "sp": "dun", "r": "石", "sc": 9, "l": 5, "f": 6160},
  {"h": "谳", "p": "yàn", "sp": "yan", "r": "言", "sc": 15, "l": 5, "f": 6161},
  {"h": "钽", "p": "tǎn", "sp": "tan", "r": "金", "sc": 10, "l": 5, "f": 6162},
  {"h": "鈤", "p": "rì", "sp": "ri", "r": "金", "sc": 12, "l": 5, "f": 6163},
  {
    "h": "脹",
    "p": "zhàng",
    "sp": "zhang",
    "r": "肉",
    "sc": 12,
    "l": 5,
    "f": 6164
  },
  {"h": "遺", "p": "yí", "sp": "yi", "r": "辵", "sc": 15, "l": 5, "f": 6165},
  {"h": "鄹", "p": "zōu", "sp": "zou", "r": "邑", "sc": 16, "l": 5, "f": 6166},
  {"h": "菡", "p": "hàn", "sp": "han", "r": "艸", "sc": 11, "l": 5, "f": 6167},
  {"h": "崦", "p": "yān", "sp": "yan", "r": "山", "sc": 11, "l": 5, "f": 6168},
  {"h": "狺", "p": "yín", "sp": "yin", "r": "犬", "sc": 10, "l": 5, "f": 6169},
  {"h": "杌", "p": "wù", "sp": "wu", "r": "木", "sc": 7, "l": 5, "f": 6170},
  {"h": "秭", "p": "zǐ", "sp": "zi", "r": "禾", "sc": 9, "l": 5, "f": 6171},
  {
    "h": "舯",
    "p": "zhōng",
    "sp": "zhong",
    "r": "舟",
    "sc": 10,
    "l": 5,
    "f": 6172
  },
  {
    "h": "豇",
    "p": "jiāng",
    "sp": "jiang",
    "r": "豆",
    "sc": 10,
    "l": 5,
    "f": 6173
  },
  {"h": "鲵", "p": "ní", "sp": "ni", "r": "魚", "sc": 16, "l": 5, "f": 6174},
  {"h": "蘧", "p": "qú", "sp": "qu", "r": "艸", "sc": 19, "l": 5, "f": 6175},
  {"h": "阚", "p": "hǎn", "sp": "han", "r": "門", "sc": 14, "l": 5, "f": 6176},
  {"h": "悝", "p": "kuī", "sp": "kui", "r": "心", "sc": 10, "l": 5, "f": 6177},
  {"h": "铌", "p": "ní", "sp": "ni", "r": "金", "sc": 10, "l": 5, "f": 6178},
  {"h": "預", "p": "yù", "sp": "yu", "r": "頁", "sc": 13, "l": 5, "f": 6179},
  {"h": "缑", "p": "gōu", "sp": "gou", "r": "糸", "sc": 12, "l": 5, "f": 6180},
  {"h": "瑗", "p": "yuàn", "sp": "yuan", "r": "玉", "sc": 13, "l": 5, "f": 6181},
  {"h": "椹", "p": "shèn", "sp": "shen", "r": "木", "sc": 13, "l": 5, "f": 6182},
  {"h": "旰", "p": "gàn", "sp": "gan", "r": "日", "sc": 7, "l": 5, "f": 6183},
  {"h": "魆", "p": "xū", "sp": "xu", "r": "鬼", "sc": 14, "l": 5, "f": 6184},
  {"h": "蓍", "p": "shī", "sp": "shi", "r": "艸", "sc": 13, "l": 5, "f": 6185},
  {"h": "蕹", "p": "wèng", "sp": "weng", "r": "艸", "sc": 16, "l": 5, "f": 6186},
  {"h": "岜", "p": "bā", "sp": "ba", "r": "山", "sc": 7, "l": 5, "f": 6187},
  {"h": "钐", "p": "shān", "sp": "shan", "r": "金", "sc": 8, "l": 5, "f": 6188},
  {"h": "锬", "p": "tán", "sp": "tan", "r": "金", "sc": 13, "l": 5, "f": 6189},
  {
    "h": "镪",
    "p": "qiāng",
    "sp": "qiang",
    "r": "金",
    "sc": 17,
    "l": 5,
    "f": 6190
  },
  {"h": "绡", "p": "xiāo", "sp": "xiao", "r": "糸", "sc": 10, "l": 5, "f": 6191},
  {"h": "檎", "p": "qín", "sp": "qin", "r": "木", "sc": 16, "l": 5, "f": 6192},
  {"h": "穀", "p": "gǔ", "sp": "gu", "r": "禾", "sc": 15, "l": 5, "f": 6193},
  {"h": "茕", "p": "qióng", "sp": "qiong", "r": "艸", "sc": 8, "l": 5, "f": 6194},
  {"h": "話", "p": "huà", "sp": "hua", "r": "言", "sc": 13, "l": 5, "f": 6195},
  {"h": "縮", "p": "suō", "sp": "suo", "r": "糸", "sc": 17, "l": 5, "f": 6196},
  {"h": "顧", "p": "gù", "sp": "gu", "r": "頁", "sc": 21, "l": 5, "f": 6197},
  {"h": "荈", "p": "chuǎn", "sp": "chuan", "r": "艸", "sc": 9, "l": 5, "f": 6198},
  {"h": "侉", "p": "kuǎ", "sp": "kua", "r": "人", "sc": 8, "l": 5, "f": 6199},
  {"h": "渫", "p": "xiè", "sp": "xie", "r": "水", "sc": 12, "l": 5, "f": 6200},
  {"h": "辂", "p": "lù", "sp": "lu", "r": "車", "sc": 10, "l": 5, "f": 6201},
  {"h": "欷", "p": "xī", "sp": "xi", "r": "欠", "sc": 11, "l": 5, "f": 6202},
  {"h": "轾", "p": "zhì", "sp": "zhi", "r": "車", "sc": 10, "l": 5, "f": 6203},
  {"h": "麴", "p": "qū", "sp": "qu", "r": "麥", "sc": 19, "l": 5, "f": 6204},
  {"h": "戢", "p": "jí", "sp": "ji", "r": "戈", "sc": 12, "l": 5, "f": 6205},
  {"h": "趼", "p": "jiǎn", "sp": "jian", "r": "足", "sc": 11, "l": 5, "f": 6206},
  {"h": "誤", "p": "wù", "sp": "wu", "r": "言", "sc": 14, "l": 5, "f": 6207},
  {"h": "芈", "p": "mǐ", "sp": "mi", "r": "羊", "sc": 7, "l": 5, "f": 6208},
  {"h": "苈", "p": "lì", "sp": "li", "r": "艸", "sc": 7, "l": 5, "f": 6209},
  {"h": "缍", "p": "duǒ", "sp": "duo", "r": "糸", "sc": 11, "l": 5, "f": 6210},
  {"h": "歃", "p": "shà", "sp": "sha", "r": "欠", "sc": 13, "l": 5, "f": 6211},
  {"h": "燠", "p": "yù", "sp": "yu", "r": "火", "sc": 16, "l": 5, "f": 6212},
  {"h": "耜", "p": "sì", "sp": "si", "r": "耒", "sc": 11, "l": 5, "f": 6213},
  {"h": "顸", "p": "hān", "sp": "han", "r": "干", "sc": 9, "l": 5, "f": 6214},
  {"h": "蚵", "p": "hé", "sp": "he", "r": "虫", "sc": 11, "l": 5, "f": 6215},
  {"h": "哓", "p": "xiāo", "sp": "xiao", "r": "口", "sc": 9, "l": 5, "f": 6216},
  {"h": "澌", "p": "sī", "sp": "si", "r": "水", "sc": 15, "l": 5, "f": 6217},
  {"h": "郜", "p": "gào", "sp": "gao", "r": "邑", "sc": 9, "l": 5, "f": 6218},
  {"h": "徼", "p": "jiǎo", "sp": "jiao", "r": "彳", "sc": 16, "l": 5, "f": 6219},
  {"h": "籀", "p": "zhòu", "sp": "zhou", "r": "竹", "sc": 19, "l": 5, "f": 6220},
  {"h": "陬", "p": "zōu", "sp": "zou", "r": "阜", "sc": 10, "l": 5, "f": 6221},
  {"h": "蕤", "p": "ruí", "sp": "rui", "r": "艸", "sc": 15, "l": 5, "f": 6222},
  {"h": "叻", "p": "lè", "sp": "le", "r": "口", "sc": 5, "l": 5, "f": 6223},
  {"h": "猞", "p": "shē", "sp": "she", "r": "犬", "sc": 11, "l": 5, "f": 6224},
  {"h": "洇", "p": "yīn", "sp": "yin", "r": "水", "sc": 9, "l": 5, "f": 6225},
  {"h": "锎", "p": "kāi", "sp": "kai", "r": "金", "sc": 12, "l": 5, "f": 6226},
  {"h": "舳", "p": "zhú", "sp": "zhu", "r": "舟", "sc": 11, "l": 5, "f": 6227},
  {"h": "缋", "p": "huì", "sp": "hui", "r": "糸", "sc": 12, "l": 5, "f": 6228},
  {"h": "镗", "p": "tāng", "sp": "tang", "r": "金", "sc": 16, "l": 5, "f": 6229},
  {"h": "褙", "p": "bèi", "sp": "bei", "r": "衣", "sc": 14, "l": 5, "f": 6230},
  {"h": "耋", "p": "dié", "sp": "die", "r": "老", "sc": 12, "l": 5, "f": 6231},
  {"h": "聬", "p": "wěng", "sp": "weng", "r": "耳", "sc": 16, "l": 5, "f": 6232},
  {"h": "鳡", "p": "gǎn", "sp": "gan", "r": "魚", "sc": 21, "l": 5, "f": 6233},
  {"h": "鲌", "p": "bà", "sp": "ba", "r": "魚", "sc": 13, "l": 5, "f": 6234},
  {"h": "鰽", "p": "qiú", "sp": "qiu", "r": "魚", "sc": 22, "l": 5, "f": 6235},
  {"h": "筆", "p": "bǐ", "sp": "bi", "r": "竹", "sc": 12, "l": 5, "f": 6236},
  {"h": "簡", "p": "jiǎn", "sp": "jian", "r": "竹", "sc": 18, "l": 5, "f": 6237},
  {"h": "計", "p": "jì", "sp": "ji", "r": "言", "sc": 9, "l": 5, "f": 6238},
  {"h": "語", "p": "yǔ", "sp": "yu", "r": "言", "sc": 14, "l": 5, "f": 6239},
  {"h": "謂", "p": "wèi", "sp": "wei", "r": "言", "sc": 16, "l": 5, "f": 6240},
  {"h": "袴", "p": "kù", "sp": "ku", "r": "衣", "sc": 11, "l": 5, "f": 6241},
  {"h": "巯", "p": "qiú", "sp": "qiu", "r": "工", "sc": 12, "l": 5, "f": 6242},
  {"h": "吖", "p": "yā", "sp": "ya", "r": "口", "sc": 6, "l": 5, "f": 6243},
  {"h": "猁", "p": "lì", "sp": "li", "r": "犬", "sc": 10, "l": 5, "f": 6244},
  {"h": "桷", "p": "jué", "sp": "jue", "r": "木", "sc": 11, "l": 5, "f": 6245},
  {"h": "砼", "p": "tóng", "sp": "tong", "r": "石", "sc": 10, "l": 5, "f": 6246},
  {"h": "锾", "p": "huán", "sp": "huan", "r": "金", "sc": 14, "l": 5, "f": 6247},
  {"h": "颟", "p": "mān", "sp": "man", "r": "頁", "sc": 16, "l": 5, "f": 6248},
  {"h": "跽", "p": "jì", "sp": "ji", "r": "足", "sc": 14, "l": 5, "f": 6249},
  {"h": "悛", "p": "quān", "sp": "quan", "r": "心", "sc": 10, "l": 5, "f": 6250},
  {"h": "桕", "p": "jiù", "sp": "jiu", "r": "木", "sc": 10, "l": 5, "f": 6251},
  {"h": "艨", "p": "méng", "sp": "meng", "r": "舟", "sc": 19, "l": 5, "f": 6252},
  {"h": "颡", "p": "sǎng", "sp": "sang", "r": "頁", "sc": 16, "l": 5, "f": 6253},
  {"h": "髹", "p": "xiū", "sp": "xiu", "r": "髟", "sc": 16, "l": 5, "f": 6254},
  {"h": "桫", "p": "suō", "sp": "suo", "r": "木", "sc": 11, "l": 5, "f": 6255},
  {"h": "鬆", "p": "sōng", "sp": "song", "r": "髟", "sc": 18, "l": 5, "f": 6256},
  {"h": "練", "p": "liàn", "sp": "lian", "r": "糸", "sc": 15, "l": 5, "f": 6257},
  {"h": "總", "p": "zǒng", "sp": "zong", "r": "糸", "sc": 17, "l": 5, "f": 6258},
  {"h": "膽", "p": "dǎn", "sp": "dan", "r": "肉", "sc": 17, "l": 5, "f": 6259},
  {"h": "貝", "p": "bèi", "sp": "bei", "r": "貝", "sc": 7, "l": 5, "f": 6260},
  {"h": "葸", "p": "xǐ", "sp": "xi", "r": "艸", "sc": 12, "l": 5, "f": 6261},
  {"h": "屣", "p": "xǐ", "sp": "xi", "r": "尸", "sc": 14, "l": 5, "f": 6262},
  {"h": "獒", "p": "áo", "sp": "ao", "r": "犬", "sc": 14, "l": 5, "f": 6263},
  {"h": "瘌", "p": "là", "sp": "la", "r": "疒", "sc": 14, "l": 5, "f": 6264},
  {"h": "佴", "p": "èr", "sp": "er", "r": "人", "sc": 8, "l": 5, "f": 6265},
  {"h": "锫", "p": "péi", "sp": "pei", "r": "金", "sc": 13, "l": 5, "f": 6266},
  {"h": "隰", "p": "xí", "sp": "xi", "r": "阜", "sc": 16, "l": 5, "f": 6267},
  {"h": "醌", "p": "kūn", "sp": "kun", "r": "酉", "sc": 15, "l": 5, "f": 6268},
  {"h": "萬", "p": "wàn", "sp": "wan", "r": "艸", "sc": 12, "l": 5, "f": 6269},
  {"h": "矯", "p": "jiǎo", "sp": "jiao", "r": "矢", "sc": 17, "l": 5, "f": 6270},
  {"h": "鰕", "p": "xiā", "sp": "xia", "r": "魚", "sc": 20, "l": 5, "f": 6271},
  {"h": "岽", "p": "dōng", "sp": "dong", "r": "山", "sc": 8, "l": 5, "f": 6272},
  {"h": "昃", "p": "zè", "sp": "ze", "r": "日", "sc": 8, "l": 5, "f": 6273},
  {"h": "螓", "p": "qín", "sp": "qin", "r": "虫", "sc": 16, "l": 5, "f": 6274},
  {"h": "垧", "p": "shǎng", "sp": "shang", "r": "土", "sc": 9, "l": 5, "f": 6275},
  {"h": "廨", "p": "xiè", "sp": "xie", "r": "广", "sc": 16, "l": 5, "f": 6276},
  {"h": "骺", "p": "hóu", "sp": "hou", "r": "骨", "sc": 15, "l": 5, "f": 6277},
  {"h": "黥", "p": "qíng", "sp": "qing", "r": "黑", "sc": 20, "l": 5, "f": 6278},
  {"h": "豳", "p": "bīn", "sp": "bin", "r": "豕", "sc": 17, "l": 5, "f": 6279},
  {"h": "議", "p": "yì", "sp": "yi", "r": "言", "sc": 20, "l": 5, "f": 6280},
  {"h": "絡", "p": "luò", "sp": "luo", "r": "糸", "sc": 12, "l": 5, "f": 6281},
  {"h": "護", "p": "hù", "sp": "hu", "r": "言", "sc": 20, "l": 5, "f": 6282},
  {"h": "釋", "p": "shì", "sp": "shi", "r": "釆", "sc": 20, "l": 5, "f": 6283},
  {"h": "隨", "p": "suí", "sp": "sui", "r": "阜", "sc": 14, "l": 5, "f": 6284},
  {"h": "荑", "p": "tí", "sp": "ti", "r": "艸", "sc": 9, "l": 5, "f": 6285},
  {"h": "莶", "p": "xiān", "sp": "xian", "r": "艸", "sc": 10, "l": 5, "f": 6286},
  {"h": "蓊", "p": "wěng", "sp": "weng", "r": "艸", "sc": 13, "l": 5, "f": 6287},
  {"h": "揎", "p": "xuān", "sp": "xuan", "r": "手", "sc": 12, "l": 5, "f": 6288},
  {"h": "娈", "p": "luán", "sp": "luan", "r": "女", "sc": 9, "l": 5, "f": 6289},
  {
    "h": "嫜",
    "p": "zhāng",
    "sp": "zhang",
    "r": "女",
    "sc": 14,
    "l": 5,
    "f": 6290
  },
  {"h": "毽", "p": "jiàn", "sp": "jian", "r": "毛", "sc": 12, "l": 5, "f": 6291},
  {"h": "砑", "p": "yà", "sp": "ya", "r": "石", "sc": 9, "l": 5, "f": 6292},
  {"h": "碶", "p": "qì", "sp": "qi", "r": "石", "sc": 14, "l": 5, "f": 6293},
  {"h": "剀", "p": "kǎi", "sp": "kai", "r": "刀", "sc": 8, "l": 5, "f": 6294},
  {"h": "薷", "p": "rú", "sp": "ru", "r": "艸", "sc": 17, "l": 5, "f": 6295},
  {"h": "怃", "p": "wǔ", "sp": "wu", "r": "心", "sc": 7, "l": 5, "f": 6296},
  {"h": "怿", "p": "yì", "sp": "yi", "r": "心", "sc": 8, "l": 5, "f": 6297},
  {"h": "骐", "p": "qí", "sp": "qi", "r": "馬", "sc": 11, "l": 5, "f": 6298},
  {"h": "腧", "p": "shù", "sp": "shu", "r": "肉", "sc": 13, "l": 5, "f": 6299},
  {"h": "霈", "p": "pèi", "sp": "pei", "r": "雨", "sc": 15, "l": 5, "f": 6300},
  {"h": "蝻", "p": "nǎn", "sp": "nan", "r": "虫", "sc": 15, "l": 5, "f": 6301},
  {"h": "鵰", "p": "diāo", "sp": "diao", "r": "鳥", "sc": 19, "l": 5, "f": 6302},
  {"h": "讀", "p": "dú", "sp": "du", "r": "言", "sc": 22, "l": 5, "f": 6303},
  {"h": "墼", "p": "jī", "sp": "ji", "r": "土", "sc": 16, "l": 5, "f": 6304},
  {"h": "坩", "p": "gān", "sp": "gan", "r": "土", "sc": 8, "l": 5, "f": 6305},
  {"h": "甏", "p": "bèng", "sp": "beng", "r": "瓦", "sc": 16, "l": 5, "f": 6306},
  {"h": "鹇", "p": "xián", "sp": "xian", "r": "鳥", "sc": 12, "l": 5, "f": 6307},
  {"h": "鹣", "p": "jiān", "sp": "jian", "r": "鳥", "sc": 15, "l": 5, "f": 6308},
  {"h": "舴", "p": "zé", "sp": "ze", "r": "舟", "sc": 11, "l": 5, "f": 6309},
  {"h": "鲭", "p": "qīng", "sp": "qing", "r": "魚", "sc": 16, "l": 5, "f": 6310},
  {"h": "傺", "p": "chì", "sp": "chi", "r": "人", "sc": 13, "l": 5, "f": 6311},
  {"h": "枘", "p": "ruì", "sp": "rui", "r": "木", "sc": 8, "l": 5, "f": 6312},
  {"h": "秕", "p": "bǐ", "sp": "bi", "r": "禾", "sc": 9, "l": 5, "f": 6313},
  {
    "h": "糨",
    "p": "jiàng",
    "sp": "jiang",
    "r": "米",
    "sc": 18,
    "l": 5,
    "f": 6314
  },
  {"h": "忉", "p": "dāo", "sp": "dao", "r": "心", "sc": 5, "l": 5, "f": 6315},
  {"h": "磲", "p": "qú", "sp": "qu", "r": "石", "sc": 16, "l": 5, "f": 6316},
  {"h": "顒", "p": "yóng", "sp": "yong", "r": "頁", "sc": 18, "l": 5, "f": 6317},
  {
    "h": "蟲",
    "p": "chóng",
    "sp": "chong",
    "r": "虫",
    "sc": 18,
    "l": 5,
    "f": 6318
  },
  {"h": "訓", "p": "xùn", "sp": "xun", "r": "言", "sc": 10, "l": 5, "f": 6319},
  {"h": "侑", "p": "yòu", "sp": "you", "r": "人", "sc": 8, "l": 5, "f": 6320},
  {"h": "萁", "p": "qí", "sp": "qi", "r": "艸", "sc": 11, "l": 5, "f": 6321},
  {"h": "葳", "p": "wēi", "sp": "wei", "r": "艸", "sc": 12, "l": 5, "f": 6322},
  {"h": "呙", "p": "guō", "sp": "guo", "r": "口", "sc": 7, "l": 5, "f": 6323},
  {"h": "鹋", "p": "miáo", "sp": "miao", "r": "鳥", "sc": 13, "l": 5, "f": 6324},
  {"h": "貘", "p": "mò", "sp": "mo", "r": "豸", "sc": 17, "l": 5, "f": 6325},
  {"h": "葶", "p": "tíng", "sp": "ting", "r": "艸", "sc": 12, "l": 5, "f": 6326},
  {"h": "枨", "p": "chéng", "sp": "cheng", "r": "木", "sc": 8, "l": 5, "f": 6327},
  {"h": "棹", "p": "zhào", "sp": "zhao", "r": "木", "sc": 12, "l": 5, "f": 6328},
  {"h": "纛", "p": "dào", "sp": "dao", "r": "糸", "sc": 25, "l": 5, "f": 6329},
  {"h": "籤", "p": "qiān", "sp": "qian", "r": "竹", "sc": 23, "l": 5, "f": 6330},
  {"h": "規", "p": "guī", "sp": "gui", "r": "見", "sc": 11, "l": 5, "f": 6331},
  {"h": "資", "p": "zī", "sp": "zi", "r": "貝", "sc": 13, "l": 5, "f": 6332},
  {"h": "嶷", "p": "yí", "sp": "yi", "r": "山", "sc": 17, "l": 5, "f": 6333},
  {"h": "妁", "p": "shuò", "sp": "shuo", "r": "女", "sc": 6, "l": 5, "f": 6334},
  {"h": "杩", "p": "mà", "sp": "ma", "r": "木", "sc": 7, "l": 5, "f": 6335},
  {"h": "鹛", "p": "méi", "sp": "mei", "r": "鳥", "sc": 14, "l": 5, "f": 6336},
  {"h": "羝", "p": "dī", "sp": "di", "r": "羊", "sc": 11, "l": 5, "f": 6337},
  {"h": "骖", "p": "cān", "sp": "can", "r": "馬", "sc": 11, "l": 5, "f": 6338},
  {"h": "祧", "p": "tiāo", "sp": "tiao", "r": "示", "sc": 10, "l": 5, "f": 6339},
  {"h": "給", "p": "gěi", "sp": "gei", "r": "糸", "sc": 12, "l": 5, "f": 6340},
  {"h": "淠", "p": "pì", "sp": "pi", "r": "水", "sc": 11, "l": 5, "f": 6341},
  {"h": "畎", "p": "quǎn", "sp": "quan", "r": "田", "sc": 9, "l": 5, "f": 6342},
  {"h": "箨", "p": "tuò", "sp": "tuo", "r": "竹", "sc": 14, "l": 5, "f": 6343},
  {"h": "鳐", "p": "yáo", "sp": "yao", "r": "魚", "sc": 18, "l": 5, "f": 6344},
  {"h": "識", "p": "shí", "sp": "shi", "r": "言", "sc": 19, "l": 5, "f": 6345},
  {"h": "遠", "p": "yuǎn", "sp": "yuan", "r": "辵", "sc": 13, "l": 5, "f": 6346},
  {"h": "谿", "p": "xī", "sp": "xi", "r": "谷", "sc": 17, "l": 5, "f": 6347},
  {"h": "剡", "p": "shàn", "sp": "shan", "r": "刀", "sc": 10, "l": 5, "f": 6348},
  {"h": "吲", "p": "yǐn", "sp": "yin", "r": "口", "sc": 7, "l": 5, "f": 6349},
  {"h": "嫫", "p": "mó", "sp": "mo", "r": "女", "sc": 13, "l": 5, "f": 6350},
  {"h": "戽", "p": "hù", "sp": "hu", "r": "戶", "sc": 8, "l": 5, "f": 6351},
  {"h": "糗", "p": "qiǔ", "sp": "qiu", "r": "米", "sc": 16, "l": 5, "f": 6352},
  {"h": "鸮", "p": "xiāo", "sp": "xiao", "r": "鳥", "sc": 10, "l": 5, "f": 6353},
  {"h": "蕰", "p": "wēn", "sp": "wen", "r": "艸", "sc": 15, "l": 5, "f": 6354},
  {"h": "坼", "p": "chè", "sp": "che", "r": "土", "sc": 8, "l": 5, "f": 6355},
  {"h": "璺", "p": "wèn", "sp": "wen", "r": "玉", "sc": 20, "l": 5, "f": 6356},
  {"h": "栊", "p": "lóng", "sp": "long", "r": "木", "sc": 9, "l": 5, "f": 6357},
  {"h": "镝", "p": "dī", "sp": "di", "r": "金", "sc": 16, "l": 5, "f": 6358},
  {"h": "疔", "p": "dīng", "sp": "ding", "r": "疒", "sc": 7, "l": 5, "f": 6359},
  {"h": "瘰", "p": "luǒ", "sp": "luo", "r": "疒", "sc": 16, "l": 5, "f": 6360},
  {"h": "谮", "p": "zèn", "sp": "zen", "r": "言", "sc": 14, "l": 5, "f": 6361},
  {"h": "辋", "p": "wǎng", "sp": "wang", "r": "車", "sc": 12, "l": 5, "f": 6362},
  {"h": "碥", "p": "biǎn", "sp": "bian", "r": "石", "sc": 14, "l": 5, "f": 6363},
  {"h": "豂", "p": "liáo", "sp": "liao", "r": "谷", "sc": 18, "l": 5, "f": 6364},
  {"h": "竅", "p": "qiào", "sp": "qiao", "r": "穴", "sc": 18, "l": 5, "f": 6365},
  {"h": "讓", "p": "ràng", "sp": "rang", "r": "言", "sc": 24, "l": 5, "f": 6366},
  {"h": "黒", "p": "hēi", "sp": "hei", "r": "黑", "sc": 11, "l": 5, "f": 6367},
  {"h": "劂", "p": "jué", "sp": "jue", "r": "刀", "sc": 14, "l": 5, "f": 6368},
  {"h": "潋", "p": "liàn", "sp": "lian", "r": "水", "sc": 14, "l": 5, "f": 6369},
  {"h": "孓", "p": "jué", "sp": "jue", "r": "子", "sc": 3, "l": 5, "f": 6370},
  {"h": "螋", "p": "sōu", "sp": "sou", "r": "虫", "sc": 15, "l": 5, "f": 6371},
  {
    "h": "鲳",
    "p": "chāng",
    "sp": "chang",
    "r": "魚",
    "sc": 16,
    "l": 5,
    "f": 6372
  },
  {"h": "铙", "p": "náo", "sp": "nao", "r": "金", "sc": 11, "l": 5, "f": 6373},
  {"h": "塄", "p": "léng", "sp": "leng", "r": "土", "sc": 12, "l": 5, "f": 6374},
  {"h": "萆", "p": "bì", "sp": "bi", "r": "艸", "sc": 11, "l": 5, "f": 6375},
  {"h": "碲", "p": "dì", "sp": "di", "r": "石", "sc": 14, "l": 5, "f": 6376},
  {"h": "粋", "p": "cuì", "sp": "cui", "r": "米", "sc": 10, "l": 5, "f": 6377},
  {"h": "纻", "p": "zhù", "sp": "zhu", "r": "糸", "sc": 7, "l": 5, "f": 6378},
  {"h": "沭", "p": "shù", "sp": "shu", "r": "水", "sc": 8, "l": 5, "f": 6379},
  {"h": "祏", "p": "shí", "sp": "shi", "r": "示", "sc": 9, "l": 5, "f": 6380},
  {"h": "電", "p": "diàn", "sp": "dian", "r": "雨", "sc": 13, "l": 5, "f": 6381},
  {"h": "鱲", "p": "liè", "sp": "lie", "r": "魚", "sc": 26, "l": 5, "f": 6382},
  {"h": "頻", "p": "pín", "sp": "pin", "r": "頁", "sc": 16, "l": 5, "f": 6383},
  {"h": "設", "p": "shè", "sp": "she", "r": "言", "sc": 11, "l": 5, "f": 6384},
  {"h": "試", "p": "shì", "sp": "shi", "r": "言", "sc": 13, "l": 5, "f": 6385},
  {
    "h": "項",
    "p": "xiàng",
    "sp": "xiang",
    "r": "頁",
    "sc": 12,
    "l": 5,
    "f": 6386
  },
  {"h": "鄩", "p": "xún", "sp": "xun", "r": "邑", "sc": 14, "l": 5, "f": 6387},
  {"h": "鱖", "p": "guì", "sp": "gui", "r": "魚", "sc": 23, "l": 5, "f": 6388},
  {"h": "鱧", "p": "lǐ", "sp": "li", "r": "魚", "sc": 24, "l": 5, "f": 6389},
  {"h": "怍", "p": "zuò", "sp": "zuo", "r": "心", "sc": 8, "l": 5, "f": 6390},
  {"h": "糍", "p": "cí", "sp": "ci", "r": "米", "sc": 15, "l": 5, "f": 6391},
  {"h": "鸻", "p": "héng", "sp": "heng", "r": "鳥", "sc": 11, "l": 5, "f": 6392},
  {"h": "仵", "p": "wǔ", "sp": "wu", "r": "人", "sc": 6, "l": 5, "f": 6393},
  {"h": "柰", "p": "nài", "sp": "nai", "r": "木", "sc": 9, "l": 5, "f": 6394},
  {"h": "棼", "p": "fén", "sp": "fen", "r": "木", "sc": 12, "l": 5, "f": 6395},
  {"h": "氇", "p": "lu", "sp": "lu", "r": "毛", "sc": 16, "l": 5, "f": 6396},
  {"h": "氆", "p": "pǔ", "sp": "pu", "r": "毛", "sc": 16, "l": 5, "f": 6397},
  {"h": "烀", "p": "hū", "sp": "hu", "r": "火", "sc": 9, "l": 5, "f": 6398},
  {"h": "镔", "p": "bīn", "sp": "bin", "r": "金", "sc": 15, "l": 5, "f": 6399},
  {"h": "珉", "p": "mín", "sp": "min", "r": "玉", "sc": 9, "l": 5, "f": 6400},
  {"h": "鞬", "p": "jiān", "sp": "jian", "r": "革", "sc": 17, "l": 5, "f": 6401},
  {"h": "範", "p": "fàn", "sp": "fan", "r": "竹", "sc": 15, "l": 5, "f": 6402},
  {"h": "膠", "p": "jiāo", "sp": "jiao", "r": "肉", "sc": 15, "l": 5, "f": 6403},
  {
    "h": "眾",
    "p": "zhòng",
    "sp": "zhong",
    "r": "目",
    "sc": 11,
    "l": 5,
    "f": 6404
  },
  {"h": "鶇", "p": "dōng", "sp": "dong", "r": "鳥", "sc": 19, "l": 5, "f": 6405},
  {"h": "囝", "p": "jiǎn", "sp": "jian", "r": "囗", "sc": 6, "l": 5, "f": 6406},
  {"h": "犴", "p": "àn", "sp": "an", "r": "犬", "sc": 6, "l": 5, "f": 6407},
  {"h": "镠", "p": "liú", "sp": "liu", "r": "金", "sc": 16, "l": 5, "f": 6408},
  {"h": "觇", "p": "chān", "sp": "chan", "r": "見", "sc": 9, "l": 5, "f": 6409},
  {"h": "钆", "p": "gá", "sp": "ga", "r": "金", "sc": 6, "l": 5, "f": 6410},
  {"h": "蚺", "p": "rán", "sp": "ran", "r": "虫", "sc": 11, "l": 5, "f": 6411},
  {"h": "蝓", "p": "yú", "sp": "yu", "r": "虫", "sc": 15, "l": 5, "f": 6412},
  {"h": "趑", "p": "zī", "sp": "zi", "r": "走", "sc": 13, "l": 5, "f": 6413},
  {"h": "貔", "p": "pí", "sp": "pi", "r": "豸", "sc": 17, "l": 5, "f": 6414},
  {"h": "躞", "p": "xiè", "sp": "xie", "r": "足", "sc": 24, "l": 5, "f": 6415},
  {"h": "氽", "p": "tǔn", "sp": "tun", "r": "人", "sc": 6, "l": 5, "f": 6416},
  {"h": "绂", "p": "fú", "sp": "fu", "r": "糸", "sc": 8, "l": 5, "f": 6417},
  {"h": "栲", "p": "kǎo", "sp": "kao", "r": "木", "sc": 10, "l": 5, "f": 6418},
  {"h": "鹎", "p": "bēi", "sp": "bei", "r": "鳥", "sc": 13, "l": 5, "f": 6419},
  {"h": "襻", "p": "pàn", "sp": "pan", "r": "衣", "sc": 24, "l": 5, "f": 6420},
  {"h": "蓚", "p": "tiáo", "sp": "tiao", "r": "艸", "sc": 12, "l": 5, "f": 6421},
  {"h": "貪", "p": "tān", "sp": "tan", "r": "貝", "sc": 11, "l": 5, "f": 6422},
  {"h": "祓", "p": "fú", "sp": "fu", "r": "示", "sc": 9, "l": 5, "f": 6423},
  {"h": "繼", "p": "jì", "sp": "ji", "r": "糸", "sc": 20, "l": 5, "f": 6424},
  {"h": "脫", "p": "tuō", "sp": "tuo", "r": "肉", "sc": 11, "l": 5, "f": 6425},
  {"h": "郿", "p": "méi", "sp": "mei", "r": "邑", "sc": 11, "l": 5, "f": 6426},
  {"h": "骝", "p": "liú", "sp": "liu", "r": "馬", "sc": 13, "l": 5, "f": 6427},
  {"h": "檫", "p": "chá", "sp": "cha", "r": "木", "sc": 18, "l": 5, "f": 6428},
  {"h": "砟", "p": "zhǎ", "sp": "zha", "r": "石", "sc": 10, "l": 5, "f": 6429},
  {"h": "蝰", "p": "kuí", "sp": "kui", "r": "虫", "sc": 15, "l": 5, "f": 6430},
  {"h": "猱", "p": "náo", "sp": "nao", "r": "犬", "sc": 12, "l": 5, "f": 6431},
  {"h": "缵", "p": "zuǎn", "sp": "zuan", "r": "糸", "sc": 19, "l": 5, "f": 6432},
  {"h": "沔", "p": "miǎn", "sp": "mian", "r": "水", "sc": 7, "l": 5, "f": 6433},
  {"h": "彖", "p": "tuàn", "sp": "tuan", "r": "彐", "sc": 9, "l": 5, "f": 6434},
  {"h": "黻", "p": "fú", "sp": "fu", "r": "黹", "sc": 17, "l": 5, "f": 6435},
  {"h": "眢", "p": "yuān", "sp": "yuan", "r": "目", "sc": 10, "l": 5, "f": 6436},
  {"h": "車", "p": "chē", "sp": "che", "r": "車", "sc": 7, "l": 5, "f": 6437},
  {"h": "陳", "p": "chén", "sp": "chen", "r": "阜", "sc": 10, "l": 5, "f": 6438},
  {"h": "雜", "p": "zá", "sp": "za", "r": "隹", "sc": 18, "l": 5, "f": 6439},
  {"h": "驚", "p": "jīng", "sp": "jing", "r": "馬", "sc": 22, "l": 5, "f": 6440},
  {"h": "龍", "p": "lóng", "sp": "long", "r": "龍", "sc": 16, "l": 5, "f": 6441},
  {"h": "鮠", "p": "wéi", "sp": "wei", "r": "魚", "sc": 17, "l": 5, "f": 6442},
  {"h": "茛", "p": "gèn", "sp": "gen", "r": "艸", "sc": 9, "l": 5, "f": 6443},
  {"h": "蕈", "p": "xùn", "sp": "xun", "r": "艸", "sc": 15, "l": 5, "f": 6444},
  {"h": "骀", "p": "dài", "sp": "dai", "r": "馬", "sc": 8, "l": 5, "f": 6445},
  {"h": "樨", "p": "xī", "sp": "xi", "r": "木", "sc": 16, "l": 5, "f": 6446},
  {"h": "昝", "p": "zǎn", "sp": "zan", "r": "日", "sc": 9, "l": 5, "f": 6447},
  {"h": "餂", "p": "tiǎn", "sp": "tian", "r": "食", "sc": 14, "l": 5, "f": 6448},
  {"h": "镎", "p": "ná", "sp": "na", "r": "金", "sc": 15, "l": 5, "f": 6449},
  {"h": "酡", "p": "tuó", "sp": "tuo", "r": "酉", "sc": 12, "l": 5, "f": 6450},
  {"h": "鳜", "p": "guì", "sp": "gui", "r": "魚", "sc": 20, "l": 5, "f": 6451},
  {"h": "觋", "p": "xí", "sp": "xi", "r": "見", "sc": 11, "l": 5, "f": 6452},
  {"h": "礅", "p": "dūn", "sp": "dun", "r": "石", "sc": 17, "l": 5, "f": 6453},
  {"h": "垆", "p": "lú", "sp": "lu", "r": "土", "sc": 8, "l": 5, "f": 6454},
  {"h": "坻", "p": "chí", "sp": "chi", "r": "土", "sc": 8, "l": 5, "f": 6455},
  {"h": "荇", "p": "xìng", "sp": "xing", "r": "艸", "sc": 9, "l": 5, "f": 6456},
  {"h": "扃", "p": "jiōng", "sp": "jiong", "r": "戶", "sc": 9, "l": 5, "f": 6457},
  {"h": "磻", "p": "pán", "sp": "pan", "r": "石", "sc": 17, "l": 5, "f": 6458},
  {"h": "郗", "p": "xī", "sp": "xi", "r": "邑", "sc": 9, "l": 5, "f": 6459},
  {"h": "阇", "p": "dū", "sp": "du", "r": "門", "sc": 11, "l": 5, "f": 6460},
  {"h": "腈", "p": "jīng", "sp": "jing", "r": "肉", "sc": 12, "l": 5, "f": 6461},
  {"h": "铈", "p": "shì", "sp": "shi", "r": "金", "sc": 10, "l": 5, "f": 6462},
  {"h": "臥", "p": "wò", "sp": "wo", "r": "臣", "sc": 8, "l": 5, "f": 6463},
  {"h": "顱", "p": "lú", "sp": "lu", "r": "頁", "sc": 25, "l": 5, "f": 6464},
  {"h": "級", "p": "jí", "sp": "ji", "r": "糸", "sc": 9, "l": 5, "f": 6465},
  {"h": "階", "p": "jiē", "sp": "jie", "r": "阜", "sc": 11, "l": 5, "f": 6466},
  {"h": "顟", "p": "láo", "sp": "lao", "r": "頁", "sc": 20, "l": 5, "f": 6467},
  {
    "h": "衖",
    "p": "xiàng",
    "sp": "xiang",
    "r": "行",
    "sc": 12,
    "l": 5,
    "f": 6468
  },
  {"h": "珙", "p": "gǒng", "sp": "gong", "r": "玉", "sc": 10, "l": 5, "f": 6469},
  {"h": "襞", "p": "bì", "sp": "bi", "r": "衣", "sc": 19, "l": 5, "f": 6470},
  {"h": "劁", "p": "qiāo", "sp": "qiao", "r": "刀", "sc": 14, "l": 5, "f": 6471},
  {"h": "隈", "p": "wēi", "sp": "wei", "r": "阜", "sc": 11, "l": 5, "f": 6472},
  {"h": "胬", "p": "nǔ", "sp": "nu", "r": "肉", "sc": 11, "l": 5, "f": 6473},
  {"h": "纮", "p": "hóng", "sp": "hong", "r": "糸", "sc": 7, "l": 5, "f": 6474},
  {"h": "翙", "p": "huì", "sp": "hui", "r": "羽", "sc": 12, "l": 5, "f": 6475},
  {"h": "茀", "p": "fú", "sp": "fu", "r": "艸", "sc": 8, "l": 5, "f": 6476},
  {"h": "荜", "p": "bì", "sp": "bi", "r": "艸", "sc": 9, "l": 5, "f": 6477},
  {"h": "溱", "p": "qín", "sp": "qin", "r": "水", "sc": 13, "l": 5, "f": 6478},
  {"h": "酤", "p": "gū", "sp": "gu", "r": "酉", "sc": 12, "l": 5, "f": 6479},
  {"h": "腳", "p": "jiǎo", "sp": "jiao", "r": "肉", "sc": 13, "l": 5, "f": 6480},
  {"h": "記", "p": "jì", "sp": "ji", "r": "言", "sc": 10, "l": 5, "f": 6481},
  {"h": "許", "p": "xǔ", "sp": "xu", "r": "言", "sc": 11, "l": 5, "f": 6482},
  {"h": "錯", "p": "cuò", "sp": "cuo", "r": "金", "sc": 16, "l": 5, "f": 6483},
  {"h": "鱇", "p": "kāng", "sp": "kang", "r": "魚", "sc": 22, "l": 5, "f": 6484},
  {"h": "圯", "p": "yí", "sp": "yi", "r": "土", "sc": 6, "l": 5, "f": 6485},
  {"h": "柽", "p": "chēng", "sp": "cheng", "r": "木", "sc": 9, "l": 5, "f": 6486},
  {"h": "瞀", "p": "mào", "sp": "mao", "r": "目", "sc": 14, "l": 5, "f": 6487},
  {
    "h": "铖",
    "p": "chéng",
    "sp": "cheng",
    "r": "金",
    "sc": 11,
    "l": 5,
    "f": 6488
  },
  {"h": "粞", "p": "xī", "sp": "xi", "r": "米", "sc": 12, "l": 5, "f": 6489},
  {"h": "撙", "p": "zǔn", "sp": "zun", "r": "手", "sc": 15, "l": 5, "f": 6490},
  {"h": "铊", "p": "tā", "sp": "ta", "r": "金", "sc": 10, "l": 5, "f": 6491},
  {"h": "圹", "p": "kuàng", "sp": "kuang", "r": "土", "sc": 6, "l": 5, "f": 6492},
  {"h": "菝", "p": "bá", "sp": "ba", "r": "艸", "sc": 11, "l": 5, "f": 6493},
  {"h": "迴", "p": "huí", "sp": "hui", "r": "辵", "sc": 9, "l": 5, "f": 6494},
  {"h": "岙", "p": "ào", "sp": "ao", "r": "山", "sc": 7, "l": 5, "f": 6495},
  {"h": "漶", "p": "huàn", "sp": "huan", "r": "水", "sc": 14, "l": 5, "f": 6496},
  {"h": "缏", "p": "biàn", "sp": "bian", "r": "糸", "sc": 12, "l": 5, "f": 6497},
  {"h": "貊", "p": "mò", "sp": "mo", "r": "豸", "sc": 13, "l": 5, "f": 6498},
  {"h": "粫", "p": "ér", "sp": "er", "r": "米", "sc": 12, "l": 5, "f": 6499},
  {"h": "徂", "p": "cú", "sp": "cu", "r": "彳", "sc": 8, "l": 5, "f": 6500},
  {"h": "洎", "p": "jì", "sp": "ji", "r": "水", "sc": 9, "l": 5, "f": 6501},
  {"h": "礌", "p": "léi", "sp": "lei", "r": "石", "sc": 18, "l": 5, "f": 6502},
  {"h": "膿", "p": "nóng", "sp": "nong", "r": "肉", "sc": 17, "l": 5, "f": 6503},
  {"h": "葉", "p": "yè", "sp": "ye", "r": "艸", "sc": 12, "l": 5, "f": 6504},
  {"h": "險", "p": "xiǎn", "sp": "xian", "r": "阜", "sc": 15, "l": 5, "f": 6505},
  {
    "h": "雙",
    "p": "shuāng",
    "sp": "shuang",
    "r": "隹",
    "sc": 18,
    "l": 5,
    "f": 6506
  },
  {"h": "餘", "p": "yú", "sp": "yu", "r": "食", "sc": 15, "l": 5, "f": 6507},
  {"h": "邠", "p": "bīn", "sp": "bin", "r": "邑", "sc": 6, "l": 5, "f": 6508},
  {"h": "邽", "p": "guī", "sp": "gui", "r": "邑", "sc": 8, "l": 5, "f": 6509},
  {"h": "溽", "p": "rù", "sp": "ru", "r": "水", "sc": 13, "l": 5, "f": 6510},
  {"h": "棰", "p": "chuí", "sp": "chui", "r": "木", "sc": 12, "l": 5, "f": 6511},
  {"h": "舭", "p": "bǐ", "sp": "bi", "r": "舟", "sc": 10, "l": 5, "f": 6512},
  {"h": "艋", "p": "měng", "sp": "meng", "r": "舟", "sc": 14, "l": 5, "f": 6513},
  {"h": "靜", "p": "jìng", "sp": "jing", "r": "青", "sc": 16, "l": 5, "f": 6514},
  {"h": "聱", "p": "áo", "sp": "ao", "r": "耳", "sc": 16, "l": 5, "f": 6515},
  {"h": "尢", "p": "yóu", "sp": "you", "r": "尢", "sc": 3, "l": 5, "f": 6516},
  {"h": "藚", "p": "xù", "sp": "xu", "r": "艸", "sc": 18, "l": 5, "f": 6517},
  {"h": "佾", "p": "yì", "sp": "yi", "r": "人", "sc": 8, "l": 5, "f": 6518},
  {"h": "蕻", "p": "hóng", "sp": "hong", "r": "艸", "sc": 16, "l": 5, "f": 6519},
  {"h": "狁", "p": "yǔn", "sp": "yun", "r": "犬", "sc": 7, "l": 5, "f": 6520},
  {"h": "绁", "p": "xiè", "sp": "xie", "r": "糸", "sc": 8, "l": 5, "f": 6521},
  {"h": "牾", "p": "wǔ", "sp": "wu", "r": "牛", "sc": 11, "l": 5, "f": 6522},
  {"h": "慝", "p": "tè", "sp": "te", "r": "心", "sc": 14, "l": 5, "f": 6523},
  {"h": "蜱", "p": "pí", "sp": "pi", "r": "虫", "sc": 14, "l": 5, "f": 6524},
  {"h": "鬘", "p": "mán", "sp": "man", "r": "髟", "sc": 21, "l": 5, "f": 6525},
  {"h": "膚", "p": "fū", "sp": "fu", "r": "肉", "sc": 15, "l": 5, "f": 6526},
  {"h": "虧", "p": "kuī", "sp": "kui", "r": "虍", "sc": 17, "l": 5, "f": 6527},
  {"h": "邊", "p": "biān", "sp": "bian", "r": "辵", "sc": 18, "l": 5, "f": 6528},
  {"h": "髮", "p": "fà", "sp": "fa", "r": "髟", "sc": 15, "l": 5, "f": 6529},
  {"h": "埚", "p": "guō", "sp": "guo", "r": "土", "sc": 10, "l": 5, "f": 6530},
  {"h": "堙", "p": "yīn", "sp": "yin", "r": "土", "sc": 12, "l": 5, "f": 6531},
  {"h": "菥", "p": "xī", "sp": "xi", "r": "艸", "sc": 11, "l": 5, "f": 6532},
  {"h": "萏", "p": "dàn", "sp": "dan", "r": "艸", "sc": 11, "l": 5, "f": 6533},
  {"h": "玟", "p": "wén", "sp": "wen", "r": "玉", "sc": 8, "l": 5, "f": 6534},
  {"h": "椟", "p": "dú", "sp": "du", "r": "木", "sc": 12, "l": 5, "f": 6535},
  {"h": "铗", "p": "jiá", "sp": "jia", "r": "金", "sc": 11, "l": 5, "f": 6536},
  {"h": "颙", "p": "yóng", "sp": "yong", "r": "頁", "sc": 15, "l": 5, "f": 6537},
  {"h": "踬", "p": "zhì", "sp": "zhi", "r": "足", "sc": 15, "l": 5, "f": 6538},
  {"h": "筚", "p": "bì", "sp": "bi", "r": "竹", "sc": 12, "l": 5, "f": 6539},
  {"h": "矻", "p": "kū", "sp": "ku", "r": "石", "sc": 8, "l": 5, "f": 6540},
  {"h": "搠", "p": "shuò", "sp": "shuo", "r": "手", "sc": 13, "l": 5, "f": 6541},
  {"h": "圉", "p": "yǔ", "sp": "yu", "r": "囗", "sc": 11, "l": 5, "f": 6542},
  {"h": "浍", "p": "huì", "sp": "hui", "r": "水", "sc": 9, "l": 5, "f": 6543},
  {"h": "閱", "p": "yuè", "sp": "yue", "r": "門", "sc": 15, "l": 5, "f": 6544},
  {"h": "蛸", "p": "shāo", "sp": "shao", "r": "虫", "sc": 13, "l": 5, "f": 6545},
  {"h": "礎", "p": "chǔ", "sp": "chu", "r": "石", "sc": 18, "l": 5, "f": 6546},
  {"h": "討", "p": "tǎo", "sp": "tao", "r": "言", "sc": 10, "l": 5, "f": 6547},
  {"h": "鎮", "p": "zhèn", "sp": "zhen", "r": "金", "sc": 18, "l": 5, "f": 6548},
  {"h": "鮊", "p": "bà", "sp": "ba", "r": "魚", "sc": 16, "l": 5, "f": 6549},
  {"h": "芘", "p": "pí", "sp": "pi", "r": "艸", "sc": 7, "l": 5, "f": 6550},
  {"h": "萜", "p": "tiē", "sp": "tie", "r": "艸", "sc": 11, "l": 5, "f": 6551},
  {"h": "鄗", "p": "hào", "sp": "hao", "r": "邑", "sc": 12, "l": 5, "f": 6552},
  {
    "h": "惝",
    "p": "chǎng",
    "sp": "chang",
    "r": "心",
    "sc": 11,
    "l": 5,
    "f": 6553
  },
  {"h": "毵", "p": "sān", "sp": "san", "r": "毛", "sc": 12, "l": 5, "f": 6554},
  {"h": "鲽", "p": "dié", "sp": "die", "r": "魚", "sc": 17, "l": 5, "f": 6555},
  {"h": "酞", "p": "tài", "sp": "tai", "r": "酉", "sc": 11, "l": 5, "f": 6556},
  {"h": "絕", "p": "jué", "sp": "jue", "r": "糸", "sc": 12, "l": 5, "f": 6557},
  {"h": "哚", "p": "duǒ", "sp": "duo", "r": "口", "sc": 9, "l": 5, "f": 6558},
  {"h": "溷", "p": "hùn", "sp": "hun", "r": "水", "sc": 13, "l": 5, "f": 6559},
  {"h": "缞", "p": "cuī", "sp": "cui", "r": "糸", "sc": 13, "l": 5, "f": 6560},
  {"h": "踰", "p": "yú", "sp": "yu", "r": "足", "sc": 16, "l": 5, "f": 6561},
  {"h": "搦", "p": "nuò", "sp": "nuo", "r": "手", "sc": 13, "l": 5, "f": 6562},
  {"h": "帻", "p": "zé", "sp": "ze", "r": "巾", "sc": 11, "l": 5, "f": 6563},
  {"h": "铦", "p": "xiān", "sp": "xian", "r": "金", "sc": 11, "l": 5, "f": 6564},
  {"h": "鲴", "p": "gù", "sp": "gu", "r": "魚", "sc": 16, "l": 5, "f": 6565},
  {"h": "髫", "p": "tiáo", "sp": "tiao", "r": "髟", "sc": 15, "l": 5, "f": 6566},
  {"h": "盡", "p": "jǐn", "sp": "jin", "r": "皿", "sc": 14, "l": 5, "f": 6567},
  {"h": "蓋", "p": "gài", "sp": "gai", "r": "艸", "sc": 13, "l": 5, "f": 6568},
  {"h": "製", "p": "zhì", "sp": "zhi", "r": "衣", "sc": 14, "l": 5, "f": 6569},
  {"h": "載", "p": "zài", "sp": "zai", "r": "車", "sc": 13, "l": 5, "f": 6570},
  {"h": "週", "p": "zhōu", "sp": "zhou", "r": "辵", "sc": 11, "l": 5, "f": 6571},
  {"h": "陸", "p": "lù", "sp": "lu", "r": "阜", "sc": 10, "l": 5, "f": 6572},
  {"h": "醡", "p": "zhà", "sp": "zha", "r": "酉", "sc": 17, "l": 5, "f": 6573},
  {"h": "儇", "p": "xuān", "sp": "xuan", "r": "人", "sc": 15, "l": 5, "f": 6574},
  {"h": "狴", "p": "bì", "sp": "bi", "r": "犬", "sc": 10, "l": 5, "f": 6575},
  {"h": "迮", "p": "zé", "sp": "ze", "r": "辵", "sc": 8, "l": 5, "f": 6576},
  {"h": "鍬", "p": "qiāo", "sp": "qiao", "r": "金", "sc": 17, "l": 5, "f": 6577},
  {"h": "趱", "p": "zǎn", "sp": "zan", "r": "走", "sc": 23, "l": 5, "f": 6578},
  {"h": "椠", "p": "qiàn", "sp": "qian", "r": "木", "sc": 12, "l": 5, "f": 6579},
  {"h": "赟", "p": "yūn", "sp": "yun", "r": "貝", "sc": 16, "l": 5, "f": 6580},
  {"h": "陔", "p": "gāi", "sp": "gai", "r": "阜", "sc": 8, "l": 5, "f": 6581},
  {"h": "坌", "p": "bèn", "sp": "ben", "r": "土", "sc": 7, "l": 5, "f": 6582},
  {"h": "悃", "p": "kǔn", "sp": "kun", "r": "心", "sc": 10, "l": 5, "f": 6583},
  {
    "h": "塍",
    "p": "chéng",
    "sp": "cheng",
    "r": "土",
    "sc": 13,
    "l": 5,
    "f": 6584
  },
  {"h": "疬", "p": "lì", "sp": "li", "r": "疒", "sc": 9, "l": 5, "f": 6585},
  {"h": "麈", "p": "zhǔ", "sp": "zhu", "r": "鹿", "sc": 16, "l": 5, "f": 6586},
  {"h": "絜", "p": "jié", "sp": "jie", "r": "糸", "sc": 12, "l": 5, "f": 6587},
  {"h": "脩", "p": "xiū", "sp": "xiu", "r": "肉", "sc": 10, "l": 5, "f": 6588},
  {"h": "赒", "p": "zhōu", "sp": "zhou", "r": "貝", "sc": 12, "l": 5, "f": 6589},
  {"h": "埒", "p": "liè", "sp": "lie", "r": "土", "sc": 10, "l": 5, "f": 6590},
  {"h": "镡", "p": "chán", "sp": "chan", "r": "金", "sc": 17, "l": 5, "f": 6591},
  {"h": "镨", "p": "pǔ", "sp": "pu", "r": "金", "sc": 17, "l": 5, "f": 6592},
  {"h": "骎", "p": "qīn", "sp": "qin", "r": "馬", "sc": 10, "l": 5, "f": 6593},
  {"h": "骙", "p": "kuí", "sp": "kui", "r": "馬", "sc": 12, "l": 5, "f": 6594},
  {
    "h": "薑",
    "p": "jiāng",
    "sp": "jiang",
    "r": "艸",
    "sc": 16,
    "l": 5,
    "f": 6595
  },
  {"h": "鼩", "p": "qú", "sp": "qu", "r": "鼠", "sc": 18, "l": 5, "f": 6596},
  {"h": "雰", "p": "fēn", "sp": "fen", "r": "雨", "sc": 12, "l": 5, "f": 6597},
  {"h": "穩", "p": "wěn", "sp": "wen", "r": "禾", "sc": 19, "l": 5, "f": 6598},
  {"h": "藴", "p": "yùn", "sp": "yun", "r": "艸", "sc": 18, "l": 5, "f": 6599},
  {"h": "責", "p": "zé", "sp": "ze", "r": "貝", "sc": 11, "l": 5, "f": 6600},
  {"h": "鮮", "p": "xiān", "sp": "xian", "r": "魚", "sc": 17, "l": 5, "f": 6601},
  {"h": "鳑", "p": "páng", "sp": "pang", "r": "魚", "sc": 18, "l": 5, "f": 6602},
  {"h": "豨", "p": "xī", "sp": "xi", "r": "豕", "sc": 14, "l": 5, "f": 6603},
  {"h": "琚", "p": "jū", "sp": "ju", "r": "玉", "sc": 12, "l": 5, "f": 6604},
  {"h": "觫", "p": "sù", "sp": "su", "r": "角", "sc": 14, "l": 5, "f": 6605},
  {"h": "鲦", "p": "tiáo", "sp": "tiao", "r": "魚", "sc": 15, "l": 5, "f": 6606},
  {"h": "刳", "p": "kū", "sp": "ku", "r": "刀", "sc": 8, "l": 5, "f": 6607},
  {"h": "拶", "p": "zā", "sp": "za", "r": "手", "sc": 9, "l": 5, "f": 6608},
  {"h": "掊", "p": "póu", "sp": "pou", "r": "手", "sc": 11, "l": 5, "f": 6609},
  {"h": "泐", "p": "lè", "sp": "le", "r": "水", "sc": 7, "l": 5, "f": 6610},
  {"h": "瘗", "p": "yì", "sp": "yi", "r": "疒", "sc": 14, "l": 5, "f": 6611},
  {"h": "雩", "p": "yú", "sp": "yu", "r": "雨", "sc": 11, "l": 5, "f": 6612},
  {"h": "麿", "p": "mo", "sp": "mo", "r": "麻", "sc": 17, "l": 5, "f": 6613},
  {"h": "檗", "p": "bò", "sp": "bo", "r": "木", "sc": 17, "l": 5, "f": 6614},
  {"h": "鼂", "p": "cháo", "sp": "chao", "r": "黽", "sc": 18, "l": 5, "f": 6615},
  {
    "h": "詳",
    "p": "xiáng",
    "sp": "xiang",
    "r": "言",
    "sc": 13,
    "l": 5,
    "f": 6616
  },
  {"h": "際", "p": "jì", "sp": "ji", "r": "阜", "sc": 13, "l": 5, "f": 6617},
  {"h": "褔", "p": "fù", "sp": "fu", "r": "衣", "sc": 14, "l": 5, "f": 6618},
  {"h": "铟", "p": "yīn", "sp": "yin", "r": "金", "sc": 11, "l": 5, "f": 6619},
  {"h": "聽", "p": "tīng", "sp": "ting", "r": "耳", "sc": 22, "l": 5, "f": 6620},
  {"h": "觸", "p": "chù", "sp": "chu", "r": "角", "sc": 20, "l": 5, "f": 6621},
  {"h": "靈", "p": "líng", "sp": "ling", "r": "雨", "sc": 24, "l": 5, "f": 6622},
  {"h": "鲏", "p": "pí", "sp": "pi", "r": "魚", "sc": 13, "l": 5, "f": 6623},
  {"h": "蒗", "p": "làng", "sp": "lang", "r": "艸", "sc": 13, "l": 5, "f": 6624},
  {"h": "魟", "p": "hóng", "sp": "hong", "r": "魚", "sc": 14, "l": 5, "f": 6625},
  {"h": "遹", "p": "yù", "sp": "yu", "r": "辵", "sc": 15, "l": 5, "f": 6626},
  {"h": "攴", "p": "pū", "sp": "pu", "r": "攴", "sc": 4, "l": 5, "f": 6627},
  {"h": "蝥", "p": "máo", "sp": "mao", "r": "虫", "sc": 15, "l": 5, "f": 6628},
  {"h": "茆", "p": "máo", "sp": "mao", "r": "艸", "sc": 8, "l": 5, "f": 6629},
  {"h": "蓁", "p": "zhēn", "sp": "zhen", "r": "艸", "sc": 13, "l": 5, "f": 6630},
  {"h": "嚆", "p": "hāo", "sp": "hao", "r": "口", "sc": 16, "l": 5, "f": 6631},
  {"h": "忮", "p": "zhì", "sp": "zhi", "r": "心", "sc": 7, "l": 5, "f": 6632},
  {"h": "篌", "p": "hóu", "sp": "hou", "r": "竹", "sc": 15, "l": 5, "f": 6633},
  {"h": "菧", "p": "dǐ", "sp": "di", "r": "艸", "sc": 11, "l": 5, "f": 6634},
  {"h": "卺", "p": "jǐn", "sp": "jin", "r": "卩", "sc": 8, "l": 5, "f": 6635},
  {"h": "郫", "p": "pí", "sp": "pi", "r": "邑", "sc": 10, "l": 5, "f": 6636},
  {"h": "篠", "p": "xiǎo", "sp": "xiao", "r": "竹", "sc": 16, "l": 5, "f": 6637},
  {"h": "複", "p": "fù", "sp": "fu", "r": "衣", "sc": 14, "l": 5, "f": 6638},
  {"h": "該", "p": "gāi", "sp": "gai", "r": "言", "sc": 13, "l": 5, "f": 6639},
  {"h": "貴", "p": "guì", "sp": "gui", "r": "貝", "sc": 12, "l": 5, "f": 6640},
  {"h": "辯", "p": "biàn", "sp": "bian", "r": "辛", "sc": 21, "l": 5, "f": 6641},
  {"h": "飛", "p": "fēi", "sp": "fei", "r": "飛", "sc": 9, "l": 5, "f": 6642},
  {"h": "犰", "p": "qiú", "sp": "qiu", "r": "犬", "sc": 5, "l": 5, "f": 6643},
  {"h": "狳", "p": "yú", "sp": "yu", "r": "犬", "sc": 10, "l": 5, "f": 6644},
  {"h": "陠", "p": "pū", "sp": "pu", "r": "阜", "sc": 9, "l": 5, "f": 6645},
  {"h": "蹁", "p": "pián", "sp": "pian", "r": "足", "sc": 16, "l": 5, "f": 6646},
  {"h": "蓐", "p": "rù", "sp": "ru", "r": "艸", "sc": 13, "l": 5, "f": 6647},
  {"h": "稜", "p": "léng", "sp": "leng", "r": "禾", "sc": 13, "l": 5, "f": 6648},
  {"h": "菃", "p": "qú", "sp": "qu", "r": "艸", "sc": 10, "l": 5, "f": 6649},
  {"h": "埝", "p": "niàn", "sp": "nian", "r": "土", "sc": 11, "l": 5, "f": 6650},
  {"h": "蕞", "p": "zuì", "sp": "zui", "r": "艸", "sc": 15, "l": 5, "f": 6651},
  {"h": "喈", "p": "jiē", "sp": "jie", "r": "口", "sc": 12, "l": 5, "f": 6652},
  {"h": "戋", "p": "jiān", "sp": "jian", "r": "戈", "sc": 5, "l": 5, "f": 6653},
  {"h": "赙", "p": "fù", "sp": "fu", "r": "貝", "sc": 14, "l": 5, "f": 6654},
  {"h": "磉", "p": "sǎng", "sp": "sang", "r": "石", "sc": 15, "l": 5, "f": 6655},
  {"h": "郤", "p": "xì", "sp": "xi", "r": "邑", "sc": 9, "l": 5, "f": 6656},
  {"h": "聍", "p": "níng", "sp": "ning", "r": "耳", "sc": 11, "l": 5, "f": 6657},
  {"h": "铷", "p": "rú", "sp": "ru", "r": "金", "sc": 11, "l": 5, "f": 6658},
  {"h": "硏", "p": "yán", "sp": "yan", "r": "石", "sc": 11, "l": 5, "f": 6659},
  {"h": "腘", "p": "guó", "sp": "guo", "r": "肉", "sc": 12, "l": 5, "f": 6660},
  {"h": "興", "p": "xìng", "sp": "xing", "r": "臼", "sc": 16, "l": 5, "f": 6661},
  {"h": "雞", "p": "jī", "sp": "ji", "r": "隹", "sc": 18, "l": 5, "f": 6662},
  {"h": "鬱", "p": "yù", "sp": "yu", "r": "鬯", "sc": 29, "l": 5, "f": 6663},
  {"h": "鳴", "p": "míng", "sp": "ming", "r": "鳥", "sc": 14, "l": 5, "f": 6664},
  {"h": "麥", "p": "mài", "sp": "mai", "r": "麥", "sc": 11, "l": 5, "f": 6665},
  {"h": "鮟", "p": "àn", "sp": "an", "r": "魚", "sc": 17, "l": 5, "f": 6666},
  {"h": "鳈", "p": "quán", "sp": "quan", "r": "魚", "sc": 17, "l": 5, "f": 6667},
  {"h": "秾", "p": "nóng", "sp": "nong", "r": "禾", "sc": 11, "l": 5, "f": 6668},
  {"h": "邗", "p": "hán", "sp": "han", "r": "邑", "sc": 5, "l": 5, "f": 6669},
  {"h": "釐", "p": "lí", "sp": "li", "r": "里", "sc": 18, "l": 5, "f": 6670},
  {"h": "矧", "p": "shěn", "sp": "shen", "r": "矢", "sc": 9, "l": 5, "f": 6671},
  {"h": "舁", "p": "yú", "sp": "yu", "r": "臼", "sc": 9, "l": 5, "f": 6672},
  {"h": "貅", "p": "xiū", "sp": "xiu", "r": "豸", "sc": 13, "l": 5, "f": 6673},
  {"h": "瞇", "p": "mī", "sp": "mi", "r": "目", "sc": 14, "l": 5, "f": 6674},
  {"h": "骢", "p": "cōng", "sp": "cong", "r": "馬", "sc": 14, "l": 5, "f": 6675},
  {"h": "阼", "p": "zuò", "sp": "zuo", "r": "阜", "sc": 7, "l": 5, "f": 6676},
  {"h": "垸", "p": "yuàn", "sp": "yuan", "r": "土", "sc": 10, "l": 5, "f": 6677},
  {"h": "墁", "p": "màn", "sp": "man", "r": "土", "sc": 14, "l": 5, "f": 6678},
  {"h": "攮", "p": "nǎng", "sp": "nang", "r": "手", "sc": 25, "l": 5, "f": 6679},
  {"h": "柝", "p": "tuò", "sp": "tuo", "r": "木", "sc": 9, "l": 5, "f": 6680},
  {"h": "梃", "p": "tǐng", "sp": "ting", "r": "木", "sc": 10, "l": 5, "f": 6681},
  {
    "h": "眚",
    "p": "shěng",
    "sp": "sheng",
    "r": "目",
    "sc": 10,
    "l": 5,
    "f": 6682
  },
  {"h": "箜", "p": "kōng", "sp": "kong", "r": "竹", "sc": 14, "l": 5, "f": 6683},
  {"h": "簏", "p": "lù", "sp": "lu", "r": "竹", "sc": 17, "l": 5, "f": 6684},
  {"h": "綮", "p": "qǐ", "sp": "qi", "r": "糸", "sc": 14, "l": 5, "f": 6685},
  {"h": "鲩", "p": "huàn", "sp": "huan", "r": "魚", "sc": 15, "l": 5, "f": 6686},
  {"h": "諡", "p": "shì", "sp": "shi", "r": "言", "sc": 16, "l": 5, "f": 6687},
  {"h": "茇", "p": "bá", "sp": "ba", "r": "艸", "sc": 8, "l": 5, "f": 6688},
  {"h": "摅", "p": "shū", "sp": "shu", "r": "手", "sc": 13, "l": 5, "f": 6689},
  {"h": "顐", "p": "wèn", "sp": "wen", "r": "頁", "sc": 18, "l": 5, "f": 6690},
  {"h": "蜾", "p": "guǒ", "sp": "guo", "r": "虫", "sc": 14, "l": 5, "f": 6691},
  {"h": "龂", "p": "yín", "sp": "yin", "r": "齒", "sc": 12, "l": 5, "f": 6692},
  {"h": "䦟", "p": "wěn", "sp": "wen", "r": "門", "sc": 15, "l": 5, "f": 6693},
  {"h": "臍", "p": "qí", "sp": "qi", "r": "肉", "sc": 18, "l": 5, "f": 6694},
  {"h": "膩", "p": "nì", "sp": "ni", "r": "肉", "sc": 16, "l": 5, "f": 6695},
  {"h": "諸", "p": "zhū", "sp": "zhu", "r": "言", "sc": 15, "l": 5, "f": 6696},
  {"h": "讉", "p": "yí", "sp": "yi", "r": "言", "sc": 22, "l": 5, "f": 6697},
  {"h": "鎯", "p": "láng", "sp": "lang", "r": "金", "sc": 16, "l": 5, "f": 6698},
  {"h": "脒", "p": "mǐ", "sp": "mi", "r": "肉", "sc": 10, "l": 5, "f": 6699},
  {"h": "箬", "p": "ruò", "sp": "ruo", "r": "竹", "sc": 14, "l": 5, "f": 6700},
  {"h": "鲻", "p": "zī", "sp": "zi", "r": "魚", "sc": 16, "l": 5, "f": 6701},
  {"h": "埸", "p": "yì", "sp": "yi", "r": "土", "sc": 11, "l": 5, "f": 6702},
  {"h": "颃", "p": "háng", "sp": "hang", "r": "頁", "sc": 10, "l": 5, "f": 6703},
  {"h": "蒹", "p": "jiān", "sp": "jian", "r": "艸", "sc": 13, "l": 5, "f": 6704},
  {"h": "瞍", "p": "sǒu", "sp": "sou", "r": "目", "sc": 14, "l": 5, "f": 6705},
  {"h": "铫", "p": "diào", "sp": "diao", "r": "金", "sc": 11, "l": 5, "f": 6706},
  {"h": "蒐", "p": "sōu", "sp": "sou", "r": "艸", "sc": 12, "l": 5, "f": 6707},
  {"h": "蹻", "p": "juē", "sp": "jue", "r": "足", "sc": 19, "l": 5, "f": 6708},
  {"h": "幞", "p": "fú", "sp": "fu", "r": "巾", "sc": 15, "l": 5, "f": 6709},
  {"h": "阯", "p": "zhǐ", "sp": "zhi", "r": "阜", "sc": 6, "l": 5, "f": 6710},
  {"h": "觏", "p": "gòu", "sp": "gou", "r": "見", "sc": 14, "l": 5, "f": 6711},
  {"h": "媵", "p": "yìng", "sp": "ying", "r": "女", "sc": 13, "l": 5, "f": 6712},
  {"h": "钬", "p": "huǒ", "sp": "huo", "r": "金", "sc": 9, "l": 5, "f": 6713},
  {"h": "襦", "p": "rú", "sp": "ru", "r": "衣", "sc": 19, "l": 5, "f": 6714},
  {"h": "簋", "p": "guǐ", "sp": "gui", "r": "竹", "sc": 17, "l": 5, "f": 6715},
  {"h": "翮", "p": "hé", "sp": "he", "r": "羽", "sc": 16, "l": 5, "f": 6716},
  {"h": "蹯", "p": "fán", "sp": "fan", "r": "足", "sc": 19, "l": 5, "f": 6717},
  {"h": "艖", "p": "chā", "sp": "cha", "r": "舟", "sc": 15, "l": 5, "f": 6718},
  {"h": "貓", "p": "māo", "sp": "mao", "r": "豸", "sc": 15, "l": 5, "f": 6719},
  {"h": "鐮", "p": "lián", "sp": "lian", "r": "金", "sc": 21, "l": 5, "f": 6720},
  {"h": "負", "p": "fù", "sp": "fu", "r": "貝", "sc": 9, "l": 5, "f": 6721},
  {"h": "順", "p": "shùn", "sp": "shun", "r": "頁", "sc": 12, "l": 5, "f": 6722},
  {
    "h": "鲿",
    "p": "cháng",
    "sp": "chang",
    "r": "魚",
    "sc": 17,
    "l": 5,
    "f": 6723
  },
  {"h": "緊", "p": "jǐn", "sp": "jin", "r": "糸", "sc": 14, "l": 5, "f": 6724},
  {"h": "蹠", "p": "zhí", "sp": "zhi", "r": "足", "sc": 18, "l": 5, "f": 6725},
  {"h": "攉", "p": "huō", "sp": "huo", "r": "手", "sc": 19, "l": 5, "f": 6726},
  {"h": "禊", "p": "xì", "sp": "xi", "r": "示", "sc": 13, "l": 5, "f": 6727},
  {"h": "箐", "p": "qìng", "sp": "qing", "r": "竹", "sc": 14, "l": 5, "f": 6728},
  {"h": "镟", "p": "xuàn", "sp": "xuan", "r": "金", "sc": 16, "l": 5, "f": 6729},
  {"h": "癍", "p": "bān", "sp": "ban", "r": "疒", "sc": 17, "l": 5, "f": 6730},
  {"h": "霪", "p": "yín", "sp": "yin", "r": "雨", "sc": 19, "l": 5, "f": 6731},
  {"h": "袪", "p": "qū", "sp": "qu", "r": "衣", "sc": 10, "l": 5, "f": 6732},
  {"h": "劓", "p": "yì", "sp": "yi", "r": "刀", "sc": 16, "l": 5, "f": 6733},
  {"h": "殂", "p": "cú", "sp": "cu", "r": "歹", "sc": 9, "l": 5, "f": 6734},
  {"h": "脞", "p": "cuǒ", "sp": "cuo", "r": "肉", "sc": 11, "l": 5, "f": 6735},
  {"h": "霨", "p": "wèi", "sp": "wei", "r": "雨", "sc": 19, "l": 5, "f": 6736},
  {"h": "饾", "p": "dòu", "sp": "dou", "r": "食", "sc": 10, "l": 5, "f": 6737},
  {"h": "耵", "p": "dīng", "sp": "ding", "r": "耳", "sc": 8, "l": 5, "f": 6738},
  {"h": "泺", "p": "luò", "sp": "luo", "r": "水", "sc": 8, "l": 5, "f": 6739},
  {"h": "埤", "p": "pí", "sp": "pi", "r": "土", "sc": 11, "l": 5, "f": 6740},
  {"h": "碁", "p": "qí", "sp": "qi", "r": "石", "sc": 13, "l": 5, "f": 6741},
  {"h": "鲐", "p": "tái", "sp": "tai", "r": "魚", "sc": 13, "l": 5, "f": 6742},
  {"h": "鷟", "p": "zhuó", "sp": "zhuo", "r": "鳥", "sc": 22, "l": 5, "f": 6743},
  {"h": "膬", "p": "cuì", "sp": "cui", "r": "肉", "sc": 16, "l": 5, "f": 6744},
  {"h": "茲", "p": "zī", "sp": "zi", "r": "艸", "sc": 9, "l": 5, "f": 6745},
  {"h": "粯", "p": "xiàn", "sp": "xian", "r": "米", "sc": 13, "l": 5, "f": 6746},
  {"h": "糞", "p": "fèn", "sp": "fen", "r": "米", "sc": 17, "l": 5, "f": 6747},
  {
    "h": "聲",
    "p": "shēng",
    "sp": "sheng",
    "r": "耳",
    "sc": 17,
    "l": 5,
    "f": 6748
  },
  {"h": "貼", "p": "tiē", "sp": "tie", "r": "貝", "sc": 12, "l": 5, "f": 6749},
  {"h": "陴", "p": "pí", "sp": "pi", "r": "阜", "sc": 10, "l": 5, "f": 6750},
  {"h": "橼", "p": "yuán", "sp": "yuan", "r": "木", "sc": 16, "l": 5, "f": 6751},
  {"h": "驩", "p": "huān", "sp": "huan", "r": "馬", "sc": 27, "l": 5, "f": 6752},
  {"h": "唣", "p": "zào", "sp": "zao", "r": "口", "sc": 10, "l": 5, "f": 6753},
  {"h": "蜩", "p": "tiáo", "sp": "tiao", "r": "虫", "sc": 14, "l": 5, "f": 6754},
  {"h": "埯", "p": "ǎn", "sp": "an", "r": "土", "sc": 11, "l": 5, "f": 6755},
  {"h": "蓥", "p": "yíng", "sp": "ying", "r": "艸", "sc": 13, "l": 5, "f": 6756},
  {"h": "袝", "p": "fù", "sp": "fu", "r": "衣", "sc": 10, "l": 5, "f": 6757},
  {"h": "谞", "p": "xū", "sp": "xu", "r": "言", "sc": 11, "l": 5, "f": 6758},
  {"h": "馘", "p": "guó", "sp": "guo", "r": "首", "sc": 17, "l": 5, "f": 6759},
  {"h": "谫", "p": "jiǎn", "sp": "jian", "r": "言", "sc": 13, "l": 5, "f": 6760},
  {"h": "輶", "p": "yóu", "sp": "you", "r": "車", "sc": 16, "l": 5, "f": 6761},
  {"h": "廑", "p": "jǐn", "sp": "jin", "r": "广", "sc": 14, "l": 5, "f": 6762},
  {"h": "瘐", "p": "yǔ", "sp": "yu", "r": "疒", "sc": 13, "l": 5, "f": 6763},
  {"h": "躐", "p": "liè", "sp": "lie", "r": "足", "sc": 22, "l": 5, "f": 6764},
  {"h": "黟", "p": "yī", "sp": "yi", "r": "黑", "sc": 18, "l": 5, "f": 6765},
  {"h": "纔", "p": "cái", "sp": "cai", "r": "糸", "sc": 23, "l": 5, "f": 6766},
  {"h": "礙", "p": "ài", "sp": "ai", "r": "石", "sc": 19, "l": 5, "f": 6767},
  {"h": "膯", "p": "tēng", "sp": "teng", "r": "肉", "sc": 16, "l": 5, "f": 6768},
  {"h": "華", "p": "huá", "sp": "hua", "r": "艸", "sc": 10, "l": 5, "f": 6769},
  {
    "h": "藭",
    "p": "qióng",
    "sp": "qiong",
    "r": "艸",
    "sc": 18,
    "l": 5,
    "f": 6770
  },
  {"h": "禆", "p": "bēi", "sp": "bei", "r": "示", "sc": 12, "l": 5, "f": 6771},
  {"h": "號", "p": "hào", "sp": "hao", "r": "虍", "sc": 13, "l": 5, "f": 6772},
  {"h": "談", "p": "tán", "sp": "tan", "r": "言", "sc": 15, "l": 5, "f": 6773},
  {"h": "費", "p": "fèi", "sp": "fei", "r": "貝", "sc": 12, "l": 5, "f": 6774},
  {"h": "顏", "p": "yán", "sp": "yan", "r": "頁", "sc": 18, "l": 5, "f": 6775},
  {"h": "飯", "p": "fàn", "sp": "fan", "r": "食", "sc": 12, "l": 5, "f": 6776},
  {"h": "稓", "p": "zuó", "sp": "zuo", "r": "禾", "sc": 13, "l": 5, "f": 6777},
  {"h": "縠", "p": "hú", "sp": "hu", "r": "糸", "sc": 16, "l": 5, "f": 6778},
  {"h": "揠", "p": "yà", "sp": "ya", "r": "手", "sc": 12, "l": 5, "f": 6779},
  {"h": "绐", "p": "dài", "sp": "dai", "r": "糸", "sc": 8, "l": 5, "f": 6780},
  {"h": "鸰", "p": "líng", "sp": "ling", "r": "鳥", "sc": 10, "l": 5, "f": 6781},
  {"h": "搛", "p": "jiān", "sp": "jian", "r": "手", "sc": 13, "l": 5, "f": 6782},
  {"h": "黹", "p": "zhǐ", "sp": "zhi", "r": "黹", "sc": 12, "l": 5, "f": 6783},
  {"h": "恧", "p": "nǜ", "sp": "nu", "r": "心", "sc": 10, "l": 5, "f": 6784},
  {"h": "粎", "p": "chǐ", "sp": "chi", "r": "米", "sc": 10, "l": 5, "f": 6785},
  {"h": "柒", "p": "qī", "sp": "qi", "r": "木", "sc": 9, "l": 5, "f": 6786},
  {"h": "裒", "p": "póu", "sp": "pou", "r": "衣", "sc": 12, "l": 5, "f": 6787},
  {"h": "栳", "p": "lǎo", "sp": "lao", "r": "木", "sc": 10, "l": 5, "f": 6788},
  {"h": "钸", "p": "bù", "sp": "bu", "r": "金", "sc": 10, "l": 5, "f": 6789},
  {"h": "饸", "p": "hé", "sp": "he", "r": "食", "sc": 9, "l": 5, "f": 6790},
  {"h": "譞", "p": "xuān", "sp": "xuan", "r": "言", "sc": 20, "l": 5, "f": 6791},
  {"h": "販", "p": "fàn", "sp": "fan", "r": "貝", "sc": 11, "l": 5, "f": 6792},
  {"h": "铕", "p": "yǒu", "sp": "you", "r": "金", "sc": 11, "l": 5, "f": 6793},
  {"h": "絷", "p": "zhí", "sp": "zhi", "r": "糸", "sc": 12, "l": 5, "f": 6794},
  {"h": "蒨", "p": "qiàn", "sp": "qian", "r": "艸", "sc": 13, "l": 5, "f": 6795},
  {"h": "蓣", "p": "yù", "sp": "yu", "r": "艸", "sc": 13, "l": 5, "f": 6796},
  {"h": "迻", "p": "yí", "sp": "yi", "r": "辵", "sc": 9, "l": 5, "f": 6797},
  {"h": "釆", "p": "biàn", "sp": "bian", "r": "釆", "sc": 7, "l": 5, "f": 6798},
  {"h": "鲕", "p": "ér", "sp": "er", "r": "魚", "sc": 14, "l": 5, "f": 6799},
  {"h": "頗", "p": "pō", "sp": "po", "r": "頁", "sc": 14, "l": 5, "f": 6800},
  {"h": "鍏", "p": "wéi", "sp": "wei", "r": "金", "sc": 17, "l": 5, "f": 6801},
  {"h": "鷉", "p": "tī", "sp": "ti", "r": "鳥", "sc": 21, "l": 5, "f": 6802},
  {"h": "郄", "p": "qiè", "sp": "qie", "r": "邑", "sc": 8, "l": 5, "f": 6803},
  {"h": "埘", "p": "shí", "sp": "shi", "r": "土", "sc": 10, "l": 5, "f": 6804},
  {"h": "薤", "p": "xiè", "sp": "xie", "r": "艸", "sc": 16, "l": 5, "f": 6805},
  {
    "h": "鲞",
    "p": "xiǎng",
    "sp": "xiang",
    "r": "魚",
    "sc": 14,
    "l": 5,
    "f": 6806
  },
  {"h": "鼯", "p": "wú", "sp": "wu", "r": "鼠", "sc": 20, "l": 5, "f": 6807},
  {"h": "坭", "p": "ní", "sp": "ni", "r": "土", "sc": 8, "l": 5, "f": 6808},
  {"h": "蔌", "p": "sù", "sp": "su", "r": "艸", "sc": 14, "l": 5, "f": 6809},
  {"h": "缳", "p": "huán", "sp": "huan", "r": "糸", "sc": 16, "l": 5, "f": 6810},
  {"h": "靦", "p": "tiǎn", "sp": "tian", "r": "面", "sc": 16, "l": 5, "f": 6811},
  {"h": "筸", "p": "gān", "sp": "gan", "r": "竹", "sc": 13, "l": 5, "f": 6812},
  {"h": "蘋", "p": "píng", "sp": "ping", "r": "艸", "sc": 19, "l": 5, "f": 6813},
  {"h": "夼", "p": "kuǎng", "sp": "kuang", "r": "大", "sc": 6, "l": 5, "f": 6814},
  {"h": "椐", "p": "jū", "sp": "ju", "r": "木", "sc": 12, "l": 5, "f": 6815},
  {"h": "轹", "p": "lì", "sp": "li", "r": "車", "sc": 9, "l": 5, "f": 6816},
  {"h": "脬", "p": "pāo", "sp": "pao", "r": "肉", "sc": 11, "l": 5, "f": 6817},
  {"h": "眍", "p": "kōu", "sp": "kou", "r": "目", "sc": 9, "l": 5, "f": 6818},
  {"h": "饹", "p": "le", "sp": "le", "r": "食", "sc": 9, "l": 5, "f": 6819},
  {"h": "瘥", "p": "chài", "sp": "chai", "r": "疒", "sc": 14, "l": 5, "f": 6820},
  {"h": "瘵", "p": "zhài", "sp": "zhai", "r": "疒", "sc": 16, "l": 5, "f": 6821},
  {"h": "蝦", "p": "xiā", "sp": "xia", "r": "虫", "sc": 15, "l": 5, "f": 6822},
  {"h": "誌", "p": "zhì", "sp": "zhi", "r": "言", "sc": 14, "l": 5, "f": 6823},
  {"h": "邴", "p": "bǐng", "sp": "bing", "r": "邑", "sc": 7, "l": 5, "f": 6824},
  {"h": "鉁", "p": "zhēn", "sp": "zhen", "r": "金", "sc": 13, "l": 5, "f": 6825},
  {"h": "鞮", "p": "dī", "sp": "di", "r": "革", "sc": 18, "l": 5, "f": 6826},
  {"h": "舄", "p": "xì", "sp": "xi", "r": "臼", "sc": 12, "l": 5, "f": 6827},
  {"h": "麄", "p": "cū", "sp": "cu", "r": "鹿", "sc": 15, "l": 5, "f": 6828},
  {"h": "翛", "p": "xiāo", "sp": "xiao", "r": "羽", "sc": 12, "l": 5, "f": 6829},
  {"h": "覩", "p": "dǔ", "sp": "du", "r": "見", "sc": 15, "l": 5, "f": 6830},
  {"h": "纩", "p": "kuàng", "sp": "kuang", "r": "糸", "sc": 6, "l": 5, "f": 6831},
  {"h": "蓮", "p": "lián", "sp": "lian", "r": "艸", "sc": 13, "l": 5, "f": 6832},
  {"h": "唛", "p": "mà", "sp": "ma", "r": "口", "sc": 10, "l": 5, "f": 6833},
  {"h": "袓", "p": "jù", "sp": "ju", "r": "衣", "sc": 10, "l": 5, "f": 6834},
  {"h": "祂", "p": "tā", "sp": "ta", "r": "示", "sc": 7, "l": 5, "f": 6835},
  {"h": "轎", "p": "jiào", "sp": "jiao", "r": "車", "sc": 19, "l": 5, "f": 6836},
  {"h": "紀", "p": "jì", "sp": "ji", "r": "糸", "sc": 9, "l": 5, "f": 6837},
  {"h": "纖", "p": "xiān", "sp": "xian", "r": "糸", "sc": 23, "l": 5, "f": 6838},
  {"h": "輔", "p": "fǔ", "sp": "fu", "r": "車", "sc": 14, "l": 5, "f": 6839},
  {"h": "鉀", "p": "jiǎ", "sp": "jia", "r": "金", "sc": 13, "l": 5, "f": 6840},
  {"h": "領", "p": "lǐng", "sp": "ling", "r": "頁", "sc": 14, "l": 5, "f": 6841},
  {"h": "螵", "p": "piāo", "sp": "piao", "r": "虫", "sc": 17, "l": 5, "f": 6842},
  {"h": "鸊", "p": "pì", "sp": "pi", "r": "鳥", "sc": 24, "l": 5, "f": 6843},
  {"h": "鮡", "p": "zhào", "sp": "zhao", "r": "魚", "sc": 17, "l": 5, "f": 6844},
  {"h": "禛", "p": "zhēn", "sp": "zhen", "r": "示", "sc": 14, "l": 5, "f": 6845},
  {"h": "袆", "p": "huī", "sp": "hui", "r": "衣", "sc": 9, "l": 5, "f": 6846},
  {"h": "謴", "p": "gùn", "sp": "gun", "r": "言", "sc": 18, "l": 5, "f": 6847},
  {"h": "邾", "p": "zhū", "sp": "zhu", "r": "邑", "sc": 8, "l": 5, "f": 6848},
  {"h": "鮣", "p": "yìn", "sp": "yin", "r": "魚", "sc": 16, "l": 5, "f": 6849},
  {
    "h": "酲",
    "p": "chéng",
    "sp": "cheng",
    "r": "酉",
    "sc": 14,
    "l": 5,
    "f": 6850
  },
  {"h": "龁", "p": "hé", "sp": "he", "r": "齒", "sc": 11, "l": 5, "f": 6851},
  {"h": "汔", "p": "qì", "sp": "qi", "r": "水", "sc": 6, "l": 5, "f": 6852},
  {"h": "眄", "p": "miǎn", "sp": "mian", "r": "目", "sc": 9, "l": 5, "f": 6853},
  {"h": "眊", "p": "mào", "sp": "mao", "r": "目", "sc": 9, "l": 5, "f": 6854},
  {"h": "鬲", "p": "gé", "sp": "ge", "r": "鬲", "sc": 10, "l": 5, "f": 6855},
  {"h": "阨", "p": "è", "sp": "e", "r": "阜", "sc": 6, "l": 5, "f": 6856},
  {"h": "靸", "p": "sǎ", "sp": "sa", "r": "革", "sc": 12, "l": 5, "f": 6857},
  {"h": "臉", "p": "liǎn", "sp": "lian", "r": "肉", "sc": 17, "l": 5, "f": 6858},
  {"h": "贇", "p": "yūn", "sp": "yun", "r": "貝", "sc": 19, "l": 5, "f": 6859},
  {"h": "酺", "p": "pú", "sp": "pu", "r": "酉", "sc": 14, "l": 5, "f": 6860},
  {"h": "狃", "p": "niǔ", "sp": "niu", "r": "犬", "sc": 7, "l": 5, "f": 6861},
  {"h": "轺", "p": "yáo", "sp": "yao", "r": "車", "sc": 9, "l": 5, "f": 6862},
  {"h": "霅", "p": "zhà", "sp": "zha", "r": "雨", "sc": 15, "l": 5, "f": 6863},
  {"h": "鬂", "p": "bìn", "sp": "bin", "r": "髟", "sc": 17, "l": 5, "f": 6864},
  {"h": "鼷", "p": "xī", "sp": "xi", "r": "鼠", "sc": 23, "l": 5, "f": 6865},
  {"h": "蠃", "p": "luǒ", "sp": "luo", "r": "虫", "sc": 19, "l": 5, "f": 6866},
  {"h": "埴", "p": "zhí", "sp": "zhi", "r": "土", "sc": 11, "l": 5, "f": 6867},
  {"h": "鴥", "p": "yù", "sp": "yu", "r": "鳥", "sc": 16, "l": 5, "f": 6868},
  {"h": "隊", "p": "duì", "sp": "dui", "r": "阜", "sc": 11, "l": 5, "f": 6869},
  {"h": "髆", "p": "bó", "sp": "bo", "r": "骨", "sc": 19, "l": 5, "f": 6870},
  {"h": "綱", "p": "gāng", "sp": "gang", "r": "糸", "sc": 14, "l": 5, "f": 6871},
  {"h": "膮", "p": "xiāo", "sp": "xiao", "r": "肉", "sc": 16, "l": 5, "f": 6872},
  {"h": "謹", "p": "jǐn", "sp": "jin", "r": "言", "sc": 18, "l": 5, "f": 6873},
  {"h": "铪", "p": "hā", "sp": "ha", "r": "金", "sc": 11, "l": 5, "f": 6874},
  {"h": "鹼", "p": "jiǎn", "sp": "jian", "r": "鹵", "sc": 24, "l": 5, "f": 6875},
  {"h": "簳", "p": "gǎn", "sp": "gan", "r": "竹", "sc": 19, "l": 5, "f": 6876},
  {
    "h": "終",
    "p": "zhōng",
    "sp": "zhong",
    "r": "糸",
    "sc": 11,
    "l": 5,
    "f": 6877
  },
  {"h": "綜", "p": "zōng", "sp": "zong", "r": "糸", "sc": 14, "l": 5, "f": 6878},
  {
    "h": "裝",
    "p": "zhuāng",
    "sp": "zhuang",
    "r": "衣",
    "sc": 13,
    "l": 5,
    "f": 6879
  },
  {"h": "訉", "p": "fān", "sp": "fan", "r": "言", "sc": 10, "l": 5, "f": 6880},
  {"h": "銀", "p": "yín", "sp": "yin", "r": "金", "sc": 14, "l": 5, "f": 6881},
  {"h": "鲉", "p": "yóu", "sp": "you", "r": "魚", "sc": 13, "l": 5, "f": 6882},
  {"h": "蕿", "p": "xuān", "sp": "xuan", "r": "艸", "sc": 16, "l": 5, "f": 6883},
  {"h": "袊", "p": "lǐng", "sp": "ling", "r": "衣", "sc": 10, "l": 5, "f": 6884},
  {"h": "汆", "p": "cuān", "sp": "cuan", "r": "水", "sc": 6, "l": 5, "f": 6885},
  {"h": "硃", "p": "zhū", "sp": "zhu", "r": "石", "sc": 11, "l": 5, "f": 6886},
  {"h": "仝", "p": "tóng", "sp": "tong", "r": "人", "sc": 5, "l": 5, "f": 6887},
  {"h": "趽", "p": "fàng", "sp": "fang", "r": "足", "sc": 11, "l": 5, "f": 6888},
  {"h": "堍", "p": "tù", "sp": "tu", "r": "土", "sc": 11, "l": 5, "f": 6889},
  {"h": "揸", "p": "zhā", "sp": "zha", "r": "手", "sc": 12, "l": 5, "f": 6890},
  {"h": "忭", "p": "biàn", "sp": "bian", "r": "心", "sc": 7, "l": 5, "f": 6891},
  {
    "h": "銧",
    "p": "guāng",
    "sp": "guang",
    "r": "金",
    "sc": 14,
    "l": 5,
    "f": 6892
  },
  {"h": "妤", "p": "yú", "sp": "yu", "r": "女", "sc": 7, "l": 5, "f": 6893},
  {"h": "頫", "p": "fǔ", "sp": "fu", "r": "頁", "sc": 15, "l": 5, "f": 6894},
  {"h": "緡", "p": "mín", "sp": "min", "r": "糸", "sc": 15, "l": 5, "f": 6895},
  {"h": "讟", "p": "dú", "sp": "du", "r": "言", "sc": 29, "l": 5, "f": 6896},
  {"h": "龠", "p": "yuè", "sp": "yue", "r": "龠", "sc": 17, "l": 5, "f": 6897},
  {"h": "垤", "p": "dié", "sp": "die", "r": "土", "sc": 9, "l": 5, "f": 6898},
  {"h": "擐", "p": "huàn", "sp": "huan", "r": "手", "sc": 16, "l": 5, "f": 6899},
  {"h": "噍", "p": "jiào", "sp": "jiao", "r": "口", "sc": 15, "l": 5, "f": 6900},
  {"h": "洫", "p": "xù", "sp": "xu", "r": "水", "sc": 9, "l": 5, "f": 6901},
  {"h": "滏", "p": "fǔ", "sp": "fu", "r": "水", "sc": 13, "l": 5, "f": 6902},
  {"h": "璁", "p": "cōng", "sp": "cong", "r": "玉", "sc": 15, "l": 5, "f": 6903},
  {"h": "癃", "p": "lóng", "sp": "long", "r": "疒", "sc": 16, "l": 5, "f": 6904},
  {"h": "邡", "p": "fāng", "sp": "fang", "r": "邑", "sc": 6, "l": 5, "f": 6905},
  {"h": "絁", "p": "shī", "sp": "shi", "r": "糸", "sc": 11, "l": 5, "f": 6906},
  {"h": "鍩", "p": "tiǎn", "sp": "tian", "r": "金", "sc": 16, "l": 5, "f": 6907},
  {"h": "趨", "p": "qū", "sp": "qu", "r": "走", "sc": 17, "l": 5, "f": 6908},
  {"h": "蒽", "p": "ēn", "sp": "en", "r": "艸", "sc": 13, "l": 5, "f": 6909},
  {"h": "絲", "p": "sī", "sp": "si", "r": "糸", "sc": 12, "l": 5, "f": 6910},
  {"h": "膫", "p": "liáo", "sp": "liao", "r": "肉", "sc": 16, "l": 5, "f": 6911},
  {"h": "臅", "p": "chù", "sp": "chu", "r": "肉", "sc": 17, "l": 5, "f": 6912},
  {"h": "茖", "p": "gé", "sp": "ge", "r": "艸", "sc": 9, "l": 5, "f": 6913},
  {"h": "蘇", "p": "sū", "sp": "su", "r": "艸", "sc": 19, "l": 5, "f": 6914},
  {"h": "詞", "p": "cí", "sp": "ci", "r": "言", "sc": 12, "l": 5, "f": 6915},
  {
    "h": "詾",
    "p": "xiōng",
    "sp": "xiong",
    "r": "言",
    "sc": 13,
    "l": 5,
    "f": 6916
  },
  {"h": "贆", "p": "biāo", "sp": "biao", "r": "貝", "sc": 19, "l": 5, "f": 6917},
  {"h": "蹨", "p": "niǎn", "sp": "nian", "r": "足", "sc": 19, "l": 5, "f": 6918},
  {"h": "農", "p": "nóng", "sp": "nong", "r": "辰", "sc": 13, "l": 5, "f": 6919},
  {"h": "錄", "p": "lù", "sp": "lu", "r": "金", "sc": 16, "l": 5, "f": 6920},
  {
    "h": "鐘",
    "p": "zhōng",
    "sp": "zhong",
    "r": "金",
    "sc": 20,
    "l": 5,
    "f": 6921
  },
  {"h": "鰜", "p": "qiàn", "sp": "qian", "r": "魚", "sc": 21, "l": 5, "f": 6922},
  {"h": "鹽", "p": "yán", "sp": "yan", "r": "鹵", "sc": 24, "l": 5, "f": 6923},
  {"h": "赕", "p": "dǎn", "sp": "dan", "r": "貝", "sc": 12, "l": 5, "f": 6924},
  {"h": "鯔", "p": "zī", "sp": "zi", "r": "魚", "sc": 19, "l": 5, "f": 6925},
  {"h": "鯮", "p": "zōng", "sp": "zong", "r": "魚", "sc": 19, "l": 5, "f": 6926},
  {"h": "蔯", "p": "chén", "sp": "chen", "r": "艸", "sc": 13, "l": 5, "f": 6927},
  {"h": "貙", "p": "chū", "sp": "chu", "r": "豸", "sc": 18, "l": 5, "f": 6928},
  {"h": "餬", "p": "hú", "sp": "hu", "r": "食", "sc": 17, "l": 5, "f": 6929},
  {"h": "躜", "p": "zuān", "sp": "zuan", "r": "足", "sc": 23, "l": 5, "f": 6930},
  {"h": "鳀", "p": "tí", "sp": "ti", "r": "魚", "sc": 17, "l": 5, "f": 6931},
  {"h": "鹡", "p": "jí", "sp": "ji", "r": "鳥", "sc": 15, "l": 5, "f": 6932},
  {"h": "䴘", "p": "tī", "sp": "ti", "r": "", "sc": 21, "l": 5, "f": 6933},
  {"h": "蠋", "p": "zhú", "sp": "zhu", "r": "虫", "sc": 19, "l": 5, "f": 6934},
  {
    "h": "钲",
    "p": "zhēng",
    "sp": "zheng",
    "r": "金",
    "sc": 10,
    "l": 5,
    "f": 6935
  },
  {"h": "蜁", "p": "xuán", "sp": "xuan", "r": "虫", "sc": 13, "l": 5, "f": 6936},
  {"h": "兕", "p": "sì", "sp": "si", "r": "儿", "sc": 7, "l": 5, "f": 6937},
  {"h": "阽", "p": "diàn", "sp": "dian", "r": "阜", "sc": 7, "l": 5, "f": 6938},
  {"h": "鎰", "p": "yì", "sp": "yi", "r": "金", "sc": 18, "l": 5, "f": 6939},
  {"h": "绋", "p": "fú", "sp": "fu", "r": "糸", "sc": 8, "l": 5, "f": 6940},
  {"h": "铓", "p": "máng", "sp": "mang", "r": "金", "sc": 11, "l": 5, "f": 6941},
  {"h": "爝", "p": "jué", "sp": "jue", "r": "火", "sc": 21, "l": 5, "f": 6942},
  {"h": "磾", "p": "dī", "sp": "di", "r": "石", "sc": 17, "l": 5, "f": 6943},
  {"h": "葷", "p": "hūn", "sp": "hun", "r": "艸", "sc": 12, "l": 5, "f": 6944},
  {"h": "螘", "p": "yǐ", "sp": "yi", "r": "虫", "sc": 16, "l": 5, "f": 6945},
  {"h": "仂", "p": "lè", "sp": "le", "r": "人", "sc": 4, "l": 5, "f": 6946},
  {"h": "雲", "p": "yún", "sp": "yun", "r": "雨", "sc": 12, "l": 5, "f": 6947},
  {"h": "顗", "p": "yǐ", "sp": "yi", "r": "頁", "sc": 19, "l": 5, "f": 6948},
  {"h": "蟫", "p": "yín", "sp": "yin", "r": "虫", "sc": 18, "l": 5, "f": 6949},
  {"h": "蕺", "p": "jí", "sp": "ji", "r": "艸", "sc": 15, "l": 5, "f": 6950},
  {"h": "骃", "p": "yīn", "sp": "yin", "r": "馬", "sc": 9, "l": 5, "f": 6951},
  {"h": "蔀", "p": "bù", "sp": "bu", "r": "艸", "sc": 13, "l": 5, "f": 6952},
  {"h": "诪", "p": "zhōu", "sp": "zhou", "r": "言", "sc": 9, "l": 5, "f": 6953},
  {"h": "猹", "p": "chá", "sp": "cha", "r": "犬", "sc": 12, "l": 5, "f": 6954},
  {"h": "鋬", "p": "pàn", "sp": "pan", "r": "金", "sc": 15, "l": 5, "f": 6955},
  {"h": "盤", "p": "pán", "sp": "pan", "r": "皿", "sc": 15, "l": 5, "f": 6956},
  {"h": "齾", "p": "yà", "sp": "ya", "r": "齒", "sc": 35, "l": 5, "f": 6957},
  {
    "h": "埕",
    "p": "chéng",
    "sp": "cheng",
    "r": "土",
    "sc": 10,
    "l": 5,
    "f": 6958
  },
  {"h": "聞", "p": "wén", "sp": "wen", "r": "耳", "sc": 14, "l": 5, "f": 6959},
  {"h": "舉", "p": "jǔ", "sp": "ju", "r": "臼", "sc": 16, "l": 5, "f": 6960},
  {"h": "虡", "p": "jù", "sp": "ju", "r": "虍", "sc": 13, "l": 5, "f": 6961},
  {"h": "評", "p": "píng", "sp": "ping", "r": "言", "sc": 12, "l": 5, "f": 6962},
  {"h": "鍋", "p": "guō", "sp": "guo", "r": "金", "sc": 16, "l": 5, "f": 6963},
  {"h": "陣", "p": "zhèn", "sp": "zhen", "r": "阜", "sc": 9, "l": 5, "f": 6964},
  {"h": "紬", "p": "chóu", "sp": "chou", "r": "糸", "sc": 11, "l": 5, "f": 6965},
  {"h": "訢", "p": "xīn", "sp": "xin", "r": "言", "sc": 11, "l": 5, "f": 6966},
  {"h": "蔎", "p": "shè", "sp": "she", "r": "艸", "sc": 14, "l": 5, "f": 6967},
  {"h": "繸", "p": "suì", "sp": "sui", "r": "糸", "sc": 18, "l": 5, "f": 6968},
  {"h": "绖", "p": "dié", "sp": "die", "r": "糸", "sc": 9, "l": 5, "f": 6969},
  {"h": "埭", "p": "dài", "sp": "dai", "r": "土", "sc": 11, "l": 5, "f": 6970},
  {"h": "駮", "p": "bó", "sp": "bo", "r": "馬", "sc": 16, "l": 5, "f": 6971},
  {"h": "鰡", "p": "liú", "sp": "liu", "r": "魚", "sc": 21, "l": 5, "f": 6972},
  {"h": "鵳", "p": "jiān", "sp": "jian", "r": "鳥", "sc": 19, "l": 5, "f": 6973},
  {"h": "柁", "p": "duò", "sp": "duo", "r": "木", "sc": 9, "l": 5, "f": 6974},
  {"h": "筴", "p": "cè", "sp": "ce", "r": "竹", "sc": 13, "l": 5, "f": 6975},
  {"h": "珧", "p": "yáo", "sp": "yao", "r": "玉", "sc": 10, "l": 5, "f": 6976},
  {"h": "贳", "p": "shì", "sp": "shi", "r": "貝", "sc": 9, "l": 5, "f": 6977},
  {"h": "贶", "p": "kuàng", "sp": "kuang", "r": "貝", "sc": 9, "l": 5, "f": 6978},
  {"h": "腩", "p": "nǎn", "sp": "nan", "r": "肉", "sc": 13, "l": 5, "f": 6979},
  {"h": "臌", "p": "gǔ", "sp": "gu", "r": "肉", "sc": 17, "l": 5, "f": 6980},
  {
    "h": "礓",
    "p": "jiāng",
    "sp": "jiang",
    "r": "石",
    "sc": 18,
    "l": 5,
    "f": 6981
  },
  {"h": "袛", "p": "dī", "sp": "di", "r": "衣", "sc": 10, "l": 5, "f": 6982},
  {"h": "貐", "p": "yǔ", "sp": "yu", "r": "豸", "sc": 16, "l": 5, "f": 6983},
  {"h": "蹐", "p": "jí", "sp": "ji", "r": "足", "sc": 17, "l": 5, "f": 6984},
  {"h": "圬", "p": "wū", "sp": "wu", "r": "土", "sc": 6, "l": 5, "f": 6985},
  {"h": "堠", "p": "hòu", "sp": "hou", "r": "土", "sc": 12, "l": 5, "f": 6986},
  {"h": "屦", "p": "jù", "sp": "ju", "r": "尸", "sc": 15, "l": 5, "f": 6987},
  {"h": "驵", "p": "zǎng", "sp": "zang", "r": "馬", "sc": 8, "l": 5, "f": 6988},
  {"h": "鑑", "p": "jiàn", "sp": "jian", "r": "金", "sc": 22, "l": 5, "f": 6989},
  {
    "h": "桄",
    "p": "guāng",
    "sp": "guang",
    "r": "木",
    "sc": 10,
    "l": 5,
    "f": 6990
  },
  {"h": "霑", "p": "zhān", "sp": "zhan", "r": "雨", "sc": 16, "l": 5, "f": 6991},
  {"h": "憝", "p": "duì", "sp": "dui", "r": "心", "sc": 16, "l": 5, "f": 6992},
  {"h": "瘳", "p": "chōu", "sp": "chou", "r": "疒", "sc": 16, "l": 5, "f": 6993},
  {"h": "騃", "p": "ái", "sp": "ai", "r": "馬", "sc": 17, "l": 5, "f": 6994},
  {"h": "茌", "p": "chí", "sp": "chi", "r": "艸", "sc": 8, "l": 5, "f": 6995},
  {"h": "辀", "p": "zhōu", "sp": "zhou", "r": "車", "sc": 10, "l": 5, "f": 6996},
  {"h": "胠", "p": "qū", "sp": "qu", "r": "肉", "sc": 9, "l": 5, "f": 6997},
  {"h": "埏", "p": "shān", "sp": "shan", "r": "土", "sc": 9, "l": 5, "f": 6998},
  {"h": "腠", "p": "còu", "sp": "cou", "r": "肉", "sc": 13, "l": 5, "f": 6999},
  {"h": "靰", "p": "wù", "sp": "wu", "r": "革", "sc": 12, "l": 5, "f": 7000},
  {"h": "鞡", "p": "la", "sp": "la", "r": "革", "sc": 17, "l": 5, "f": 7001},
  {"h": "缻", "p": "fǒu", "sp": "fou", "r": "缶", "sc": 10, "l": 5, "f": 7002},
  {"h": "輳", "p": "còu", "sp": "cou", "r": "車", "sc": 16, "l": 5, "f": 7003},
  {"h": "闆", "p": "bǎn", "sp": "ban", "r": "門", "sc": 17, "l": 5, "f": 7004},
  {"h": "铼", "p": "lái", "sp": "lai", "r": "金", "sc": 12, "l": 5, "f": 7005},
  {"h": "紤", "p": "jiǔ", "sp": "jiu", "r": "糸", "sc": 10, "l": 5, "f": 7006},
  {"h": "翹", "p": "qiào", "sp": "qiao", "r": "羽", "sc": 18, "l": 5, "f": 7007},
  {"h": "聯", "p": "lián", "sp": "lian", "r": "耳", "sc": 17, "l": 5, "f": 7008},
  {"h": "詢", "p": "xún", "sp": "xun", "r": "言", "sc": 13, "l": 5, "f": 7009},
  {"h": "課", "p": "kè", "sp": "ke", "r": "言", "sc": 15, "l": 5, "f": 7010},
  {"h": "豐", "p": "fēng", "sp": "feng", "r": "豆", "sc": 18, "l": 5, "f": 7011},
  {"h": "賴", "p": "lài", "sp": "lai", "r": "貝", "sc": 16, "l": 5, "f": 7012},
  {"h": "輸", "p": "shū", "sp": "shu", "r": "車", "sc": 16, "l": 5, "f": 7013},
  {"h": "遲", "p": "chí", "sp": "chi", "r": "辵", "sc": 15, "l": 5, "f": 7014},
  {
    "h": "鐣",
    "p": "zhang",
    "sp": "zhang",
    "r": "金",
    "sc": 20,
    "l": 5,
    "f": 7015
  },
  {"h": "齡", "p": "líng", "sp": "ling", "r": "齒", "sc": 20, "l": 5, "f": 7016},
  {"h": "莤", "p": "sù", "sp": "su", "r": "艸", "sc": 10, "l": 5, "f": 7017},
  {"h": "鱯", "p": "hù", "sp": "hu", "r": "魚", "sc": 24, "l": 5, "f": 7018},
  {"h": "缽", "p": "bō", "sp": "bo", "r": "缶", "sc": 11, "l": 5, "f": 7019},
  {"h": "邲", "p": "bì", "sp": "bi", "r": "邑", "sc": 7, "l": 5, "f": 7020},
  {"h": "唼", "p": "shà", "sp": "sha", "r": "口", "sc": 11, "l": 5, "f": 7021},
  {"h": "鐻", "p": "jù", "sp": "ju", "r": "金", "sc": 21, "l": 5, "f": 7022},
  {"h": "閟", "p": "bì", "sp": "bi", "r": "門", "sc": 13, "l": 5, "f": 7023},
  {"h": "駚", "p": "yǎng", "sp": "yang", "r": "馬", "sc": 15, "l": 5, "f": 7024},
  {"h": "篥", "p": "lì", "sp": "li", "r": "竹", "sc": 16, "l": 5, "f": 7025},
  {"h": "茈", "p": "cí", "sp": "ci", "r": "艸", "sc": 9, "l": 5, "f": 7026},
  {"h": "肏", "p": "cào", "sp": "cao", "r": "肉", "sc": 8, "l": 5, "f": 7027},
  {"h": "萩", "p": "qiū", "sp": "qiu", "r": "艸", "sc": 12, "l": 5, "f": 7028},
  {"h": "跐", "p": "cī", "sp": "ci", "r": "足", "sc": 13, "l": 5, "f": 7029},
  {"h": "蹅", "p": "chǎ", "sp": "cha", "r": "足", "sc": 16, "l": 5, "f": 7030},
  {"h": "瞢", "p": "méng", "sp": "meng", "r": "目", "sc": 15, "l": 5, "f": 7031},
  {"h": "罽", "p": "jì", "sp": "ji", "r": "网", "sc": 17, "l": 5, "f": 7032},
  {"h": "聭", "p": "kuì", "sp": "kui", "r": "耳", "sc": 15, "l": 5, "f": 7033},
  {"h": "辻", "p": "shí", "sp": "shi", "r": "辵", "sc": 5, "l": 5, "f": 7034},
  {"h": "铚", "p": "zhì", "sp": "zhi", "r": "金", "sc": 11, "l": 5, "f": 7035},
  {"h": "箦", "p": "zé", "sp": "ze", "r": "竹", "sc": 14, "l": 5, "f": 7036},
  {"h": "踣", "p": "bó", "sp": "bo", "r": "足", "sc": 15, "l": 5, "f": 7037},
  {"h": "菑", "p": "zāi", "sp": "zai", "r": "艸", "sc": 11, "l": 5, "f": 7038},
  {"h": "蹟", "p": "jī", "sp": "ji", "r": "足", "sc": 18, "l": 5, "f": 7039},
  {"h": "捃", "p": "jùn", "sp": "jun", "r": "手", "sc": 10, "l": 5, "f": 7040},
  {"h": "瘕", "p": "jiǎ", "sp": "jia", "r": "疒", "sc": 14, "l": 5, "f": 7041},
  {"h": "睺", "p": "hóu", "sp": "hou", "r": "目", "sc": 14, "l": 5, "f": 7042},
  {"h": "閒", "p": "xián", "sp": "xian", "r": "門", "sc": 12, "l": 5, "f": 7043},
  {"h": "鈦", "p": "tài", "sp": "tai", "r": "金", "sc": 12, "l": 5, "f": 7044},
  {"h": "顩", "p": "yǎn", "sp": "yan", "r": "頁", "sc": 22, "l": 5, "f": 7045},
  {"h": "鳓", "p": "lè", "sp": "le", "r": "魚", "sc": 19, "l": 5, "f": 7046},
  {"h": "闇", "p": "àn", "sp": "an", "r": "門", "sc": 17, "l": 5, "f": 7047},
  {"h": "鹹", "p": "xián", "sp": "xian", "r": "鹵", "sc": 20, "l": 5, "f": 7048},
  {"h": "䥇", "p": "shàn", "sp": "shan", "r": "金", "sc": 18, "l": 5, "f": 7049},
  {"h": "禍", "p": "huò", "sp": "huo", "r": "示", "sc": 12, "l": 5, "f": 7050},
  {"h": "窩", "p": "wō", "sp": "wo", "r": "穴", "sc": 13, "l": 5, "f": 7051},
  {"h": "純", "p": "chún", "sp": "chun", "r": "糸", "sc": 10, "l": 5, "f": 7052},
  {"h": "縱", "p": "zòng", "sp": "zong", "r": "糸", "sc": 17, "l": 5, "f": 7053},
  {
    "h": "聖",
    "p": "shèng",
    "sp": "sheng",
    "r": "耳",
    "sc": 13,
    "l": 5,
    "f": 7054
  },
  {"h": "聛", "p": "bǐ", "sp": "bi", "r": "耳", "sc": 14, "l": 5, "f": 7055},
  {"h": "脮", "p": "něi", "sp": "nei", "r": "肉", "sc": 11, "l": 5, "f": 7056},
  {
    "h": "衝",
    "p": "chōng",
    "sp": "chong",
    "r": "行",
    "sc": 15,
    "l": 5,
    "f": 7057
  },
  {"h": "訊", "p": "xùn", "sp": "xun", "r": "言", "sc": 10, "l": 5, "f": 7058},
  {"h": "馬", "p": "mǎ", "sp": "ma", "r": "馬", "sc": 10, "l": 5, "f": 7059},
  {"h": "諹", "p": "yáng", "sp": "yang", "r": "言", "sc": 16, "l": 5, "f": 7060},
  {"h": "誖", "p": "bèi", "sp": "bei", "r": "言", "sc": 14, "l": 5, "f": 7061},
  {"h": "蕅", "p": "ǒu", "sp": "ou", "r": "艸", "sc": 15, "l": 5, "f": 7062},
  {"h": "鰨", "p": "tǎ", "sp": "ta", "r": "魚", "sc": 21, "l": 5, "f": 7063},
  {"h": "睒", "p": "shǎn", "sp": "shan", "r": "目", "sc": 13, "l": 5, "f": 7064},
  {"h": "笰", "p": "fú", "sp": "fu", "r": "竹", "sc": 11, "l": 5, "f": 7065},
  {"h": "籥", "p": "yuè", "sp": "yue", "r": "竹", "sc": 23, "l": 5, "f": 7066},
  {"h": "翧", "p": "xuān", "sp": "xuan", "r": "羽", "sc": 15, "l": 5, "f": 7067},
  {"h": "镩", "p": "cuān", "sp": "cuan", "r": "金", "sc": 17, "l": 5, "f": 7068},
  {"h": "耧", "p": "lóu", "sp": "lou", "r": "耒", "sc": 15, "l": 5, "f": 7069},
  {"h": "蜞", "p": "qí", "sp": "qi", "r": "虫", "sc": 14, "l": 5, "f": 7070},
  {"h": "舡", "p": "chuán", "sp": "chuan", "r": "舟", "sc": 9, "l": 5, "f": 7071},
  {"h": "鰲", "p": "áo", "sp": "ao", "r": "魚", "sc": 21, "l": 5, "f": 7072},
  {"h": "鹀", "p": "wú", "sp": "wu", "r": "鳥", "sc": 12, "l": 5, "f": 7073},
  {"h": "䴙", "p": "pì", "sp": "pi", "r": "", "sc": 24, "l": 5, "f": 7074},
  {"h": "罣", "p": "guà", "sp": "gua", "r": "网", "sc": 11, "l": 5, "f": 7075},
  {"h": "觍", "p": "tiǎn", "sp": "tian", "r": "見", "sc": 12, "l": 5, "f": 7076},
  {"h": "絽", "p": "lǚ", "sp": "lu", "r": "糸", "sc": 12, "l": 5, "f": 7077},
  {"h": "羶", "p": "shān", "sp": "shan", "r": "羊", "sc": 19, "l": 5, "f": 7078},
  {"h": "臜", "p": "zā", "sp": "za", "r": "肉", "sc": 20, "l": 5, "f": 7079},
  {"h": "薙", "p": "tì", "sp": "ti", "r": "艸", "sc": 16, "l": 5, "f": 7080},
  {"h": "觔", "p": "jīn", "sp": "jin", "r": "角", "sc": 9, "l": 5, "f": 7081},
  {"h": "躼", "p": "lào", "sp": "lao", "r": "身", "sc": 15, "l": 5, "f": 7082},
  {"h": "岈", "p": "yá", "sp": "ya", "r": "山", "sc": 7, "l": 5, "f": 7083},
  {"h": "潲", "p": "shào", "sp": "shao", "r": "水", "sc": 15, "l": 5, "f": 7084},
  {"h": "枧", "p": "jiǎn", "sp": "jian", "r": "木", "sc": 8, "l": 5, "f": 7085},
  {"h": "觌", "p": "dí", "sp": "di", "r": "見", "sc": 12, "l": 5, "f": 7086},
  {"h": "膦", "p": "lìn", "sp": "lin", "r": "肉", "sc": 16, "l": 5, "f": 7087},
  {"h": "礞", "p": "méng", "sp": "meng", "r": "石", "sc": 18, "l": 5, "f": 7088},
  {"h": "蝤", "p": "qiú", "sp": "qiu", "r": "虫", "sc": 15, "l": 5, "f": 7089},
  {"h": "螗", "p": "táng", "sp": "tang", "r": "虫", "sc": 16, "l": 5, "f": 7090},
  {"h": "酹", "p": "lèi", "sp": "lei", "r": "酉", "sc": 14, "l": 5, "f": 7091},
  {"h": "醅", "p": "pēi", "sp": "pei", "r": "酉", "sc": 15, "l": 5, "f": 7092},
  {
    "h": "絅",
    "p": "jiōng",
    "sp": "jiong",
    "r": "糸",
    "sc": 11,
    "l": 5,
    "f": 7093
  },
  {"h": "纁", "p": "xūn", "sp": "xun", "r": "糸", "sc": 20, "l": 5, "f": 7094},
  {"h": "葰", "p": "suī", "sp": "sui", "r": "艸", "sc": 12, "l": 5, "f": 7095},
  {"h": "郇", "p": "huán", "sp": "huan", "r": "邑", "sc": 8, "l": 5, "f": 7096},
  {"h": "躄", "p": "bì", "sp": "bi", "r": "足", "sc": 20, "l": 5, "f": 7097},
  {"h": "茼", "p": "tóng", "sp": "tong", "r": "艸", "sc": 9, "l": 5, "f": 7098},
  {"h": "洧", "p": "wěi", "sp": "wei", "r": "水", "sc": 9, "l": 5, "f": 7099},
  {"h": "滠", "p": "shè", "sp": "she", "r": "水", "sc": 13, "l": 5, "f": 7100},
  {"h": "澉", "p": "gǎn", "sp": "gan", "r": "水", "sc": 14, "l": 5, "f": 7101},
  {"h": "鎭", "p": "zhèn", "sp": "zhen", "r": "金", "sc": 18, "l": 5, "f": 7102},
  {"h": "铏", "p": "xíng", "sp": "xing", "r": "金", "sc": 11, "l": 5, "f": 7103},
  {"h": "閤", "p": "gé", "sp": "ge", "r": "門", "sc": 14, "l": 5, "f": 7104},
  {"h": "赆", "p": "jìn", "sp": "jin", "r": "貝", "sc": 10, "l": 5, "f": 7105},
  {"h": "赇", "p": "qiú", "sp": "qiu", "r": "貝", "sc": 11, "l": 5, "f": 7106},
  {"h": "顣", "p": "cù", "sp": "cu", "r": "頁", "sc": 20, "l": 5, "f": 7107},
  {"h": "骕", "p": "sù", "sp": "su", "r": "馬", "sc": 11, "l": 5, "f": 7108},
  {"h": "飑", "p": "biāo", "sp": "biao", "r": "風", "sc": 9, "l": 5, "f": 7109},
  {"h": "訄", "p": "qiú", "sp": "qiu", "r": "言", "sc": 9, "l": 5, "f": 7110},
  {"h": "輓", "p": "wǎn", "sp": "wan", "r": "車", "sc": 14, "l": 5, "f": 7111},
  {"h": "鉤", "p": "gōu", "sp": "gou", "r": "金", "sc": 13, "l": 5, "f": 7112},
  {
    "h": "鋹",
    "p": "chǎng",
    "sp": "chang",
    "r": "金",
    "sc": 16,
    "l": 5,
    "f": 7113
  },
  {"h": "钘", "p": "xíng", "sp": "xing", "r": "金", "sc": 9, "l": 5, "f": 7114},
  {"h": "陁", "p": "tuó", "sp": "tuo", "r": "阜", "sc": 7, "l": 5, "f": 7115},
  {"h": "耱", "p": "mò", "sp": "mo", "r": "耒", "sc": 22, "l": 5, "f": 7116},
  {"h": "繫", "p": "xì", "sp": "xi", "r": "糸", "sc": 19, "l": 5, "f": 7117},
  {"h": "鋵", "p": "tū", "sp": "tu", "r": "金", "sc": 15, "l": 5, "f": 7118},
  {"h": "鱵", "p": "zhēn", "sp": "zhen", "r": "魚", "sc": 26, "l": 5, "f": 7119},
  {"h": "黨", "p": "dǎng", "sp": "dang", "r": "黑", "sc": 20, "l": 5, "f": 7120},
  {"h": "禿", "p": "tū", "sp": "tu", "r": "禾", "sc": 7, "l": 5, "f": 7121},
  {"h": "貧", "p": "pín", "sp": "pin", "r": "貝", "sc": 11, "l": 5, "f": 7122},
  {"h": "買", "p": "mǎi", "sp": "mai", "r": "貝", "sc": 12, "l": 5, "f": 7123},
  {"h": "賈", "p": "jiǎ", "sp": "jia", "r": "貝", "sc": 13, "l": 5, "f": 7124},
  {"h": "榀", "p": "pǐn", "sp": "pin", "r": "木", "sc": 13, "l": 5, "f": 7125},
  {"h": "緣", "p": "yuán", "sp": "yuan", "r": "糸", "sc": 15, "l": 5, "f": 7126},
  {"h": "膶", "p": "rùn", "sp": "run", "r": "月", "sc": 16, "l": 5, "f": 7127},
  {"h": "茐", "p": "cong", "sp": "cong", "r": "艸", "sc": 8, "l": 5, "f": 7128},
  {"h": "螝", "p": "guī", "sp": "gui", "r": "虫", "sc": 15, "l": 5, "f": 7129},
  {"h": "訙", "p": "xùn", "sp": "xun", "r": "言", "sc": 10, "l": 5, "f": 7130},
  {"h": "誘", "p": "yòu", "sp": "you", "r": "言", "sc": 14, "l": 5, "f": 7131},
  {"h": "賦", "p": "fù", "sp": "fu", "r": "貝", "sc": 15, "l": 5, "f": 7132},
  {"h": "贅", "p": "zhuì", "sp": "zhui", "r": "貝", "sc": 17, "l": 5, "f": 7133},
  {"h": "鄬", "p": "wéi", "sp": "wei", "r": "邑", "sc": 14, "l": 5, "f": 7134},
  {"h": "釤", "p": "shàn", "sp": "shan", "r": "金", "sc": 11, "l": 5, "f": 7135},
  {"h": "鍵", "p": "jiàn", "sp": "jian", "r": "金", "sc": 16, "l": 5, "f": 7136},
  {"h": "閮", "p": "tíng", "sp": "ting", "r": "門", "sc": 14, "l": 5, "f": 7137},
  {"h": "隻", "p": "zhī", "sp": "zhi", "r": "隹", "sc": 10, "l": 5, "f": 7138},
  {"h": "飽", "p": "bǎo", "sp": "bao", "r": "食", "sc": 13, "l": 5, "f": 7139},
  {"h": "齒", "p": "chǐ", "sp": "chi", "r": "齒", "sc": 15, "l": 5, "f": 7140},
  {"h": "鯆", "p": "pū", "sp": "pu", "r": "魚", "sc": 18, "l": 5, "f": 7141},
  {"h": "逿", "p": "dàng", "sp": "dang", "r": "辵", "sc": 12, "l": 5, "f": 7142},
  {"h": "鮀", "p": "tuó", "sp": "tuo", "r": "魚", "sc": 16, "l": 5, "f": 7143},
  {"h": "鯝", "p": "gù", "sp": "gu", "r": "魚", "sc": 19, "l": 5, "f": 7144},
  {"h": "鰁", "p": "quán", "sp": "quan", "r": "魚", "sc": 20, "l": 5, "f": 7145},
  {"h": "鱤", "p": "gǎn", "sp": "gan", "r": "魚", "sc": 24, "l": 5, "f": 7146},
  {"h": "郕", "p": "chéng", "sp": "cheng", "r": "邑", "sc": 8, "l": 5, "f": 7147},
  {"h": "镴", "p": "là", "sp": "la", "r": "金", "sc": 20, "l": 5, "f": 7148},
  {"h": "韻", "p": "yùn", "sp": "yun", "r": "音", "sc": 19, "l": 5, "f": 7149},
  {"h": "鮄", "p": "fú", "sp": "fu", "r": "魚", "sc": 16, "l": 5, "f": 7150},
  {"h": "鲙", "p": "kuài", "sp": "kuai", "r": "魚", "sc": 14, "l": 5, "f": 7151},
  {"h": "鳁", "p": "wēn", "sp": "wen", "r": "魚", "sc": 17, "l": 5, "f": 7152},
  {"h": "齁", "p": "hōu", "sp": "hou", "r": "鼻", "sc": 19, "l": 5, "f": 7153},
  {"h": "舖", "p": "pù", "sp": "pu", "r": "舌", "sc": 15, "l": 5, "f": 7154},
  {"h": "膹", "p": "fèn", "sp": "fen", "r": "肉", "sc": 16, "l": 5, "f": 7155},
  {"h": "袣", "p": "yì", "sp": "yi", "r": "衣", "sc": 10, "l": 5, "f": 7156},
  {"h": "謿", "p": "cháo", "sp": "chao", "r": "言", "sc": 19, "l": 5, "f": 7157},
  {"h": "陧", "p": "niè", "sp": "nie", "r": "阜", "sc": 9, "l": 5, "f": 7158},
  {"h": "坜", "p": "lì", "sp": "li", "r": "土", "sc": 7, "l": 5, "f": 7159},
  {"h": "坫", "p": "diàn", "sp": "dian", "r": "土", "sc": 8, "l": 5, "f": 7160},
  {"h": "甍", "p": "méng", "sp": "meng", "r": "瓦", "sc": 14, "l": 5, "f": 7161},
  {"h": "绠", "p": "gěng", "sp": "geng", "r": "糸", "sc": 10, "l": 5, "f": 7162},
  {"h": "恝", "p": "jiá", "sp": "jia", "r": "心", "sc": 10, "l": 5, "f": 7163},
  {"h": "飗", "p": "liú", "sp": "liu", "r": "風", "sc": 14, "l": 5, "f": 7164},
  {"h": "镙", "p": "luó", "sp": "luo", "r": "金", "sc": 16, "l": 5, "f": 7165},
  {"h": "镱", "p": "yì", "sp": "yi", "r": "金", "sc": 18, "l": 5, "f": 7166},
  {"h": "騾", "p": "luó", "sp": "luo", "r": "馬", "sc": 21, "l": 5, "f": 7167},
  {"h": "魋", "p": "tuí", "sp": "tui", "r": "鬼", "sc": 17, "l": 5, "f": 7168},
  {"h": "硱", "p": "kǔn", "sp": "kun", "r": "石", "sc": 12, "l": 5, "f": 7169},
  {"h": "祫", "p": "xiá", "sp": "xia", "r": "示", "sc": 10, "l": 5, "f": 7170},
  {"h": "筍", "p": "sǔn", "sp": "sun", "r": "竹", "sc": 12, "l": 5, "f": 7171},
  {"h": "箖", "p": "lín", "sp": "lin", "r": "竹", "sc": 14, "l": 5, "f": 7172},
  {"h": "袧", "p": "gōu", "sp": "gou", "r": "衣", "sc": 10, "l": 5, "f": 7173},
  {"h": "迯", "p": "táo", "sp": "tao", "r": "辵", "sc": 8, "l": 5, "f": 7174},
  {"h": "郙", "p": "fǔ", "sp": "fu", "r": "邑", "sc": 9, "l": 5, "f": 7175},
  {"h": "酖", "p": "zhèn", "sp": "zhen", "r": "酉", "sc": 11, "l": 5, "f": 7176},
  {"h": "遘", "p": "gòu", "sp": "gou", "r": "辵", "sc": 13, "l": 5, "f": 7177},
  {"h": "阘", "p": "dá", "sp": "da", "r": "門", "sc": 13, "l": 5, "f": 7178},
  {"h": "眭", "p": "suī", "sp": "sui", "r": "目", "sc": 11, "l": 5, "f": 7179},
  {"h": "餗", "p": "sù", "sp": "su", "r": "食", "sc": 15, "l": 5, "f": 7180},
  {"h": "耖", "p": "chào", "sp": "chao", "r": "耒", "sc": 10, "l": 5, "f": 7181},
  {"h": "鯒", "p": "yǒng", "sp": "yong", "r": "魚", "sc": 18, "l": 5, "f": 7182},
  {"h": "邶", "p": "bèi", "sp": "bei", "r": "邑", "sc": 7, "l": 5, "f": 7183},
  {"h": "絼", "p": "zhěn", "sp": "zhen", "r": "糸", "sc": 13, "l": 5, "f": 7184},
  {"h": "颥", "p": "rú", "sp": "ru", "r": "頁", "sc": 20, "l": 5, "f": 7185},
  {"h": "篚", "p": "fěi", "sp": "fei", "r": "竹", "sc": 16, "l": 5, "f": 7186},
  {"h": "鳘", "p": "mǐn", "sp": "min", "r": "魚", "sc": 19, "l": 5, "f": 7187},
  {"h": "䥺", "p": "yé", "sp": "ye", "r": "", "sc": 12, "l": 5, "f": 7188},
  {"h": "虒", "p": "sī", "sp": "si", "r": "虍", "sc": 10, "l": 5, "f": 7189},
  {"h": "裈", "p": "kūn", "sp": "kun", "r": "衣", "sc": 11, "l": 5, "f": 7190},
  {"h": "遆", "p": "tí", "sp": "ti", "r": "辵", "sc": 12, "l": 5, "f": 7191},
  {"h": "麑", "p": "ní", "sp": "ni", "r": "鹿", "sc": 19, "l": 5, "f": 7192},
  {"h": "粚", "p": "lí", "sp": "li", "r": "米", "sc": 11, "l": 5, "f": 7193},
  {"h": "託", "p": "tuō", "sp": "tuo", "r": "言", "sc": 10, "l": 5, "f": 7194},
  {"h": "鄚", "p": "mào", "sp": "mao", "r": "邑", "sc": 12, "l": 5, "f": 7195},
  {"h": "軌", "p": "guǐ", "sp": "gui", "r": "車", "sc": 9, "l": 5, "f": 7196},
  {"h": "饋", "p": "kuì", "sp": "kui", "r": "食", "sc": 20, "l": 5, "f": 7197},
  {"h": "銳", "p": "ruì", "sp": "rui", "r": "金", "sc": 15, "l": 5, "f": 7198},
  {"h": "粌", "p": "yin", "sp": "yin", "r": "米", "sc": 10, "l": 5, "f": 7199},
  {"h": "紙", "p": "zhǐ", "sp": "zhi", "r": "糸", "sc": 10, "l": 5, "f": 7200},
  {"h": "職", "p": "zhí", "sp": "zhi", "r": "耳", "sc": 18, "l": 5, "f": 7201},
  {"h": "臏", "p": "bìn", "sp": "bin", "r": "肉", "sc": 18, "l": 5, "f": 7202},
  {"h": "莖", "p": "jīng", "sp": "jing", "r": "艸", "sc": 10, "l": 5, "f": 7203},
  {"h": "葌", "p": "jiān", "sp": "jian", "r": "艸", "sc": 12, "l": 5, "f": 7204},
  {"h": "豬", "p": "zhū", "sp": "zhu", "r": "豕", "sc": 15, "l": 5, "f": 7205},
  {"h": "貏", "p": "bǐ", "sp": "bi", "r": "豸", "sc": 15, "l": 5, "f": 7206},
  {"h": "違", "p": "wéi", "sp": "wei", "r": "辵", "sc": 12, "l": 5, "f": 7207},
  {
    "h": "鄉",
    "p": "xiāng",
    "sp": "xiang",
    "r": "邑",
    "sc": 11,
    "l": 5,
    "f": 7208
  },
  {"h": "闅", "p": "wén", "sp": "wen", "r": "門", "sc": 17, "l": 5, "f": 7209},
  {"h": "闊", "p": "kuò", "sp": "kuo", "r": "門", "sc": 17, "l": 5, "f": 7210},
  {"h": "頂", "p": "dǐng", "sp": "ding", "r": "頁", "sc": 11, "l": 5, "f": 7211},
  {"h": "飢", "p": "jī", "sp": "ji", "r": "食", "sc": 10, "l": 5, "f": 7212},
  {"h": "驅", "p": "qū", "sp": "qu", "r": "馬", "sc": 21, "l": 5, "f": 7213},
  {"h": "魚", "p": "yú", "sp": "yu", "r": "魚", "sc": 11, "l": 5, "f": 7214},
  {
    "h": "鳉",
    "p": "jiāng",
    "sp": "jiang",
    "r": "魚",
    "sc": 17,
    "l": 5,
    "f": 7215
  },
  {"h": "鴱", "p": "ài", "sp": "ai", "r": "鳥", "sc": 16, "l": 5, "f": 7216},
  {"h": "鶎", "p": "zun", "sp": "zun", "r": "鳥", "sc": 19, "l": 5, "f": 7217},
  {"h": "黴", "p": "méi", "sp": "mei", "r": "黑", "sc": 23, "l": 5, "f": 7218},
  {"h": "祼", "p": "guàn", "sp": "guan", "r": "示", "sc": 12, "l": 5, "f": 7219},
  {"h": "鬛", "p": "liè", "sp": "lie", "r": "髟", "sc": 22, "l": 5, "f": 7220},
  {"h": "镈", "p": "bó", "sp": "bo", "r": "金", "sc": 15, "l": 5, "f": 7221},
  {"h": "瞓", "p": "shuì", "sp": "shui", "r": "目", "sc": 15, "l": 5, "f": 7222},
  {"h": "罶", "p": "liǔ", "sp": "liu", "r": "网", "sc": 15, "l": 5, "f": 7223},
  {"h": "羑", "p": "yǒu", "sp": "you", "r": "羊", "sc": 9, "l": 5, "f": 7224},
  {"h": "黉", "p": "hóng", "sp": "hong", "r": "黃", "sc": 16, "l": 5, "f": 7225},
  {"h": "霚", "p": "wù", "sp": "wu", "r": "雨", "sc": 17, "l": 5, "f": 7226},
  {"h": "笹", "p": "ti", "sp": "ti", "r": "竹", "sc": 11, "l": 5, "f": 7227},
  {"h": "郐", "p": "kuài", "sp": "kuai", "r": "邑", "sc": 8, "l": 5, "f": 7228},
  {"h": "溻", "p": "tā", "sp": "ta", "r": "水", "sc": 13, "l": 5, "f": 7229},
  {"h": "骯", "p": "āng", "sp": "ang", "r": "骨", "sc": 13, "l": 5, "f": 7230},
  {"h": "礡", "p": "bó", "sp": "bo", "r": "石", "sc": 18, "l": 5, "f": 7231},
  {"h": "袉", "p": "tuó", "sp": "tuo", "r": "衣", "sc": 10, "l": 5, "f": 7232},
  {"h": "訚", "p": "yín", "sp": "yin", "r": "言", "sc": 10, "l": 5, "f": 7233},
  {"h": "谉", "p": "shěn", "sp": "shen", "r": "言", "sc": 10, "l": 5, "f": 7234},
  {"h": "莛", "p": "tíng", "sp": "ting", "r": "艸", "sc": 9, "l": 5, "f": 7235},
  {"h": "哕", "p": "huì", "sp": "hui", "r": "口", "sc": 9, "l": 5, "f": 7236},
  {"h": "滗", "p": "bì", "sp": "bi", "r": "水", "sc": 13, "l": 5, "f": 7237},
  {"h": "瀹", "p": "yuè", "sp": "yue", "r": "水", "sc": 20, "l": 5, "f": 7238},
  {"h": "鏰", "p": "bèng", "sp": "beng", "r": "金", "sc": 19, "l": 5, "f": 7239},
  {"h": "稃", "p": "fū", "sp": "fu", "r": "禾", "sc": 12, "l": 5, "f": 7240},
  {"h": "眰", "p": "diè", "sp": "die", "r": "目", "sc": 11, "l": 5, "f": 7241},
  {"h": "眳", "p": "míng", "sp": "ming", "r": "目", "sc": 11, "l": 5, "f": 7242},
  {"h": "矞", "p": "yù", "sp": "yu", "r": "矛", "sc": 12, "l": 5, "f": 7243},
  {"h": "硚", "p": "qiáo", "sp": "qiao", "r": "石", "sc": 11, "l": 5, "f": 7244},
  {"h": "禋", "p": "yīn", "sp": "yin", "r": "示", "sc": 13, "l": 5, "f": 7245},
  {
    "h": "翀",
    "p": "chōng",
    "sp": "chong",
    "r": "羽",
    "sc": 10,
    "l": 5,
    "f": 7246
  },
  {"h": "苾", "p": "bì", "sp": "bi", "r": "艸", "sc": 8, "l": 5, "f": 7247},
  {"h": "襕", "p": "lán", "sp": "lan", "r": "衣", "sc": 17, "l": 5, "f": 7248},
  {"h": "趥", "p": "qiū", "sp": "qiu", "r": "走", "sc": 16, "l": 5, "f": 7249},
  {"h": "茺", "p": "chōng", "sp": "chong", "r": "艸", "sc": 9, "l": 5, "f": 7250},
  {"h": "軿", "p": "píng", "sp": "ping", "r": "車", "sc": 13, "l": 5, "f": 7251},
  {"h": "逌", "p": "yōu", "sp": "you", "r": "辵", "sc": 10, "l": 5, "f": 7252},
  {"h": "唪", "p": "fěng", "sp": "feng", "r": "口", "sc": 11, "l": 5, "f": 7253},
  {"h": "庳", "p": "bì", "sp": "bi", "r": "广", "sc": 11, "l": 5, "f": 7254},
  {"h": "殪", "p": "yì", "sp": "yi", "r": "歹", "sc": 16, "l": 5, "f": 7255},
  {"h": "隃", "p": "shù", "sp": "shu", "r": "阜", "sc": 11, "l": 5, "f": 7256},
  {"h": "糈", "p": "xǔ", "sp": "xu", "r": "米", "sc": 15, "l": 5, "f": 7257},
  {"h": "鲄", "p": "hé", "sp": "he", "r": "魚", "sc": 13, "l": 5, "f": 7258},
  {"h": "龜", "p": "guī", "sp": "gui", "r": "龜", "sc": 17, "l": 5, "f": 7259},
  {"h": "碕", "p": "qí", "sp": "qi", "r": "石", "sc": 13, "l": 5, "f": 7260},
  {"h": "鲣", "p": "jiān", "sp": "jian", "r": "魚", "sc": 15, "l": 5, "f": 7261},
  {"h": "綎", "p": "tīng", "sp": "ting", "r": "糸", "sc": 12, "l": 5, "f": 7262},
  {"h": "栝", "p": "guā", "sp": "gua", "r": "木", "sc": 10, "l": 5, "f": 7263},
  {"h": "轵", "p": "zhǐ", "sp": "zhi", "r": "車", "sc": 9, "l": 5, "f": 7264},
  {"h": "雔", "p": "chóu", "sp": "chou", "r": "隹", "sc": 16, "l": 5, "f": 7265},
  {"h": "頠", "p": "wěi", "sp": "wei", "r": "頁", "sc": 15, "l": 5, "f": 7266},
  {"h": "穈", "p": "méi", "sp": "mei", "r": "禾", "sc": 16, "l": 5, "f": 7267},
  {"h": "翃", "p": "hóng", "sp": "hong", "r": "羽", "sc": 10, "l": 5, "f": 7268},
  {"h": "膭", "p": "guī", "sp": "gui", "r": "肉", "sc": 16, "l": 5, "f": 7269},
  {"h": "衒", "p": "xuàn", "sp": "xuan", "r": "行", "sc": 11, "l": 5, "f": 7270},
  {"h": "盇", "p": "hé", "sp": "he", "r": "皿", "sc": 9, "l": 5, "f": 7271},
  {
    "h": "礦",
    "p": "kuàng",
    "sp": "kuang",
    "r": "石",
    "sc": 19,
    "l": 5,
    "f": 7272
  },
  {"h": "譢", "p": "suì", "sp": "sui", "r": "言", "sc": 19, "l": 5, "f": 7273},
  {"h": "鳰", "p": "ru", "sp": "ru", "r": "鳥", "sc": 13, "l": 5, "f": 7274},
  {"h": "遞", "p": "dì", "sp": "di", "r": "辵", "sc": 13, "l": 5, "f": 7275},
  {"h": "鴫", "p": "tian", "sp": "tian", "r": "鳥", "sc": 16, "l": 5, "f": 7276},
  {"h": "笟", "p": "gū", "sp": "gu", "r": "竹", "sc": 11, "l": 5, "f": 7277},
  {"h": "莾", "p": "mǎng", "sp": "mang", "r": "艸", "sc": 11, "l": 5, "f": 7278},
  {"h": "蒼", "p": "cāng", "sp": "cang", "r": "艸", "sc": 13, "l": 5, "f": 7279},
  {"h": "蟬", "p": "chán", "sp": "chan", "r": "虫", "sc": 18, "l": 5, "f": 7280},
  {"h": "袑", "p": "shào", "sp": "shao", "r": "衣", "sc": 10, "l": 5, "f": 7281},
  {"h": "訒", "p": "rèn", "sp": "ren", "r": "言", "sc": 10, "l": 5, "f": 7282},
  {
    "h": "講",
    "p": "jiǎng",
    "sp": "jiang",
    "r": "言",
    "sc": 17,
    "l": 5,
    "f": 7283
  },
  {"h": "讝", "p": "zhán", "sp": "zhan", "r": "言", "sc": 26, "l": 5, "f": 7284},
  {"h": "購", "p": "gòu", "sp": "gou", "r": "貝", "sc": 17, "l": 5, "f": 7285},
  {"h": "趕", "p": "gǎn", "sp": "gan", "r": "走", "sc": 14, "l": 5, "f": 7286},
  {"h": "鈧", "p": "kàng", "sp": "kang", "r": "金", "sc": 12, "l": 5, "f": 7287},
  {"h": "鈩", "p": "lu", "sp": "lu", "r": "金", "sc": 12, "l": 5, "f": 7288},
  {"h": "閃", "p": "shǎn", "sp": "shan", "r": "門", "sc": 10, "l": 5, "f": 7289},
  {"h": "閷", "p": "shài", "sp": "shai", "r": "門", "sc": 16, "l": 5, "f": 7290},
  {"h": "頉", "p": "yí", "sp": "yi", "r": "頁", "sc": 13, "l": 5, "f": 7291},
  {"h": "髤", "p": "xiū", "sp": "xiu", "r": "髟", "sc": 14, "l": 5, "f": 7292},
  {"h": "龤", "p": "xié", "sp": "xie", "r": "龠", "sc": 26, "l": 5, "f": 7293},
  {"h": "薿", "p": "nǐ", "sp": "ni", "r": "艸", "sc": 17, "l": 5, "f": 7294},
  {"h": "鱥", "p": "guì", "sp": "gui", "r": "魚", "sc": 24, "l": 5, "f": 7295},
  {"h": "觝", "p": "dǐ", "sp": "di", "r": "角", "sc": 12, "l": 5, "f": 7296},
  {"h": "鮻", "p": "suō", "sp": "suo", "r": "魚", "sc": 18, "l": 5, "f": 7297},
  {"h": "鳤", "p": "guǎn", "sp": "guan", "r": "魚", "sc": 22, "l": 5, "f": 7298},
  {"h": "穉", "p": "zhì", "sp": "zhi", "r": "禾", "sc": 17, "l": 5, "f": 7299},
  {"h": "觱", "p": "bì", "sp": "bi", "r": "角", "sc": 16, "l": 5, "f": 7300},
  {"h": "轘", "p": "huán", "sp": "huan", "r": "車", "sc": 20, "l": 5, "f": 7301},
  {"h": "鐓", "p": "duì", "sp": "dui", "r": "金", "sc": 20, "l": 5, "f": 7302},
  {"h": "筘", "p": "kòu", "sp": "kou", "r": "竹", "sc": 12, "l": 5, "f": 7303},
  {"h": "鲝", "p": "zhǎ", "sp": "zha", "r": "魚", "sc": 14, "l": 5, "f": 7304},
  {"h": "睏", "p": "kùn", "sp": "kun", "r": "目", "sc": 12, "l": 5, "f": 7305},
  {"h": "蔸", "p": "dōu", "sp": "dou", "r": "艸", "sc": 14, "l": 5, "f": 7306},
  {"h": "镚", "p": "bèng", "sp": "beng", "r": "金", "sc": 16, "l": 5, "f": 7307},
  {"h": "痦", "p": "wù", "sp": "wu", "r": "疒", "sc": 12, "l": 5, "f": 7308},
  {"h": "秵", "p": "yīn", "sp": "yin", "r": "禾", "sc": 11, "l": 5, "f": 7309},
  {"h": "臎", "p": "cuì", "sp": "cui", "r": "肉", "sc": 18, "l": 5, "f": 7310},
  {"h": "葴", "p": "zhēn", "sp": "zhen", "r": "艸", "sc": 12, "l": 5, "f": 7311},
  {"h": "蝟", "p": "wèi", "sp": "wei", "r": "虫", "sc": 15, "l": 5, "f": 7312},
  {"h": "訂", "p": "dìng", "sp": "ding", "r": "言", "sc": 9, "l": 5, "f": 7313},
  {"h": "泖", "p": "mǎo", "sp": "mao", "r": "水", "sc": 8, "l": 5, "f": 7314},
  {"h": "骟", "p": "shàn", "sp": "shan", "r": "馬", "sc": 13, "l": 5, "f": 7315},
  {
    "h": "锽",
    "p": "huáng",
    "sp": "huang",
    "r": "金",
    "sc": 14,
    "l": 5,
    "f": 7316
  },
  {"h": "犋", "p": "jù", "sp": "ju", "r": "牛", "sc": 12, "l": 5, "f": 7317},
  {"h": "眵", "p": "chī", "sp": "chi", "r": "目", "sc": 11, "l": 5, "f": 7318},
  {
    "h": "耩",
    "p": "jiǎng",
    "sp": "jiang",
    "r": "耒",
    "sc": 16,
    "l": 5,
    "f": 7319
  },
  {"h": "筢", "p": "pá", "sp": "pa", "r": "竹", "sc": 13, "l": 5, "f": 7320},
  {"h": "躔", "p": "chán", "sp": "chan", "r": "足", "sc": 22, "l": 5, "f": 7321},
  {"h": "鬏", "p": "jiū", "sp": "jiu", "r": "髟", "sc": 19, "l": 5, "f": 7322},
  {"h": "黪", "p": "cǎn", "sp": "can", "r": "黑", "sc": 20, "l": 5, "f": 7323},
  {"h": "䞍", "p": "qíng", "sp": "qing", "r": "", "sc": 15, "l": 5, "f": 7324},
  {"h": "祔", "p": "fù", "sp": "fu", "r": "示", "sc": 9, "l": 5, "f": 7325},
  {"h": "箠", "p": "chuí", "sp": "chui", "r": "竹", "sc": 14, "l": 5, "f": 7326},
  {"h": "舲", "p": "líng", "sp": "ling", "r": "舟", "sc": 11, "l": 5, "f": 7327},
  {"h": "蒷", "p": "yún", "sp": "yun", "r": "艸", "sc": 13, "l": 5, "f": 7328},
  {"h": "蜺", "p": "ní", "sp": "ni", "r": "虫", "sc": 14, "l": 5, "f": 7329},
  {"h": "襢", "p": "tǎn", "sp": "tan", "r": "衣", "sc": 18, "l": 5, "f": 7330},
  {"h": "謺", "p": "zhé", "sp": "zhe", "r": "言", "sc": 18, "l": 5, "f": 7331},
  {"h": "劐", "p": "huō", "sp": "huo", "r": "刀", "sc": 15, "l": 5, "f": 7332},
  {"h": "貯", "p": "zhù", "sp": "zhu", "r": "貝", "sc": 12, "l": 5, "f": 7333},
  {
    "h": "搋",
    "p": "chuāi",
    "sp": "chuai",
    "r": "手",
    "sc": 13,
    "l": 5,
    "f": 7334
  },
  {"h": "醊", "p": "zhuì", "sp": "zhui", "r": "酉", "sc": 15, "l": 5, "f": 7335},
  {"h": "廒", "p": "áo", "sp": "ao", "r": "广", "sc": 13, "l": 5, "f": 7336},
  {"h": "洚", "p": "jiàng", "sp": "jiang", "r": "水", "sc": 9, "l": 5, "f": 7337},
  {"h": "胲", "p": "hǎi", "sp": "hai", "r": "肉", "sc": 10, "l": 5, "f": 7338},
  {"h": "韥", "p": "dú", "sp": "du", "r": "韋", "sc": 24, "l": 5, "f": 7339},
  {"h": "饘", "p": "zhān", "sp": "zhan", "r": "食", "sc": 21, "l": 5, "f": 7340},
  {"h": "馂", "p": "jùn", "sp": "jun", "r": "食", "sc": 10, "l": 5, "f": 7341},
  {"h": "筅", "p": "xiǎn", "sp": "xian", "r": "竹", "sc": 12, "l": 5, "f": 7342},
  {"h": "鳆", "p": "fù", "sp": "fu", "r": "魚", "sc": 17, "l": 5, "f": 7343},
  {"h": "鲊", "p": "zhǎ", "sp": "zha", "r": "魚", "sc": 13, "l": 5, "f": 7344},
  {"h": "麗", "p": "lì", "sp": "li", "r": "鹿", "sc": 19, "l": 5, "f": 7345},
  {"h": "篸", "p": "cǎn", "sp": "can", "r": "竹", "sc": 17, "l": 5, "f": 7346},
  {"h": "腄", "p": "chuí", "sp": "chui", "r": "肉", "sc": 12, "l": 5, "f": 7347},
  {"h": "臑", "p": "nào", "sp": "nao", "r": "肉", "sc": 18, "l": 5, "f": 7348},
  {"h": "诐", "p": "bì", "sp": "bi", "r": "言", "sc": 7, "l": 5, "f": 7349},
  {"h": "仳", "p": "pǐ", "sp": "pi", "r": "人", "sc": 6, "l": 5, "f": 7350},
  {"h": "趩", "p": "chì", "sp": "chi", "r": "走", "sc": 18, "l": 5, "f": 7351},
  {"h": "篩", "p": "shāi", "sp": "shai", "r": "竹", "sc": 16, "l": 5, "f": 7352},
  {"h": "簾", "p": "lián", "sp": "lian", "r": "竹", "sc": 19, "l": 5, "f": 7353},
  {"h": "絪", "p": "yīn", "sp": "yin", "r": "糸", "sc": 12, "l": 5, "f": 7354},
  {"h": "绹", "p": "táo", "sp": "tao", "r": "糸", "sc": 11, "l": 5, "f": 7355},
  {"h": "膵", "p": "cuì", "sp": "cui", "r": "肉", "sc": 15, "l": 5, "f": 7356},
  {"h": "蕗", "p": "lù", "sp": "lu", "r": "艸", "sc": 16, "l": 5, "f": 7357},
  {"h": "蝘", "p": "yǎn", "sp": "yan", "r": "虫", "sc": 15, "l": 5, "f": 7358},
  {"h": "袠", "p": "zhì", "sp": "zhi", "r": "衣", "sc": 11, "l": 5, "f": 7359},
  {
    "h": "賸",
    "p": "shèng",
    "sp": "sheng",
    "r": "貝",
    "sc": 17,
    "l": 5,
    "f": 7360
  },
  {
    "h": "葙",
    "p": "xiāng",
    "sp": "xiang",
    "r": "艸",
    "sc": 12,
    "l": 5,
    "f": 7361
  },
  {"h": "遊", "p": "yóu", "sp": "you", "r": "辵", "sc": 12, "l": 5, "f": 7362},
  {"h": "饑", "p": "jī", "sp": "ji", "r": "食", "sc": 20, "l": 5, "f": 7363},
  {"h": "驀", "p": "mò", "sp": "mo", "r": "馬", "sc": 20, "l": 5, "f": 7364},
  {"h": "祪", "p": "guǐ", "sp": "gui", "r": "示", "sc": 10, "l": 5, "f": 7365},
  {"h": "禘", "p": "dì", "sp": "di", "r": "示", "sc": 13, "l": 5, "f": 7366},
  {"h": "臺", "p": "tái", "sp": "tai", "r": "至", "sc": 14, "l": 5, "f": 7367},
  {"h": "菍", "p": "niè", "sp": "nie", "r": "艸", "sc": 11, "l": 5, "f": 7368},
  {"h": "髒", "p": "zāng", "sp": "zang", "r": "骨", "sc": 21, "l": 5, "f": 7369},
  {"h": "軸", "p": "zhóu", "sp": "zhou", "r": "車", "sc": 12, "l": 5, "f": 7370},
  {"h": "釘", "p": "dīng", "sp": "ding", "r": "金", "sc": 10, "l": 5, "f": 7371},
  {"h": "頃", "p": "qǐng", "sp": "qing", "r": "頁", "sc": 11, "l": 5, "f": 7372},
  {"h": "仉", "p": "zhǎng", "sp": "zhang", "r": "人", "sc": 4, "l": 5, "f": 7373},
  {"h": "氕", "p": "piē", "sp": "pie", "r": "气", "sc": 5, "l": 5, "f": 7374},
  {"h": "矔", "p": "guàn", "sp": "guan", "r": "目", "sc": 22, "l": 5, "f": 7375},
  {"h": "緒", "p": "xù", "sp": "xu", "r": "糸", "sc": 14, "l": 5, "f": 7376},
  {"h": "脅", "p": "xié", "sp": "xie", "r": "肉", "sc": 10, "l": 5, "f": 7377},
  {"h": "臒", "p": "wò", "sp": "wo", "r": "肉", "sc": 17, "l": 5, "f": 7378},
  {"h": "苿", "p": "wèi", "sp": "wei", "r": "艸", "sc": 8, "l": 5, "f": 7379},
  {"h": "菢", "p": "bào", "sp": "bao", "r": "艸", "sc": 11, "l": 5, "f": 7380},
  {"h": "蔂", "p": "léi", "sp": "lei", "r": "艸", "sc": 14, "l": 5, "f": 7381},
  {"h": "藞", "p": "lǎ", "sp": "la", "r": "艸", "sc": 18, "l": 5, "f": 7382},
  {"h": "袩", "p": "zhé", "sp": "zhe", "r": "衣", "sc": 10, "l": 5, "f": 7383},
  {"h": "貞", "p": "zhēn", "sp": "zhen", "r": "貝", "sc": 9, "l": 5, "f": 7384},
  {"h": "賣", "p": "mài", "sp": "mai", "r": "貝", "sc": 15, "l": 5, "f": 7385},
  {"h": "跍", "p": "kū", "sp": "ku", "r": "足", "sc": 12, "l": 5, "f": 7386},
  {"h": "跓", "p": "zhù", "sp": "zhu", "r": "足", "sc": 12, "l": 5, "f": 7387},
  {"h": "蹤", "p": "zōng", "sp": "zong", "r": "足", "sc": 18, "l": 5, "f": 7388},
  {"h": "軍", "p": "jūn", "sp": "jun", "r": "車", "sc": 9, "l": 5, "f": 7389},
  {"h": "邿", "p": "shī", "sp": "shi", "r": "邑", "sc": 8, "l": 5, "f": 7390},
  {"h": "鄰", "p": "lín", "sp": "lin", "r": "邑", "sc": 14, "l": 5, "f": 7391},
  {"h": "醿", "p": "mí", "sp": "mi", "r": "酉", "sc": 24, "l": 5, "f": 7392},
  {"h": "鈨", "p": "yuan", "sp": "yuan", "r": "金", "sc": 12, "l": 5, "f": 7393},
  {"h": "鋼", "p": "gāng", "sp": "gang", "r": "金", "sc": 16, "l": 5, "f": 7394},
  {"h": "駵", "p": "liú", "sp": "liu", "r": "馬", "sc": 17, "l": 5, "f": 7395},
  {"h": "駸", "p": "qīn", "sp": "qin", "r": "馬", "sc": 17, "l": 5, "f": 7396},
  {"h": "鰪", "p": "é", "sp": "e", "r": "魚", "sc": 21, "l": 5, "f": 7397},
  {"h": "鱾", "p": "jǐ", "sp": "ji", "r": "魚", "sc": 11, "l": 5, "f": 7398},
  {"h": "䝼", "p": "jìng", "sp": "jing", "r": "貝", "sc": 15, "l": 5, "f": 7399},
  {"h": "賨", "p": "cóng", "sp": "cong", "r": "貝", "sc": 15, "l": 5, "f": 7400},
  {"h": "艞", "p": "yào", "sp": "yao", "r": "舟", "sc": 18, "l": 5, "f": 7401},
  {"h": "鹟", "p": "wēng", "sp": "weng", "r": "鳥", "sc": 15, "l": 5, "f": 7402},
  {"h": "钁", "p": "jué", "sp": "jue", "r": "金", "sc": 28, "l": 5, "f": 7403},
  {"h": "鞓", "p": "tīng", "sp": "ting", "r": "革", "sc": 16, "l": 5, "f": 7404},
  {
    "h": "鰉",
    "p": "huáng",
    "sp": "huang",
    "r": "魚",
    "sc": 20,
    "l": 5,
    "f": 7405
  },
  {"h": "監", "p": "jiān", "sp": "jian", "r": "皿", "sc": 14, "l": 5, "f": 7406},
  {"h": "穊", "p": "jì", "sp": "ji", "r": "禾", "sc": 14, "l": 5, "f": 7407},
  {"h": "蕡", "p": "fén", "sp": "fen", "r": "艸", "sc": 15, "l": 5, "f": 7408},
  {"h": "鑨", "p": "lóng", "sp": "long", "r": "金", "sc": 24, "l": 5, "f": 7409},
  {"h": "楱", "p": "zòu", "sp": "zou", "r": "木", "sc": 13, "l": 5, "f": 7410},
  {"h": "靹", "p": "nà", "sp": "na", "r": "革", "sc": 13, "l": 5, "f": 7411},
  {"h": "鷃", "p": "yàn", "sp": "yan", "r": "鳥", "sc": 21, "l": 5, "f": 7412},
  {"h": "鼱", "p": "jīng", "sp": "jing", "r": "鼠", "sc": 21, "l": 5, "f": 7413},
  {"h": "眫", "p": "mǐ", "sp": "mi", "r": "目", "sc": 10, "l": 5, "f": 7414},
  {"h": "艆", "p": "láng", "sp": "lang", "r": "舟", "sc": 13, "l": 5, "f": 7415},
  {"h": "褯", "p": "jiè", "sp": "jie", "r": "衣", "sc": 15, "l": 5, "f": 7416},
  {"h": "剞", "p": "jī", "sp": "ji", "r": "刀", "sc": 10, "l": 5, "f": 7417},
  {"h": "蔟", "p": "cù", "sp": "cu", "r": "艸", "sc": 14, "l": 5, "f": 7418},
  {"h": "靿", "p": "yào", "sp": "yao", "r": "革", "sc": 14, "l": 5, "f": 7419},
  {"h": "飭", "p": "chì", "sp": "chi", "r": "食", "sc": 12, "l": 5, "f": 7420},
  {"h": "瘼", "p": "mò", "sp": "mo", "r": "疒", "sc": 15, "l": 5, "f": 7421},
  {"h": "酴", "p": "tú", "sp": "tu", "r": "酉", "sc": 14, "l": 5, "f": 7422},
  {"h": "鼢", "p": "fén", "sp": "fen", "r": "鼠", "sc": 17, "l": 5, "f": 7423},
  {"h": "睆", "p": "huàn", "sp": "huan", "r": "目", "sc": 12, "l": 5, "f": 7424},
  {"h": "硁", "p": "kēng", "sp": "keng", "r": "石", "sc": 10, "l": 5, "f": 7425},
  {"h": "秬", "p": "jù", "sp": "ju", "r": "禾", "sc": 9, "l": 5, "f": 7426},
  {
    "h": "穪",
    "p": "chēng",
    "sp": "cheng",
    "r": "禾",
    "sc": 19,
    "l": 5,
    "f": 7427
  },
  {"h": "筦", "p": "guǎn", "sp": "guan", "r": "竹", "sc": 13, "l": 5, "f": 7428},
  {"h": "粡", "p": "tóng", "sp": "tong", "r": "米", "sc": 12, "l": 5, "f": 7429},
  {"h": "粣", "p": "cè", "sp": "ce", "r": "米", "sc": 11, "l": 5, "f": 7430},
  {"h": "綖", "p": "yán", "sp": "yan", "r": "糸", "sc": 12, "l": 5, "f": 7431},
  {"h": "裏", "p": "lǐ", "sp": "li", "r": "衣", "sc": 13, "l": 5, "f": 7432},
  {"h": "刭", "p": "jǐng", "sp": "jing", "r": "刀", "sc": 7, "l": 5, "f": 7433},
  {"h": "跼", "p": "jú", "sp": "ju", "r": "足", "sc": 14, "l": 5, "f": 7434},
  {"h": "蒇", "p": "chǎn", "sp": "chan", "r": "艸", "sc": 12, "l": 5, "f": 7435},
  {"h": "挢", "p": "jiǎo", "sp": "jiao", "r": "手", "sc": 9, "l": 5, "f": 7436},
  {"h": "邆", "p": "téng", "sp": "teng", "r": "辵", "sc": 15, "l": 5, "f": 7437},
  {"h": "彡", "p": "shān", "sp": "shan", "r": "彡", "sc": 3, "l": 5, "f": 7438},
  {"h": "猃", "p": "xiǎn", "sp": "xian", "r": "犬", "sc": 10, "l": 5, "f": 7439},
  {"h": "闉", "p": "yīn", "sp": "yin", "r": "門", "sc": 17, "l": 5, "f": 7440},
  {"h": "阓", "p": "huì", "sp": "hui", "r": "門", "sc": 12, "l": 5, "f": 7441},
  {"h": "阛", "p": "huán", "sp": "huan", "r": "門", "sc": 16, "l": 5, "f": 7442},
  {"h": "靺", "p": "mò", "sp": "mo", "r": "革", "sc": 14, "l": 5, "f": 7443},
  {"h": "鞨", "p": "hé", "sp": "he", "r": "革", "sc": 18, "l": 5, "f": 7444},
  {"h": "铥", "p": "diū", "sp": "diu", "r": "金", "sc": 11, "l": 5, "f": 7445},
  {"h": "瘅", "p": "dān", "sp": "dan", "r": "疒", "sc": 13, "l": 5, "f": 7446},
  {"h": "駰", "p": "yīn", "sp": "yin", "r": "馬", "sc": 16, "l": 5, "f": 7447},
  {
    "h": "騻",
    "p": "shuāng",
    "sp": "shuang",
    "r": "馬",
    "sc": 21,
    "l": 5,
    "f": 7448
  },
  {"h": "箅", "p": "bì", "sp": "bi", "r": "竹", "sc": 14, "l": 5, "f": 7449},
  {"h": "箢", "p": "yuān", "sp": "yuan", "r": "竹", "sc": 14, "l": 5, "f": 7450},
  {"h": "觖", "p": "jué", "sp": "jue", "r": "角", "sc": 11, "l": 5, "f": 7451},
  {"h": "麆", "p": "zhù", "sp": "zhu", "r": "鹿", "sc": 16, "l": 5, "f": 7452},
  {"h": "齮", "p": "yǐ", "sp": "yi", "r": "齒", "sc": 23, "l": 5, "f": 7453},
  {"h": "糀", "p": "hua", "sp": "hua", "r": "米", "sc": 13, "l": 5, "f": 7454},
  {"h": "虯", "p": "qiú", "sp": "qiu", "r": "虫", "sc": 8, "l": 5, "f": 7455},
  {"h": "瘊", "p": "hóu", "sp": "hou", "r": "疒", "sc": 14, "l": 5, "f": 7456},
  {"h": "絛", "p": "tāo", "sp": "tao", "r": "糸", "sc": 12, "l": 5, "f": 7457},
  {"h": "聴", "p": "tīng", "sp": "ting", "r": "耳", "sc": 17, "l": 5, "f": 7458},
  {"h": "蘭", "p": "lán", "sp": "lan", "r": "艸", "sc": 20, "l": 5, "f": 7459},
  {"h": "讬", "p": "tuō", "sp": "tuo", "r": "言", "sc": 5, "l": 5, "f": 7460},
  {"h": "揲", "p": "dié", "sp": "die", "r": "手", "sc": 12, "l": 5, "f": 7461},
  {"h": "镮", "p": "huán", "sp": "huan", "r": "金", "sc": 18, "l": 5, "f": 7462},
  {"h": "钷", "p": "pǒ", "sp": "po", "r": "金", "sc": 10, "l": 5, "f": 7463},
  {"h": "睪", "p": "yì", "sp": "yi", "r": "目", "sc": 13, "l": 5, "f": 7464},
  {"h": "砲", "p": "pào", "sp": "pao", "r": "石", "sc": 10, "l": 5, "f": 7465},
  {"h": "稅", "p": "shuì", "sp": "shui", "r": "禾", "sc": 12, "l": 5, "f": 7466},
  {"h": "翫", "p": "wán", "sp": "wan", "r": "羽", "sc": 15, "l": 5, "f": 7467},
  {"h": "荅", "p": "dā", "sp": "da", "r": "艸", "sc": 9, "l": 5, "f": 7468},
  {"h": "菉", "p": "lù", "sp": "lu", "r": "艸", "sc": 11, "l": 5, "f": 7469},
  {"h": "袚", "p": "bō", "sp": "bo", "r": "衣", "sc": 10, "l": 5, "f": 7470},
  {"h": "褋", "p": "dié", "sp": "die", "r": "衣", "sc": 14, "l": 5, "f": 7471},
  {"h": "贠", "p": "yuán", "sp": "yuan", "r": "貝", "sc": 6, "l": 5, "f": 7472},
  {"h": "跡", "p": "jī", "sp": "ji", "r": "足", "sc": 13, "l": 5, "f": 7473},
  {"h": "踦", "p": "yǐ", "sp": "yi", "r": "足", "sc": 15, "l": 5, "f": 7474},
  {"h": "遯", "p": "dùn", "sp": "dun", "r": "辵", "sc": 14, "l": 5, "f": 7475},
  {"h": "鉅", "p": "jù", "sp": "ju", "r": "金", "sc": 12, "l": 5, "f": 7476},
  {"h": "鎌", "p": "lián", "sp": "lian", "r": "金", "sc": 18, "l": 5, "f": 7477},
  {"h": "靉", "p": "ài", "sp": "ai", "r": "雨", "sc": 25, "l": 5, "f": 7478},
  {"h": "韲", "p": "jī", "sp": "ji", "r": "韭", "sc": 19, "l": 5, "f": 7479},
  {"h": "餖", "p": "dòu", "sp": "dou", "r": "食", "sc": 15, "l": 5, "f": 7480},
  {"h": "鶺", "p": "jí", "sp": "ji", "r": "鳥", "sc": 21, "l": 5, "f": 7481},
  {"h": "麕", "p": "jūn", "sp": "jun", "r": "鹿", "sc": 19, "l": 5, "f": 7482},
  {"h": "瞼", "p": "jiǎn", "sp": "jian", "r": "目", "sc": 18, "l": 5, "f": 7483},
  {"h": "菵", "p": "wǎng", "sp": "wang", "r": "艸", "sc": 11, "l": 5, "f": 7484},
  {"h": "蟭", "p": "jiāo", "sp": "jiao", "r": "虫", "sc": 18, "l": 5, "f": 7485},
  {"h": "錏", "p": "yā", "sp": "ya", "r": "金", "sc": 16, "l": 5, "f": 7486},
  {"h": "踐", "p": "jiàn", "sp": "jian", "r": "足", "sc": 15, "l": 5, "f": 7487},
  {"h": "輾", "p": "zhǎn", "sp": "zhan", "r": "車", "sc": 17, "l": 5, "f": 7488},
  {"h": "銷", "p": "xiāo", "sp": "xiao", "r": "金", "sc": 15, "l": 5, "f": 7489},
  {"h": "閣", "p": "gé", "sp": "ge", "r": "門", "sc": 14, "l": 5, "f": 7490},
  {"h": "頁", "p": "yè", "sp": "ye", "r": "頁", "sc": 9, "l": 5, "f": 7491},
  {"h": "驟", "p": "zhòu", "sp": "zhou", "r": "馬", "sc": 24, "l": 5, "f": 7492},
  {"h": "碼", "p": "mǎ", "sp": "ma", "r": "石", "sc": 15, "l": 5, "f": 7493},
  {"h": "禎", "p": "zhēn", "sp": "zhen", "r": "示", "sc": 13, "l": 5, "f": 7494},
  {"h": "禦", "p": "yù", "sp": "yu", "r": "示", "sc": 17, "l": 5, "f": 7495},
  {"h": "稖", "p": "bàng", "sp": "bang", "r": "禾", "sc": 13, "l": 5, "f": 7496},
  {"h": "稟", "p": "bǐng", "sp": "bing", "r": "禾", "sc": 13, "l": 5, "f": 7497},
  {"h": "竇", "p": "dòu", "sp": "dou", "r": "穴", "sc": 20, "l": 5, "f": 7498},
  {"h": "笇", "p": "suàn", "sp": "suan", "r": "竹", "sc": 10, "l": 5, "f": 7499},
  {"h": "篂", "p": "xīng", "sp": "xing", "r": "竹", "sc": 15, "l": 5, "f": 7500},
  {"h": "緛", "p": "ruǎn", "sp": "ruan", "r": "糸", "sc": 15, "l": 5, "f": 7501},
  {"h": "編", "p": "biān", "sp": "bian", "r": "糸", "sc": 15, "l": 5, "f": 7502},
  {"h": "縿", "p": "shān", "sp": "shan", "r": "糸", "sc": 17, "l": 5, "f": 7503},
  {"h": "舊", "p": "jiù", "sp": "jiu", "r": "臼", "sc": 17, "l": 5, "f": 7504},
  {"h": "蓴", "p": "chún", "sp": "chun", "r": "艸", "sc": 14, "l": 5, "f": 7505},
  {"h": "蕚", "p": "è", "sp": "e", "r": "艸", "sc": 15, "l": 5, "f": 7506},
  {"h": "蘆", "p": "lú", "sp": "lu", "r": "艸", "sc": 19, "l": 5, "f": 7507},
  {"h": "蝿", "p": "yíng", "sp": "ying", "r": "虫", "sc": 15, "l": 5, "f": 7508},
  {"h": "袃", "p": "chài", "sp": "chai", "r": "衣", "sc": 10, "l": 5, "f": 7509},
  {"h": "親", "p": "qīn", "sp": "qin", "r": "見", "sc": 16, "l": 5, "f": 7510},
  {"h": "訴", "p": "sù", "sp": "su", "r": "言", "sc": 12, "l": 5, "f": 7511},
  {"h": "讕", "p": "lán", "sp": "lan", "r": "言", "sc": 24, "l": 5, "f": 7512},
  {"h": "讚", "p": "zàn", "sp": "zan", "r": "言", "sc": 26, "l": 5, "f": 7513},
  {"h": "財", "p": "cái", "sp": "cai", "r": "貝", "sc": 10, "l": 5, "f": 7514},
  {"h": "贏", "p": "yíng", "sp": "ying", "r": "貝", "sc": 20, "l": 5, "f": 7515},
  {"h": "軅", "p": "yan", "sp": "yan", "r": "身", "sc": 19, "l": 5, "f": 7516},
  {"h": "軓", "p": "fàn", "sp": "fan", "r": "車", "sc": 10, "l": 5, "f": 7517},
  {"h": "釙", "p": "pò", "sp": "po", "r": "金", "sc": 10, "l": 5, "f": 7518},
  {"h": "鉄", "p": "zhí", "sp": "zhi", "r": "金", "sc": 13, "l": 5, "f": 7519},
  {
    "h": "錆",
    "p": "qiāng",
    "sp": "qiang",
    "r": "金",
    "sc": 16,
    "l": 5,
    "f": 7520
  },
  {
    "h": "閶",
    "p": "chāng",
    "sp": "chang",
    "r": "門",
    "sc": 16,
    "l": 5,
    "f": 7521
  },
  {"h": "鯕", "p": "qí", "sp": "qi", "r": "魚", "sc": 19, "l": 5, "f": 7522},
  {"h": "鶕", "p": "ān", "sp": "an", "r": "鳥", "sc": 20, "l": 5, "f": 7523},
  {"h": "麨", "p": "chǎo", "sp": "chao", "r": "麥", "sc": 15, "l": 5, "f": 7524},
  {"h": "䜩", "p": "yàn", "sp": "yan", "r": "", "sc": 23, "l": 5, "f": 7525},
  {"h": "䴗", "p": "jú", "sp": "ju", "r": "", "sc": 20, "l": 5, "f": 7526},
  {
    "h": "脭",
    "p": "chéng",
    "sp": "cheng",
    "r": "肉",
    "sc": 11,
    "l": 5,
    "f": 7527
  },
  {"h": "鲯", "p": "qí", "sp": "qi", "r": "魚", "sc": 16, "l": 5, "f": 7528},
  {"h": "鐏", "p": "zūn", "sp": "zun", "r": "金", "sc": 20, "l": 5, "f": 7529},
  {"h": "魾", "p": "pī", "sp": "pi", "r": "魚", "sc": 16, "l": 5, "f": 7530},
  {"h": "眎", "p": "shì", "sp": "shi", "r": "目", "sc": 10, "l": 5, "f": 7531},
  {"h": "睞", "p": "lài", "sp": "lai", "r": "目", "sc": 13, "l": 5, "f": 7532},
  {"h": "窋", "p": "zhú", "sp": "zhu", "r": "穴", "sc": 10, "l": 5, "f": 7533},
  {"h": "籠", "p": "lóng", "sp": "long", "r": "竹", "sc": 22, "l": 5, "f": 7534},
  {"h": "綠", "p": "lǜ", "sp": "lu", "r": "糸", "sc": 14, "l": 5, "f": 7535},
  {"h": "縡", "p": "zài", "sp": "zai", "r": "糸", "sc": 16, "l": 5, "f": 7536},
  {"h": "縫", "p": "fèng", "sp": "feng", "r": "糸", "sc": 16, "l": 5, "f": 7537},
  {"h": "肅", "p": "sù", "sp": "su", "r": "聿", "sc": 13, "l": 5, "f": 7538},
  {"h": "鉏", "p": "chú", "sp": "chu", "r": "金", "sc": 13, "l": 5, "f": 7539},
  {"h": "謦", "p": "qìng", "sp": "qing", "r": "言", "sc": 18, "l": 5, "f": 7540},
  {"h": "麹", "p": "qū", "sp": "qu", "r": "麥", "sc": 15, "l": 5, "f": 7541},
  {"h": "蔔", "p": "bó", "sp": "bo", "r": "艸", "sc": 14, "l": 5, "f": 7542},
  {"h": "繐", "p": "suì", "sp": "sui", "r": "糸", "sc": 18, "l": 5, "f": 7543},
  {"h": "眥", "p": "zì", "sp": "zi", "r": "目", "sc": 11, "l": 5, "f": 7544},
  {"h": "粺", "p": "bài", "sp": "bai", "r": "米", "sc": 14, "l": 5, "f": 7545},
  {"h": "缊", "p": "yūn", "sp": "yun", "r": "糸", "sc": 12, "l": 5, "f": 7546},
  {"h": "艈", "p": "yù", "sp": "yu", "r": "舟", "sc": 13, "l": 5, "f": 7547},
  {"h": "蓇", "p": "gǔ", "sp": "gu", "r": "艸", "sc": 12, "l": 5, "f": 7548},
  {"h": "蛣", "p": "qī", "sp": "qi", "r": "虫", "sc": 12, "l": 5, "f": 7549},
  {"h": "蹾", "p": "dūn", "sp": "dun", "r": "足", "sc": 19, "l": 5, "f": 7550},
  {"h": "掎", "p": "jǐ", "sp": "ji", "r": "手", "sc": 11, "l": 5, "f": 7551},
  {"h": "哜", "p": "jì", "sp": "ji", "r": "口", "sc": 9, "l": 5, "f": 7552},
  {"h": "鄦", "p": "xǔ", "sp": "xu", "r": "邑", "sc": 14, "l": 5, "f": 7553},
  {"h": "陞", "p": "shēng", "sp": "sheng", "r": "阜", "sc": 9, "l": 5, "f": 7554},
  {"h": "韰", "p": "xiè", "sp": "xie", "r": "韭", "sc": 16, "l": 5, "f": 7555},
  {"h": "飖", "p": "yáo", "sp": "yao", "r": "風", "sc": 14, "l": 5, "f": 7556},
  {"h": "镲", "p": "chǎ", "sp": "cha", "r": "金", "sc": 19, "l": 5, "f": 7557},
  {"h": "麃", "p": "páo", "sp": "pao", "r": "鹿", "sc": 15, "l": 5, "f": 7558},
  {"h": "禑", "p": "wú", "sp": "wu", "r": "示", "sc": 13, "l": 5, "f": 7559},
  {"h": "筰", "p": "zuó", "sp": "zuo", "r": "竹", "sc": 13, "l": 5, "f": 7560},
  {"h": "聠", "p": "pīng", "sp": "ping", "r": "耳", "sc": 12, "l": 5, "f": 7561},
  {"h": "茷", "p": "fá", "sp": "fa", "r": "艸", "sc": 9, "l": 5, "f": 7562},
  {
    "h": "螀",
    "p": "jiāng",
    "sp": "jiang",
    "r": "虫",
    "sc": 15,
    "l": 5,
    "f": 7563
  },
  {"h": "袞", "p": "gǔn", "sp": "gun", "r": "衣", "sc": 11, "l": 5, "f": 7564},
  {
    "h": "褧",
    "p": "jiǒng",
    "sp": "jiong",
    "r": "衣",
    "sc": 16,
    "l": 5,
    "f": 7565
  },
  {"h": "賮", "p": "jìn", "sp": "jin", "r": "貝", "sc": 16, "l": 5, "f": 7566},
  {"h": "赗", "p": "fèng", "sp": "feng", "r": "貝", "sc": 13, "l": 5, "f": 7567},
  {"h": "踡", "p": "quán", "sp": "quan", "r": "足", "sc": 15, "l": 5, "f": 7568},
  {"h": "揞", "p": "ǎn", "sp": "an", "r": "手", "sc": 12, "l": 5, "f": 7569},
  {"h": "遝", "p": "tà", "sp": "ta", "r": "辵", "sc": 13, "l": 5, "f": 7570},
  {"h": "酂", "p": "cuó", "sp": "cuo", "r": "邑", "sc": 18, "l": 5, "f": 7571},
  {"h": "酇", "p": "zàn", "sp": "zan", "r": "邑", "sc": 21, "l": 5, "f": 7572},
  {"h": "醜", "p": "chǒu", "sp": "chou", "r": "酉", "sc": 16, "l": 5, "f": 7573},
  {"h": "檑", "p": "léi", "sp": "lei", "r": "木", "sc": 17, "l": 5, "f": 7574},
  {"h": "雮", "p": "mù", "sp": "mu", "r": "雨", "sc": 12, "l": 5, "f": 7575},
  {"h": "氍", "p": "qú", "sp": "qu", "r": "毛", "sc": 22, "l": 5, "f": 7576},
  {"h": "腙", "p": "zōng", "sp": "zong", "r": "肉", "sc": 12, "l": 5, "f": 7577},
  {"h": "飩", "p": "tún", "sp": "tun", "r": "食", "sc": 12, "l": 5, "f": 7578},
  {"h": "痖", "p": "yǎ", "sp": "ya", "r": "疒", "sc": 11, "l": 5, "f": 7579},
  {"h": "髣", "p": "fǎng", "sp": "fang", "r": "髟", "sc": 14, "l": 5, "f": 7580},
  {"h": "髳", "p": "máo", "sp": "mao", "r": "髟", "sc": 15, "l": 5, "f": 7581},
  {"h": "酎", "p": "zhòu", "sp": "zhou", "r": "酉", "sc": 10, "l": 5, "f": 7582},
  {"h": "鲒", "p": "jié", "sp": "jie", "r": "魚", "sc": 14, "l": 5, "f": 7583},
  {"h": "鲪", "p": "jūn", "sp": "jun", "r": "魚", "sc": 15, "l": 5, "f": 7584},
  {"h": "鹯", "p": "zhān", "sp": "zhan", "r": "鳥", "sc": 18, "l": 5, "f": 7585},
  {"h": "麜", "p": "lì", "sp": "li", "r": "鹿", "sc": 21, "l": 5, "f": 7586},
  {"h": "齸", "p": "yì", "sp": "yi", "r": "齒", "sc": 25, "l": 5, "f": 7587},
  {"h": "齻", "p": "diān", "sp": "dian", "r": "齒", "sc": 25, "l": 5, "f": 7588},
  {"h": "粩", "p": "lao", "sp": "lao", "r": "米", "sc": 12, "l": 5, "f": 7589},
  {"h": "詣", "p": "yì", "sp": "yi", "r": "言", "sc": 13, "l": 5, "f": 7590},
  {"h": "隄", "p": "dī", "sp": "di", "r": "阜", "sc": 11, "l": 5, "f": 7591},
  {"h": "簰", "p": "pái", "sp": "pai", "r": "竹", "sc": 18, "l": 5, "f": 7592},
  {"h": "蒾", "p": "mí", "sp": "mi", "r": "艸", "sc": 12, "l": 5, "f": 7593},
  {"h": "蓏", "p": "luǒ", "sp": "luo", "r": "艸", "sc": 13, "l": 5, "f": 7594},
  {"h": "薳", "p": "wěi", "sp": "wei", "r": "艸", "sc": 16, "l": 5, "f": 7595},
  {"h": "藛", "p": "xiě", "sp": "xie", "r": "艸", "sc": 18, "l": 5, "f": 7596},
  {"h": "蚢", "p": "háng", "sp": "hang", "r": "虫", "sc": 10, "l": 5, "f": 7597},
  {"h": "蓠", "p": "lí", "sp": "li", "r": "艸", "sc": 13, "l": 5, "f": 7598},
  {"h": "蓰", "p": "xǐ", "sp": "xi", "r": "艸", "sc": 14, "l": 5, "f": 7599},
  {"h": "遡", "p": "sù", "sp": "su", "r": "辵", "sc": 13, "l": 5, "f": 7600},
  {"h": "郋", "p": "xí", "sp": "xi", "r": "邑", "sc": 8, "l": 5, "f": 7601},
  {"h": "鄜", "p": "fū", "sp": "fu", "r": "邑", "sc": 13, "l": 5, "f": 7602},
  {"h": "錞", "p": "chún", "sp": "chun", "r": "金", "sc": 16, "l": 5, "f": 7603},
  {"h": "靣", "p": "miàn", "sp": "mian", "r": "面", "sc": 8, "l": 5, "f": 7604},
  {"h": "鸂", "p": "xī", "sp": "xi", "r": "鳥", "sc": 24, "l": 5, "f": 7605},
  {"h": "麐", "p": "lín", "sp": "lin", "r": "鹿", "sc": 18, "l": 5, "f": 7606},
  {"h": "麚", "p": "jiā", "sp": "jia", "r": "鹿", "sc": 20, "l": 5, "f": 7607},
  {"h": "齧", "p": "niè", "sp": "nie", "r": "齒", "sc": 21, "l": 5, "f": 7608},
  {"h": "䌷", "p": "chōu", "sp": "chou", "r": "", "sc": 11, "l": 5, "f": 7609},
  {"h": "睊", "p": "juàn", "sp": "juan", "r": "目", "sc": 12, "l": 5, "f": 7610},
  {"h": "砢", "p": "kē", "sp": "ke", "r": "石", "sc": 10, "l": 5, "f": 7611},
  {"h": "碻", "p": "què", "sp": "que", "r": "石", "sc": 15, "l": 5, "f": 7612},
  {"h": "絃", "p": "xián", "sp": "xian", "r": "糸", "sc": 11, "l": 5, "f": 7613},
  {"h": "茝", "p": "chǎi", "sp": "chai", "r": "艸", "sc": 10, "l": 5, "f": 7614},
  {"h": "菴", "p": "ān", "sp": "an", "r": "艸", "sc": 11, "l": 5, "f": 7615},
  {"h": "袥", "p": "tuō", "sp": "tuo", "r": "衣", "sc": 10, "l": 5, "f": 7616},
  {"h": "襍", "p": "zá", "sp": "za", "r": "衣", "sc": 17, "l": 5, "f": 7617},
  {"h": "覗", "p": "sì", "sp": "si", "r": "見", "sc": 12, "l": 5, "f": 7618},
  {"h": "詩", "p": "shī", "sp": "shi", "r": "言", "sc": 13, "l": 5, "f": 7619},
  {
    "h": "諥",
    "p": "zhòng",
    "sp": "zhong",
    "r": "言",
    "sc": 16,
    "l": 5,
    "f": 7620
  },
  {"h": "跂", "p": "qí", "sp": "qi", "r": "足", "sc": 11, "l": 5, "f": 7621},
  {"h": "辵", "p": "chuò", "sp": "chuo", "r": "辵", "sc": 7, "l": 5, "f": 7622},
  {"h": "邅", "p": "zhān", "sp": "zhan", "r": "辵", "sc": 16, "l": 5, "f": 7623},
  {"h": "醨", "p": "lí", "sp": "li", "r": "酉", "sc": 17, "l": 5, "f": 7624},
  {"h": "頍", "p": "kuǐ", "sp": "kui", "r": "頁", "sc": 13, "l": 5, "f": 7625},
  {"h": "疰", "p": "zhù", "sp": "zhu", "r": "疒", "sc": 10, "l": 5, "f": 7626},
  {"h": "觜", "p": "zī", "sp": "zi", "r": "角", "sc": 13, "l": 5, "f": 7627},
  {"h": "鴿", "p": "gē", "sp": "ge", "r": "鳥", "sc": 17, "l": 5, "f": 7628},
  {"h": "笭", "p": "líng", "sp": "ling", "r": "竹", "sc": 11, "l": 5, "f": 7629},
  {"h": "糷", "p": "làn", "sp": "lan", "r": "米", "sc": 26, "l": 5, "f": 7630},
  {"h": "纘", "p": "zuǎn", "sp": "zuan", "r": "糸", "sc": 25, "l": 5, "f": 7631},
  {"h": "翉", "p": "běn", "sp": "ben", "r": "羽", "sc": 11, "l": 5, "f": 7632},
  {"h": "胵", "p": "chī", "sp": "chi", "r": "肉", "sc": 10, "l": 5, "f": 7633},
  {"h": "芖", "p": "zhì", "sp": "zhi", "r": "艸", "sc": 6, "l": 5, "f": 7634},
  {
    "h": "蔣",
    "p": "jiǎng",
    "sp": "jiang",
    "r": "艸",
    "sc": 14,
    "l": 5,
    "f": 7635
  },
  {"h": "蔦", "p": "niǎo", "sp": "niao", "r": "艸", "sc": 14, "l": 5, "f": 7636},
  {"h": "蛚", "p": "liè", "sp": "lie", "r": "虫", "sc": 12, "l": 5, "f": 7637},
  {"h": "蛯", "p": "lao", "sp": "lao", "r": "虫", "sc": 12, "l": 5, "f": 7638},
  {"h": "蜏", "p": "yǒu", "sp": "you", "r": "虫", "sc": 13, "l": 5, "f": 7639},
  {"h": "讞", "p": "yàn", "sp": "yan", "r": "言", "sc": 27, "l": 5, "f": 7640},
  {"h": "賲", "p": "bǎo", "sp": "bao", "r": "貝", "sc": 16, "l": 5, "f": 7641},
  {"h": "钃", "p": "shǔ", "sp": "shu", "r": "金", "sc": 29, "l": 5, "f": 7642},
  {"h": "餈", "p": "cí", "sp": "ci", "r": "食", "sc": 15, "l": 5, "f": 7643},
  {"h": "馧", "p": "yūn", "sp": "yun", "r": "香", "sc": 18, "l": 5, "f": 7644},
  {"h": "麤", "p": "cū", "sp": "cu", "r": "鹿", "sc": 33, "l": 5, "f": 7645},
  {"h": "鼶", "p": "sī", "sp": "si", "r": "鼠", "sc": 23, "l": 5, "f": 7646},
  {"h": "禱", "p": "dǎo", "sp": "dao", "r": "示", "sc": 18, "l": 5, "f": 7647},
  {"h": "臓", "p": "zàng", "sp": "zang", "r": "肉", "sc": 18, "l": 5, "f": 7648},
  {"h": "鈍", "p": "dùn", "sp": "dun", "r": "金", "sc": 12, "l": 5, "f": 7649},
  {"h": "錫", "p": "xī", "sp": "xi", "r": "金", "sc": 16, "l": 5, "f": 7650},
  {"h": "紛", "p": "fēn", "sp": "fen", "r": "糸", "sc": 10, "l": 5, "f": 7651},
  {"h": "遙", "p": "yáo", "sp": "yao", "r": "辵", "sc": 13, "l": 5, "f": 7652},
  {"h": "螣", "p": "tè", "sp": "te", "r": "虫", "sc": 16, "l": 5, "f": 7653},
  {"h": "铽", "p": "tè", "sp": "te", "r": "金", "sc": 12, "l": 5, "f": 7654},
  {"h": "竝", "p": "bìng", "sp": "bing", "r": "立", "sc": 10, "l": 5, "f": 7655},
  {"h": "粊", "p": "bì", "sp": "bi", "r": "米", "sc": 10, "l": 5, "f": 7656},
  {"h": "紗", "p": "shā", "sp": "sha", "r": "糸", "sc": 10, "l": 5, "f": 7657},
  {"h": "紮", "p": "zā", "sp": "za", "r": "糸", "sc": 11, "l": 5, "f": 7658},
  {"h": "絺", "p": "chī", "sp": "chi", "r": "糸", "sc": 13, "l": 5, "f": 7659},
  {"h": "絿", "p": "qiú", "sp": "qiu", "r": "糸", "sc": 13, "l": 5, "f": 7660},
  {"h": "緯", "p": "wěi", "sp": "wei", "r": "糸", "sc": 15, "l": 5, "f": 7661},
  {"h": "脦", "p": "de", "sp": "de", "r": "肉", "sc": 11, "l": 5, "f": 7662},
  {"h": "臇", "p": "juǎn", "sp": "juan", "r": "肉", "sc": 16, "l": 5, "f": 7663},
  {"h": "臋", "p": "tún", "sp": "tun", "r": "肉", "sc": 19, "l": 5, "f": 7664},
  {"h": "葍", "p": "fú", "sp": "fu", "r": "艸", "sc": 12, "l": 5, "f": 7665},
  {"h": "葞", "p": "mǐ", "sp": "mi", "r": "艸", "sc": 12, "l": 5, "f": 7666},
  {"h": "葯", "p": "yào", "sp": "yao", "r": "艸", "sc": 12, "l": 5, "f": 7667},
  {"h": "葻", "p": "lán", "sp": "lan", "r": "艸", "sc": 12, "l": 5, "f": 7668},
  {"h": "藙", "p": "yì", "sp": "yi", "r": "艸", "sc": 18, "l": 5, "f": 7669},
  {"h": "蜔", "p": "diàn", "sp": "dian", "r": "虫", "sc": 13, "l": 5, "f": 7670},
  {"h": "螕", "p": "bī", "sp": "bi", "r": "虫", "sc": 16, "l": 5, "f": 7671},
  {"h": "蟻", "p": "yǐ", "sp": "yi", "r": "虫", "sc": 19, "l": 5, "f": 7672},
  {"h": "襲", "p": "xí", "sp": "xi", "r": "衣", "sc": 22, "l": 5, "f": 7673},
  {"h": "觼", "p": "jué", "sp": "jue", "r": "角", "sc": 21, "l": 5, "f": 7674},
  {"h": "觿", "p": "xī", "sp": "xi", "r": "角", "sc": 25, "l": 5, "f": 7675},
  {"h": "詼", "p": "huī", "sp": "hui", "r": "言", "sc": 13, "l": 5, "f": 7676},
  {"h": "諝", "p": "xū", "sp": "xu", "r": "言", "sc": 16, "l": 5, "f": 7677},
  {"h": "貨", "p": "huò", "sp": "huo", "r": "貝", "sc": 11, "l": 5, "f": 7678},
  {"h": "賀", "p": "hè", "sp": "he", "r": "貝", "sc": 12, "l": 5, "f": 7679},
  {"h": "趧", "p": "tí", "sp": "ti", "r": "走", "sc": 16, "l": 5, "f": 7680},
  {"h": "跉", "p": "líng", "sp": "ling", "r": "足", "sc": 12, "l": 5, "f": 7681},
  {"h": "躍", "p": "yuè", "sp": "yue", "r": "足", "sc": 21, "l": 5, "f": 7682},
  {"h": "軒", "p": "xuān", "sp": "xuan", "r": "車", "sc": 10, "l": 5, "f": 7683},
  {"h": "輪", "p": "lún", "sp": "lun", "r": "車", "sc": 15, "l": 5, "f": 7684},
  {"h": "邉", "p": "biān", "sp": "bian", "r": "辵", "sc": 17, "l": 5, "f": 7685},
  {"h": "鄶", "p": "kuài", "sp": "kuai", "r": "邑", "sc": 15, "l": 5, "f": 7686},
  {"h": "酁", "p": "chán", "sp": "chan", "r": "邑", "sc": 19, "l": 5, "f": 7687},
  {
    "h": "釀",
    "p": "niàng",
    "sp": "niang",
    "r": "酉",
    "sc": 24,
    "l": 5,
    "f": 7688
  },
  {"h": "鈾", "p": "yóu", "sp": "you", "r": "金", "sc": 13, "l": 5, "f": 7689},
  {"h": "鈿", "p": "tián", "sp": "tian", "r": "金", "sc": 13, "l": 5, "f": 7690},
  {"h": "鍊", "p": "liàn", "sp": "lian", "r": "金", "sc": 17, "l": 5, "f": 7691},
  {
    "h": "闀",
    "p": "xiàng",
    "sp": "xiang",
    "r": "門",
    "sc": 17,
    "l": 5,
    "f": 7692
  },
  {"h": "闄", "p": "yǎo", "sp": "yao", "r": "門", "sc": 17, "l": 5, "f": 7693},
  {"h": "腽", "p": "wà", "sp": "wa", "r": "肉", "sc": 13, "l": 5, "f": 7694},
  {"h": "霧", "p": "wù", "sp": "wu", "r": "雨", "sc": 18, "l": 5, "f": 7695},
  {"h": "韌", "p": "rèn", "sp": "ren", "r": "韋", "sc": 12, "l": 5, "f": 7696},
  {"h": "頎", "p": "qí", "sp": "qi", "r": "頁", "sc": 13, "l": 5, "f": 7697},
  {"h": "額", "p": "é", "sp": "e", "r": "頁", "sc": 18, "l": 5, "f": 7698},
  {"h": "顫", "p": "chàn", "sp": "chan", "r": "頁", "sc": 22, "l": 5, "f": 7699},
  {"h": "顬", "p": "rú", "sp": "ru", "r": "頁", "sc": 23, "l": 5, "f": 7700},
  {"h": "髿", "p": "suō", "sp": "suo", "r": "髟", "sc": 17, "l": 5, "f": 7701},
  {"h": "鴣", "p": "gū", "sp": "gu", "r": "鳥", "sc": 16, "l": 5, "f": 7702},
  {"h": "鶝", "p": "fú", "sp": "fu", "r": "鳥", "sc": 20, "l": 5, "f": 7703},
  {"h": "鷣", "p": "yín", "sp": "yin", "r": "鳥", "sc": 23, "l": 5, "f": 7704},
  {"h": "齋", "p": "zhāi", "sp": "zhai", "r": "齊", "sc": 17, "l": 5, "f": 7705},
  {"h": "鋊", "p": "yù", "sp": "yu", "r": "金", "sc": 15, "l": 5, "f": 7706},
  {"h": "隤", "p": "tuí", "sp": "tui", "r": "阜", "sc": 14, "l": 5, "f": 7707},
  {"h": "艦", "p": "jiàn", "sp": "jian", "r": "舟", "sc": 20, "l": 5, "f": 7708},
  {"h": "譟", "p": "zào", "sp": "zao", "r": "言", "sc": 20, "l": 5, "f": 7709},
  {"h": "蒉", "p": "kuì", "sp": "kui", "r": "艸", "sc": 12, "l": 5, "f": 7710},
  {"h": "磜", "p": "qì", "sp": "qi", "r": "石", "sc": 16, "l": 5, "f": 7711},
  {"h": "綼", "p": "bì", "sp": "bi", "r": "糸", "sc": 14, "l": 5, "f": 7712},
  {"h": "輋", "p": "shē", "sp": "she", "r": "車", "sc": 13, "l": 5, "f": 7713},
  {"h": "頨", "p": "yǔ", "sp": "yu", "r": "頁", "sc": 15, "l": 5, "f": 7714},
  {"h": "閗", "p": "dòu", "sp": "dou", "r": "門", "sc": 12, "l": 5, "f": 7715},
  {"h": "鮍", "p": "pī", "sp": "pi", "r": "魚", "sc": 16, "l": 5, "f": 7716},
  {"h": "鯡", "p": "fèi", "sp": "fei", "r": "魚", "sc": 19, "l": 5, "f": 7717},
  {"h": "鲹", "p": "shēn", "sp": "shen", "r": "魚", "sc": 16, "l": 5, "f": 7718},
  {"h": "禣", "p": "fu", "sp": "fu", "r": "示", "sc": 14, "l": 5, "f": 7719},
  {"h": "粏", "p": "tai", "sp": "tai", "r": "米", "sc": 10, "l": 5, "f": 7720},
  {"h": "綿", "p": "mián", "sp": "mian", "r": "糸", "sc": 14, "l": 5, "f": 7721},
  {"h": "臐", "p": "xūn", "sp": "xun", "r": "肉", "sc": 18, "l": 5, "f": 7722},
  {"h": "迏", "p": "tì", "sp": "ti", "r": "辵", "sc": 7, "l": 5, "f": 7723},
  {"h": "鋾", "p": "táo", "sp": "tao", "r": "金", "sc": 16, "l": 5, "f": 7724},
  {"h": "鐇", "p": "fán", "sp": "fan", "r": "金", "sc": 20, "l": 5, "f": 7725},
  {"h": "铻", "p": "wú", "sp": "wu", "r": "金", "sc": 12, "l": 5, "f": 7726},
  {"h": "鮼", "p": "qīn", "sp": "qin", "r": "魚", "sc": 18, "l": 5, "f": 7727},
  {"h": "鼈", "p": "biē", "sp": "bie", "r": "黽", "sc": 24, "l": 5, "f": 7728},
  {"h": "込", "p": "ru", "sp": "ru", "r": "辵", "sc": 5, "l": 5, "f": 7729},
  {"h": "篰", "p": "bù", "sp": "bu", "r": "竹", "sc": 16, "l": 5, "f": 7730},
  {"h": "痄", "p": "zhà", "sp": "zha", "r": "疒", "sc": 10, "l": 5, "f": 7731},
  {"h": "睄", "p": "shào", "sp": "shao", "r": "目", "sc": 12, "l": 5, "f": 7732},
  {"h": "磡", "p": "kàn", "sp": "kan", "r": "石", "sc": 16, "l": 5, "f": 7733},
  {"h": "窔", "p": "yào", "sp": "yao", "r": "穴", "sc": 11, "l": 5, "f": 7734},
  {"h": "筊", "p": "xiáo", "sp": "xiao", "r": "竹", "sc": 12, "l": 5, "f": 7735},
  {"h": "羺", "p": "nóu", "sp": "nou", "r": "羊", "sc": 20, "l": 5, "f": 7736},
  {"h": "腪", "p": "yùn", "sp": "yun", "r": "肉", "sc": 13, "l": 5, "f": 7737},
  {"h": "臢", "p": "zā", "sp": "za", "r": "肉", "sc": 23, "l": 5, "f": 7738},
  {"h": "荄", "p": "gāi", "sp": "gai", "r": "艸", "sc": 9, "l": 5, "f": 7739},
  {"h": "蓡", "p": "shēn", "sp": "shen", "r": "艸", "sc": 13, "l": 5, "f": 7740},
  {"h": "蝀", "p": "dōng", "sp": "dong", "r": "虫", "sc": 14, "l": 5, "f": 7741},
  {"h": "詥", "p": "hé", "sp": "he", "r": "言", "sc": 13, "l": 5, "f": 7742},
  {"h": "豀", "p": "xī", "sp": "xi", "r": "谷", "sc": 17, "l": 5, "f": 7743},
  {"h": "苠", "p": "mín", "sp": "min", "r": "艸", "sc": 8, "l": 5, "f": 7744},
  {"h": "郞", "p": "láng", "sp": "lang", "r": "邑", "sc": 9, "l": 5, "f": 7745},
  {"h": "骒", "p": "kè", "sp": "ke", "r": "馬", "sc": 11, "l": 5, "f": 7746},
  {"h": "戤", "p": "gài", "sp": "gai", "r": "戈", "sc": 13, "l": 5, "f": 7747},
  {"h": "铞", "p": "diào", "sp": "diao", "r": "金", "sc": 11, "l": 5, "f": 7748},
  {"h": "耢", "p": "lào", "sp": "lao", "r": "耒", "sc": 13, "l": 5, "f": 7749},
  {"h": "麯", "p": "qū", "sp": "qu", "r": "麥", "sc": 17, "l": 5, "f": 7750},
  {"h": "秱", "p": "tóng", "sp": "tong", "r": "禾", "sc": 11, "l": 5, "f": 7751},
  {"h": "筼", "p": "yún", "sp": "yun", "r": "竹", "sc": 13, "l": 5, "f": 7752},
  {"h": "綈", "p": "tí", "sp": "ti", "r": "糸", "sc": 13, "l": 5, "f": 7753},
  {"h": "翬", "p": "huī", "sp": "hui", "r": "羽", "sc": 15, "l": 5, "f": 7754},
  {"h": "脧", "p": "juān", "sp": "juan", "r": "肉", "sc": 11, "l": 5, "f": 7755},
  {"h": "茣", "p": "wú", "sp": "wu", "r": "艸", "sc": 10, "l": 5, "f": 7756},
  {"h": "菒", "p": "gǎo", "sp": "gao", "r": "艸", "sc": 11, "l": 5, "f": 7757},
  {"h": "蓯", "p": "cōng", "sp": "cong", "r": "艸", "sc": 14, "l": 5, "f": 7758},
  {"h": "蕝", "p": "jué", "sp": "jue", "r": "艸", "sc": 15, "l": 5, "f": 7759},
  {"h": "詝", "p": "zhǔ", "sp": "zhu", "r": "言", "sc": 12, "l": 5, "f": 7760},
  {"h": "謱", "p": "lóu", "sp": "lou", "r": "言", "sc": 18, "l": 5, "f": 7761},
  {"h": "冁", "p": "chǎn", "sp": "chan", "r": "八", "sc": 18, "l": 5, "f": 7762},
  {"h": "踧", "p": "cù", "sp": "cu", "r": "足", "sc": 15, "l": 5, "f": 7763},
  {"h": "茚", "p": "yìn", "sp": "yin", "r": "艸", "sc": 8, "l": 5, "f": 7764},
  {
    "h": "辌",
    "p": "liáng",
    "sp": "liang",
    "r": "車",
    "sc": 12,
    "l": 5,
    "f": 7765
  },
  {"h": "辒", "p": "wēn", "sp": "wen", "r": "車", "sc": 13, "l": 5, "f": 7766},
  {"h": "郪", "p": "qī", "sp": "qi", "r": "邑", "sc": 10, "l": 5, "f": 7767},
  {"h": "鄘", "p": "yōng", "sp": "yong", "r": "邑", "sc": 13, "l": 5, "f": 7768},
  {"h": "鄠", "p": "hù", "sp": "hu", "r": "邑", "sc": 13, "l": 5, "f": 7769},
  {"h": "帱", "p": "chóu", "sp": "chou", "r": "巾", "sc": 10, "l": 5, "f": 7770},
  {"h": "鈇", "p": "fū", "sp": "fu", "r": "金", "sc": 12, "l": 5, "f": 7771},
  {"h": "缲", "p": "qiāo", "sp": "qiao", "r": "糸", "sc": 16, "l": 5, "f": 7772},
  {"h": "雚", "p": "guàn", "sp": "guan", "r": "隹", "sc": 17, "l": 5, "f": 7773},
  {"h": "雟", "p": "xī", "sp": "xi", "r": "隹", "sc": 18, "l": 5, "f": 7774},
  {"h": "韺", "p": "yīng", "sp": "ying", "r": "音", "sc": 17, "l": 5, "f": 7775},
  {"h": "頵", "p": "yūn", "sp": "yun", "r": "頁", "sc": 16, "l": 5, "f": 7776},
  {"h": "顝", "p": "kuī", "sp": "kui", "r": "頁", "sc": 18, "l": 5, "f": 7777},
  {"h": "镘", "p": "màn", "sp": "man", "r": "金", "sc": 16, "l": 5, "f": 7778},
  {"h": "鬋", "p": "jiǎn", "sp": "jian", "r": "髟", "sc": 19, "l": 5, "f": 7779},
  {"h": "酾", "p": "shāi", "sp": "shai", "r": "酉", "sc": 14, "l": 5, "f": 7780},
  {"h": "鳣", "p": "zhān", "sp": "zhan", "r": "魚", "sc": 21, "l": 5, "f": 7781},
  {"h": "鴦", "p": "yāng", "sp": "yang", "r": "鳥", "sc": 16, "l": 5, "f": 7782},
  {"h": "麀", "p": "yōu", "sp": "you", "r": "鹿", "sc": 13, "l": 5, "f": 7783},
  {"h": "麉", "p": "jiān", "sp": "jian", "r": "鹿", "sc": 17, "l": 5, "f": 7784},
  {"h": "麊", "p": "mí", "sp": "mi", "r": "鹿", "sc": 17, "l": 5, "f": 7785},
  {"h": "麣", "p": "yán", "sp": "yan", "r": "鹿", "sc": 30, "l": 5, "f": 7786},
  {"h": "藬", "p": "tuī", "sp": "tui", "r": "艸", "sc": 17, "l": 5, "f": 7787},
  {
    "h": "鏘",
    "p": "qiāng",
    "sp": "qiang",
    "r": "金",
    "sc": 19,
    "l": 5,
    "f": 7788
  },
  {"h": "鞚", "p": "kòng", "sp": "kong", "r": "革", "sc": 17, "l": 5, "f": 7789},
  {"h": "飔", "p": "sī", "sp": "si", "r": "風", "sc": 13, "l": 5, "f": 7790},
  {"h": "鴃", "p": "jué", "sp": "jue", "r": "鳥", "sc": 15, "l": 5, "f": 7791},
  {"h": "齏", "p": "jī", "sp": "ji", "r": "齊", "sc": 23, "l": 5, "f": 7792},
  {"h": "礄", "p": "qiáo", "sp": "qiao", "r": "石", "sc": 17, "l": 5, "f": 7793},
  {"h": "誥", "p": "gào", "sp": "gao", "r": "言", "sc": 14, "l": 5, "f": 7794},
  {"h": "醲", "p": "nóng", "sp": "nong", "r": "酉", "sc": 20, "l": 5, "f": 7795},
  {"h": "毳", "p": "cuì", "sp": "cui", "r": "毛", "sc": 12, "l": 5, "f": 7796},
  {"h": "築", "p": "zhú", "sp": "zhu", "r": "竹", "sc": 16, "l": 5, "f": 7797},
  {"h": "絫", "p": "lěi", "sp": "lei", "r": "糸", "sc": 12, "l": 5, "f": 7798},
  {"h": "脺", "p": "cuì", "sp": "cui", "r": "肉", "sc": 12, "l": 5, "f": 7799},
  {"h": "苧", "p": "níng", "sp": "ning", "r": "艸", "sc": 8, "l": 5, "f": 7800},
  {"h": "蚘", "p": "huí", "sp": "hui", "r": "虫", "sc": 10, "l": 5, "f": 7801},
  {"h": "蝺", "p": "qǔ", "sp": "qu", "r": "虫", "sc": 15, "l": 5, "f": 7802},
  {"h": "蠶", "p": "cán", "sp": "can", "r": "虫", "sc": 24, "l": 5, "f": 7803},
  {"h": "裀", "p": "yīn", "sp": "yin", "r": "衣", "sc": 11, "l": 5, "f": 7804},
  {"h": "詧", "p": "chá", "sp": "cha", "r": "言", "sc": 13, "l": 5, "f": 7805},
  {"h": "貍", "p": "lí", "sp": "li", "r": "豸", "sc": 14, "l": 5, "f": 7806},
  {"h": "踘", "p": "jū", "sp": "ju", "r": "足", "sc": 15, "l": 5, "f": 7807},
  {"h": "蹷", "p": "jué", "sp": "jue", "r": "足", "sc": 19, "l": 5, "f": 7808},
  {
    "h": "蟓",
    "p": "xiàng",
    "sp": "xiang",
    "r": "虫",
    "sc": 17,
    "l": 5,
    "f": 7809
  },
  {"h": "盦", "p": "ān", "sp": "an", "r": "皿", "sc": 16, "l": 5, "f": 7810},
  {"h": "瞆", "p": "kuì", "sp": "kui", "r": "目", "sc": 14, "l": 5, "f": 7811},
  {"h": "矰", "p": "zēng", "sp": "zeng", "r": "矢", "sc": 17, "l": 5, "f": 7812},
  {"h": "磥", "p": "lěi", "sp": "lei", "r": "石", "sc": 16, "l": 5, "f": 7813},
  {"h": "祮", "p": "gào", "sp": "gao", "r": "示", "sc": 10, "l": 5, "f": 7814},
  {"h": "稛", "p": "kǔn", "sp": "kun", "r": "禾", "sc": 13, "l": 5, "f": 7815},
  {"h": "箾", "p": "shuò", "sp": "shuo", "r": "竹", "sc": 15, "l": 5, "f": 7816},
  {"h": "翂", "p": "fēn", "sp": "fen", "r": "羽", "sc": 10, "l": 5, "f": 7817},
  {"h": "胊", "p": "qú", "sp": "qu", "r": "肉", "sc": 9, "l": 5, "f": 7818},
  {"h": "蓺", "p": "yì", "sp": "yi", "r": "艸", "sc": 14, "l": 5, "f": 7819},
  {"h": "薾", "p": "ěr", "sp": "er", "r": "艸", "sc": 17, "l": 5, "f": 7820},
  {"h": "襱", "p": "lóng", "sp": "long", "r": "衣", "sc": 21, "l": 5, "f": 7821},
  {"h": "覅", "p": "fiào", "sp": "fiao", "r": "襾", "sc": 13, "l": 5, "f": 7822},
  {"h": "讏", "p": "wèi", "sp": "wei", "r": "言", "sc": 23, "l": 5, "f": 7823},
  {"h": "跅", "p": "tuò", "sp": "tuo", "r": "足", "sc": 12, "l": 5, "f": 7824},
  {"h": "踆", "p": "cūn", "sp": "cun", "r": "足", "sc": 14, "l": 5, "f": 7825},
  {"h": "蹔", "p": "zàn", "sp": "zan", "r": "足", "sc": 18, "l": 5, "f": 7826},
  {"h": "軐", "p": "xìn", "sp": "xin", "r": "車", "sc": 10, "l": 5, "f": 7827},
  {"h": "轪", "p": "dài", "sp": "dai", "r": "車", "sc": 7, "l": 5, "f": 7828},
  {"h": "鄤", "p": "màn", "sp": "man", "r": "邑", "sc": 13, "l": 5, "f": 7829},
  {"h": "醰", "p": "tán", "sp": "tan", "r": "酉", "sc": 19, "l": 5, "f": 7830},
  {"h": "鑟", "p": "dú", "sp": "du", "r": "金", "sc": 23, "l": 5, "f": 7831},
  {"h": "铔", "p": "yà", "sp": "ya", "r": "金", "sc": 11, "l": 5, "f": 7832},
  {"h": "閛", "p": "pēng", "sp": "peng", "r": "門", "sc": 13, "l": 5, "f": 7833},
  {"h": "陙", "p": "chún", "sp": "chun", "r": "阜", "sc": 9, "l": 5, "f": 7834},
  {"h": "霳", "p": "lóng", "sp": "long", "r": "雨", "sc": 19, "l": 5, "f": 7835},
  {"h": "顑", "p": "kǎn", "sp": "kan", "r": "頁", "sc": 18, "l": 5, "f": 7836},
  {"h": "镤", "p": "pú", "sp": "pu", "r": "金", "sc": 17, "l": 5, "f": 7837},
  {"h": "鵲", "p": "què", "sp": "que", "r": "鳥", "sc": 19, "l": 5, "f": 7838},
  {"h": "竚", "p": "zhù", "sp": "zhu", "r": "立", "sc": 10, "l": 5, "f": 7839},
  {"h": "紩", "p": "zhì", "sp": "zhi", "r": "糸", "sc": 11, "l": 5, "f": 7840},
  {"h": "聦", "p": "cōng", "sp": "cong", "r": "耳", "sc": 15, "l": 5, "f": 7841},
  {"h": "餅", "p": "bǐng", "sp": "bing", "r": "食", "sc": 14, "l": 5, "f": 7842},
  {"h": "鶄", "p": "jīng", "sp": "jing", "r": "鳥", "sc": 19, "l": 5, "f": 7843},
  {"h": "競", "p": "jìng", "sp": "jing", "r": "立", "sc": 20, "l": 5, "f": 7844},
  {"h": "簷", "p": "yán", "sp": "yan", "r": "竹", "sc": 19, "l": 5, "f": 7845},
  {"h": "糵", "p": "niè", "sp": "nie", "r": "米", "sc": 22, "l": 5, "f": 7846},
  {"h": "紖", "p": "zhèn", "sp": "zhen", "r": "糸", "sc": 10, "l": 5, "f": 7847},
  {"h": "訪", "p": "fǎng", "sp": "fang", "r": "言", "sc": 11, "l": 5, "f": 7848},
  {"h": "豎", "p": "shù", "sp": "shu", "r": "豆", "sc": 15, "l": 5, "f": 7849},
  {"h": "趦", "p": "zī", "sp": "zi", "r": "走", "sc": 16, "l": 5, "f": 7850},
  {"h": "隸", "p": "lì", "sp": "li", "r": "隶", "sc": 17, "l": 5, "f": 7851},
  {"h": "禠", "p": "sī", "sp": "si", "r": "示", "sc": 14, "l": 5, "f": 7852},
  {"h": "籓", "p": "fān", "sp": "fan", "r": "竹", "sc": 21, "l": 5, "f": 7853},
  {"h": "镥", "p": "lǔ", "sp": "lu", "r": "金", "sc": 17, "l": 5, "f": 7854},
  {"h": "盜", "p": "dào", "sp": "dao", "r": "皿", "sc": 12, "l": 5, "f": 7855},
  {"h": "矝", "p": "lín", "sp": "lin", "r": "矛", "sc": 10, "l": 5, "f": 7856},
  {"h": "笖", "p": "yǐ", "sp": "yi", "r": "竹", "sc": 10, "l": 5, "f": 7857},
  {"h": "笧", "p": "cè", "sp": "ce", "r": "竹", "sc": 11, "l": 5, "f": 7858},
  {"h": "簢", "p": "mǐn", "sp": "min", "r": "竹", "sc": 18, "l": 5, "f": 7859},
  {"h": "簻", "p": "zhuā", "sp": "zhua", "r": "竹", "sc": 17, "l": 5, "f": 7860},
  {"h": "籌", "p": "chóu", "sp": "chou", "r": "竹", "sc": 20, "l": 5, "f": 7861},
  {"h": "粍", "p": "zhé", "sp": "zhe", "r": "米", "sc": 10, "l": 5, "f": 7862},
  {"h": "綍", "p": "fú", "sp": "fu", "r": "糸", "sc": 13, "l": 5, "f": 7863},
  {"h": "縕", "p": "yūn", "sp": "yun", "r": "糸", "sc": 16, "l": 5, "f": 7864},
  {"h": "罋", "p": "wèng", "sp": "weng", "r": "缶", "sc": 19, "l": 5, "f": 7865},
  {"h": "罷", "p": "bà", "sp": "ba", "r": "网", "sc": 15, "l": 5, "f": 7866},
  {"h": "翜", "p": "shà", "sp": "sha", "r": "羽", "sc": 13, "l": 5, "f": 7867},
  {"h": "膰", "p": "fán", "sp": "fan", "r": "肉", "sc": 16, "l": 5, "f": 7868},
  {"h": "膲", "p": "jiāo", "sp": "jiao", "r": "肉", "sc": 16, "l": 5, "f": 7869},
  {"h": "菂", "p": "dì", "sp": "di", "r": "艸", "sc": 11, "l": 5, "f": 7870},
  {"h": "葦", "p": "wěi", "sp": "wei", "r": "艸", "sc": 12, "l": 5, "f": 7871},
  {"h": "薢", "p": "xiè", "sp": "xie", "r": "艸", "sc": 16, "l": 5, "f": 7872},
  {"h": "藷", "p": "shǔ", "sp": "shu", "r": "艸", "sc": 18, "l": 5, "f": 7873},
  {"h": "螞", "p": "mǎ", "sp": "ma", "r": "虫", "sc": 16, "l": 5, "f": 7874},
  {"h": "蠻", "p": "mán", "sp": "man", "r": "虫", "sc": 25, "l": 5, "f": 7875},
  {"h": "袀", "p": "jūn", "sp": "jun", "r": "衣", "sc": 9, "l": 5, "f": 7876},
  {"h": "袐", "p": "bì", "sp": "bi", "r": "衣", "sc": 10, "l": 5, "f": 7877},
  {"h": "覍", "p": "biàn", "sp": "bian", "r": "見", "sc": 10, "l": 5, "f": 7878},
  {"h": "訌", "p": "hòng", "sp": "hong", "r": "言", "sc": 10, "l": 5, "f": 7879},
  {"h": "訛", "p": "é", "sp": "e", "r": "言", "sc": 11, "l": 5, "f": 7880},
  {"h": "詿", "p": "guà", "sp": "gua", "r": "言", "sc": 13, "l": 5, "f": 7881},
  {"h": "諄", "p": "zhūn", "sp": "zhun", "r": "言", "sc": 15, "l": 5, "f": 7882},
  {"h": "譜", "p": "pǔ", "sp": "pu", "r": "言", "sc": 19, "l": 5, "f": 7883},
  {"h": "豈", "p": "qǐ", "sp": "qi", "r": "豆", "sc": 10, "l": 5, "f": 7884},
  {"h": "賢", "p": "xián", "sp": "xian", "r": "貝", "sc": 15, "l": 5, "f": 7885},
  {"h": "賹", "p": "yì", "sp": "yi", "r": "貝", "sc": 17, "l": 5, "f": 7886},
  {"h": "遷", "p": "qiān", "sp": "qian", "r": "辵", "sc": 15, "l": 5, "f": 7887},
  {"h": "釛", "p": "bā", "sp": "ba", "r": "金", "sc": 10, "l": 5, "f": 7888},
  {"h": "鈣", "p": "gài", "sp": "gai", "r": "金", "sc": 12, "l": 5, "f": 7889},
  {"h": "鈱", "p": "mín", "sp": "min", "r": "金", "sc": 13, "l": 5, "f": 7890},
  {"h": "鈶", "p": "sì", "sp": "si", "r": "金", "sc": 13, "l": 5, "f": 7891},
  {"h": "鍛", "p": "duàn", "sp": "duan", "r": "金", "sc": 17, "l": 5, "f": 7892},
  {"h": "鍦", "p": "shī", "sp": "shi", "r": "金", "sc": 17, "l": 5, "f": 7893},
  {"h": "鐵", "p": "tiě", "sp": "tie", "r": "金", "sc": 21, "l": 5, "f": 7894},
  {"h": "閯", "p": "shà", "sp": "sha", "r": "門", "sc": 15, "l": 5, "f": 7895},
  {"h": "閺", "p": "wén", "sp": "wen", "r": "門", "sc": 16, "l": 5, "f": 7896},
  {"h": "閻", "p": "yán", "sp": "yan", "r": "門", "sc": 16, "l": 5, "f": 7897},
  {"h": "隂", "p": "yīn", "sp": "yin", "r": "阜", "sc": 11, "l": 5, "f": 7898},
  {"h": "頌", "p": "sòng", "sp": "song", "r": "頁", "sc": 13, "l": 5, "f": 7899},
  {"h": "頰", "p": "jiá", "sp": "jia", "r": "頁", "sc": 16, "l": 5, "f": 7900},
  {"h": "顆", "p": "kē", "sp": "ke", "r": "頁", "sc": 17, "l": 5, "f": 7901},
  {"h": "顪", "p": "huì", "sp": "hui", "r": "頁", "sc": 22, "l": 5, "f": 7902},
  {"h": "飼", "p": "sì", "sp": "si", "r": "食", "sc": 13, "l": 5, "f": 7903},
  {"h": "駕", "p": "jià", "sp": "jia", "r": "馬", "sc": 15, "l": 5, "f": 7904},
  {"h": "駝", "p": "tuó", "sp": "tuo", "r": "馬", "sc": 15, "l": 5, "f": 7905},
  {"h": "驌", "p": "sù", "sp": "su", "r": "馬", "sc": 23, "l": 5, "f": 7906},
  {"h": "驘", "p": "luó", "sp": "luo", "r": "馬", "sc": 23, "l": 5, "f": 7907},
  {"h": "鰠", "p": "sāo", "sp": "sao", "r": "魚", "sc": 20, "l": 5, "f": 7908},
  {"h": "鰹", "p": "jiān", "sp": "jian", "r": "魚", "sc": 22, "l": 5, "f": 7909},
  {"h": "鲗", "p": "zéi", "sp": "zei", "r": "魚", "sc": 14, "l": 5, "f": 7910},
  {"h": "鶒", "p": "chì", "sp": "chi", "r": "鳥", "sc": 20, "l": 5, "f": 7911},
  {"h": "鶗", "p": "tí", "sp": "ti", "r": "鳥", "sc": 20, "l": 5, "f": 7912},
  {"h": "鷥", "p": "sī", "sp": "si", "r": "鳥", "sc": 23, "l": 5, "f": 7913},
  {"h": "鷹", "p": "yīng", "sp": "ying", "r": "鳥", "sc": 24, "l": 5, "f": 7914},
  {"h": "齊", "p": "qí", "sp": "qi", "r": "齊", "sc": 14, "l": 5, "f": 7915},
  {"h": "硙", "p": "wéi", "sp": "wei", "r": "石", "sc": 11, "l": 5, "f": 7916},
  {"h": "禜", "p": "yǒng", "sp": "yong", "r": "示", "sc": 15, "l": 5, "f": 7917},
  {"h": "糢", "p": "mó", "sp": "mo", "r": "米", "sc": 16, "l": 5, "f": 7918},
  {"h": "藠", "p": "jiào", "sp": "jiao", "r": "艸", "sc": 18, "l": 5, "f": 7919},
  {"h": "镦", "p": "duì", "sp": "dui", "r": "金", "sc": 17, "l": 5, "f": 7920},
  {
    "h": "蕂",
    "p": "shèng",
    "sp": "sheng",
    "r": "艸",
    "sc": 15,
    "l": 5,
    "f": 7921
  },
  {"h": "霮", "p": "dàn", "sp": "dan", "r": "雨", "sc": 20, "l": 5, "f": 7922},
  {"h": "沩", "p": "wéi", "sp": "wei", "r": "水", "sc": 7, "l": 5, "f": 7923},
  {"h": "颼", "p": "sōu", "sp": "sou", "r": "風", "sc": 18, "l": 5, "f": 7924},
  {"h": "禬", "p": "guì", "sp": "gui", "r": "示", "sc": 17, "l": 5, "f": 7925},
  {"h": "鴪", "p": "yù", "sp": "yu", "r": "鳥", "sc": 16, "l": 5, "f": 7926},
  {"h": "鮨", "p": "yì", "sp": "yi", "r": "魚", "sc": 17, "l": 5, "f": 7927},
  {"h": "鱣", "p": "zhān", "sp": "zhan", "r": "魚", "sc": 24, "l": 5, "f": 7928},
  {"h": "盝", "p": "lù", "sp": "lu", "r": "皿", "sc": 13, "l": 5, "f": 7929},
  {"h": "眹", "p": "zhèn", "sp": "zhen", "r": "目", "sc": 11, "l": 5, "f": 7930},
  {"h": "睖", "p": "lèng", "sp": "leng", "r": "目", "sc": 13, "l": 5, "f": 7931},
  {"h": "磫", "p": "zōng", "sp": "zong", "r": "石", "sc": 16, "l": 5, "f": 7932},
  {"h": "禫", "p": "dàn", "sp": "dan", "r": "示", "sc": 16, "l": 5, "f": 7933},
  {"h": "繨", "p": "da", "sp": "da", "r": "糸", "sc": 18, "l": 5, "f": 7934},
  {"h": "薦", "p": "jiàn", "sp": "jian", "r": "艸", "sc": 16, "l": 5, "f": 7935},
  {"h": "褱", "p": "huái", "sp": "huai", "r": "衣", "sc": 16, "l": 5, "f": 7936},
  {"h": "鏏", "p": "wèi", "sp": "wei", "r": "金", "sc": 19, "l": 5, "f": 7937},
  {"h": "鮁", "p": "bō", "sp": "bo", "r": "魚", "sc": 16, "l": 5, "f": 7938},
  {"h": "鹓", "p": "yuān", "sp": "yuan", "r": "鳥", "sc": 13, "l": 5, "f": 7939},
  {"h": "繟", "p": "chǎn", "sp": "chan", "r": "糸", "sc": 18, "l": 5, "f": 7940},
  {"h": "繺", "p": "shǎi", "sp": "shai", "r": "糸", "sc": 19, "l": 5, "f": 7941},
  {"h": "沲", "p": "tuó", "sp": "tuo", "r": "水", "sc": 8, "l": 5, "f": 7942},
  {
    "h": "餦",
    "p": "zhāng",
    "sp": "zhang",
    "r": "食",
    "sc": 16,
    "l": 5,
    "f": 7943
  },
  {"h": "篤", "p": "dǔ", "sp": "du", "r": "竹", "sc": 16, "l": 5, "f": 7944},
  {"h": "踺", "p": "jiàn", "sp": "jian", "r": "足", "sc": 15, "l": 5, "f": 7945},
  {"h": "祦", "p": "wú", "sp": "wu", "r": "示", "sc": 11, "l": 5, "f": 7946},
  {"h": "篲", "p": "huì", "sp": "hui", "r": "竹", "sc": 17, "l": 5, "f": 7947},
  {"h": "糰", "p": "tuán", "sp": "tuan", "r": "米", "sc": 20, "l": 5, "f": 7948},
  {"h": "絙", "p": "huán", "sp": "huan", "r": "糸", "sc": 12, "l": 5, "f": 7949},
  {"h": "纄", "p": "péng", "sp": "peng", "r": "糸", "sc": 19, "l": 5, "f": 7950},
  {"h": "脝", "p": "hēng", "sp": "heng", "r": "肉", "sc": 11, "l": 5, "f": 7951},
  {"h": "茊", "p": "zī", "sp": "zi", "r": "艸", "sc": 8, "l": 5, "f": 7952},
  {"h": "菫", "p": "jǐn", "sp": "jin", "r": "艸", "sc": 11, "l": 5, "f": 7953},
  {"h": "蓞", "p": "dàn", "sp": "dan", "r": "艸", "sc": 16, "l": 5, "f": 7954},
  {"h": "酦", "p": "fā", "sp": "fa", "r": "酉", "sc": 12, "l": 5, "f": 7955},
  {"h": "醾", "p": "mí", "sp": "mi", "r": "酉", "sc": 24, "l": 5, "f": 7956},
  {"h": "锳", "p": "yīng", "sp": "ying", "r": "金", "sc": 13, "l": 5, "f": 7957},
  {"h": "钶", "p": "ē", "sp": "e", "r": "金", "sc": 10, "l": 5, "f": 7958},
  {"h": "鲼", "p": "fèn", "sp": "fen", "r": "魚", "sc": 17, "l": 5, "f": 7959},
  {"h": "齄", "p": "zhā", "sp": "zha", "r": "鼻", "sc": 23, "l": 5, "f": 7960},
  {"h": "鴨", "p": "yā", "sp": "ya", "r": "鳥", "sc": 16, "l": 5, "f": 7961},
  {"h": "鷧", "p": "yì", "sp": "yi", "r": "鳥", "sc": 23, "l": 5, "f": 7962},
  {"h": "㧟", "p": "kuǎi", "sp": "kuai", "r": "手", "sc": 8, "l": 5, "f": 7963},
  {"h": "祋", "p": "duì", "sp": "dui", "r": "示", "sc": 8, "l": 5, "f": 7964},
  {"h": "箯", "p": "biān", "sp": "bian", "r": "竹", "sc": 15, "l": 5, "f": 7965},
  {"h": "粐", "p": "hu", "sp": "hu", "r": "米", "sc": 10, "l": 5, "f": 7966},
  {"h": "粦", "p": "lín", "sp": "lin", "r": "米", "sc": 12, "l": 5, "f": 7967},
  {"h": "粭", "p": "he", "sp": "he", "r": "米", "sc": 12, "l": 5, "f": 7968},
  {"h": "紃", "p": "xún", "sp": "xun", "r": "糸", "sc": 9, "l": 5, "f": 7969},
  {"h": "脰", "p": "dòu", "sp": "dou", "r": "肉", "sc": 11, "l": 5, "f": 7970},
  {"h": "芃", "p": "péng", "sp": "peng", "r": "艸", "sc": 6, "l": 5, "f": 7971},
  {"h": "苽", "p": "gū", "sp": "gu", "r": "艸", "sc": 8, "l": 5, "f": 7972},
  {"h": "莿", "p": "cì", "sp": "ci", "r": "艸", "sc": 11, "l": 5, "f": 7973},
  {"h": "菆", "p": "zōu", "sp": "zou", "r": "艸", "sc": 11, "l": 5, "f": 7974},
  {"h": "藗", "p": "sù", "sp": "su", "r": "艸", "sc": 17, "l": 5, "f": 7975},
  {"h": "藧", "p": "huàn", "sp": "huan", "r": "艸", "sc": 18, "l": 5, "f": 7976},
  {"h": "袵", "p": "rèn", "sp": "ren", "r": "衣", "sc": 11, "l": 5, "f": 7977},
  {"h": "裯", "p": "chóu", "sp": "chou", "r": "衣", "sc": 13, "l": 5, "f": 7978},
  {"h": "褵", "p": "lí", "sp": "li", "r": "衣", "sc": 15, "l": 5, "f": 7979},
  {"h": "襆", "p": "fú", "sp": "fu", "r": "衣", "sc": 17, "l": 5, "f": 7980},
  {
    "h": "詤",
    "p": "huǎng",
    "sp": "huang",
    "r": "言",
    "sc": 13,
    "l": 5,
    "f": 7981
  },
  {"h": "譓", "p": "huì", "sp": "hui", "r": "言", "sc": 19, "l": 5, "f": 7982},
  {"h": "跊", "p": "mèi", "sp": "mei", "r": "足", "sc": 12, "l": 5, "f": 7983},
  {"h": "軈", "p": "ying", "sp": "ying", "r": "身", "sc": 24, "l": 5, "f": 7984},
  {"h": "苘", "p": "qǐng", "sp": "qing", "r": "艸", "sc": 8, "l": 5, "f": 7985},
  {"h": "迺", "p": "nǎi", "sp": "nai", "r": "辵", "sc": 9, "l": 5, "f": 7986},
  {"h": "掭", "p": "tiàn", "sp": "tian", "r": "手", "sc": 11, "l": 5, "f": 7987},
  {"h": "邇", "p": "ěr", "sp": "er", "r": "辵", "sc": 17, "l": 5, "f": 7988},
  {"h": "邘", "p": "yú", "sp": "yu", "r": "邑", "sc": 5, "l": 5, "f": 7989},
  {"h": "鄳", "p": "méng", "sp": "meng", "r": "邑", "sc": 15, "l": 5, "f": 7990},
  {"h": "猸", "p": "méi", "sp": "mei", "r": "犬", "sc": 12, "l": 5, "f": 7991},
  {"h": "艴", "p": "fú", "sp": "fu", "r": "色", "sc": 11, "l": 5, "f": 7992},
  {"h": "鑕", "p": "zhì", "sp": "zhi", "r": "金", "sc": 23, "l": 5, "f": 7993},
  {"h": "柃", "p": "líng", "sp": "ling", "r": "木", "sc": 9, "l": 5, "f": 7994},
  {"h": "闒", "p": "tà", "sp": "ta", "r": "門", "sc": 18, "l": 5, "f": 7995},
  {"h": "毹", "p": "shū", "sp": "shu", "r": "毛", "sc": 13, "l": 5, "f": 7996},
  {"h": "脶", "p": "luó", "sp": "luo", "r": "肉", "sc": 11, "l": 5, "f": 7997},
  {"h": "颋", "p": "tǐng", "sp": "ting", "r": "頁", "sc": 12, "l": 5, "f": 7998},
  {"h": "鵀", "p": "rén", "sp": "ren", "r": "鳥", "sc": 17, "l": 5, "f": 7999},
  {"h": "鶜", "p": "máo", "sp": "mao", "r": "鳥", "sc": 19, "l": 5, "f": 8000},
  {"h": "鶟", "p": "tú", "sp": "tu", "r": "鳥", "sc": 20, "l": 5, "f": 8001},
  {"h": "鷱", "p": "gāo", "sp": "gao", "r": "鳥", "sc": 23, "l": 5, "f": 8002},
  {"h": "鷲", "p": "jiù", "sp": "jiu", "r": "鳥", "sc": 23, "l": 5, "f": 8003},
  {"h": "鹍", "p": "kūn", "sp": "kun", "r": "鳥", "sc": 13, "l": 5, "f": 8004},
  {"h": "麖", "p": "jīng", "sp": "jing", "r": "鹿", "sc": 19, "l": 5, "f": 8005},
  {"h": "䦆", "p": "jué", "sp": "jue", "r": "", "sc": 28, "l": 5, "f": 8006},
  {"h": "軎", "p": "wèi", "sp": "wei", "r": "車", "sc": 10, "l": 5, "f": 8007},
  {"h": "稂", "p": "láng", "sp": "lang", "r": "禾", "sc": 12, "l": 5, "f": 8008},
  {"h": "蒞", "p": "lì", "sp": "li", "r": "艸", "sc": 13, "l": 5, "f": 8009},
  {"h": "蠆", "p": "chài", "sp": "chai", "r": "虫", "sc": 18, "l": 5, "f": 8010},
  {"h": "褲", "p": "kù", "sp": "ku", "r": "衣", "sc": 15, "l": 5, "f": 8011},
  {"h": "佧", "p": "kǎ", "sp": "ka", "r": "人", "sc": 7, "l": 5, "f": 8012},
  {"h": "紵", "p": "zhù", "sp": "zhu", "r": "糸", "sc": 11, "l": 5, "f": 8013},
  {"h": "緌", "p": "ruí", "sp": "rui", "r": "糸", "sc": 14, "l": 5, "f": 8014},
  {"h": "蠾", "p": "zhú", "sp": "zhu", "r": "虫", "sc": 27, "l": 5, "f": 8015},
  {
    "h": "裲",
    "p": "liǎng",
    "sp": "liang",
    "r": "衣",
    "sc": 13,
    "l": 5,
    "f": 8016
  },
  {"h": "郈", "p": "hòu", "sp": "hou", "r": "邑", "sc": 8, "l": 5, "f": 8017},
  {"h": "鋐", "p": "hóng", "sp": "hong", "r": "金", "sc": 15, "l": 5, "f": 8018},
  {"h": "餱", "p": "hóu", "sp": "hou", "r": "食", "sc": 17, "l": 5, "f": 8019},
  {"h": "瓞", "p": "dié", "sp": "die", "r": "瓜", "sc": 10, "l": 5, "f": 8020},
  {"h": "驙", "p": "zhān", "sp": "zhan", "r": "馬", "sc": 23, "l": 5, "f": 8021},
  {"h": "鯙", "p": "chún", "sp": "chun", "r": "魚", "sc": 19, "l": 5, "f": 8022},
  {"h": "眆", "p": "fǎng", "sp": "fang", "r": "目", "sc": 9, "l": 5, "f": 8023},
  {"h": "睠", "p": "juàn", "sp": "juan", "r": "目", "sc": 13, "l": 5, "f": 8024},
  {"h": "瞫", "p": "shěn", "sp": "shen", "r": "目", "sc": 17, "l": 5, "f": 8025},
  {"h": "矅", "p": "yào", "sp": "yao", "r": "目", "sc": 19, "l": 5, "f": 8026},
  {"h": "碃", "p": "qìng", "sp": "qing", "r": "石", "sc": 13, "l": 5, "f": 8027},
  {"h": "禃", "p": "zhí", "sp": "zhi", "r": "示", "sc": 12, "l": 5, "f": 8028},
  {"h": "秔", "p": "jīng", "sp": "jing", "r": "禾", "sc": 9, "l": 5, "f": 8029},
  {"h": "穻", "p": "yū", "sp": "yu", "r": "穴", "sc": 8, "l": 5, "f": 8030},
  {"h": "箇", "p": "gè", "sp": "ge", "r": "竹", "sc": 14, "l": 5, "f": 8031},
  {"h": "籐", "p": "téng", "sp": "teng", "r": "竹", "sc": 21, "l": 5, "f": 8032},
  {"h": "絏", "p": "xiè", "sp": "xie", "r": "糸", "sc": 12, "l": 5, "f": 8033},
  {"h": "纕", "p": "rǎng", "sp": "rang", "r": "糸", "sc": 23, "l": 5, "f": 8034},
  {"h": "翐", "p": "zhì", "sp": "zhi", "r": "羽", "sc": 11, "l": 5, "f": 8035},
  {"h": "耈", "p": "gǒu", "sp": "gou", "r": "老", "sc": 11, "l": 5, "f": 8036},
  {"h": "耎", "p": "ruǎn", "sp": "ruan", "r": "而", "sc": 9, "l": 5, "f": 8037},
  {"h": "脷", "p": "lì", "sp": "li", "r": "肉", "sc": 11, "l": 5, "f": 8038},
  {"h": "膕", "p": "guó", "sp": "guo", "r": "肉", "sc": 15, "l": 5, "f": 8039},
  {"h": "艐", "p": "kè", "sp": "ke", "r": "舟", "sc": 15, "l": 5, "f": 8040},
  {"h": "艷", "p": "yàn", "sp": "yan", "r": "色", "sc": 24, "l": 5, "f": 8041},
  {"h": "苙", "p": "lì", "sp": "li", "r": "艸", "sc": 8, "l": 5, "f": 8042},
  {"h": "葹", "p": "shī", "sp": "shi", "r": "艸", "sc": 12, "l": 5, "f": 8043},
  {"h": "蓂", "p": "míng", "sp": "ming", "r": "艸", "sc": 13, "l": 5, "f": 8044},
  {"h": "蔲", "p": "kòu", "sp": "kou", "r": "艸", "sc": 14, "l": 5, "f": 8045},
  {"h": "薋", "p": "cí", "sp": "ci", "r": "艸", "sc": 16, "l": 5, "f": 8046},
  {"h": "蛷", "p": "qiú", "sp": "qiu", "r": "虫", "sc": 13, "l": 5, "f": 8047},
  {"h": "蜑", "p": "dàn", "sp": "dan", "r": "虫", "sc": 12, "l": 5, "f": 8048},
  {"h": "衂", "p": "niù", "sp": "niu", "r": "血", "sc": 9, "l": 5, "f": 8049},
  {"h": "衋", "p": "xì", "sp": "xi", "r": "血", "sc": 24, "l": 5, "f": 8050},
  {"h": "袿", "p": "guī", "sp": "gui", "r": "衣", "sc": 11, "l": 5, "f": 8051},
  {"h": "訕", "p": "shàn", "sp": "shan", "r": "言", "sc": 10, "l": 5, "f": 8052},
  {"h": "訝", "p": "yà", "sp": "ya", "r": "言", "sc": 11, "l": 5, "f": 8053},
  {"h": "詠", "p": "yǒng", "sp": "yong", "r": "言", "sc": 12, "l": 5, "f": 8054},
  {"h": "誳", "p": "qū", "sp": "qu", "r": "言", "sc": 15, "l": 5, "f": 8055},
  {"h": "諆", "p": "qī", "sp": "qi", "r": "言", "sc": 15, "l": 5, "f": 8056},
  {"h": "讱", "p": "rèn", "sp": "ren", "r": "言", "sc": 5, "l": 5, "f": 8057},
  {"h": "軃", "p": "duǒ", "sp": "duo", "r": "身", "sc": 19, "l": 5, "f": 8058},
  {"h": "轝", "p": "yù", "sp": "yu", "r": "車", "sc": 20, "l": 5, "f": 8059},
  {"h": "郀", "p": "kū", "sp": "ku", "r": "邑", "sc": 8, "l": 5, "f": 8060},
  {"h": "郵", "p": "yóu", "sp": "you", "r": "邑", "sc": 10, "l": 5, "f": 8061},
  {"h": "鉐", "p": "shí", "sp": "shi", "r": "金", "sc": 13, "l": 5, "f": 8062},
  {"h": "鍧", "p": "hōng", "sp": "hong", "r": "金", "sc": 17, "l": 5, "f": 8063},
  {"h": "鍼", "p": "zhēn", "sp": "zhen", "r": "金", "sc": 17, "l": 5, "f": 8064},
  {"h": "鏦", "p": "cōng", "sp": "cong", "r": "金", "sc": 19, "l": 5, "f": 8065},
  {"h": "镵", "p": "chán", "sp": "chan", "r": "金", "sc": 22, "l": 5, "f": 8066},
  {"h": "閭", "p": "lǘ", "sp": "lu", "r": "門", "sc": 14, "l": 5, "f": 8067},
  {"h": "闢", "p": "pì", "sp": "pi", "r": "門", "sc": 21, "l": 5, "f": 8068},
  {"h": "陗", "p": "qiào", "sp": "qiao", "r": "阜", "sc": 9, "l": 5, "f": 8069},
  {"h": "陼", "p": "zhǔ", "sp": "zhu", "r": "阜", "sc": 10, "l": 5, "f": 8070},
  {"h": "韈", "p": "wà", "sp": "wa", "r": "革", "sc": 23, "l": 5, "f": 8071},
  {"h": "骫", "p": "wěi", "sp": "wei", "r": "骨", "sc": 12, "l": 5, "f": 8072},
  {"h": "骳", "p": "bèi", "sp": "bei", "r": "骨", "sc": 14, "l": 5, "f": 8073},
  {"h": "骹", "p": "qiāo", "sp": "qiao", "r": "骨", "sc": 15, "l": 5, "f": 8074},
  {"h": "髠", "p": "kūn", "sp": "kun", "r": "髟", "sc": 12, "l": 5, "f": 8075},
  {"h": "髥", "p": "rán", "sp": "ran", "r": "髟", "sc": 14, "l": 5, "f": 8076},
  {"h": "鳒", "p": "jiān", "sp": "jian", "r": "魚", "sc": 18, "l": 5, "f": 8077},
  {"h": "鶡", "p": "hé", "sp": "he", "r": "鳥", "sc": 20, "l": 5, "f": 8078},
  {"h": "龒", "p": "lóng", "sp": "long", "r": "龍", "sc": 21, "l": 5, "f": 8079},
  {"h": "籂", "p": "shi", "sp": "shi", "r": "竹", "sc": 19, "l": 5, "f": 8080},
  {"h": "翯", "p": "hè", "sp": "he", "r": "羽", "sc": 16, "l": 5, "f": 8081},
  {"h": "莚", "p": "yán", "sp": "yan", "r": "艸", "sc": 9, "l": 5, "f": 8082},
  {"h": "萗", "p": "cè", "sp": "ce", "r": "艸", "sc": 11, "l": 5, "f": 8083},
  {
    "h": "転",
    "p": "zhuǎn",
    "sp": "zhuan",
    "r": "車",
    "sc": 11,
    "l": 5,
    "f": 8084
  },
  {"h": "鄷", "p": "fēng", "sp": "feng", "r": "邑", "sc": 15, "l": 5, "f": 8085},
  {"h": "靭", "p": "rèn", "sp": "ren", "r": "革", "sc": 12, "l": 5, "f": 8086},
  {"h": "馉", "p": "gǔ", "sp": "gu", "r": "食", "sc": 12, "l": 5, "f": 8087},
  {"h": "騙", "p": "piàn", "sp": "pian", "r": "馬", "sc": 19, "l": 5, "f": 8088},
  {"h": "驊", "p": "huá", "sp": "hua", "r": "馬", "sc": 20, "l": 5, "f": 8089},
  {"h": "鼰", "p": "niàn", "sp": "nian", "r": "鼠", "sc": 20, "l": 5, "f": 8090},
  {"h": "綑", "p": "kǔn", "sp": "kun", "r": "糸", "sc": 13, "l": 5, "f": 8091},
  {"h": "績", "p": "jī", "sp": "ji", "r": "糸", "sc": 17, "l": 5, "f": 8092},
  {"h": "臘", "p": "là", "sp": "la", "r": "肉", "sc": 19, "l": 5, "f": 8093},
  {"h": "臛", "p": "huò", "sp": "huo", "r": "肉", "sc": 20, "l": 5, "f": 8094},
  {"h": "賜", "p": "cì", "sp": "ci", "r": "貝", "sc": 15, "l": 5, "f": 8095},
  {"h": "贈", "p": "zèng", "sp": "zeng", "r": "貝", "sc": 19, "l": 5, "f": 8096},
  {"h": "蹣", "p": "pán", "sp": "pan", "r": "足", "sc": 18, "l": 5, "f": 8097},
  {"h": "軡", "p": "qián", "sp": "qian", "r": "車", "sc": 11, "l": 5, "f": 8098},
  {"h": "鋸", "p": "jù", "sp": "ju", "r": "金", "sc": 16, "l": 5, "f": 8099},
  {"h": "錐", "p": "zhuī", "sp": "zhui", "r": "金", "sc": 16, "l": 5, "f": 8100},
  {"h": "顎", "p": "è", "sp": "e", "r": "頁", "sc": 18, "l": 5, "f": 8101},
  {"h": "瞐", "p": "mò", "sp": "mo", "r": "目", "sc": 15, "l": 5, "f": 8102},
  {"h": "銶", "p": "qiú", "sp": "qiu", "r": "金", "sc": 15, "l": 5, "f": 8103},
  {"h": "钖", "p": "yáng", "sp": "yang", "r": "金", "sc": 8, "l": 5, "f": 8104},
  {"h": "矊", "p": "mián", "sp": "mian", "r": "目", "sc": 19, "l": 5, "f": 8105},
  {"h": "矘", "p": "tǎng", "sp": "tang", "r": "目", "sc": 25, "l": 5, "f": 8106},
  {"h": "矚", "p": "zhǔ", "sp": "zhu", "r": "目", "sc": 26, "l": 5, "f": 8107},
  {"h": "矡", "p": "jué", "sp": "jue", "r": "矛", "sc": 25, "l": 5, "f": 8108},
  {"h": "硽", "p": "yān", "sp": "yan", "r": "石", "sc": 13, "l": 5, "f": 8109},
  {"h": "磿", "p": "lì", "sp": "li", "r": "石", "sc": 17, "l": 5, "f": 8110},
  {"h": "礈", "p": "zhuì", "sp": "zhui", "r": "石", "sc": 17, "l": 5, "f": 8111},
  {"h": "礜", "p": "yù", "sp": "yu", "r": "石", "sc": 18, "l": 5, "f": 8112},
  {"h": "秲", "p": "zhì", "sp": "zhi", "r": "禾", "sc": 11, "l": 5, "f": 8113},
  {
    "h": "窮",
    "p": "qióng",
    "sp": "qiong",
    "r": "穴",
    "sc": 15,
    "l": 5,
    "f": 8114
  },
  {"h": "竄", "p": "cuàn", "sp": "cuan", "r": "穴", "sc": 18, "l": 5, "f": 8115},
  {"h": "笗", "p": "dōng", "sp": "dong", "r": "竹", "sc": 11, "l": 5, "f": 8116},
  {"h": "箘", "p": "jùn", "sp": "jun", "r": "竹", "sc": 14, "l": 5, "f": 8117},
  {"h": "箽", "p": "dǒng", "sp": "dong", "r": "竹", "sc": 15, "l": 5, "f": 8118},
  {"h": "簍", "p": "lǒu", "sp": "lou", "r": "竹", "sc": 17, "l": 5, "f": 8119},
  {"h": "簜", "p": "dàng", "sp": "dang", "r": "竹", "sc": 18, "l": 5, "f": 8120},
  {"h": "粷", "p": "jú", "sp": "ju", "r": "米", "sc": 14, "l": 5, "f": 8121},
  {"h": "絨", "p": "róng", "sp": "rong", "r": "糸", "sc": 12, "l": 5, "f": 8122},
  {"h": "絹", "p": "juàn", "sp": "juan", "r": "糸", "sc": 13, "l": 5, "f": 8123},
  {"h": "緷", "p": "yùn", "sp": "yun", "r": "糸", "sc": 15, "l": 5, "f": 8124},
  {"h": "緻", "p": "zhì", "sp": "zhi", "r": "糸", "sc": 16, "l": 5, "f": 8125},
  {"h": "繉", "p": "hún", "sp": "hun", "r": "糸", "sc": 17, "l": 5, "f": 8126},
  {"h": "繙", "p": "fān", "sp": "fan", "r": "糸", "sc": 18, "l": 5, "f": 8127},
  {"h": "罁", "p": "gāng", "sp": "gang", "r": "缶", "sc": 14, "l": 5, "f": 8128},
  {"h": "羅", "p": "luó", "sp": "luo", "r": "网", "sc": 19, "l": 5, "f": 8129},
  {"h": "羇", "p": "jī", "sp": "ji", "r": "网", "sc": 22, "l": 5, "f": 8130},
  {"h": "聣", "p": "ní", "sp": "ni", "r": "耳", "sc": 14, "l": 5, "f": 8131},
  {"h": "聰", "p": "cōng", "sp": "cong", "r": "耳", "sc": 17, "l": 5, "f": 8132},
  {"h": "脙", "p": "xiū", "sp": "xiu", "r": "肉", "sc": 11, "l": 5, "f": 8133},
  {"h": "脛", "p": "jìng", "sp": "jing", "r": "肉", "sc": 11, "l": 5, "f": 8134},
  {"h": "脻", "p": "jiē", "sp": "jie", "r": "肉", "sc": 12, "l": 5, "f": 8135},
  {
    "h": "脼",
    "p": "liǎng",
    "sp": "liang",
    "r": "肉",
    "sc": 12,
    "l": 5,
    "f": 8136
  },
  {"h": "腗", "p": "pí", "sp": "pi", "r": "肉", "sc": 12, "l": 5, "f": 8137},
  {"h": "艜", "p": "dài", "sp": "dai", "r": "舟", "sc": 17, "l": 5, "f": 8138},
  {"h": "萊", "p": "lái", "sp": "lai", "r": "艸", "sc": 11, "l": 5, "f": 8139},
  {"h": "葘", "p": "zī", "sp": "zi", "r": "艸", "sc": 12, "l": 5, "f": 8140},
  {"h": "葢", "p": "gài", "sp": "gai", "r": "艸", "sc": 12, "l": 5, "f": 8141},
  {"h": "蔥", "p": "cōng", "sp": "cong", "r": "艸", "sc": 14, "l": 5, "f": 8142},
  {"h": "藍", "p": "lán", "sp": "lan", "r": "艸", "sc": 17, "l": 5, "f": 8143},
  {"h": "藘", "p": "lǘ", "sp": "lu", "r": "艸", "sc": 18, "l": 5, "f": 8144},
  {"h": "藝", "p": "yì", "sp": "yi", "r": "艸", "sc": 18, "l": 5, "f": 8145},
  {"h": "蘿", "p": "luó", "sp": "luo", "r": "艸", "sc": 22, "l": 5, "f": 8146},
  {
    "h": "蛬",
    "p": "qióng",
    "sp": "qiong",
    "r": "虫",
    "sc": 12,
    "l": 5,
    "f": 8147
  },
  {"h": "螁", "p": "ban", "sp": "ban", "r": "虫", "sc": 15, "l": 5, "f": 8148},
  {"h": "螚", "p": "nài", "sp": "nai", "r": "虫", "sc": 16, "l": 5, "f": 8149},
  {
    "h": "螤",
    "p": "zhōng",
    "sp": "zhong",
    "r": "虫",
    "sc": 16,
    "l": 5,
    "f": 8150
  },
  {"h": "袇", "p": "rán", "sp": "ran", "r": "衣", "sc": 9, "l": 5, "f": 8151},
  {"h": "袌", "p": "bào", "sp": "bao", "r": "衣", "sc": 11, "l": 5, "f": 8152},
  {"h": "袔", "p": "hè", "sp": "he", "r": "衣", "sc": 10, "l": 5, "f": 8153},
  {"h": "袦", "p": "nà", "sp": "na", "r": "衣", "sc": 10, "l": 5, "f": 8154},
  {"h": "袨", "p": "xuàn", "sp": "xuan", "r": "衣", "sc": 10, "l": 5, "f": 8155},
  {"h": "覂", "p": "fěng", "sp": "feng", "r": "襾", "sc": 10, "l": 5, "f": 8156},
  {"h": "訃", "p": "fù", "sp": "fu", "r": "言", "sc": 9, "l": 5, "f": 8157},
  {"h": "訋", "p": "diào", "sp": "diao", "r": "言", "sc": 10, "l": 5, "f": 8158},
  {"h": "詫", "p": "chà", "sp": "cha", "r": "言", "sc": 13, "l": 5, "f": 8159},
  {"h": "詬", "p": "gòu", "sp": "gou", "r": "言", "sc": 13, "l": 5, "f": 8160},
  {"h": "詮", "p": "quán", "sp": "quan", "r": "言", "sc": 13, "l": 5, "f": 8161},
  {"h": "詻", "p": "è", "sp": "e", "r": "言", "sc": 13, "l": 5, "f": 8162},
  {
    "h": "誠",
    "p": "chéng",
    "sp": "cheng",
    "r": "言",
    "sc": 13,
    "l": 5,
    "f": 8163
  },
  {"h": "謬", "p": "miù", "sp": "miu", "r": "言", "sc": 18, "l": 5, "f": 8164},
  {"h": "譃", "p": "xū", "sp": "xu", "r": "言", "sc": 18, "l": 5, "f": 8165},
  {"h": "譄", "p": "zēng", "sp": "zeng", "r": "言", "sc": 19, "l": 5, "f": 8166},
  {"h": "譯", "p": "yì", "sp": "yi", "r": "言", "sc": 20, "l": 5, "f": 8167},
  {"h": "譽", "p": "yù", "sp": "yu", "r": "言", "sc": 20, "l": 5, "f": 8168},
  {"h": "賰", "p": "chǔn", "sp": "chun", "r": "貝", "sc": 16, "l": 5, "f": 8169},
  {"h": "趮", "p": "zào", "sp": "zao", "r": "走", "sc": 20, "l": 5, "f": 8170},
  {
    "h": "蹱",
    "p": "zhōng",
    "sp": "zhong",
    "r": "足",
    "sc": 19,
    "l": 5,
    "f": 8171
  },
  {"h": "軙", "p": "chén", "sp": "chen", "r": "車", "sc": 11, "l": 5, "f": 8172},
  {"h": "軠", "p": "rèn", "sp": "ren", "r": "車", "sc": 11, "l": 5, "f": 8173},
  {"h": "輩", "p": "bèi", "sp": "bei", "r": "車", "sc": 15, "l": 5, "f": 8174},
  {"h": "辭", "p": "cí", "sp": "ci", "r": "辛", "sc": 19, "l": 5, "f": 8175},
  {"h": "遜", "p": "xùn", "sp": "xun", "r": "辵", "sc": 13, "l": 5, "f": 8176},
  {"h": "鄧", "p": "dèng", "sp": "deng", "r": "邑", "sc": 14, "l": 5, "f": 8177},
  {"h": "鄫", "p": "zēng", "sp": "zeng", "r": "邑", "sc": 14, "l": 5, "f": 8178},
  {"h": "鄿", "p": "jī", "sp": "ji", "r": "邑", "sc": 17, "l": 5, "f": 8179},
  {"h": "酀", "p": "yàn", "sp": "yan", "r": "邑", "sc": 18, "l": 5, "f": 8180},
  {"h": "釡", "p": "fǔ", "sp": "fu", "r": "金", "sc": 10, "l": 5, "f": 8181},
  {"h": "釢", "p": "nǎi", "sp": "nai", "r": "金", "sc": 10, "l": 5, "f": 8182},
  {"h": "鉃", "p": "shì", "sp": "shi", "r": "金", "sc": 13, "l": 5, "f": 8183},
  {"h": "鍍", "p": "dù", "sp": "du", "r": "金", "sc": 17, "l": 5, "f": 8184},
  {"h": "鍎", "p": "tú", "sp": "tu", "r": "金", "sc": 17, "l": 5, "f": 8185},
  {"h": "鎖", "p": "suǒ", "sp": "suo", "r": "金", "sc": 18, "l": 5, "f": 8186},
  {"h": "闃", "p": "qù", "sp": "qu", "r": "門", "sc": 17, "l": 5, "f": 8187},
  {"h": "鞗", "p": "tiáo", "sp": "tiao", "r": "革", "sc": 15, "l": 5, "f": 8188},
  {"h": "韱", "p": "xiān", "sp": "xian", "r": "韭", "sc": 17, "l": 5, "f": 8189},
  {"h": "頹", "p": "tuí", "sp": "tui", "r": "頁", "sc": 16, "l": 5, "f": 8190},
  {"h": "願", "p": "yuàn", "sp": "yuan", "r": "頁", "sc": 19, "l": 5, "f": 8191},
  {"h": "顳", "p": "niè", "sp": "nie", "r": "頁", "sc": 27, "l": 5, "f": 8192},
  {"h": "餁", "p": "rèn", "sp": "ren", "r": "食", "sc": 14, "l": 5, "f": 8193},
  {"h": "餃", "p": "jiǎo", "sp": "jiao", "r": "食", "sc": 14, "l": 5, "f": 8194},
  {"h": "餵", "p": "wèi", "sp": "wei", "r": "食", "sc": 17, "l": 5, "f": 8195},
  {"h": "馮", "p": "féng", "sp": "feng", "r": "馬", "sc": 12, "l": 5, "f": 8196},
  {"h": "馳", "p": "chí", "sp": "chi", "r": "馬", "sc": 13, "l": 5, "f": 8197},
  {"h": "駛", "p": "shǐ", "sp": "shi", "r": "馬", "sc": 15, "l": 5, "f": 8198},
  {"h": "騎", "p": "qí", "sp": "qi", "r": "馬", "sc": 18, "l": 5, "f": 8199},
  {"h": "騺", "p": "zhì", "sp": "zhi", "r": "馬", "sc": 21, "l": 5, "f": 8200},
  {"h": "鯚", "p": "jì", "sp": "ji", "r": "魚", "sc": 19, "l": 5, "f": 8201},
  {"h": "鰟", "p": "fáng", "sp": "fang", "r": "魚", "sc": 21, "l": 5, "f": 8202},
  {"h": "鱉", "p": "biē", "sp": "bie", "r": "魚", "sc": 22, "l": 5, "f": 8203},
  {"h": "鳛", "p": "xí", "sp": "xi", "r": "魚", "sc": 19, "l": 5, "f": 8204},
  {"h": "鴟", "p": "chī", "sp": "chi", "r": "鳥", "sc": 16, "l": 5, "f": 8205},
  {"h": "鴮", "p": "wū", "sp": "wu", "r": "鳥", "sc": 17, "l": 5, "f": 8206},
  {"h": "鴲", "p": "zhī", "sp": "zhi", "r": "鳥", "sc": 17, "l": 5, "f": 8207},
  {"h": "鴷", "p": "liè", "sp": "lie", "r": "鳥", "sc": 17, "l": 5, "f": 8208},
  {"h": "鷷", "p": "zūn", "sp": "zun", "r": "鳥", "sc": 23, "l": 5, "f": 8209},
  {"h": "鸀", "p": "chǔ", "sp": "chu", "r": "鳥", "sc": 24, "l": 5, "f": 8210},
  {"h": "麎", "p": "chén", "sp": "chen", "r": "鹿", "sc": 18, "l": 5, "f": 8211},
  {"h": "麡", "p": "qí", "sp": "qi", "r": "鹿", "sc": 25, "l": 5, "f": 8212},
  {"h": "麧", "p": "hé", "sp": "he", "r": "麥", "sc": 14, "l": 5, "f": 8213},
  {"h": "鼇", "p": "áo", "sp": "ao", "r": "黽", "sc": 23, "l": 5, "f": 8214},
  {"h": "鼞", "p": "tāng", "sp": "tang", "r": "鼓", "sc": 24, "l": 5, "f": 8215},
  {"h": "鼤", "p": "wén", "sp": "wen", "r": "鼠", "sc": 17, "l": 5, "f": 8216},
  {"h": "齺", "p": "zōu", "sp": "zou", "r": "齒", "sc": 25, "l": 5, "f": 8217},
  {"h": "齽", "p": "jìn", "sp": "jin", "r": "齒", "sc": 28, "l": 5, "f": 8218},
  {"h": "龑", "p": "yǎn", "sp": "yan", "r": "龍", "sc": 20, "l": 5, "f": 8219},
  {"h": "裏", "p": "", "sp": "", "r": "衣", "sc": 13, "l": 5, "f": 8220},
  {"h": "䦶", "p": "zhèng", "sp": "zheng", "r": "", "sc": 14, "l": 5, "f": 8221},
  {"h": "䲡", "p": "qiū", "sp": "qiu", "r": "", "sc": 20, "l": 5, "f": 8222},
  {"h": "砮", "p": "nǔ", "sp": "nu", "r": "石", "sc": 10, "l": 5, "f": 8223},
  {"h": "綟", "p": "lì", "sp": "li", "r": "糸", "sc": 14, "l": 5, "f": 8224},
  {"h": "緜", "p": "mián", "sp": "mian", "r": "糸", "sc": 15, "l": 5, "f": 8225},
  {"h": "繑", "p": "qiāo", "sp": "qiao", "r": "糸", "sc": 18, "l": 5, "f": 8226},
  {"h": "纑", "p": "lú", "sp": "lu", "r": "糸", "sc": 22, "l": 5, "f": 8227},
  {"h": "罈", "p": "tán", "sp": "tan", "r": "缶", "sc": 18, "l": 5, "f": 8228},
  {"h": "蠙", "p": "pín", "sp": "pin", "r": "虫", "sc": 20, "l": 5, "f": 8229},
  {"h": "輴", "p": "chūn", "sp": "chun", "r": "車", "sc": 16, "l": 5, "f": 8230},
  {"h": "酨", "p": "zài", "sp": "zai", "r": "酉", "sc": 13, "l": 5, "f": 8231},
  {"h": "鏂", "p": "ōu", "sp": "ou", "r": "金", "sc": 19, "l": 5, "f": 8232},
  {"h": "駹", "p": "máng", "sp": "mang", "r": "馬", "sc": 17, "l": 5, "f": 8233},
  {"h": "鎪", "p": "sōu", "sp": "sou", "r": "金", "sc": 17, "l": 5, "f": 8234},
  {
    "h": "瞕",
    "p": "zhàng",
    "sp": "zhang",
    "r": "目",
    "sc": 16,
    "l": 5,
    "f": 8235
  },
  {"h": "窪", "p": "wā", "sp": "wa", "r": "穴", "sc": 14, "l": 5, "f": 8236},
  {"h": "肻", "p": "kěn", "sp": "ken", "r": "肉", "sc": 8, "l": 5, "f": 8237},
  {"h": "蠔", "p": "háo", "sp": "hao", "r": "虫", "sc": 20, "l": 5, "f": 8238},
  {"h": "鐰", "p": "qiāo", "sp": "qiao", "r": "金", "sc": 21, "l": 5, "f": 8239},
  {"h": "綵", "p": "cǎi", "sp": "cai", "r": "糸", "sc": 14, "l": 5, "f": 8240},
  {"h": "鮰", "p": "huí", "sp": "hui", "r": "魚", "sc": 17, "l": 5, "f": 8241},
  {"h": "鳾", "p": "shī", "sp": "shi", "r": "鳥", "sc": 15, "l": 5, "f": 8242},
  {"h": "虰", "p": "dīng", "sp": "ding", "r": "虫", "sc": 8, "l": 5, "f": 8243},
  {"h": "驎", "p": "lín", "sp": "lin", "r": "馬", "sc": 22, "l": 5, "f": 8244},
  {"h": "荬", "p": "mǎi", "sp": "mai", "r": "艸", "sc": 9, "l": 5, "f": 8245},
  {
    "h": "鑛",
    "p": "kuàng",
    "sp": "kuang",
    "r": "金",
    "sc": 22,
    "l": 5,
    "f": 8246
  },
  {"h": "犏", "p": "piān", "sp": "pian", "r": "牛", "sc": 13, "l": 5, "f": 8247},
  {"h": "鞧", "p": "qiū", "sp": "qiu", "r": "革", "sc": 18, "l": 5, "f": 8248},
  {"h": "鮋", "p": "yóu", "sp": "you", "r": "魚", "sc": 16, "l": 5, "f": 8249},
  {"h": "鰷", "p": "tiáo", "sp": "tiao", "r": "魚", "sc": 21, "l": 5, "f": 8250},
  {"h": "鶲", "p": "wēng", "sp": "weng", "r": "鳥", "sc": 21, "l": 5, "f": 8251},
  {"h": "盕", "p": "fàn", "sp": "fan", "r": "皿", "sc": 11, "l": 5, "f": 8252},
  {"h": "盩", "p": "zhōu", "sp": "zhou", "r": "皿", "sc": 17, "l": 5, "f": 8253},
  {"h": "盻", "p": "xì", "sp": "xi", "r": "目", "sc": 9, "l": 5, "f": 8254},
  {
    "h": "眖",
    "p": "kuàng",
    "sp": "kuang",
    "r": "目",
    "sc": 10,
    "l": 5,
    "f": 8255
  },
  {"h": "砛", "p": "jin", "sp": "jin", "r": "石", "sc": 9, "l": 5, "f": 8256},
  {"h": "硈", "p": "qià", "sp": "qia", "r": "石", "sc": 11, "l": 5, "f": 8257},
  {"h": "禢", "p": "tà", "sp": "ta", "r": "示", "sc": 14, "l": 5, "f": 8258},
  {"h": "禥", "p": "qí", "sp": "qi", "r": "示", "sc": 15, "l": 5, "f": 8259},
  {"h": "窼", "p": "zhāo", "sp": "zhao", "r": "穴", "sc": 16, "l": 5, "f": 8260},
  {"h": "竏", "p": "qian", "sp": "qian", "r": "立", "sc": 8, "l": 5, "f": 8261},
  {"h": "筈", "p": "kuò", "sp": "kuo", "r": "竹", "sc": 12, "l": 5, "f": 8262},
  {"h": "筽", "p": "o", "sp": "o", "r": "竹", "sc": 13, "l": 5, "f": 8263},
  {"h": "篴", "p": "dí", "sp": "di", "r": "竹", "sc": 16, "l": 5, "f": 8264},
  {"h": "籆", "p": "yuè", "sp": "yue", "r": "竹", "sc": 19, "l": 5, "f": 8265},
  {"h": "籇", "p": "háo", "sp": "hao", "r": "竹", "sc": 20, "l": 5, "f": 8266},
  {"h": "紏", "p": "tǒu", "sp": "tou", "r": "糸", "sc": 10, "l": 5, "f": 8267},
  {"h": "肣", "p": "hán", "sp": "han", "r": "肉", "sc": 8, "l": 5, "f": 8268},
  {"h": "胑", "p": "zhī", "sp": "zhi", "r": "肉", "sc": 9, "l": 5, "f": 8269},
  {"h": "臮", "p": "jì", "sp": "ji", "r": "自", "sc": 12, "l": 5, "f": 8270},
  {"h": "苲", "p": "zhǎ", "sp": "zha", "r": "艸", "sc": 8, "l": 5, "f": 8271},
  {
    "h": "蒃",
    "p": "zhuàn",
    "sp": "zhuan",
    "r": "艸",
    "sc": 12,
    "l": 5,
    "f": 8272
  },
  {"h": "虗", "p": "xū", "sp": "xu", "r": "虍", "sc": 11, "l": 5, "f": 8273},
  {"h": "蜌", "p": "bì", "sp": "bi", "r": "虫", "sc": 13, "l": 5, "f": 8274},
  {"h": "跘", "p": "pán", "sp": "pan", "r": "足", "sc": 12, "l": 5, "f": 8275},
  {"h": "鞦", "p": "qiū", "sp": "qiu", "r": "革", "sc": 18, "l": 5, "f": 8276},
  {"h": "驄", "p": "cōng", "sp": "cong", "r": "馬", "sc": 21, "l": 5, "f": 8277},
  {"h": "鷭", "p": "fán", "sp": "fan", "r": "鳥", "sc": 23, "l": 5, "f": 8278},
  {"h": "鸼", "p": "zhōu", "sp": "zhou", "r": "鳥", "sc": 11, "l": 5, "f": 8279},
  {"h": "鹐", "p": "qiān", "sp": "qian", "r": "", "sc": 13, "l": 5, "f": 8280},
  {"h": "矄", "p": "xūn", "sp": "xun", "r": "目", "sc": 19, "l": 5, "f": 8281},
  {"h": "魖", "p": "xū", "sp": "xu", "r": "鬼", "sc": 20, "l": 5, "f": 8282},
  {"h": "鴐", "p": "gē", "sp": "ge", "r": "鳥", "sc": 16, "l": 5, "f": 8283},
  {"h": "薀", "p": "yùn", "sp": "yun", "r": "艸", "sc": 16, "l": 5, "f": 8284},
  {"h": "謏", "p": "xiǎo", "sp": "xiao", "r": "言", "sc": 16, "l": 5, "f": 8285},
  {"h": "脎", "p": "sà", "sp": "sa", "r": "肉", "sc": 10, "l": 5, "f": 8286},
  {"h": "締", "p": "dì", "sp": "di", "r": "糸", "sc": 15, "l": 5, "f": 8287},
  {"h": "聧", "p": "kuī", "sp": "kui", "r": "耳", "sc": 15, "l": 5, "f": 8288},
  {"h": "脜", "p": "yǒu", "sp": "you", "r": "肉", "sc": 11, "l": 5, "f": 8289},
  {"h": "艤", "p": "yǐ", "sp": "yi", "r": "舟", "sc": 19, "l": 5, "f": 8290},
  {"h": "荖", "p": "lǎo", "sp": "lao", "r": "艸", "sc": 9, "l": 5, "f": 8291},
  {"h": "蕌", "p": "lěi", "sp": "lei", "r": "艸", "sc": 15, "l": 5, "f": 8292},
  {"h": "蠷", "p": "qú", "sp": "qu", "r": "虫", "sc": 24, "l": 5, "f": 8293},
  {"h": "褏", "p": "xiù", "sp": "xiu", "r": "衣", "sc": 14, "l": 5, "f": 8294},
  {"h": "觽", "p": "xī", "sp": "xi", "r": "角", "sc": 22, "l": 5, "f": 8295},
  {"h": "踫", "p": "pèng", "sp": "peng", "r": "足", "sc": 15, "l": 5, "f": 8296},
  {"h": "蹽", "p": "liāo", "sp": "liao", "r": "足", "sc": 19, "l": 5, "f": 8297},
  {"h": "軉", "p": "yù", "sp": "yu", "r": "身", "sc": 27, "l": 5, "f": 8298},
  {"h": "軛", "p": "è", "sp": "e", "r": "車", "sc": 11, "l": 5, "f": 8299},
  {
    "h": "墚",
    "p": "liáng",
    "sp": "liang",
    "r": "土",
    "sc": 14,
    "l": 5,
    "f": 8300
  },
  {"h": "搌", "p": "zhǎn", "sp": "zhan", "r": "手", "sc": 13, "l": 5, "f": 8301},
  {"h": "擗", "p": "pǐ", "sp": "pi", "r": "手", "sc": 16, "l": 5, "f": 8302},
  {"h": "哳", "p": "zhā", "sp": "zha", "r": "口", "sc": 10, "l": 5, "f": 8303},
  {"h": "鋻", "p": "jiàn", "sp": "jian", "r": "金", "sc": 16, "l": 5, "f": 8304},
  {"h": "楗", "p": "jiàn", "sp": "jian", "r": "木", "sc": 12, "l": 5, "f": 8305},
  {"h": "韂", "p": "chàn", "sp": "chan", "r": "革", "sc": 22, "l": 5, "f": 8306},
  {"h": "飆", "p": "biāo", "sp": "biao", "r": "風", "sc": 21, "l": 5, "f": 8307},
  {"h": "饏", "p": "dàn", "sp": "dan", "r": "食", "sc": 20, "l": 5, "f": 8308},
  {"h": "耥", "p": "tāng", "sp": "tang", "r": "耒", "sc": 14, "l": 5, "f": 8309},
  {"h": "髼", "p": "péng", "sp": "peng", "r": "髟", "sc": 17, "l": 5, "f": 8310},
  {"h": "篼", "p": "dōu", "sp": "dou", "r": "竹", "sc": 17, "l": 5, "f": 8311},
  {"h": "鬑", "p": "lián", "sp": "lian", "r": "髟", "sc": 20, "l": 5, "f": 8312},
  {"h": "醑", "p": "xǔ", "sp": "xu", "r": "酉", "sc": 16, "l": 5, "f": 8313},
  {"h": "鴧", "p": "ròng", "sp": "rong", "r": "鳥", "sc": 16, "l": 5, "f": 8314},
  {"h": "眜", "p": "mò", "sp": "mo", "r": "目", "sc": 10, "l": 5, "f": 8315},
  {"h": "瞷", "p": "jiàn", "sp": "jian", "r": "目", "sc": 17, "l": 5, "f": 8316},
  {"h": "磦", "p": "biāo", "sp": "biao", "r": "石", "sc": 16, "l": 5, "f": 8317},
  {"h": "礨", "p": "lěi", "sp": "lei", "r": "石", "sc": 20, "l": 5, "f": 8318},
  {"h": "祃", "p": "mà", "sp": "ma", "r": "示", "sc": 7, "l": 5, "f": 8319},
  {"h": "祲", "p": "jìn", "sp": "jin", "r": "示", "sc": 11, "l": 5, "f": 8320},
  {"h": "祻", "p": "gù", "sp": "gu", "r": "示", "sc": 12, "l": 5, "f": 8321},
  {"h": "窊", "p": "wā", "sp": "wa", "r": "穴", "sc": 10, "l": 5, "f": 8322},
  {"h": "簥", "p": "jiāo", "sp": "jiao", "r": "竹", "sc": 18, "l": 5, "f": 8323},
  {"h": "籲", "p": "yù", "sp": "yu", "r": "竹", "sc": 32, "l": 5, "f": 8324},
  {"h": "縢", "p": "téng", "sp": "teng", "r": "糸", "sc": 16, "l": 5, "f": 8325},
  {"h": "繌", "p": "sha", "sp": "sha", "r": "糸", "sc": 17, "l": 5, "f": 8326},
  {"h": "繒", "p": "zēng", "sp": "zeng", "r": "糸", "sc": 18, "l": 5, "f": 8327},
  {"h": "聜", "p": "dǐ", "sp": "di", "r": "耳", "sc": 14, "l": 5, "f": 8328},
  {"h": "胔", "p": "zì", "sp": "zi", "r": "肉", "sc": 12, "l": 5, "f": 8329},
  {"h": "胾", "p": "zì", "sp": "zi", "r": "肉", "sc": 12, "l": 5, "f": 8330},
  {"h": "脨", "p": "cù", "sp": "cu", "r": "肉", "sc": 11, "l": 5, "f": 8331},
  {"h": "膼", "p": "zhuā", "sp": "zhua", "r": "肉", "sc": 15, "l": 5, "f": 8332},
  {"h": "艕", "p": "bàng", "sp": "bang", "r": "舟", "sc": 16, "l": 5, "f": 8333},
  {"h": "苼", "p": "shēng", "sp": "sheng", "r": "艸", "sc": 8, "l": 5, "f": 8334},
  {"h": "莻", "p": "gòng", "sp": "gong", "r": "匕", "sc": 13, "l": 5, "f": 8335},
  {"h": "蓨", "p": "tiáo", "sp": "tiao", "r": "艸", "sc": 13, "l": 5, "f": 8336},
  {"h": "蓵", "p": "jié", "sp": "jie", "r": "艸", "sc": 14, "l": 5, "f": 8337},
  {"h": "蓾", "p": "lǔ", "sp": "lu", "r": "艸", "sc": 14, "l": 5, "f": 8338},
  {"h": "藦", "p": "mò", "sp": "mo", "r": "艸", "sc": 18, "l": 5, "f": 8339},
  {"h": "虀", "p": "jī", "sp": "ji", "r": "艸", "sc": 22, "l": 5, "f": 8340},
  {"h": "蚈", "p": "qiān", "sp": "qian", "r": "虫", "sc": 10, "l": 5, "f": 8341},
  {"h": "蛡", "p": "yì", "sp": "yi", "r": "虫", "sc": 12, "l": 5, "f": 8342},
  {"h": "蛻", "p": "tuì", "sp": "tui", "r": "虫", "sc": 13, "l": 5, "f": 8343},
  {"h": "蝝", "p": "yuán", "sp": "yuan", "r": "虫", "sc": 15, "l": 5, "f": 8344},
  {"h": "蝹", "p": "yūn", "sp": "yun", "r": "虫", "sc": 15, "l": 5, "f": 8345},
  {"h": "蟿", "p": "qì", "sp": "qi", "r": "虫", "sc": 19, "l": 5, "f": 8346},
  {"h": "衼", "p": "zhī", "sp": "zhi", "r": "衣", "sc": 9, "l": 5, "f": 8347},
  {"h": "褦", "p": "nài", "sp": "nai", "r": "衣", "sc": 15, "l": 5, "f": 8348},
  {"h": "襶", "p": "dài", "sp": "dai", "r": "衣", "sc": 22, "l": 5, "f": 8349},
  {"h": "覐", "p": "jué", "sp": "jue", "r": "見", "sc": 11, "l": 5, "f": 8350},
  {"h": "詯", "p": "huì", "sp": "hui", "r": "言", "sc": 13, "l": 5, "f": 8351},
  {"h": "誴", "p": "cóng", "sp": "cong", "r": "言", "sc": 15, "l": 5, "f": 8352},
  {"h": "諓", "p": "jiàn", "sp": "jian", "r": "言", "sc": 15, "l": 5, "f": 8353},
  {"h": "謧", "p": "lí", "sp": "li", "r": "言", "sc": 17, "l": 5, "f": 8354},
  {
    "h": "謪",
    "p": "shāng",
    "sp": "shang",
    "r": "言",
    "sc": 18,
    "l": 5,
    "f": 8355
  },
  {"h": "譁", "p": "huá", "sp": "hua", "r": "言", "sc": 17, "l": 5, "f": 8356},
  {"h": "貤", "p": "yí", "sp": "yi", "r": "貝", "sc": 10, "l": 5, "f": 8357},
  {
    "h": "赪",
    "p": "chēng",
    "sp": "cheng",
    "r": "赤",
    "sc": 13,
    "l": 5,
    "f": 8358
  },
  {"h": "踖", "p": "jí", "sp": "ji", "r": "足", "sc": 15, "l": 5, "f": 8359},
  {"h": "軆", "p": "tǐ", "sp": "ti", "r": "身", "sc": 20, "l": 5, "f": 8360},
  {"h": "輈", "p": "zhōu", "sp": "zhou", "r": "車", "sc": 13, "l": 5, "f": 8361},
  {"h": "鄛", "p": "cháo", "sp": "chao", "r": "邑", "sc": 13, "l": 5, "f": 8362},
  {"h": "鄡", "p": "qiāo", "sp": "qiao", "r": "邑", "sc": 13, "l": 5, "f": 8363},
  {"h": "醖", "p": "yùn", "sp": "yun", "r": "酉", "sc": 16, "l": 5, "f": 8364},
  {"h": "屺", "p": "qǐ", "sp": "qi", "r": "山", "sc": 6, "l": 5, "f": 8365},
  {"h": "鍑", "p": "fù", "sp": "fu", "r": "金", "sc": 17, "l": 5, "f": 8366},
  {"h": "鏉", "p": "shòu", "sp": "shou", "r": "金", "sc": 19, "l": 5, "f": 8367},
  {"h": "雊", "p": "gòu", "sp": "gou", "r": "隹", "sc": 13, "l": 5, "f": 8368},
  {
    "h": "顓",
    "p": "zhuān",
    "sp": "zhuan",
    "r": "頁",
    "sc": 18,
    "l": 5,
    "f": 8369
  },
  {"h": "飐", "p": "zhǎn", "sp": "zhan", "r": "風", "sc": 9, "l": 5, "f": 8370},
  {"h": "稆", "p": "lǚ", "sp": "lu", "r": "禾", "sc": 11, "l": 5, "f": 8371},
  {"h": "駳", "p": "dàn", "sp": "dan", "r": "馬", "sc": 16, "l": 5, "f": 8372},
  {"h": "騽", "p": "xí", "sp": "xi", "r": "馬", "sc": 21, "l": 5, "f": 8373},
  {"h": "骍", "p": "xīng", "sp": "xing", "r": "馬", "sc": 10, "l": 5, "f": 8374},
  {"h": "筻", "p": "gàng", "sp": "gang", "r": "竹", "sc": 13, "l": 5, "f": 8375},
  {"h": "魱", "p": "hú", "sp": "hu", "r": "魚", "sc": 15, "l": 5, "f": 8376},
  {"h": "舣", "p": "yǐ", "sp": "yi", "r": "舟", "sc": 9, "l": 5, "f": 8377},
  {"h": "鮾", "p": "něi", "sp": "nei", "r": "魚", "sc": 18, "l": 5, "f": 8378},
  {"h": "鯋", "p": "shā", "sp": "sha", "r": "魚", "sc": 18, "l": 5, "f": 8379},
  {"h": "鯏", "p": "li", "sp": "li", "r": "魚", "sc": 18, "l": 5, "f": 8380},
  {"h": "鶃", "p": "yì", "sp": "yi", "r": "鳥", "sc": 19, "l": 5, "f": 8381},
  {"h": "鶫", "p": "dong", "sp": "dong", "r": "鳥", "sc": 20, "l": 5, "f": 8382},
  {"h": "齛", "p": "xiè", "sp": "xie", "r": "齒", "sc": 20, "l": 5, "f": 8383},
  {"h": "矙", "p": "kàn", "sp": "kan", "r": "目", "sc": 24, "l": 5, "f": 8384},
  {
    "h": "絾",
    "p": "chéng",
    "sp": "cheng",
    "r": "糸",
    "sc": 12,
    "l": 5,
    "f": 8385
  },
  {"h": "袗", "p": "zhěn", "sp": "zhen", "r": "衣", "sc": 10, "l": 5, "f": 8386},
  {"h": "襬", "p": "bǎi", "sp": "bai", "r": "衣", "sc": 20, "l": 5, "f": 8387},
  {"h": "漤", "p": "lǎn", "sp": "lan", "r": "水", "sc": 14, "l": 5, "f": 8388},
  {"h": "闕", "p": "què", "sp": "que", "r": "門", "sc": 18, "l": 5, "f": 8389},
  {"h": "鬐", "p": "qí", "sp": "qi", "r": "髟", "sc": 20, "l": 5, "f": 8390},
  {"h": "鲾", "p": "bī", "sp": "bi", "r": "魚", "sc": 17, "l": 5, "f": 8391},
  {"h": "鴾", "p": "móu", "sp": "mou", "r": "鳥", "sc": 17, "l": 5, "f": 8392},
  {"h": "稭", "p": "jiē", "sp": "jie", "r": "禾", "sc": 14, "l": 5, "f": 8393},
  {"h": "籄", "p": "kuì", "sp": "kui", "r": "竹", "sc": 20, "l": 5, "f": 8394},
  {"h": "聺", "p": "qié", "sp": "qie", "r": "耳", "sc": 20, "l": 5, "f": 8395},
  {"h": "褹", "p": "yì", "sp": "yi", "r": "衣", "sc": 16, "l": 5, "f": 8396},
  {"h": "訟", "p": "sòng", "sp": "song", "r": "言", "sc": 11, "l": 5, "f": 8397},
  {"h": "鏕", "p": "lù", "sp": "lu", "r": "金", "sc": 19, "l": 5, "f": 8398},
  {"h": "齆", "p": "wèng", "sp": "weng", "r": "鼻", "sc": 24, "l": 5, "f": 8399},
  {"h": "盫", "p": "ān", "sp": "an", "r": "皿", "sc": 18, "l": 5, "f": 8400},
  {"h": "矎", "p": "xuān", "sp": "xuan", "r": "目", "sc": 19, "l": 5, "f": 8401},
  {"h": "礋", "p": "zé", "sp": "ze", "r": "石", "sc": 18, "l": 5, "f": 8402},
  {"h": "窌", "p": "jiào", "sp": "jiao", "r": "穴", "sc": 10, "l": 5, "f": 8403},
  {"h": "簼", "p": "lóng", "sp": "long", "r": "竹", "sc": 19, "l": 5, "f": 8404},
  {"h": "糴", "p": "dí", "sp": "di", "r": "米", "sc": 22, "l": 5, "f": 8405},
  {"h": "纴", "p": "rèn", "sp": "ren", "r": "糸", "sc": 7, "l": 5, "f": 8406},
  {"h": "罵", "p": "mà", "sp": "ma", "r": "网", "sc": 15, "l": 5, "f": 8407},
  {"h": "翚", "p": "huī", "sp": "hui", "r": "羽", "sc": 12, "l": 5, "f": 8408},
  {"h": "耰", "p": "yōu", "sp": "you", "r": "耒", "sc": 21, "l": 5, "f": 8409},
  {"h": "艧", "p": "huò", "sp": "huo", "r": "舟", "sc": 19, "l": 5, "f": 8410},
  {"h": "藇", "p": "xù", "sp": "xu", "r": "艸", "sc": 16, "l": 5, "f": 8411},
  {"h": "虖", "p": "hū", "sp": "hu", "r": "虍", "sc": 11, "l": 5, "f": 8412},
  {"h": "蚖", "p": "yuán", "sp": "yuan", "r": "虫", "sc": 10, "l": 5, "f": 8413},
  {"h": "蜎", "p": "yuān", "sp": "yuan", "r": "虫", "sc": 13, "l": 5, "f": 8414},
  {"h": "蜨", "p": "dié", "sp": "die", "r": "虫", "sc": 14, "l": 5, "f": 8415},
  {"h": "蠭", "p": "fēng", "sp": "feng", "r": "虫", "sc": 22, "l": 5, "f": 8416},
  {"h": "詟", "p": "zhé", "sp": "zhe", "r": "言", "sc": 12, "l": 5, "f": 8417},
  {"h": "詵", "p": "shēn", "sp": "shen", "r": "言", "sc": 13, "l": 5, "f": 8418},
  {
    "h": "譔",
    "p": "zhuàn",
    "sp": "zhuan",
    "r": "言",
    "sc": 19,
    "l": 5,
    "f": 8419
  },
  {"h": "豭", "p": "jiā", "sp": "jia", "r": "豕", "sc": 16, "l": 5, "f": 8420},
  {"h": "趙", "p": "zhào", "sp": "zhao", "r": "走", "sc": 14, "l": 5, "f": 8421},
  {"h": "趬", "p": "qiāo", "sp": "qiao", "r": "走", "sc": 19, "l": 5, "f": 8422},
  {"h": "阢", "p": "wù", "sp": "wu", "r": "阜", "sc": 5, "l": 5, "f": 8423},
  {"h": "圊", "p": "qīng", "sp": "qing", "r": "囗", "sc": 11, "l": 5, "f": 8424},
  {"h": "醃", "p": "yān", "sp": "yan", "r": "酉", "sc": 15, "l": 5, "f": 8425},
  {"h": "録", "p": "lù", "sp": "lu", "r": "金", "sc": 16, "l": 5, "f": 8426},
  {"h": "鍮", "p": "tōu", "sp": "tou", "r": "金", "sc": 17, "l": 5, "f": 8427},
  {"h": "鍸", "p": "hú", "sp": "hu", "r": "金", "sc": 17, "l": 5, "f": 8428},
  {"h": "闚", "p": "kuī", "sp": "kui", "r": "門", "sc": 19, "l": 5, "f": 8429},
  {"h": "阤", "p": "zhì", "sp": "zhi", "r": "阜", "sc": 5, "l": 5, "f": 8430},
  {"h": "霫", "p": "xí", "sp": "xi", "r": "雨", "sc": 19, "l": 5, "f": 8431},
  {"h": "鞎", "p": "hén", "sp": "hen", "r": "革", "sc": 15, "l": 5, "f": 8432},
  {"h": "韗", "p": "yùn", "sp": "yun", "r": "韋", "sc": 18, "l": 5, "f": 8433},
  {"h": "鬖", "p": "sān", "sp": "san", "r": "髟", "sc": 21, "l": 5, "f": 8434},
  {"h": "鳠", "p": "hù", "sp": "hu", "r": "魚", "sc": 21, "l": 5, "f": 8435},
  {"h": "鴼", "p": "luò", "sp": "luo", "r": "鳥", "sc": 17, "l": 5, "f": 8436},
  {"h": "鷇", "p": "kòu", "sp": "kou", "r": "鳥", "sc": 21, "l": 5, "f": 8437},
  {"h": "黇", "p": "tiān", "sp": "tian", "r": "黃", "sc": 16, "l": 5, "f": 8438},
  {"h": "黡", "p": "yǎn", "sp": "yan", "r": "黑", "sc": 18, "l": 5, "f": 8439},
  {"h": "齣", "p": "chū", "sp": "chu", "r": "齒", "sc": 20, "l": 5, "f": 8440},
  {"h": "盞", "p": "zhǎn", "sp": "zhan", "r": "皿", "sc": 13, "l": 5, "f": 8441},
  {"h": "眅", "p": "pān", "sp": "pan", "r": "目", "sc": 9, "l": 5, "f": 8442},
  {"h": "磈", "p": "wěi", "sp": "wei", "r": "石", "sc": 14, "l": 5, "f": 8443},
  {"h": "磳", "p": "zēng", "sp": "zeng", "r": "石", "sc": 17, "l": 5, "f": 8444},
  {"h": "穫", "p": "huò", "sp": "huo", "r": "禾", "sc": 18, "l": 5, "f": 8445},
  {"h": "窅", "p": "yǎo", "sp": "yao", "r": "穴", "sc": 10, "l": 5, "f": 8446},
  {"h": "窾", "p": "kuǎn", "sp": "kuan", "r": "穴", "sc": 17, "l": 5, "f": 8447},
  {"h": "筭", "p": "suàn", "sp": "suan", "r": "竹", "sc": 13, "l": 5, "f": 8448},
  {"h": "箣", "p": "cè", "sp": "ce", "r": "竹", "sc": 14, "l": 5, "f": 8449},
  {"h": "簃", "p": "yí", "sp": "yi", "r": "竹", "sc": 17, "l": 5, "f": 8450},
  {"h": "簺", "p": "sài", "sp": "sai", "r": "竹", "sc": 19, "l": 5, "f": 8451},
  {"h": "粴", "p": "lǐ", "sp": "li", "r": "米", "sc": 13, "l": 5, "f": 8452},
  {"h": "糦", "p": "xī", "sp": "xi", "r": "米", "sc": 18, "l": 5, "f": 8453},
  {"h": "絟", "p": "quán", "sp": "quan", "r": "糸", "sc": 12, "l": 5, "f": 8454},
  {"h": "緺", "p": "guā", "sp": "gua", "r": "糸", "sc": 14, "l": 5, "f": 8455},
  {"h": "縩", "p": "cài", "sp": "cai", "r": "糸", "sc": 17, "l": 5, "f": 8456},
  {"h": "罍", "p": "léi", "sp": "lei", "r": "缶", "sc": 21, "l": 5, "f": 8457},
  {"h": "翣", "p": "shà", "sp": "sha", "r": "羽", "sc": 14, "l": 5, "f": 8458},
  {"h": "聟", "p": "xù", "sp": "xu", "r": "耳", "sc": 14, "l": 5, "f": 8459},
  {"h": "肊", "p": "yì", "sp": "yi", "r": "肉", "sc": 5, "l": 5, "f": 8460},
  {"h": "肙", "p": "yuàn", "sp": "yuan", "r": "肉", "sc": 7, "l": 5, "f": 8461},
  {"h": "肞", "p": "chā", "sp": "cha", "r": "肉", "sc": 7, "l": 5, "f": 8462},
  {"h": "臞", "p": "qú", "sp": "qu", "r": "肉", "sc": 22, "l": 5, "f": 8463},
  {"h": "苶", "p": "nié", "sp": "nie", "r": "艸", "sc": 8, "l": 5, "f": 8464},
  {"h": "菭", "p": "tái", "sp": "tai", "r": "艸", "sc": 11, "l": 5, "f": 8465},
  {"h": "葖", "p": "tū", "sp": "tu", "r": "艸", "sc": 12, "l": 5, "f": 8466},
  {"h": "蕳", "p": "jiān", "sp": "jian", "r": "艸", "sc": 15, "l": 5, "f": 8467},
  {"h": "蚒", "p": "tóng", "sp": "tong", "r": "虫", "sc": 10, "l": 5, "f": 8468},
  {"h": "蜪", "p": "táo", "sp": "tao", "r": "虫", "sc": 14, "l": 5, "f": 8469},
  {"h": "蝕", "p": "shí", "sp": "shi", "r": "虫", "sc": 14, "l": 5, "f": 8470},
  {"h": "蠜", "p": "fán", "sp": "fan", "r": "虫", "sc": 21, "l": 5, "f": 8471},
  {"h": "袕", "p": "xué", "sp": "xue", "r": "衣", "sc": 10, "l": 5, "f": 8472},
  {"h": "裋", "p": "shù", "sp": "shu", "r": "衣", "sc": 12, "l": 5, "f": 8473},
  {"h": "裩", "p": "kūn", "sp": "kun", "r": "衣", "sc": 13, "l": 5, "f": 8474},
  {"h": "褌", "p": "kūn", "sp": "kun", "r": "衣", "sc": 14, "l": 5, "f": 8475},
  {"h": "褎", "p": "xiù", "sp": "xiu", "r": "衣", "sc": 14, "l": 5, "f": 8476},
  {"h": "襫", "p": "shì", "sp": "shi", "r": "衣", "sc": 20, "l": 5, "f": 8477},
  {"h": "覈", "p": "hé", "sp": "he", "r": "襾", "sc": 19, "l": 5, "f": 8478},
  {"h": "觭", "p": "jī", "sp": "ji", "r": "角", "sc": 15, "l": 5, "f": 8479},
  {"h": "訆", "p": "jiào", "sp": "jiao", "r": "言", "sc": 9, "l": 5, "f": 8480},
  {"h": "訏", "p": "xū", "sp": "xu", "r": "言", "sc": 10, "l": 5, "f": 8481},
  {"h": "訽", "p": "gòu", "sp": "gou", "r": "言", "sc": 12, "l": 5, "f": 8482},
  {"h": "詨", "p": "xiào", "sp": "xiao", "r": "言", "sc": 13, "l": 5, "f": 8483},
  {"h": "譻", "p": "yīng", "sp": "ying", "r": "言", "sc": 21, "l": 5, "f": 8484},
  {"h": "诇", "p": "xiòng", "sp": "xiong", "r": "言", "sc": 7, "l": 5, "f": 8485},
  {"h": "貑", "p": "jiā", "sp": "jia", "r": "豸", "sc": 16, "l": 5, "f": 8486},
  {"h": "跱", "p": "zhì", "sp": "zhi", "r": "足", "sc": 13, "l": 5, "f": 8487},
  {"h": "坶", "p": "mǔ", "sp": "mu", "r": "土", "sc": 8, "l": 5, "f": 8488},
  {"h": "躘", "p": "lóng", "sp": "long", "r": "足", "sc": 23, "l": 5, "f": 8489},
  {"h": "輖", "p": "zhōu", "sp": "zhou", "r": "車", "sc": 15, "l": 5, "f": 8490},
  {"h": "轇", "p": "jiāo", "sp": "jiao", "r": "車", "sc": 18, "l": 5, "f": 8491},
  {"h": "轋", "p": "hún", "sp": "hun", "r": "車", "sc": 17, "l": 5, "f": 8492},
  {"h": "轕", "p": "gé", "sp": "ge", "r": "車", "sc": 19, "l": 5, "f": 8493},
  {"h": "迒", "p": "háng", "sp": "hang", "r": "辵", "sc": 7, "l": 5, "f": 8494},
  {"h": "郘", "p": "lǘ", "sp": "lu", "r": "邑", "sc": 8, "l": 5, "f": 8495},
  {"h": "醄", "p": "táo", "sp": "tao", "r": "酉", "sc": 15, "l": 5, "f": 8496},
  {"h": "醼", "p": "yàn", "sp": "yan", "r": "酉", "sc": 23, "l": 5, "f": 8497},
  {"h": "鋋", "p": "chán", "sp": "chan", "r": "金", "sc": 14, "l": 5, "f": 8498},
  {"h": "鏁", "p": "suǒ", "sp": "suo", "r": "金", "sc": 19, "l": 5, "f": 8499},
  {"h": "鏭", "p": "xī", "sp": "xi", "r": "金", "sc": 19, "l": 5, "f": 8500},
  {
    "h": "鐄",
    "p": "huáng",
    "sp": "huang",
    "r": "金",
    "sc": 19,
    "l": 5,
    "f": 8501
  },
  {"h": "鑣", "p": "biāo", "sp": "biao", "r": "金", "sc": 23, "l": 5, "f": 8502},
  {"h": "鑷", "p": "niè", "sp": "nie", "r": "金", "sc": 26, "l": 5, "f": 8503},
  {"h": "辁", "p": "quán", "sp": "quan", "r": "車", "sc": 10, "l": 5, "f": 8504},
  {"h": "阸", "p": "è", "sp": "e", "r": "阜", "sc": 7, "l": 5, "f": 8505},
  {"h": "牿", "p": "gù", "sp": "gu", "r": "牛", "sc": 11, "l": 5, "f": 8506},
  {
    "h": "鞝",
    "p": "shàng",
    "sp": "shang",
    "r": "革",
    "sc": 17,
    "l": 5,
    "f": 8507
  },
  {"h": "頞", "p": "è", "sp": "e", "r": "頁", "sc": 15, "l": 5, "f": 8508},
  {"h": "瞵", "p": "lín", "sp": "lin", "r": "目", "sc": 17, "l": 5, "f": 8509},
  {"h": "饤", "p": "dìng", "sp": "ding", "r": "食", "sc": 5, "l": 5, "f": 8510},
  {"h": "駃", "p": "jué", "sp": "jue", "r": "馬", "sc": 14, "l": 5, "f": 8511},
  {
    "h": "駉",
    "p": "jiōng",
    "sp": "jiong",
    "r": "馬",
    "sc": 15,
    "l": 5,
    "f": 8512
  },
  {"h": "騠", "p": "tí", "sp": "ti", "r": "馬", "sc": 19, "l": 5, "f": 8513},
  {"h": "驡", "p": "lóng", "sp": "long", "r": "馬", "sc": 26, "l": 5, "f": 8514},
  {"h": "髪", "p": "fà", "sp": "fa", "r": "髟", "sc": 14, "l": 5, "f": 8515},
  {"h": "鬚", "p": "xū", "sp": "xu", "r": "髟", "sc": 22, "l": 5, "f": 8516},
  {"h": "鬿", "p": "qí", "sp": "qi", "r": "鬼", "sc": 13, "l": 5, "f": 8517},
  {"h": "鮌", "p": "gǔn", "sp": "gun", "r": "魚", "sc": 16, "l": 5, "f": 8518},
  {"h": "鯁", "p": "gěng", "sp": "geng", "r": "魚", "sc": 18, "l": 5, "f": 8519},
  {"h": "鱻", "p": "xiān", "sp": "xian", "r": "魚", "sc": 33, "l": 5, "f": 8520},
  {"h": "鶚", "p": "è", "sp": "e", "r": "鳥", "sc": 20, "l": 5, "f": 8521},
  {"h": "鷫", "p": "sù", "sp": "su", "r": "鳥", "sc": 24, "l": 5, "f": 8522},
  {"h": "鹖", "p": "hé", "sp": "he", "r": "鳥", "sc": 14, "l": 5, "f": 8523},
  {"h": "磩", "p": "qì", "sp": "qi", "r": "石", "sc": 16, "l": 5, "f": 8524},
  {"h": "磭", "p": "chuò", "sp": "chuo", "r": "石", "sc": 16, "l": 5, "f": 8525},
  {"h": "稌", "p": "tú", "sp": "tu", "r": "禾", "sc": 12, "l": 5, "f": 8526},
  {"h": "穃", "p": "rong", "sp": "rong", "r": "禾", "sc": 15, "l": 5, "f": 8527},
  {"h": "綂", "p": "tǒng", "sp": "tong", "r": "糸", "sc": 13, "l": 5, "f": 8528},
  {"h": "膥", "p": "chūn", "sp": "chun", "r": "肉", "sc": 17, "l": 5, "f": 8529},
  {"h": "膱", "p": "zhí", "sp": "zhi", "r": "肉", "sc": 16, "l": 5, "f": 8530},
  {"h": "臿", "p": "chā", "sp": "cha", "r": "臼", "sc": 9, "l": 5, "f": 8531},
  {
    "h": "艎",
    "p": "huáng",
    "sp": "huang",
    "r": "舟",
    "sc": 15,
    "l": 5,
    "f": 8532
  },
  {"h": "苃", "p": "yǒu", "sp": "you", "r": "艸", "sc": 7, "l": 5, "f": 8533},
  {"h": "茍", "p": "jì", "sp": "ji", "r": "艸", "sc": 9, "l": 5, "f": 8534},
  {"h": "蒟", "p": "jǔ", "sp": "ju", "r": "艸", "sc": 13, "l": 5, "f": 8535},
  {"h": "藣", "p": "bēi", "sp": "bei", "r": "艸", "sc": 18, "l": 5, "f": 8536},
  {"h": "虈", "p": "xiāo", "sp": "xiao", "r": "艸", "sc": 24, "l": 5, "f": 8537},
  {"h": "蜐", "p": "jié", "sp": "jie", "r": "虫", "sc": 13, "l": 5, "f": 8538},
  {"h": "訅", "p": "qiú", "sp": "qiu", "r": "言", "sc": 9, "l": 5, "f": 8539},
  {"h": "詭", "p": "guǐ", "sp": "gui", "r": "言", "sc": 13, "l": 5, "f": 8540},
  {"h": "讈", "p": "lì", "sp": "li", "r": "言", "sc": 23, "l": 5, "f": 8541},
  {"h": "貿", "p": "mào", "sp": "mao", "r": "貝", "sc": 12, "l": 5, "f": 8542},
  {"h": "趷", "p": "kē", "sp": "ke", "r": "足", "sc": 10, "l": 5, "f": 8543},
  {
    "h": "蹌",
    "p": "qiāng",
    "sp": "qiang",
    "r": "足",
    "sc": 17,
    "l": 5,
    "f": 8544
  },
  {"h": "蹹", "p": "tá", "sp": "ta", "r": "足", "sc": 19, "l": 5, "f": 8545},
  {"h": "鈰", "p": "shì", "sp": "shi", "r": "金", "sc": 13, "l": 5, "f": 8546},
  {"h": "銑", "p": "xiǎn", "sp": "xian", "r": "金", "sc": 14, "l": 5, "f": 8547},
  {"h": "鋫", "p": "lí", "sp": "li", "r": "金", "sc": 15, "l": 5, "f": 8548},
  {"h": "鍔", "p": "è", "sp": "e", "r": "金", "sc": 17, "l": 5, "f": 8549},
  {"h": "鐨", "p": "fèi", "sp": "fei", "r": "金", "sc": 20, "l": 5, "f": 8550},
  {
    "h": "餉",
    "p": "xiǎng",
    "sp": "xiang",
    "r": "食",
    "sc": 14,
    "l": 5,
    "f": 8551
  },
  {"h": "騛", "p": "fēi", "sp": "fei", "r": "馬", "sc": 19, "l": 5, "f": 8552},
  {"h": "鯛", "p": "diāo", "sp": "diao", "r": "魚", "sc": 19, "l": 5, "f": 8553},
  {"h": "鰤", "p": "shī", "sp": "shi", "r": "魚", "sc": 21, "l": 5, "f": 8554},
  {"h": "鳳", "p": "fèng", "sp": "feng", "r": "鳥", "sc": 14, "l": 5, "f": 8555},
  {"h": "鵂", "p": "xiū", "sp": "xiu", "r": "鳥", "sc": 17, "l": 5, "f": 8556},
  {"h": "鶢", "p": "yuán", "sp": "yuan", "r": "鳥", "sc": 20, "l": 5, "f": 8557},
  {"h": "鶳", "p": "shī", "sp": "shi", "r": "鳥", "sc": 21, "l": 5, "f": 8558},
  {
    "h": "黋",
    "p": "kuàng",
    "sp": "kuang",
    "r": "黃",
    "sc": 17,
    "l": 5,
    "f": 8559
  },
  {"h": "齢", "p": "líng", "sp": "ling", "r": "齒", "sc": 17, "l": 5, "f": 8560},
  {
    "h": "磚",
    "p": "zhuān",
    "sp": "zhuan",
    "r": "石",
    "sc": 16,
    "l": 5,
    "f": 8561
  },
  {"h": "緲", "p": "miǎo", "sp": "miao", "r": "糸", "sc": 15, "l": 5, "f": 8562},
  {"h": "繳", "p": "jiǎo", "sp": "jiao", "r": "糸", "sc": 19, "l": 5, "f": 8563},
  {"h": "聝", "p": "guó", "sp": "guo", "r": "耳", "sc": 14, "l": 5, "f": 8564},
  {"h": "膴", "p": "hū", "sp": "hu", "r": "肉", "sc": 16, "l": 5, "f": 8565},
  {"h": "芵", "p": "jué", "sp": "jue", "r": "艸", "sc": 7, "l": 5, "f": 8566},
  {"h": "藫", "p": "tán", "sp": "tan", "r": "艸", "sc": 18, "l": 5, "f": 8567},
  {"h": "覓", "p": "mì", "sp": "mi", "r": "見", "sc": 11, "l": 5, "f": 8568},
  {"h": "誕", "p": "dàn", "sp": "dan", "r": "言", "sc": 13, "l": 5, "f": 8569},
  {"h": "誡", "p": "jiè", "sp": "jie", "r": "言", "sc": 14, "l": 5, "f": 8570},
  {"h": "貎", "p": "ní", "sp": "ni", "r": "豸", "sc": 15, "l": 5, "f": 8571},
  {"h": "賄", "p": "huì", "sp": "hui", "r": "貝", "sc": 13, "l": 5, "f": 8572},
  {"h": "跰", "p": "pián", "sp": "pian", "r": "足", "sc": 13, "l": 5, "f": 8573},
  {"h": "踥", "p": "qiè", "sp": "qie", "r": "足", "sc": 15, "l": 5, "f": 8574},
  {"h": "迆", "p": "yí", "sp": "yi", "r": "辵", "sc": 6, "l": 5, "f": 8575},
  {"h": "逕", "p": "jìng", "sp": "jing", "r": "辵", "sc": 10, "l": 5, "f": 8576},
  {"h": "逤", "p": "suò", "sp": "suo", "r": "辵", "sc": 10, "l": 5, "f": 8577},
  {"h": "鏢", "p": "biāo", "sp": "biao", "r": "金", "sc": 19, "l": 5, "f": 8578},
  {"h": "鏤", "p": "lòu", "sp": "lou", "r": "金", "sc": 19, "l": 5, "f": 8579},
  {"h": "鑄", "p": "zhù", "sp": "zhu", "r": "金", "sc": 22, "l": 5, "f": 8580},
  {"h": "鑽", "p": "zuān", "sp": "zuan", "r": "金", "sc": 27, "l": 5, "f": 8581},
  {"h": "頷", "p": "hàn", "sp": "han", "r": "頁", "sc": 16, "l": 5, "f": 8582},
  {"h": "鴉", "p": "yā", "sp": "ya", "r": "鳥", "sc": 15, "l": 5, "f": 8583},
  {"h": "鴰", "p": "guā", "sp": "gua", "r": "鳥", "sc": 17, "l": 5, "f": 8584},
  {"h": "麅", "p": "páo", "sp": "pao", "r": "鹿", "sc": 16, "l": 5, "f": 8585},
  {"h": "礱", "p": "lóng", "sp": "long", "r": "石", "sc": 21, "l": 5, "f": 8586},
  {"h": "祿", "p": "lù", "sp": "lu", "r": "示", "sc": 12, "l": 5, "f": 8587},
  {"h": "缷", "p": "xiè", "sp": "xie", "r": "缶", "sc": 8, "l": 5, "f": 8588},
  {"h": "虙", "p": "fú", "sp": "fu", "r": "虍", "sc": 11, "l": 5, "f": 8589},
  {"h": "齗", "p": "yín", "sp": "yin", "r": "齒", "sc": 19, "l": 5, "f": 8590},
  {"h": "裓", "p": "gé", "sp": "ge", "r": "衣", "sc": 12, "l": 5, "f": 8591},
  {"h": "铹", "p": "láo", "sp": "lao", "r": "金", "sc": 12, "l": 5, "f": 8592},
  {"h": "睙", "p": "liè", "sp": "lie", "r": "目", "sc": 13, "l": 5, "f": 8593},
  {"h": "矟", "p": "shuò", "sp": "shuo", "r": "矛", "sc": 12, "l": 5, "f": 8594},
  {"h": "硣", "p": "xiāo", "sp": "xiao", "r": "石", "sc": 12, "l": 5, "f": 8595},
  {"h": "磹", "p": "tán", "sp": "tan", "r": "石", "sc": 17, "l": 5, "f": 8596},
  {"h": "礀", "p": "jian", "sp": "jian", "r": "石", "sc": 17, "l": 5, "f": 8597},
  {"h": "禮", "p": "lǐ", "sp": "li", "r": "示", "sc": 17, "l": 5, "f": 8598},
  {"h": "秶", "p": "zī", "sp": "zi", "r": "禾", "sc": 11, "l": 5, "f": 8599},
  {"h": "窇", "p": "báo", "sp": "bao", "r": "穴", "sc": 10, "l": 5, "f": 8600},
  {"h": "竷", "p": "kǎn", "sp": "kan", "r": "立", "sc": 20, "l": 5, "f": 8601},
  {"h": "笣", "p": "bāo", "sp": "bao", "r": "竹", "sc": 11, "l": 5, "f": 8602},
  {"h": "篃", "p": "mèi", "sp": "mei", "r": "竹", "sc": 15, "l": 5, "f": 8603},
  {"h": "簬", "p": "lù", "sp": "lu", "r": "竹", "sc": 19, "l": 5, "f": 8604},
  {"h": "簯", "p": "qi", "sp": "qi", "r": "竹", "sc": 18, "l": 5, "f": 8605},
  {"h": "粓", "p": "gān", "sp": "gan", "r": "米", "sc": 11, "l": 5, "f": 8606},
  {"h": "粶", "p": "lù", "sp": "lu", "r": "米", "sc": 14, "l": 5, "f": 8607},
  {"h": "紋", "p": "wén", "sp": "wen", "r": "糸", "sc": 10, "l": 5, "f": 8608},
  {"h": "紓", "p": "shū", "sp": "shu", "r": "糸", "sc": 10, "l": 5, "f": 8609},
  {"h": "絞", "p": "jiǎo", "sp": "jiao", "r": "糸", "sc": 12, "l": 5, "f": 8610},
  {"h": "緹", "p": "tí", "sp": "ti", "r": "糸", "sc": 15, "l": 5, "f": 8611},
  {"h": "縷", "p": "lǚ", "sp": "lu", "r": "糸", "sc": 17, "l": 5, "f": 8612},
  {"h": "縼", "p": "xuàn", "sp": "xuan", "r": "糸", "sc": 17, "l": 5, "f": 8613},
  {"h": "繆", "p": "móu", "sp": "mou", "r": "糸", "sc": 17, "l": 5, "f": 8614},
  {"h": "繕", "p": "shàn", "sp": "shan", "r": "糸", "sc": 18, "l": 5, "f": 8615},
  {"h": "繚", "p": "liáo", "sp": "liao", "r": "糸", "sc": 18, "l": 5, "f": 8616},
  {"h": "繞", "p": "rào", "sp": "rao", "r": "糸", "sc": 18, "l": 5, "f": 8617},
  {"h": "绤", "p": "xì", "sp": "xi", "r": "糸", "sc": 10, "l": 5, "f": 8618},
  {"h": "罆", "p": "guàn", "sp": "guan", "r": "缶", "sc": 17, "l": 5, "f": 8619},
  {"h": "羈", "p": "jī", "sp": "ji", "r": "网", "sc": 24, "l": 5, "f": 8620},
  {"h": "翆", "p": "cuì", "sp": "cui", "r": "羽", "sc": 10, "l": 5, "f": 8621},
  {"h": "翑", "p": "qú", "sp": "qu", "r": "羽", "sc": 11, "l": 5, "f": 8622},
  {"h": "聵", "p": "kuì", "sp": "kui", "r": "耳", "sc": 18, "l": 5, "f": 8623},
  {"h": "脵", "p": "gǔ", "sp": "gu", "r": "肉", "sc": 11, "l": 5, "f": 8624},
  {"h": "臄", "p": "jué", "sp": "jue", "r": "肉", "sc": 17, "l": 5, "f": 8625},
  {"h": "茽", "p": "zhòng", "sp": "zhong", "r": "艸", "sc": 9, "l": 5, "f": 8626},
  {
    "h": "莊",
    "p": "zhuāng",
    "sp": "zhuang",
    "r": "艸",
    "sc": 10,
    "l": 5,
    "f": 8627
  },
  {"h": "莢", "p": "jiá", "sp": "jia", "r": "艸", "sc": 10, "l": 5, "f": 8628},
  {"h": "莵", "p": "tu", "sp": "tu", "r": "艸", "sc": 10, "l": 5, "f": 8629},
  {"h": "菕", "p": "lún", "sp": "lun", "r": "艸", "sc": 11, "l": 5, "f": 8630},
  {"h": "蓛", "p": "cè", "sp": "ce", "r": "艸", "sc": 13, "l": 5, "f": 8631},
  {"h": "蓷", "p": "tuī", "sp": "tui", "r": "艸", "sc": 14, "l": 5, "f": 8632},
  {"h": "蓸", "p": "cáo", "sp": "cao", "r": "艸", "sc": 14, "l": 5, "f": 8633},
  {"h": "蔘", "p": "shēn", "sp": "shen", "r": "艸", "sc": 14, "l": 5, "f": 8634},
  {"h": "蔞", "p": "lóu", "sp": "lou", "r": "艸", "sc": 14, "l": 5, "f": 8635},
  {"h": "蕘", "p": "ráo", "sp": "rao", "r": "艸", "sc": 15, "l": 5, "f": 8636},
  {"h": "蕼", "p": "sì", "sp": "si", "r": "艸", "sc": 16, "l": 5, "f": 8637},
  {
    "h": "薔",
    "p": "qiáng",
    "sp": "qiang",
    "r": "艸",
    "sc": 16,
    "l": 5,
    "f": 8638
  },
  {"h": "藖", "p": "xián", "sp": "xian", "r": "艸", "sc": 18, "l": 5, "f": 8639},
  {"h": "藢", "p": "zhǐ", "sp": "zhi", "r": "艸", "sc": 18, "l": 5, "f": 8640},
  {"h": "藶", "p": "lì", "sp": "li", "r": "艸", "sc": 19, "l": 5, "f": 8641},
  {"h": "蛢", "p": "píng", "sp": "ping", "r": "虫", "sc": 12, "l": 5, "f": 8642},
  {"h": "蟯", "p": "náo", "sp": "nao", "r": "虫", "sc": 18, "l": 5, "f": 8643},
  {"h": "蠄", "p": "qín", "sp": "qin", "r": "虫", "sc": 18, "l": 5, "f": 8644},
  {"h": "蠈", "p": "zéi", "sp": "zei", "r": "虫", "sc": 19, "l": 5, "f": 8645},
  {"h": "蠎", "p": "mǎng", "sp": "mang", "r": "虫", "sc": 17, "l": 5, "f": 8646},
  {"h": "蠣", "p": "lì", "sp": "li", "r": "虫", "sc": 20, "l": 5, "f": 8647},
  {"h": "襪", "p": "wà", "sp": "wa", "r": "衣", "sc": 19, "l": 5, "f": 8648},
  {"h": "覙", "p": "luó", "sp": "luo", "r": "見", "sc": 12, "l": 5, "f": 8649},
  {"h": "訑", "p": "yí", "sp": "yi", "r": "言", "sc": 10, "l": 5, "f": 8650},
  {"h": "訣", "p": "jué", "sp": "jue", "r": "言", "sc": 11, "l": 5, "f": 8651},
  {"h": "詰", "p": "jié", "sp": "jie", "r": "言", "sc": 13, "l": 5, "f": 8652},
  {"h": "詸", "p": "mí", "sp": "mi", "r": "言", "sc": 13, "l": 5, "f": 8653},
  {"h": "誰", "p": "shuí", "sp": "shui", "r": "言", "sc": 15, "l": 5, "f": 8654},
  {"h": "諏", "p": "zōu", "sp": "zou", "r": "言", "sc": 15, "l": 5, "f": 8655},
  {"h": "諛", "p": "yú", "sp": "yu", "r": "言", "sc": 15, "l": 5, "f": 8656},
  {"h": "諮", "p": "zī", "sp": "zi", "r": "言", "sc": 16, "l": 5, "f": 8657},
  {"h": "諱", "p": "huì", "sp": "hui", "r": "言", "sc": 16, "l": 5, "f": 8658},
  {"h": "謥", "p": "còng", "sp": "cong", "r": "言", "sc": 18, "l": 5, "f": 8659},
  {"h": "謨", "p": "mó", "sp": "mo", "r": "言", "sc": 17, "l": 5, "f": 8660},
  {"h": "譫", "p": "zhān", "sp": "zhan", "r": "言", "sc": 20, "l": 5, "f": 8661},
  {"h": "貋", "p": "hàn", "sp": "han", "r": "豸", "sc": 14, "l": 5, "f": 8662},
  {"h": "貛", "p": "huān", "sp": "huan", "r": "豸", "sc": 24, "l": 5, "f": 8663},
  {"h": "貫", "p": "guàn", "sp": "guan", "r": "貝", "sc": 11, "l": 5, "f": 8664},
  {"h": "賓", "p": "bīn", "sp": "bin", "r": "貝", "sc": 14, "l": 5, "f": 8665},
  {
    "h": "賞",
    "p": "shǎng",
    "sp": "shang",
    "r": "貝",
    "sc": 15,
    "l": 5,
    "f": 8666
  },
  {"h": "贁", "p": "bài", "sp": "bai", "r": "貝", "sc": 18, "l": 5, "f": 8667},
  {"h": "贂", "p": "chěn", "sp": "chen", "r": "貝", "sc": 18, "l": 5, "f": 8668},
  {"h": "贉", "p": "dàn", "sp": "dan", "r": "貝", "sc": 19, "l": 5, "f": 8669},
  {"h": "趤", "p": "dàng", "sp": "dang", "r": "走", "sc": 15, "l": 5, "f": 8670},
  {"h": "趯", "p": "tì", "sp": "ti", "r": "走", "sc": 21, "l": 5, "f": 8671},
  {"h": "趶", "p": "kù", "sp": "ku", "r": "足", "sc": 10, "l": 5, "f": 8672},
  {"h": "跔", "p": "jū", "sp": "ju", "r": "足", "sc": 12, "l": 5, "f": 8673},
  {"h": "踴", "p": "yǒng", "sp": "yong", "r": "足", "sc": 16, "l": 5, "f": 8674},
  {"h": "蹆", "p": "tuǐ", "sp": "tui", "r": "足", "sc": 16, "l": 5, "f": 8675},
  {"h": "塥", "p": "gé", "sp": "ge", "r": "土", "sc": 13, "l": 5, "f": 8676},
  {"h": "輝", "p": "huī", "sp": "hui", "r": "車", "sc": 15, "l": 5, "f": 8677},
  {"h": "輰", "p": "yáng", "sp": "yang", "r": "車", "sc": 16, "l": 5, "f": 8678},
  {"h": "輿", "p": "yú", "sp": "yu", "r": "車", "sc": 17, "l": 5, "f": 8679},
  {"h": "轂", "p": "gǔ", "sp": "gu", "r": "車", "sc": 17, "l": 5, "f": 8680},
  {"h": "邁", "p": "mài", "sp": "mai", "r": "辵", "sc": 15, "l": 5, "f": 8681},
  {"h": "邼", "p": "kuāng", "sp": "kuang", "r": "邑", "sc": 8, "l": 5, "f": 8682},
  {"h": "釟", "p": "bā", "sp": "ba", "r": "金", "sc": 10, "l": 5, "f": 8683},
  {"h": "釠", "p": "luàn", "sp": "luan", "r": "金", "sc": 10, "l": 5, "f": 8684},
  {"h": "釣", "p": "diào", "sp": "diao", "r": "金", "sc": 11, "l": 5, "f": 8685},
  {"h": "鈉", "p": "nà", "sp": "na", "r": "金", "sc": 12, "l": 5, "f": 8686},
  {"h": "鈴", "p": "líng", "sp": "ling", "r": "金", "sc": 13, "l": 5, "f": 8687},
  {"h": "鈺", "p": "yù", "sp": "yu", "r": "金", "sc": 13, "l": 5, "f": 8688},
  {"h": "鉂", "p": "shǐ", "sp": "shi", "r": "金", "sc": 13, "l": 5, "f": 8689},
  {"h": "銊", "p": "xù", "sp": "xu", "r": "金", "sc": 14, "l": 5, "f": 8690},
  {"h": "銘", "p": "míng", "sp": "ming", "r": "金", "sc": 14, "l": 5, "f": 8691},
  {"h": "鋤", "p": "chú", "sp": "chu", "r": "金", "sc": 15, "l": 5, "f": 8692},
  {"h": "鋭", "p": "ruì", "sp": "rui", "r": "金", "sc": 15, "l": 5, "f": 8693},
  {"h": "錃", "p": "bēi", "sp": "bei", "r": "金", "sc": 16, "l": 5, "f": 8694},
  {"h": "錈", "p": "juǎn", "sp": "juan", "r": "金", "sc": 16, "l": 5, "f": 8695},
  {"h": "漭", "p": "mǎng", "sp": "mang", "r": "水", "sc": 13, "l": 5, "f": 8696},
  {"h": "錨", "p": "máo", "sp": "mao", "r": "金", "sc": 16, "l": 5, "f": 8697},
  {"h": "鍤", "p": "chā", "sp": "cha", "r": "金", "sc": 17, "l": 5, "f": 8698},
  {"h": "鏆", "p": "guàn", "sp": "guan", "r": "金", "sc": 19, "l": 5, "f": 8699},
  {"h": "鐖", "p": "jī", "sp": "ji", "r": "金", "sc": 20, "l": 5, "f": 8700},
  {"h": "鐧", "p": "jiān", "sp": "jian", "r": "金", "sc": 20, "l": 5, "f": 8701},
  {"h": "鐳", "p": "léi", "sp": "lei", "r": "金", "sc": 21, "l": 5, "f": 8702},
  {"h": "鑱", "p": "chán", "sp": "chan", "r": "金", "sc": 25, "l": 5, "f": 8703},
  {"h": "鑾", "p": "luán", "sp": "luan", "r": "金", "sc": 27, "l": 5, "f": 8704},
  {"h": "镃", "p": "zī", "sp": "zi", "r": "金", "sc": 14, "l": 5, "f": 8705},
  {"h": "閬", "p": "làng", "sp": "lang", "r": "門", "sc": 15, "l": 5, "f": 8706},
  {"h": "閾", "p": "yù", "sp": "yu", "r": "門", "sc": 16, "l": 5, "f": 8707},
  {"h": "閿", "p": "wén", "sp": "wen", "r": "門", "sc": 16, "l": 5, "f": 8708},
  {"h": "陜", "p": "xiá", "sp": "xia", "r": "阜", "sc": 9, "l": 5, "f": 8709},
  {"h": "陝", "p": "shǎn", "sp": "shan", "r": "阜", "sc": 9, "l": 5, "f": 8710},
  {"h": "霠", "p": "yīn", "sp": "yin", "r": "雨", "sc": 17, "l": 5, "f": 8711},
  {"h": "鞏", "p": "gǒng", "sp": "gong", "r": "革", "sc": 15, "l": 5, "f": 8712},
  {"h": "鞖", "p": "suī", "sp": "sui", "r": "革", "sc": 16, "l": 5, "f": 8713},
  {"h": "韨", "p": "fú", "sp": "fu", "r": "韋", "sc": 9, "l": 5, "f": 8714},
  {"h": "韼", "p": "péng", "sp": "peng", "r": "音", "sc": 19, "l": 5, "f": 8715},
  {"h": "韽", "p": "ān", "sp": "an", "r": "音", "sc": 20, "l": 5, "f": 8716},
  {"h": "頒", "p": "bān", "sp": "ban", "r": "頁", "sc": 13, "l": 5, "f": 8717},
  {"h": "頤", "p": "yí", "sp": "yi", "r": "頁", "sc": 16, "l": 5, "f": 8718},
  {"h": "顛", "p": "diān", "sp": "dian", "r": "頁", "sc": 19, "l": 5, "f": 8719},
  {"h": "飳", "p": "tǒu", "sp": "tou", "r": "食", "sc": 13, "l": 5, "f": 8720},
  {"h": "飸", "p": "tāo", "sp": "tao", "r": "食", "sc": 14, "l": 5, "f": 8721},
  {"h": "餆", "p": "yáo", "sp": "yao", "r": "食", "sc": 14, "l": 5, "f": 8722},
  {"h": "館", "p": "guǎn", "sp": "guan", "r": "食", "sc": 16, "l": 5, "f": 8723},
  {"h": "饅", "p": "mán", "sp": "man", "r": "食", "sc": 19, "l": 5, "f": 8724},
  {"h": "駐", "p": "zhù", "sp": "zhu", "r": "馬", "sc": 15, "l": 5, "f": 8725},
  {"h": "駡", "p": "mà", "sp": "ma", "r": "口", "sc": 16, "l": 5, "f": 8726},
  {"h": "駷", "p": "sǒng", "sp": "song", "r": "馬", "sc": 17, "l": 5, "f": 8727},
  {
    "h": "騁",
    "p": "chěng",
    "sp": "cheng",
    "r": "馬",
    "sc": 17,
    "l": 5,
    "f": 8728
  },
  {"h": "騫", "p": "qiān", "sp": "qian", "r": "馬", "sc": 20, "l": 5, "f": 8729},
  {"h": "騷", "p": "sāo", "sp": "sao", "r": "馬", "sc": 19, "l": 5, "f": 8730},
  {"h": "騼", "p": "lù", "sp": "lu", "r": "馬", "sc": 21, "l": 5, "f": 8731},
  {"h": "髕", "p": "bìn", "sp": "bin", "r": "骨", "sc": 23, "l": 5, "f": 8732},
  {"h": "魜", "p": "rén", "sp": "ren", "r": "魚", "sc": 13, "l": 5, "f": 8733},
  {"h": "鰫", "p": "yóng", "sp": "yong", "r": "魚", "sc": 21, "l": 5, "f": 8734},
  {"h": "鱗", "p": "lín", "sp": "lin", "r": "魚", "sc": 23, "l": 5, "f": 8735},
  {"h": "鱽", "p": "dāo", "sp": "dao", "r": "魚", "sc": 10, "l": 5, "f": 8736},
  {"h": "鲬", "p": "yǒng", "sp": "yong", "r": "魚", "sc": 15, "l": 5, "f": 8737},
  {"h": "鴻", "p": "hóng", "sp": "hong", "r": "鳥", "sc": 17, "l": 5, "f": 8738},
  {"h": "鶆", "p": "lái", "sp": "lai", "r": "鳥", "sc": 19, "l": 5, "f": 8739},
  {"h": "鶊", "p": "gēng", "sp": "geng", "r": "鳥", "sc": 19, "l": 5, "f": 8740},
  {"h": "鷳", "p": "xián", "sp": "xian", "r": "鳥", "sc": 23, "l": 5, "f": 8741},
  {"h": "鷿", "p": "pì", "sp": "pi", "r": "鳥", "sc": 24, "l": 5, "f": 8742},
  {"h": "鸁", "p": "luó", "sp": "luo", "r": "鳥", "sc": 24, "l": 5, "f": 8743},
  {
    "h": "麞",
    "p": "zhāng",
    "sp": "zhang",
    "r": "鹿",
    "sc": 22,
    "l": 5,
    "f": 8744
  },
  {"h": "鼏", "p": "mì", "sp": "mi", "r": "鼎", "sc": 14, "l": 5, "f": 8745},
  {"h": "鼜", "p": "qì", "sp": "qi", "r": "鼓", "sc": 22, "l": 5, "f": 8746},
  {"h": "鼫", "p": "shí", "sp": "shi", "r": "鼠", "sc": 18, "l": 5, "f": 8747},
  {"h": "齦", "p": "kěn", "sp": "ken", "r": "齒", "sc": 21, "l": 5, "f": 8748},
  {"h": "䦂", "p": "shàn", "sp": "shan", "r": "", "sc": 18, "l": 5, "f": 8749},
  {"h": "䲟", "p": "yìn", "sp": "yin", "r": "", "sc": 16, "l": 5, "f": 8750},
  {"h": "䴓", "p": "shī", "sp": "shi", "r": "", "sc": 15, "l": 5, "f": 8751},
  {"h": "䴕", "p": "liè", "sp": "lie", "r": "", "sc": 17, "l": 5, "f": 8752},
  {"h": "䶮", "p": "yǎn", "sp": "yan", "r": "", "sc": 20, "l": 5, "f": 8753},
  {"h": "盬", "p": "gǔ", "sp": "gu", "r": "皿", "sc": 18, "l": 5, "f": 8754},
  {"h": "祅", "p": "yāo", "sp": "yao", "r": "示", "sc": 8, "l": 5, "f": 8755},
  {"h": "禡", "p": "mà", "sp": "ma", "r": "示", "sc": 14, "l": 5, "f": 8756},
  {"h": "禭", "p": "suì", "sp": "sui", "r": "示", "sc": 16, "l": 5, "f": 8757},
  {"h": "糱", "p": "niè", "sp": "nie", "r": "米", "sc": 22, "l": 5, "f": 8758},
  {"h": "縂", "p": "zǒng", "sp": "zong", "r": "糸", "sc": 15, "l": 5, "f": 8759},
  {"h": "蠥", "p": "niè", "sp": "nie", "r": "虫", "sc": 22, "l": 5, "f": 8760},
  {"h": "軷", "p": "bá", "sp": "ba", "r": "車", "sc": 12, "l": 5, "f": 8761},
  {"h": "鄮", "p": "mào", "sp": "mao", "r": "邑", "sc": 14, "l": 5, "f": 8762},
  {"h": "釱", "p": "dì", "sp": "di", "r": "金", "sc": 11, "l": 5, "f": 8763},
  {"h": "銦", "p": "yīn", "sp": "yin", "r": "金", "sc": 14, "l": 5, "f": 8764},
  {"h": "錣", "p": "zhuì", "sp": "zhui", "r": "金", "sc": 16, "l": 5, "f": 8765},
  {"h": "鑢", "p": "lǜ", "sp": "lu", "r": "金", "sc": 23, "l": 5, "f": 8766},
  {"h": "钑", "p": "jí", "sp": "ji", "r": "金", "sc": 8, "l": 5, "f": 8767},
  {
    "h": "饟",
    "p": "xiǎng",
    "sp": "xiang",
    "r": "食",
    "sc": 25,
    "l": 5,
    "f": 8768
  },
  {"h": "鯅", "p": "shān", "sp": "shan", "r": "魚", "sc": 17, "l": 5, "f": 8769},
  {"h": "鼊", "p": "bì", "sp": "bi", "r": "黽", "sc": 26, "l": 5, "f": 8770},
  {"h": "䴖", "p": "jīng", "sp": "jing", "r": "", "sc": 19, "l": 5, "f": 8771},
  {"h": "簫", "p": "xiāo", "sp": "xiao", "r": "竹", "sc": 19, "l": 5, "f": 8772},
  {"h": "粛", "p": "sù", "sp": "su", "r": "聿", "sc": 11, "l": 5, "f": 8773},
  {
    "h": "粻",
    "p": "zhāng",
    "sp": "zhang",
    "r": "米",
    "sc": 14,
    "l": 5,
    "f": 8774
  },
  {"h": "饈", "p": "xiū", "sp": "xiu", "r": "食", "sc": 18, "l": 5, "f": 8775},
  {"h": "盌", "p": "wǎn", "sp": "wan", "r": "皿", "sc": 10, "l": 5, "f": 8776},
  {"h": "瞂", "p": "fá", "sp": "fa", "r": "目", "sc": 14, "l": 5, "f": 8777},
  {"h": "鋅", "p": "xīn", "sp": "xin", "r": "金", "sc": 15, "l": 5, "f": 8778},
  {"h": "靮", "p": "dí", "sp": "di", "r": "革", "sc": 12, "l": 5, "f": 8779},
  {"h": "盨", "p": "xǔ", "sp": "xu", "r": "皿", "sc": 17, "l": 5, "f": 8780},
  {"h": "県", "p": "xiàn", "sp": "xian", "r": "目", "sc": 9, "l": 5, "f": 8781},
  {"h": "箥", "p": "bǒ", "sp": "bo", "r": "竹", "sc": 14, "l": 5, "f": 8782},
  {"h": "簕", "p": "lè", "sp": "le", "r": "竹", "sc": 17, "l": 5, "f": 8783},
  {"h": "臖", "p": "xìng", "sp": "xing", "r": "肉", "sc": 20, "l": 5, "f": 8784},
  {"h": "鎔", "p": "róng", "sp": "rong", "r": "金", "sc": 18, "l": 5, "f": 8785},
  {"h": "鏐", "p": "liú", "sp": "liu", "r": "金", "sc": 19, "l": 5, "f": 8786},
  {"h": "顔", "p": "yán", "sp": "yan", "r": "頁", "sc": 18, "l": 5, "f": 8787},
  {"h": "盿", "p": "mín", "sp": "min", "r": "目", "sc": 9, "l": 5, "f": 8788},
  {"h": "萵", "p": "wō", "sp": "wo", "r": "艸", "sc": 11, "l": 5, "f": 8789},
  {"h": "蔊", "p": "hǎn", "sp": "han", "r": "艸", "sc": 14, "l": 5, "f": 8790},
  {"h": "覥", "p": "tiǎn", "sp": "tian", "r": "見", "sc": 15, "l": 5, "f": 8791},
  {"h": "貲", "p": "zī", "sp": "zi", "r": "貝", "sc": 13, "l": 5, "f": 8792},
  {"h": "驢", "p": "lǘ", "sp": "lu", "r": "馬", "sc": 26, "l": 5, "f": 8793},
  {"h": "鶸", "p": "ruò", "sp": "ruo", "r": "鳥", "sc": 21, "l": 5, "f": 8794},
  {"h": "磑", "p": "wéi", "sp": "wei", "r": "石", "sc": 15, "l": 5, "f": 8795},
  {"h": "窎", "p": "diào", "sp": "diao", "r": "穴", "sc": 10, "l": 5, "f": 8796},
  {"h": "詒", "p": "yí", "sp": "yi", "r": "言", "sc": 12, "l": 5, "f": 8797},
  {"h": "鉔", "p": "zā", "sp": "za", "r": "金", "sc": 13, "l": 5, "f": 8798},
  {
    "h": "鏱",
    "p": "zhang",
    "sp": "zhang",
    "r": "金",
    "sc": 19,
    "l": 5,
    "f": 8799
  },
  {"h": "隺", "p": "hú", "sp": "hu", "r": "隹", "sc": 10, "l": 5, "f": 8800},
  {"h": "饜", "p": "yàn", "sp": "yan", "r": "食", "sc": 23, "l": 5, "f": 8801},
  {
    "h": "荘",
    "p": "zhuāng",
    "sp": "zhuang",
    "r": "艸",
    "sc": 9,
    "l": 5,
    "f": 8802
  },
  {"h": "蒻", "p": "ruò", "sp": "ruo", "r": "艸", "sc": 13, "l": 5, "f": 8803},
  {"h": "蔴", "p": "má", "sp": "ma", "r": "艸", "sc": 14, "l": 5, "f": 8804},
  {"h": "裛", "p": "yì", "sp": "yi", "r": "衣", "sc": 13, "l": 5, "f": 8805},
  {"h": "鈙", "p": "qín", "sp": "qin", "r": "金", "sc": 12, "l": 5, "f": 8806},
  {"h": "鎋", "p": "xiá", "sp": "xia", "r": "金", "sc": 18, "l": 5, "f": 8807},
  {"h": "鑀", "p": "ài", "sp": "ai", "r": "金", "sc": 21, "l": 5, "f": 8808},
  {"h": "髴", "p": "fú", "sp": "fu", "r": "髟", "sc": 15, "l": 5, "f": 8809},
  {"h": "鴓", "p": "miè", "sp": "mie", "r": "鳥", "sc": 16, "l": 5, "f": 8810},
  {"h": "鸴", "p": "xué", "sp": "xue", "r": "鳥", "sc": 10, "l": 5, "f": 8811},
  {"h": "黿", "p": "yuán", "sp": "yuan", "r": "黽", "sc": 17, "l": 5, "f": 8812},
  {"h": "秹", "p": "rěn", "sp": "ren", "r": "禾", "sc": 11, "l": 5, "f": 8813},
  {"h": "茢", "p": "liè", "sp": "lie", "r": "艸", "sc": 9, "l": 5, "f": 8814},
  {"h": "蘁", "p": "wù", "sp": "wu", "r": "艸", "sc": 19, "l": 5, "f": 8815},
  {"h": "蚑", "p": "qí", "sp": "qi", "r": "虫", "sc": 10, "l": 5, "f": 8816},
  {"h": "鏻", "p": "lín", "sp": "lin", "r": "金", "sc": 20, "l": 5, "f": 8817},
  {"h": "霤", "p": "liù", "sp": "liu", "r": "雨", "sc": 18, "l": 5, "f": 8818},
  {"h": "霩", "p": "kuò", "sp": "kuo", "r": "雨", "sc": 18, "l": 5, "f": 8819},
  {"h": "頺", "p": "tuí", "sp": "tui", "r": "頁", "sc": 16, "l": 5, "f": 8820},
  {"h": "餚", "p": "yáo", "sp": "yao", "r": "食", "sc": 16, "l": 5, "f": 8821},
  {"h": "鯂", "p": "su", "sp": "su", "r": "魚", "sc": 18, "l": 5, "f": 8822},
  {"h": "籔", "p": "sǒu", "sp": "sou", "r": "竹", "sc": 21, "l": 5, "f": 8823},
  {"h": "蕁", "p": "qián", "sp": "qian", "r": "艸", "sc": 15, "l": 5, "f": 8824},
  {"h": "蘘", "p": "ráng", "sp": "rang", "r": "艸", "sc": 20, "l": 5, "f": 8825},
  {"h": "雑", "p": "zá", "sp": "za", "r": "隹", "sc": 14, "l": 5, "f": 8826},
  {"h": "鬶", "p": "guī", "sp": "gui", "r": "鬲", "sc": 18, "l": 5, "f": 8827},
  {"h": "鯵", "p": "shēn", "sp": "shen", "r": "魚", "sc": 19, "l": 5, "f": 8828},
  {"h": "鰧", "p": "téng", "sp": "teng", "r": "魚", "sc": 21, "l": 5, "f": 8829},
  {
    "h": "鱂",
    "p": "jiāng",
    "sp": "jiang",
    "r": "魚",
    "sc": 22,
    "l": 5,
    "f": 8830
  },
  {"h": "鳂", "p": "wēi", "sp": "wei", "r": "魚", "sc": 17, "l": 5, "f": 8831},
  {
    "h": "鵟",
    "p": "kuáng",
    "sp": "kuang",
    "r": "鳥",
    "sc": 18,
    "l": 5,
    "f": 8832
  },
  {"h": "盠", "p": "lí", "sp": "li", "r": "皿", "sc": 14, "l": 5, "f": 8833},
  {"h": "盳", "p": "wàng", "sp": "wang", "r": "目", "sc": 8, "l": 5, "f": 8834},
  {"h": "盽", "p": "fēng", "sp": "feng", "r": "目", "sc": 9, "l": 5, "f": 8835},
  {"h": "眃", "p": "yún", "sp": "yun", "r": "目", "sc": 9, "l": 5, "f": 8836},
  {"h": "眏", "p": "yāng", "sp": "yang", "r": "目", "sc": 10, "l": 5, "f": 8837},
  {"h": "睧", "p": "hūn", "sp": "hun", "r": "目", "sc": 13, "l": 5, "f": 8838},
  {
    "h": "硄",
    "p": "kuāng",
    "sp": "kuang",
    "r": "石",
    "sc": 11,
    "l": 5,
    "f": 8839
  },
  {"h": "碔", "p": "wǔ", "sp": "wu", "r": "石", "sc": 13, "l": 5, "f": 8840},
  {"h": "碦", "p": "kè", "sp": "ke", "r": "石", "sc": 14, "l": 5, "f": 8841},
  {"h": "碸", "p": "fēng", "sp": "feng", "r": "石", "sc": 14, "l": 5, "f": 8842},
  {"h": "磃", "p": "sī", "sp": "si", "r": "石", "sc": 15, "l": 5, "f": 8843},
  {"h": "礘", "p": "è", "sp": "e", "r": "石", "sc": 19, "l": 5, "f": 8844},
  {"h": "礮", "p": "pào", "sp": "pao", "r": "石", "sc": 21, "l": 5, "f": 8845},
  {"h": "祘", "p": "suàn", "sp": "suan", "r": "示", "sc": 10, "l": 5, "f": 8846},
  {"h": "禤", "p": "xuān", "sp": "xuan", "r": "示", "sc": 15, "l": 5, "f": 8847},
  {"h": "秌", "p": "qiū", "sp": "qiu", "r": "火", "sc": 9, "l": 5, "f": 8848},
  {"h": "秙", "p": "kù", "sp": "ku", "r": "禾", "sc": 10, "l": 5, "f": 8849},
  {"h": "秢", "p": "líng", "sp": "ling", "r": "禾", "sc": 10, "l": 5, "f": 8850},
  {"h": "稧", "p": "xì", "sp": "xi", "r": "禾", "sc": 14, "l": 5, "f": 8851},
  {"h": "稬", "p": "nuò", "sp": "nuo", "r": "禾", "sc": 14, "l": 5, "f": 8852},
  {"h": "稲", "p": "dào", "sp": "dao", "r": "禾", "sc": 14, "l": 5, "f": 8853},
  {"h": "稸", "p": "xù", "sp": "xu", "r": "禾", "sc": 15, "l": 5, "f": 8854},
  {
    "h": "穜",
    "p": "zhǒng",
    "sp": "zhong",
    "r": "禾",
    "sc": 17,
    "l": 5,
    "f": 8855
  },
  {"h": "竒", "p": "qí", "sp": "qi", "r": "立", "sc": 9, "l": 5, "f": 8856},
  {"h": "筶", "p": "gào", "sp": "gao", "r": "竹", "sc": 13, "l": 5, "f": 8857},
  {
    "h": "篬",
    "p": "qiāng",
    "sp": "qiang",
    "r": "竹",
    "sc": 16,
    "l": 5,
    "f": 8858
  },
  {"h": "籒", "p": "zhòu", "sp": "zhou", "r": "竹", "sc": 21, "l": 5, "f": 8859},
  {"h": "籜", "p": "tuò", "sp": "tuo", "r": "竹", "sc": 22, "l": 5, "f": 8860},
  {"h": "籞", "p": "yù", "sp": "yu", "r": "竹", "sc": 23, "l": 5, "f": 8861},
  {"h": "籶", "p": "shēn", "sp": "shen", "r": "米", "sc": 8, "l": 5, "f": 8862},
  {"h": "糂", "p": "sǎn", "sp": "san", "r": "米", "sc": 15, "l": 5, "f": 8863},
  {"h": "糉", "p": "zòng", "sp": "zong", "r": "米", "sc": 15, "l": 5, "f": 8864},
  {
    "h": "綡",
    "p": "liáng",
    "sp": "liang",
    "r": "糸",
    "sc": 14,
    "l": 5,
    "f": 8865
  },
  {"h": "綶", "p": "guǒ", "sp": "guo", "r": "糸", "sc": 14, "l": 5, "f": 8866},
  {"h": "緄", "p": "gǔn", "sp": "gun", "r": "糸", "sc": 14, "l": 5, "f": 8867},
  {
    "h": "緗",
    "p": "xiāng",
    "sp": "xiang",
    "r": "糸",
    "sc": 15,
    "l": 5,
    "f": 8868
  },
  {"h": "緼", "p": "yùn", "sp": "yun", "r": "糸", "sc": 15, "l": 5, "f": 8869},
  {"h": "縁", "p": "yuán", "sp": "yuan", "r": "糸", "sc": 15, "l": 5, "f": 8870},
  {"h": "縋", "p": "zhuì", "sp": "zhui", "r": "糸", "sc": 15, "l": 5, "f": 8871},
  {"h": "縞", "p": "gǎo", "sp": "gao", "r": "糸", "sc": 16, "l": 5, "f": 8872},
  {"h": "繤", "p": "zuǎn", "sp": "zuan", "r": "糸", "sc": 17, "l": 5, "f": 8873},
  {"h": "繧", "p": "yun", "sp": "yun", "r": "糸", "sc": 18, "l": 5, "f": 8874},
  {"h": "繬", "p": "sè", "sp": "se", "r": "糸", "sc": 19, "l": 5, "f": 8875},
  {"h": "纃", "p": "qí", "sp": "qi", "r": "糸", "sc": 20, "l": 5, "f": 8876},
  {"h": "纉", "p": "zuǎn", "sp": "zuan", "r": "糸", "sc": 21, "l": 5, "f": 8877},
  {"h": "纒", "p": "chán", "sp": "chan", "r": "糸", "sc": 22, "l": 5, "f": 8878},
  {"h": "纗", "p": "zuī", "sp": "zui", "r": "糸", "sc": 24, "l": 5, "f": 8879},
  {"h": "罯", "p": "ǎn", "sp": "an", "r": "网", "sc": 14, "l": 5, "f": 8880},
  {"h": "羀", "p": "liǔ", "sp": "liu", "r": "网", "sc": 18, "l": 5, "f": 8881},
  {"h": "耴", "p": "yì", "sp": "yi", "r": "耳", "sc": 7, "l": 5, "f": 8882},
  {"h": "肍", "p": "qiú", "sp": "qiu", "r": "肉", "sc": 6, "l": 5, "f": 8883},
  {
    "h": "腨",
    "p": "shuàn",
    "sp": "shuan",
    "r": "肉",
    "sc": 13,
    "l": 5,
    "f": 8884
  },
  {
    "h": "膗",
    "p": "chuái",
    "sp": "chuai",
    "r": "肉",
    "sc": 15,
    "l": 5,
    "f": 8885
  },
  {"h": "膟", "p": "lǜ", "sp": "lu", "r": "肉", "sc": 15, "l": 5, "f": 8886},
  {"h": "舦", "p": "tài", "sp": "tai", "r": "舟", "sc": 10, "l": 5, "f": 8887},
  {"h": "艊", "p": "bó", "sp": "bo", "r": "舟", "sc": 14, "l": 5, "f": 8888},
  {"h": "菄", "p": "dōng", "sp": "dong", "r": "艸", "sc": 11, "l": 5, "f": 8889},
  {"h": "葾", "p": "yuān", "sp": "yuan", "r": "艸", "sc": 12, "l": 5, "f": 8890},
  {"h": "蒕", "p": "yūn", "sp": "yun", "r": "艸", "sc": 13, "l": 5, "f": 8891},
  {"h": "蒘", "p": "rú", "sp": "ru", "r": "艸", "sc": 13, "l": 5, "f": 8892},
  {"h": "蒧", "p": "diǎn", "sp": "dian", "r": "艸", "sc": 13, "l": 5, "f": 8893},
  {"h": "蓃", "p": "sōu", "sp": "sou", "r": "艸", "sc": 12, "l": 5, "f": 8894},
  {"h": "蓆", "p": "xí", "sp": "xi", "r": "艸", "sc": 13, "l": 5, "f": 8895},
  {"h": "蔱", "p": "shā", "sp": "sha", "r": "艸", "sc": 13, "l": 5, "f": 8896},
  {
    "h": "蘔",
    "p": "jiōng",
    "sp": "jiong",
    "r": "艸",
    "sc": 19,
    "l": 5,
    "f": 8897
  },
  {"h": "蟨", "p": "jué", "sp": "jue", "r": "虫", "sc": 18, "l": 5, "f": 8898},
  {"h": "蠪", "p": "lóng", "sp": "long", "r": "虫", "sc": 22, "l": 5, "f": 8899},
  {"h": "襖", "p": "ǎo", "sp": "ao", "r": "衣", "sc": 17, "l": 5, "f": 8900},
  {"h": "誇", "p": "kuā", "sp": "kua", "r": "言", "sc": 13, "l": 5, "f": 8901},
  {"h": "謟", "p": "tāo", "sp": "tao", "r": "言", "sc": 17, "l": 5, "f": 8902},
  {"h": "讙", "p": "huān", "sp": "huan", "r": "言", "sc": 24, "l": 5, "f": 8903},
  {"h": "谽", "p": "hān", "sp": "han", "r": "谷", "sc": 14, "l": 5, "f": 8904},
  {"h": "踨", "p": "zōng", "sp": "zong", "r": "足", "sc": 15, "l": 5, "f": 8905},
  {"h": "躉", "p": "dǔn", "sp": "dun", "r": "足", "sc": 19, "l": 5, "f": 8906},
  {"h": "躱", "p": "duǒ", "sp": "duo", "r": "身", "sc": 13, "l": 5, "f": 8907},
  {"h": "輗", "p": "ní", "sp": "ni", "r": "車", "sc": 15, "l": 5, "f": 8908},
  {"h": "釔", "p": "yǐ", "sp": "yi", "r": "金", "sc": 9, "l": 5, "f": 8909},
  {"h": "鋕", "p": "zhì", "sp": "zhi", "r": "金", "sc": 15, "l": 5, "f": 8910},
  {"h": "鎚", "p": "chuí", "sp": "chui", "r": "金", "sc": 17, "l": 5, "f": 8911},
  {"h": "韆", "p": "qiān", "sp": "qian", "r": "革", "sc": 24, "l": 5, "f": 8912},
  {"h": "鯶", "p": "hǔn", "sp": "hun", "r": "魚", "sc": 20, "l": 5, "f": 8913},
  {"h": "鳋", "p": "sāo", "sp": "sao", "r": "魚", "sc": 17, "l": 5, "f": 8914},
  {"h": "鹒", "p": "gēng", "sp": "geng", "r": "鳥", "sc": 13, "l": 5, "f": 8915},
  {"h": "篭", "p": "lóng", "sp": "long", "r": "竹", "sc": 16, "l": 5, "f": 8916},
  {"h": "縜", "p": "yún", "sp": "yun", "r": "糸", "sc": 16, "l": 5, "f": 8917},
  {"h": "腣", "p": "dì", "sp": "di", "r": "肉", "sc": 13, "l": 5, "f": 8918},
  {"h": "赲", "p": "lì", "sp": "li", "r": "走", "sc": 9, "l": 5, "f": 8919},
  {"h": "馻", "p": "yǔn", "sp": "yun", "r": "馬", "sc": 14, "l": 5, "f": 8920},
  {"h": "磞", "p": "pēng", "sp": "peng", "r": "石", "sc": 16, "l": 5, "f": 8921},
  {"h": "睕", "p": "wǎn", "sp": "wan", "r": "目", "sc": 13, "l": 5, "f": 8922},
  {"h": "瞜", "p": "lōu", "sp": "lou", "r": "目", "sc": 16, "l": 5, "f": 8923},
  {"h": "碠", "p": "dìng", "sp": "ding", "r": "石", "sc": 14, "l": 5, "f": 8924},
  {"h": "磽", "p": "qiāo", "sp": "qiao", "r": "石", "sc": 17, "l": 5, "f": 8925},
  {"h": "簘", "p": "xiāo", "sp": "xiao", "r": "竹", "sc": 17, "l": 5, "f": 8926},
  {"h": "腃", "p": "kuì", "sp": "kui", "r": "肉", "sc": 12, "l": 5, "f": 8927},
  {"h": "腵", "p": "jiā", "sp": "jia", "r": "肉", "sc": 13, "l": 5, "f": 8928},
  {"h": "蓳", "p": "jǐn", "sp": "jin", "r": "艸", "sc": 14, "l": 5, "f": 8929},
  {"h": "覇", "p": "bà", "sp": "ba", "r": "襾", "sc": 19, "l": 5, "f": 8930},
  {"h": "醱", "p": "fā", "sp": "fa", "r": "酉", "sc": 19, "l": 5, "f": 8931},
  {"h": "銝", "p": "xiū", "sp": "xiu", "r": "金", "sc": 14, "l": 5, "f": 8932},
  {"h": "巛", "p": "chuān", "sp": "chuan", "r": "巛", "sc": 3, "l": 5, "f": 8933},
  {"h": "鐴", "p": "bì", "sp": "bi", "r": "金", "sc": 21, "l": 5, "f": 8934},
  {"h": "鑵", "p": "guàn", "sp": "guan", "r": "金", "sc": 25, "l": 5, "f": 8935},
  {"h": "隩", "p": "ào", "sp": "ao", "r": "阜", "sc": 14, "l": 5, "f": 8936},
  {"h": "颺", "p": "yáng", "sp": "yang", "r": "風", "sc": 18, "l": 5, "f": 8937},
  {"h": "飮", "p": "yǐn", "sp": "yin", "r": "食", "sc": 13, "l": 5, "f": 8938},
  {"h": "飰", "p": "fàn", "sp": "fan", "r": "食", "sc": 12, "l": 5, "f": 8939},
  {"h": "餇", "p": "tóng", "sp": "tong", "r": "食", "sc": 14, "l": 5, "f": 8940},
  {"h": "饐", "p": "yì", "sp": "yi", "r": "食", "sc": 20, "l": 5, "f": 8941},
  {"h": "痃", "p": "xuán", "sp": "xuan", "r": "疒", "sc": 10, "l": 5, "f": 8942},
  {"h": "駞", "p": "tuó", "sp": "tuo", "r": "馬", "sc": 15, "l": 5, "f": 8943},
  {"h": "鴜", "p": "cí", "sp": "ci", "r": "鳥", "sc": 17, "l": 5, "f": 8944},
  {"h": "㳠", "p": "tà", "sp": "ta", "r": "水", "sc": 9, "l": 5, "f": 8945},
  {"h": "䁖", "p": "lōu", "sp": "lou", "r": "目", "sc": 14, "l": 5, "f": 8946},
  {"h": "睅", "p": "hàn", "sp": "han", "r": "目", "sc": 12, "l": 5, "f": 8947},
  {"h": "睍", "p": "xiàn", "sp": "xian", "r": "目", "sc": 12, "l": 5, "f": 8948},
  {"h": "碭", "p": "dàng", "sp": "dang", "r": "石", "sc": 14, "l": 5, "f": 8949},
  {"h": "碽", "p": "gōng", "sp": "gong", "r": "石", "sc": 15, "l": 5, "f": 8950},
  {"h": "磸", "p": "dìng", "sp": "ding", "r": "石", "sc": 17, "l": 5, "f": 8951},
  {"h": "礉", "p": "hé", "sp": "he", "r": "石", "sc": 18, "l": 5, "f": 8952},
  {"h": "礽", "p": "réng", "sp": "reng", "r": "示", "sc": 6, "l": 5, "f": 8953},
  {"h": "祾", "p": "líng", "sp": "ling", "r": "示", "sc": 12, "l": 5, "f": 8954},
  {"h": "禌", "p": "zī", "sp": "zi", "r": "示", "sc": 13, "l": 5, "f": 8955},
  {"h": "筥", "p": "jǔ", "sp": "ju", "r": "竹", "sc": 12, "l": 5, "f": 8956},
  {"h": "簴", "p": "jù", "sp": "ju", "r": "竹", "sc": 19, "l": 5, "f": 8957},
  {"h": "紣", "p": "cuì", "sp": "cui", "r": "糸", "sc": 10, "l": 5, "f": 8958},
  {"h": "絍", "p": "rèn", "sp": "ren", "r": "糸", "sc": 12, "l": 5, "f": 8959},
  {"h": "縆", "p": "gēng", "sp": "geng", "r": "糸", "sc": 15, "l": 5, "f": 8960},
  {"h": "繀", "p": "suì", "sp": "sui", "r": "糸", "sc": 17, "l": 5, "f": 8961},
  {"h": "繃", "p": "běng", "sp": "beng", "r": "糸", "sc": 17, "l": 5, "f": 8962},
  {"h": "羕", "p": "yàng", "sp": "yang", "r": "羊", "sc": 11, "l": 5, "f": 8963},
  {"h": "羵", "p": "fén", "sp": "fen", "r": "羊", "sc": 18, "l": 5, "f": 8964},
  {"h": "耇", "p": "gǒu", "sp": "gou", "r": "老", "sc": 9, "l": 5, "f": 8965},
  {"h": "聫", "p": "lián", "sp": "lian", "r": "耳", "sc": 15, "l": 5, "f": 8966},
  {"h": "脗", "p": "wěn", "sp": "wen", "r": "肉", "sc": 11, "l": 5, "f": 8967},
  {"h": "臯", "p": "gāo", "sp": "gao", "r": "自", "sc": 12, "l": 5, "f": 8968},
  {"h": "艁", "p": "zào", "sp": "zao", "r": "舟", "sc": 13, "l": 5, "f": 8969},
  {"h": "艂", "p": "féng", "sp": "feng", "r": "舟", "sc": 13, "l": 5, "f": 8970},
  {"h": "茞", "p": "chén", "sp": "chen", "r": "艸", "sc": 9, "l": 5, "f": 8971},
  {"h": "葨", "p": "wēi", "sp": "wei", "r": "艸", "sc": 12, "l": 5, "f": 8972},
  {"h": "蒄", "p": "guān", "sp": "guan", "r": "艸", "sc": 12, "l": 5, "f": 8973},
  {"h": "蒭", "p": "chú", "sp": "chu", "r": "艸", "sc": 13, "l": 5, "f": 8974},
  {"h": "薂", "p": "xí", "sp": "xi", "r": "艸", "sc": 16, "l": 5, "f": 8975},
  {"h": "藯", "p": "wèi", "sp": "wei", "r": "艸", "sc": 18, "l": 5, "f": 8976},
  {"h": "蜆", "p": "xiàn", "sp": "xian", "r": "虫", "sc": 13, "l": 5, "f": 8977},
  {"h": "螑", "p": "xiù", "sp": "xiu", "r": "虫", "sc": 16, "l": 5, "f": 8978},
  {"h": "蟼", "p": "jǐng", "sp": "jing", "r": "虫", "sc": 18, "l": 5, "f": 8979},
  {"h": "衟", "p": "dào", "sp": "dao", "r": "行", "sc": 16, "l": 5, "f": 8980},
  {"h": "褑", "p": "yuàn", "sp": "yuan", "r": "衣", "sc": 14, "l": 5, "f": 8981},
  {"h": "褠", "p": "gōu", "sp": "gou", "r": "衣", "sc": 15, "l": 5, "f": 8982},
  {"h": "訐", "p": "jié", "sp": "jie", "r": "言", "sc": 10, "l": 5, "f": 8983},
  {"h": "訔", "p": "yín", "sp": "yin", "r": "言", "sc": 10, "l": 5, "f": 8984},
  {"h": "詖", "p": "bì", "sp": "bi", "r": "言", "sc": 12, "l": 5, "f": 8985},
  {
    "h": "詗",
    "p": "xiòng",
    "sp": "xiong",
    "r": "言",
    "sc": 12,
    "l": 5,
    "f": 8986
  },
  {"h": "詪", "p": "hěn", "sp": "hen", "r": "言", "sc": 13, "l": 5, "f": 8987},
  {"h": "諈", "p": "zhuì", "sp": "zhui", "r": "言", "sc": 15, "l": 5, "f": 8988},
  {"h": "諑", "p": "zhuó", "sp": "zhuo", "r": "言", "sc": 15, "l": 5, "f": 8989},
  {"h": "讌", "p": "yàn", "sp": "yan", "r": "言", "sc": 23, "l": 5, "f": 8990},
  {"h": "豗", "p": "huī", "sp": "hui", "r": "豕", "sc": 10, "l": 5, "f": 8991},
  {"h": "豷", "p": "yì", "sp": "yi", "r": "豕", "sc": 19, "l": 5, "f": 8992},
  {"h": "貗", "p": "jù", "sp": "ju", "r": "豸", "sc": 18, "l": 5, "f": 8993},
  {"h": "貣", "p": "tè", "sp": "te", "r": "貝", "sc": 10, "l": 5, "f": 8994},
  {
    "h": "賯",
    "p": "xiōng",
    "sp": "xiong",
    "r": "貝",
    "sc": 16,
    "l": 5,
    "f": 8995
  },
  {"h": "贋", "p": "yàn", "sp": "yan", "r": "貝", "sc": 19, "l": 5, "f": 8996},
  {"h": "冱", "p": "hù", "sp": "hu", "r": "冫", "sc": 6, "l": 5, "f": 8997},
  {"h": "輀", "p": "ér", "sp": "er", "r": "車", "sc": 13, "l": 5, "f": 8998},
  {"h": "蒈", "p": "kǎi", "sp": "kai", "r": "艸", "sc": 12, "l": 5, "f": 8999},
  {"h": "轓", "p": "fān", "sp": "fan", "r": "車", "sc": 19, "l": 5, "f": 9000},
  {"h": "迋", "p": "wàng", "sp": "wang", "r": "辵", "sc": 7, "l": 5, "f": 9001},
  {"h": "逓", "p": "dì", "sp": "di", "r": "辵", "sc": 10, "l": 5, "f": 9002},
  {
    "h": "鄺",
    "p": "kuàng",
    "sp": "kuang",
    "r": "邑",
    "sc": 16,
    "l": 5,
    "f": 9003
  },
  {"h": "酳", "p": "yìn", "sp": "yin", "r": "酉", "sc": 14, "l": 5, "f": 9004},
  {"h": "鈪", "p": "è", "sp": "e", "r": "金", "sc": 12, "l": 5, "f": 9005},
  {"h": "尜", "p": "gá", "sp": "ga", "r": "小", "sc": 9, "l": 5, "f": 9006},
  {"h": "鐐", "p": "liáo", "sp": "liao", "r": "金", "sc": 20, "l": 5, "f": 9007},
  {"h": "轷", "p": "hū", "sp": "hu", "r": "車", "sc": 9, "l": 5, "f": 9008},
  {"h": "霘", "p": "dòng", "sp": "dong", "r": "雨", "sc": 17, "l": 5, "f": 9009},
  {"h": "鞄", "p": "páo", "sp": "pao", "r": "革", "sc": 14, "l": 5, "f": 9010},
  {"h": "騧", "p": "guā", "sp": "gua", "r": "馬", "sc": 18, "l": 5, "f": 9011},
  {"h": "髽", "p": "zhuā", "sp": "zhua", "r": "髟", "sc": 17, "l": 5, "f": 9012},
  {"h": "魣", "p": "xù", "sp": "xu", "r": "魚", "sc": 15, "l": 5, "f": 9013},
  {"h": "鯀", "p": "gǔn", "sp": "gun", "r": "魚", "sc": 18, "l": 5, "f": 9014},
  {"h": "鰰", "p": "shen", "sp": "shen", "r": "魚", "sc": 20, "l": 5, "f": 9015},
  {"h": "鴹", "p": "yáng", "sp": "yang", "r": "鳥", "sc": 17, "l": 5, "f": 9016},
  {
    "h": "鼪",
    "p": "shēng",
    "sp": "sheng",
    "r": "鼠",
    "sc": 18,
    "l": 5,
    "f": 9017
  },
  {"h": "龔", "p": "gōng", "sp": "gong", "r": "龍", "sc": 22, "l": 5, "f": 9018},
  {"h": "秊", "p": "", "sp": "", "r": "禾", "sc": 8, "l": 5, "f": 9019},
  {"h": "䴔", "p": "jiāo", "sp": "jiao", "r": "", "sc": 17, "l": 5, "f": 9020},
  {"h": "眮", "p": "tóng", "sp": "tong", "r": "目", "sc": 11, "l": 5, "f": 9021},
  {"h": "笀", "p": "máng", "sp": "mang", "r": "竹", "sc": 9, "l": 5, "f": 9022},
  {"h": "籣", "p": "lán", "sp": "lan", "r": "竹", "sc": 23, "l": 5, "f": 9023},
  {
    "h": "藑",
    "p": "qióng",
    "sp": "qiong",
    "r": "艸",
    "sc": 17,
    "l": 5,
    "f": 9024
  },
  {"h": "垲", "p": "kǎi", "sp": "kai", "r": "土", "sc": 9, "l": 5, "f": 9025},
  {"h": "鍳", "p": "jiàn", "sp": "jian", "r": "金", "sc": 17, "l": 5, "f": 9026},
  {"h": "鵬", "p": "péng", "sp": "peng", "r": "鳥", "sc": 19, "l": 5, "f": 9027},
  {"h": "禔", "p": "zhī", "sp": "zhi", "r": "示", "sc": 13, "l": 5, "f": 9028},
  {"h": "穯", "p": "se", "sp": "se", "r": "禾", "sc": 20, "l": 5, "f": 9029},
  {"h": "茿", "p": "zhú", "sp": "zhu", "r": "艸", "sc": 9, "l": 5, "f": 9030},
  {"h": "莂", "p": "bié", "sp": "bie", "r": "艸", "sc": 10, "l": 5, "f": 9031},
  {"h": "薃", "p": "hào", "sp": "hao", "r": "艸", "sc": 16, "l": 5, "f": 9032},
  {"h": "襛", "p": "nóng", "sp": "nong", "r": "衣", "sc": 18, "l": 5, "f": 9033},
  {"h": "谻", "p": "jí", "sp": "ji", "r": "谷", "sc": 11, "l": 5, "f": 9034},
  {"h": "賚", "p": "lài", "sp": "lai", "r": "貝", "sc": 15, "l": 5, "f": 9035},
  {"h": "魳", "p": "zā", "sp": "za", "r": "魚", "sc": 15, "l": 5, "f": 9036},
  {"h": "鵄", "p": "zhī", "sp": "zhi", "r": "鳥", "sc": 17, "l": 5, "f": 9037},
  {"h": "㭎", "p": "gāng", "sp": "gang", "r": "木", "sc": 8, "l": 5, "f": 9038},
  {"h": "穄", "p": "jì", "sp": "ji", "r": "禾", "sc": 16, "l": 5, "f": 9039},
  {"h": "穵", "p": "wā", "sp": "wa", "r": "穴", "sc": 6, "l": 5, "f": 9040},
  {"h": "篛", "p": "ruò", "sp": "ruo", "r": "竹", "sc": 16, "l": 5, "f": 9041},
  {"h": "簑", "p": "suō", "sp": "suo", "r": "竹", "sc": 16, "l": 5, "f": 9042},
  {"h": "絭", "p": "juàn", "sp": "juan", "r": "糸", "sc": 12, "l": 5, "f": 9043},
  {"h": "緅", "p": "zōu", "sp": "zou", "r": "糸", "sc": 14, "l": 5, "f": 9044},
  {"h": "緍", "p": "mín", "sp": "min", "r": "糸", "sc": 14, "l": 5, "f": 9045},
  {"h": "罇", "p": "zūn", "sp": "zun", "r": "缶", "sc": 18, "l": 5, "f": 9046},
  {"h": "耑", "p": "duān", "sp": "duan", "r": "而", "sc": 9, "l": 5, "f": 9047},
  {"h": "艒", "p": "mù", "sp": "mu", "r": "舟", "sc": 15, "l": 5, "f": 9048},
  {"h": "豠", "p": "chú", "sp": "chu", "r": "豕", "sc": 12, "l": 5, "f": 9049},
  {"h": "蹎", "p": "diān", "sp": "dian", "r": "足", "sc": 17, "l": 5, "f": 9050},
  {"h": "辢", "p": "là", "sp": "la", "r": "辛", "sc": 14, "l": 5, "f": 9051},
  {"h": "醻", "p": "chóu", "sp": "chou", "r": "酉", "sc": 21, "l": 5, "f": 9052},
  {"h": "鐍", "p": "jué", "sp": "jue", "r": "金", "sc": 20, "l": 5, "f": 9053},
  {"h": "鑮", "p": "bó", "sp": "bo", "r": "金", "sc": 24, "l": 5, "f": 9054},
  {"h": "闍", "p": "dū", "sp": "du", "r": "門", "sc": 16, "l": 5, "f": 9055},
  {
    "h": "頳",
    "p": "chēng",
    "sp": "cheng",
    "r": "頁",
    "sc": 16,
    "l": 5,
    "f": 9056
  },
  {"h": "驈", "p": "yù", "sp": "yu", "r": "馬", "sc": 22, "l": 5, "f": 9057},
  {"h": "髲", "p": "bì", "sp": "bi", "r": "髟", "sc": 15, "l": 5, "f": 9058},
  {"h": "鱮", "p": "xù", "sp": "xu", "r": "魚", "sc": 24, "l": 5, "f": 9059},
  {"h": "麰", "p": "móu", "sp": "mou", "r": "麥", "sc": 17, "l": 5, "f": 9060},
  {"h": "鼄", "p": "zhū", "sp": "zhu", "r": "黽", "sc": 19, "l": 5, "f": 9061},
  {"h": "鼅", "p": "zhī", "sp": "zhi", "r": "黽", "sc": 21, "l": 5, "f": 9062},
  {"h": "鼖", "p": "fén", "sp": "fen", "r": "鼓", "sc": 18, "l": 5, "f": 9063},
  {"h": "㩳", "p": "sǒng", "sp": "song", "r": "手", "sc": 21, "l": 5, "f": 9064},
  {"h": "眣", "p": "dié", "sp": "die", "r": "目", "sc": 10, "l": 5, "f": 9065},
  {"h": "睎", "p": "xī", "sp": "xi", "r": "目", "sc": 12, "l": 5, "f": 9066},
  {"h": "瞣", "p": "huan", "sp": "huan", "r": "目", "sc": 16, "l": 5, "f": 9067},
  {"h": "矀", "p": "méi", "sp": "mei", "r": "目", "sc": 18, "l": 5, "f": 9068},
  {"h": "硊", "p": "wěi", "sp": "wei", "r": "石", "sc": 11, "l": 5, "f": 9069},
  {"h": "碅", "p": "jūn", "sp": "jun", "r": "石", "sc": 13, "l": 5, "f": 9070},
  {"h": "磪", "p": "cuī", "sp": "cui", "r": "石", "sc": 16, "l": 5, "f": 9071},
  {"h": "礒", "p": "yǐ", "sp": "yi", "r": "石", "sc": 18, "l": 5, "f": 9072},
  {"h": "禞", "p": "gào", "sp": "gao", "r": "示", "sc": 14, "l": 5, "f": 9073},
  {"h": "秠", "p": "pī", "sp": "pi", "r": "禾", "sc": 10, "l": 5, "f": 9074},
  {"h": "穙", "p": "pú", "sp": "pu", "r": "禾", "sc": 17, "l": 5, "f": 9075},
  {"h": "穱", "p": "zhuō", "sp": "zhuo", "r": "禾", "sc": 22, "l": 5, "f": 9076},
  {"h": "窯", "p": "yáo", "sp": "yao", "r": "穴", "sc": 15, "l": 5, "f": 9077},
  {"h": "竢", "p": "sì", "sp": "si", "r": "立", "sc": 12, "l": 5, "f": 9078},
  {"h": "笍", "p": "zhuì", "sp": "zhui", "r": "竹", "sc": 10, "l": 5, "f": 9079},
  {"h": "筩", "p": "tóng", "sp": "tong", "r": "竹", "sc": 13, "l": 5, "f": 9080},
  {"h": "篳", "p": "bì", "sp": "bi", "r": "竹", "sc": 16, "l": 5, "f": 9081},
  {"h": "簠", "p": "fǔ", "sp": "fu", "r": "竹", "sc": 18, "l": 5, "f": 9082},
  {"h": "粃", "p": "bǐ", "sp": "bi", "r": "米", "sc": 10, "l": 5, "f": 9083},
  {"h": "粖", "p": "mò", "sp": "mo", "r": "米", "sc": 11, "l": 5, "f": 9084},
  {"h": "糓", "p": "gǔ", "sp": "gu", "r": "米", "sc": 16, "l": 5, "f": 9085},
  {"h": "糺", "p": "jiū", "sp": "jiu", "r": "糸", "sc": 7, "l": 5, "f": 9086},
  {"h": "紭", "p": "hóng", "sp": "hong", "r": "糸", "sc": 11, "l": 5, "f": 9087},
  {"h": "紾", "p": "zhěn", "sp": "zhen", "r": "糸", "sc": 11, "l": 5, "f": 9088},
  {"h": "絶", "p": "jué", "sp": "jue", "r": "糸", "sc": 12, "l": 5, "f": 9089},
  {"h": "綅", "p": "qīn", "sp": "qin", "r": "糸", "sc": 13, "l": 5, "f": 9090},
  {"h": "繂", "p": "lǜ", "sp": "lu", "r": "糸", "sc": 17, "l": 5, "f": 9091},
  {"h": "繋", "p": "jì", "sp": "ji", "r": "糸", "sc": 17, "l": 5, "f": 9092},
  {"h": "繣", "p": "huà", "sp": "hua", "r": "糸", "sc": 18, "l": 5, "f": 9093},
  {"h": "罙", "p": "mí", "sp": "mi", "r": "网", "sc": 8, "l": 5, "f": 9094},
  {"h": "罥", "p": "juàn", "sp": "juan", "r": "网", "sc": 12, "l": 5, "f": 9095},
  {"h": "罼", "p": "bì", "sp": "bi", "r": "网", "sc": 15, "l": 5, "f": 9096},
  {"h": "羗", "p": "qiāng", "sp": "qiang", "r": "羊", "sc": 9, "l": 5, "f": 9097},
  {"h": "羜", "p": "zhù", "sp": "zhu", "r": "羊", "sc": 11, "l": 5, "f": 9098},
  {"h": "翭", "p": "hóu", "sp": "hou", "r": "羽", "sc": 15, "l": 5, "f": 9099},
  {"h": "耮", "p": "lào", "sp": "lao", "r": "耒", "sc": 18, "l": 5, "f": 9100},
  {"h": "聸", "p": "dān", "sp": "dan", "r": "耳", "sc": 19, "l": 5, "f": 9101},
  {"h": "肸", "p": "xī", "sp": "xi", "r": "肉", "sc": 8, "l": 5, "f": 9102},
  {"h": "胹", "p": "ér", "sp": "er", "r": "肉", "sc": 10, "l": 5, "f": 9103},
  {"h": "脣", "p": "chún", "sp": "chun", "r": "肉", "sc": 11, "l": 5, "f": 9104},
  {"h": "腬", "p": "ròu", "sp": "rou", "r": "肉", "sc": 13, "l": 5, "f": 9105},
  {"h": "膄", "p": "sòu", "sp": "sou", "r": "肉", "sc": 13, "l": 5, "f": 9106},
  {"h": "膆", "p": "sù", "sp": "su", "r": "肉", "sc": 14, "l": 5, "f": 9107},
  {"h": "膋", "p": "liáo", "sp": "liao", "r": "肉", "sc": 14, "l": 5, "f": 9108},
  {"h": "臕", "p": "biāo", "sp": "biao", "r": "肉", "sc": 19, "l": 5, "f": 9109},
  {"h": "舃", "p": "xì", "sp": "xi", "r": "臼", "sc": 12, "l": 5, "f": 9110},
  {"h": "艫", "p": "lú", "sp": "lu", "r": "舟", "sc": 22, "l": 5, "f": 9111},
  {
    "h": "艭",
    "p": "shuāng",
    "sp": "shuang",
    "r": "舟",
    "sc": 24,
    "l": 5,
    "f": 9112
  },
  {"h": "艶", "p": "yàn", "sp": "yan", "r": "色", "sc": 19, "l": 5, "f": 9113},
  {"h": "艼", "p": "tīng", "sp": "ting", "r": "艸", "sc": 5, "l": 5, "f": 9114},
  {"h": "荢", "p": "yu", "sp": "yu", "r": "艸", "sc": 9, "l": 5, "f": 9115},
  {"h": "荳", "p": "dòu", "sp": "dou", "r": "艸", "sc": 10, "l": 5, "f": 9116},
  {"h": "萚", "p": "tuò", "sp": "tuo", "r": "艸", "sc": 11, "l": 5, "f": 9117},
  {"h": "蓗", "p": "zǒng", "sp": "zong", "r": "艸", "sc": 13, "l": 5, "f": 9118},
  {"h": "蓱", "p": "píng", "sp": "ping", "r": "艸", "sc": 12, "l": 5, "f": 9119},
  {"h": "蕋", "p": "ruǐ", "sp": "rui", "r": "艸", "sc": 15, "l": 5, "f": 9120},
  {"h": "蕟", "p": "fà", "sp": "fa", "r": "艸", "sc": 15, "l": 5, "f": 9121},
  {"h": "薐", "p": "léng", "sp": "leng", "r": "艸", "sc": 16, "l": 5, "f": 9122},
  {"h": "薠", "p": "fán", "sp": "fan", "r": "艸", "sc": 16, "l": 5, "f": 9123},
  {"h": "藺", "p": "lìn", "sp": "lin", "r": "艸", "sc": 19, "l": 5, "f": 9124},
  {"h": "蘓", "p": "sū", "sp": "su", "r": "艸", "sc": 19, "l": 5, "f": 9125},
  {"h": "蘪", "p": "méi", "sp": "mei", "r": "艸", "sc": 20, "l": 5, "f": 9126},
  {"h": "虵", "p": "shé", "sp": "she", "r": "虫", "sc": 9, "l": 5, "f": 9127},
  {"h": "蚾", "p": "pí", "sp": "pi", "r": "虫", "sc": 11, "l": 5, "f": 9128},
  {"h": "蛖", "p": "máng", "sp": "mang", "r": "虫", "sc": 13, "l": 5, "f": 9129},
  {"h": "蛜", "p": "yī", "sp": "yi", "r": "虫", "sc": 12, "l": 5, "f": 9130},
  {"h": "蝍", "p": "jié", "sp": "jie", "r": "虫", "sc": 13, "l": 5, "f": 9131},
  {"h": "蝨", "p": "shī", "sp": "shi", "r": "虫", "sc": 15, "l": 5, "f": 9132},
  {"h": "蟂", "p": "xiāo", "sp": "xiao", "r": "虫", "sc": 17, "l": 5, "f": 9133},
  {"h": "蟜", "p": "jiǎo", "sp": "jiao", "r": "虫", "sc": 18, "l": 5, "f": 9134},
  {"h": "蠏", "p": "xiè", "sp": "xie", "r": "虫", "sc": 19, "l": 5, "f": 9135},
  {"h": "衕", "p": "tòng", "sp": "tong", "r": "行", "sc": 12, "l": 5, "f": 9136},
  {"h": "衘", "p": "xián", "sp": "xian", "r": "行", "sc": 13, "l": 5, "f": 9137},
  {"h": "袟", "p": "zhì", "sp": "zhi", "r": "衣", "sc": 10, "l": 5, "f": 9138},
  {"h": "袬", "p": "gǔn", "sp": "gun", "r": "衣", "sc": 11, "l": 5, "f": 9139},
  {"h": "誾", "p": "yín", "sp": "yin", "r": "言", "sc": 15, "l": 5, "f": 9140},
  {"h": "譅", "p": "sè", "sp": "se", "r": "言", "sc": 21, "l": 5, "f": 9141},
  {"h": "譌", "p": "é", "sp": "e", "r": "言", "sc": 19, "l": 5, "f": 9142},
  {"h": "譸", "p": "zhōu", "sp": "zhou", "r": "言", "sc": 21, "l": 5, "f": 9143},
  {"h": "讜", "p": "dǎng", "sp": "dang", "r": "言", "sc": 27, "l": 5, "f": 9144},
  {"h": "趝", "p": "jiàn", "sp": "jian", "r": "走", "sc": 15, "l": 5, "f": 9145},
  {"h": "趡", "p": "cuǐ", "sp": "cui", "r": "走", "sc": 15, "l": 5, "f": 9146},
  {"h": "跾", "p": "shū", "sp": "shu", "r": "足", "sc": 14, "l": 5, "f": 9147},
  {"h": "踀", "p": "chù", "sp": "chu", "r": "足", "sc": 14, "l": 5, "f": 9148},
  {"h": "踸", "p": "chěn", "sp": "chen", "r": "足", "sc": 16, "l": 5, "f": 9149},
  {
    "h": "蹡",
    "p": "qiāng",
    "sp": "qiang",
    "r": "足",
    "sc": 18,
    "l": 5,
    "f": 9150
  },
  {"h": "垌", "p": "dòng", "sp": "dong", "r": "土", "sc": 9, "l": 5, "f": 9151},
  {"h": "蹳", "p": "bō", "sp": "bo", "r": "足", "sc": 19, "l": 5, "f": 9152},
  {"h": "躛", "p": "wèi", "sp": "wei", "r": "足", "sc": 23, "l": 5, "f": 9153},
  {"h": "躩", "p": "jué", "sp": "jue", "r": "足", "sc": 27, "l": 5, "f": 9154},
  {"h": "躶", "p": "luǒ", "sp": "luo", "r": "身", "sc": 15, "l": 5, "f": 9155},
  {"h": "軏", "p": "yuè", "sp": "yue", "r": "車", "sc": 10, "l": 5, "f": 9156},
  {"h": "軬", "p": "fàn", "sp": "fan", "r": "車", "sc": 12, "l": 5, "f": 9157},
  {"h": "輵", "p": "gé", "sp": "ge", "r": "車", "sc": 16, "l": 5, "f": 9158},
  {"h": "轗", "p": "kǎn", "sp": "kan", "r": "車", "sc": 20, "l": 5, "f": 9159},
  {"h": "轥", "p": "lìn", "sp": "lin", "r": "車", "sc": 26, "l": 5, "f": 9160},
  {"h": "逷", "p": "tì", "sp": "ti", "r": "辵", "sc": 11, "l": 5, "f": 9161},
  {"h": "遈", "p": "shí", "sp": "shi", "r": "辵", "sc": 12, "l": 5, "f": 9162},
  {"h": "邌", "p": "lí", "sp": "li", "r": "辵", "sc": 18, "l": 5, "f": 9163},
  {
    "h": "鄟",
    "p": "zhuān",
    "sp": "zhuan",
    "r": "邑",
    "sc": 13,
    "l": 5,
    "f": 9164
  },
  {"h": "醎", "p": "xián", "sp": "xian", "r": "酉", "sc": 16, "l": 5, "f": 9165},
  {"h": "釭", "p": "gāng", "sp": "gang", "r": "金", "sc": 11, "l": 5, "f": 9166},
  {"h": "鈸", "p": "bó", "sp": "bo", "r": "金", "sc": 13, "l": 5, "f": 9167},
  {"h": "怊", "p": "chāo", "sp": "chao", "r": "心", "sc": 8, "l": 5, "f": 9168},
  {"h": "銌", "p": "zùn", "sp": "zun", "r": "金", "sc": 14, "l": 5, "f": 9169},
  {"h": "鋂", "p": "méi", "sp": "mei", "r": "金", "sc": 15, "l": 5, "f": 9170},
  {"h": "鋎", "p": "hàn", "sp": "han", "r": "金", "sc": 15, "l": 5, "f": 9171},
  {"h": "鍐", "p": "zōng", "sp": "zong", "r": "金", "sc": 17, "l": 5, "f": 9172},
  {"h": "鎴", "p": "xí", "sp": "xi", "r": "金", "sc": 18, "l": 5, "f": 9173},
  {"h": "鎸", "p": "juān", "sp": "juan", "r": "金", "sc": 18, "l": 5, "f": 9174},
  {"h": "鑘", "p": "léi", "sp": "lei", "r": "金", "sc": 23, "l": 5, "f": 9175},
  {"h": "閔", "p": "mǐn", "sp": "min", "r": "門", "sc": 12, "l": 5, "f": 9176},
  {"h": "雋", "p": "juàn", "sp": "juan", "r": "隹", "sc": 12, "l": 5, "f": 9177},
  {
    "h": "膪",
    "p": "chuài",
    "sp": "chuai",
    "r": "肉",
    "sc": 16,
    "l": 5,
    "f": 9178
  },
  {"h": "霣", "p": "yǔn", "sp": "yun", "r": "雨", "sc": 18, "l": 5, "f": 9179},
  {"h": "靃", "p": "huò", "sp": "huo", "r": "雨", "sc": 24, "l": 5, "f": 9180},
  {"h": "韍", "p": "fú", "sp": "fu", "r": "韋", "sc": 14, "l": 5, "f": 9181},
  {"h": "颩", "p": "biāo", "sp": "biao", "r": "風", "sc": 12, "l": 5, "f": 9182},
  {"h": "飦", "p": "zhān", "sp": "zhan", "r": "食", "sc": 11, "l": 5, "f": 9183},
  {"h": "餔", "p": "bù", "sp": "bu", "r": "食", "sc": 15, "l": 5, "f": 9184},
  {"h": "駋", "p": "zhāo", "sp": "zhao", "r": "馬", "sc": 15, "l": 5, "f": 9185},
  {"h": "騄", "p": "lù", "sp": "lu", "r": "馬", "sc": 18, "l": 5, "f": 9186},
  {"h": "骩", "p": "wěi", "sp": "wei", "r": "骨", "sc": 11, "l": 5, "f": 9187},
  {"h": "骻", "p": "kuà", "sp": "kua", "r": "骨", "sc": 15, "l": 5, "f": 9188},
  {"h": "骾", "p": "gěng", "sp": "geng", "r": "骨", "sc": 16, "l": 5, "f": 9189},
  {"h": "鬙", "p": "sēng", "sp": "seng", "r": "髟", "sc": 22, "l": 5, "f": 9190},
  {"h": "鬽", "p": "mèi", "sp": "mei", "r": "鬼", "sc": 12, "l": 5, "f": 9191},
  {"h": "魫", "p": "shěn", "sp": "shen", "r": "魚", "sc": 15, "l": 5, "f": 9192},
  {"h": "鯈", "p": "tiáo", "sp": "tiao", "r": "魚", "sc": 17, "l": 5, "f": 9193},
  {"h": "鯩", "p": "lún", "sp": "lun", "r": "魚", "sc": 19, "l": 5, "f": 9194},
  {"h": "鯬", "p": "lí", "sp": "li", "r": "魚", "sc": 19, "l": 5, "f": 9195},
  {"h": "鰌", "p": "qiū", "sp": "qiu", "r": "魚", "sc": 20, "l": 5, "f": 9196},
  {"h": "鲓", "p": "kǎo", "sp": "kao", "r": "魚", "sc": 14, "l": 5, "f": 9197},
  {"h": "鴂", "p": "jué", "sp": "jue", "r": "鳥", "sc": 15, "l": 5, "f": 9198},
  {"h": "鶍", "p": "yi", "sp": "yi", "r": "鳥", "sc": 19, "l": 5, "f": 9199},
  {"h": "鶠", "p": "yǎn", "sp": "yan", "r": "鳥", "sc": 20, "l": 5, "f": 9200},
  {"h": "鷁", "p": "yì", "sp": "yi", "r": "鳥", "sc": 21, "l": 5, "f": 9201},
  {"h": "鷖", "p": "yī", "sp": "yi", "r": "鳥", "sc": 22, "l": 5, "f": 9202},
  {"h": "鸇", "p": "zhān", "sp": "zhan", "r": "鳥", "sc": 24, "l": 5, "f": 9203},
  {
    "h": "鸘",
    "p": "shuāng",
    "sp": "shuang",
    "r": "鳥",
    "sc": 28,
    "l": 5,
    "f": 9204
  },
  {"h": "鸧", "p": "cāng", "sp": "cang", "r": "鳥", "sc": 9, "l": 5, "f": 9205},
  {"h": "鹙", "p": "qiū", "sp": "qiu", "r": "鳥", "sc": 14, "l": 5, "f": 9206},
  {"h": "麏", "p": "jūn", "sp": "jun", "r": "鹿", "sc": 18, "l": 5, "f": 9207},
  {"h": "黮", "p": "dǎn", "sp": "dan", "r": "黑", "sc": 21, "l": 5, "f": 9208},
  {"h": "齯", "p": "ní", "sp": "ni", "r": "齒", "sc": 23, "l": 5, "f": 9209},
  {"h": "龓", "p": "lǒng", "sp": "long", "r": "龍", "sc": 22, "l": 5, "f": 9210},
  {"h": "盓", "p": "yū", "sp": "yu", "r": "皿", "sc": 11, "l": 5, "f": 9211},
  {"h": "眕", "p": "zhěn", "sp": "zhen", "r": "目", "sc": 10, "l": 5, "f": 9212},
  {
    "h": "睘",
    "p": "qióng",
    "sp": "qiong",
    "r": "目",
    "sc": 13,
    "l": 5,
    "f": 9213
  },
  {"h": "睩", "p": "lù", "sp": "lu", "r": "目", "sc": 13, "l": 5, "f": 9214},
  {"h": "矂", "p": "sào", "sp": "sao", "r": "目", "sc": 18, "l": 5, "f": 9215},
  {"h": "矵", "p": "qì", "sp": "qi", "r": "石", "sc": 7, "l": 5, "f": 9216},
  {"h": "矺", "p": "zhé", "sp": "zhe", "r": "石", "sc": 8, "l": 5, "f": 9217},
  {"h": "砪", "p": "mǔ", "sp": "mu", "r": "石", "sc": 10, "l": 5, "f": 9218},
  {"h": "硂", "p": "quán", "sp": "quan", "r": "石", "sc": 11, "l": 5, "f": 9219},
  {"h": "碒", "p": "yín", "sp": "yin", "r": "石", "sc": 13, "l": 5, "f": 9220},
  {"h": "磣", "p": "chěn", "sp": "chen", "r": "石", "sc": 16, "l": 5, "f": 9221},
  {"h": "祤", "p": "yǔ", "sp": "yu", "r": "示", "sc": 10, "l": 5, "f": 9222},
  {"h": "禩", "p": "sì", "sp": "si", "r": "示", "sc": 15, "l": 5, "f": 9223},
  {"h": "稑", "p": "lù", "sp": "lu", "r": "禾", "sc": 13, "l": 5, "f": 9224},
  {"h": "穐", "p": "qiū", "sp": "qiu", "r": "禾", "sc": 16, "l": 5, "f": 9225},
  {"h": "穛", "p": "zhuō", "sp": "zhuo", "r": "禾", "sc": 17, "l": 5, "f": 9226},
  {"h": "竂", "p": "piáo", "sp": "piao", "r": "穴", "sc": 17, "l": 5, "f": 9227},
  {"h": "笉", "p": "qǐn", "sp": "qin", "r": "竹", "sc": 10, "l": 5, "f": 9228},
  {"h": "篟", "p": "qiàn", "sp": "qian", "r": "竹", "sc": 16, "l": 5, "f": 9229},
  {"h": "篺", "p": "pí", "sp": "pi", "r": "竹", "sc": 17, "l": 5, "f": 9230},
  {"h": "紂", "p": "zhòu", "sp": "zhou", "r": "糸", "sc": 9, "l": 5, "f": 9231},
  {"h": "紻", "p": "yǎng", "sp": "yang", "r": "糸", "sc": 11, "l": 5, "f": 9232},
  {"h": "綹", "p": "liǔ", "sp": "liu", "r": "糸", "sc": 14, "l": 5, "f": 9233},
  {"h": "緐", "p": "fán", "sp": "fan", "r": "糸", "sc": 14, "l": 5, "f": 9234},
  {"h": "縈", "p": "yíng", "sp": "ying", "r": "糸", "sc": 16, "l": 5, "f": 9235},
  {"h": "繖", "p": "sǎn", "sp": "san", "r": "糸", "sc": 18, "l": 5, "f": 9236},
  {"h": "繯", "p": "huán", "sp": "huan", "r": "糸", "sc": 19, "l": 5, "f": 9237},
  {"h": "繴", "p": "bì", "sp": "bi", "r": "糸", "sc": 19, "l": 5, "f": 9238},
  {"h": "纀", "p": "pú", "sp": "pu", "r": "糸", "sc": 20, "l": 5, "f": 9239},
  {"h": "羉", "p": "luán", "sp": "luan", "r": "网", "sc": 24, "l": 5, "f": 9240},
  {"h": "翈", "p": "xiá", "sp": "xia", "r": "羽", "sc": 11, "l": 5, "f": 9241},
  {"h": "耡", "p": "chú", "sp": "chu", "r": "耒", "sc": 13, "l": 5, "f": 9242},
  {"h": "腀", "p": "lún", "sp": "lun", "r": "肉", "sc": 12, "l": 5, "f": 9243},
  {"h": "艩", "p": "qí", "sp": "qi", "r": "舟", "sc": 20, "l": 5, "f": 9244},
  {"h": "苅", "p": "yì", "sp": "yi", "r": "艸", "sc": 7, "l": 5, "f": 9245},
  {"h": "莁", "p": "wú", "sp": "wu", "r": "艸", "sc": 10, "l": 5, "f": 9246},
  {"h": "菞", "p": "lí", "sp": "li", "r": "艸", "sc": 11, "l": 5, "f": 9247},
  {"h": "菳", "p": "qín", "sp": "qin", "r": "艸", "sc": 11, "l": 5, "f": 9248},
  {"h": "萈", "p": "huán", "sp": "huan", "r": "艸", "sc": 11, "l": 5, "f": 9249},
  {"h": "蒁", "p": "shù", "sp": "shu", "r": "艸", "sc": 11, "l": 5, "f": 9250},
  {"h": "蒑", "p": "yìn", "sp": "yin", "r": "艸", "sc": 13, "l": 5, "f": 9251},
  {"h": "蓔", "p": "yǎo", "sp": "yao", "r": "艸", "sc": 13, "l": 5, "f": 9252},
  {"h": "蓕", "p": "guì", "sp": "gui", "r": "艸", "sc": 13, "l": 5, "f": 9253},
  {"h": "蔙", "p": "xuàn", "sp": "xuan", "r": "艸", "sc": 14, "l": 5, "f": 9254},
  {"h": "蔿", "p": "wěi", "sp": "wei", "r": "艸", "sc": 15, "l": 5, "f": 9255},
  {"h": "蕧", "p": "fù", "sp": "fu", "r": "艸", "sc": 15, "l": 5, "f": 9256},
  {"h": "薆", "p": "ài", "sp": "ai", "r": "艸", "sc": 16, "l": 5, "f": 9257},
  {"h": "薗", "p": "yuán", "sp": "yuan", "r": "艸", "sc": 16, "l": 5, "f": 9258},
  {"h": "薝", "p": "zhān", "sp": "zhan", "r": "艸", "sc": 16, "l": 5, "f": 9259},
  {"h": "藀", "p": "yíng", "sp": "ying", "r": "艸", "sc": 17, "l": 5, "f": 9260},
  {"h": "蘌", "p": "yǔ", "sp": "yu", "r": "艸", "sc": 20, "l": 5, "f": 9261},
  {"h": "虋", "p": "mén", "sp": "men", "r": "艸", "sc": 29, "l": 5, "f": 9262},
  {"h": "蚿", "p": "xián", "sp": "xian", "r": "虫", "sc": 11, "l": 5, "f": 9263},
  {"h": "蜫", "p": "kūn", "sp": "kun", "r": "虫", "sc": 14, "l": 5, "f": 9264},
  {
    "h": "蜽",
    "p": "liǎng",
    "sp": "liang",
    "r": "虫",
    "sc": 14,
    "l": 5,
    "f": 9265
  },
  {"h": "螥", "p": "cāng", "sp": "cang", "r": "虫", "sc": 16, "l": 5, "f": 9266},
  {"h": "蠚", "p": "hē", "sp": "he", "r": "虫", "sc": 20, "l": 5, "f": 9267},
  {"h": "袽", "p": "rú", "sp": "ru", "r": "衣", "sc": 11, "l": 5, "f": 9268},
  {"h": "裶", "p": "fēi", "sp": "fei", "r": "衣", "sc": 13, "l": 5, "f": 9269},
  {"h": "裺", "p": "yǎn", "sp": "yan", "r": "衣", "sc": 13, "l": 5, "f": 9270},
  {"h": "褣", "p": "róng", "sp": "rong", "r": "衣", "sc": 15, "l": 5, "f": 9271},
  {"h": "褸", "p": "lǚ", "sp": "lu", "r": "衣", "sc": 16, "l": 5, "f": 9272},
  {"h": "詅", "p": "líng", "sp": "ling", "r": "言", "sc": 12, "l": 5, "f": 9273},
  {"h": "詇", "p": "yàng", "sp": "yang", "r": "言", "sc": 12, "l": 5, "f": 9274},
  {"h": "詐", "p": "zhà", "sp": "zha", "r": "言", "sc": 12, "l": 5, "f": 9275},
  {"h": "誻", "p": "tà", "sp": "ta", "r": "言", "sc": 15, "l": 5, "f": 9276},
  {"h": "諧", "p": "xié", "sp": "xie", "r": "言", "sc": 16, "l": 5, "f": 9277},
  {"h": "謚", "p": "shì", "sp": "shi", "r": "言", "sc": 17, "l": 5, "f": 9278},
  {"h": "豧", "p": "fū", "sp": "fu", "r": "豕", "sc": 14, "l": 5, "f": 9279},
  {"h": "貮", "p": "èr", "sp": "er", "r": "貝", "sc": 11, "l": 5, "f": 9280},
  {"h": "賡", "p": "gēng", "sp": "geng", "r": "貝", "sc": 15, "l": 5, "f": 9281},
  {"h": "賱", "p": "yǔn", "sp": "yun", "r": "貝", "sc": 16, "l": 5, "f": 9282},
  {"h": "赑", "p": "bì", "sp": "bi", "r": "貝", "sc": 12, "l": 5, "f": 9283},
  {
    "h": "赹",
    "p": "qióng",
    "sp": "qiong",
    "r": "走",
    "sc": 11,
    "l": 5,
    "f": 9284
  },
  {"h": "赽", "p": "jué", "sp": "jue", "r": "走", "sc": 11, "l": 5, "f": 9285},
  {"h": "踃", "p": "xiāo", "sp": "xiao", "r": "足", "sc": 14, "l": 5, "f": 9286},
  {"h": "蹞", "p": "kuǐ", "sp": "kui", "r": "足", "sc": 18, "l": 5, "f": 9287},
  {"h": "蹧", "p": "zāo", "sp": "zao", "r": "足", "sc": 18, "l": 5, "f": 9288},
  {"h": "躻", "p": "kong", "sp": "kong", "r": "身", "sc": 15, "l": 5, "f": 9289},
  {"h": "躿", "p": "kāng", "sp": "kang", "r": "身", "sc": 18, "l": 5, "f": 9290},
  {"h": "輻", "p": "fú", "sp": "fu", "r": "車", "sc": 16, "l": 5, "f": 9291},
  {"h": "郟", "p": "jiá", "sp": "jia", "r": "邑", "sc": 9, "l": 5, "f": 9292},
  {"h": "鄑", "p": "zī", "sp": "zi", "r": "邑", "sc": 12, "l": 5, "f": 9293},
  {"h": "酼", "p": "hǎi", "sp": "hai", "r": "酉", "sc": 14, "l": 5, "f": 9294},
  {"h": "醁", "p": "lù", "sp": "lu", "r": "酉", "sc": 15, "l": 5, "f": 9295},
  {"h": "醕", "p": "chún", "sp": "chun", "r": "酉", "sc": 16, "l": 5, "f": 9296},
  {"h": "醥", "p": "piǎo", "sp": "piao", "r": "酉", "sc": 18, "l": 5, "f": 9297},
  {"h": "釪", "p": "huá", "sp": "hua", "r": "金", "sc": 11, "l": 5, "f": 9298},
  {"h": "鈵", "p": "bǐng", "sp": "bing", "r": "金", "sc": 13, "l": 5, "f": 9299},
  {"h": "銒", "p": "xíng", "sp": "xing", "r": "金", "sc": 14, "l": 5, "f": 9300},
  {"h": "銙", "p": "kuǎ", "sp": "kua", "r": "金", "sc": 14, "l": 5, "f": 9301},
  {"h": "鋺", "p": "yuǎn", "sp": "yuan", "r": "金", "sc": 16, "l": 5, "f": 9302},
  {"h": "錗", "p": "nèi", "sp": "nei", "r": "金", "sc": 16, "l": 5, "f": 9303},
  {"h": "錥", "p": "yù", "sp": "yu", "r": "金", "sc": 16, "l": 5, "f": 9304},
  {"h": "鍓", "p": "jí", "sp": "ji", "r": "金", "sc": 17, "l": 5, "f": 9305},
  {"h": "鍟", "p": "xīng", "sp": "xing", "r": "金", "sc": 17, "l": 5, "f": 9306},
  {"h": "鎍", "p": "suǒ", "sp": "suo", "r": "金", "sc": 18, "l": 5, "f": 9307},
  {"h": "鎘", "p": "lì", "sp": "li", "r": "金", "sc": 18, "l": 5, "f": 9308},
  {
    "h": "鏿",
    "p": "chēng",
    "sp": "cheng",
    "r": "金",
    "sc": 20,
    "l": 5,
    "f": 9309
  },
  {"h": "鐎", "p": "jiāo", "sp": "jiao", "r": "金", "sc": 20, "l": 5, "f": 9310},
  {"h": "鐤", "p": "dǐng", "sp": "ding", "r": "金", "sc": 20, "l": 5, "f": 9311},
  {"h": "鐿", "p": "yì", "sp": "yi", "r": "金", "sc": 21, "l": 5, "f": 9312},
  {"h": "鑊", "p": "huò", "sp": "huo", "r": "金", "sc": 21, "l": 5, "f": 9313},
  {"h": "锧", "p": "zhì", "sp": "zhi", "r": "金", "sc": 13, "l": 5, "f": 9314},
  {"h": "閧", "p": "hòng", "sp": "hong", "r": "門", "sc": 14, "l": 5, "f": 9315},
  {"h": "闛", "p": "táng", "sp": "tang", "r": "門", "sc": 19, "l": 5, "f": 9316},
  {"h": "隝", "p": "dǎo", "sp": "dao", "r": "阜", "sc": 13, "l": 5, "f": 9317},
  {"h": "隥", "p": "dèng", "sp": "deng", "r": "阜", "sc": 14, "l": 5, "f": 9318},
  {"h": "隦", "p": "jiǎo", "sp": "jiao", "r": "阜", "sc": 15, "l": 5, "f": 9319},
  {"h": "靄", "p": "ǎi", "sp": "ai", "r": "雨", "sc": 24, "l": 5, "f": 9320},
  {"h": "靱", "p": "rèn", "sp": "ren", "r": "革", "sc": 12, "l": 5, "f": 9321},
  {"h": "鞪", "p": "mù", "sp": "mu", "r": "革", "sc": 18, "l": 5, "f": 9322},
  {"h": "鞺", "p": "tāng", "sp": "tang", "r": "革", "sc": 20, "l": 5, "f": 9323},
  {"h": "鞿", "p": "jī", "sp": "ji", "r": "革", "sc": 21, "l": 5, "f": 9324},
  {"h": "韟", "p": "gāo", "sp": "gao", "r": "韋", "sc": 19, "l": 5, "f": 9325},
  {"h": "顀", "p": "chuí", "sp": "chui", "r": "頁", "sc": 17, "l": 5, "f": 9326},
  {"h": "飀", "p": "liú", "sp": "liu", "r": "風", "sc": 19, "l": 5, "f": 9327},
  {"h": "飥", "p": "tuō", "sp": "tuo", "r": "食", "sc": 11, "l": 5, "f": 9328},
  {"h": "飪", "p": "rèn", "sp": "ren", "r": "食", "sc": 12, "l": 5, "f": 9329},
  {"h": "餒", "p": "něi", "sp": "nei", "r": "食", "sc": 15, "l": 5, "f": 9330},
  {"h": "饆", "p": "bì", "sp": "bi", "r": "食", "sc": 18, "l": 5, "f": 9331},
  {"h": "馿", "p": "lǘ", "sp": "lu", "r": "馬", "sc": 14, "l": 5, "f": 9332},
  {
    "h": "癀",
    "p": "huáng",
    "sp": "huang",
    "r": "疒",
    "sc": 16,
    "l": 5,
    "f": 9333
  },
  {
    "h": "驦",
    "p": "shuāng",
    "sp": "shuang",
    "r": "馬",
    "sc": 27,
    "l": 5,
    "f": 9334
  },
  {"h": "髉", "p": "bó", "sp": "bo", "r": "骨", "sc": 19, "l": 5, "f": 9335},
  {"h": "鬉", "p": "zōng", "sp": "zong", "r": "髟", "sc": 19, "l": 5, "f": 9336},
  {"h": "鬸", "p": "liù", "sp": "liu", "r": "鬲", "sc": 20, "l": 5, "f": 9337},
  {"h": "魛", "p": "dāo", "sp": "dao", "r": "魚", "sc": 13, "l": 5, "f": 9338},
  {"h": "鳷", "p": "zhī", "sp": "zhi", "r": "鳥", "sc": 15, "l": 5, "f": 9339},
  {"h": "鵋", "p": "jì", "sp": "ji", "r": "鳥", "sc": 18, "l": 5, "f": 9340},
  {"h": "鵏", "p": "bǔ", "sp": "bu", "r": "鳥", "sc": 18, "l": 5, "f": 9341},
  {"h": "鵨", "p": "shū", "sp": "shu", "r": "鳥", "sc": 19, "l": 5, "f": 9342},
  {"h": "鵼", "p": "kōng", "sp": "kong", "r": "鳥", "sc": 19, "l": 5, "f": 9343},
  {"h": "鷵", "p": "tú", "sp": "tu", "r": "鳥", "sc": 22, "l": 5, "f": 9344},
  {"h": "鷼", "p": "xián", "sp": "xian", "r": "鳥", "sc": 23, "l": 5, "f": 9345},
  {"h": "鸍", "p": "mí", "sp": "mi", "r": "鳥", "sc": 25, "l": 5, "f": 9346},
  {"h": "鸓", "p": "lěi", "sp": "lei", "r": "鳥", "sc": 26, "l": 5, "f": 9347},
  {"h": "鸖", "p": "hè", "sp": "he", "r": "鳥", "sc": 27, "l": 5, "f": 9348},
  {"h": "鹠", "p": "liú", "sp": "liu", "r": "鳥", "sc": 15, "l": 5, "f": 9349},
  {"h": "黶", "p": "yǎn", "sp": "yan", "r": "黑", "sc": 26, "l": 5, "f": 9350},
  {"h": "鼉", "p": "tuó", "sp": "tuo", "r": "黽", "sc": 25, "l": 5, "f": 9351},
  {"h": "齰", "p": "zé", "sp": "ze", "r": "齒", "sc": 23, "l": 5, "f": 9352},
  {
    "h": "碀",
    "p": "chéng",
    "sp": "cheng",
    "r": "石",
    "sc": 11,
    "l": 5,
    "f": 9353
  },
  {"h": "礇", "p": "yù", "sp": "yu", "r": "石", "sc": 17, "l": 5, "f": 9354},
  {"h": "禪", "p": "chán", "sp": "chan", "r": "示", "sc": 16, "l": 5, "f": 9355},
  {"h": "筣", "p": "lí", "sp": "li", "r": "竹", "sc": 13, "l": 5, "f": 9356},
  {
    "h": "箏",
    "p": "zhēng",
    "sp": "zheng",
    "r": "竹",
    "sc": 14,
    "l": 5,
    "f": 9357
  },
  {
    "h": "糧",
    "p": "liáng",
    "sp": "liang",
    "r": "米",
    "sc": 18,
    "l": 5,
    "f": 9358
  },
  {"h": "紐", "p": "niǔ", "sp": "niu", "r": "糸", "sc": 10, "l": 5, "f": 9359},
  {"h": "紳", "p": "shēn", "sp": "shen", "r": "糸", "sc": 11, "l": 5, "f": 9360},
  {"h": "綀", "p": "shū", "sp": "shu", "r": "糸", "sc": 13, "l": 5, "f": 9361},
  {"h": "緱", "p": "gōu", "sp": "gou", "r": "糸", "sc": 15, "l": 5, "f": 9362},
  {"h": "缾", "p": "píng", "sp": "ping", "r": "缶", "sc": 12, "l": 5, "f": 9363},
  {"h": "羷", "p": "liǎn", "sp": "lian", "r": "羊", "sc": 19, "l": 5, "f": 9364},
  {"h": "耊", "p": "diè", "sp": "die", "r": "老", "sc": 10, "l": 5, "f": 9365},
  {"h": "臚", "p": "lú", "sp": "lu", "r": "肉", "sc": 20, "l": 5, "f": 9366},
  {"h": "艍", "p": "jū", "sp": "ju", "r": "舟", "sc": 14, "l": 5, "f": 9367},
  {"h": "葏", "p": "jīng", "sp": "jing", "r": "艸", "sc": 12, "l": 5, "f": 9368},
  {"h": "蠅", "p": "yíng", "sp": "ying", "r": "虫", "sc": 19, "l": 5, "f": 9369},
  {"h": "詛", "p": "zǔ", "sp": "zu", "r": "言", "sc": 12, "l": 5, "f": 9370},
  {"h": "誽", "p": "nì", "sp": "ni", "r": "言", "sc": 15, "l": 5, "f": 9371},
  {"h": "諂", "p": "chǎn", "sp": "chan", "r": "言", "sc": 15, "l": 5, "f": 9372},
  {"h": "賂", "p": "lù", "sp": "lu", "r": "貝", "sc": 13, "l": 5, "f": 9373},
  {"h": "輧", "p": "píng", "sp": "ping", "r": "車", "sc": 15, "l": 5, "f": 9374},
  {"h": "輽", "p": "bèn", "sp": "ben", "r": "車", "sc": 17, "l": 5, "f": 9375},
  {"h": "釦", "p": "kòu", "sp": "kou", "r": "金", "sc": 11, "l": 5, "f": 9376},
  {"h": "鈕", "p": "niǔ", "sp": "niu", "r": "金", "sc": 12, "l": 5, "f": 9377},
  {"h": "鏍", "p": "luó", "sp": "luo", "r": "金", "sc": 19, "l": 5, "f": 9378},
  {"h": "鏽", "p": "xiù", "sp": "xiu", "r": "金", "sc": 21, "l": 5, "f": 9379},
  {"h": "鐲", "p": "zhuó", "sp": "zhuo", "r": "金", "sc": 21, "l": 5, "f": 9380},
  {"h": "颱", "p": "tái", "sp": "tai", "r": "風", "sc": 14, "l": 5, "f": 9381},
  {"h": "饝", "p": "mó", "sp": "mo", "r": "食", "sc": 24, "l": 5, "f": 9382},
  {"h": "騇", "p": "shè", "sp": "she", "r": "馬", "sc": 18, "l": 5, "f": 9383},
  {"h": "髏", "p": "lóu", "sp": "lou", "r": "骨", "sc": 20, "l": 5, "f": 9384},
  {"h": "腂", "p": "lěi", "sp": "lei", "r": "肉", "sc": 12, "l": 5, "f": 9385},
  {"h": "臗", "p": "kuān", "sp": "kuan", "r": "肉", "sc": 18, "l": 5, "f": 9386},
  {"h": "眀", "p": "míng", "sp": "ming", "r": "目", "sc": 9, "l": 5, "f": 9387},
  {"h": "砫", "p": "zhù", "sp": "zhu", "r": "石", "sc": 10, "l": 5, "f": 9388},
  {"h": "祙", "p": "mèi", "sp": "mei", "r": "示", "sc": 9, "l": 5, "f": 9389},
  {"h": "祳", "p": "shèn", "sp": "shen", "r": "示", "sc": 11, "l": 5, "f": 9390},
  {
    "h": "粧",
    "p": "zhuāng",
    "sp": "zhuang",
    "r": "米",
    "sc": 12,
    "l": 5,
    "f": 9391
  },
  {"h": "蕣", "p": "shùn", "sp": "shun", "r": "艸", "sc": 15, "l": 5, "f": 9392},
  {"h": "觓", "p": "qiú", "sp": "qiu", "r": "角", "sc": 9, "l": 5, "f": 9393},
  {"h": "趹", "p": "jué", "sp": "jue", "r": "足", "sc": 11, "l": 5, "f": 9394},
  {"h": "鄍", "p": "míng", "sp": "ming", "r": "邑", "sc": 12, "l": 5, "f": 9395},
  {
    "h": "鈡",
    "p": "zhōng",
    "sp": "zhong",
    "r": "金",
    "sc": 12,
    "l": 5,
    "f": 9396
  },
  {"h": "羭", "p": "yú", "sp": "yu", "r": "羊", "sc": 15, "l": 5, "f": 9397},
  {"h": "蝄", "p": "wǎng", "sp": "wang", "r": "虫", "sc": 14, "l": 5, "f": 9398},
  {"h": "衸", "p": "jiè", "sp": "jie", "r": "衣", "sc": 9, "l": 5, "f": 9399},
  {"h": "謡", "p": "yáo", "sp": "yao", "r": "言", "sc": 17, "l": 5, "f": 9400},
  {"h": "閏", "p": "rùn", "sp": "run", "r": "門", "sc": 12, "l": 5, "f": 9401},
  {"h": "盪", "p": "dàng", "sp": "dang", "r": "皿", "sc": 17, "l": 5, "f": 9402},
  {"h": "眲", "p": "nè", "sp": "ne", "r": "目", "sc": 11, "l": 5, "f": 9403},
  {"h": "睋", "p": "é", "sp": "e", "r": "目", "sc": 12, "l": 5, "f": 9404},
  {"h": "矃", "p": "nǐng", "sp": "ning", "r": "目", "sc": 19, "l": 5, "f": 9405},
  {"h": "矒", "p": "méng", "sp": "meng", "r": "目", "sc": 20, "l": 5, "f": 9406},
  {"h": "硟", "p": "chàn", "sp": "chan", "r": "石", "sc": 11, "l": 5, "f": 9407},
  {"h": "碩", "p": "shuò", "sp": "shuo", "r": "石", "sc": 14, "l": 5, "f": 9408},
  {"h": "磄", "p": "táng", "sp": "tang", "r": "石", "sc": 15, "l": 5, "f": 9409},
  {"h": "礂", "p": "xi", "sp": "xi", "r": "石", "sc": 17, "l": 5, "f": 9410},
  {"h": "礊", "p": "kè", "sp": "ke", "r": "石", "sc": 18, "l": 5, "f": 9411},
  {"h": "祰", "p": "gào", "sp": "gao", "r": "示", "sc": 11, "l": 5, "f": 9412},
  {"h": "祹", "p": "táo", "sp": "tao", "r": "示", "sc": 12, "l": 5, "f": 9413},
  {"h": "禐", "p": "yuàn", "sp": "yuan", "r": "示", "sc": 13, "l": 5, "f": 9414},
  {"h": "秳", "p": "huó", "sp": "huo", "r": "禾", "sc": 11, "l": 5, "f": 9415},
  {"h": "秼", "p": "zhū", "sp": "zhu", "r": "禾", "sc": 11, "l": 5, "f": 9416},
  {"h": "稄", "p": "xùn", "sp": "xun", "r": "禾", "sc": 12, "l": 5, "f": 9417},
  {"h": "稇", "p": "kǔn", "sp": "kun", "r": "禾", "sc": 12, "l": 5, "f": 9418},
  {"h": "稈", "p": "gǎn", "sp": "gan", "r": "禾", "sc": 12, "l": 5, "f": 9419},
  {"h": "稘", "p": "jī", "sp": "ji", "r": "禾", "sc": 13, "l": 5, "f": 9420},
  {"h": "稙", "p": "zhī", "sp": "zhi", "r": "禾", "sc": 13, "l": 5, "f": 9421},
  {"h": "穢", "p": "huì", "sp": "hui", "r": "禾", "sc": 18, "l": 5, "f": 9422},
  {"h": "竸", "p": "jìng", "sp": "jing", "r": "立", "sc": 22, "l": 5, "f": 9423},
  {"h": "竾", "p": "chí", "sp": "chi", "r": "竹", "sc": 9, "l": 5, "f": 9424},
  {"h": "笐", "p": "háng", "sp": "hang", "r": "竹", "sc": 10, "l": 5, "f": 9425},
  {"h": "笜", "p": "zhú", "sp": "zhu", "r": "竹", "sc": 11, "l": 5, "f": 9426},
  {"h": "筂", "p": "chí", "sp": "chi", "r": "竹", "sc": 12, "l": 5, "f": 9427},
  {"h": "箒", "p": "zhǒu", "sp": "zhou", "r": "竹", "sc": 14, "l": 5, "f": 9428},
  {"h": "箛", "p": "gū", "sp": "gu", "r": "竹", "sc": 14, "l": 5, "f": 9429},
  {"h": "箞", "p": "qiān", "sp": "qian", "r": "竹", "sc": 14, "l": 5, "f": 9430},
  {"h": "箤", "p": "zú", "sp": "zu", "r": "竹", "sc": 14, "l": 5, "f": 9431},
  {"h": "箷", "p": "yí", "sp": "yi", "r": "竹", "sc": 15, "l": 5, "f": 9432},
  {"h": "箹", "p": "yuē", "sp": "yue", "r": "竹", "sc": 15, "l": 5, "f": 9433},
  {"h": "箿", "p": "jí", "sp": "ji", "r": "竹", "sc": 15, "l": 5, "f": 9434},
  {
    "h": "篜",
    "p": "zhēng",
    "sp": "zheng",
    "r": "竹",
    "sc": 16,
    "l": 5,
    "f": 9435
  },
  {"h": "簞", "p": "dān", "sp": "dan", "r": "竹", "sc": 18, "l": 5, "f": 9436},
  {"h": "簤", "p": "dài", "sp": "dai", "r": "竹", "sc": 18, "l": 5, "f": 9437},
  {"h": "簨", "p": "sǔn", "sp": "sun", "r": "竹", "sc": 18, "l": 5, "f": 9438},
  {"h": "簹", "p": "dāng", "sp": "dang", "r": "竹", "sc": 19, "l": 5, "f": 9439},
  {"h": "簽", "p": "qiān", "sp": "qian", "r": "竹", "sc": 19, "l": 5, "f": 9440},
  {"h": "籬", "p": "lí", "sp": "li", "r": "竹", "sc": 24, "l": 5, "f": 9441},
  {"h": "粔", "p": "jù", "sp": "ju", "r": "米", "sc": 10, "l": 5, "f": 9442},
  {"h": "粸", "p": "qí", "sp": "qi", "r": "米", "sc": 14, "l": 5, "f": 9443},
  {"h": "紇", "p": "hé", "sp": "he", "r": "糸", "sc": 9, "l": 5, "f": 9444},
  {"h": "紝", "p": "rèn", "sp": "ren", "r": "糸", "sc": 10, "l": 5, "f": 9445},
  {"h": "紺", "p": "gàn", "sp": "gan", "r": "糸", "sc": 11, "l": 5, "f": 9446},
  {"h": "絆", "p": "bàn", "sp": "ban", "r": "糸", "sc": 11, "l": 5, "f": 9447},
  {
    "h": "絳",
    "p": "jiàng",
    "sp": "jiang",
    "r": "糸",
    "sc": 12,
    "l": 5,
    "f": 9448
  },
  {"h": "綄", "p": "huán", "sp": "huan", "r": "糸", "sc": 13, "l": 5, "f": 9449},
  {"h": "綊", "p": "xié", "sp": "xie", "r": "糸", "sc": 13, "l": 5, "f": 9450},
  {"h": "綏", "p": "suī", "sp": "sui", "r": "糸", "sc": 13, "l": 5, "f": 9451},
  {"h": "綒", "p": "fū", "sp": "fu", "r": "糸", "sc": 13, "l": 5, "f": 9452},
  {"h": "綔", "p": "hù", "sp": "hu", "r": "糸", "sc": 13, "l": 5, "f": 9453},
  {
    "h": "綗",
    "p": "jiǒng",
    "sp": "jiong",
    "r": "糸",
    "sc": 13,
    "l": 5,
    "f": 9454
  },
  {"h": "綴", "p": "zhuì", "sp": "zhui", "r": "糸", "sc": 14, "l": 5, "f": 9455},
  {"h": "緝", "p": "jī", "sp": "ji", "r": "糸", "sc": 15, "l": 5, "f": 9456},
  {"h": "緞", "p": "duàn", "sp": "duan", "r": "糸", "sc": 15, "l": 5, "f": 9457},
  {"h": "緦", "p": "sī", "sp": "si", "r": "糸", "sc": 15, "l": 5, "f": 9458},
  {"h": "緰", "p": "tóu", "sp": "tou", "r": "糸", "sc": 15, "l": 5, "f": 9459},
  {"h": "緶", "p": "biàn", "sp": "bian", "r": "糸", "sc": 15, "l": 5, "f": 9460},
  {"h": "縣", "p": "xiàn", "sp": "xian", "r": "糸", "sc": 16, "l": 5, "f": 9461},
  {"h": "縹", "p": "piǎo", "sp": "piao", "r": "糸", "sc": 17, "l": 5, "f": 9462},
  {"h": "繜", "p": "zūn", "sp": "zun", "r": "糸", "sc": 18, "l": 5, "f": 9463},
  {"h": "繝", "p": "jiàn", "sp": "jian", "r": "糸", "sc": 18, "l": 5, "f": 9464},
  {
    "h": "繩",
    "p": "shéng",
    "sp": "sheng",
    "r": "糸",
    "sc": 19,
    "l": 5,
    "f": 9465
  },
  {"h": "纏", "p": "chán", "sp": "chan", "r": "糸", "sc": 21, "l": 5, "f": 9466},
  {"h": "罃", "p": "yīng", "sp": "ying", "r": "缶", "sc": 16, "l": 5, "f": 9467},
  {"h": "罝", "p": "jū", "sp": "ju", "r": "网", "sc": 10, "l": 5, "f": 9468},
  {"h": "罰", "p": "fá", "sp": "fa", "r": "网", "sc": 14, "l": 5, "f": 9469},
  {"h": "羴", "p": "shān", "sp": "shan", "r": "羊", "sc": 18, "l": 5, "f": 9470},
  {"h": "翢", "p": "dào", "sp": "dao", "r": "羽", "sc": 14, "l": 5, "f": 9471},
  {"h": "聡", "p": "cōng", "sp": "cong", "r": "耳", "sc": 14, "l": 5, "f": 9472},
  {"h": "聹", "p": "níng", "sp": "ning", "r": "耳", "sc": 20, "l": 5, "f": 9473},
  {"h": "聻", "p": "nǐ", "sp": "ni", "r": "耳", "sc": 20, "l": 5, "f": 9474},
  {"h": "聾", "p": "lóng", "sp": "long", "r": "耳", "sc": 22, "l": 5, "f": 9475},
  {"h": "脕", "p": "wàn", "sp": "wan", "r": "肉", "sc": 11, "l": 5, "f": 9476},
  {"h": "脟", "p": "liè", "sp": "lie", "r": "肉", "sc": 11, "l": 5, "f": 9477},
  {"h": "脳", "p": "nao", "sp": "nao", "r": "肉", "sc": 11, "l": 5, "f": 9478},
  {"h": "脽", "p": "shuí", "sp": "shui", "r": "肉", "sc": 12, "l": 5, "f": 9479},
  {"h": "脿", "p": "biāo", "sp": "biao", "r": "肉", "sc": 12, "l": 5, "f": 9480},
  {"h": "臈", "p": "là", "sp": "la", "r": "肉", "sc": 16, "l": 5, "f": 9481},
  {"h": "艑", "p": "biàn", "sp": "bian", "r": "舟", "sc": 15, "l": 5, "f": 9482},
  {"h": "艙", "p": "cāng", "sp": "cang", "r": "舟", "sc": 16, "l": 5, "f": 9483},
  {
    "h": "艢",
    "p": "qiáng",
    "sp": "qiang",
    "r": "舟",
    "sc": 19,
    "l": 5,
    "f": 9484
  },
  {"h": "芻", "p": "chú", "sp": "chu", "r": "艸", "sc": 10, "l": 5, "f": 9485},
  {"h": "茻", "p": "mǎng", "sp": "mang", "r": "艸", "sc": 12, "l": 5, "f": 9486},
  {"h": "荊", "p": "jīng", "sp": "jing", "r": "艸", "sc": 9, "l": 5, "f": 9487},
  {"h": "菐", "p": "pú", "sp": "pu", "r": "艸", "sc": 12, "l": 5, "f": 9488},
  {"h": "菓", "p": "guǒ", "sp": "guo", "r": "艸", "sc": 11, "l": 5, "f": 9489},
  {"h": "菙", "p": "chuí", "sp": "chui", "r": "艸", "sc": 11, "l": 5, "f": 9490},
  {"h": "菚", "p": "zhàn", "sp": "zhan", "r": "艸", "sc": 11, "l": 5, "f": 9491},
  {"h": "菣", "p": "qìn", "sp": "qin", "r": "艸", "sc": 11, "l": 5, "f": 9492},
  {"h": "菦", "p": "qín", "sp": "qin", "r": "艸", "sc": 10, "l": 5, "f": 9493},
  {"h": "菨", "p": "jiē", "sp": "jie", "r": "艸", "sc": 11, "l": 5, "f": 9494},
  {"h": "葔", "p": "hóu", "sp": "hou", "r": "艸", "sc": 12, "l": 5, "f": 9495},
  {"h": "葠", "p": "shēn", "sp": "shen", "r": "艸", "sc": 12, "l": 5, "f": 9496},
  {"h": "葮", "p": "duàn", "sp": "duan", "r": "艸", "sc": 12, "l": 5, "f": 9497},
  {"h": "葲", "p": "quán", "sp": "quan", "r": "艸", "sc": 12, "l": 5, "f": 9498},
  {"h": "蓙", "p": "zuo", "sp": "zuo", "r": "艸", "sc": 13, "l": 5, "f": 9499},
  {
    "h": "蔁",
    "p": "zhāng",
    "sp": "zhang",
    "r": "艸",
    "sc": 14,
    "l": 5,
    "f": 9500
  },
  {"h": "蕛", "p": "tí", "sp": "ti", "r": "艸", "sc": 15, "l": 5, "f": 9501},
  {"h": "蕜", "p": "fěi", "sp": "fei", "r": "艸", "sc": 15, "l": 5, "f": 9502},
  {"h": "蕦", "p": "xū", "sp": "xu", "r": "艸", "sc": 15, "l": 5, "f": 9503},
  {"h": "蕵", "p": "sūn", "sp": "sun", "r": "艸", "sc": 15, "l": 5, "f": 9504},
  {"h": "蕷", "p": "yù", "sp": "yu", "r": "艸", "sc": 16, "l": 5, "f": 9505},
  {"h": "薖", "p": "kē", "sp": "ke", "r": "艸", "sc": 14, "l": 5, "f": 9506},
  {"h": "藨", "p": "biāo", "sp": "biao", "r": "艸", "sc": 18, "l": 5, "f": 9507},
  {"h": "藮", "p": "qiáo", "sp": "qiao", "r": "艸", "sc": 19, "l": 5, "f": 9508},
  {"h": "藰", "p": "liú", "sp": "liu", "r": "艸", "sc": 18, "l": 5, "f": 9509},
  {"h": "藳", "p": "gǎo", "sp": "gao", "r": "艸", "sc": 18, "l": 5, "f": 9510},
  {"h": "藵", "p": "bao", "sp": "bao", "r": "艸", "sc": 18, "l": 5, "f": 9511},
  {"h": "虃", "p": "jiān", "sp": "jian", "r": "艸", "sc": 23, "l": 5, "f": 9512},
  {"h": "虘", "p": "cuó", "sp": "cuo", "r": "虍", "sc": 11, "l": 5, "f": 9513},
  {"h": "虜", "p": "lǔ", "sp": "lu", "r": "虍", "sc": 13, "l": 5, "f": 9514},
  {"h": "虝", "p": "hǔ", "sp": "hu", "r": "虍", "sc": 12, "l": 5, "f": 9515},
  {"h": "蛥", "p": "shé", "sp": "she", "r": "虫", "sc": 12, "l": 5, "f": 9516},
  {"h": "蛶", "p": "jiè", "sp": "jie", "r": "虫", "sc": 13, "l": 5, "f": 9517},
  {"h": "蜠", "p": "jùn", "sp": "jun", "r": "虫", "sc": 14, "l": 5, "f": 9518},
  {"h": "蝸", "p": "wō", "sp": "wo", "r": "虫", "sc": 14, "l": 5, "f": 9519},
  {"h": "螐", "p": "wū", "sp": "wu", "r": "虫", "sc": 16, "l": 5, "f": 9520},
  {"h": "螖", "p": "huá", "sp": "hua", "r": "虫", "sc": 15, "l": 5, "f": 9521},
  {"h": "螡", "p": "wén", "sp": "wen", "r": "虫", "sc": 16, "l": 5, "f": 9522},
  {"h": "螢", "p": "yíng", "sp": "ying", "r": "虫", "sc": 16, "l": 5, "f": 9523},
  {"h": "蟳", "p": "xún", "sp": "xun", "r": "虫", "sc": 18, "l": 5, "f": 9524},
  {"h": "蠌", "p": "zé", "sp": "ze", "r": "虫", "sc": 19, "l": 5, "f": 9525},
  {"h": "蠍", "p": "xiē", "sp": "xie", "r": "虫", "sc": 19, "l": 5, "f": 9526},
  {"h": "蠒", "p": "jiǎn", "sp": "jian", "r": "虫", "sc": 20, "l": 5, "f": 9527},
  {"h": "蠗", "p": "zhuó", "sp": "zhuo", "r": "虫", "sc": 20, "l": 5, "f": 9528},
  {"h": "蠝", "p": "léi", "sp": "lei", "r": "虫", "sc": 21, "l": 5, "f": 9529},
  {"h": "蠟", "p": "là", "sp": "la", "r": "虫", "sc": 21, "l": 5, "f": 9530},
  {"h": "袘", "p": "yí", "sp": "yi", "r": "衣", "sc": 10, "l": 5, "f": 9531},
  {"h": "褜", "p": "pao", "sp": "pao", "r": "衣", "sc": 15, "l": 5, "f": 9532},
  {"h": "褨", "p": "suǒ", "sp": "suo", "r": "衣", "sc": 14, "l": 5, "f": 9533},
  {"h": "襾", "p": "yà", "sp": "ya", "r": "襾", "sc": 6, "l": 5, "f": 9534},
  {"h": "覊", "p": "jī", "sp": "ji", "r": "襾", "sc": 25, "l": 5, "f": 9535},
  {"h": "覚", "p": "jué", "sp": "jue", "r": "見", "sc": 12, "l": 5, "f": 9536},
  {"h": "覛", "p": "mì", "sp": "mi", "r": "見", "sc": 13, "l": 5, "f": 9537},
  {"h": "覟", "p": "zhì", "sp": "zhi", "r": "見", "sc": 14, "l": 5, "f": 9538},
  {"h": "覠", "p": "jūn", "sp": "jun", "r": "見", "sc": 14, "l": 5, "f": 9539},
  {"h": "覯", "p": "gòu", "sp": "gou", "r": "見", "sc": 17, "l": 5, "f": 9540},
  {"h": "訜", "p": "fēn", "sp": "fen", "r": "言", "sc": 11, "l": 5, "f": 9541},
  {"h": "訡", "p": "yín", "sp": "yin", "r": "言", "sc": 11, "l": 5, "f": 9542},
  {"h": "訮", "p": "xiān", "sp": "xian", "r": "言", "sc": 11, "l": 5, "f": 9543},
  {
    "h": "証",
    "p": "zhèng",
    "sp": "zheng",
    "r": "言",
    "sc": 12,
    "l": 5,
    "f": 9544
  },
  {"h": "詷", "p": "tóng", "sp": "tong", "r": "言", "sc": 13, "l": 5, "f": 9545},
  {"h": "詺", "p": "mìng", "sp": "ming", "r": "言", "sc": 13, "l": 5, "f": 9546},
  {"h": "諉", "p": "wěi", "sp": "wei", "r": "言", "sc": 15, "l": 5, "f": 9547},
  {
    "h": "諒",
    "p": "liàng",
    "sp": "liang",
    "r": "言",
    "sc": 15,
    "l": 5,
    "f": 9548
  },
  {"h": "諘", "p": "biǎo", "sp": "biao", "r": "言", "sc": 15, "l": 5, "f": 9549},
  {"h": "諦", "p": "dì", "sp": "di", "r": "言", "sc": 16, "l": 5, "f": 9550},
  {"h": "諺", "p": "yàn", "sp": "yan", "r": "言", "sc": 16, "l": 5, "f": 9551},
  {"h": "謀", "p": "móu", "sp": "mou", "r": "言", "sc": 16, "l": 5, "f": 9552},
  {"h": "謠", "p": "yáo", "sp": "yao", "r": "言", "sc": 17, "l": 5, "f": 9553},
  {"h": "謩", "p": "mò", "sp": "mo", "r": "言", "sc": 17, "l": 5, "f": 9554},
  {"h": "謸", "p": "áo", "sp": "ao", "r": "言", "sc": 17, "l": 5, "f": 9555},
  {
    "h": "謽",
    "p": "jiàng",
    "sp": "jiang",
    "r": "言",
    "sc": 19,
    "l": 5,
    "f": 9556
  },
  {"h": "讍", "p": "è", "sp": "e", "r": "言", "sc": 23, "l": 5, "f": 9557},
  {"h": "讛", "p": "yì", "sp": "yi", "r": "言", "sc": 25, "l": 5, "f": 9558},
  {"h": "貦", "p": "wàn", "sp": "wan", "r": "貝", "sc": 11, "l": 5, "f": 9559},
  {"h": "貸", "p": "dài", "sp": "dai", "r": "貝", "sc": 12, "l": 5, "f": 9560},
  {"h": "貽", "p": "yí", "sp": "yi", "r": "貝", "sc": 12, "l": 5, "f": 9561},
  {"h": "賅", "p": "gāi", "sp": "gai", "r": "貝", "sc": 13, "l": 5, "f": 9562},
  {"h": "賊", "p": "zéi", "sp": "zei", "r": "貝", "sc": 13, "l": 5, "f": 9563},
  {"h": "賳", "p": "zāi", "sp": "zai", "r": "貝", "sc": 16, "l": 5, "f": 9564},
  {
    "h": "賺",
    "p": "zhuàn",
    "sp": "zhuan",
    "r": "貝",
    "sc": 17,
    "l": 5,
    "f": 9565
  },
  {"h": "贀", "p": "yì", "sp": "yi", "r": "貝", "sc": 18, "l": 5, "f": 9566},
  {"h": "贄", "p": "zhì", "sp": "zhi", "r": "貝", "sc": 18, "l": 5, "f": 9567},
  {"h": "贊", "p": "zàn", "sp": "zan", "r": "貝", "sc": 19, "l": 5, "f": 9568},
  {"h": "贌", "p": "pu", "sp": "pu", "r": "貝", "sc": 19, "l": 5, "f": 9569},
  {
    "h": "趪",
    "p": "huáng",
    "sp": "huang",
    "r": "走",
    "sc": 18,
    "l": 5,
    "f": 9570
  },
  {"h": "跁", "p": "bà", "sp": "ba", "r": "足", "sc": 11, "l": 5, "f": 9571},
  {"h": "跈", "p": "niǎn", "sp": "nian", "r": "足", "sc": 12, "l": 5, "f": 9572},
  {"h": "跒", "p": "qiǎ", "sp": "qia", "r": "足", "sc": 12, "l": 5, "f": 9573},
  {
    "h": "蹖",
    "p": "chōng",
    "sp": "chong",
    "r": "足",
    "sc": 18,
    "l": 5,
    "f": 9574
  },
  {"h": "軀", "p": "qū", "sp": "qu", "r": "身", "sc": 18, "l": 5, "f": 9575},
  {"h": "軇", "p": "dào", "sp": "dao", "r": "身", "sc": 21, "l": 5, "f": 9576},
  {"h": "軔", "p": "rèn", "sp": "ren", "r": "車", "sc": 10, "l": 5, "f": 9577},
  {"h": "軘", "p": "tún", "sp": "tun", "r": "車", "sc": 11, "l": 5, "f": 9578},
  {"h": "軣", "p": "hōng", "sp": "hong", "r": "車", "sc": 11, "l": 5, "f": 9579},
  {"h": "軻", "p": "kē", "sp": "ke", "r": "車", "sc": 12, "l": 5, "f": 9580},
  {"h": "輒", "p": "zhé", "sp": "zhe", "r": "車", "sc": 14, "l": 5, "f": 9581},
  {"h": "輥", "p": "gǔn", "sp": "gun", "r": "車", "sc": 15, "l": 5, "f": 9582},
  {"h": "輦", "p": "niǎn", "sp": "nian", "r": "車", "sc": 15, "l": 5, "f": 9583},
  {"h": "轟", "p": "hōng", "sp": "hong", "r": "車", "sc": 21, "l": 5, "f": 9584},
  {"h": "逇", "p": "dùn", "sp": "dun", "r": "辵", "sc": 9, "l": 5, "f": 9585},
  {"h": "邔", "p": "qǐ", "sp": "qi", "r": "邑", "sc": 5, "l": 5, "f": 9586},
  {"h": "邥", "p": "shěn", "sp": "shen", "r": "邑", "sc": 6, "l": 5, "f": 9587},
  {"h": "邫", "p": "bāng", "sp": "bang", "r": "邑", "sc": 11, "l": 5, "f": 9588},
  {"h": "鄽", "p": "chán", "sp": "chan", "r": "邑", "sc": 17, "l": 5, "f": 9589},
  {"h": "醆", "p": "zhǎn", "sp": "zhan", "r": "酉", "sc": 15, "l": 5, "f": 9590},
  {"h": "醞", "p": "yùn", "sp": "yun", "r": "酉", "sc": 17, "l": 5, "f": 9591},
  {"h": "醽", "p": "líng", "sp": "ling", "r": "酉", "sc": 24, "l": 5, "f": 9592},
  {"h": "釥", "p": "qiǎo", "sp": "qiao", "r": "金", "sc": 11, "l": 5, "f": 9593},
  {"h": "鈂", "p": "chén", "sp": "chen", "r": "金", "sc": 12, "l": 5, "f": 9594},
  {"h": "鈽", "p": "bū", "sp": "bu", "r": "金", "sc": 13, "l": 5, "f": 9595},
  {"h": "鉍", "p": "bì", "sp": "bi", "r": "金", "sc": 13, "l": 5, "f": 9596},
  {"h": "鉛", "p": "qiān", "sp": "qian", "r": "金", "sc": 13, "l": 5, "f": 9597},
  {"h": "銓", "p": "quán", "sp": "quan", "r": "金", "sc": 14, "l": 5, "f": 9598},
  {"h": "銜", "p": "xián", "sp": "xian", "r": "金", "sc": 14, "l": 5, "f": 9599},
  {"h": "鋇", "p": "bèi", "sp": "bei", "r": "金", "sc": 15, "l": 5, "f": 9600},
  {"h": "鋒", "p": "fēng", "sp": "feng", "r": "金", "sc": 15, "l": 5, "f": 9601},
  {"h": "鋰", "p": "lǐ", "sp": "li", "r": "金", "sc": 15, "l": 5, "f": 9602},
  {"h": "鋷", "p": "zuì", "sp": "zui", "r": "金", "sc": 16, "l": 5, "f": 9603},
  {"h": "錠", "p": "dìng", "sp": "ding", "r": "金", "sc": 16, "l": 5, "f": 9604},
  {"h": "錦", "p": "jǐn", "sp": "jin", "r": "金", "sc": 16, "l": 5, "f": 9605},
  {"h": "鎲", "p": "tǎng", "sp": "tang", "r": "金", "sc": 18, "l": 5, "f": 9606},
  {"h": "鏈", "p": "liàn", "sp": "lian", "r": "金", "sc": 18, "l": 5, "f": 9607},
  {"h": "鏟", "p": "chǎn", "sp": "chan", "r": "金", "sc": 19, "l": 5, "f": 9608},
  {"h": "鐑", "p": "qiè", "sp": "qie", "r": "金", "sc": 20, "l": 5, "f": 9609},
  {"h": "鐬", "p": "huì", "sp": "hui", "r": "金", "sc": 21, "l": 5, "f": 9610},
  {
    "h": "鑲",
    "p": "xiāng",
    "sp": "xiang",
    "r": "金",
    "sc": 25,
    "l": 5,
    "f": 9611
  },
  {"h": "鑴", "p": "xī", "sp": "xi", "r": "金", "sc": 26, "l": 5, "f": 9612},
  {"h": "鑿", "p": "záo", "sp": "zao", "r": "金", "sc": 28, "l": 5, "f": 9613},
  {"h": "镻", "p": "dié", "sp": "die", "r": "長", "sc": 12, "l": 5, "f": 9614},
  {"h": "閑", "p": "xián", "sp": "xian", "r": "門", "sc": 12, "l": 5, "f": 9615},
  {"h": "閰", "p": "jú", "sp": "ju", "r": "門", "sc": 14, "l": 5, "f": 9616},
  {"h": "閳", "p": "chǎn", "sp": "chan", "r": "門", "sc": 15, "l": 5, "f": 9617},
  {"h": "閴", "p": "qù", "sp": "qu", "r": "門", "sc": 15, "l": 5, "f": 9618},
  {"h": "閹", "p": "yān", "sp": "yan", "r": "門", "sc": 16, "l": 5, "f": 9619},
  {"h": "闌", "p": "lán", "sp": "lan", "r": "門", "sc": 17, "l": 5, "f": 9620},
  {"h": "闡", "p": "chǎn", "sp": "chan", "r": "門", "sc": 20, "l": 5, "f": 9621},
  {"h": "陮", "p": "duì", "sp": "dui", "r": "阜", "sc": 10, "l": 5, "f": 9622},
  {"h": "陻", "p": "yīn", "sp": "yin", "r": "阜", "sc": 11, "l": 5, "f": 9623},
  {"h": "霂", "p": "mù", "sp": "mu", "r": "雨", "sc": 15, "l": 5, "f": 9624},
  {"h": "靵", "p": "niǔ", "sp": "niu", "r": "革", "sc": 13, "l": 5, "f": 9625},
  {"h": "鞊", "p": "jié", "sp": "jie", "r": "革", "sc": 15, "l": 5, "f": 9626},
  {"h": "鞛", "p": "běng", "sp": "beng", "r": "革", "sc": 17, "l": 5, "f": 9627},
  {"h": "韓", "p": "hán", "sp": "han", "r": "韋", "sc": 17, "l": 5, "f": 9628},
  {"h": "韴", "p": "zá", "sp": "za", "r": "音", "sc": 13, "l": 5, "f": 9629},
  {"h": "韷", "p": "lè", "sp": "le", "r": "音", "sc": 14, "l": 5, "f": 9630},
  {"h": "頊", "p": "xū", "sp": "xu", "r": "頁", "sc": 13, "l": 5, "f": 9631},
  {"h": "頋", "p": "ě", "sp": "e", "r": "頁", "sc": 13, "l": 5, "f": 9632},
  {"h": "頓", "p": "dùn", "sp": "dun", "r": "頁", "sc": 13, "l": 5, "f": 9633},
  {"h": "頜", "p": "hé", "sp": "he", "r": "頁", "sc": 15, "l": 5, "f": 9634},
  {"h": "顕", "p": "xiǎn", "sp": "xian", "r": "頁", "sc": 18, "l": 5, "f": 9635},
  {"h": "顨", "p": "xùn", "sp": "xun", "r": "頁", "sc": 21, "l": 5, "f": 9636},
  {"h": "顴", "p": "quán", "sp": "quan", "r": "頁", "sc": 26, "l": 5, "f": 9637},
  {"h": "飷", "p": "jiě", "sp": "jie", "r": "食", "sc": 13, "l": 5, "f": 9638},
  {"h": "飺", "p": "cí", "sp": "ci", "r": "食", "sc": 15, "l": 5, "f": 9639},
  {"h": "飾", "p": "shì", "sp": "shi", "r": "食", "sc": 13, "l": 5, "f": 9640},
  {"h": "餀", "p": "hài", "sp": "hai", "r": "食", "sc": 13, "l": 5, "f": 9641},
  {"h": "餌", "p": "ěr", "sp": "er", "r": "食", "sc": 14, "l": 5, "f": 9642},
  {"h": "餓", "p": "è", "sp": "e", "r": "食", "sc": 15, "l": 5, "f": 9643},
  {"h": "饒", "p": "ráo", "sp": "rao", "r": "食", "sc": 20, "l": 5, "f": 9644},
  {"h": "饡", "p": "zàn", "sp": "zan", "r": "食", "sc": 27, "l": 5, "f": 9645},
  {"h": "饳", "p": "duò", "sp": "duo", "r": "食", "sc": 8, "l": 5, "f": 9646},
  {"h": "饻", "p": "xī", "sp": "xi", "r": "食", "sc": 9, "l": 5, "f": 9647},
  {"h": "馴", "p": "xún", "sp": "xun", "r": "馬", "sc": 13, "l": 5, "f": 9648},
  {"h": "駜", "p": "bì", "sp": "bi", "r": "馬", "sc": 15, "l": 5, "f": 9649},
  {"h": "駣", "p": "táo", "sp": "tao", "r": "馬", "sc": 16, "l": 5, "f": 9650},
  {"h": "駥", "p": "róng", "sp": "rong", "r": "馬", "sc": 16, "l": 5, "f": 9651},
  {"h": "駨", "p": "xūn", "sp": "xun", "r": "馬", "sc": 16, "l": 5, "f": 9652},
  {"h": "駶", "p": "jú", "sp": "ju", "r": "馬", "sc": 17, "l": 5, "f": 9653},
  {"h": "騰", "p": "téng", "sp": "teng", "r": "馬", "sc": 20, "l": 5, "f": 9654},
  {
    "h": "騿",
    "p": "zhāng",
    "sp": "zhang",
    "r": "馬",
    "sc": 21,
    "l": 5,
    "f": 9655
  },
  {"h": "驁", "p": "ào", "sp": "ao", "r": "馬", "sc": 20, "l": 5, "f": 9656},
  {"h": "驇", "p": "zhì", "sp": "zhi", "r": "馬", "sc": 21, "l": 5, "f": 9657},
  {"h": "驑", "p": "liú", "sp": "liu", "r": "馬", "sc": 22, "l": 5, "f": 9658},
  {"h": "驕", "p": "jiāo", "sp": "jiao", "r": "馬", "sc": 22, "l": 5, "f": 9659},
  {"h": "驖", "p": "tiě", "sp": "tie", "r": "馬", "sc": 23, "l": 5, "f": 9660},
  {"h": "驥", "p": "jì", "sp": "ji", "r": "馬", "sc": 26, "l": 5, "f": 9661},
  {"h": "髞", "p": "sào", "sp": "sao", "r": "高", "sc": 23, "l": 5, "f": 9662},
  {"h": "鬀", "p": "tì", "sp": "ti", "r": "髟", "sc": 17, "l": 5, "f": 9663},
  {"h": "鬄", "p": "dí", "sp": "di", "r": "髟", "sc": 18, "l": 5, "f": 9664},
  {"h": "鬅", "p": "péng", "sp": "peng", "r": "髟", "sc": 18, "l": 5, "f": 9665},
  {"h": "鬢", "p": "bìn", "sp": "bin", "r": "髟", "sc": 24, "l": 5, "f": 9666},
  {"h": "鬧", "p": "nào", "sp": "nao", "r": "鬥", "sc": 15, "l": 5, "f": 9667},
  {"h": "魰", "p": "wén", "sp": "wen", "r": "魚", "sc": 15, "l": 5, "f": 9668},
  {"h": "魵", "p": "fén", "sp": "fen", "r": "魚", "sc": 15, "l": 5, "f": 9669},
  {"h": "魸", "p": "pian", "sp": "pian", "r": "魚", "sc": 15, "l": 5, "f": 9670},
  {"h": "魺", "p": "hé", "sp": "he", "r": "魚", "sc": 16, "l": 5, "f": 9671},
  {"h": "鯊", "p": "shā", "sp": "sha", "r": "魚", "sc": 18, "l": 5, "f": 9672},
  {"h": "鯐", "p": "zou", "sp": "zou", "r": "魚", "sc": 18, "l": 5, "f": 9673},
  {"h": "鯓", "p": "ní", "sp": "ni", "r": "魚", "sc": 18, "l": 5, "f": 9674},
  {
    "h": "鯖",
    "p": "zhēng",
    "sp": "zheng",
    "r": "魚",
    "sc": 19,
    "l": 5,
    "f": 9675
  },
  {
    "h": "鯗",
    "p": "xiǎng",
    "sp": "xiang",
    "r": "魚",
    "sc": 18,
    "l": 5,
    "f": 9676
  },
  {"h": "鰞", "p": "wū", "sp": "wu", "r": "魚", "sc": 21, "l": 5, "f": 9677},
  {"h": "鰬", "p": "qián", "sp": "qian", "r": "魚", "sc": 21, "l": 5, "f": 9678},
  {"h": "鰵", "p": "mǐn", "sp": "min", "r": "魚", "sc": 22, "l": 5, "f": 9679},
  {"h": "鰸", "p": "qū", "sp": "qu", "r": "魚", "sc": 22, "l": 5, "f": 9680},
  {"h": "鰻", "p": "mán", "sp": "man", "r": "魚", "sc": 22, "l": 5, "f": 9681},
  {
    "h": "鱌",
    "p": "xiàng",
    "sp": "xiang",
    "r": "魚",
    "sc": 22,
    "l": 5,
    "f": 9682
  },
  {"h": "鱺", "p": "lí", "sp": "li", "r": "魚", "sc": 30, "l": 5, "f": 9683},
  {"h": "鳩", "p": "jiū", "sp": "jiu", "r": "鳥", "sc": 13, "l": 5, "f": 9684},
  {"h": "鳪", "p": "bú", "sp": "bu", "r": "鳥", "sc": 13, "l": 5, "f": 9685},
  {"h": "鳶", "p": "yuān", "sp": "yuan", "r": "鳥", "sc": 14, "l": 5, "f": 9686},
  {"h": "鴭", "p": "duī", "sp": "dui", "r": "鳥", "sc": 17, "l": 5, "f": 9687},
  {"h": "鴳", "p": "yàn", "sp": "yan", "r": "鳥", "sc": 17, "l": 5, "f": 9688},
  {"h": "鴴", "p": "héng", "sp": "heng", "r": "鳥", "sc": 17, "l": 5, "f": 9689},
  {"h": "鴵", "p": "xiāo", "sp": "xiao", "r": "鳥", "sc": 17, "l": 5, "f": 9690},
  {"h": "鴺", "p": "tí", "sp": "ti", "r": "鳥", "sc": 17, "l": 5, "f": 9691},
  {"h": "鴽", "p": "rú", "sp": "ru", "r": "鳥", "sc": 17, "l": 5, "f": 9692},
  {"h": "鵁", "p": "jiāo", "sp": "jiao", "r": "鳥", "sc": 17, "l": 5, "f": 9693},
  {"h": "鵝", "p": "é", "sp": "e", "r": "鳥", "sc": 18, "l": 5, "f": 9694},
  {"h": "鶈", "p": "qī", "sp": "qi", "r": "鳥", "sc": 19, "l": 5, "f": 9695},
  {"h": "鶐", "p": "shù", "sp": "shu", "r": "鳥", "sc": 19, "l": 5, "f": 9696},
  {"h": "鶓", "p": "miáo", "sp": "miao", "r": "鳥", "sc": 19, "l": 5, "f": 9697},
  {"h": "鶩", "p": "wù", "sp": "wu", "r": "鳥", "sc": 20, "l": 5, "f": 9698},
  {"h": "鶴", "p": "hè", "sp": "he", "r": "鳥", "sc": 21, "l": 5, "f": 9699},
  {"h": "鷩", "p": "bì", "sp": "bi", "r": "鳥", "sc": 22, "l": 5, "f": 9700},
  {"h": "鷴", "p": "xián", "sp": "xian", "r": "鳥", "sc": 23, "l": 5, "f": 9701},
  {"h": "鷽", "p": "xué", "sp": "xue", "r": "鳥", "sc": 24, "l": 5, "f": 9702},
  {"h": "麩", "p": "fū", "sp": "fu", "r": "麥", "sc": 15, "l": 5, "f": 9703},
  {"h": "麵", "p": "miàn", "sp": "mian", "r": "麥", "sc": 20, "l": 5, "f": 9704},
  {"h": "鼑", "p": "dǐng", "sp": "ding", "r": "鼎", "sc": 14, "l": 5, "f": 9705},
  {"h": "鼘", "p": "yuān", "sp": "yuan", "r": "鼓", "sc": 22, "l": 5, "f": 9706},
  {"h": "鼮", "p": "tíng", "sp": "ting", "r": "鼠", "sc": 19, "l": 5, "f": 9707},
  {"h": "齲", "p": "qǔ", "sp": "qu", "r": "齒", "sc": 24, "l": 5, "f": 9708},
  {"h": "齼", "p": "chǔ", "sp": "chu", "r": "齒", "sc": 28, "l": 5, "f": 9709},
  {"h": "龝", "p": "qiū", "sp": "qiu", "r": "禾", "sc": 22, "l": 5, "f": 9710},
  {"h": "䦃", "p": "zhuō", "sp": "zhuo", "r": "", "sc": 20, "l": 5, "f": 9711},
  {"h": "䲣", "p": "yú", "sp": "yu", "r": "", "sc": 21, "l": 5, "f": 9712},
  {"h": "䲠", "p": "chūn", "sp": "chun", "r": "", "sc": 20, "l": 5, "f": 9713},
  {"h": "禖", "p": "méi", "sp": "mei", "r": "示", "sc": 13, "l": 5, "f": 9714},
  {"h": "禵", "p": "tí", "sp": "ti", "r": "示", "sc": 22, "l": 5, "f": 9715},
  {"h": "秅", "p": "chá", "sp": "cha", "r": "禾", "sc": 8, "l": 5, "f": 9716},
  {"h": "竌", "p": "chù", "sp": "chu", "r": "立", "sc": 7, "l": 5, "f": 9717},
  {"h": "籙", "p": "lù", "sp": "lu", "r": "竹", "sc": 22, "l": 5, "f": 9718},
  {"h": "粵", "p": "yuè", "sp": "yue", "r": "米", "sc": 13, "l": 5, "f": 9719},
  {"h": "緋", "p": "fēi", "sp": "fei", "r": "糸", "sc": 14, "l": 5, "f": 9720},
  {"h": "羠", "p": "yí", "sp": "yi", "r": "羊", "sc": 12, "l": 5, "f": 9721},
  {"h": "翖", "p": "xì", "sp": "xi", "r": "羽", "sc": 12, "l": 5, "f": 9722},
  {"h": "艣", "p": "lǔ", "sp": "lu", "r": "舟", "sc": 19, "l": 5, "f": 9723},
  {"h": "蚡", "p": "fén", "sp": "fen", "r": "虫", "sc": 10, "l": 5, "f": 9724},
  {"h": "訹", "p": "xù", "sp": "xu", "r": "言", "sc": 12, "l": 5, "f": 9725},
  {"h": "諠", "p": "xuān", "sp": "xuan", "r": "言", "sc": 16, "l": 5, "f": 9726},
  {"h": "輂", "p": "jú", "sp": "ju", "r": "車", "sc": 13, "l": 5, "f": 9727},
  {"h": "輨", "p": "guǎn", "sp": "guan", "r": "車", "sc": 15, "l": 5, "f": 9728},
  {"h": "遶", "p": "rào", "sp": "rao", "r": "辵", "sc": 15, "l": 5, "f": 9729},
  {"h": "鈆", "p": "qiān", "sp": "qian", "r": "金", "sc": 12, "l": 5, "f": 9730},
  {"h": "鈖", "p": "fēn", "sp": "fen", "r": "金", "sc": 12, "l": 5, "f": 9731},
  {"h": "銅", "p": "tóng", "sp": "tong", "r": "金", "sc": 14, "l": 5, "f": 9732},
  {"h": "鍥", "p": "qiè", "sp": "qie", "r": "金", "sc": 17, "l": 5, "f": 9733},
  {"h": "鎇", "p": "méi", "sp": "mei", "r": "金", "sc": 17, "l": 5, "f": 9734},
  {"h": "绗", "p": "háng", "sp": "hang", "r": "糸", "sc": 9, "l": 5, "f": 9735},
  {"h": "鐷", "p": "yè", "sp": "ye", "r": "金", "sc": 20, "l": 5, "f": 9736},
  {"h": "閽", "p": "hūn", "sp": "hun", "r": "門", "sc": 16, "l": 5, "f": 9737},
  {"h": "闟", "p": "xì", "sp": "xi", "r": "門", "sc": 20, "l": 5, "f": 9738},
  {"h": "韮", "p": "jiǔ", "sp": "jiu", "r": "韭", "sc": 12, "l": 5, "f": 9739},
  {"h": "騣", "p": "zōng", "sp": "zong", "r": "馬", "sc": 19, "l": 5, "f": 9740},
  {"h": "髧", "p": "dàn", "sp": "dan", "r": "髟", "sc": 14, "l": 5, "f": 9741},
  {"h": "鴢", "p": "yǎo", "sp": "yao", "r": "鳥", "sc": 16, "l": 5, "f": 9742},
  {"h": "鷘", "p": "chì", "sp": "chi", "r": "鳥", "sc": 22, "l": 5, "f": 9743},
  {"h": "黺", "p": "fěn", "sp": "fen", "r": "黹", "sc": 16, "l": 5, "f": 9744},
  {"h": "祶", "p": "dì", "sp": "di", "r": "示", "sc": 11, "l": 5, "f": 9745},
  {"h": "荮", "p": "zhòu", "sp": "zhou", "r": "艸", "sc": 9, "l": 5, "f": 9746},
  {
    "h": "瞏",
    "p": "qióng",
    "sp": "qiong",
    "r": "目",
    "sc": 15,
    "l": 5,
    "f": 9747
  },
  {"h": "稴", "p": "xián", "sp": "xian", "r": "禾", "sc": 15, "l": 5, "f": 9748},
  {"h": "笲", "p": "fán", "sp": "fan", "r": "竹", "sc": 11, "l": 5, "f": 9749},
  {"h": "笿", "p": "luò", "sp": "luo", "r": "竹", "sc": 12, "l": 5, "f": 9750},
  {"h": "縪", "p": "bì", "sp": "bi", "r": "糸", "sc": 16, "l": 5, "f": 9751},
  {"h": "纍", "p": "léi", "sp": "lei", "r": "糸", "sc": 21, "l": 5, "f": 9752},
  {"h": "翸", "p": "pěn", "sp": "pen", "r": "羽", "sc": 18, "l": 5, "f": 9753},
  {"h": "肧", "p": "pēi", "sp": "pei", "r": "肉", "sc": 8, "l": 5, "f": 9754},
  {"h": "臶", "p": "jiàn", "sp": "jian", "r": "至", "sc": 12, "l": 5, "f": 9755},
  {"h": "蜸", "p": "qiǎn", "sp": "qian", "r": "虫", "sc": 14, "l": 5, "f": 9756},
  {"h": "豊", "p": "lǐ", "sp": "li", "r": "豆", "sc": 13, "l": 5, "f": 9757},
  {"h": "賁", "p": "bì", "sp": "bi", "r": "貝", "sc": 12, "l": 5, "f": 9758},
  {"h": "輐", "p": "wàn", "sp": "wan", "r": "車", "sc": 14, "l": 5, "f": 9759},
  {"h": "輘", "p": "léng", "sp": "leng", "r": "車", "sc": 15, "l": 5, "f": 9760},
  {"h": "邏", "p": "luó", "sp": "luo", "r": "辵", "sc": 22, "l": 5, "f": 9761},
  {"h": "雘", "p": "wò", "sp": "wo", "r": "隹", "sc": 17, "l": 5, "f": 9762},
  {"h": "鯽", "p": "zéi", "sp": "zei", "r": "魚", "sc": 18, "l": 5, "f": 9763},
  {
    "h": "䓖",
    "p": "qióng",
    "sp": "qiong",
    "r": "艸",
    "sc": 13,
    "l": 5,
    "f": 9764
  },
  {"h": "眞", "p": "zhēn", "sp": "zhen", "r": "目", "sc": 10, "l": 5, "f": 9765},
  {"h": "睴", "p": "gùn", "sp": "gun", "r": "目", "sc": 14, "l": 5, "f": 9766},
  {"h": "瞡", "p": "guǐ", "sp": "gui", "r": "目", "sc": 16, "l": 5, "f": 9767},
  {"h": "砏", "p": "bīn", "sp": "bin", "r": "石", "sc": 9, "l": 5, "f": 9768},
  {"h": "磯", "p": "jī", "sp": "ji", "r": "石", "sc": 17, "l": 5, "f": 9769},
  {"h": "粿", "p": "guǒ", "sp": "guo", "r": "米", "sc": 14, "l": 5, "f": 9770},
  {"h": "紘", "p": "hóng", "sp": "hong", "r": "糸", "sc": 10, "l": 5, "f": 9771},
  {"h": "舋", "p": "xìn", "sp": "xin", "r": "臼", "sc": 20, "l": 5, "f": 9772},
  {"h": "葋", "p": "qú", "sp": "qu", "r": "艸", "sc": 12, "l": 5, "f": 9773},
  {"h": "薭", "p": "bai", "sp": "bai", "r": "艸", "sc": 16, "l": 5, "f": 9774},
  {"h": "虲", "p": "xiā", "sp": "xia", "r": "虫", "sc": 8, "l": 5, "f": 9775},
  {"h": "蠩", "p": "zhū", "sp": "zhu", "r": "虫", "sc": 21, "l": 5, "f": 9776},
  {"h": "覵", "p": "jiàn", "sp": "jian", "r": "見", "sc": 19, "l": 5, "f": 9777},
  {"h": "註", "p": "zhù", "sp": "zhu", "r": "言", "sc": 12, "l": 5, "f": 9778},
  {"h": "詄", "p": "dié", "sp": "die", "r": "言", "sc": 12, "l": 5, "f": 9779},
  {"h": "譗", "p": "zhá", "sp": "zha", "r": "言", "sc": 19, "l": 5, "f": 9780},
  {"h": "趰", "p": "ěr", "sp": "er", "r": "走", "sc": 21, "l": 5, "f": 9781},
  {"h": "軧", "p": "dǐ", "sp": "di", "r": "車", "sc": 12, "l": 5, "f": 9782},
  {"h": "岍", "p": "qiān", "sp": "qian", "r": "山", "sc": 7, "l": 5, "f": 9783},
  {"h": "鉼", "p": "bǐng", "sp": "bing", "r": "金", "sc": 14, "l": 5, "f": 9784},
  {"h": "鋪", "p": "pù", "sp": "pu", "r": "金", "sc": 15, "l": 5, "f": 9785},
  {"h": "鞻", "p": "lóu", "sp": "lou", "r": "革", "sc": 20, "l": 5, "f": 9786},
  {"h": "颵", "p": "shāo", "sp": "shao", "r": "風", "sc": 16, "l": 5, "f": 9787},
  {"h": "餼", "p": "xì", "sp": "xi", "r": "食", "sc": 18, "l": 5, "f": 9788},
  {"h": "騍", "p": "kè", "sp": "ke", "r": "馬", "sc": 18, "l": 5, "f": 9789},
  {"h": "驂", "p": "cān", "sp": "can", "r": "馬", "sc": 21, "l": 5, "f": 9790},
  {"h": "鯘", "p": "něi", "sp": "nei", "r": "魚", "sc": 19, "l": 5, "f": 9791},
  {"h": "鳿", "p": "yù", "sp": "yu", "r": "鳥", "sc": 15, "l": 5, "f": 9792},
  {"h": "龎", "p": "páng", "sp": "pang", "r": "龍", "sc": 18, "l": 5, "f": 9793},
  {"h": "睉", "p": "cuó", "sp": "cuo", "r": "目", "sc": 12, "l": 5, "f": 9794},
  {"h": "砤", "p": "tuó", "sp": "tuo", "r": "石", "sc": 10, "l": 5, "f": 9795},
  {"h": "礑", "p": "dàng", "sp": "dang", "r": "石", "sc": 18, "l": 5, "f": 9796},
  {"h": "禰", "p": "mí", "sp": "mi", "r": "示", "sc": 18, "l": 5, "f": 9797},
  {"h": "窤", "p": "guān", "sp": "guan", "r": "穴", "sc": 13, "l": 5, "f": 9798},
  {"h": "竈", "p": "zào", "sp": "zao", "r": "穴", "sc": 21, "l": 5, "f": 9799},
  {"h": "聶", "p": "niè", "sp": "nie", "r": "耳", "sc": 18, "l": 5, "f": 9800},
  {"h": "蕒", "p": "mǎi", "sp": "mai", "r": "艸", "sc": 15, "l": 5, "f": 9801},
  {"h": "説", "p": "shuō", "sp": "shuo", "r": "言", "sc": 14, "l": 5, "f": 9802},
  {"h": "趌", "p": "jí", "sp": "ji", "r": "走", "sc": 13, "l": 5, "f": 9803},
  {"h": "踄", "p": "bù", "sp": "bu", "r": "足", "sc": 14, "l": 5, "f": 9804},
  {"h": "鑰", "p": "yào", "sp": "yao", "r": "金", "sc": 25, "l": 5, "f": 9805},
  {"h": "隣", "p": "lín", "sp": "lin", "r": "阜", "sc": 14, "l": 5, "f": 9806},
  {
    "h": "饌",
    "p": "zhuàn",
    "sp": "zhuan",
    "r": "食",
    "sc": 20,
    "l": 5,
    "f": 9807
  },
  {"h": "驛", "p": "yì", "sp": "yi", "r": "馬", "sc": 23, "l": 5, "f": 9808},
  {"h": "骿", "p": "pián", "sp": "pian", "r": "骨", "sc": 15, "l": 5, "f": 9809},
  {"h": "鼥", "p": "bá", "sp": "ba", "r": "鼠", "sc": 18, "l": 5, "f": 9810},
  {"h": "碵", "p": "tian", "sp": "tian", "r": "石", "sc": 14, "l": 5, "f": 9811},
  {"h": "竤", "p": "hóng", "sp": "hong", "r": "立", "sc": 12, "l": 5, "f": 9812},
  {"h": "絒", "p": "chóu", "sp": "chou", "r": "糸", "sc": 12, "l": 5, "f": 9813},
  {"h": "絓", "p": "guà", "sp": "gua", "r": "糸", "sc": 12, "l": 5, "f": 9814},
  {"h": "芼", "p": "mào", "sp": "mao", "r": "艸", "sc": 7, "l": 5, "f": 9815},
  {"h": "萢", "p": "pao", "sp": "pao", "r": "艸", "sc": 11, "l": 5, "f": 9816},
  {"h": "蓧", "p": "diào", "sp": "diao", "r": "艸", "sc": 13, "l": 5, "f": 9817},
  {"h": "蔄", "p": "màn", "sp": "man", "r": "艸", "sc": 14, "l": 5, "f": 9818},
  {"h": "螦", "p": "sao", "sp": "sao", "r": "虫", "sc": 16, "l": 5, "f": 9819},
  {"h": "詀", "p": "zhān", "sp": "zhan", "r": "言", "sc": 12, "l": 5, "f": 9820},
  {"h": "躦", "p": "cuó", "sp": "cuo", "r": "足", "sc": 26, "l": 5, "f": 9821},
  {"h": "狨", "p": "róng", "sp": "rong", "r": "犬", "sc": 9, "l": 5, "f": 9822},
  {"h": "錋", "p": "péng", "sp": "peng", "r": "金", "sc": 16, "l": 5, "f": 9823},
  {"h": "鎝", "p": "dā", "sp": "da", "r": "金", "sc": 17, "l": 5, "f": 9824},
  {"h": "鐩", "p": "suì", "sp": "sui", "r": "金", "sc": 20, "l": 5, "f": 9825},
  {"h": "隲", "p": "é", "sp": "e", "r": "阜", "sc": 16, "l": 5, "f": 9826},
  {"h": "霝", "p": "líng", "sp": "ling", "r": "雨", "sc": 17, "l": 5, "f": 9827},
  {"h": "鞟", "p": "kuò", "sp": "kuo", "r": "革", "sc": 17, "l": 5, "f": 9828},
  {"h": "礐", "p": "què", "sp": "que", "r": "石", "sc": 18, "l": 5, "f": 9829},
  {"h": "禗", "p": "sī", "sp": "si", "r": "示", "sc": 13, "l": 5, "f": 9830},
  {"h": "稡", "p": "zuì", "sp": "zui", "r": "禾", "sc": 13, "l": 5, "f": 9831},
  {"h": "穽", "p": "jǐng", "sp": "jing", "r": "穴", "sc": 9, "l": 5, "f": 9832},
  {"h": "糮", "p": "xiàn", "sp": "xian", "r": "米", "sc": 20, "l": 5, "f": 9833},
  {"h": "羮", "p": "gēng", "sp": "geng", "r": "羊", "sc": 15, "l": 5, "f": 9834},
  {"h": "肰", "p": "rán", "sp": "ran", "r": "肉", "sc": 8, "l": 5, "f": 9835},
  {"h": "艸", "p": "cǎo", "sp": "cao", "r": "艸", "sc": 6, "l": 5, "f": 9836},
  {"h": "葇", "p": "róu", "sp": "rou", "r": "艸", "sc": 12, "l": 5, "f": 9837},
  {
    "h": "葟",
    "p": "huáng",
    "sp": "huang",
    "r": "艸",
    "sc": 12,
    "l": 5,
    "f": 9838
  },
  {"h": "蒪", "p": "pò", "sp": "po", "r": "艸", "sc": 13, "l": 5, "f": 9839},
  {"h": "蝑", "p": "xū", "sp": "xu", "r": "虫", "sc": 15, "l": 5, "f": 9840},
  {"h": "螧", "p": "qi", "sp": "qi", "r": "虫", "sc": 16, "l": 5, "f": 9841},
  {"h": "蟅", "p": "zhè", "sp": "zhe", "r": "虫", "sc": 17, "l": 5, "f": 9842},
  {"h": "蟢", "p": "xǐ", "sp": "xi", "r": "虫", "sc": 18, "l": 5, "f": 9843},
  {"h": "貒", "p": "tuān", "sp": "tuan", "r": "豸", "sc": 16, "l": 5, "f": 9844},
  {"h": "銫", "p": "sè", "sp": "se", "r": "金", "sc": 14, "l": 5, "f": 9845},
  {"h": "関", "p": "guān", "sp": "guan", "r": "門", "sc": 14, "l": 5, "f": 9846},
  {"h": "礤", "p": "cǎ", "sp": "ca", "r": "石", "sc": 19, "l": 5, "f": 9847},
  {"h": "鮭", "p": "guī", "sp": "gui", "r": "魚", "sc": 17, "l": 5, "f": 9848},
  {"h": "鴙", "p": "zhì", "sp": "zhi", "r": "鳥", "sc": 16, "l": 5, "f": 9849},
  {"h": "鴞", "p": "xiāo", "sp": "xiao", "r": "鳥", "sc": 16, "l": 5, "f": 9850},
  {"h": "鵣", "p": "lài", "sp": "lai", "r": "鳥", "sc": 18, "l": 5, "f": 9851},
  {"h": "鵪", "p": "ān", "sp": "an", "r": "鳥", "sc": 19, "l": 5, "f": 9852},
  {"h": "鷦", "p": "jiāo", "sp": "jiao", "r": "鳥", "sc": 23, "l": 5, "f": 9853},
  {"h": "鷺", "p": "lù", "sp": "lu", "r": "鳥", "sc": 24, "l": 5, "f": 9854},
  {"h": "齜", "p": "chái", "sp": "chai", "r": "齒", "sc": 21, "l": 5, "f": 9855},
  {"h": "盢", "p": "xù", "sp": "xu", "r": "皿", "sc": 14, "l": 5, "f": 9856},
  {"h": "眘", "p": "shèn", "sp": "shen", "r": "目", "sc": 10, "l": 5, "f": 9857},
  {"h": "瞞", "p": "mán", "sp": "man", "r": "目", "sc": 16, "l": 5, "f": 9858},
  {"h": "礿", "p": "yuè", "sp": "yue", "r": "示", "sc": 7, "l": 5, "f": 9859},
  {"h": "祩", "p": "zhù", "sp": "zhu", "r": "示", "sc": 10, "l": 5, "f": 9860},
  {"h": "禨", "p": "jī", "sp": "ji", "r": "示", "sc": 16, "l": 5, "f": 9861},
  {"h": "秺", "p": "dù", "sp": "du", "r": "禾", "sc": 11, "l": 5, "f": 9862},
  {"h": "穟", "p": "suì", "sp": "sui", "r": "禾", "sc": 17, "l": 5, "f": 9863},
  {"h": "窫", "p": "yà", "sp": "ya", "r": "穴", "sc": 14, "l": 5, "f": 9864},
  {"h": "笎", "p": "yuán", "sp": "yuan", "r": "竹", "sc": 10, "l": 5, "f": 9865},
  {"h": "篨", "p": "chú", "sp": "chu", "r": "竹", "sc": 15, "l": 5, "f": 9866},
  {"h": "籋", "p": "niè", "sp": "nie", "r": "竹", "sc": 20, "l": 5, "f": 9867},
  {"h": "緤", "p": "xiè", "sp": "xie", "r": "糸", "sc": 15, "l": 5, "f": 9868},
  {"h": "緵", "p": "zōng", "sp": "zong", "r": "糸", "sc": 15, "l": 5, "f": 9869},
  {"h": "縚", "p": "tāo", "sp": "tao", "r": "糸", "sc": 16, "l": 5, "f": 9870},
  {"h": "繿", "p": "lán", "sp": "lan", "r": "糸", "sc": 20, "l": 5, "f": 9871},
  {"h": "羙", "p": "gāo", "sp": "gao", "r": "羊", "sc": 10, "l": 5, "f": 9872},
  {"h": "蘂", "p": "ruǐ", "sp": "rui", "r": "艸", "sc": 19, "l": 5, "f": 9873},
  {"h": "蜼", "p": "wèi", "sp": "wei", "r": "虫", "sc": 14, "l": 5, "f": 9874},
  {"h": "襘", "p": "guì", "sp": "gui", "r": "衣", "sc": 18, "l": 5, "f": 9875},
  {"h": "訞", "p": "yāo", "sp": "yao", "r": "言", "sc": 11, "l": 5, "f": 9876},
  {"h": "謄", "p": "téng", "sp": "teng", "r": "言", "sc": 17, "l": 5, "f": 9877},
  {"h": "谼", "p": "hóng", "sp": "hong", "r": "谷", "sc": 13, "l": 5, "f": 9878},
  {"h": "躒", "p": "lì", "sp": "li", "r": "足", "sc": 22, "l": 5, "f": 9879},
  {"h": "躭", "p": "dān", "sp": "dan", "r": "身", "sc": 11, "l": 5, "f": 9880},
  {"h": "迍", "p": "zhūn", "sp": "zhun", "r": "辵", "sc": 7, "l": 5, "f": 9881},
  {"h": "釿", "p": "jīn", "sp": "jin", "r": "金", "sc": 12, "l": 5, "f": 9882},
  {"h": "鏸", "p": "huì", "sp": "hui", "r": "金", "sc": 20, "l": 5, "f": 9883},
  {"h": "颻", "p": "yáo", "sp": "yao", "r": "風", "sc": 19, "l": 5, "f": 9884},
  {"h": "餟", "p": "zhuì", "sp": "zhui", "r": "食", "sc": 16, "l": 5, "f": 9885},
  {"h": "餫", "p": "yùn", "sp": "yun", "r": "食", "sc": 17, "l": 5, "f": 9886},
  {"h": "餽", "p": "kuì", "sp": "kui", "r": "食", "sc": 17, "l": 5, "f": 9887},
  {"h": "驆", "p": "bì", "sp": "bi", "r": "馬", "sc": 20, "l": 5, "f": 9888},
  {"h": "鬳", "p": "yàn", "sp": "yan", "r": "鬲", "sc": 16, "l": 5, "f": 9889},
  {"h": "魒", "p": "piāo", "sp": "piao", "r": "鬼", "sc": 20, "l": 5, "f": 9890},
  {"h": "盉", "p": "hé", "sp": "he", "r": "皿", "sc": 10, "l": 5, "f": 9891},
  {"h": "睱", "p": "xià", "sp": "xia", "r": "目", "sc": 14, "l": 5, "f": 9892},
  {"h": "睻", "p": "xuān", "sp": "xuan", "r": "目", "sc": 14, "l": 5, "f": 9893},
  {"h": "磠", "p": "lǔ", "sp": "lu", "r": "石", "sc": 16, "l": 5, "f": 9894},
  {"h": "箋", "p": "jiān", "sp": "jian", "r": "竹", "sc": 14, "l": 5, "f": 9895},
  {"h": "絣", "p": "bēng", "sp": "beng", "r": "糸", "sc": 12, "l": 5, "f": 9896},
  {"h": "縀", "p": "xiá", "sp": "xia", "r": "糸", "sc": 15, "l": 5, "f": 9897},
  {"h": "縛", "p": "fù", "sp": "fu", "r": "糸", "sc": 16, "l": 5, "f": 9898},
  {"h": "縧", "p": "tāo", "sp": "tao", "r": "糸", "sc": 16, "l": 5, "f": 9899},
  {"h": "荙", "p": "dá", "sp": "da", "r": "艸", "sc": 9, "l": 5, "f": 9900},
  {"h": "莕", "p": "xìng", "sp": "xing", "r": "艸", "sc": 10, "l": 5, "f": 9901},
  {"h": "莙", "p": "jūn", "sp": "jun", "r": "艸", "sc": 10, "l": 5, "f": 9902},
  {"h": "螠", "p": "yì", "sp": "yi", "r": "虫", "sc": 16, "l": 5, "f": 9903},
  {"h": "蟧", "p": "láo", "sp": "lao", "r": "虫", "sc": 18, "l": 5, "f": 9904},
  {"h": "襌", "p": "dān", "sp": "dan", "r": "衣", "sc": 17, "l": 5, "f": 9905},
  {"h": "襽", "p": "lan", "sp": "lan", "r": "衣", "sc": 25, "l": 5, "f": 9906},
  {"h": "豰", "p": "bó", "sp": "bo", "r": "豕", "sc": 17, "l": 5, "f": 9907},
  {"h": "賷", "p": "jī", "sp": "ji", "r": "貝", "sc": 17, "l": 5, "f": 9908},
  {"h": "鉽", "p": "shì", "sp": "shi", "r": "金", "sc": 14, "l": 5, "f": 9909},
  {"h": "銲", "p": "hàn", "sp": "han", "r": "金", "sc": 15, "l": 5, "f": 9910},
  {"h": "錧", "p": "guǎn", "sp": "guan", "r": "金", "sc": 16, "l": 5, "f": 9911},
  {"h": "鍚", "p": "yáng", "sp": "yang", "r": "金", "sc": 17, "l": 5, "f": 9912},
  {"h": "鍱", "p": "yè", "sp": "ye", "r": "金", "sc": 17, "l": 5, "f": 9913},
  {"h": "骣", "p": "chǎn", "sp": "chan", "r": "馬", "sc": 15, "l": 5, "f": 9914},
  {"h": "鑪", "p": "lú", "sp": "lu", "r": "金", "sc": 24, "l": 5, "f": 9915},
  {"h": "靷", "p": "yǐn", "sp": "yin", "r": "革", "sc": 13, "l": 5, "f": 9916},
  {"h": "頦", "p": "hái", "sp": "hai", "r": "頁", "sc": 15, "l": 5, "f": 9917},
  {"h": "騊", "p": "táo", "sp": "tao", "r": "馬", "sc": 18, "l": 5, "f": 9918},
  {"h": "骔", "p": "zōng", "sp": "zong", "r": "馬", "sc": 11, "l": 5, "f": 9919},
  {"h": "魽", "p": "hán", "sp": "han", "r": "魚", "sc": 16, "l": 5, "f": 9920},
  {"h": "魿", "p": "líng", "sp": "ling", "r": "魚", "sc": 16, "l": 5, "f": 9921},
  {"h": "鮅", "p": "bì", "sp": "bi", "r": "魚", "sc": 16, "l": 5, "f": 9922},
  {"h": "鮇", "p": "wèi", "sp": "wei", "r": "魚", "sc": 16, "l": 5, "f": 9923},
  {"h": "鮎", "p": "nián", "sp": "nian", "r": "魚", "sc": 16, "l": 5, "f": 9924},
  {"h": "鮘", "p": "dai", "sp": "dai", "r": "魚", "sc": 16, "l": 5, "f": 9925},
  {"h": "鮪", "p": "wěi", "sp": "wei", "r": "魚", "sc": 17, "l": 5, "f": 9926},
  {"h": "鯨", "p": "jīng", "sp": "jing", "r": "魚", "sc": 19, "l": 5, "f": 9927},
  {"h": "鰿", "p": "jì", "sp": "ji", "r": "魚", "sc": 22, "l": 5, "f": 9928},
  {"h": "鱀", "p": "jì", "sp": "ji", "r": "魚", "sc": 20, "l": 5, "f": 9929},
  {
    "h": "鱄",
    "p": "zhuān",
    "sp": "zhuan",
    "r": "魚",
    "sc": 22,
    "l": 5,
    "f": 9930
  },
  {"h": "鱓", "p": "shàn", "sp": "shan", "r": "魚", "sc": 23, "l": 5, "f": 9931},
  {"h": "鲖", "p": "tóng", "sp": "tong", "r": "魚", "sc": 14, "l": 5, "f": 9932},
  {"h": "鴒", "p": "líng", "sp": "ling", "r": "鳥", "sc": 16, "l": 5, "f": 9933},
];
